/*****************************************************************************
******************************************************************************
*
*               Camera object
*
*               $Revision: 1.1.2.1 $
*               $Date: 2009/01/22 15:56:03 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright 1998-2009 HUMUSOFT s.r.o.

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#ifndef CAMERA_H
#define CAMERA_H

#include <math.h>
#include "Math3d.h"

/**
 * This object represents the camera in a virtual world.
 */
class Camera
{
private:

  /** The position. */
  Point3 pos;

  /** The direction vector. */
  Vector3 dir;

  /** The up vector. */
  Vector3 up;

  /** The zoom factor. */
  double zoom;

public:

  /**
   * Default constructor.
   * The camera position is initialized to (0, 0, 0),
   * the direction vector to (0, 0, -1),
   * the up vector to (0, 1, 0),
   * the zoom factor to 1.
   */
  Camera() : pos(0, 0, 0), dir(0, 0, -1), up(0, 1, 0), zoom(1) { }

  /**
   * Constructor.
   * @param position The camera position.
   * @param direction The direction vector (where the camera looks).
   * @param upvec The up vector.
   * @param z The zoom factor.
   */
  Camera(const Point3 &position,
         const Vector3 &direction,
         const Vector3 &upvec,
         double z)
    : pos(position), dir(direction), up(upvec), zoom(z) { }

  /**
   * Returns the camera position.
   * @return The camera position.
   */
  const Point3& getPosition() const { return pos; };

  /**
   * Sets the camera position.
   * @param position The new position.
   */
  void setPosition(const Point3 &position) { pos = position; }

  /**
   * Returns the camera direction.
   * @return The camera direction.
   */
  const Vector3& getDirection() const { return dir; };

  /**
   * Sets the camera direction.
   * @param direction The new direction.
   */
  void setDirection(const Vector3 &direction) { dir = direction; }

  /**
   * Returns the up vector of the camera.
   * @return The up vector of the camera.
   */
  const Vector3& getUpVector() const { return up; }

  /**
   * Sets the up vector of the camera.
   * @param direction The new up vector.
   */
  void setUpVector(const Vector3 &upvec) { up = upvec; }

  /**
   * Returns the zoom factor of the camera.
   * @return The zoom factor.
   */
  double getZoom() const { return zoom; }

  /**
   * Sets the zoom factor of the camera.
   * @param z The new zoom factor (the value is clamped
   * to range 0.001 .. 1000).
   */
  void setZoom(double z)
    { zoom = (z > 0.001) ? ((z < 1000) ? z : 1000) : 0.001; }

  /**
   * Returns the axis of rotation between the current camera direction
   * and the default direction (0, 0, -1).
   * @return The rotation axis.
   */
  Vector3 getYawAxis() const
    {
      Vector3 result = Vector3(0, 0, -1).cross(dir);
      if (result.length() < 1e-6) result = Vector3(0, 1, 0);
      return result;
    }

  /**
   * Returns the angle between the current camera direction
   * and the default direction (0, 0, -1).
   * @return The angle.
   */
  double getYawAngle() const
  {
    return acos(std::max(std::min(Vector3(0, 0, -1).dot(dir.normalize()), 1.), -1.));
  }

  /**
   * Computes the viewing matrix representing this camera view.
   * @return The resulting matrix.
   */
  Matrix4x4 computeViewingMatrix() const
    { Matrix4x4 m; m.setToViewing(pos, dir, up); return m; }

  /**
   * Resets the camera to defaults.
   */
  void reset()
    {
      pos = Point3(0, 0, 0);
      dir = Vector3(0, 0, -1);
      up = Vector3(0, 1, 0);
      zoom = 1;
    }

  /**
   * Transforms the camera by the given matrix.
   * @param m The matrix.
   * @return A reference to this object, so that calls can be chained.
   */
  Camera &transform(const Matrix4x4 &m)
    { pos = m*pos; dir = m*dir; up = m*up; return *this; }

  /**
   * Moves the camera forwards or backwards by the given distance.
   * @param distance The distance (negative to step backwards).
   * @return A reference to this object, so that calls can be chained.
   */
  Camera &step(double distance)
    { pos += dir.scale(distance); return *this; }

  /**
   * Slides the camera horizontally by the given distance.
   * @param distance The distance (positive to move to the right,
   * negative to move to the left).
   * @return A reference to this object, so that calls can be chained.
   */
  Camera &slideH(double distance)
    { pos += dir.cross(up).scale(distance); return *this; }

  /**
   * Slides the camera vertically by the given distance.
   * @param distance The distance (positive to slide upwards,
   * negative to slide downwards).
   * @return A reference to this object, so that calls can be chained.
   */
  Camera &slideV(double distance)
    { pos += up.scale(distance); return *this; }

  /**
   * Rotates the camera in the horizontal plane (to the left
   * or to the right) by the given angle.
   * @param angle The angle.
   * @return A reference to this object, so that calls can be chained.
   */
  Camera &rotateH(double angle)
    { Matrix4x4 m; m.setToRotation(angle, up); dir = m*dir; return *this; }

  /**
   * Rotates the camera in the vertical plane
   * (up or down) by the given angle.
   * @param angle The angle.
   * @return A reference to this object, so that calls can be chained.
   */
  Camera &rotateV(double angle)
    { Matrix4x4 m; m.setToRotation(angle, dir.cross(up)); dir = m*dir; up = m*up; return *this; }

  /**
   * Rotates the camera in the screen plane by the given angle.
   * @param angle The angle.
   * @return A reference to this object, so that calls can be chained.
   */
  Camera &rotateS(double angle)
    { Matrix4x4 m; m.setToRotation(angle, dir); up = m*up; return *this; }

  /**
   * Rotates the camera around the given center in the horizontal plane
   * by the given angle.
   * @param center The center point.
   * @param angle The angle.
   * @return A reference to this object, so that calls can be chained.
   */
  Camera &orbitH(const Point3 &center, double angle)
    {
      // we first move the camera to the center of rotation, then rotate it
      // horizontally as needed, and then move it back by the same distance
      Vector3 radius = center - pos;
      Matrix4x4 m; m.setToTranslation(radius);
      Matrix4x4 temp; temp.setToRotation(angle, up);
      m = m*temp; temp.setToTranslation(-radius); m = m*temp;
      pos = center - m*radius; dir = m*dir; up = m*up;
      return *this;
    }

  /**
   * Rotates the camera around the given center in the vertical plane
   * by the given angle.
   * @param center The center point.
   * @param angle The angle.
   * @return A reference to this object, so that calls can be chained.
   */
  Camera &orbitV(const Point3 &center, double angle)
    {
      // we first move the camera to the center of rotation, then rotate it
      // vertically as needed, and then move it back by the same distance
      Vector3 radius = center - pos;
      Matrix4x4 m; m.setToTranslation(radius);
      Matrix4x4 temp; temp.setToRotation(angle, dir.cross(up));
      m = m*temp; temp.setToTranslation(-radius); m = m*temp;
      pos = center - m*radius; dir = m*dir; up = m*up;
      return *this;
    }

  /**
   * Makes the camera stand straight on the horizontal plane.
   * @return A reference to this object, so that calls can be chained.
   */
  Camera &straightenUp()
    {
      up = Vector3(0, 1, 0);
      dir.setY(0);
      return *this;
    }
};

#endif
