/*****************************************************************************
******************************************************************************
*
*               Document class for I/O operations
*
*               $Revision: 1.1.2.10 $
*               $Date: 2011/06/01 17:24:47 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright (C) 1998  Chris Morley
// Modified by HUMUSOFT s.r.o.  Copyright 1998-2011 HUMUSOFT s.r.o.

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "Doc.h"

#include <string>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>


#define _CHARBUF_LEN 1024


Doc::Doc(const char* url, const Doc* relative) :
  d_url(0),
  d_tmpfile(0),
  d_fp(0)
{
  if (url != 0)
    this->seturl(url, relative);
}

Doc::Doc(const Doc* doc) :
  d_url(0),
  d_tmpfile(0),
  d_fp(0)
{
  if (doc)
    this->seturl(doc->url());
}

Doc::Doc(const Doc& other) :
  d_url(0),
  d_tmpfile(0),
  d_fp(0)
{
  *this = other;
}

Doc::~Doc()
{
  delete [] d_url;
  if (d_tmpfile)
  {
    theSystem->removeFile(d_tmpfile);
    delete [] d_tmpfile;
  }
}

Doc& Doc::operator=(const Doc& other)
{
  delete [] d_url;
  d_url = 0;
  int len = strlen(other.d_url);
  if (len>0)
  {
    d_url = new char[len+1];
    strcpy(d_url,other.d_url);
  }

  delete [] d_tmpfile;
  d_tmpfile = 0;
  len = strlen(other.d_tmpfile);
  if (len>0)
  {
    d_tmpfile = new char[len+1];
    strcpy(d_tmpfile,other.d_tmpfile);
  }

  delete d_fp;
  d_fp = 0;

  return *this;
}

const char* Doc::stripProtocol(const char* url)
{
  const char *s = url;

#ifdef _WIN32
  if (strncmp(s+1,":/",2) == 0) return url;
#endif

  // strip off protocol if any
  while (*s && isalpha(*s)) ++s;

  if (*s == COLON)
    return s + 1;

  return url;
}


bool Doc::isAbsolute(const char* url)
{
  const char *s = stripProtocol(url);
  const std::string& libraryToken(theSystem->getLibraryToken());

  // libraryToken is shortcut for sl3d library which expands to full path
  return (*s == SLASH || *(s+1) == ':' || strncmp(s, libraryToken.c_str(), libraryToken.size())==0);
}


char const* Doc::localName()
{
  static char buf[_CHARBUF_LEN];
  if (filename(buf, sizeof(buf)))
    return buf;
  return 0;
}


char const* Doc::localPath()
{
  static char buf[_CHARBUF_LEN];
  if (filename(buf, sizeof(buf)))
  {
    char *s = strrchr(buf, SLASH);
    if (s)
      *(s+1) = '\0';
    else
      buf[0] = '\0';

    return(buf);
  }
  return 0;
}


// Converts a url into a local filename
bool Doc::filename(char* fn, int nfn)
{
  const std::string& libraryToken(theSystem->getLibraryToken());

  fn[0] = '\0';

  char* e = 0;
  char* s = (char*) stripProtocol(d_url);      // fix for const correctness on Sun

  if ((e = strrchr(s,'#')) != 0)
    *e = '\0';

  const char* protocol = urlProtocol();

  // Get a local copy of http files
  if (strcmp(protocol, "http") == 0)
  {
    if (d_tmpfile)// Already fetched it
    {
      s = d_tmpfile;
    }
    else if ((s = (char *) theSystem->httpFetch(d_url)))
    {
      d_tmpfile = new char[strlen(s)+1];
      strcpy(d_tmpfile, s);
      s = d_tmpfile;
    }
  }

  // Unrecognized protocol (need ftp here...)
  else if (strcmp(protocol, "file") != 0)
    s = 0;

#ifdef _WIN32
  // Does not like "//C:" skip "// "
  if (s && (strlen(s)>4 && s[0] == '/' && s[1] == '/' && s[3] == ':')) s=s+2;
#endif

  if (s)
  {
    if (strncmp(s, libraryToken.c_str(), sizeof(libraryToken.c_str())-1)==0)
    {
      std::string path(s);
      // libraryToken is replaced by real path to library
      path.replace(0, libraryToken.size(), theSystem->getLibDir().c_str());

      strncpy(fn, path.c_str(), path.size());
      fn[path.size()] = '\0';
      s = fn;
    }
    else
    {
      strncpy( fn, s, nfn-1 );
      fn[nfn-1] = '\0';
    }
  }

  if (e)
    *e = '#';

  return s && *s;
}


char const* Doc::urlProtocol() const
{
  if (d_url)
  {
    static char protocol[12];
    const char *s = d_url;

#ifdef _WIN32
    if (strncmp(s+1,":/",2) == 0) return "file";
#endif

    for (unsigned int i=0; i<sizeof(protocol); ++i, ++s)
    {
      if (*s == 0 || ! isalpha(*s))
      {
        protocol[i] = '\0';
        break;
      }
      protocol[i] = tolower(*s);
    }

    protocol[sizeof(protocol)-1] = '\0';
    if (*s == COLON)
      return protocol;
  }
  return "file";
}


void Doc::seturl(const char* url, const Doc* relative)
{
  // clear old URL
  delete[] d_url;
  d_url = 0;
  if (!url)
    return;

  // convert Windows path to URL
  const std::string urlStr(backslashesToSlashes(url));
  const char* urlptr = urlStr.c_str();

  // get relative path
  const char* path = "";
  if (relative && !isAbsolute(urlptr))
    path = relative->urlPath();

  // strip leading "./"
  if (strlen(urlptr)>2 && urlptr[0]=='.' && urlptr[1]==SLASH)
    urlptr += 2;

  // allocate new URL and store path and URL
  d_url = new char[strlen(path) + strlen(urlptr) + 1];
  strcpy(d_url, path);
  strcat(d_url, urlptr);
}


const char* Doc::url() const
{
  return this->d_url;
}

const char* Doc::urlBase() const
{
  if (!d_url)
    return("");

  static char path[_CHARBUF_LEN];
  char* p;
  char* s = path;
  strncpy(path, d_url, sizeof(path) - 1);
  path[sizeof(path) - 1] = '\0';
  if ((p = strrchr(s, SLASH)) != 0)
  {
    s = p + 1;
  }
  else if ((p = strchr(s, COLON)) != 0)
  {
    s = p + 1;
  }

  // do not return extension
  if ((p = strrchr(s, '.')) != 0)
  {
    *p = '\0';
  }

  // do not return url modifier
  p = path;
  if ((p = strrchr(p, '#')) != 0)
    *p = '\0';

  return s;
}

const char* Doc::urlExt() const
{
  if (!d_url)
    return("");

  static char ext[20];
  char* p;

  if ((p = strrchr(d_url, '.')) != 0)
  {
    strncpy(ext, p + 1, sizeof(ext) - 1);
    ext[sizeof(ext)-1] = '\0';
  }
  else
    ext[0] = '\0';

  return(ext);
}


const char* Doc::urlPath() const
{
  if (!d_url) return "";

  static char path[_CHARBUF_LEN];

  strcpy(path, d_url);
  char * slash;
  if ((slash = strrchr(path, SLASH)) != 0)
    *(slash+1) = '\0';
  else
    path[0] = '\0';

  return(path);
}


const char* Doc::urlModifier() const
{
  const char* mod = d_url ? strrchr(d_url, '#') : 0;
  return(mod ? mod : "");
}


// Having both fopen and outputStream is dumb...

FILE *Doc::fopen(const char* mode)
{
  if (d_fp)
    theSystem->error(System::ERRORTYPE_UNKNOWN, "Doc::fopen: %s is already open.\n", d_url ? d_url : "");

  char fn[256];
  if (filename(fn, sizeof(fn)))
  {
    if (strcmp(fn, "-") == 0)
    {
      if (*mode == 'r')
        d_fp = stdin;
      else if (*mode == 'w')
        d_fp = stdout;
    }
    else
      d_fp = ::fopen( fn, mode );
  }

  return(d_fp);
}


void Doc::fclose()
{
  if (d_fp && (strcmp(d_url, "-") != 0) && (strncmp(d_url, "-#", 2) != 0))
    ::fclose(d_fp);

  d_fp = 0;
  if (d_tmpfile)
  {
    theSystem->removeFile(d_tmpfile);
    delete [] d_tmpfile;
    d_tmpfile = 0;
  }
}
