/*****************************************************************************
******************************************************************************
*
*               3D Math routines
*
*               $Revision: 1.1.2.1 $
*               $Date: 2009/01/22 13:26:00 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright 1998-2009 HUMUSOFT s.r.o.

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//



#include <math.h>
#include <string.h>     // memcpy

#include "Math3d.h"

Coordinates3::Coordinates3() { set(0.0, 0.0, 0.0); }
Coordinates3::Coordinates3(double x, double y, double z) { set(x, y, z); }
Coordinates3::Coordinates3(const double *values)         { set(values); }
Coordinates3::Coordinates3(const float *values)          { set(values); }

double Coordinates3::getX() const { return elem[0]; }
double Coordinates3::getY() const { return elem[1]; }
double Coordinates3::getZ() const { return elem[2]; }

void Coordinates3::setX(double x) { elem[0] = x; }
void Coordinates3::setY(double y) { elem[1] = y; }
void Coordinates3::setZ(double z) { elem[2] = z; }

void Coordinates3::set(double x, double y, double z)
{
    elem[0] = x;
    elem[1] = y;
    elem[2] = z;
}

void Coordinates3::set(const double *values)
{
    memcpy(elem, values, 3*sizeof(double));
}

void Coordinates3::set(const float *values)
{
    elem[0] = (double) values[0];
    elem[1] = (double) values[1];
    elem[2] = (double) values[2];
}

double Coordinates3::euclidDist(const Coordinates3* aCoord) const
{
  return sqrt( (aCoord->getX()-elem[0]) * (aCoord->getX()-elem[0]) +
               (aCoord->getY()-elem[1]) * (aCoord->getY()-elem[1]) +
	       (aCoord->getZ()-elem[2]) * (aCoord->getZ()-elem[2]) );
}

Point3::Point3() : Coordinates3() {}
Point3::Point3(double x, double y, double z) : Coordinates3(x, y, z) {}
Point3::Point3(const double *values) : Coordinates3(values) {}
Point3::Point3(const float *values) : Coordinates3(values) {}

Point3 Point3::operator +(const Vector3 &v) const
{
    return Point3(
        elem[0] + v.elem[0],
        elem[1] + v.elem[1],
        elem[2] + v.elem[2]);
}

Point3 Point3::operator -(const Vector3 &v) const
{
    return Point3(
        elem[0] - v.elem[0],
        elem[1] - v.elem[1],
        elem[2] - v.elem[2]);
}

Point3 &Point3::operator +=(const Vector3 &v)
{
    elem[0] += v.elem[0];
    elem[1] += v.elem[1];
    elem[2] += v.elem[2];
    return *this;
}

Point3 &Point3::operator -=(const Vector3 &v)
{
    elem[0] -= v.elem[0];
    elem[1] -= v.elem[1];
    elem[2] -= v.elem[2];
    return *this;
}

Vector3 Point3::operator -(const Point3 &p) const
{
    return Vector3(
        elem[0] - p.elem[0],
        elem[1] - p.elem[1],
        elem[2] - p.elem[2]);
}

Vector3::Vector3() : Coordinates3() {}
Vector3::Vector3(double x, double y, double z) : Coordinates3(x, y, z) {}
Vector3::Vector3(const double *values) : Coordinates3(values) {}
Vector3::Vector3(const float *values) : Coordinates3(values) {}

Vector3 Vector3::operator *(const Matrix4x4 &m) const
{
    double result[3];
    int i;
    for (i=0; i<3; i++)
    {
        result[i] =
              m.elem[0][i]*elem[0]
            + m.elem[1][i]*elem[1]
            + m.elem[2][i]*elem[2]
            + m.elem[3][i];
    }
    return Vector3(result[0], result[1], result[2]);
}

Vector3 Vector3::operator +(const Vector3 &v) const
{
    return Vector3(
        elem[0] + v.elem[0],
        elem[1] + v.elem[1],
        elem[2] + v.elem[2]);
}

Vector3 Vector3::operator -() const
{
    return Vector3(-elem[0], -elem[1], -elem[2]);
}

Vector3 Vector3::operator -(const Vector3 &v) const
{
    return Vector3(
        elem[0] - v.elem[0],
        elem[1] - v.elem[1],
        elem[2] - v.elem[2]);
}

double Vector3::length() const
{
    return sqrt(elem[0]*elem[0]+elem[1]*elem[1]+elem[2]*elem[2]);
}

Vector3 Vector3::scale(double value) const
{
    return Vector3(elem[0]*value, elem[1]*value, elem[2]*value);
}

Vector3 Vector3::normalize() const
{
    double len = length();
    if (fabs(len) <= 1e-10) return Vector3(0.0, 0.0, 0.0);
    return scale(1/len);
}

Vector3 Vector3::cross(const Vector3 &a) const
{
    return Vector3(
        elem[1]*a.elem[2] - elem[2]*a.elem[1],
        elem[2]*a.elem[0] - elem[0]*a.elem[2],
        elem[0]*a.elem[1] - elem[1]*a.elem[0]);
}

double Vector3::dot(const Vector3 &a) const
{
    return
          elem[0]*a.elem[0]
        + elem[1]*a.elem[1]
        + elem[2]*a.elem[2];
}

double Matrix4x4::getElement(int i, int j) const
{
    return elem[i][j];
}

void Matrix4x4::setElement(int i, int j, double value)
{
    elem[i][j] = value;
}

void Matrix4x4::getElements(double values[4][4]) const
{
    int i, j;
    for (i=0; i<4; i++)
        for (j=0; j<4; j++)
            values[i][j] = elem[i][j];
}

void Matrix4x4::setElements(const double values[4][4])
{
    int i, j;
    for (i=0; i<4; i++)
        for (j=0; j<4; j++)
            elem[i][j] = values[i][j];
}

void Matrix4x4::setRow(int i, double x0, double x1, double x2, double x3)
{
    elem[i][0] = x0;
    elem[i][1] = x1;
    elem[i][2] = x2;
    elem[i][3] = x3;
}

void Matrix4x4::setRow(int i, const double *values)
{
    elem[i][0] = values[0];
    elem[i][1] = values[1];
    elem[i][2] = values[2];
    elem[i][3] = values[3];
}

void Matrix4x4::setColumn(int i, double x0, double x1, double x2, double x3)
{
    elem[0][i] = x0;
    elem[1][i] = x1;
    elem[2][i] = x2;
    elem[3][i] = x3;
}

void Matrix4x4::setColumn(int i, const double *values)
{
    elem[0][i] = values[0];
    elem[1][i] = values[1];
    elem[2][i] = values[2];
    elem[3][i] = values[3];
}

void Matrix4x4::getRows(double values[16]) const
{
    memcpy(values, elem, sizeof(double)*16);
}

void Matrix4x4::setRows(const double values[16])
{
    memcpy(elem, values, sizeof(double)*16);
}

void Matrix4x4::getColumns(double values[16]) const
{
    int i, j;
    for (i=0; i<4; i++)
        for (j=0; j<4; j++)
            values[i*4+j] = elem[j][i];
}

void Matrix4x4::setColumns(const double values[16])
{
    int i, j;
    for (i=0; i<4; i++)
        for (j=0; j<4; j++)
            elem[j][i] = values[i*4+j];
}

Matrix4x4 Matrix4x4::transpose() const
{
    Matrix4x4 result;
    int i, j;
    for (i=0; i<4; i++)
        for (j=0; j<4; j++)
            result.elem[i][j] = elem[j][i];
    return result;
}

Matrix4x4 Matrix4x4::operator *(const Matrix4x4 &m) const
{
    Matrix4x4 result;
    int i, j;
    for (i=0; i<4; ++i)
        for (j=0; j<4; ++j)
            result.elem[i][j] =
                  elem[i][0]*m.elem[0][j]
                + elem[i][1]*m.elem[1][j]
                + elem[i][2]*m.elem[2][j]
                + elem[i][3]*m.elem[3][j];
    return result;
}

Matrix4x4 &Matrix4x4::operator *=(const Matrix4x4 &m)
{
    Matrix4x4 result;
    int i, j;
    for (i=0; i<4; ++i)
        for (j=0; j<4; ++j)
            result.elem[i][j] =
                  elem[i][0]*m.elem[0][j]
                + elem[i][1]*m.elem[1][j]
                + elem[i][2]*m.elem[2][j]
                + elem[i][3]*m.elem[3][j];
    memcpy(elem, result.elem, sizeof(double)*16);
    return *this;
}

Vector3 Matrix4x4::operator *(const Vector3 &v) const
{
    Vector3 result;
    int i;
    for (i=0; i<3; i++)
    {
        result.elem[i] =
              elem[i][0]*v.elem[0]
            + elem[i][1]*v.elem[1]
            + elem[i][2]*v.elem[2];
    }
    return result;
}

Point3 Matrix4x4::operator *(const Point3 &p) const
{
    Point3 result;
    int i;
    for (i=0; i<3; i++)
    {
        result.elem[i] =
              elem[i][0]*p.elem[0]
            + elem[i][1]*p.elem[1]
            + elem[i][2]*p.elem[2]
            + elem[i][3];
    }
    return result;
}

double &Matrix4x4::operator[](int index)
{
  // work around illegal index values
  if (index < 0) index = -index;
  if (index > 15) index %= 16;

  // this is pretty ugly, we should use linear indexing everywhere.. later
  return elem[index%4][index/4];
}

const double &Matrix4x4::operator[](int index) const
{
  return const_cast<Matrix4x4*>(this)->operator[](index);
}

Matrix4x4 &Matrix4x4::setToZeros()
{
    int i, j;
    for (i=0; i<4; i++)
        for (j=0; j<4; j++)
            elem[i][j] = 0.0;
    return *this;
}

Matrix4x4 &Matrix4x4::setToIdentity()
{
    int i, j;
    for (i=0; i<4; i++)
        for (j=0; j<4; j++)
            elem[i][j] = (i==j) ? 1.0 : 0.0;
    return *this;
}

Matrix4x4 &Matrix4x4::setToRotation(double angle, const Vector3 &axis)
{
    Vector3 naxis = axis.normalize();

    double s = sin(angle);
    double c = cos(angle);
    double t = 1.0 - c;
    double x = naxis.elem[0];
    double y = naxis.elem[1];
    double z = naxis.elem[2];

    double values[16] =
    {
        t*x*x + c,    t*x*y - s*z,  t*x*z + s*y,  0.0,
        t*x*y + s*z,  t*y*y + c,    t*y*z - s*x,  0.0,
        t*x*z - s*y,  t*y*z + s*x,  t*z*z + c,    0.0,
        0.0,          0.0,          0.0,          1.0,
    };
    setRows(values);
    return *this;
}

Matrix4x4 &Matrix4x4::setToTranslation(const Vector3 &trans)
{
    double values[16] =
    {
        1.0,          0.0,          0.0,          trans.elem[0],
        0.0,          1.0,          0.0,          trans.elem[1],
        0.0,          0.0,          1.0,          trans.elem[2],
        0.0,          0.0,          0.0,          1.0
    };
    setRows(values);
    return *this;
}

Matrix4x4 &Matrix4x4::setToScaling(const Vector3 &scale)
{
    double values[16] =
    {
        scale.elem[0], 0.0,           0.0,           0.0,
        0.0,           scale.elem[1], 0.0,           0.0,
        0.0,           0.0,           scale.elem[2], 0.0,
        0.0,           0.0,           0.0,           1.0
    };
    setRows(values);
    return *this;
}

Matrix4x4 &Matrix4x4::setToViewing(
    const Point3 &viewer,
    const Vector3 &direction,
    const Vector3 &up)
{
    // calculate the new Z axis
    Vector3 za = -direction.normalize();

    // calculate the new Y axis
    Vector3 ya = up.normalize();

    // apply a correction when up vector and direction are not perpendicular
    ya = ya - za.scale(ya.dot(za));
    ya = ya.normalize();

    // calculate the new X axis
    Vector3 xa = ya.cross(za);
    xa = xa.normalize();

    // convert the position to a vector (I'm a formalist and proud of it!)
    Vector3 pos = viewer - Point3(0.0, 0.0, 0.0);

    // fill in the transformation matrix
    double values[16] =
    {
        xa.getX(),    xa.getY(),    xa.getZ(),    -xa.dot(pos),
        ya.getX(),    ya.getY(),    ya.getZ(),    -ya.dot(pos),
        za.getX(),    za.getY(),    za.getZ(),    -za.dot(pos),
        0.0,          0.0,          0.0,          1.0
    };
    setRows(values);
    return *this;
}

Matrix4x4 &Matrix4x4::setToPerspective(double l, double r,
    double b, double t, double n, double f)
{
    double values[16] =
    {
        (2*n)/(r-l),    0.0,            (r+l)/(r-l),        0.0,
        0.0,            (2*n)/(t-b),    (t+b)/(t-b),        0.0,
        0.0,            0.0,            -(f+n)/(f-n),       (-2*f*n)/(f-n),
        0.0,            0.0,            -1.0,               0.0
    };
    setRows(values);
    return *this;
}

Matrix4x4 &Matrix4x4::setToInvPerspective(double l, double r,
    double b, double t, double n, double f)
{
    double values[16] =
    {
        (r-l)/(2*n),    0.0,            0.0,            (r+l)/(2*n),
        0.0,            (t-b)/(2*n),    0.0,            (t+b)/(2*n),
        0.0,            0.0,            0.0,            -1.0,
        0.0,            0.0,            -(f-n)/(2*f*n), (f+n)/(2*f*n)
    };
    setRows(values);
    return *this;
}

Matrix4x4 &Matrix4x4::setToOrthogonal(double l, double r,
    double b, double t, double n, double f)
{
    double values[16] =
    {
        2/(r-l),        0.0,            0.0,            (r+l)/(r-l),
        0.0,            2/(t-b),        0.0,            (t+b)/(t-b),
        0.0,            0.0,            -2/(f-n),       (f+n)/(f-n),
        0.0,            0.0,            0.0,            1.0
    };
    setRows(values);
    return *this;
}

Matrix4x4 &Matrix4x4::setToInvOrthogonal(double l, double r,
    double b, double t, double n, double f)
{
    double values[16] =
    {
        (r-l)/2,        0.0,            0.0,            (r+l)/2,
        0.0,            (t-b)/2,        0.0,            (t+b)/2,
        0.0,            0.0,            (f-n)/-2,       (n+f)/2,
        0.0,            0.0,            0.0,            1.0
    };
    setRows(values);
    return *this;
}



//---------------------------------------------------------------------------
// Quaternion
//---------------------------------------------------------------------------



Quaternion::Quaternion(double q1, double q2, double q3, double q4)
{
  elem[0] = q1;
  elem[1] = q2;
  elem[2] = q3;
  elem[3] = q4;
}



Quaternion::Quaternion(const Vector3 &axis, double angle)
{
  const double sintd2 = sin(angle * 0.5);
  const double len = axis.length();
  const double f = sintd2 / len;
  elem[0] = axis.getX() * f;
  elem[1] = axis.getY() * f;
  elem[2] = axis.getZ() * f;
  elem[3] = cos(angle * 0.5);
}



Quaternion::Quaternion(const Matrix4x4 &m)
{
  // trace is calculated from the three elements on the main diagonal
  const double trace = 1 + m[0] + m[5] + m[10];

  double s, x, y, z, w;
  if (trace > 1e-8)
  {
    s = sqrt(trace) * 2;
    x = (m[9] - m[6]) / s,
    y = (m[2] - m[8]) / s,
    z = (m[4] - m[1]) / s,
    w = 0.25 * s;
  }
  else
  {
    // the largest coefficient on the diagonal determines which column to use
    if (m[0] > m[5] && m[0] > m[10])
    {
      s  = sqrt(1.0 + m[0] - m[5] - m[10]) * 2;
      x = 0.25 * s;
      y = (m[4] + m[1]) / s;
      z = (m[2] + m[8]) / s;
      w = (m[9] - m[6]) / s;
    }
    else if (m[5] > m[10])
    {
      s  = sqrt(1.0 + m[5] - m[0] - m[10]) * 2;
      x = (m[4] + m[1]) / s;
      y = 0.25 * s;
      z = (m[9] + m[6]) / s;
      w = (m[2] - m[8]) / s;
    }
    else
    {
      s  = sqrt(1.0 + m[10] - m[0] - m[5]) * 2;
      x = (m[2] + m[8]) / s;
      y = (m[9] + m[6]) / s;
      z = 0.25 * s;
      w = (m[4] - m[1]) / s;
    }
  }

  // use the results
  elem[0] = x;
  elem[1] = y;
  elem[2] = z;
  elem[3] = w;
}



Quaternion Quaternion::operator*(const Quaternion &q)
{
  return Quaternion(
    elem[3]*q.elem[0] + elem[0]*q.elem[3]
      + elem[1]*q.elem[2] - elem[2]*q.elem[1],
    elem[3]*q.elem[1] + elem[1]*q.elem[3]
      + elem[2]*q.elem[0] - elem[0]*q.elem[2],
    elem[3]*q.elem[2] + elem[2]*q.elem[3]
      + elem[0]*q.elem[1] - elem[1]*q.elem[0],
    elem[3]*q.elem[3] - elem[0]*q.elem[0]
       - elem[1]*q.elem[1] - elem[2]*q.elem[2]);
}



double &Quaternion::operator[](int n)
{
  n = n % 4;
  if (n < 0) n = -n;
  return elem[n];
}



const double &Quaternion::operator[](int n) const
{
  n = n % 4;
  if (n < 0) n = -n;
  return elem[n];
}



double Quaternion::magnitude() const
{
  return sqrt(elem[0]*elem[0]
              + elem[1]*elem[1]
              + elem[2]*elem[2]
              + elem[3]*elem[3]);
}



void Quaternion::toAxisAngle(Vector3 &axis, double &angle)
{
  double sina2 = sqrt(elem[0]*elem[0] + elem[1]*elem[1] + elem[2]*elem[2]);
  angle = (2.0*atan2(sina2, (double) elem[3]));
  if (sina2 >= 1e-8)
  {
    sina2 = 1.0/sina2;
    axis = Vector3(elem[0] * sina2,
                   elem[1] * sina2,
                   elem[2] * sina2);
  }
  else
  {
    axis = Vector3(0, 1, 0);
    angle = 0.0;
  }
}
