/*****************************************************************************
******************************************************************************
*
*               3D Math routines
*
*               $Revision: 1.1.2.1 $
*               $Date: 2009/01/22 16:01:10 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright 1998-2009 HUMUSOFT s.r.o.

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//



#ifndef MATH3D_H
#define MATH3D_H

#include "System.h"   // OPENVRMLDLLEXPORT

class Matrix4x4;

/** This object represents a coordinate triplet for use in 3d space. */
class OPENVRMLDLLEXPORT Coordinates3
{
    protected:

        /** Array containing the X, Y, Z coordinates, respectively. */
        double elem[3];

    public:

        /** Default constructor.
          * Coordinate values are set to (0, 0, 0).
          */
        Coordinates3();

        /** Constructor.
          * Coordinate values are set as specified.
          * @param x Value for the X coordinate.
          * @param y Value for the Y coordinate.
          * @param z Value for the Z coordinate.
          */
        Coordinates3(double x, double y, double z);

        /** Constructor.
          * Coordinate values are set as specified.
          * @param values Array of coordinate values (X, Y, Z, respectively).
          */
        Coordinates3(const double *values);

        /** Constructor.
          * Coordinate values are set as specified.
          * @param values Array of coordinate values (X, Y, Z, respectively).
          */
        Coordinates3(const float *values);

        /** Gets the value of the X coordinate.
          * @returns The X coordinate value.
          */
        double getX() const;

        /** Gets the value of the Y coordinate.
          * @returns The Y coordinate value.
          */
        double getY() const;

        /** Gets the value of the Z coordinate.
          * @returns The Z coordinate value.
          */
        double getZ() const;

        /** Sets the value of the X coordinate.
          * @param x New X coordinate value.
          */
        void setX(double x);

        /** Sets the value of the Y coordinate.
          * @param y New Y coordinate value.
          */
        void setY(double y);

        /** Sets the value of the Z coordinate.
          * @param z New Z coordinate value.
          */
        void setZ(double z);

        /** Sets all three coordinates.
          * @param x New X coordinate value.
          * @param y New Y coordinate value.
          * @param z New Z coordinate value.
          */
        void set(double x, double y, double z);

        /** Sets all three coordinates.
          * @values Array of the coordinate values (X, Y, Z, respectively).
          */
        void set(const double *values);

        /** Sets all three coordinates.
          * @values Array of the coordinate values (X, Y, Z, respectively).
          */
        void set(const float *values);

	/**
	 * Computes and returns euclidean distance between this and aCoord  coordinates
	 * @param aCoord coordinate to which is distance computed
	 8 @returns euclidean distance
	 */
	double euclidDist(const Coordinates3* aCoord) const;

};

class OPENVRMLDLLEXPORT Vector3;

/** This object represents a point in 3d space. */
class OPENVRMLDLLEXPORT Point3 : public Coordinates3
{
    friend class Matrix4x4;
    public:

        /** Default constructor.
          * Point coordinates are set to (0, 0, 0).
          */
        Point3();

        /** Constructor.
          * Point coordinates are set as specified.
          * @param x Value for the X coordinate.
          * @param y Value for the Y coordinate.
          * @param z Value for the Z coordinate.
          */
        Point3(double x, double y, double z);

        /** Constructor.
          * Point coordinates are set as specified.
          * @param values Array of coordinate values (X, Y, Z, respectively).
          */
        Point3(const double *values);

        /** Constructor.
          * Point coordinates are set as specified.
          * @param values Array of coordinate values (X, Y, Z, respectively).
          */
        Point3(const float *values);

        /** Addition operator.
          * Adds a vector to the point, effectively translating it
          * by the coordinates of the vector.
          * @param v Vector operand.
          * @returns Resulting point.
          */
        Point3 operator +(const Vector3 &v) const;

        /** Subtraction operator.
          * Subtracts a vector to the point, effectively translating it
          * in the reverse direction of the vector.
          * @param v Vector operand.
          * @returns Resulting point.
          */
        Point3 operator -(const Vector3 &v) const;

        /** Addition-assignment operator.
          * Adds a vector to the point and stores the result back to it.
          * @param v Vector operand.
          * @returns Reference to this point.
          */
        Point3 &operator +=(const Vector3 &v);

        /** Subtraction-assignment operator.
          * Subtracts a vector from the point and stores the result back
          * to it.
          * @param v Vector operand.
          * @returns Reference to this point.
          */
        Point3 &operator -=(const Vector3 &v);

        /** Subtraction operator.
          * The result is a difference vector between this point
          * and the given point, respectively.
          * @param p Point operand.
          * @returns Resulting vector.
          */
        Vector3 operator -(const Point3 &p) const;
};

/** This object represents a vector in 3d space. */
class Vector3 : public Coordinates3
{
    friend class Matrix4x4;
    friend class Point3;
    public:

        /** Default constructor.
          * Vector coordinates are set to (0, 0, 0).
          */
        Vector3();

        /** Constructor.
          * Vector coordinates are set as specified.
          * @param x Value for the X coordinate.
          * @param y Value for the Y coordinate.
          * @param z Value for the Z coordinate.
          */
        Vector3(double x, double y, double z);

        /** Constructor.
          * Vector coordinates are set as specified.
          * @param values Array of coordinate values (X, Y, Z, respectively).
          */
        Vector3(const double *values);

        /** Constructor.
          * Vector coordinates are set as specified.
          * @param values Array of coordinate values (X, Y, Z, respectively).
          */
        Vector3(const float *values);

        /** Multiplication operator.
          * Multiplies the vector by the specified matrix,
          * considering the vector to be a row vector
          * (X, Y, Z, 1).
          * @param m Matrix operand.
          * @returns Resulting vector.
          */
        Vector3 operator *(const Matrix4x4 &m) const;

        /** Addition operator.
          * @param v Vector operand.
          * @returns Resulting vector.
          */
        Vector3 operator +(const Vector3 &v) const;

        /** Negation operator.
          * @returns Resulting vector.
          */
        Vector3 operator -() const;

        /** Subtraction operator.
          * @param v Vector operand.
          * @returns Resulting vector.
          */
        Vector3 operator -(const Vector3 &v) const;

        /** Calculates the length of this vector.
          */
        double length() const;

        /** Scales the vector by the given factor.
          * @param x Scaling factor.
          * @returns The scaled vector.
          */
        Vector3 scale(double x) const;

        /** Normalizes this vector.
          * @returns The normalized vector.
          */
        Vector3 normalize() const;

        /** Calculates the cross product of this vector and the given
          * vector.
          * @param vec Second vector operand.
          * @returns The cross product vector.
          */
        Vector3 cross(const Vector3 &vec) const;

        /** Calculates the dot product of this vector
          * and the specified vector.
          * @param vec Second vector operand.
          * @returns The dot product.
          */
        double dot(const Vector3 &vec) const;
};

/** A 4x4 matrix. */
class OPENVRMLDLLEXPORT Matrix4x4
{
    friend class Vector3;
    protected:

        /** Matrix elements. */
        double elem[4][4];

    public:

        /** Multiplication operator.
          * @param m Multiplication operand.
          * @returns Result of the multiplication.
          */
        Matrix4x4 operator *(const Matrix4x4 &m) const;

        /** Multiplication-assignment operator.
          * @param m Multiplication operand.
          * @returns Reference to this matrix.
          */
        Matrix4x4 &operator *=(const Matrix4x4 &m);

        /** Multiplication operator.
          * Multiplies the matrix by the specified vector,
          * considering it to be a column vector
          * (X, Y, Z, 0).
          * @param v Vector operand.
          * @returns Resulting vector.
          */
        Vector3 operator *(const Vector3 &v) const;

        /** Multiplication operator.
          * Multiplies the matrix by the specified point,
          * considering it to be a column vector
          * (X, Y, Z, 1).
          * @param p Point operand.
          * @returns Resulting point.
          */
        Point3 operator *(const Point3 &p) const;

        /**
         * Operator for accessing the elements of the matrix
         * by a single index. Elements are numbered as follows:
         * <pre>
         * M = [
         *        0   4   8  12
         *        1   5   9  13
         *        2   6  10  14
         *        3   7  11  15
         *      ];
         * </pre>
         * @param index Element index (0..15).
         * @return A reference to the given element.
         */
        double &operator[](int index);

        /**
         * A read-only variant of the operator.
         * @param index Element index (0..15).
         * @return A read-only reference to the given element.
         */
        const double &operator[](int index) const;

        /** Reads a given element from the matrix.
          * @param i Row index (0..3).
          * @param j Column index (0..3).
          * @returns Element value.
          */
        double getElement(int i, int j) const;

        /** Sets a given element in the matrix.
          * @param i Row index (0..3).
          * @param j Column index (0..3).
          * @param value Value to be set.
          */
        void setElement(int i, int j, double value);

        /** Reads all elements from the matrix.
          * @param values Array to receive the values.
          */
        void getElements(double values[4][4]) const;

        /** Sets all elements in the matrix.
          * @param values Values to be set.
          */
        void setElements(const double values[4][4]);

        /** Sets a row of elements.
          * @param i Row index (0..3).
          * @param x0 Value for the 0th element in a row.
          * @param x1 Value for the 1th element in a row.
          * @param x2 Value for the 2th element in a row.
          * @param x3 Value for the 3th element in a row.
          */
        void setRow(int i, double x0, double x1, double x2, double x3);

        /** Sets a row of elements.
          * @param i Row index (0..3).
          * @param values Array of values to be set (4 values).
          */
        void setRow(int i, const double *values);

        /** Sets a column of elements.
          * @param i Row index (0..3).
          * @param x0 Value for the 0th element in a column.
          * @param x1 Value for the 1th element in a column.
          * @param x2 Value for the 2th element in a column.
          * @param x3 Value for the 3th element in a column.
          */
        void setColumn(int i, double x0, double x1, double x2, double x3);

        /** Sets a column of elements.
          * @param i Column index (0..3).
          * @param values Array of values to be set (4 values).
          */
        void setColumn(int i, const double *values);

        /** Reads all rows of the matrix (reads all elements
          * rowwise).
          * @param values Array to receive the values (16 values).
          */
        void getRows(double *values) const;

        /** Sets all rows of the matrix (sets all elements
          * rowwise).
          * @param values Array of values to be set (16 values).
          */
        void setRows(const double *values);

        /** Reads all columns of the matrix (reads all elements
          * columnwise).
          * @param values Array to receive the values (16 values).
          */
        void getColumns(double *values) const;

        /** Sets all columns of the matrix (sets all elements
          * columnwise).
          * @param values Array of values to be set (16 values).
          */
        void setColumns(const double *values);

        /** Transposes the matrix.
          * @returns Transposed version of this matrix.
          */
        Matrix4x4 transpose() const;

        /** Sets this matrix to all zeros.
          * @returns Reference to this matrix.
          */
        Matrix4x4 &setToZeros();

        /** Sets this matrix to identity.
          * @returns Reference to this matrix.
          */
        Matrix4x4 &setToIdentity();

        /** Sets this matrix to a rotation matrix.
          * @param angle Angle of rotation, in radians.
          * @param axis Axis of rotation.
          * @returns Reference to this matrix.
          */
        Matrix4x4 &setToRotation(double angle, const Vector3 &axis);

        /** Sets this matrix to a translation matrix.
          * @param trans Translation vector.
          * @returns Reference to this matrix.
          */
        Matrix4x4 &setToTranslation(const Vector3 &trans);

        /** Sets this matrix to a scaling matrix.
          * @param scale Scale vector.
          * @returns Reference to this matrix.
          */
        Matrix4x4 &setToScaling(const Vector3 &scale);

        /** Sets this matrix to a viewing matrix with given camera
          * position and orientation. The matrix transforms coordinates
          * so that the camera is at (0, 0, 0) and looks down
          * the -Z axis.
          * @param viewer Position of the camera.
          * @param direction Direction of the camera.
          * @param up Up vector of the camera.
          * @returns Reference to this matrix.
          */
        Matrix4x4 &setToViewing(
            const Point3 &viewer,
            const Vector3 &direction,
            const Vector3 &up);

        /** Sets this matrix to a perspective projection matrix.
          * @param left Leftmost visible X coordinate.
          * @param right Rightmost visible X coordinate.
          * @param bottom Bottommost visible Y coordinate.
          * @param top Topmost visible Y coordinate.
          * @param near Nearest visible Z coordinate.
          * @param far Farthest visible Z coordinate.
          * @returns Reference to this matrix.
          */
        Matrix4x4 &setToPerspective(double left, double right,
            double bottom, double top, double nearest, double farthest);

        /** Sets this matrix to an inverse perspective projection matrix.
          * @param left Leftmost visible X coordinate.
          * @param right Rightmost visible X coordinate.
          * @param bottom Bottommost visible Y coordinate.
          * @param top Topmost visible Y coordinate.
          * @param near Nearest visible Z coordinate.
          * @param far Farthest visible Z coordinate.
          * @returns Reference to this matrix.
          */
        Matrix4x4 &setToInvPerspective(double left, double right,
            double bottom, double top, double nearest, double farthest);

        /** Sets this matrix to an orthogonal projection matrix.
          * @param left Leftmost visible X coordinate.
          * @param right Rightmost visible X coordinate.
          * @param bottom Bottommost visible Y coordinate.
          * @param top Topmost visible Y coordinate.
          * @param near Nearest visible Z coordinate.
          * @param far Farthest visible Z coordinate.
          * @returns Reference to this matrix.
          */
        Matrix4x4 &setToOrthogonal(double left, double right,
            double bottom, double top, double nearest, double farthest);

        /** Sets this matrix to an inverse orthogonal projection matrix.
          * @param left Leftmost visible X coordinate.
          * @param right Rightmost visible X coordinate.
          * @param bottom Bottommost visible Y coordinate.
          * @param top Topmost visible Y coordinate.
          * @param near Nearest visible Z coordinate.
          * @param far Farthest visible Z coordinate.
          * @returns Reference to this matrix.
          */
        Matrix4x4 &setToInvOrthogonal(double left, double right,
            double bottom, double top, double nearest, double farthest);
};



/** A quaternion. */
class OPENVRMLDLLEXPORT Quaternion
{
private:

  /** The elements. */
  double elem[4];

public:

  /**
   * Constructor.
   * @param q1 The first element.
   * @param q2 The second element.
   * @param q3 The third element.
   * @param q4 The fourth element.
   */
  Quaternion(double q1, double q2, double q3, double q4);

  /**
   * Constructor. Builds a quaternion representing the rotation
   * by the given angle around the given axis.
   * @param axis The rotation axis.
   * @param angle The rotation angle.
   */
  Quaternion(const Vector3 &axis, double angle);

  /**
   * Constructor. Builds a quaternion from the rotational part
   * of the given matrix.
   * @param m The matrix.
   */
  explicit Quaternion(const Matrix4x4 &m);

  /**
   * The multiplication operator.
   * @param q The second operand.
   * @return The resulting quaternion.
   */
  Quaternion operator*(const Quaternion &q);

  /**
   * Element access operator.
   * @param n The element index (0..3, other values are wrapped).
   * @return A reference to the element.
   */
  double &operator[](int n);

  /**
   * Constant element access operator.
   * @param n The element index (0..3, other values are wrapped).
   * @return A constant reference to the element.
   */
  const double &operator[](int n) const;

  /**
   * Calculates the magnitude of the quaternion.
   * @return The magnitude.
   */
  double magnitude() const;

  /**
   * Converts the quaternion to a rotation defined by
   * an axis and an angle.
   * @param axis A variable to receive the axis.
   * @param angle A variable to receive the angle.
   */
  void toAxisAngle(Vector3 &axis, double &angle);
};

#endif
