/*****************************************************************************
******************************************************************************
*
*               Recording information
*
*               $Revision: 1.1.2.1 $
*               $Date: 2009/01/22 16:01:54 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright 1998-2009 HUMUSOFT s.r.o.

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#ifndef RECORDINGINFO_H
#define RECORDINGINFO_H

/**
 * This object holds information about recording set up for a given scene.
 * <p>
 * There are two modes of recording for a VRML scene: 3D recording,
 * which is done in the scene itself and results in a VRML file containing
 * the animation, and 2D recording, which is done by the viewers
 * and results in planar movie of the animation as seen by the given viewer.
 * Both modes can be combined simultaneously.
 */
class RecordingInfo
{
private:

  /** True if recording is scheduled, false if not. */
  bool scheduled;

  /** Time when the scheduled recording should start. */
  double sched_start;

  /** Time when the scheduled recording should stop. */
  double sched_stop;

  /** Last tick time. */
  double last_time;

  /** True if recording is currently in progress, false if not. */
  bool in_progress;

  /** Time when the recording last started, -1 if not yet started. */
  double start_time;

  /** Time when the recording finished, -1 if not yet started. */
  double stop_time;

  /** True if 3d recording mode is enabled, false if not. */
  bool mode_3d;

public:

  /**
   * Constructor.
   */
  RecordingInfo()
    : scheduled(false),
      sched_start(-1),
      sched_stop(-1),
      last_time(0),
      in_progress(false),
      start_time(-1),
      stop_time(-1),
      mode_3d(false)
    {}

  /**
   * Returns the time when the recording last started.
   * @return The start time.
   */
  double GetStartTime() const { return start_time; }

  /**
   * Returns the time when the recording last stopped.
   * @return The stop time.
   */
  double GetStopTime() const { return stop_time; }

  /**
   * Checks whether a recording is scheduled.
   * @return True if a recording is scheduled, false if it is not.
   */
  bool IsScheduled() { return scheduled; }

  /**
   * Enables or disables scheduled recording.
   * @param onoff True to enable scheduled recording, false to disable it.
   */
  void SetScheduled(bool onoff) { scheduled = onoff; }

  /**
   * Returns the start time for the scheduled recording.
   * @return The scheduled start time.
   */
  double GetScheduledStartTime() { return sched_start; }

  /**
   * Sets the start time for the scheduled recording.
   * @param val The stop time.
   */
  void SetScheduledStartTime(double val) { sched_start = val; }

  /**
   * Returns the stop time for the scheduled recording.
   * @return The scheduled stop time.
   */
  double GetScheduledStopTime() { return sched_stop; }

  /**
   * Sets the start time for the scheduled recording.
   * @param val The stop time.
   */
  void SetScheduledStopTime(double val) { sched_stop = val; }

  /**
   * Checks whether the 3D recording mode is enabled.
   * @return True if the 3D recording mode is enabled, false if not.
   */
  bool Is3DModeEnabled() const { return mode_3d; }

  /**
   * Enables or disables the 3D recording mode.
   * @param onoff True to enable the 3D recording, false to disable it.
   */
  void Enable3DMode(bool onoff) { mode_3d = onoff; }

  /**
   * Checks whether recording is currently in progress.
   * @return True if recording is in progress, false if it is not.
   */
  bool IsRunning() const { return in_progress; }

  /**
   * Returns true if the recording has just started in this time element.
   * @return True if the recording has just started, false otherwise.
   */
  bool HasJustStarted() const { return start_time == last_time; }

  /**
   * Returns true if the recording has just stopped in this time element.
   * @return True if the recording has just stopped, false otherwise.
   */
  bool HasJustStopped() const { return stop_time == last_time; }

  /**
   * Starts recording (does nothing if recording is already running).
   */
  void Start()
  {
    if (!in_progress)
    {
      start_time = last_time;
      in_progress = true;
    }
  }

  /**
   * Stops recording (does nothing if recording is not running).
   */
  void Stop()
  {
    if (in_progress)
    {
      stop_time = last_time;
      in_progress = false;
    }
  }

  /**
   * Updates the stored last time value to the given value,
   * possibly starting or stopping the recording if scheduled.
   * @param time The new time value.
   * @note The recording is started or stopped only when the appropriate
   * time mark is really passed (when the previous time is before the mark
   * and the new time is after the mark). Merely setting the time
   * after the mark is not enough. This is necessary to allow manual
   * switching of recording.
   */
  void Tick(double time)
  {
    if (scheduled && time >= sched_start && last_time <= sched_start) Start();
    if (scheduled && time >= sched_stop && last_time <= sched_stop) Stop();
    last_time = time;
  }

  /**
   * Sets the last time to the given value without affecting
   * the scheduled recording.
   */
  void SetTime(double time) { last_time = time; }
};

#endif
