/*****************************************************************************
******************************************************************************
*
*               Reference-counted array
*
*               $Revision: 1.1.2.2 $
*               $Date: 2010/09/20 10:47:58 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright 1998-2009 HUMUSOFT s.r.o.

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#ifndef REFCNTARRAY_H
#define REFCNTARRAY_H



/** A shareable, reference counted array of data of type T. */
template <typename T>
class RefCntArray
{
private:

  /** Reference counter. */
  int refs;

  /** Number of values. */
  size_t length;

  /** A dynamically allocated array of values. */
  T* values;

protected:

  /** Protected default constructor. Client code should use create(). */
  RefCntArray()
    : refs(1), length(0), values(0) {}

  /** Protected constructor. Client code should use create(). */
  RefCntArray(size_t len)
    : refs(1), length(len), values(0)
    {
      if (length < 0) length = 0;
      if (length > 0) values = new T[length];
    }

  /** Protected copy constructor. Client code should use clone(). */
  RefCntArray(const RefCntArray &d)
    : refs(1), length(d.length), values(0)
    {
      if (length > 0) values = new T[length];
      for (size_t i=0; i<length; i++) values[i] = d.values[i];
    }

public:

  ~RefCntArray()
    { delete[] values; }

  /**
   * Returns a newly allocated array of the given size, with reference
   * counter set to one.
   */
  static RefCntArray* create(int length)
    { return new RefCntArray(length); }

  /**
   * Returns a newly allocated array of the given size, initialized
   * to the given content, with reference counter set to one.
   */
  static RefCntArray* create(size_t length, const T* initdata)
    {
      RefCntArray* result = new RefCntArray(length);
      if (initdata != 0)
        for (size_t i=0; i<length; i++) result->values[i] = initdata[i];
      return result;
    }

  /**
   * Returns a newly allocated array of the given size, initialized
   * as follows: the first n elements of data are initialized
   * using the given data, the rest is filled with the given value.
   */
  static RefCntArray* create(size_t length, const RefCntArray* src, T initval)
    {
      RefCntArray *result = new RefCntArray(length);
      if (src->length > result->length)
      {
        size_t i;
        for (i=0; i<length; i++) result->values[i] = src->values[i];
      }
      else
      {
        size_t i;
        for (i=0; i<src->length; i++) result->values[i] = src->values[i];
        for (i=src->length; i<length; i++) result->values[i] = initval;
      }
      return result;
    }

  /**
   * Increments the reference counter by one.
   */
  void incRef()
    { refs++; }

  /**
   * Decrements the reference counter by one, deleting this object
   * if the reference counter reaches zero.
   */
  void decRef()
    { refs--; if (refs == 0) delete this; }

  /**
   * Returns a newly allocated copy of this object with reference count
   * reset to one.
   */
  RefCntArray* clone() const
    { return new RefCntArray(*this); }

  /**
   * Ensures that the data part of this object is unique (not shared).
   * This is done by returning the same object if not shared and by
   * "un-sharing" (returning a clone, derefencing the original) if it is shared.
   */
  RefCntArray* makeUnique()
    { if (refs<=1) return(this); RefCntArray* result = clone(); decRef(); return(result); }

  /**
   * Returns the value of the reference counter.
   */
  int getRefCount() const
    { return refs; }

  /**
   * Returns the length of the data.
   */
  size_t getLength() const
    { return length; }

  /**
   * Returns the n-th element of the data (0 if the element does not exist).
   */
  T getValue(size_t n) const
    { return (n < length) ? values[n] : 0; }

  /**
   * Sets the n-th element of the data (does nothing if the element
   * does not exist).
   */
  void setValue(size_t n, T value)
    { if (n < length) values[n] = value; }

  /**
   * Returns the pointer to the data (0 if length is zero).
   */
  T* getValues()
    { return values; }

  /**
   * Returns a pointer to the data (0 if length is zero).
   */
  const T* getValues() const
    { return values; }

  /**
   * Returns true if the contents of the arrays are the same,
   * false otherwise.
   */
  bool equals(const RefCntArray* a)
    {
      if (a->length != length) return false;      // an easy case
      for (size_t i = 0; i<length; i++)
        if (values[i] != a->values[i]) return false;
      return true;
    }
};


#endif
