/*****************************************************************************
******************************************************************************
*
*               JavaScript interface
*
*               $Revision: 1.1.2.9 $
*               $Date: 2012/02/22 11:39:37 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright 1998-2012 HUMUSOFT s.r.o.
// Portions Copyright (C) 1998  Chris Morley

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

# ifndef OPENVRML_DISABLE_JAVASCRIPT

# include <assert.h>
# include <stdio.h>
# include <string.h>

#include "ScriptJS.h"
#include "Doc.h"
#include "MathUtils.h"
#include "System.h"
#include "VrmlException.h"
#include "VrmlNodeScript.h"
#include "VrmlNamespace.h"
#include "VrmlNodeType.h"
#include "VrmlScene.h"
#include "VrmlField.h"

// This is nominally a private include but I want to subclass Arrays...
#include <js/jsarray.h>


#include <vector>
#include <sstream>
using namespace std;



# define MAX_HEAP_BYTES 4L * 1024L * 1024L
# define STACK_CHUNK_BYTES 4024L

//#define SCRIPTJS_DEBUG

JSRuntime  *ScriptJS::rt = 0;	// Javascript runtime singleton object
int ScriptJS::nInstances = 0;	// Number of distinct script objects


// Global class and functions

static JSClass globalClass = {
    "global", 0,
    JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,
    JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,   JS_FinalizeStub
};


static JSBool Print(JSContext *, JSObject *, uintN, jsval *, jsval *);
static VrmlField *jsvalToVrmlField( JSContext *cx,
				    jsval v,
				    VrmlField::VrmlFieldType expectType );


static JSFunctionSpec globalFunctions[] = {
    {"print",           Print,          0},
    { 0 }
};

// Browser class and functions

static JSClass browserClass = {
    "Browser", 0,
    JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub, JS_PropertyStub,
    JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  JS_FinalizeStub
};

static JSBool getName(JSContext*, JSObject*, uintN, jsval *, jsval *);
static JSBool getVersion(JSContext*, JSObject*, uintN, jsval *, jsval *);
static JSBool getCurrentSpeed(JSContext*, JSObject*, uintN, jsval *, jsval *);
static JSBool getCurrentFrameRate(JSContext*, JSObject*, uintN, jsval*, jsval*);
static JSBool getWorldURL(JSContext*, JSObject*, uintN, jsval *, jsval *);
static JSBool replaceWorld(JSContext*, JSObject*, uintN, jsval *, jsval *);
static JSBool createVrmlFromString(JSContext*, JSObject*, uintN, jsval *, jsval *);
static JSBool createVrmlFromURL(JSContext*, JSObject*, uintN, jsval *, jsval *);

static JSBool addRoute(JSContext*, JSObject*, uintN, jsval *, jsval *);
static JSBool deleteRoute(JSContext*, JSObject*, uintN, jsval *, jsval *);
static JSBool loadURL(JSContext*, JSObject*, uintN, jsval *, jsval *);
static JSBool setDescription(JSContext*, JSObject*, uintN, jsval *, jsval *);


static JSFunctionSpec browserFunctions[] = {
  {"getName", getName, 0},
  {"getVersion", getVersion, 0},
  {"getCurrentSpeed", getCurrentSpeed, 0},
  {"getCurrentFrameRate", getCurrentFrameRate, 0},
  {"getWorldURL", getWorldURL, 0},

  {"replaceWorld", replaceWorld, 0},

  {"createVrmlFromString", createVrmlFromString, 0},
  {"createVrmlFromURL", createVrmlFromURL, 0},
  {"addRoute", addRoute, 0},
  {"deleteRoute", deleteRoute, 0},
  {"loadURL", loadURL, 0},
  {"setDescription", setDescription, 0},
  {0}
};



static void ErrorReporter(JSContext *, const char *, JSErrorReport *);


/* Pointer to jsval conversion (and back) */

static std::vector<void*> PtrTable;

static void* JSValToPtr(jsval x)
{
  return(PtrTable.at(JSVAL_TO_INT(x)));
}

static jsval PtrToJSVal(void* x)
{
  // the pointer table grows with each stored pointer
  // it cleared only when there are no ScriptJS objects left
  PtrTable.push_back(x);
  return(INT_TO_JSVAL(PtrTable.size()-1));
}



// Construct from inline script

ScriptJS::ScriptJS( VrmlNodeScript *node, const char *source ) :
  d_node(node), d_cx(0), d_globalObj(0), d_browserObj(0)
{
  if (! rt)
    rt = JS_Init( MAX_HEAP_BYTES );

  ++nInstances;

  if (rt)
    d_cx = JS_NewContext( rt, STACK_CHUNK_BYTES );

  if (d_cx)
    {
      JS_SetErrorReporter(d_cx, ErrorReporter);

      // Define the global objects (builtins, Browser, SF*, MF*) ...
      d_globalObj = JS_NewObject( d_cx, &globalClass, 0, 0 );
      JS_InitStandardClasses( d_cx, d_globalObj );
      JS_DefineFunctions( d_cx, d_globalObj, globalFunctions );

      // VRML-like TRUE, FALSE syntax
      if (! JS_DefineProperty( d_cx, d_globalObj, "FALSE",
			       BOOLEAN_TO_JSVAL(false), 0, 0,
			       JSPROP_READONLY | JSPROP_PERMANENT ))
	theSystem->error(System::ERRORTYPE_UNKNOWN, "JS_DefineProp FALSE failed\n");
      if (! JS_DefineProperty( d_cx, d_globalObj, "TRUE",
			       BOOLEAN_TO_JSVAL(true), 0, 0,
			       JSPROP_READONLY | JSPROP_PERMANENT ))
	theSystem->error(System::ERRORTYPE_UNKNOWN, "JS_DefineProp TRUE failed\n");

      jsval thisjsval(PtrToJSVal(this));
      if (! JS_DefineProperty( d_cx, d_globalObj, "_script",
			       thisjsval, 0, 0,
			       JSPROP_READONLY | JSPROP_PERMANENT ))
	theSystem->error(System::ERRORTYPE_UNKNOWN, "JS_DefineProp _script failed\n");

      // Browser object
      d_browserObj = JS_DefineObject( d_cx, d_globalObj, "Browser",
				    &browserClass, 0, 0 );
      if (! JS_DefineProperty( d_cx, d_browserObj, "_script",
			       thisjsval, 0, 0,
			       JSPROP_READONLY | JSPROP_PERMANENT ))
	theSystem->error(System::ERRORTYPE_UNKNOWN, "JS_DefineProp _script failed\n");

      if (! JS_DefineFunctions( d_cx, d_browserObj, browserFunctions ))
	theSystem->error(System::ERRORTYPE_UNKNOWN, "JS_DefineFunctions failed\n");

      // Define SF*/MF* classes
      defineAPI();

      // Define field/eventOut vars for this script
      defineFields();

      /* These should indicate source location for diagnostics. */
      char *filename = 0;
      uintN lineno = 0;

      jsval rval;
      if (! JS_EvaluateScript( d_cx, d_globalObj, source, strlen(source),
			       filename, lineno, &rval))
	theSystem->error(System::ERRORTYPE_UNKNOWN, "JS_EvaluateScript failed\n");
    }
}


ScriptJS::~ScriptJS()
{
  JS_DestroyContext(d_cx);
  if (--nInstances == 0)
  {
    PtrTable.clear();
    JS_Finish(rt);
    rt = 0;
  }
}


static double s_timeStamp;	// go away...

// Run a specified script

void ScriptJS::activate(double timeStamp, const char *fname, int argc,
   const VrmlField *argv[])
{
  if (! d_cx) return;

  jsval fval, rval;

  if (! JS_LookupProperty( d_cx, d_globalObj, fname, &fval ))
    theSystem->error(System::ERRORTYPE_UNKNOWN, "JS_LookupProperty %s failed\n", fname);

  else if (! JSVAL_IS_VOID(fval))
  {
    jsval *jsargv = new jsval[argc];

#if defined(__ORBISNAP__)

    // list of pointers to objects we want to protect against gc
    // (of course we then need to free them manually at the end).
    jsval **protectedPtrs = new jsval *[argc];

#endif

    d_timeStamp = timeStamp;
    s_timeStamp = timeStamp;	// this won't work for long...

#if defined(__ORBISNAP__)

    // convert VrmlField*'s to (gc-protected) jsvals
    int i;
    for (i=0; i<argc; ++i)
    {
      if (argv[i] == 0)
      {
        jsargv[i] = 0;
        protectedPtrs[i] = 0;
      }
      else
      {
        jsargv[i] = vrmlFieldToJSVal(argv[i]->fieldType(), argv[i], false);
        protectedPtrs[i] = &jsargv[i];
        JS_AddNamedRoot(d_cx, protectedPtrs[i], "ScriptJS argument");
      }
    }

#else

    // convert VrmlField*'s to (gc-protected) jsvals
    int i;
    for (i=0; i<argc; ++i)
      jsargv[i] = argv[i] ? vrmlFieldToJSVal(argv[i]->fieldType(), argv[i], true) : JSVAL_NULL;

#endif

    if (! JS_CallFunctionValue( d_cx, d_globalObj, fval, argc, jsargv, &rval))
    {
      theSystem->error(System::ERRORTYPE_UNKNOWN, "JS_CallFunctionName(%s) failed\n", fname);
    }

#if defined(__ORBISNAP__)

    // Free up args
    for (i=0; i<argc; ++i)
    {
      if (protectedPtrs[i] != 0)
      {
        JS_RemoveRoot(d_cx, protectedPtrs[i]);
      }
    }
    delete[] protectedPtrs;

#else

    // Free up args
    for (i=0; i<argc; ++i)
      if (JSVAL_IS_GCTHING(jsargv[i]))
        JS_RemoveRoot( d_cx, JSVAL_TO_GCTHING(jsargv[i]));

#endif

    delete [] jsargv;
  }
}

// Get a handle to the scene from a ScriptJS
VrmlScene *ScriptJS::browser() { return d_node->browser(); }

// Get a handle to the ScriptJS from the global object
static ScriptJS *objToScript( JSContext *cx, JSObject *obj )
{
  jsval val;
  if (! JS_LookupProperty( cx, obj, "_script", &val ))
    theSystem->error(System::ERRORTYPE_UNKNOWN, "JS_LookupProperty _script failed\n");

  else if (! JSVAL_IS_INT(val)) // Actually JSVAL_IS_PRIVATE
    theSystem->error(System::ERRORTYPE_UNKNOWN, "JS_LookupProperty _script failed- NULL _script\n");

  return static_cast<ScriptJS*>(JSValToPtr(val));
}

// Get a pointer to an eventOut from an object
static char *objToEventOut( JSContext *cx, JSObject *obj )
{
  jsval val;
  if (JS_LookupProperty( cx, obj, "_eventOut", &val ) &&
      JSVAL_IS_INT(val))
    return(static_cast<char*>(JSValToPtr(val)));

  return(0);
}

// Check whether we are modifying a prop on an eventOut object, and if so,
// notify the script.

static void checkEventOut( JSContext *cx, JSObject *obj, VrmlField *val )
{
  char *eventOut = 0;
  ScriptJS *script = 0;

  if ((eventOut = objToEventOut( cx, obj )) != 0 &&
      (script = objToScript( cx, JS_GetParent( cx, obj ))) != 0)
    {
      VrmlNodeScript *scriptNode = script->scriptNode();
      scriptNode->setEventOut( eventOut, val );
    }
}

// data converters

static JSBool floatsToJSArray( int n, const float *f, JSContext *cx, jsval *rval )
{
  static const int MAX_FIXED = 20;
  jsval jsfixed[MAX_FIXED];
  jsval *jsvec = (n > MAX_FIXED) ? new jsval[n] : jsfixed;

  int i;
  for (i=0; i<n; ++i)
    {
      jsdouble *d = JS_NewDouble( cx, f[i] );
      if (d && JS_AddRoot( cx, d ))
	jsvec[i] = DOUBLE_TO_JSVAL(d);
      else
	break;
    }

  if (i == n)
    {
      JSObject *arr = JS_NewArrayObject( cx, n, jsvec );
      if (arr) *rval = OBJECT_TO_JSVAL(arr);
    }

  for (int j=0; j<i; ++j)
    JS_RemoveRoot( cx, JSVAL_TO_GCTHING(jsvec[j]) );
  if (n > MAX_FIXED) delete [] jsvec;
  return (i == n) ? JS_TRUE : JS_FALSE;
}

// Generic VrmlField finalize

static void
finalize(JSContext *cx, JSObject *obj)
{
  VrmlField *f = (VrmlField *)JS_GetPrivate( cx, obj );
  if (f) delete f;
  JS_SetPrivate( cx, obj, 0 );
}

// Generic VrmlField toString

static JSBool
toString(JSContext *cx, JSObject *obj, uintN, jsval *, jsval *rval)
{
  VrmlField *f = (VrmlField *)JS_GetPrivate( cx, obj );
  if (f)
    {
      ostringstream os;
      os << (*f);
      const char *ss = os.str().c_str();
      JSString *s = JS_NewStringCopyZ( cx, ss );
      if (s) { *rval = STRING_TO_JSVAL(s); return JS_TRUE; }
    }
  return JS_FALSE;
}

//
//  Classes precede the methods since the methods may access
//  objects of the various classes.
//

// SFColor class

static JSBool
SFColorCons(JSContext *cx, JSObject *obj,
	    uintN argc, jsval *argv, jsval *)
{
  jsdouble r = 0.0, g = 0.0, b = 0.0;
  for (unsigned int i=0; i<argc; ++i)
    switch (i) {
    case 0: JS_ValueToNumber( cx, argv[i], &r ); break;
    case 1: JS_ValueToNumber( cx, argv[i], &g ); break;
    case 2: JS_ValueToNumber( cx, argv[i], &b ); break;
    default: /* Too many args */ break;
    }

  VrmlSFColor *c = new VrmlSFColor( r, g, b );
  JS_SetPrivate(cx, obj, c);
  return JS_TRUE;
}


static JSBool
color_getProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
{
  VrmlSFColor *c = (VrmlSFColor *)JS_GetPrivate( cx, obj );

  if (c && JSVAL_IS_INT(id))
    switch (JSVAL_TO_INT(id)) {
    case 0: *vp = DOUBLE_TO_JSVAL(JS_NewDouble( cx, c->getR())); break;
    case 1: *vp = DOUBLE_TO_JSVAL(JS_NewDouble( cx, c->getG())); break;
    case 2: *vp = DOUBLE_TO_JSVAL(JS_NewDouble( cx, c->getB())); break;
    default: return JS_FALSE;
    }
  return JS_TRUE;
}


static JSBool
color_setProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
{
  VrmlSFColor * const sfColor =
          reinterpret_cast<VrmlSFColor *>(JS_GetPrivate(cx, obj));
  jsdouble d;

  if (sfColor && JSVAL_IS_INT(id) &&
      JSVAL_TO_INT(id) >= 0 && JSVAL_TO_INT(id) < 3 &&
      JS_ValueToNumber( cx, *vp, &d ))
    {
      (*sfColor)[JSVAL_TO_INT(id)] = d;
      checkEventOut(cx, obj, sfColor);
      return JS_TRUE;
    }

  return JS_FALSE;
}

static JSClass SFColorClass = {
  "SFColor", JSCLASS_HAS_PRIVATE,
  JS_PropertyStub,  JS_PropertyStub,  color_getProperty, color_setProperty,
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  finalize
};


// SFImage class
// This is ugly because the spec requires an MFInt array for the pixel data
// and because giving write access to the image size parameters can cause
// the library code to crash unless they are validated somehow...

static JSBool
SFImageCons(JSContext *cx, JSObject *obj,
	    uintN argc, jsval *argv, jsval *)
{
  int w = 0, h = 0, nc = 0;
  unsigned char *pixels = 0;
  jsdouble x;

  for (unsigned int i=0; i<argc; ++i)
    switch (i) {
    case 0: JS_ValueToNumber( cx, argv[i], &x ); w = (int) x; break;
    case 1: JS_ValueToNumber( cx, argv[i], &x ); h = (int) x; break;
    case 2: JS_ValueToNumber( cx, argv[i], &x ); nc = (int) x; break;
    case 3: /* MFInt image data */
      if ( JSVAL_IS_OBJECT(argv[i]) &&
	   JS_IsArrayObject( cx, JSVAL_TO_OBJECT(argv[i])) )
	{
	  jsuint len;
	  JS_GetArrayLength( cx, JSVAL_TO_OBJECT(argv[i]), &len );
	  if ((int)len != w*h)
	    return JS_FALSE;
	  pixels = new unsigned char[ nc*w*h ];
	  unsigned char *pp = pixels;
	  for (int ii=0; ii<(int)len; ++ii, pp+=nc)
	    {
	      static int byteMask[] = { 0x000000FF, 0x0000FF00, 0x00FF0000, 0xFF000000 };
	      jsval elt;
	      JS_GetElement( cx, JSVAL_TO_OBJECT(argv[i]), (jsint) ii, &elt );
	      if (JSVAL_IS_NUMBER(elt)) {
		jsdouble factor;
		JS_ValueToNumber( cx, elt, &factor );
		for (int j=0; j<nc; ++j)
		  pp[j] = (((int) factor) & byteMask[j]) >> (8*j);
	      }
	    }
	}
      break;
    default: /* Too many args */ break;
    }

  VrmlSFImage *c = new VrmlSFImage( w, h, nc, pixels );
  delete[] pixels;
  JS_SetPrivate(cx, obj, c);
  return JS_TRUE;
}


static JSBool
image_getProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
{
  VrmlSFImage *c = (VrmlSFImage *)JS_GetPrivate( cx, obj );
  if (! c) return JS_FALSE;
  if (JSVAL_IS_INT(id))
    switch (JSVAL_TO_INT(id)) {
    case 0: *vp = INT_TO_JSVAL( c->getWidth() ); break;
    case 1: *vp = INT_TO_JSVAL( c->getHeight() ); break;
    case 2: *vp = INT_TO_JSVAL( c->getComponents() ); break;
    case 3: break; // *vp = convert pixels to MFInt...
    default: return JS_FALSE;
    }

  return JS_TRUE;
}


static JSBool
image_setProperty(JSContext *, JSObject *, jsval , jsval *)
{
  // ...
  return JS_FALSE;
}

static JSClass SFImageClass = {
  "SFImage", JSCLASS_HAS_PRIVATE,
  JS_PropertyStub,  JS_PropertyStub,  image_getProperty, image_setProperty,
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  finalize
};


// SFNode class

static JSBool
SFNodeCons(JSContext *cx, JSObject *obj,
	   uintN argc, jsval *argv, jsval *)
{
  JSString *str;

  if (argc == 1 && JSVAL_IS_STRING(argv[0]) &&
      (str = JSVAL_TO_STRING(argv[0])) != 0)
    {
      // If the nodes created contain scripts, they might want to access the
      // namespace...
      VrmlNamespace ns;
      VrmlMFNode* nodes;
      try { nodes = VrmlScene::readString(JS_GetStringBytes(str), &ns); }
      catch (VrmlLoadingError &error)
      {
        theSystem->error(System::ERRORTYPE_UNKNOWN, "Javascript: Error reading VRML string: %s",
                         error.what());
        return JS_FALSE;
      }
      if (!nodes || nodes->getLength() == 0) return JS_FALSE;

      // If there are multiple top-level nodes, wrap them in a Group. SPEC?
      VrmlNode *n;
      if (nodes->getLength() == 1)
	n = (*nodes)[0];
      else
	{
	  VrmlNodeGroup *g = new VrmlNodeGroup();
	  g->addChildren( *nodes);
	  n = g;
	}

      VrmlSFNode *sfnode = new VrmlSFNode(n);
      JS_SetPrivate(cx, obj, sfnode);
      delete nodes;
      return JS_TRUE;
    }

  return JS_FALSE;
}

// SFNode getProperty reads eventOut values, setProperty sends eventIns.

static JSBool
node_getProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
{
  VrmlSFNode *n = (VrmlSFNode *)JS_GetPrivate( cx, obj );
  JSString *str = 0;
  ScriptJS *script = 0;

  if (n && n->get() &&
      JSVAL_IS_STRING(id) && ((str = JSVAL_TO_STRING(id)) != 0) &&
      (script = objToScript( cx, JS_GetParent( cx, obj ))) != 0)
    {
      char *eventOut = JS_GetStringBytes(str);
      const VrmlField *fieldVal = n->get()->getEventOut( eventOut );

      // convert event out value to jsval...
      if (fieldVal)
	{
	  *vp = script->vrmlFieldToJSVal(fieldVal->fieldType(), fieldVal, false);
	  //return JS_TRUE;
	}

      // If JS_FALSE is returned, apparently the parent object is not searched for
      // the property, so stuff like toString() will fail... don't know how to
      // distinguish between a valid prop and an invalid eventOut...
      // theSystem->error("unrecognized exposedField/eventOut:node.getProperty(%s)\n",
      //                 JS_GetStringBytes(str));
      return JS_TRUE;
    }

  return JS_FALSE;
}

static JSBool
node_setProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
{
  VrmlSFNode *sfn = (VrmlSFNode *)JS_GetPrivate( cx, obj );
  VrmlNode *n = sfn ? sfn->get() : 0;

  if (n && JSVAL_IS_STRING(id))
    {
      JSString *str = JSVAL_TO_STRING(id);
      char *eventIn = str ? JS_GetStringBytes(str) : 0;
      VrmlField::VrmlFieldType expect;
      expect = n->nodeType().hasEventIn( eventIn );

      if (expect && eventIn)	// convert vp to field, send eventIn to n
	{

	  VrmlField *f = jsvalToVrmlField( cx, *vp, expect );
	  if ( ! f ) return JS_FALSE;

	  // This should only happen if directOutput is set...
#ifndef __ORBISNAP__  // disabled because it uses cout instead of theSystem
#ifdef SCRIPTJS_DEBUG
	  cout << "ScriptJS::node_setProperty sending " << eventIn
	       << " (" << (*f) << ") to "
	       << n->nodeType().getName() << "::"
	       << n->name() << endl;
#endif
#endif

	  // the timestamp should be stored as a global property and
	  // looked up via obj somehow...
	  n->eventIn(s_timeStamp, eventIn, f);
	  delete f;
	}

      checkEventOut( cx, obj, sfn );
    }
  return JS_TRUE;
}

static JSClass SFNodeClass = {
  "SFNode", JSCLASS_HAS_PRIVATE,
  JS_PropertyStub,  JS_PropertyStub,  node_getProperty, node_setProperty,
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  finalize
};



// SFVec2f class

static JSBool
SFVec2fCons(JSContext *cx, JSObject *obj,
	    uintN argc, jsval *argv, jsval *)
{
  jsdouble x = 0.0, y = 0.0;
  if (argc > 0) JS_ValueToNumber( cx, argv[0], &x );
  if (argc > 1) JS_ValueToNumber( cx, argv[1], &y );
  VrmlSFVec2f *v = new VrmlSFVec2f( x, y );
  JS_SetPrivate(cx, obj, v);
  return JS_TRUE;
}


static JSBool
vec2f_getProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
{
  VrmlSFVec2f *v = (VrmlSFVec2f *)JS_GetPrivate( cx, obj );
  if (v && JSVAL_IS_INT(id))
    switch (JSVAL_TO_INT(id)) {
    case 0: *vp = DOUBLE_TO_JSVAL(JS_NewDouble( cx, v->getX())); break;
    case 1: *vp = DOUBLE_TO_JSVAL(JS_NewDouble( cx, v->getY())); break;
    default: return JS_FALSE;
    }
  return JS_TRUE;
}

static JSBool
vec2f_setProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
{
  VrmlSFVec2f * const sfVec2f =
          reinterpret_cast<VrmlSFVec2f *>(JS_GetPrivate(cx, obj));
  jsdouble d;

  if (sfVec2f && JSVAL_IS_INT(id) &&
      JSVAL_TO_INT(id) >= 0 && JSVAL_TO_INT(id) < 2 &&
      JS_ValueToNumber( cx, *vp, &d ))
    {
      (*sfVec2f)[JSVAL_TO_INT(id)] = d;
      checkEventOut(cx, obj, sfVec2f);
      return JS_TRUE;
    }

  return JS_FALSE;
}

static JSClass SFVec2fClass = {
  "SFVec2f", JSCLASS_HAS_PRIVATE,
  JS_PropertyStub,  JS_PropertyStub,  vec2f_getProperty, vec2f_setProperty,
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  finalize
};


// SFVec3f class

static JSBool
SFVec3fCons(JSContext *cx, JSObject *obj,
	    uintN argc, jsval *argv, jsval *)
{
  jsdouble x = 0.0, y = 0.0, z = 0.0;
  if (argc > 0) JS_ValueToNumber( cx, argv[0], &x );
  if (argc > 1) JS_ValueToNumber( cx, argv[1], &y );
  if (argc > 2) JS_ValueToNumber( cx, argv[2], &z );
  VrmlSFVec3f *v = new VrmlSFVec3f( x, y, z );
  JS_SetPrivate(cx, obj, v);
  return JS_TRUE;
}


static JSBool
vec3f_getProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
{
  VrmlSFVec3f *v = (VrmlSFVec3f *)JS_GetPrivate( cx, obj );
  if (v && JSVAL_IS_INT(id))
    switch (JSVAL_TO_INT(id)) {
    case 0: *vp = DOUBLE_TO_JSVAL(JS_NewDouble( cx, v->getX())); break;
    case 1: *vp = DOUBLE_TO_JSVAL(JS_NewDouble( cx, v->getY())); break;
    case 2: *vp = DOUBLE_TO_JSVAL(JS_NewDouble( cx, v->getZ())); break;
    default: return JS_FALSE;
    }
  return JS_TRUE;
}

static JSBool
vec3f_setProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
{
  VrmlSFVec3f *v = (VrmlSFVec3f *)JS_GetPrivate( cx, obj );
  jsdouble d;

  if (v && JSVAL_IS_INT(id) &&
      JSVAL_TO_INT(id) >= 0 && JSVAL_TO_INT(id) < 3 &&
      JS_ValueToNumber( cx, *vp, &d ))
    {
      (*v)[JSVAL_TO_INT(id)] = d;
      checkEventOut( cx, obj, v );
      return JS_TRUE;
    }

  return JS_FALSE;
}

static JSClass SFVec3fClass = {
  "SFVec3f", JSCLASS_HAS_PRIVATE,
  JS_PropertyStub,  JS_PropertyStub,  vec3f_getProperty, vec3f_setProperty,
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  finalize
};


// SFRotation class (needs to be after SFVec3f)

static JSBool
SFRotationCons(JSContext *cx, JSObject *obj,
	       uintN argc, jsval *argv, jsval *)
{
  jsdouble x = 0.0, y = 1.0, z = 0.0, angle = 0.0;

  if (argc > 0)
    {
      if ( JSVAL_IS_NUMBER(argv[0]) )
	{
	  JS_ValueToNumber( cx, argv[0], &x );
	  if (argc > 1) JS_ValueToNumber( cx, argv[1], &y );
	  if (argc > 2) JS_ValueToNumber( cx, argv[2], &z );
	  if (argc > 3) JS_ValueToNumber( cx, argv[3], &angle );
	}
      else if (argc == 2 &&
	       JSVAL_IS_OBJECT(argv[0]) &&
	       &SFVec3fClass == JS_GetClass( JSVAL_TO_OBJECT(argv[0]) ))
	{
	  JSObject *v1obj = JSVAL_TO_OBJECT(argv[0]);
	  VrmlSFVec3f *v1 = (VrmlSFVec3f *)JS_GetPrivate( cx, v1obj );
	  if (JSVAL_IS_NUMBER( argv[1] ))
	    {
	      x = v1->getX();	// axis/angle
	      y = v1->getY();
	      z = v1->getZ();
	      JS_ValueToNumber( cx, argv[1], &angle );
	    }
	  else if ( JSVAL_IS_OBJECT(argv[1]) &&
		    &SFVec3fClass == JS_GetClass( JSVAL_TO_OBJECT(argv[1]) ))
	    {
	      JSObject *v2obj = JSVAL_TO_OBJECT(argv[1]);
	      VrmlSFVec3f *v2 = (VrmlSFVec3f *)JS_GetPrivate( cx, v2obj );
	      VrmlSFVec3f v3(*v1);  // find rotation that takes v1 to v2
	      v3 = v3.cross(*v2);
	      x = v3.getX();
	      y = v3.getY();
	      z = v3.getZ();
	      angle = acos( v1->dot(*v2) / (v1->length() * v2->length()) );
	    }
	  else
	    return JS_FALSE;
	}
      else
	return JS_FALSE;
    }

  VrmlSFRotation *v = new VrmlSFRotation( x, y, z, angle );
  JS_SetPrivate(cx, obj, v);
  return JS_TRUE;
}

static JSBool
rot_getProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
{
  VrmlSFRotation *v = (VrmlSFRotation *)JS_GetPrivate( cx, obj );
  if (v && JSVAL_IS_INT(id))
    switch (JSVAL_TO_INT(id)) {
    case 0: *vp = DOUBLE_TO_JSVAL(JS_NewDouble( cx, v->getX())); break;
    case 1: *vp = DOUBLE_TO_JSVAL(JS_NewDouble( cx, v->getY())); break;
    case 2: *vp = DOUBLE_TO_JSVAL(JS_NewDouble( cx, v->getZ())); break;
    case 3: *vp = DOUBLE_TO_JSVAL(JS_NewDouble( cx, v->getAngle())); break;
    default: return JS_FALSE;
    }
  return JS_TRUE;
}

static JSBool
rot_setProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
{
  VrmlSFRotation *v = (VrmlSFRotation *)JS_GetPrivate( cx, obj );
  jsdouble d;

  if (v && JSVAL_IS_INT(id) &&
      JSVAL_TO_INT(id) >= 0 && JSVAL_TO_INT(id) < 4 &&
      JS_ValueToNumber( cx, *vp, &d ))
    {
      (*v)[ JSVAL_TO_INT(id) ] = d;
      checkEventOut( cx, obj, v );
      return JS_TRUE;
    }

  return JS_FALSE;
}

static JSClass SFRotationClass = {
  "SFRotation", JSCLASS_HAS_PRIVATE,
  JS_PropertyStub,  JS_PropertyStub,  rot_getProperty, rot_setProperty,
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  finalize
};

// MFColor class

#if defined(__ORBISNAP__)

static JSBool
MFColorCons(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
  // create the internal VrmlMFColor object
  VrmlMFColor *v = new VrmlMFColor();
  JS_SetPrivate(cx, obj, v);

  // load the initial values
  int i;
  v->setLength(argc);
  for (i=0; i<argc; i++)
  {
    if (!JSVAL_IS_OBJECT(argv[i])) return JS_FALSE;
    JSObject *val = JSVAL_TO_OBJECT(argv[i]);
    if (!val) return JS_FALSE;
    if (strcmp(JS_GetClass(val)->name, "SFColor") != 0) return JS_FALSE;
    VrmlSFColor *vv = (VrmlSFColor *) JS_GetPrivate(cx, val);
    const float *data = vv->get();
    v->setElement(i, data[0], data[1], data[2]);
  }
  return JS_TRUE;
}

static JSBool
mfcolor_setProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
{
  // get the internal MFColor object
  VrmlMFColor *v = (VrmlMFColor *)JS_GetPrivate( cx, obj );
  if (!v) return JS_FALSE;

  // are we setting an indexed value?
  if (JSVAL_IS_INT(id))
  {
    int index = JSVAL_TO_INT(id);
    if (index < 0) return JS_FALSE;

    // get the value to be set
    if (!JSVAL_TO_OBJECT(*vp)) return JS_FALSE;
    JSObject *val = JSVAL_TO_OBJECT(*vp);
    if (!val) return JS_FALSE;
    if (strcmp(JS_GetClass(val)->name, "SFColor") != 0) return JS_FALSE;

    // set the value
    VrmlSFColor *vv = (VrmlSFColor *) JS_GetPrivate(cx, val);
    const float *data = vv->get();
    v->setElement(index, data[0], data[1], data[2]);

    // if this is a VRML field, send an appropriate event
    checkEventOut(cx, obj, v);
    return JS_TRUE;
  }

  // are we setting a named property?
  else if (JSVAL_IS_STRING(id))
  {
    JSString *str = JSVAL_TO_STRING(id);
    char *propname = JS_GetStringBytes(str);
    if (strcmp(propname, "length") == 0)
    {
      if (!JSVAL_IS_INT(*vp)) return JS_FALSE;
      else v->setLength(JSVAL_TO_INT(*vp));
      return JS_TRUE;
    }
  }

  // unrecognized property
  return JS_FALSE;
}

static JSBool
mfcolor_getProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
{
  // get the internal MFColor object
  VrmlMFColor *v = (VrmlMFColor *)JS_GetPrivate( cx, obj );
  if (!v) return JS_FALSE;

  // are we reading an indexed value?
  if (JSVAL_IS_INT(id))
  {
    int index = JSVAL_TO_INT(id);
    if (index < 0) return JS_FALSE;

    // read the values and return them as a new SFColor object
    JSObject *retval = JS_NewObject(cx, &SFColorClass, 0,
      JS_GetParent(cx, obj));
    if (index < v->getLength())
      JS_SetPrivate(cx, retval, new VrmlSFColor(v->get() + 3*index));
    else
      JS_SetPrivate(cx, retval, new VrmlSFColor(0.0, 0.0, 0.0));
    *vp = OBJECT_TO_JSVAL(retval);
    return JS_TRUE;
  }

  // are we reading a named property?
  else if (JSVAL_IS_STRING(id))
  {
    JSString *str = JSVAL_TO_STRING(id);
    char *propname = JS_GetStringBytes(str);
    if (strcmp(propname, "length") == 0)
    {
       *vp = INT_TO_JSVAL(v->getLength());
       return JS_TRUE;
    }
  }

  // unrecognized property
  return JS_FALSE;
}

static JSClass MFColorClass = {
  "MFColor", JSCLASS_HAS_PRIVATE,
  JS_PropertyStub,  JS_PropertyStub,  mfcolor_getProperty,  mfcolor_setProperty,
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  finalize,
};

#else

static JSBool
MFColorCons(JSContext *cx, JSObject *obj,
	    uintN argc, jsval *argv, jsval *rval)
{
  obj = JS_NewArrayObject(cx, (jsint) argc, argv);
  *rval = OBJECT_TO_JSVAL(obj);
  return obj == 0 ? JS_FALSE : JS_TRUE;
}

static JSClass MFColorClafss = {
  "MFColor", 0,
  JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  JS_FinalizeStub,
};

#endif

// MFFloat class

#if defined(__ORBISNAP__)

static JSBool
MFFloatCons(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
  // create the internal VrmlMFColor object
  VrmlMFFloat *v = new VrmlMFFloat();
  JS_SetPrivate(cx, obj, v);

  // load the initial values
  int i;
  v->setLength(argc);
  for (i=0; i<argc; i++)
  {
    double x;
    if (JSVAL_IS_DOUBLE(argv[i]))   x = (double) *(JSVAL_TO_DOUBLE(argv[i]));
    else if (JSVAL_IS_INT(argv[i])) x = (double) JSVAL_TO_INT(argv[i]);
    else return JS_FALSE;
    v->setElement(i, x);
  }
  return JS_TRUE;
}

static JSBool
mffloat_setProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
{
  // get the internal MFFloat object
  VrmlMFFloat *v = (VrmlMFFloat *) JS_GetPrivate( cx, obj );
  if (!v) return JS_FALSE;

  // are we setting an indexed value?
  if (JSVAL_IS_INT(id))
  {
    int index = JSVAL_TO_INT(id);
    if (index < 0) return JS_FALSE;

    // get the value to be set
    double x;
    if (JSVAL_IS_DOUBLE(*vp))   x = (double) *(JSVAL_TO_DOUBLE(*vp));
    else if (JSVAL_IS_INT(*vp)) x = (double) JSVAL_TO_INT(*vp);
    else return JS_FALSE;

    // set the value
    v->setElement(index, x);

    // if this is a VRML field, send an appropriate event
    checkEventOut(cx, obj, v);
    return JS_TRUE;
  }

  // are we setting a named property?
  else if (JSVAL_IS_STRING(id))
  {
    JSString *str = JSVAL_TO_STRING(id);
    char *propname = JS_GetStringBytes(str);
    if (strcmp(propname, "length") == 0)
    {
      if (!JSVAL_IS_INT(*vp)) return JS_FALSE;
      else v->setLength(JSVAL_TO_INT(*vp));
      return JS_TRUE;
    }
  }

  // unrecognized property
  return JS_FALSE;
}

static JSBool
mffloat_getProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
{
  // get the internal MFFloat object
  VrmlMFFloat *v = (VrmlMFFloat *) JS_GetPrivate(cx, obj);
  if (!v) return JS_FALSE;

  // are we reading an indexed value?
  if (JSVAL_IS_INT(id))
  {
    int index = JSVAL_TO_INT(id);
    if (index < 0) return JS_FALSE;

    // read the values and return them as a new double value
    double x = (index < v->getLength()) ? *(v->get() + index) : 0.0;
    JS_NewDoubleValue(cx, x, vp);
    return JS_TRUE;
  }

  // are we reading a named property?
  else if (JSVAL_IS_STRING(id))
  {
    JSString *str = JSVAL_TO_STRING(id);
    char *propname = JS_GetStringBytes(str);
    if (strcmp(propname, "length") == 0)
    {
       *vp = INT_TO_JSVAL(v->getLength());
       return JS_TRUE;
    }
  }

  // unrecognized property
  return JS_FALSE;
}

static JSClass MFFloatClass = {
  "MFFloat", JSCLASS_HAS_PRIVATE,
  JS_PropertyStub,  JS_PropertyStub,  mffloat_getProperty,  mffloat_setProperty,
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  finalize,
};

#else

static JSBool
MFFloatCons(JSContext *cx, JSObject *obj,
	    uintN argc, jsval *argv, jsval *rval)
{
  obj = JS_NewArrayObject(cx, (jsint) argc, argv);
  *rval = OBJECT_TO_JSVAL(obj);
  return obj == 0 ? JS_FALSE : JS_TRUE;
}

static JSClass MFFloatClass = {
  "MFFloat", 0,
  JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  JS_FinalizeStub,
};

#endif

// MFInt32 class

#if defined(__ORBISNAP__)

static JSBool
MFInt32Cons(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
  // create the internal VrmlMFColor object
  VrmlMFInt32 *v = new VrmlMFInt32();
  JS_SetPrivate(cx, obj, v);

  // load the initial values
  int i;
  v->setLength(argc);
  for (i=0; i<argc; i++)
  {
    int n;
    if (JSVAL_IS_DOUBLE(argv[i]))   n = (int) *(JSVAL_TO_DOUBLE(argv[i]));
    else if (JSVAL_IS_INT(argv[i])) n = (int) JSVAL_TO_INT(argv[i]);
    else return JS_FALSE;
    v->setElement(i, n);
  }
  return JS_TRUE;
}

static JSBool
mfint32_setProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
{
  // get the internal MFFloat object
  VrmlMFInt32 *v = (VrmlMFInt32 *) JS_GetPrivate( cx, obj );
  if (!v) return JS_FALSE;

  // are we setting an indexed value?
  if (JSVAL_IS_INT(id))
  {
    int index = JSVAL_TO_INT(id);
    if (index < 0) return JS_FALSE;

    // get the value to be set
    int n;
    if (JSVAL_IS_DOUBLE(*vp))   n = (int) *(JSVAL_TO_DOUBLE(*vp));
    else if (JSVAL_IS_INT(*vp)) n = (int) JSVAL_TO_INT(*vp);
    else return JS_FALSE;

    // set the value
    v->setElement(index, n);

    // if this is a VRML field, send an appropriate event
    checkEventOut(cx, obj, v);
    return JS_TRUE;
  }

  // are we setting a named property?
  else if (JSVAL_IS_STRING(id))
  {
    JSString *str = JSVAL_TO_STRING(id);
    char *propname = JS_GetStringBytes(str);
    if (strcmp(propname, "length") == 0)
    {
      if (!JSVAL_IS_INT(*vp)) return JS_FALSE;
      else v->setLength(JSVAL_TO_INT(*vp));
      return JS_TRUE;
    }
  }

  // unrecognized property
  return JS_FALSE;
}

static JSBool
mfint32_getProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
{
  // get the internal MFFloat object
  VrmlMFInt32 *v = (VrmlMFInt32 *) JS_GetPrivate(cx, obj);
  if (!v) return JS_FALSE;

  // are we reading an indexed value?
  if (JSVAL_IS_INT(id))
  {
    int index = JSVAL_TO_INT(id);
    if (index < 0) return JS_FALSE;

    // read the value and return it as a new integer value
    int n = (index < v->getLength()) ? *(v->get() + index) : 0;
    *vp = INT_TO_JSVAL(n);
    return JS_TRUE;
  }

  // are we reading a named property?
  else if (JSVAL_IS_STRING(id))
  {
    JSString *str = JSVAL_TO_STRING(id);
    char *propname = JS_GetStringBytes(str);
    if (strcmp(propname, "length") == 0)
    {
       *vp = INT_TO_JSVAL(v->getLength());
       return JS_TRUE;
    }
  }

  // unrecognized property
  return JS_FALSE;
}

static JSClass MFInt32Class = {
  "MFInt32", JSCLASS_HAS_PRIVATE,
  JS_PropertyStub,  JS_PropertyStub,  mfint32_getProperty,  mfint32_setProperty,
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  finalize,
};

#else

static JSBool
MFInt32Cons(JSContext *cx, JSObject *obj,
	    uintN argc, jsval *argv, jsval *rval)
{
  obj = JS_NewArrayObject(cx, (jsint) argc, argv);
  *rval = OBJECT_TO_JSVAL(obj);
  return obj == 0 ? JS_FALSE : JS_TRUE;
}

static JSClass MFInt32Class = {
  "MFInt32", 0,
  JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  JS_FinalizeStub,
};

#endif

// MFNode class

static JSBool
MFNodeCons(JSContext *cx, JSObject *obj,
	   uintN argc, jsval *argv, jsval *rval)
{
  obj = JS_NewArrayObject(cx, (jsint) argc, argv);
  *rval = OBJECT_TO_JSVAL(obj);
  return obj == 0 ? JS_FALSE : JS_TRUE;
}

static JSClass MFNodeClass = {
  "MFNode", 0,
  JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  JS_FinalizeStub,
};

// MFRotation class

#if defined(__ORBISNAP__)

static JSBool
MFRotationCons(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
  jsval *rval)
{
  // create the internal VrmlMFRotation object
  VrmlMFRotation *v = new VrmlMFRotation();
  JS_SetPrivate(cx, obj, v);

  // load the initial values
  int i;
  v->setLength(argc);
  for (i=0; i<argc; i++)
  {
    if (!JSVAL_IS_OBJECT(argv[i])) return JS_FALSE;
    JSObject *val = JSVAL_TO_OBJECT(argv[i]);
    if (!val) return JS_FALSE;
    if (strcmp(JS_GetClass(val)->name, "SFRotation") != 0) return JS_FALSE;
    VrmlSFRotation *vv = (VrmlSFRotation *) JS_GetPrivate(cx, val);
    const float *data = vv->get();
    v->setElement(i, data[0], data[1], data[2], data[3]);
  }
  return JS_TRUE;
}

static JSBool
mfrotation_setProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
{
  // get the internal MFRotation object
  VrmlMFRotation *v = (VrmlMFRotation *)JS_GetPrivate( cx, obj );
  if (!v) return JS_FALSE;

  // are we setting an indexed value?
  if (JSVAL_IS_INT(id))
  {
    int index = JSVAL_TO_INT(id);
    if (index < 0) return JS_FALSE;

    // get the value to be set
    if (!JSVAL_TO_OBJECT(*vp)) return JS_FALSE;
    JSObject *val = JSVAL_TO_OBJECT(*vp);
    if (!val) return JS_FALSE;
    if (strcmp(JS_GetClass(val)->name, "SFRotation") != 0) return JS_FALSE;

    // set the value
    VrmlSFRotation *vv = (VrmlSFRotation *) JS_GetPrivate(cx, val);
    const float *data = vv->get();
    v->setElement(index, data[0], data[1], data[2], data[3]);

    // if this is a VRML field, send an appropriate event
    checkEventOut(cx, obj, v);
    return JS_TRUE;
  }

  // are we setting a named property?
  else if (JSVAL_IS_STRING(id))
  {
    JSString *str = JSVAL_TO_STRING(id);
    char *propname = JS_GetStringBytes(str);
    if (strcmp(propname, "length") == 0)
    {
      if (!JSVAL_IS_INT(*vp)) return JS_FALSE;
      else v->setLength(JSVAL_TO_INT(*vp));
      return JS_TRUE;
    }
  }

  // unrecognized property
  return JS_FALSE;
}

static JSBool
mfrotation_getProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
{
  // get the internal MFRotation object
  VrmlMFRotation *v = (VrmlMFRotation *)JS_GetPrivate( cx, obj );
  if (!v) return JS_FALSE;

  // are we reading an indexed value?
  if (JSVAL_IS_INT(id))
  {
    int index = JSVAL_TO_INT(id);
    if (index < 0) return JS_FALSE;

    // read the values and return them as a new SFRotation object
    JSObject *retval = JS_NewObject(cx, &SFRotationClass, 0,
      JS_GetParent(cx, obj));
    if (index < v->getLength())
      JS_SetPrivate(cx, retval, new VrmlSFRotation(v->get() + 4*index));
    else
      JS_SetPrivate(cx, retval, new VrmlSFRotation(0.0, 1.0, 0.0, 0.0));
    *vp = OBJECT_TO_JSVAL(retval);
    return JS_TRUE;
  }

  // are we reading a named property?
  else if (JSVAL_IS_STRING(id))
  {
    JSString *str = JSVAL_TO_STRING(id);
    char *propname = JS_GetStringBytes(str);
    if (strcmp(propname, "length") == 0)
    {
       *vp = INT_TO_JSVAL(v->getLength());
       return JS_TRUE;
    }
  }

  // unrecognized property
  return JS_FALSE;
}

static JSClass MFRotationClass = {
  "MFRotation", JSCLASS_HAS_PRIVATE,
  JS_PropertyStub,  JS_PropertyStub,  mfrotation_getProperty,  mfrotation_setProperty,
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  finalize,
};

#else

static JSBool
MFRotationCons(JSContext *cx, JSObject *obj,
	       uintN argc, jsval *argv, jsval *rval)
{
  obj = JS_NewArrayObject(cx, (jsint) argc, argv);
  *rval = OBJECT_TO_JSVAL(obj);
  return obj == 0 ? JS_FALSE : JS_TRUE;
}

static JSClass MFRotationClass = {
  "MFRotation", 0,
  JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  JS_FinalizeStub,
};

#endif

// MFString class

#if defined(__ORBISNAP__)

static JSBool
MFStringCons(JSContext *cx, JSObject *obj,
	     uintN argc, jsval *argv, jsval *rval)
{
  char **s = new char*[ argc ];
  // load the initial values
  int i;
  for (i=0; i<argc; i++)
  {
    JSString *str = 0;
    if (JSVAL_IS_STRING(argv[i]) && (str = JSVAL_TO_STRING(argv[i])) != 0)
      s[i] = JS_GetStringBytes(str);
    else
      s[i] = 0;
  }

  VrmlMFString* v = new VrmlMFString( argc, s );
  delete [] s;
  JS_SetPrivate(cx, obj, v);
  return JS_TRUE;
}

static JSBool
mfstring_setProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
{
  // get the internal MFFloat object
  VrmlMFString *v = (VrmlMFString *) JS_GetPrivate( cx, obj );
  if (!v) return JS_FALSE;

  // are we setting an indexed value?
  if (JSVAL_IS_INT(id))
  {
    int index = JSVAL_TO_INT(id);
    if (index < 0) return JS_FALSE;

    // get the value to be set
    JSString *str = JS_ValueToString(cx,*vp);
    if (!str) return JS_FALSE;
    const char* n = JS_GetStringBytes(str);

    // set the value
    v->set(index, n);

    // if this is a VRML field, send an appropriate event
    checkEventOut(cx, obj, v);
    return JS_TRUE;
  }

  // are we setting a named property?
  else if (JSVAL_IS_STRING(id))
  {
    JSString *str = JSVAL_TO_STRING(id);
    char *propname = JS_GetStringBytes(str);
    if (strcmp(propname, "length") == 0)
    {
      if (!JSVAL_IS_INT(*vp)) return JS_FALSE;
      else v->set(JSVAL_TO_INT(*vp)-1, static_cast<const char*> (0)); //;!!
      return JS_TRUE;
    }
  }

  // unrecognized property
  return JS_FALSE;
}

static JSBool
mfstring_getProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
{
  // get the internal MFFloat object
  VrmlMFString *v = (VrmlMFString *) JS_GetPrivate(cx, obj);
  if (!v) return JS_FALSE;

  // are we reading an indexed value?
  if (JSVAL_IS_INT(id))
  {
    int index = JSVAL_TO_INT(id);
    if (index < 0) return JS_FALSE;

    // read the value and return it as a new integer value
    const char* n = (index < v->getLength()) ? *(v->get() + index) : 0;
    *vp = STRING_TO_JSVAL(n);
    return JS_TRUE;
  }

  // are we reading a named property?
  else if (JSVAL_IS_STRING(id))
  {
    JSString *str = JSVAL_TO_STRING(id);
    char *propname = JS_GetStringBytes(str);
    if (strcmp(propname, "length") == 0)
    {
       *vp = INT_TO_JSVAL(v->getLength());
       return JS_TRUE;
    }
  }

  // unrecognized property
  return JS_FALSE;
}

static JSClass MFStringClass = {
  "MFString", JSCLASS_HAS_PRIVATE,
  JS_PropertyStub,  JS_PropertyStub,  mfstring_getProperty,  mfstring_setProperty,
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  JS_FinalizeStub,
};

#else

static JSBool
MFStringCons(JSContext *cx, JSObject *obj,
	     uintN argc, jsval *argv, jsval *rval)
{
  obj = JS_NewArrayObject(cx, (jsint) argc, argv);
  *rval = OBJECT_TO_JSVAL(obj);
  return obj == 0 ? JS_FALSE : JS_TRUE;
}

static JSClass MFStringClass = {
  "MFString", 0,
  JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  JS_FinalizeStub,
};

#endif

// MFVec2f class

#if defined(__ORBISNAP__)

static JSBool
MFVec2fCons(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
  // create the internal VrmlMFVec2f object
  VrmlMFVec2f *v = new VrmlMFVec2f();
  JS_SetPrivate(cx, obj, v);

  // load the initial values
  int i;
  v->setLength(argc);
  for (i=0; i<argc; i++)
  {
    if (!JSVAL_IS_OBJECT(argv[i])) return JS_FALSE;
    JSObject *val = JSVAL_TO_OBJECT(argv[i]);
    if (!val) return JS_FALSE;
    if (strcmp(JS_GetClass(val)->name, "SFVec2f") != 0) return JS_FALSE;
    VrmlSFVec2f *vv = (VrmlSFVec2f *) JS_GetPrivate(cx, val);
    const float *data = vv->get();
    v->setElement(i, data[0], data[1]);
  }
  return JS_TRUE;
}

static JSBool
mfvec2f_setProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
{
  // get the internal MFVec2f object
  VrmlMFVec2f *v = (VrmlMFVec2f *)JS_GetPrivate( cx, obj );
  if (!v) return JS_FALSE;

  // are we setting an indexed value?
  if (JSVAL_IS_INT(id))
  {
    int index = JSVAL_TO_INT(id);
    if (index < 0) return JS_FALSE;

    // get the value to be set
    if (!JSVAL_TO_OBJECT(*vp)) return JS_FALSE;
    JSObject *val = JSVAL_TO_OBJECT(*vp);
    if (!val) return JS_FALSE;
    if (strcmp(JS_GetClass(val)->name, "SFVec2f") != 0) return JS_FALSE;

    // set the value
    VrmlSFVec2f *vv = (VrmlSFVec2f *) JS_GetPrivate(cx, val);
    const float *data = vv->get();
    v->setElement(index, data[0], data[1]);

    // if this is a VRML field, send an appropriate event
    checkEventOut(cx, obj, v);
    return JS_TRUE;
  }

  // are we setting a named property?
  else if (JSVAL_IS_STRING(id))
  {
    JSString *str = JSVAL_TO_STRING(id);
    char *propname = JS_GetStringBytes(str);
    if (strcmp(propname, "length") == 0)
    {
      if (!JSVAL_IS_INT(*vp)) return JS_FALSE;
      else v->setLength(JSVAL_TO_INT(*vp));
      return JS_TRUE;
    }
  }

  // unrecognized property
  return JS_FALSE;
}

static JSBool
mfvec2f_getProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
{
  // get the internal MFVec2f object
  VrmlMFVec2f *v = (VrmlMFVec2f *)JS_GetPrivate( cx, obj );
  if (!v) return JS_FALSE;

  // are we reading an indexed value?
  if (JSVAL_IS_INT(id))
  {
    int index = JSVAL_TO_INT(id);
    if (index < 0) return JS_FALSE;

    // read the values and return them as a new SFVec2f object
    JSObject *retval = JS_NewObject(cx, &SFVec2fClass, 0,
      JS_GetParent(cx, obj));
    if (index < v->getLength())
      JS_SetPrivate(cx, retval, new VrmlSFVec2f(v->get() + 2*index));
    else
      JS_SetPrivate(cx, retval, new VrmlSFVec2f(0.0, 0.0));
    *vp = OBJECT_TO_JSVAL(retval);
    return JS_TRUE;
  }

  // are we reading a named property?
  else if (JSVAL_IS_STRING(id))
  {
    JSString *str = JSVAL_TO_STRING(id);
    char *propname = JS_GetStringBytes(str);
    if (strcmp(propname, "length") == 0)
    {
       *vp = INT_TO_JSVAL(v->getLength());
       return JS_TRUE;
    }
  }

  // unrecognized property
  return JS_FALSE;
}

static JSClass MFVec2fClass = {
  "MFVec2f", JSCLASS_HAS_PRIVATE,
  JS_PropertyStub,  JS_PropertyStub,  mfvec2f_getProperty,  mfvec2f_setProperty,
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  finalize,
};

#else

static JSBool
MFVec2fCons(JSContext *cx, JSObject *obj,
	     uintN argc, jsval *argv, jsval *rval)
{
  obj = JS_NewArrayObject(cx, (jsint) argc, argv);
  *rval = OBJECT_TO_JSVAL(obj);
  return obj == 0 ? JS_FALSE : JS_TRUE;
}

static JSClass MFVec2fClass = {
  "MFVec2f", 0,
  JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  JS_FinalizeStub,
};

#endif

// MFVec3f class

#if defined(__ORBISNAP__)

static JSBool
MFVec3fCons(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
  // create the internal VrmlMFVec3f object
  VrmlMFVec3f *v = new VrmlMFVec3f();
  JS_SetPrivate(cx, obj, v);

  // load the initial values
  int i;
  v->setLength(argc);
  for (i=0; i<argc; i++)
  {
    if (!JSVAL_IS_OBJECT(argv[i])) return JS_FALSE;
    JSObject *val = JSVAL_TO_OBJECT(argv[i]);
    if (!val) return JS_FALSE;
    if (strcmp(JS_GetClass(val)->name, "SFVec3f") != 0) return JS_FALSE;
    VrmlSFVec3f *vv = (VrmlSFVec3f *) JS_GetPrivate(cx, val);
    const float *data = vv->get();
    v->setElement(i, data[0], data[1], data[2]);
  }
  return JS_TRUE;
}

static JSBool
mfvec3f_setProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
{
  // get the internal MFVec3f object
  VrmlMFVec3f *v = (VrmlMFVec3f *)JS_GetPrivate( cx, obj );
  if (!v) return JS_FALSE;

  // are we setting an indexed value?
  if (JSVAL_IS_INT(id))
  {
    int index = JSVAL_TO_INT(id);
    if (index < 0) return JS_FALSE;

    // get the value to be set
    if (!JSVAL_TO_OBJECT(*vp)) return JS_FALSE;
    JSObject *val = JSVAL_TO_OBJECT(*vp);
    if (!val) return JS_FALSE;
    if (strcmp(JS_GetClass(val)->name, "SFVec3f") != 0) return JS_FALSE;

    // set the value
    VrmlSFVec3f *vv = (VrmlSFVec3f *) JS_GetPrivate(cx, val);
    const float *data = vv->get();
    v->setElement(index, data[0], data[1], data[2]);

    // if this is a VRML field, send an appropriate event
    checkEventOut(cx, obj, v);
    return JS_TRUE;
  }

  // are we setting a named property?
  else if (JSVAL_IS_STRING(id))
  {
    JSString *str = JSVAL_TO_STRING(id);
    char *propname = JS_GetStringBytes(str);
    if (strcmp(propname, "length") == 0)
    {
      if (!JSVAL_IS_INT(*vp)) return JS_FALSE;
      else v->setLength(JSVAL_TO_INT(*vp));
      return JS_TRUE;
    }
  }

  // unrecognized property
  return JS_FALSE;
}

static JSBool
mfvec3f_getProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
{
  // get the internal MFVec3f object
  VrmlMFVec3f *v = (VrmlMFVec3f *)JS_GetPrivate( cx, obj );
  if (!v) return JS_FALSE;

  // are we reading an indexed value?
  if (JSVAL_IS_INT(id))
  {
    int index = JSVAL_TO_INT(id);
    if (index < 0) return JS_FALSE;

    // read the values and return them as a new SFVec3f object
    JSObject *retval = JS_NewObject(cx, &SFVec3fClass, 0,
      JS_GetParent(cx, obj));
    if (index < v->getLength())
      JS_SetPrivate(cx, retval, new VrmlSFVec3f(v->get() + 3*index));
    else
      JS_SetPrivate(cx, retval, new VrmlSFVec3f(0.0, 0.0, 0.0));
    *vp = OBJECT_TO_JSVAL(retval);
    return JS_TRUE;
  }

  // are we reading a named property?
  else if (JSVAL_IS_STRING(id))
  {
    JSString *str = JSVAL_TO_STRING(id);
    char *propname = JS_GetStringBytes(str);
    if (strcmp(propname, "length") == 0)
    {
       *vp = INT_TO_JSVAL(v->getLength());
       return JS_TRUE;
    }
  }

  // unrecognized property
  return JS_FALSE;
}

static JSClass MFVec3fClass = {
  "MFVec3f", JSCLASS_HAS_PRIVATE,
  JS_PropertyStub,  JS_PropertyStub,  mfvec3f_getProperty,  mfvec3f_setProperty,
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  finalize,
};

#else


static JSBool
MFVec3fCons(JSContext *cx, JSObject *obj,
	     uintN argc, jsval *argv, jsval *rval)
{
  obj = JS_NewArrayObject(cx, (jsint) argc, argv);
  *rval = OBJECT_TO_JSVAL(obj);
  return obj == 0 ? JS_FALSE : JS_TRUE;
}

static JSClass MFVec3fClass = {
  "MFVec3f", 0,
  JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  JS_FinalizeStub,
};

#endif

// VrmlMatrix class ...


//
//  Methods
//

// SFColor methods

static JSBool
color_setHSV(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
	     jsval *)
{
  jsdouble h, s, v;
  VrmlSFColor *c = (VrmlSFColor *)JS_GetPrivate( cx, obj );
  if ( c && argc == 3 &&
       JS_ValueToNumber( cx, argv[0], &h ) &&
       JS_ValueToNumber( cx, argv[1], &s ) &&
       JS_ValueToNumber( cx, argv[2], &v ) )
    {
      c->setHSV( h, s, v );
      return JS_TRUE;
    }

  return JS_FALSE;
}

static JSBool
color_getHSV(JSContext *cx, JSObject *obj, uintN , jsval *,
	     jsval *rval)
{
  VrmlSFColor *c = (VrmlSFColor *)JS_GetPrivate( cx, obj );
  if ( c )
    {
      float hsv[3];
      c->getHSV(hsv);
      return floatsToJSArray( 3, hsv, cx, rval );
    }
  return JS_FALSE;
}

// SFRotation methods

static JSBool
rot_getAxis(JSContext *cx, JSObject *obj, uintN , jsval *,
	    jsval *rval)
{
    const VrmlSFRotation * const rot =
            reinterpret_cast<const VrmlSFRotation *>(JS_GetPrivate(cx, obj));
    assert(rot);

    JSObject* v3obj = 0;
    if ((v3obj = JS_NewObject(cx, &SFVec3fClass, 0, JS_GetParent( cx, obj )))!=0)
    {
      VrmlSFVec3f* v3 = new VrmlSFVec3f(rot->getX(), rot->getY(), rot->getZ());
      JS_SetPrivate( cx, v3obj, v3 );
      *rval = OBJECT_TO_JSVAL(v3obj);
      return JS_TRUE;
    }
    return JS_FALSE;
}

static JSBool
rot_inverse(JSContext *cx, JSObject *obj, uintN , jsval *,
	    jsval *rval)
{
  VrmlSFRotation *r = (VrmlSFRotation *)JS_GetPrivate( cx, obj );
  JSObject *robj = 0;

  if (r &&
      (robj = JS_NewObject( cx, &SFRotationClass, 0,
			    JS_GetParent( cx, obj ) )) != 0)
    {
      JS_SetPrivate(cx, robj, new VrmlSFRotation(r->inverse()));
      *rval = OBJECT_TO_JSVAL(robj);
      return JS_TRUE;
    }

  return JS_FALSE;
}

static JSBool
rot_multiply(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
	     jsval *rval)
{
  const VrmlSFRotation * const r1 =
          reinterpret_cast<const VrmlSFRotation *>(JS_GetPrivate(cx, obj));
  const VrmlSFRotation * r2 = 0;
  JSObject *r3obj = 0;

  // Verify that we have 2 SFRotation's && create a result
  if (r1 && argc > 0 &&
      JSVAL_IS_OBJECT(argv[0]) &&
      &SFRotationClass == JS_GetClass( JSVAL_TO_OBJECT(argv[0]) ) &&
      (r2 = reinterpret_cast<const VrmlSFRotation *>(JS_GetPrivate( cx, JSVAL_TO_OBJECT(argv[0])))) &&
      (((VrmlField*)r2)->fieldType() == VrmlField::SFROTATION) &&
      (r3obj = JS_NewObject( cx, &SFRotationClass, 0,
			     JS_GetParent( cx, obj ) )) != 0)
    {
      JS_SetPrivate(cx, r3obj, new VrmlSFRotation(r1->multiply(*r2)));
      *rval = OBJECT_TO_JSVAL(r3obj);
      return JS_TRUE;
    }

  return JS_FALSE;
}

static JSBool
rot_multVec(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
	    jsval *rval)
{
  VrmlSFRotation *r1 = (VrmlSFRotation *)JS_GetPrivate( cx, obj );
  VrmlSFVec3f *v2 = 0;
  JSObject *v3obj = 0;

  // Verify that we have an SFVec3f arg && create a result
  if (r1 && argc > 0 &&
      JSVAL_IS_OBJECT(argv[0]) &&
      &SFVec3fClass == JS_GetClass( JSVAL_TO_OBJECT(argv[0]) ) &&
      (v2 = (VrmlSFVec3f *)JS_GetPrivate( cx, JSVAL_TO_OBJECT(argv[0]))) &&
      (((VrmlField*)v2)->fieldType() == VrmlField::SFVEC3F) &&
      (v3obj = JS_NewObject( cx, &SFVec3fClass, 0,
			     JS_GetParent( cx, obj ) )) != 0)
    {
      VrmlSFVec3f *v3 = new VrmlSFVec3f;
      // multiply v2 by rotation matrix...
      //Vmatrix( v3->get(), r1->getMatrix(), v2->get() );
      JS_SetPrivate( cx, v3obj, v3 );
      *rval = OBJECT_TO_JSVAL(v3obj);
      return JS_TRUE;
    }

  return JS_FALSE;
}


// Should this create a new object or modify obj?...

static JSBool
rot_setAxis(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
	    jsval *rval)
{
  VrmlSFRotation *r1 = (VrmlSFRotation *)JS_GetPrivate( cx, obj );
  VrmlSFVec3f *v2 = 0;

  // Verify that we have an SFVec3f arg && create a result
  if (r1 && argc > 0 &&
      JSVAL_IS_OBJECT(argv[0]) &&
      &SFVec3fClass == JS_GetClass( JSVAL_TO_OBJECT(argv[0]) ) &&
      (v2 = (VrmlSFVec3f *)JS_GetPrivate( cx, JSVAL_TO_OBJECT(argv[0]))) &&
      (((VrmlField*)v2)->fieldType() == VrmlField::SFVEC3F) )
    {
      r1->setAxis(*v2);
      *rval = OBJECT_TO_JSVAL(obj); // return the same object
      return JS_TRUE;
    }

  return JS_FALSE;
}

static JSBool
rot_slerp(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
	  jsval *rval)
{
  VrmlSFRotation *r1 = (VrmlSFRotation *)JS_GetPrivate( cx, obj );
  VrmlSFRotation *r2 = 0;
  JSObject *r3obj = 0;

  // Verify that we have 2 SFRotation's, a number, & create a result
  if (r1 && argc > 1 &&
      JSVAL_IS_OBJECT(argv[0]) &&
      &SFRotationClass == JS_GetClass( JSVAL_TO_OBJECT(argv[0]) ) &&
      (r2 = (VrmlSFRotation *)JS_GetPrivate( cx, JSVAL_TO_OBJECT(argv[0]))) &&
      (((VrmlField*)r2)->fieldType() == VrmlField::SFROTATION) &&
      JSVAL_IS_NUMBER(argv[1]) &&
      (r3obj = JS_NewObject( cx, &SFRotationClass, 0,
			     JS_GetParent( cx, obj ) )) != 0)
    {
      jsdouble factor;
      JS_ValueToNumber( cx, argv[1], &factor );
      JS_SetPrivate( cx, r3obj, new VrmlSFRotation(r1->slerp(*r2, factor)));
      *rval = OBJECT_TO_JSVAL(r3obj);
      return JS_TRUE;
    }

  return JS_FALSE;
}


// SFVec2f methods

static JSBool
vec2f_add(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
  VrmlSFVec2f *v1 = (VrmlSFVec2f *)JS_GetPrivate( cx, obj );
  VrmlSFVec2f *v2 = 0;
  JSObject *v3obj = 0;

  // Verify that we have 2 SFVec2f's && create a result
  if (v1 && argc > 0 &&
      JSVAL_IS_OBJECT(argv[0]) &&
      &SFVec2fClass == JS_GetClass( JSVAL_TO_OBJECT(argv[0]) ) &&
      (v2 = (VrmlSFVec2f *)JS_GetPrivate( cx, JSVAL_TO_OBJECT(argv[0]))) &&
      (((VrmlField*)v2)->fieldType() == VrmlField::SFVEC2F) &&
      (v3obj = JS_NewObject( cx, &SFVec2fClass, 0,
			     JS_GetParent( cx, obj ) )) != 0)
    {
      JS_SetPrivate(cx, v3obj, new VrmlSFVec2f(v1->add(*v2)));
      *rval = OBJECT_TO_JSVAL(v3obj);
      return JS_TRUE;
    }

  return JS_FALSE;
}

static JSBool
vec2f_divide(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
  VrmlSFVec2f *v1 = (VrmlSFVec2f *)JS_GetPrivate( cx, obj );
  jsdouble d = 0.0;
  JSObject *v3obj = 0;

  // Verify that we have a number && create a result
  if (v1 && argc > 0 &&
      JS_ValueToNumber( cx, argv[0], &d ) &&
      (v3obj = JS_NewObject( cx, &SFVec2fClass, 0,
			     JS_GetParent( cx, obj ) )) != 0)
    {
      JS_SetPrivate(cx, v3obj, new VrmlSFVec2f(v1->divide(d)));
      *rval = OBJECT_TO_JSVAL(v3obj);
      return JS_TRUE;
    }

  return JS_FALSE;
}

static JSBool vec2f_dot(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
                        jsval *rval) {
    if (argc > 0 && JSVAL_IS_OBJECT(argv[0])
            && (&SFVec2fClass == JS_GetClass(JSVAL_TO_OBJECT(argv[0])))) {

        const VrmlSFVec2f * const thisVec =
                reinterpret_cast<const VrmlSFVec2f *>(JS_GetPrivate(cx, obj));
        assert(thisVec);

        const VrmlSFVec2f * const argVec =
                reinterpret_cast<const VrmlSFVec2f *>
                    (JS_GetPrivate(cx, JSVAL_TO_OBJECT(argv[0])));
        assert(argVec);

        *rval = DOUBLE_TO_JSVAL(thisVec->dot(*argVec));

        return JS_TRUE;
    }

    return JS_FALSE;
}

static JSBool
vec2f_length(JSContext *cx, JSObject *obj, uintN, jsval *, jsval *rval)
{
  VrmlSFVec2f *v = (VrmlSFVec2f *)JS_GetPrivate( cx, obj );
  if (v)
    {
      *rval = DOUBLE_TO_JSVAL(JS_NewDouble( cx, v->length()));
      return JS_TRUE;
    }
  return JS_FALSE;
}

static JSBool
vec2f_multiply(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
  VrmlSFVec2f *v1 = (VrmlSFVec2f *)JS_GetPrivate( cx, obj );
  jsdouble d = 0.0;
  JSObject *v3obj = 0;

  // Verify that we have a number && create a result
  if (v1 && argc > 0 &&
      JS_ValueToNumber( cx, argv[0], &d ) &&
      (v3obj = JS_NewObject( cx, &SFVec2fClass, 0,
			     JS_GetParent( cx, obj ) )) != 0)
    {
      JS_SetPrivate(cx, v3obj, new VrmlSFVec2f(v1->multiply(d)));
      *rval = OBJECT_TO_JSVAL(v3obj);
      return JS_TRUE;
    }

  return JS_FALSE;
}

static JSBool
vec2f_negate(JSContext *cx, JSObject *obj, uintN, jsval *, jsval *rval)
{
  VrmlSFVec2f *v1 = (VrmlSFVec2f *)JS_GetPrivate( cx, obj );
  JSObject *v3obj = 0;

  if (v1 &&
      (v3obj = JS_NewObject( cx, &SFVec2fClass, 0,
			     JS_GetParent( cx, obj ) )) != 0)
    {
      JS_SetPrivate(cx, v3obj, new VrmlSFVec2f(v1->negate()));
      *rval = OBJECT_TO_JSVAL(v3obj);
      return JS_TRUE;
    }

  return JS_FALSE;
}

static JSBool
vec2f_normalize(JSContext *cx, JSObject *obj, uintN, jsval *, jsval *rval)
{
  VrmlSFVec2f *v1 = (VrmlSFVec2f *)JS_GetPrivate( cx, obj );
  JSObject *v3obj = 0;

  if (v1 &&
      (v3obj = JS_NewObject( cx, &SFVec2fClass, 0,
			     JS_GetParent( cx, obj ) )) != 0)
    {
      JS_SetPrivate(cx, v3obj, new VrmlSFVec2f(v1->normalize()));
      *rval = OBJECT_TO_JSVAL(v3obj);
      return JS_TRUE;
    }

  return JS_FALSE;
}

static JSBool
vec2f_subtract(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
  const VrmlSFVec2f * const v1 =
          reinterpret_cast<const VrmlSFVec2f *>(JS_GetPrivate(cx, obj));
  const VrmlSFVec2f * v2 = 0;
  JSObject *v3obj = 0;

  // Verify that we have 2 SFVec2f's && create a result
  if (v1 && argc > 0 &&
      JSVAL_IS_OBJECT(argv[0]) &&
      &SFVec2fClass == JS_GetClass( JSVAL_TO_OBJECT(argv[0]) ) &&
      (v2 = (VrmlSFVec2f *)JS_GetPrivate( cx, JSVAL_TO_OBJECT(argv[0]))) &&
      (((VrmlField*)v2)->fieldType() == VrmlField::SFVEC2F) &&
      (v3obj = JS_NewObject( cx, &SFVec2fClass, 0,
			     JS_GetParent( cx, obj ) )) != 0)
    {
      JS_SetPrivate(cx, v3obj, new VrmlSFVec2f(v1->subtract(*v2)));
      *rval = OBJECT_TO_JSVAL(v3obj);
      return JS_TRUE;
    }

  return JS_FALSE;
}


// SFVec3f methods

static JSBool
vec3f_add(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
  const VrmlSFVec3f * const v1 =
          reinterpret_cast<VrmlSFVec3f *>(JS_GetPrivate(cx, obj));
  const VrmlSFVec3f * v2 = 0;
  JSObject *v3obj = 0;

  // Verify that we have 2 SFVec3f's && create a result
  if (v1 && argc > 0 &&
      JSVAL_IS_OBJECT(argv[0]) &&
      &SFVec3fClass == JS_GetClass( JSVAL_TO_OBJECT(argv[0]) ) &&
      (v2 = reinterpret_cast<const VrmlSFVec3f *>(JS_GetPrivate(cx, JSVAL_TO_OBJECT(argv[0])))) &&
      (((VrmlField*)v2)->fieldType() == VrmlField::SFVEC3F) &&
      (v3obj = JS_NewObject( cx, &SFVec3fClass, 0,
			     JS_GetParent( cx, obj ) )) != 0)
    {
      JS_SetPrivate(cx, v3obj, new VrmlSFVec3f(v1->add(*v2)));
      *rval = OBJECT_TO_JSVAL(v3obj);
      return JS_TRUE;
    }

  return JS_FALSE;
}

static JSBool
vec3f_cross(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
  VrmlSFVec3f *v1 = (VrmlSFVec3f *)JS_GetPrivate( cx, obj );
  VrmlSFVec3f *v2 = 0;
  JSObject *v3obj = 0;

  // Verify that we have 2 SFVec3f's && create a result
  if (v1 && argc > 0 &&
      JSVAL_IS_OBJECT(argv[0]) &&
      &SFVec3fClass == JS_GetClass( JSVAL_TO_OBJECT(argv[0]) ) &&
      (v2 = (VrmlSFVec3f *)JS_GetPrivate( cx, JSVAL_TO_OBJECT(argv[0]))) &&
      (((VrmlField*)v2)->fieldType() == VrmlField::SFVEC3F) &&
      (v3obj = JS_NewObject( cx, &SFVec3fClass, 0,
			     JS_GetParent( cx, obj ) )) != 0)
    {
      JS_SetPrivate(cx, v3obj, new VrmlSFVec3f(v1->cross(*v2)));
      *rval = OBJECT_TO_JSVAL(v3obj);
      return JS_TRUE;
    }

  return JS_FALSE;
}

static JSBool
vec3f_divide(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
  VrmlSFVec3f *v1 = (VrmlSFVec3f *)JS_GetPrivate( cx, obj );
  jsdouble d = 0.0;
  JSObject *v3obj = 0;

  // Verify that we have a number && create a result
  if (v1 && argc > 0 &&
      JS_ValueToNumber( cx, argv[0], &d ) &&
      (v3obj = JS_NewObject( cx, &SFVec3fClass, 0,
			     JS_GetParent( cx, obj ) )) != 0)
    {
      JS_SetPrivate(cx, v3obj, new VrmlSFVec3f(v1->divide(d)));
      *rval = OBJECT_TO_JSVAL(v3obj);
      return JS_TRUE;
    }

  return JS_FALSE;
}

static JSBool vec3f_dot(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
                        jsval *rval) {
    if (argc > 0 && JSVAL_IS_OBJECT(argv[0])
            && (&SFVec3fClass == JS_GetClass(JSVAL_TO_OBJECT(argv[0])))) {

        const VrmlSFVec3f * const thisVec =
                reinterpret_cast<const VrmlSFVec3f *>(JS_GetPrivate(cx, obj));
        assert(thisVec);

        const VrmlSFVec3f * const argVec =
                reinterpret_cast<const VrmlSFVec3f *>
                    (JS_GetPrivate(cx, JSVAL_TO_OBJECT(argv[0])));
        assert(argVec);

        *rval = DOUBLE_TO_JSVAL(thisVec->dot(*argVec));

        return JS_TRUE;
    }

    return JS_FALSE;
}

static JSBool
vec3f_length(JSContext *cx, JSObject *obj, uintN, jsval *, jsval *rval)
{
  VrmlSFVec3f *v = (VrmlSFVec3f *)JS_GetPrivate( cx, obj );
  if (v)
    {
      *rval = DOUBLE_TO_JSVAL(JS_NewDouble( cx, v->length()));
      return JS_TRUE;
    }
  return JS_FALSE;
}

static JSBool
vec3f_multiply(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
  const VrmlSFVec3f * const v1 =
          reinterpret_cast<const VrmlSFVec3f *>(JS_GetPrivate(cx, obj));
  jsdouble d = 0.0;
  JSObject *v3obj = 0;

  // Verify that we have a number && create a result
  if (v1 && argc > 0 &&
      JS_ValueToNumber( cx, argv[0], &d ) &&
      (v3obj = JS_NewObject( cx, &SFVec3fClass, 0,
			     JS_GetParent( cx, obj ) )) != 0)
    {
      JS_SetPrivate(cx, v3obj, new VrmlSFVec3f(v1->multiply(d)));
      *rval = OBJECT_TO_JSVAL(v3obj);
      return JS_TRUE;
    }

  return JS_FALSE;
}

static JSBool
vec3f_negate(JSContext *cx, JSObject *obj, uintN, jsval *, jsval *rval)
{
  const VrmlSFVec3f * const v1 =
          reinterpret_cast<const VrmlSFVec3f *>(JS_GetPrivate(cx, obj));
  JSObject *v3obj = 0;

  if (v1 &&
      (v3obj = JS_NewObject( cx, &SFVec3fClass, 0,
			     JS_GetParent( cx, obj ) )) != 0)
    {
      JS_SetPrivate( cx, v3obj, new VrmlSFVec3f(v1->negate()));
      *rval = OBJECT_TO_JSVAL(v3obj);
      return JS_TRUE;
    }

  return JS_FALSE;
}

static JSBool
vec3f_normalize(JSContext *cx, JSObject *obj, uintN, jsval *, jsval *rval)
{
  const VrmlSFVec3f * const v1 =
          reinterpret_cast<VrmlSFVec3f *>(JS_GetPrivate(cx, obj));
  JSObject *v3obj = 0;

  if (v1 &&
      (v3obj = JS_NewObject( cx, &SFVec3fClass, 0,
			     JS_GetParent( cx, obj ) )) != 0)
    {
      JS_SetPrivate(cx, v3obj, new VrmlSFVec3f(v1->normalize()));
      *rval = OBJECT_TO_JSVAL(v3obj);
      return JS_TRUE;
    }

  return JS_FALSE;
}

static JSBool
vec3f_subtract(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
  const VrmlSFVec3f * const v1 =
          reinterpret_cast<const VrmlSFVec3f *>(JS_GetPrivate(cx, obj));
  const VrmlSFVec3f * v2 = 0;
  JSObject *v3obj = 0;

  // Verify that we have 2 SFVec3f's && create a result
  if (v1 && argc > 0 &&
      JSVAL_IS_OBJECT(argv[0]) &&
      &SFVec3fClass == JS_GetClass( JSVAL_TO_OBJECT(argv[0]) ) &&
      (v2 = reinterpret_cast<const VrmlSFVec3f *>(JS_GetPrivate(cx, JSVAL_TO_OBJECT(argv[0])))) &&
      (((VrmlField*)v2)->fieldType() == VrmlField::SFVEC3F) &&
      (v3obj = JS_NewObject( cx, &SFVec3fClass, 0,
			     JS_GetParent( cx, obj ) )) != 0)
    {
      JS_SetPrivate(cx, v3obj, new VrmlSFVec3f(v1->subtract(*v2)));
      *rval = OBJECT_TO_JSVAL(v3obj);
      return JS_TRUE;
    }

  return JS_FALSE;
}

// Create class objects for the non-scalar field types.
// Move toString to a prototype object?

void ScriptJS::defineAPI()
{
  // SFColor
  static JSPropertySpec SFColorProps[] = {
    {"r",       0,       JSPROP_ENUMERATE},
    {"g",       1,       JSPROP_ENUMERATE},
    {"b",       2,       JSPROP_ENUMERATE},
    {0}
  };

  static JSFunctionSpec SFColorMethods[] = {
    /*    name          native          nargs    */
    {"setHSV",          color_setHSV,   3},
    {"getHSV",          color_getHSV,   0},
    {"toString",        toString,       0},
    {0}
    };

  if (! JS_InitClass( d_cx, d_globalObj, 0, &SFColorClass,
		      SFColorCons, 0, // constructor function, min arg count
		      SFColorProps, SFColorMethods,
		      0, 0))	  // static properties and methods
    theSystem->error(System::ERRORTYPE_UNKNOWN, "JS_InitClass SFColor failed\n");

  // SFImage
  static JSPropertySpec SFImageProps[] = {
    {"x",       0,       JSPROP_ENUMERATE},
    {"y",       1,       JSPROP_ENUMERATE},
    {"comp",    2,       JSPROP_ENUMERATE},
    {"array",   3,       JSPROP_ENUMERATE},
    {0}
  };

  static JSFunctionSpec SFImageMethods[] = {
    /*    name          native          nargs    */
    {"toString",        toString,       0},
    {0}
    };

  if (! JS_InitClass( d_cx, d_globalObj, 0, &SFImageClass,
		      SFImageCons, 0, // constructor function, min arg count
		      SFImageProps, SFImageMethods,
		      0, 0))	  // static properties and methods
    theSystem->error(System::ERRORTYPE_UNKNOWN, "JS_InitClass SFImage failed\n");

  // SFNode
  static JSFunctionSpec SFNodeMethods[] = {
    /*    name          native          nargs    */
    {"toString",        toString,       0},
    {0}
    };

  if (! JS_InitClass( d_cx, d_globalObj, 0, &SFNodeClass,
		      SFNodeCons, 1, // constructor function, min arg count
		      0, SFNodeMethods,
		      0, 0))	  // static properties and methods
    theSystem->error(System::ERRORTYPE_UNKNOWN, "JS_InitClass SFNode failed\n");


  // SFRotation
  static JSPropertySpec SFRotationProps[] = {
    {"x",       0,       JSPROP_ENUMERATE},
    {"y",       1,       JSPROP_ENUMERATE},
    {"z",       2,       JSPROP_ENUMERATE},
    {"angle",   3,       JSPROP_ENUMERATE},
    {0}
  };
  static JSFunctionSpec SFRotationMethods[] = {
    /*    name          native          nargs    */
    {"getAxis",         rot_getAxis,    0},
    {"inverse",         rot_inverse,    0},
    {"multiply",        rot_multiply,   1},
    {"multVec",         rot_multVec,    1},
    {"setAxis",         rot_setAxis,    1},
    {"slerp",           rot_slerp,      2},
    {"toString",        toString,       0},
    {0}
    };

  if (! JS_InitClass( d_cx, d_globalObj, 0, &SFRotationClass,
		      SFRotationCons, 0, // constructor function, min arg count
		      SFRotationProps, SFRotationMethods,
		      0, 0))	  // static properties and methods
    theSystem->error(System::ERRORTYPE_UNKNOWN, "JS_InitClass SFRotation failed\n");

  // SFVec2f
  static JSPropertySpec SFVec2fProps[] = {
    {"x",       0,       JSPROP_ENUMERATE},
    {"y",       1,       JSPROP_ENUMERATE},
    {0}
  };
  static JSFunctionSpec SFVec2fMethods[] = {
    /*    name          native          nargs    */
    {"add",             vec2f_add,      1},
    {"divide",          vec2f_divide,   1},
    {"dot",             vec2f_dot,      1},
    {"length",          vec2f_length,   0},
    {"multiply",        vec2f_multiply, 1},
    {"negate",          vec2f_negate,   0},
    {"normalize",       vec2f_normalize,0},
    {"subtract",        vec2f_subtract, 1},
    {"toString",        toString,       0},
    {0}
    };

  if (! JS_InitClass( d_cx, d_globalObj, 0, &SFVec2fClass,
		      SFVec2fCons, 0, // constructor function, min arg count
		      SFVec2fProps, SFVec2fMethods,
		      0, 0))	  // static properties and methods
    theSystem->error(System::ERRORTYPE_UNKNOWN, "JS_InitClass SFVec2f failed\n");

  // SFVec3f
  static JSPropertySpec SFVec3fProps[] = {
    {"x",       0,       JSPROP_ENUMERATE},
    {"y",       1,       JSPROP_ENUMERATE},
    {"z",       2,       JSPROP_ENUMERATE},
    {0}
  };
  static JSFunctionSpec SFVec3fMethods[] = {
    /*    name          native          nargs    */
    {"add",             vec3f_add,      1},
    {"cross",           vec3f_cross,    1},
    {"divide",          vec3f_divide,   1},
    {"dot",             vec3f_dot,      1},
    {"length",          vec3f_length,   0},
    {"multiply",        vec3f_multiply, 1},
    {"negate",          vec3f_negate,   0},
    {"normalize",       vec3f_normalize,0},
    {"subtract",        vec3f_subtract, 1},
    {"toString",        toString,       0},
    {0}
    };

  if (! JS_InitClass( d_cx, d_globalObj, 0, &SFVec3fClass,
		      SFVec3fCons, 0, // constructor function, min arg count
		      SFVec3fProps, SFVec3fMethods,
		      0, 0))	  // static properties and methods
    theSystem->error(System::ERRORTYPE_UNKNOWN, "JS_InitClass SFVec3f failed\n");


  // All the MF* classes are subclassed from Array
  JSObject *array_proto = js_InitArrayClass( d_cx, d_globalObj );

  // MFColor class
  static JSPropertySpec MFColorProps[] = {
    {0}
  };

  static JSFunctionSpec MFColorMethods[] = {
    {0}
  };

  if (! JS_InitClass( d_cx, d_globalObj, array_proto, &MFColorClass,
		      MFColorCons, 0, // constructor function, min arg count
		      MFColorProps, MFColorMethods,
		      0, 0))	  // static properties and methods
    theSystem->error(System::ERRORTYPE_UNKNOWN, "JS_InitClass MFColor failed\n");

  // MFFloat class
  static JSPropertySpec MFFloatProps[] = {
    {0}
  };

  static JSFunctionSpec MFFloatMethods[] = {
    {0}
  };

  if (! JS_InitClass( d_cx, d_globalObj, array_proto, &MFFloatClass,
		      MFFloatCons, 0, // constructor function, min arg count
		      MFFloatProps, MFFloatMethods,
		      0, 0))	  // static properties and methods
    theSystem->error(System::ERRORTYPE_UNKNOWN, "JS_InitClass MFFloat failed\n");

  // MFInt32 class
  static JSPropertySpec MFInt32Props[] = {
    {0}
  };

  static JSFunctionSpec MFInt32Methods[] = {
    {0}
  };

  if (! JS_InitClass( d_cx, d_globalObj, array_proto, &MFInt32Class,
		      MFInt32Cons, 0, // constructor function, min arg count
		      MFInt32Props, MFInt32Methods,
		      0, 0))	  // static properties and methods
    theSystem->error(System::ERRORTYPE_UNKNOWN, "JS_InitClass MFInt32 failed\n");

  // MFNode class
  static JSPropertySpec MFNodeProps[] = {
    {0}
  };

  static JSFunctionSpec MFNodeMethods[] = {
    {0}
  };

  if (! JS_InitClass( d_cx, d_globalObj, array_proto, &MFNodeClass,
		      MFNodeCons, 0, // constructor function, min arg count
		      MFNodeProps, MFNodeMethods,
		      0, 0))	  // static properties and methods
    theSystem->error(System::ERRORTYPE_UNKNOWN, "JS_InitClass MFNode failed\n");

  // MFRotation class
  static JSPropertySpec MFRotationProps[] = {
    {0}
  };

  static JSFunctionSpec MFRotationMethods[] = {
    {0}
  };

  if (! JS_InitClass( d_cx, d_globalObj, array_proto, &MFRotationClass,
		      MFRotationCons, 0, // constructor function, min arg count
		      MFRotationProps, MFRotationMethods,
		      0, 0))	  // static properties and methods
    theSystem->error(System::ERRORTYPE_UNKNOWN, "JS_InitClass MFRotation failed\n");

  // MFString class
  static JSPropertySpec MFStringProps[] = {
    {0}
  };

  static JSFunctionSpec MFStringMethods[] = {
    {0}
  };

  if (! JS_InitClass( d_cx, d_globalObj, array_proto, &MFStringClass,
		      MFStringCons, 0, // constructor function, min arg count
		      MFStringProps, MFStringMethods,
		      0, 0))	  // static properties and methods
    theSystem->error(System::ERRORTYPE_UNKNOWN, "JS_InitClass MFString failed\n");

  // MFVec2f class
  static JSPropertySpec MFVec2fProps[] = {
    {0}
  };

  static JSFunctionSpec MFVec2fMethods[] = {
    {0}
  };

  if (! JS_InitClass( d_cx, d_globalObj, array_proto, &MFVec2fClass,
		      MFVec2fCons, 0, // constructor function, min arg count
		      MFVec2fProps, MFVec2fMethods,
		      0, 0))	  // static properties and methods
    theSystem->error(System::ERRORTYPE_UNKNOWN, "JS_InitClass MFVec2f failed\n");

  // MFVec3f class
  static JSPropertySpec MFVec3fProps[] = {
    {0}
  };

  static JSFunctionSpec MFVec3fMethods[] = {
    {0}
  };

  if (! JS_InitClass( d_cx, d_globalObj, array_proto, &MFVec3fClass,
		      MFVec3fCons, 0, // constructor function, min arg count
		      MFVec3fProps, MFVec3fMethods,
		      0, 0))	  // static properties and methods
    theSystem->error(System::ERRORTYPE_UNKNOWN, "JS_InitClass MFVec3f failed\n");
}

// Convert a VrmlField value to a jsval, optionally protect from gc.

#if defined(__ORBISNAP__)

// ATTENTION: The gc protection is broken! Do not use it!
// The JS_AddRoot() function expects a pointer to a pointer to an object,
// not just pointer to an object as this code assumes - if this code is used,
// the garbage collector will crash.

#endif

jsval ScriptJS::vrmlFieldToJSVal( VrmlField::VrmlFieldType type,
				  const VrmlField *f,
				  bool protect )
{
  switch (type)
    {
    case VrmlField::SFBOOL:
      {
	const VrmlSFBool* sfbval = dynamic_cast<const VrmlSFBool*>(f);
	return BOOLEAN_TO_JSVAL( sfbval ? sfbval->get() : 0 );
      }

    case VrmlField::SFCOLOR:
      {
	JSObject *obj = JS_NewObject( d_cx, &SFColorClass, 0, d_globalObj );
	if (! obj) return JSVAL_NULL;
	if (protect) JS_AddRoot( d_cx, obj );
	JS_SetPrivate( d_cx, obj, f ? f->clone() : new VrmlSFColor() );
	return OBJECT_TO_JSVAL(obj);
      }

    case VrmlField::SFFLOAT:
      {
	const VrmlSFFloat* sffval = dynamic_cast<const VrmlSFFloat*>(f);
	jsdouble *d = JS_NewDouble( d_cx, sffval ? sffval->get() : 0.0 );
	if (protect) JS_AddRoot( d_cx, d );
	return DOUBLE_TO_JSVAL( d );
      }

    case VrmlField::SFIMAGE:
      {
	JSObject *obj = JS_NewObject( d_cx, &SFImageClass, 0, d_globalObj );
	if (! obj) return JSVAL_NULL;
	if (protect) JS_AddRoot( d_cx, obj );
	JS_SetPrivate( d_cx, obj, f ? f->clone() : new VrmlSFImage() );
	return OBJECT_TO_JSVAL(obj);
      }

    case VrmlField::SFINT32:
      {
	const VrmlSFInt32* sfival = dynamic_cast<const VrmlSFInt32*>(f);
	return INT_TO_JSVAL( sfival ? sfival->get() : 0 );
      }

    case VrmlField::SFNODE:
      {
	JSObject *obj = JS_NewObject( d_cx, &SFNodeClass, 0, d_globalObj );
	if (! obj) return JSVAL_NULL;
	if (protect) JS_AddRoot( d_cx, obj );
	JS_SetPrivate( d_cx, obj, f ? f->clone() : new VrmlSFNode() );
	return OBJECT_TO_JSVAL(obj);
      }

    case VrmlField::SFROTATION:
      {
	JSObject *obj = JS_NewObject( d_cx, &SFRotationClass, 0, d_globalObj );
	if (! obj) return JSVAL_NULL;
	if (protect) JS_AddRoot( d_cx, obj );
	JS_SetPrivate( d_cx, obj, f ? f->clone() : new VrmlSFRotation() );
	return OBJECT_TO_JSVAL(obj);
      }

    case VrmlField::SFSTRING:
      {
	const VrmlSFString* sfsval = dynamic_cast<const VrmlSFString*>(f);
	JSString *s = JS_NewStringCopyZ( d_cx, sfsval ? sfsval->get() : "" );
	if (protect) JS_AddRoot( d_cx, s );
	return STRING_TO_JSVAL( s );
      }

    case VrmlField::SFTIME:
      {
	const VrmlSFTime* sftval = dynamic_cast<const VrmlSFTime*>(f);
	jsdouble *d = JS_NewDouble( d_cx, sftval ? sftval->get() : 0 );
	if (protect) JS_AddRoot( d_cx, d );
	return DOUBLE_TO_JSVAL( d );
      }

    case VrmlField::SFVEC2F:
      {
	JSObject *obj = JS_NewObject( d_cx, &SFVec2fClass, 0, d_globalObj );
	if (! obj) return JSVAL_NULL;
	if (protect) JS_AddRoot( d_cx, obj );
	JS_SetPrivate( d_cx, obj, f ? f->clone() : new VrmlSFVec2f() );
	return OBJECT_TO_JSVAL(obj);
      }

    case VrmlField::SFVEC3F:
      {
	JSObject *obj = JS_NewObject( d_cx, &SFVec3fClass, 0, d_globalObj );
	if (! obj) return JSVAL_NULL;
	if (protect) JS_AddRoot( d_cx, obj );
	JS_SetPrivate( d_cx, obj, f ? f->clone() : new VrmlSFVec3f() );
	return OBJECT_TO_JSVAL(obj);
      }

    // MF*
    case VrmlField::MFCOLOR:
      {
	assert(!f || dynamic_cast<const VrmlMFColor *>(f));
        const VrmlMFColor * const mf = static_cast<const VrmlMFColor *>(f);
	int n = mf ? mf->getLength() : 0;

#if defined(__ORBISNAP__)

	JSObject *obj = JS_NewObject(d_cx, &MFColorClass, 0, d_globalObj);
	if (!obj) return JSVAL_NULL;
	if (protect) JS_AddRoot(d_cx, obj);
	VrmlMFColor *v = f ? (VrmlMFColor*) f->clone() : new VrmlMFColor();
	JS_SetPrivate(d_cx, obj, v);
	return OBJECT_TO_JSVAL(obj);

#else

	JSObject *obj = JS_NewArrayObject( d_cx, (jsint)n, 0 );
	if (! obj) return JSVAL_NULL;
	JS_AddRoot( d_cx, obj );
	const float * fn = mf ? mf->get() : 0;
	for (i=0; i<n; ++i, fn+=3)
	  {
	    JSObject *elt = JS_NewObject( d_cx, &SFColorClass,
					  0, d_globalObj );
	    if (! elt) break;
	    JS_SetPrivate( d_cx, elt, new VrmlSFColor(fn[0], fn[1], fn[2]) );
	    JS_DefineElement( d_cx, obj, (jsint)i, OBJECT_TO_JSVAL(elt),
			      JS_PropertyStub, JS_PropertyStub,
			      JSPROP_ENUMERATE );
	  }
	if (! protect) JS_RemoveRoot( d_cx, obj );
	return OBJECT_TO_JSVAL(obj);

#endif
      }

    case VrmlField::MFFLOAT:
      {

#if defined(__ORBISNAP__)

	JSObject *obj = JS_NewObject(d_cx, &MFFloatClass, 0, d_globalObj);
	if (!obj) return JSVAL_NULL;
	if (protect) JS_AddRoot(d_cx, obj);
	VrmlMFFloat *v = f ? (VrmlMFFloat *) f->clone() : new VrmlMFFloat();
	JS_SetPrivate(d_cx, obj, v);
	return OBJECT_TO_JSVAL(obj);

#else

	assert(!f || dynamic_cast<const VrmlMFFloat *>(f));
        const VrmlMFFloat * const mf = static_cast<const VrmlMFFloat *>(f);
	int i, n = mf ? mf->getLength() : 0;
	JSObject *obj = JS_NewArrayObject( d_cx, (jsint)n, 0 );
	if (! obj) return JSVAL_NULL;
	JS_AddRoot( d_cx, obj );
	const float * fn = mf ? mf->get() : 0;
	for (i=0; i<n; ++i, ++fn)
	  {
	    jsdouble *elt = JS_NewDouble( d_cx, *fn );
	    if (! elt) break;
	    JS_DefineElement( d_cx, obj, (jsint)i, DOUBLE_TO_JSVAL(elt),
			      JS_PropertyStub, JS_PropertyStub,
			      JSPROP_ENUMERATE );
	  }
	if (! protect) JS_RemoveRoot( d_cx, obj );
	return OBJECT_TO_JSVAL(obj);

#endif

      }

    case VrmlField::MFINT32:
      {

#if defined(__ORBISNAP__)

	JSObject *obj = JS_NewObject(d_cx, &MFInt32Class, 0, d_globalObj);
	if (!obj) return JSVAL_NULL;
	if (protect) JS_AddRoot(d_cx, obj);
	VrmlMFInt32 *v = f ? (VrmlMFInt32 *) f->clone() : new VrmlMFInt32();
	JS_SetPrivate(d_cx, obj, v);
	return OBJECT_TO_JSVAL(obj);

#else

	assert(!f || dynamic_cast<const VrmlMFInt32 *>(f));
        const VrmlMFInt32 * const mf = static_cast<const VrmlMFInt32 *>(f);
	int i, n = mf ? mf->getLength() : 0;
	JSObject *obj = JS_NewArrayObject( d_cx, (jsint)n, 0 );
	if (! obj) return JSVAL_NULL;
	if (protect) JS_AddRoot( d_cx, obj );
	const long *fn = mf ? mf->get() : 0;
	for (i=0; i<n; ++i, ++fn)
	  {
	    JS_DefineElement( d_cx, obj, (jsint)i, INT_TO_JSVAL(*fn),
			      JS_PropertyStub, JS_PropertyStub,
			      JSPROP_ENUMERATE );
	  }
	return OBJECT_TO_JSVAL(obj);

#endif

      }

    case VrmlField::MFNODE:
      {
	assert(!f || dynamic_cast<const VrmlMFNode *>(f));
        const VrmlMFNode * const mf = static_cast<const VrmlMFNode *>(f);
	int i, n = mf ? mf->getLength() : 0;
	JSObject *obj = JS_NewArrayObject( d_cx, (jsint)n, 0 );
	if (! obj) return JSVAL_NULL;
	JS_AddRoot( d_cx, obj );
	for (i=0; i<n; ++i)
	  {
	    JSObject *elt = JS_NewObject( d_cx, &SFNodeClass,
					  0, d_globalObj );
	    if (! elt) break;
	    JS_SetPrivate(d_cx, elt, new VrmlSFNode((*mf)[i]));
	    JS_DefineElement( d_cx, obj, (jsint)i, OBJECT_TO_JSVAL(elt),
			      JS_PropertyStub, JS_PropertyStub,
			      JSPROP_ENUMERATE );
	  }
	if (! protect) JS_RemoveRoot( d_cx, obj );
	return OBJECT_TO_JSVAL(obj);
      }

    case VrmlField::MFROTATION:
      {

#if defined(__ORBISNAP__)

  JSObject *obj = JS_NewObject(d_cx, &MFRotationClass, 0, d_globalObj);
  if (!obj) return JSVAL_NULL;
  if (protect) JS_AddRoot(d_cx, obj);
  VrmlMFRotation *v = f ? (VrmlMFRotation *) f->clone() :
    new VrmlMFRotation();
  JS_SetPrivate(d_cx, obj, v);
  return OBJECT_TO_JSVAL(obj);

#else

	assert(!f || dynamic_cast<const VrmlMFRotation *>(f));
        const VrmlMFRotation * const mf = static_cast<const VrmlMFRotation *>(f);
	int i, n = mf ? mf->getLength() : 0;
	JSObject *obj = JS_NewArrayObject( d_cx, (jsint)n, 0 );
	if (! obj) return JSVAL_NULL;
	JS_AddRoot( d_cx, obj );
	const float * fn = mf ? mf->get() : 0;
	for (i=0; i<n; ++i, fn+=4)
	  {
	    JSObject *elt = JS_NewObject( d_cx, &SFRotationClass,
					  0, d_globalObj );
	    if (! elt) break;
	    JS_SetPrivate( d_cx, elt,
			   new VrmlSFRotation(fn[0], fn[1], fn[2], fn[3]) );
	    JS_DefineElement( d_cx, obj, (jsint)i, OBJECT_TO_JSVAL(elt),
			      JS_PropertyStub, JS_PropertyStub,
			      JSPROP_ENUMERATE );
	  }
	if (! protect) JS_RemoveRoot( d_cx, obj );
	return OBJECT_TO_JSVAL(obj);

#endif
      }

    case VrmlField::MFSTRING:
      {
#if defined(__ORBISNAP__)

	JSObject *obj = JS_NewObject(d_cx, &MFStringClass, 0, d_globalObj);
	if (!obj) return JSVAL_NULL;
	if (protect) JS_AddRoot(d_cx, obj);
	VrmlMFString *v = f ? (VrmlMFString *) f->clone() : new VrmlMFString();
	JS_SetPrivate(d_cx, obj, v);
	return OBJECT_TO_JSVAL(obj);

#else

	jsval *jsvec = 0;
	int i, n = 0;
	if (f)
	  {
            assert(dynamic_cast<const VrmlMFString *>(f));
	    const VrmlMFString * const mf = static_cast<const VrmlMFString *>(f);
	    n = mf ? mf->getLength() : 0;
	    if (n > 0) jsvec = new jsval[n];
	    for (i=0; i<n; ++i)
	      {
		JSString *s = JS_NewStringCopyZ( d_cx, mf->get(i) );
		jsvec[i] = STRING_TO_JSVAL( s );
	      }
	  }
	JSObject *obj = JS_NewArrayObject( d_cx, (jsint)n, jsvec );
	if (jsvec) delete [] jsvec;
	if (! obj) return JSVAL_NULL;
	if (protect) JS_AddRoot( d_cx, obj );
	return OBJECT_TO_JSVAL(obj);

#endif
      }

    case VrmlField::MFVEC2F:
      {

#if defined(__ORBISNAP__)

  JSObject *obj = JS_NewObject(d_cx, &MFVec2fClass, 0, d_globalObj);
  if (!obj) return JSVAL_NULL;
  if (protect) JS_AddRoot(d_cx, obj);
  VrmlMFVec2f *v = f ? (VrmlMFVec2f *) f->clone() : new VrmlMFVec2f();
  JS_SetPrivate(d_cx, obj, v);
  return OBJECT_TO_JSVAL(obj);

#else

	assert(!f || dynamic_cast<const VrmlMFVec2f *>(f));
        const VrmlMFVec2f * const mf = static_cast<const VrmlMFVec2f *>(f);
	int i, n = mf ? mf->getLength() : 0;
	JSObject *obj = JS_NewArrayObject( d_cx, (jsint)n, 0 );
	if (! obj) return JSVAL_NULL;
	JS_AddRoot( d_cx, obj );
	const float * fn = mf ? mf->get() : 0;
	for (i=0; i<n; ++i, fn+=2)
	  {
	    JSObject *elt = JS_NewObject( d_cx, &SFVec2fClass,
					  0, d_globalObj );
	    if (! elt) break;
	    JS_SetPrivate( d_cx, elt, new VrmlSFVec2f(fn[0], fn[1]) );
	    JS_DefineElement( d_cx, obj, (jsint)i, OBJECT_TO_JSVAL(elt),
			      JS_PropertyStub, JS_PropertyStub,
			      JSPROP_ENUMERATE );
	  }
	if (! protect) JS_RemoveRoot( d_cx, obj );
	return OBJECT_TO_JSVAL(obj);

#endif

      }

    case VrmlField::MFVEC3F:
      {

#if defined(__ORBISNAP__)

  JSObject *obj = JS_NewObject(d_cx, &MFVec3fClass, 0, d_globalObj);
  if (!obj) return JSVAL_NULL;
  if (protect) JS_AddRoot(d_cx, obj);
  VrmlMFVec3f *v = f ? (VrmlMFVec3f *) f->clone() : new VrmlMFVec3f();
  JS_SetPrivate(d_cx, obj, v);
  return OBJECT_TO_JSVAL(obj);

#else

        assert(!f || dynamic_cast<const VrmlMFVec3f *>(f));
        const VrmlMFVec3f * const mf = static_cast<const VrmlMFVec3f *>(f);
	int i, n = mf ? mf->getLength() : 0;
	JSObject *obj = JS_NewArrayObject( d_cx, (jsint)n, 0 );
	if (! obj) return JSVAL_NULL;
	JS_AddRoot( d_cx, obj );
	const float *fn = mf ? mf->get() : 0;
	for (i=0; i<n; ++i, fn+=3)
	  {
	    JSObject *elt = JS_NewObject( d_cx, &SFVec3fClass,
					  0, d_globalObj );
	    if (! elt) break;
	    JS_SetPrivate( d_cx, elt, new VrmlSFVec3f(fn[0], fn[1], fn[2]) );
	    JS_DefineElement( d_cx, obj, (jsint)i, OBJECT_TO_JSVAL(elt),
			      JS_PropertyStub, JS_PropertyStub,
			      JSPROP_ENUMERATE );
	  }
	if (! protect) JS_RemoveRoot( d_cx, obj );
	return OBJECT_TO_JSVAL(obj);

#endif

      }

    default:
      theSystem->error(System::ERRORTYPE_UNKNOWN, "vrmlFieldToJSVal: unhandled type (%d)\n",
		       (int) type);
      return BOOLEAN_TO_JSVAL( false );	// oops...
    }
}

// Convert a jsval to a (new) VrmlField

static VrmlField *jsvalToVrmlField( JSContext *cx,
				    jsval v,
				    VrmlField::VrmlFieldType expectType )
{
  if (JSVAL_IS_NULL(v) || JSVAL_IS_VOID(v))
    return 0;

  else if (JSVAL_IS_BOOLEAN(v) && expectType == VrmlField::SFBOOL)
    return new VrmlSFBool( JSVAL_TO_BOOLEAN(v) );

#if defined(__ORBISNAP__)

  // allow both ints and floats to be fed into SFInt32
  else if (JSVAL_IS_NUMBER(v) && expectType == VrmlField::SFINT32)
    {
      jsdouble d;
      JS_ValueToNumber( cx, v, &d );
      return new VrmlSFInt32((int) d);
    }

#else

  else if (JSVAL_IS_INT(v) && expectType == VrmlField::SFINT32)
    return new VrmlSFInt32( JSVAL_TO_INT(v) );

#endif  // __ORBISNAP__

  else if (JSVAL_IS_NUMBER(v) && expectType == VrmlField::SFFLOAT)
    {
      jsdouble d;
      JS_ValueToNumber( cx, v, &d );
      return new VrmlSFFloat( d );
    }

  else if (JSVAL_IS_NUMBER(v) && expectType == VrmlField::SFTIME)
    {
      jsdouble d;
      JS_ValueToNumber( cx, v, &d );
      return new VrmlSFTime( d );
    }

  else if (JSVAL_IS_STRING(v) && expectType == VrmlField::SFSTRING)
    {
      JSString *str = JSVAL_TO_STRING(v);
      return new VrmlSFString( str ? JS_GetStringBytes(str) : "" );
    }

  else if (JSVAL_IS_OBJECT(v) &&
	   ((expectType == VrmlField::SFCOLOR &&
	     &SFColorClass == JS_GetClass( JSVAL_TO_OBJECT(v) )) ||
	    (expectType == VrmlField::SFIMAGE &&
	     &SFImageClass == JS_GetClass( JSVAL_TO_OBJECT(v) )) ||
	    (expectType == VrmlField::SFNODE &&
	     &SFNodeClass == JS_GetClass( JSVAL_TO_OBJECT(v) )) ||
	    (expectType == VrmlField::SFROTATION &&
	     &SFRotationClass == JS_GetClass( JSVAL_TO_OBJECT(v) )) ||
	    (expectType == VrmlField::SFVEC2F &&
	     &SFVec2fClass == JS_GetClass( JSVAL_TO_OBJECT(v) )) ||
	    (expectType == VrmlField::SFVEC3F &&
	     &SFVec3fClass == JS_GetClass( JSVAL_TO_OBJECT(v) )) ))
    {
      VrmlField *f = (VrmlField *)JS_GetPrivate( cx, JSVAL_TO_OBJECT(v) );
      if (! f) return 0;
      return f->clone();
    }

#if defined(__ORBISNAP__)

  else if (JSVAL_IS_OBJECT(v) &&
    ((expectType == VrmlField::MFSTRING &&
      &MFStringClass == JS_GetClass(JSVAL_TO_OBJECT(v))) ||
    (expectType == VrmlField::MFCOLOR &&
      &MFColorClass == JS_GetClass(JSVAL_TO_OBJECT(v))) ||
    (expectType == VrmlField::MFROTATION &&
      &MFRotationClass == JS_GetClass(JSVAL_TO_OBJECT(v))) ||
    (expectType == VrmlField::MFFLOAT &&
      &MFFloatClass == JS_GetClass(JSVAL_TO_OBJECT(v))) ||
    (expectType == VrmlField::MFINT32 &&
      &MFInt32Class == JS_GetClass(JSVAL_TO_OBJECT(v)))))
  {
    VrmlField *f = (VrmlField *) JS_GetPrivate(cx, JSVAL_TO_OBJECT(v));
    if (!f) return 0;
    return f->clone();
  }

#else

  // MFColor
  else if (JSVAL_IS_OBJECT(v) &&
	   JS_IsArrayObject( cx, JSVAL_TO_OBJECT(v) ) &&
	   expectType == VrmlField::MFCOLOR)
    {
      jsuint len;
      JS_GetArrayLength( cx, JSVAL_TO_OBJECT(v), &len );
      float *data = new float[ 3*len ];
      memset(data, 0, 3*len*sizeof(float));

      for (int i=0; i<(int)len; ++i)
	{
	  jsval elt;
	  JS_GetElement( cx, JSVAL_TO_OBJECT(v), (jsint) i, &elt );
	  if (JSVAL_IS_OBJECT(elt)) {
	    JSObject *obj = JSVAL_TO_OBJECT(elt);
	    VrmlField *vf = obj ? (VrmlField *)JS_GetPrivate( cx, obj ) : 0;
	    VrmlSFColor *c = vf ? vf->toSFColor() : 0;
	    if (c) {
	      data[3*i] = c->get()[0];
	      data[3*i+1] = c->get()[1];
	      data[3*i+2] = c->get()[2];
	    }
	  }
	}
      VrmlMFColor *f = new VrmlMFColor( (int)len, data );
      delete [] data;
      return f;
    }

  // MFFloat
  else if (JSVAL_IS_OBJECT(v) &&
	   JS_IsArrayObject( cx, JSVAL_TO_OBJECT(v) ) &&
	   expectType == VrmlField::MFFLOAT)
    {
      jsuint len;
      JS_GetArrayLength( cx, JSVAL_TO_OBJECT(v), &len );
      float *data = new float[ len ];
      memset(data, 0, len*sizeof(float));

      for (int i=0; i<(int)len; ++i)
	{
	  jsval elt;
	  JS_GetElement( cx, JSVAL_TO_OBJECT(v), (jsint) i, &elt );
	  if (JSVAL_IS_NUMBER(elt)) {
	    jsdouble factor;
	    JS_ValueToNumber( cx, elt, &factor );
	    data[i] = (float) factor;
	  }
	}
      VrmlMFFloat *f = new VrmlMFFloat( (int)len, data );
      delete [] data;
      return f;
    }

  // MFInt32
  else if (JSVAL_IS_OBJECT(v) &&
	   JS_IsArrayObject( cx, JSVAL_TO_OBJECT(v) ) &&
	   expectType == VrmlField::MFINT32)
    {
      jsuint len;
      JS_GetArrayLength( cx, JSVAL_TO_OBJECT(v), &len );
      long *data = new long[ len ];
      memset(data, 0, len*sizeof(long));

      for (size_t i = 0; i < len; ++i)
	{
	  jsval elt;
	  JS_GetElement( cx, JSVAL_TO_OBJECT(v), (jsint) i, &elt );
	  if (JSVAL_IS_NUMBER(elt)) {
	    jsdouble factor;
	    JS_ValueToNumber( cx, elt, &factor );
	    data[i] = static_cast<long>(factor);
	  }
	}
      VrmlMFInt32 *f = new VrmlMFInt32(len, data);
      delete [] data;
      return f;
    }

  // MFRotation
  else if (JSVAL_IS_OBJECT(v) &&
	   JS_IsArrayObject( cx, JSVAL_TO_OBJECT(v) ) &&
	   expectType == VrmlField::MFROTATION)
    {
      jsuint len;
      JS_GetArrayLength( cx, JSVAL_TO_OBJECT(v), &len );
      float *data = new float[ 4*len ];
      memset(data, 0, 4*len*sizeof(float));

      for (int i=0; i<(int)len; ++i)
	{
	  jsval elt;
	  JS_GetElement( cx, JSVAL_TO_OBJECT(v), (jsint) i, &elt );
	  if (JSVAL_IS_OBJECT(elt)) {
	    JSObject *obj = JSVAL_TO_OBJECT(elt);
	    VrmlField *vf = obj ? (VrmlField *)JS_GetPrivate( cx, obj ) : 0;
	    VrmlSFRotation *c = vf ? vf->toSFRotation() : 0;
	    if (c) {
	      data[3*i] = c->get()[0];
	      data[3*i+1] = c->get()[1];
	      data[3*i+2] = c->get()[2];
	      data[3*i+3] = c->get()[3];
	    }
	  }
	}
      VrmlMFRotation *f = new VrmlMFRotation( (int)len, data );
      delete [] data;
      return f;
    }

  // MFVec2f
  else if (JSVAL_IS_OBJECT(v) &&
	   JS_IsArrayObject( cx, JSVAL_TO_OBJECT(v) ) &&
	   expectType == VrmlField::MFVEC2F)
    {
      jsuint len;
      JS_GetArrayLength( cx, JSVAL_TO_OBJECT(v), &len );
      float *data = new float[ 2*len ];
      memset(data, 0, 2*len*sizeof(float));

      for (int i=0; i<(int)len; ++i)
	{
	  jsval elt;
	  JS_GetElement( cx, JSVAL_TO_OBJECT(v), (jsint) i, &elt );
	  if (JSVAL_IS_OBJECT(elt)) {
	    JSObject *obj = JSVAL_TO_OBJECT(elt);
	    VrmlField *vf = obj ? (VrmlField *)JS_GetPrivate( cx, obj ) : 0;
	    VrmlSFVec2f *c = vf ? vf->toSFVec2f() : 0;
	    if (c) {
	      data[3*i] = c->get()[0];
	      data[3*i+1] = c->get()[1];
	    }
	  }
	}
      VrmlMFVec2f *f = new VrmlMFVec2f( (int)len, data );
      delete [] data;
      return f;
    }

  // MFVec3f
  else if (JSVAL_IS_OBJECT(v) &&
	   JS_IsArrayObject( cx, JSVAL_TO_OBJECT(v) ) &&
	   expectType == VrmlField::MFVEC3F)
    {
      jsuint len;
      JS_GetArrayLength( cx, JSVAL_TO_OBJECT(v), &len );
      float *data = new float[ 3*len ];
      memset(data, 0, 3*len*sizeof(float));

      for (int i=0; i<(int)len; ++i)
	{
	  jsval elt;
	  JS_GetElement( cx, JSVAL_TO_OBJECT(v), (jsint) i, &elt );
	  if (JSVAL_IS_OBJECT(elt)) {
	    JSObject *obj = JSVAL_TO_OBJECT(elt);
	    VrmlField *vf = obj ? (VrmlField *)JS_GetPrivate( cx, obj ) : 0;
	    VrmlSFVec3f *c = vf ? vf->toSFVec3f() : 0;
	    if (c) {
	      data[3*i] = c->get()[0];
	      data[3*i+1] = c->get()[1];
	      data[3*i+2] = c->get()[2];
	    }
	  }
	}
      VrmlMFVec3f *f = new VrmlMFVec3f( (int)len, data );
      delete [] data;
      return f;
    }

#endif  // __ORBISNAP__

  // MFNode
  else if (JSVAL_IS_OBJECT(v) &&
	   JS_IsArrayObject( cx, JSVAL_TO_OBJECT(v) ) &&
	   expectType == VrmlField::MFNODE)
    {
      jsuint len;
      JS_GetArrayLength( cx, JSVAL_TO_OBJECT(v), &len );
      VrmlMFNode *f = new VrmlMFNode(len);
      for (size_t i = 0; i < len; ++i)
	{
	  jsval elt;
	  JS_GetElement(cx, JSVAL_TO_OBJECT(v), static_cast<jsint>(i), &elt);
	  VrmlSFNode *child = 0;
	  if (JSVAL_IS_OBJECT(elt) &&
	      &SFNodeClass == JS_GetClass( JSVAL_TO_OBJECT(elt) ))
	    child = static_cast<VrmlSFNode *>(JS_GetPrivate(cx, JSVAL_TO_OBJECT(elt)));
	  (*f)[i] = child ? child->get()->reference() : 0;
	}
      return f;
    }

  // MFString
  else if (JSVAL_IS_OBJECT(v) &&
	   JS_IsArrayObject( cx, JSVAL_TO_OBJECT(v) ) &&
	   expectType == VrmlField::MFSTRING)
    {
      jsuint len;
      JS_GetArrayLength( cx, JSVAL_TO_OBJECT(v), &len );
      char **s = new char*[ len ];
      for (int i=0; i<(int)len; ++i)
	{
	  jsval elt;
	  JS_GetElement( cx, JSVAL_TO_OBJECT(v), (jsint) i, &elt );
	  JSString *str = 0;
	  if (JSVAL_IS_STRING(elt) &&
	      (str = JSVAL_TO_STRING(elt)) != 0)
	    s[i] = JS_GetStringBytes(str);
	  else
	    s[i] = 0;
	}
      VrmlMFString *f = new VrmlMFString( (int)len, s );
      delete [] s;
      return f;
    }

  else
    {
      theSystem->error(System::ERRORTYPE_UNKNOWN, "jsvalToVrmlField: unhandled type (expected %d)\n",
	      expectType);
      JSString *str = JS_ValueToString(cx, v);
      if (str)
	theSystem->error(System::ERRORTYPE_UNKNOWN, "%s", JS_GetStringBytes(str));

      return 0;
    }
}


// Must assign the proper type to eventOuts

static JSBool
eventOut_setProperty(JSContext *cx, JSObject *obj, jsval id, jsval *val)
{
  JSString *str = JS_ValueToString(cx, id);
  if (! str) return JS_FALSE;
  const char *eventName = JS_GetStringBytes(str);

  // The ScriptJS object pointer is stored as a global property value:
  ScriptJS *script = objToScript( cx, obj );
  if (! script) return JS_FALSE;
  VrmlNodeScript *scriptNode = script->scriptNode();

  // Validate the type
  VrmlField::VrmlFieldType t = scriptNode->hasEventOut( eventName ) ;
  if (! t) return JS_FALSE;

  // Convert to a vrmlField and set the eventOut value
  VrmlField *f = jsvalToVrmlField( cx, *val, t );
  if (! f)
    {
      theSystem->error(System::ERRORTYPE_UNKNOWN, "Error: invalid type in assignment to eventOut %s\n",
	      eventName );
      return JS_FALSE;
    }

  scriptNode->setEventOut( eventName, f );
  if (f) delete f;

  // Don't overwrite the property value.
  if (JSVAL_IS_OBJECT(*val) &&
      JSVAL_TO_OBJECT(*val) != 0 &&
      ! JS_DefineProperty( cx, JSVAL_TO_OBJECT(*val), "_eventOut",
			   PtrToJSVal(const_cast<char*>(eventName)),
			   0, 0, JSPROP_READONLY | JSPROP_PERMANENT ))
    theSystem->error(System::ERRORTYPE_UNKNOWN, "JS_DefineProp _eventOut failed\n");

  return JS_TRUE;
}

// Must assign the proper type to fields
static JSBool
field_setProperty(JSContext *cx, JSObject *obj, jsval id, jsval *val)
{
  JSString *str = JS_ValueToString(cx, id);
  if (! str) return JS_FALSE;
  const char *fieldName = JS_GetStringBytes(str);

  // The ScriptJS object pointer is stored as a global property value:
  ScriptJS *script = objToScript( cx, obj );
  if (! script) return JS_FALSE;
  VrmlNodeScript *scriptNode = script->scriptNode();

  // Validate the type
  VrmlField::VrmlFieldType t = scriptNode->hasField( fieldName ) ;
  if (! t) return JS_FALSE;

  // Convert to a vrmlField and set the eventOut value
  const VrmlField *f = jsvalToVrmlField( cx, *val, t );
  if (! f)
    {
      theSystem->error(System::ERRORTYPE_UNKNOWN, "Error: invalid type in assignment to field %s\n",
	      fieldName );
      return JS_FALSE;
    }

  scriptNode->setField( fieldName, *f );
  if (f) delete f;

  return JS_TRUE;
}


// Define objects corresponding to fields/eventOuts

void ScriptJS::defineFields()
{
  for (VrmlNodeScript::FieldList::const_iterator it=d_node->fields().begin(); it!=d_node->fields().end(); ++it)
  {
    jsval val = vrmlFieldToJSVal(it->second.type, it->second.value, false);
#ifdef SCRIPTJS_DEBUG
    if ( it->second->value )
      cout << "field " << it->first << " value "
	   << *(it->second->value) << endl;
    else
      cout << "field " << it->first << " <no value>\n";
#endif

    if (! JS_DefineProperty( d_cx, d_globalObj, it->first.c_str(), val,
			     //getter, setter, ...
			     0, field_setProperty,
			     JSPROP_PERMANENT ))
      theSystem->error(System::ERRORTYPE_UNKNOWN, "JS_DefineProp %s failed\n", it->first.c_str());
  }

  for (VrmlNodeScript::FieldList::const_iterator it=d_node->eventOuts().begin(); it!=d_node->eventOuts().end(); ++it)
  {
    jsval val = vrmlFieldToJSVal(it->second.type, it->second.value, false);
    if (JSVAL_IS_OBJECT(val) && JSVAL_TO_OBJECT(val) != 0 &&
      !JS_DefineProperty(d_cx, JSVAL_TO_OBJECT(val), "_eventOut",
         PtrToJSVal(const_cast<char*>(it->first.c_str())),
         0, 0, JSPROP_READONLY | JSPROP_PERMANENT ))
    {
      theSystem->error(System::ERRORTYPE_UNKNOWN, "JS_DefineProp _eventOut failed\n");
    }

    if (!JS_DefineProperty( d_cx, d_globalObj, it->first.c_str(), val,
      //getter, setter
      0, eventOut_setProperty,
      JSPROP_PERMANENT ))
    {
      theSystem->error(System::ERRORTYPE_UNKNOWN, "JS_DefineProp %s failed\n", it->first.c_str());
    }
  }
}

// Global functions

static JSBool
Print(JSContext *cx, JSObject *, uintN argc, jsval *argv, jsval *)
{
  uintN i;

  for (i = 0; i < argc; i++) {
    JSString *str = JS_ValueToString(cx, argv[i]);
    if (!str) return JS_FALSE;
    printf("%s%s", i ? " " : "", JS_GetStringBytes(str));
  }
  putchar('\n');
  return JS_TRUE;
}


// Browser functions

static JSBool getName(JSContext *cx, JSObject *b,
		      uintN , jsval *,
		      jsval *rval)
{
  jsval p;
  if (! JS_GetProperty( cx, b, "_script", &p ))
    return JS_FALSE;

  ScriptJS* s = static_cast<ScriptJS*>(JSValToPtr(p));
  if (! s || ! s->browser() )
    return JS_FALSE;

  const char *name = s->browser()->getName();
  *rval = STRING_TO_JSVAL(JS_InternString( cx, name ));
  return JS_TRUE;
}

static JSBool getVersion(JSContext *cx, JSObject *b,
		      uintN , jsval *,
		      jsval *rval)
{
  jsval p;
  if (! JS_GetProperty( cx, b, "_script", &p ))
    return JS_FALSE;

  ScriptJS* s = static_cast<ScriptJS*>(JSValToPtr(p));
  if (! s || ! s->browser() )
    return JS_FALSE;

  const char *version = s->browser()->getVersion();
  *rval = STRING_TO_JSVAL(JS_InternString( cx, version ));
  return JS_TRUE;
}

static JSBool getCurrentSpeed(JSContext *cx, JSObject*,
			      uintN, jsval *,
			      jsval *rval)
{
  *rval = DOUBLE_TO_JSVAL(JS_NewDouble( cx, 0.0 )); //...
  return JS_TRUE;
}

static JSBool getCurrentFrameRate(JSContext *cx, JSObject* b,
				  uintN, jsval*, jsval *rval)
{
  jsval p;
  if (! JS_GetProperty( cx, b, "_script", &p ))
    return JS_FALSE;

  ScriptJS* s = static_cast<ScriptJS*>(JSValToPtr(p));
  if (! s || ! s->browser() )
    return JS_FALSE;

  *rval = DOUBLE_TO_JSVAL(JS_NewDouble( cx, s->browser()->getFrameRate() ));
  return JS_TRUE;
}

static JSBool getWorldURL(JSContext *cx, JSObject *b,
			  uintN, jsval *,
			  jsval *rval)
{
  jsval p;
  if (! JS_GetProperty( cx, b, "_script", &p ))
    return JS_FALSE;

  ScriptJS* s = static_cast<ScriptJS*>(JSValToPtr(p));
  if (! s || ! s->browser() )
    return JS_FALSE;

  const char *url = 0;
  if (s->browser()->urlDoc())
    url = s->browser()->urlDoc()->url();
  if (! url) url = "";
  *rval = STRING_TO_JSVAL(JS_InternString( cx, url ));
  return JS_TRUE;
}


// No events will be processed after loadURL.

static JSBool loadURL(JSContext* cx, JSObject* b,
		      uintN argc, jsval *argv, jsval *)
{
  if ((int)argc < 1)
    return JS_FALSE;

  jsval p;
  if (! JS_GetProperty( cx, b, "_script", &p ))
    return JS_FALSE;

  ScriptJS* s = static_cast<ScriptJS*>(JSValToPtr(p));
  if (! s || ! s->browser() )
    return JS_FALSE;

  // Get 2 arguments MFString url, MFString parameters
  VrmlMFString *url, *parameters = 0;
  url = (VrmlMFString*) jsvalToVrmlField( cx, argv[0], VrmlField::MFSTRING );
  if (! url) return JS_TRUE;	// empty URL

  if ((int)argc > 1)
    parameters = (VrmlMFString*) jsvalToVrmlField( cx, argv[1],
						   VrmlField::MFSTRING );

  s->browser()->queueLoadUrl( url, parameters );
  return JS_TRUE;
}


// This does return, but no events will be processed after it is called.

static JSBool replaceWorld(JSContext *cx, JSObject *b,
			   uintN argc, jsval *argv, jsval *)
{
  if ((int)argc < 1)
    return JS_FALSE;

  jsval p;
  if (! JS_GetProperty( cx, b, "_script", &p ))
    return JS_FALSE;

  ScriptJS* s = static_cast<ScriptJS*>(JSValToPtr(p));
  if (! s || ! s->browser() )
    return JS_FALSE;

  // Get 1 argument MFNode nodes
  VrmlMFNode *nodes = (VrmlMFNode*) jsvalToVrmlField( cx,
						      argv[0],
						      VrmlField::MFNODE );
  if (nodes)
    {
      VrmlNamespace *ns = new VrmlNamespace(); // should be stored with nodes...
      s->browser()->queueReplaceNodes( nodes, ns );
    }

  return JS_TRUE;
}

static JSBool createVrmlFromString(JSContext *cx, JSObject* bobj,
				   uintN argc, jsval *argv, jsval *rval)
{
  JSString *str;

  if (argc == 1 &&
      JSVAL_IS_STRING(argv[0]) &&
      (str = JSVAL_TO_STRING(argv[0])) != 0)
    {
      char *vrmlString = JS_GetStringBytes(str);
      VrmlNamespace ns;
      VrmlMFNode *kids;

      try
      {
        kids = VrmlScene::readString(vrmlString, &ns);
      }
      catch (VrmlLoadingError &error)
      {
        theSystem->error(System::ERRORTYPE_UNKNOWN, "Javascript: createVrmlFromString() failed: %s",
                         error.what());
	return JS_FALSE;
      }

      // Put the children from g into an MFNode and return in rval.
      // should store the namespace as well...
      int i, n = kids->getLength();
      jsval *jsvec = new jsval[n];

      for (i=0; i<n; ++i)
	{
	  JSObject *obj = JS_NewObject( cx, &SFNodeClass, 0, bobj );
	  if (! obj) return JS_FALSE;
	  //uwe JS_SetPrivate( cx, obj, k[i] ? k[i]->reference() : 0);
	  JS_SetPrivate( cx, obj, new VrmlSFNode((*kids)[i] ? (*kids)[i]->reference() : 0));
	  jsvec[i] = OBJECT_TO_JSVAL(obj);
	}
      *rval = OBJECT_TO_JSVAL(JS_NewArrayObject(cx, (jsint)n, jsvec));
      delete [] jsvec;
      delete kids;
      return JS_TRUE;
    }

  return JS_FALSE;
}


// createVrmlFromURL( MFString url, SFNode node, SFString event )

static JSBool createVrmlFromURL(JSContext* cx, JSObject* b,
				uintN argc, jsval *argv, jsval *)
{
  if ((int)argc != 3) return JS_FALSE;

  Doc * relative = 0;

  jsval p;
  ScriptJS *s = 0;		// egcs warning
  if (JS_GetProperty( cx, b, "_script", &p ) &&
      (s = static_cast<ScriptJS*>(JSValToPtr(p))) != 0 &&
      s->browser() != 0)
    relative = s->browser()->urlDoc();

  VrmlMFString *url = 0;
  VrmlField *f;
  if (JSVAL_IS_STRING(argv[0]))
    {
      JSString *str = JSVAL_TO_STRING(argv[0]);
      const char * const strBytes = JS_GetStringBytes(str);
      url = new VrmlMFString(1, &strBytes);
    }
  else if ((f = jsvalToVrmlField( cx, argv[0], VrmlField::MFSTRING )) != 0)
    {
      url = dynamic_cast<VrmlMFString*>(f);
    }

  if (url)
    {
      VrmlSFNode *node = 0;	// egcs warning
      JSString *str = 0;	// egcs warning

      f = jsvalToVrmlField( cx, argv[1], VrmlField::SFNODE );
      if ( (node = dynamic_cast<VrmlSFNode*>(f)) != 0 &&
	   JSVAL_IS_STRING(argv[2]) &&
	   (str = JSVAL_TO_STRING(argv[2])) != 0)
	{
	  char *event = JS_GetStringBytes(str);

	  VrmlNamespace ns;	// this is a problem...
          VrmlMFNode* kids;
          try
          {
            kids = VrmlScene::readWrl( url, relative, &ns );
          }
          catch (VrmlLoadingError &error)
          {
            theSystem->error(System::ERRORTYPE_UNKNOWN, "Javascript: createVrmlFromUrl() failed: %s",
                             error.what());
            delete node;
            delete url;
            return JS_FALSE;
          }
	  VrmlNode *nn = node->get();

	  if (nn)
	    nn->eventIn( s_timeStamp, // fix me...
			 event,
			 kids );
	  delete node;
	  delete kids;
	  delete url;
	  return JS_TRUE;
	}

      delete url;
      delete f;
    }
  return JS_FALSE;
}

// addRoute(SFNode fromNode, String fromEventOut, SFNode toNode, String toEvent)

static JSBool addRoute(JSContext* cx, JSObject*,
		       uintN argc, jsval *argv,
		       jsval *)
{
  if ((int)argc != 4 ||
      ! JSVAL_IS_STRING(argv[1]) ||
      ! JSVAL_IS_STRING(argv[3]))
    return JS_FALSE;
  JSString *str = JS_ValueToString(cx, argv[1]);
  if (!str) return JS_FALSE;
  char *eventOut = JS_GetStringBytes( str );
  str = JS_ValueToString(cx, argv[3]);
  if (!str) return JS_FALSE;
  char *eventIn = JS_GetStringBytes( str );

  if (!eventOut || !eventIn) return JS_FALSE;
  VrmlSFNode* from = dynamic_cast<VrmlSFNode*>(jsvalToVrmlField(cx, argv[0], VrmlField::SFNODE));
  VrmlSFNode* to = dynamic_cast<VrmlSFNode*>(jsvalToVrmlField(cx, argv[2], VrmlField::SFNODE));
  if (!to || !to->get() || !from || !from->get()) return JS_FALSE;

  theSystem->debug("addRoute(%s::%s::%s, %s::%s::%s)\n",
		   from->get()->nodeType().getName(),
		   from->get()->name(),
		   eventOut,
		   to->get()->nodeType().getName(),
		   to->get()->name(),
		   eventIn);

  from->get()->getNamespace()->addRoute(from->get(), eventOut,
    to->get(), eventIn);

  return JS_TRUE;
}

// deleteRoute(SFNode fromNode, String fromEventOut, SFNode toNode, String toEvent)

static JSBool deleteRoute(JSContext* cx, JSObject*,
			  uintN argc, jsval *argv, jsval *)
{
  if ((int)argc != 4 ||
      ! JSVAL_IS_STRING(argv[1]) ||
      ! JSVAL_IS_STRING(argv[3]))
    return JS_FALSE;
  JSString *str = JS_ValueToString(cx, argv[1]);
  if (! str) return JS_FALSE;
  char *eventOut = JS_GetStringBytes( str );
  str = JS_ValueToString(cx, argv[3]);
  if (! str) return JS_FALSE;
  char *eventIn = JS_GetStringBytes( str );

  if (! eventOut || ! eventIn) return JS_FALSE;
  VrmlSFNode* from = dynamic_cast<VrmlSFNode*>(jsvalToVrmlField(cx, argv[0], VrmlField::SFNODE));
  VrmlSFNode* to = dynamic_cast<VrmlSFNode*>(jsvalToVrmlField(cx, argv[2], VrmlField::SFNODE));
  if (!to || !to->get() || !from || !from->get()) return JS_FALSE;

  from->get()->getNamespace()->removeRoute(from->get(), eventOut,
    to->get(), eventIn);

  return JS_TRUE;
}

static JSBool setDescription(JSContext *cx, JSObject*,
			     uintN argc, jsval *argv, jsval *)
{
  if (argc < 1)
    return JS_FALSE;
  JSString *str = JS_ValueToString(cx, argv[0]);
  if (!str)
    return JS_FALSE;
  theSystem->debug("%s", JS_GetStringBytes(str));
  return JS_TRUE;
}


//

static void
ErrorReporter(JSContext *, const char *message, JSErrorReport *report)
{
    int i, j, k, n;

    theSystem->error(System::ERRORTYPE_UNKNOWN, "javascript: ");
    if (!report) {
	theSystem->error(System::ERRORTYPE_UNKNOWN, "%s\n", message);
	return;
    }

    if (report->filename)
	theSystem->error(System::ERRORTYPE_UNKNOWN, "%s, ", report->filename);
    if (report->lineno)
	theSystem->error(System::ERRORTYPE_UNKNOWN, "line %u: ", report->lineno);
    theSystem->error(System::ERRORTYPE_UNKNOWN, message);
    if (!report->linebuf) {
	theSystem->error(System::ERRORTYPE_UNKNOWN, "\n");
	return;
    }

    theSystem->error(System::ERRORTYPE_UNKNOWN, ":\n%s\n", report->linebuf);
    n = report->tokenptr - report->linebuf;
    for (i = j = 0; i < n; i++) {
	if (report->linebuf[i] == '\t') {
	    for (k = (j + 8) & ~7; j < k; j++)
	      theSystem->error(System::ERRORTYPE_UNKNOWN, ".");
	    continue;
	}
	theSystem->error(System::ERRORTYPE_UNKNOWN, ".");
	j++;
    }
    theSystem->error(System::ERRORTYPE_UNKNOWN, "\n");
}

# endif // ifndef OPENVRML_DISABLE_JAVASCRIPT
