/*****************************************************************************
******************************************************************************
*
*               System-dependent functions
*
*               $Revision: 1.1.2.13 $
*               $Date: 2012/01/13 12:56:15 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright 1998-2012 HUMUSOFT s.r.o.
// Portions Copyright (C) 1998  Chris Morley

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "System.h"
#include "Doc.h"

#include "wx/socket.h"
#include "wx/protocol/http.h"
#include "wx/url.h"

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if !defined(NDEBUG) && !defined(__ORBISNAP__)
#define SYSTEM_DEBUG
#endif

#include <sstream>
using namespace std;

#include "wx/file.h"
#include "wx/filename.h"
#include "wx/log.h"


/* Define snprintf for Windows. Correctly terminate the string
   because Windows _snprintf doesn't do that correctly.
*/
#ifdef _WIN32
int snprintf(char* buf, size_t count, const char* format, ...)
{
  va_list args;
  va_start(args, format);

  int ret = vsnprintf(buf, count, format, args);
  if (ret<0)
    ret = count;
  if (ret==count)
    buf[count-1] = '\0';

  va_end(args);
  return(ret);
}

/* Convert pathname backslashes to slashes.
   Windows only.
*/
std::string backslashesToSlashes(const std::string& str)
{
  // copy source argument to destination
  std::string retval(str);
  const char* strp = str.c_str();

  // convert backslashes to slashes, taking multibyte characters into account
  wchar_t wc;
  size_t nb;
  for (size_t i=0; (nb = mbtowc(&wc, strp, MB_CUR_MAX))>0; i+=nb, strp+=nb)
    if (wc == L'\\') retval[i] = '/';

  return(retval);
}
#endif // _WIN32


// A default System object
static System defaultSystem;

// The global System object
System *theSystem = &defaultSystem;

// constructor
System::System() :
  debugmode(false),
  libraryToken("*sl3dlib")  // the library token string
{ };


void System::setLibDir(const std::string& ld)
{
  libDir = backslashesToSlashes(ld);
}

std::string System::copyResource(const char* sourceUrl)
{
  // create url representation
  Doc urlDoc(sourceUrl);

  // return if it is not file url
  if (strcmp(urlDoc.urlProtocol(), "file")!=0)
    return std::string(sourceUrl);


  if (!::wxDirExists(exportDir.c_str()) && !::wxMkdir(exportDir.c_str()))
  {
    theSystem->warn(System::SEVERITY_LOW, System::WARNINGTYPE_UNKNOWN,
		    "Could not copy resource file %s to destination folder %s"\
		    "because this folder could not be created.", sourceUrl, exportDir.c_str());
    return std::string(sourceUrl);
  }

  // relative path to resource
  std::string relativePath;

  const char* urlBase = urlDoc.urlBase();
  if (urlBase == 0 || *urlBase == '\0')
  {
    // no file in url -- do not copy anything
    return std::string(sourceUrl);
  }
  else if (strncmp(sourceUrl, libraryToken.c_str(), sizeof(libraryToken.c_str())-1)==0)
  {
    // link to library

    // absolute link to library resource
    std::string libraryUrl(sourceUrl);

    // replace library token in url (if any)
    libraryUrl.replace(0, libraryToken.size(), getLibDir().c_str());
    urlDoc.seturl(libraryUrl.c_str());

    // get relative path of url
    relativePath.assign(sourceUrl);
    size_t lengthToReplace = (relativePath.size() > libraryToken.size()) ? libraryToken.size()+1 : libraryToken.size();
    relativePath.replace(0, lengthToReplace, "");
    Doc relativeUrlDoc(relativePath.c_str());
    relativePath.assign(relativeUrlDoc.localPath());
  }
  else if (!Doc::isAbsolute(urlDoc.url()))
  {
    // relative url

    // get relative path
    relativePath.assign(urlDoc.localPath());

    // get original wrl file path
    Doc originalUrl(sourceFileUrl.c_str());

    // do not prepend dir to relative export path when source relative path starts with "<original_file_name>_files"
    // test case: do export operation recursively
    wxFileName relativePathFN(wxString(relativePath.c_str()), wxString(""));
    wxString firstDir(originalUrl.urlBase());
    firstDir.append("_files");
    if (relativePathFN.GetDirCount()>0 && relativePathFN.GetDirs().Item(0).Cmp(firstDir)==0)
    {
      relativePathFN.RemoveDir(0);
    }

    // set full path to resource
    std::string sourceFullPath(originalUrl.localPath());

    // append relative path and file name
    sourceFullPath.append(relativePath.c_str()).append(urlDoc.urlBase()).c_str();

    // append extension (if any)
    const char* ext = 0;
    if ((ext = urlDoc.urlExt())!=0 && *ext)
      sourceFullPath.append(".").append(ext);

    // set it
    urlDoc.seturl(sourceFullPath.c_str());

    // modify relative path which is needed for building export file name
    relativePath.assign(relativePathFN.GetFullPath(wxPATH_UNIX).c_str());
  }

  // source file to copy
  wxFileName sourceFileNameFN(urlDoc.localName());

  // destination file
  std::string exportPath(exportDir);
  wxFileName exportFileNameFN(exportPath.append(relativePath).c_str(), sourceFileNameFN.GetFullName());

  // try to create directory if it does not exist
  if (!wxFileName::DirExists(exportFileNameFN.GetPath()))
    wxFileName::Mkdir(exportFileNameFN.GetPath(), 0777, wxPATH_MKDIR_FULL);

  // copy resource only if read-only destination doesn't exist already
  if ((!exportFileNameFN.FileExists() || exportFileNameFN.IsFileWritable()) &&
      !::wxCopyFile(sourceFileNameFN.GetFullPath(), exportFileNameFN.GetFullPath()))
  {
    theSystem->warn(System::SEVERITY_LOW, System::WARNINGTYPE_UNKNOWN,
		    "Could not copy resource file %s to destination folder %s.", sourceUrl, exportDir.c_str());
    return std::string(sourceUrl);
  }

  // build url and return it
  wxFileName exportDirFN(exportDir.c_str());
  const wxArrayString& dirs = exportDirFN.GetDirs();
  wxString dir = dirs[exportDirFN.GetDirCount()-1];
  dir.append(wxFileName::GetPathSeparator(wxPATH_UNIX));
  dir.append(relativePath.c_str());
  dir.append(sourceFileNameFN.GetFullName().c_str());
  return std::string(dir);
}

// Should make these iostream objects...

void System::error(ErrorType type, const char *fmt, ...)
{
  va_list ap;
  va_start(ap, fmt);
  vfprintf(stderr, fmt, ap);
  va_end(ap);
}

void System::warn(Severity severity, WarningType type, const char *fmt, ...)
{
  va_list ap;
  va_start(ap, fmt);
  fprintf(stderr,"Warning: ");
  vfprintf(stderr, fmt, ap);
  va_end(ap);
}


// Write to the browser status line...

void System::inform(const char *fmt, ...)
{
  static char lastbuf[1024] = { 0 };
  char buf[1024];

  va_list ap;
  va_start(ap, fmt);
  vsprintf(buf, fmt, ap);
  va_end(ap);
  if (strcmp(lastbuf, buf))
    {
      fprintf(stderr,"%s", buf);
      putc('\n', stderr);
      strcpy(lastbuf, buf);
    }
}

#ifdef SYSTEM_DEBUG
void System::debug(const char *fmt, ...)
{
  va_list ap;
  va_start(ap, fmt);
  vfprintf(stderr, fmt, ap);
  va_end(ap);
#else
void System::debug(const char *, ...)
{
#endif
}

#if defined(_WIN32) && !defined(__CYGWIN__)

// Windows

#include <sys/timeb.h>
#include <time.h>

double System::time()
{
   struct _timeb timebuffer;
   _ftime( &timebuffer );
	return (double) timebuffer.time + 1.e-3 * (double) timebuffer.millitm;
}

#else


// Unix/Linux
#include <sys/time.h>
#include <unistd.h>

double System::time()
{
  struct timeval tv;
  struct timezone tz;
  gettimeofday(&tv, &tz);

  return (double) tv.tv_sec + 1.e-6 * (double) tv.tv_usec;
}

#endif // _WIN32


// added for working under windows (and is not needed for mac os)...
#if defined(_WIN32) && !defined(__CYGWIN__)

// needed only for wxWidgets 2.9.x and above
#if wxCHECK_VERSION(2, 9, 0)
#include <winsock2.h>
#endif // wxCHECK_VERSION(2, 9, 0)

#include <sys/types.h>
#include <string.h>
#include <ctype.h>
#else
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <string.h>		// memset
#include <ctype.h>
#endif // _WIN32


int System::connectSocket( const char *host, int port )
{
  struct sockaddr_in sin;
  struct hostent *he;

#if defined(_WIN32) && !defined(__CYGWIN__)
  WSADATA wsaData;
  WORD wVersionRequested;
#endif

  int sockfd = -1;

#if defined(_WIN32) && !defined(__CYGWIN__)
   wVersionRequested = MAKEWORD( 1, 0 );
   if (WSAStartup(wVersionRequested,&wsaData) == SOCKET_ERROR)
   {
     theSystem->error(System::ERRORTYPE_UNKNOWN, "WSAStartup failed with error %d\n",WSAGetLastError());
     WSACleanup();
     return -1;
   }
#endif
  memset( &sin, 0, sizeof(sin) );
  sin.sin_family = AF_INET;
  sin.sin_port = htons(port);

  // Check for dotted number format
  char *s;
  for (s=(char*)host; *s; ++s)
    if ( ! (isdigit(*s) || *s == '.') ) break;

  if (*s)			// Not dotted number
    he = gethostbyname(host);
  else
    {
      u_long addr = inet_addr(host);
      he = gethostbyaddr((char *)&addr, sizeof(addr), AF_INET);
#ifdef __ORBISNAP__
      if (he==NULL && addr==inet_addr("127.0.0.1"))   // gethostbyaddr() fails on Win9x/Me for 127.0.0.1
        he = gethostbyname("localhost");	      // so try localhost
#endif
    }

  if (he)
    {
      memcpy((char *)&sin.sin_addr, he->h_addr, he->h_length);
      sockfd = socket( PF_INET, SOCK_STREAM, IPPROTO_TCP );
      if (sockfd != -1)
	if (connect( sockfd, (struct sockaddr *)&sin, sizeof(sin)) == -1)
	  {
#if defined(_WIN32) && !defined(__CYGWIN__)
	    closesocket(sockfd);
	    WSACleanup();
#else
	    close( sockfd );
#endif
	    sockfd = -1;
	  }
    }

  return sockfd;
}


#include <fcntl.h>		// open() modes
#if defined(_WIN32) && !defined(__CYGWIN__)
# include <io.h>
#else
# include <unistd.h>
#endif

#include <errno.h>

const char *System::httpHost( const char *url, int *port )
{
  static char hostname[256];
  const char *s = strstr(url,"//");
  char *p = hostname;

  if (s)
    {
      for (s+=2; *s && *s != '/' && *s != ':'; *p++ = *s++)
	/* empty */ ;
      if (*s == ':' && isdigit(*(s+1)))
	*port = atoi(s+1);
    }
  *p = '\0';
  return hostname;
}


#ifndef __APPLE__

const char *System::httpFetch( const char *url )
{
  int port = 80;
  const char *hostname = httpHost(url, &port);

  // download the file
  wxHTTP http;
  wxIPV4address addr;
  addr.Hostname(hostname);
  addr.Service(port);
  http.Connect(addr, true);
  wxURI ref(url);
  wxInputStream* istr = http.GetInputStream(ref.GetPath());
  if (istr == NULL)
  {
    theSystem->error(System::ERRORTYPE_UNKNOWN, "Could not download file \"%s\"", url);
    return "";
  }

  // make a local copy
  const char* result = tempnam(0, "VR");
  FILE* f = fopen(result, "wb");
  while (true)
  {
    char buf[1024];
    istr->Read(buf, 1024);
    if (istr->LastRead() == 0) break;
    fwrite(buf, 1, istr->LastRead(), f);
  }
  fclose(f);
  delete istr;

  return result;
}

#else // non-wxWidgets code used for Mac - Mac crashes in wxWidgets for some reason

const char *System::httpFetch( const char *url )
{
  int port = 80;
  const char *hostname = httpHost(url, &port);

  int sockfd;
  if ((sockfd = System::connectSocket( hostname, port )) != -1)
  {
    theSystem->debug("connected.");
  }
  else
  {
    theSystem->error(System::ERRORTYPE_UNKNOWN, "Connect failed: %s (errno %d).\n", strerror(errno), errno);
    return("");
  }

  // Copy to a local temp file
  char *result = 0;
  if (sockfd != -1 && (result = tempnam(0, "VR")))
  {
    int fd = open(result, O_RDWR|O_CREAT, 0777);
    if (fd != -1)
    {
      char *abspath = strstr((char *)url, "//");
      if (abspath) abspath = strchr(abspath+2, '/');
      if (! abspath) abspath = (char*) url;
      char request[1024];
      snprintf(request, sizeof(request), "GET %s HTTP/1.0\nAccept: */*\n\r\n", abspath);

      int nbytes = strlen(request);
      if (write(sockfd, request, nbytes) != nbytes)
      {
        theSystem->warn(System::SEVERITY_HIGH, System::WARNINGTYPE_UNKNOWN, "Sending HTTP GET request failed: %s (errno %d)\n", strerror(errno), errno);
        return("");
      }

      bool gothdr = false;
      int nmore;
      char *start;

      // loop while there's anything to read
      while ((nmore = read(sockfd, request, sizeof(request)-1)) > 0)
      {
        start = request;

        // skip header (should rather analyze it)
        if (!gothdr)
        {
          request[nmore] = '\0';    // add string terminator
          start = strstr(request, "\r\n\r\n");
          if (start)
            start += 4;
          else
          {
            start = strstr(request, "\n\n");
            if (start) start += 2;
          }

          /* header not found yet - wait for more data */
          if (!start)
          {
            // sleep for 10 ms to allow the HTTP server execution
            usleep(10000);
            continue;
          }

          gothdr = true;
        }

        // store the read data to a temporary file
        nmore -= (start - request);
        if (write(fd, start, nmore) != nmore)
        {
          theSystem->warn(System::SEVERITY_HIGH, System::WARNINGTYPE_UNKNOWN, "HTTP: temporary file write error\n");
          break;
        }

        // sleep for 10 ms to allow the HTTP server execution
        usleep(10000);
      }
      close(fd);
    }
  }

  if (sockfd != -1)
    close(sockfd);

  return result;
}

#endif // __APPLE__



void System::removeFile( const char *fn )
{
  remove(fn);
}



char* System::newstr(const char* str)
{
  char* result;

  // special case: NULL produces an empty string
  if (str == NULL)
  {
    result = new char[1];
    result[0] = '\0';
    return result;
  }

  size_t len = strlen(str);
  result = new char[len + 1];
  memcpy(result, str, len + 1);
  return result;
}



char* System::generateFile(const char* tmpl, const char* rel)
{
  // get the current date and time
  time_t t = ::time(NULL);
  struct tm* timestruct = localtime(&t);    // statically allocated (THREAD UNSAFE!)

  // parse the related filename
  wxFileName relFileName(rel);

  // this cycle may be repeated more than once to find a free number for '%n'
  int n = 1;
  size_t len = strlen(tmpl);
  ostringstream strbuf;
  bool containsNum = false, success = false, done = false;
  while (!done)
  {
    strbuf.str("");
    bool special = false;
    for (size_t i=0; i<len; i++)
    {
      char c = tmpl[i];

      // '%' is a special character marker, double '%%' translates to '%'
      if (c == '%')
      {
        if (special) { strbuf << '%'; special = false; }
        else special = true;
      }
      else
      {
        if (special)
        {
          // next character will be handled normally
          special = false;

          // translate the special characters
          if      (c == 'f') strbuf << (const char*) relFileName.GetName();
          else if (c == 'd') strbuf << (const char*) relFileName.GetPath(wxPATH_GET_VOLUME);
          else if (c == 'n') { strbuf << n; containsNum = true; }
          else if (c == 'h') strbuf << (const char*) wxString::Format("%02d", timestruct->tm_hour);
          else if (c == 'm') strbuf << (const char*) wxString::Format("%02d", timestruct->tm_min);
          else if (c == 's') strbuf << (const char*) wxString::Format("%02d", timestruct->tm_sec);
          else if (c == 'D') strbuf << (const char*) wxString::Format("%02d", timestruct->tm_mday);
          else if (c == 'M') strbuf << (const char*) wxString::Format("%02d", timestruct->tm_mon + 1);
          else if (c == 'Y') strbuf << timestruct->tm_year + 1900;

          // unknown special characters are printed out as if not special
          else strbuf << c;
        }
        else
        {
          strbuf << tmpl[i];
          special = false;
        }
      }
    }

    // get file path
    wxString pathonly = wxPathOnly(strbuf.str().c_str());
    wxFileName dirname = wxFileName::DirName(pathonly);

    // create directory if does not exist
    if (!dirname.DirExists() && !pathonly.IsEmpty()) // DirExists() returns false on empty string
      dirname.Mkdir(0777, wxPATH_MKDIR_FULL); // we don't care about result at this moment

    // try creating the file
    // If '%n' is not present, we can overwrite the existing file.
    wxFile* file = new wxFile();
    {
      wxLogNull hush;  // suppresses logging in this scope
      success = file->Open(strbuf.str().c_str(), containsNum ? wxFile::write_excl : wxFile::write);
      file->Close();
    }
    delete file;

    // if the template contains '%n' and the file can't be created,
    // we increment the number and try again; in all other cases we exit
    // (incrementing is limited to 99 just to be sure we don't deadlock)
    if (containsNum && !success && n<100) { done = false; n++; }
    else done = true;
  }

  char* retval = success ? newstr(strbuf.str().c_str()) : NULL;
  return retval;
}


bool System::renameToBackup(const char* filename)
{
  // true if backup operation successfull
  bool ok = true;

  // do nothing if file doesn't exist
  if (wxFileExists(filename))
  {
    // rename the file by appending .bak to the original name
    std::string bakname(filename);
    bakname.append(".bak");
    wxRemoveFile(bakname.c_str());
    ok &= wxRenameFile(filename, bakname.c_str(), false);
  }

  if (!exportDir.empty() && ::wxDirExists(exportDir.c_str()))
  {
    std::string bakname(exportDir.c_str());
    char pathSep = wxFileName::GetPathSeparator();
    if (bakname.compare(bakname.size()-1, 1, &pathSep, 1)==0)
      bakname.erase(bakname.size()-1);

    wxFileName bakDir(bakname.append(".bak").append(1, pathSep).c_str());
    bakDir.Rmdir(wxPATH_RMDIR_RECURSIVE);

    ok &= (rename(exportDir.c_str(), bakname.c_str()) == 0);
  }
  return ok;
}
