/*****************************************************************************
******************************************************************************
*
*               System-dependent functions
*
*               $Revision: 1.1.2.8 $
*               $Date: 2011/06/01 14:14:26 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright 1998-2011 HUMUSOFT s.r.o.
// Portions Copyright (C) 1998  Chris Morley

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#ifndef SYSTEM_H
#define SYSTEM_H
//
//  System dependent utilities class
//

#include <stddef.h>
#include <string>


// This needs cleaning up because:
//
// OPENVRMLDLLEXPORT definition should be placed in separate or module-wide
// config include file
//
// OPENVRMLDLLEXPORT attribute should be used only with class members that
// really need to be exported - this way we wouldn't need to disable warnings
//
#if defined(_WIN32) && defined(OPENVRML_MAKINGDLL)
#  define OPENVRMLDLLEXPORT __declspec(dllexport)
#  pragma warning(disable:4251 4275)
#elif defined(_WIN32) && defined(OPENVRML_USINGDLL)
#  define OPENVRMLDLLEXPORT __declspec(dllimport)
#  pragma warning(disable:4251 4275)
#else
#  define OPENVRMLDLLEXPORT
#endif


class OPENVRMLDLLEXPORT System {

private:
   // sl3d library directory
  std::string libDir;

  // export directory (if empty, export disabled)
  std::string exportDir;

  // source file url
  std::string sourceFileUrl;

  // library token
  std::string libraryToken;

protected:

  /**
   * True when the debug mode is enabled (debugging messages are printed),
   * false if disabled (debugging messages are ignored).
   */
  bool debugmode;

public:

  /** Type of error */
  enum ErrorType
  {
    ERRORTYPE_UNKNOWN
  };

  /** Type of warning */
  enum WarningType
  {
    WARNINGTYPE_UNKNOWN
  };

  /** severity (for warnings) */
  enum Severity
  {
    SEVERITY_LOW,
    SEVERITY_HIGH
  };

  // constructor
  System();

  // setter and getter for library directory and token
  void setLibDir(const std::string& LD);
  const std::string& getLibDir() { return libDir; }
  const std::string& getLibraryToken() { return libraryToken; }

  // setter and getter for export directory
  void setExportDir(const char* dir)
  {
    exportDir.assign(dir);
  }

  const char* getExportDir()
  {
    return (exportDir.empty() ? 0 : exportDir.c_str());
  }

  // setter for source directory
  void setSourceFileUrl(const char* url)
  {
    sourceFileUrl.assign(url);
  }

  // copy source file to destination folder
  std::string copyResource(const char* sourceUrl);

  virtual double time();

  virtual void error(ErrorType type, const char *, ...);

  virtual void warn(Severity severity, WarningType type, const char *, ...);

  virtual void inform(const char *, ...);

  virtual void debug(const char *, ...);

  bool getDebugMode() const
    { return debugmode; }

  void setDebugMode(bool onoff)
    { debugmode = onoff; }


  virtual int connectSocket( const char *host, int port );

  virtual const char *httpHost(const char *url, int *port);
  virtual const char *httpFetch( const char *url );

  virtual void removeFile( const char *fn );

  /**
   * Makes a copy of the given string via 'new'.
   * @param str The string to be copied (NULL is allowed,
   *            resulting in an empty string).
   * @return A pointer to the copy.
   */
  virtual char* newstr(const char* str);

  /**
   * Builds a file name from a template containing wildcard characters
   * and a specified related file path, and creates the empty file
   * if it does not already exist.
   * The following wildcards are possible:
   * <ul>
   * <li><em>%f</em> - replaced by the related file name</li>
   * <li><em>%d</em> - replaced by the related file directory</li>
   * <li><em>%h</em> - replaced by the current hour (0..23)</li>
   * <li><em>%m</em> - replaced by the current minute (0..59)</li>
   * <li><em>%s</em> - replaced by the current second (0..59)</li>
   * <li><em>%D</em> - replaced by the current day in the month (0..31)</li>
   * <li><em>%M</em> - replaced by the current month (0..12)</li>
   * <li><em>%Y</em> - replaced by the current year (e.g. 2003)</li>
   * </ul>
   * @param tmpl The name template with wildcards.
   * @param rel The related file path.
   * @return Name of the generated file, or NULL on failure.
   */
  virtual char* generateFile(const char* tmpl, const char* rel);

  /**
   * Rename a file to a backup file.
   * @param filename File to back up.
   * @return True if renamed to backup, false if file does not exist or on error.
  */
  bool renameToBackup(const char* filename);
};

extern OPENVRMLDLLEXPORT System *theSystem;


/* Declare snprintf for Windows. We use our own implementation
   because Windows _snprintf doesn't terminate strings correctly.
*/
#ifdef _WIN32
OPENVRMLDLLEXPORT int snprintf(char* buf, size_t count, const char* format, ...);
#endif


/* Convert pathname backslashes to slashes in-place.
   Windows only.
*/
#ifdef _WIN32
OPENVRMLDLLEXPORT std::string backslashesToSlashes(const std::string& str);
#else
inline std::string backslashesToSlashes(const std::string& str)
{ return(str); }
#endif


#endif // SYSTEM_H
