/*****************************************************************************
******************************************************************************
*
*               Abstract viewer base class
*
*               $Revision: 1.1.2.4 $
*               $Date: 2011/01/06 09:48:15 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright 1998-2010 HUMUSOFT s.r.o.
// Portions Copyright (C) 1998  Chris Morley

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "Viewer.h"

#include "MathUtils.h"
#include "VrmlScene.h"

#include "VrmlBVolume.h"
#include "VrmlBSphere.h"
#include "VrmlAABox.h"
#include "VrmlNodeViewpoint.h"
#include "VrmlNodeNavigationInfo.h"
#include "VrmlNodeBackground.h"
#include "VrmlNodeFog.h"
#include "VrmlNodeProto.h"


/* initialization of static member */
Viewer* Viewer::d_currentViewer = NULL;

/* constructor */
Viewer::Viewer(VrmlScene* scene)
  : d_scene(scene),
    d_renderSuppressed(false),
    d_headlight(true),
    d_navMode(NAVMODE_EXAMINE),
    d_renderPhase(RENDER_PHASE_NONE),
    viewpoint(NULL),
    navinfo(NULL),
    background(NULL),
    fog(NULL),
    d_cameraBound(true),
    d_cameraChanged(true),
    d_persistentInfo(NULL),

    markedForDestruction(false),
    valid(false)
{
  // return if no scene
  if (d_scene==NULL) return;

  // load the default viewpoint, navinfo, background and fog
  viewpoint = d_scene->bindableViewpointTop();
  background = d_scene->bindableBackgroundTop();
  navinfo = d_scene->bindableNavigationInfoTop();
  fog = d_scene->bindableFogTop();
}



Viewer::~Viewer()
{
  // unregister the viewer with scene if still registered
  if (d_scene) d_scene->unregisterViewer(this);
}



void Viewer::setScene(VrmlScene* scene)
{
  // relink to the new scene only if scene has changed
  if (d_scene == scene)
    return;

  if (d_scene) d_scene->unregisterViewer(this);
  d_scene = scene;
  if (d_scene) d_scene->registerViewer(this);
}



//  Build a cylinder object. It might be smarter to do just one, and reference
//  it with scaling (but the world creator could just as easily do that with
//  DEF/USE ...).

void Viewer::computeCylinder(double height,
                             double radius,
                             int numFacets,
                             float c[][3],
                             float tc[][3],
                             int faces[])
{
  double angle, x, y;
  int i, polyIndex;

  // Compute coordinates, texture coordinates:
  for (i = 0; i < numFacets; ++i) {
    angle = i * 2 * M_PI / numFacets;
    x = cos(angle);
    y = sin(angle);
    c[i][0] = radius * x;
    c[i][1] = 0.5 * height;
    c[i][2] = radius * y;
    c[numFacets+i][0] = radius * x;
    c[numFacets+i][1] = -0.5 * height;
    c[numFacets+i][2] = radius * y;

    if (tc)
      {
        double u = 0.75 - ((float) i) / numFacets;
        //double u = ((float) i) / numFacets + 0.25;
        //if ( u > 1.0 ) u -= 1.0;
        tc[i][0] = u; // ((float) i) / numFacets;
        tc[i][1] = 1.0;
        tc[i][2] = 0.0;
        tc[numFacets+i][0] = u; //((float) i) / numFacets;
        tc[numFacets+i][1] = 0.0;
        tc[numFacets+i][2] = 0.0;
      }
  }

  // And compute indices:
  for (i = 0; i < numFacets; ++i) {
    polyIndex = 5*i;
    faces[polyIndex + 0] = i;
    faces[polyIndex + 1] = (i+1) % numFacets;
    faces[polyIndex + 2] = (i+1) % numFacets + numFacets;
    faces[polyIndex + 3] = i + numFacets;
    faces[polyIndex + 4] = -1;
  }
}


void Viewer::setNavMode(int mode)
{
  d_navMode = mode;
}

void Viewer::setNavMode()
{
  if (d_scene && navinfo)
  {
    const VrmlMFString* navModes = (const VrmlMFString*) navinfo->getField("type");
    if (navModes->getLength() > 0)
    {
      const char* defNavMode = navModes->get(0);
      if (strcmp(defNavMode, "FLY") == 0)
	setNavMode(NAVMODE_FLY);
      else if (strcmp(defNavMode, "WALK") == 0)
	setNavMode(NAVMODE_WALK);
      else if (strcmp(defNavMode, "EXAMINE") == 0)
	setNavMode(NAVMODE_EXAMINE);
      else if (strcmp(defNavMode, "NONE") == 0)
	setNavMode(NAVMODE_NONE);
    }
  }
}

void Viewer::setHeadlight(bool headlight)
{
  d_headlight = headlight;
}

//  Build an extrusion.

void Viewer::computeExtrusion(int nOrientation,
                              const float *orientation,
                              int nScale,
                              const float *scale,
                              int nCrossSection,
                              const float *crossSection,
                              int nSpine,
                              const float *spine,
                              float *c,   // OUT: coordinates
                              float *tc,  // OUT: texture coords
                              int *faces)     // OUT: face list
{
  int i, j, ci;

  // Xscp, Yscp, Zscp- columns of xform matrix to align cross section
  // with spine segments.
  float Xscp[3] = { 1.0, 0.0, 0.0};
  float Yscp[3] = { 0.0, 1.0, 0.0};
  float Zscp[3] = { 0.0, 0.0, 1.0};
  float lastZ[3];

  // Is the spine a closed curve (last pt == first pt)?
  bool spineClosed = (FPZERO(spine[ 3*(nSpine-1)+0 ] - spine[0]) &&
                      FPZERO(spine[ 3*(nSpine-1)+1 ] - spine[1]) &&
                      FPZERO(spine[ 3*(nSpine-1)+2 ] - spine[2]));

  // Is the spine a straight line?
  bool spineStraight = true;
  for (i = 1; i < nSpine-1; ++i)
    {
      float v1[3], v2[3];
      v1[0] = spine[3*(i-1)+0] - spine[3*(i)+0];
      v1[1] = spine[3*(i-1)+1] - spine[3*(i)+1];
      v1[2] = spine[3*(i-1)+2] - spine[3*(i)+2];
      v2[0] = spine[3*(i+1)+0] - spine[3*(i)+0];
      v2[1] = spine[3*(i+1)+1] - spine[3*(i)+1];
      v2[2] = spine[3*(i+1)+2] - spine[3*(i)+2];
      Vcross(v1, v2, v1);
      if (Vlength(v1) != 0.0)
        {
          spineStraight = false;
          Vnorm( v1 );
          Vset( Zscp, v1 );
          break;
        }
    }

  // If the spine is a straight line, compute a constant SCP xform
  if (spineStraight)
    {
      float V1[3] = { 0.0, 1.0, 0.0}, V2[3], V3[3];
      V2[0] = spine[3*(nSpine-1)+0] - spine[0];
      V2[1] = spine[3*(nSpine-1)+1] - spine[1];
      V2[2] = spine[3*(nSpine-1)+2] - spine[2];
      Vcross( V3, V2, V1 );
      double len = Vlength(V3);
      if (len != 0.0)           // Not aligned with Y axis
        {
          Vscale(V3, 1.0/len);

          float orient[4];              // Axis/angle
          Vset(orient, V3);
          orient[3] = acos(Vdot(V1,V2));
          double scp[4][4];             // xform matrix
          Mrotation( scp, orient );
          for (int k=0; k<3; ++k) {
            Xscp[k] = scp[0][k];
            Yscp[k] = scp[1][k];
            Zscp[k] = scp[2][k];
          }
        }
    }

  // Orientation matrix
  double om[4][4];

#if defined(__ORBISNAP__)

  if (!FPZERO(orientation[3])) Mrotation(om, orientation);
  else Midentity(om);

#else

  if (nOrientation == 1 && ! FPZERO(orientation[3]) )
    Mrotation( om, orientation );

#endif

  // Compute coordinates, texture coordinates:
  for (i = 0, ci = 0; i < nSpine; ++i, ci+=nCrossSection) {

    // Scale cross section
    for (j = 0; j < nCrossSection; ++j) {
      c[3*(ci+j)+0] = scale[0] * crossSection[ 2*j ];
      c[3*(ci+j)+1] = 0.0;
      c[3*(ci+j)+2] = scale[1] * crossSection[ 2*j+1 ];
    }

    // Compute Spine-aligned Cross-section Plane (SCP)
    if (! spineStraight)
      {
        float S1[3], S2[3];     // Spine vectors [i,i-1] and [i,i+1]
        int yi1, yi2, si1, s1i2, s2i2;

        if (spineClosed && (i == 0 || i == nSpine-1))
          {
            yi1 = 3*(nSpine-2);
            yi2 = 3;
            si1 = 0;
            s1i2 = 3*(nSpine-2);
            s2i2 = 3;
          }
        else if (i == 0)
          {
            yi1 = 0;
            yi2 = 3;
            si1 = 3;
            s1i2 = 0;
            s2i2 = 6;
          }
        else if (i == nSpine-1)
          {
            yi1 = 3*(nSpine-2);
            yi2 = 3*(nSpine-1);
            si1 = 3*(nSpine-2);
            s1i2 = 3*(nSpine-3);
            s2i2 = 3*(nSpine-1);
          }
        else
          {
            yi1 = 3*(i-1);
            yi2 = 3*(i+1);
            si1 = 3*i;
            s1i2 = 3*(i-1);
            s2i2 = 3*(i+1);
          }

        Vdiff( Yscp, &spine[yi2], &spine[yi1] );
        Vdiff( S1, &spine[s1i2], &spine[si1] );
        Vdiff( S2, &spine[s2i2], &spine[si1] );

        Vnorm( Yscp );
        Vset(lastZ, Zscp);      // Save last Zscp
        Vcross( Zscp, S2, S1 );

        float VlenZ = Vlength(Zscp);
        if ( VlenZ == 0.0 )
          Vset(Zscp, lastZ);
        else
          Vscale( Zscp, 1.0/VlenZ );

        if ((i > 0) && (Vdot( Zscp, lastZ ) < 0.0))
          Vscale( Zscp, -1.0 );

        Vcross( Xscp, Yscp, Zscp );
      }

#if !defined(__ORBISNAP__)

    // Rotate cross section into SCP
    for (j = 0; j < nCrossSection; ++j) {
      float cx, cy, cz;
      cx = c[3*(ci+j)+0]*Xscp[0]+c[3*(ci+j)+1]*Yscp[0]+c[3*(ci+j)+2]*Zscp[0];
      cy = c[3*(ci+j)+0]*Xscp[1]+c[3*(ci+j)+1]*Yscp[1]+c[3*(ci+j)+2]*Zscp[1];
      cz = c[3*(ci+j)+0]*Xscp[2]+c[3*(ci+j)+1]*Yscp[2]+c[3*(ci+j)+2]*Zscp[2];

      c[3*(ci+j)+0] = cx;
      c[3*(ci+j)+1] = cy;
      c[3*(ci+j)+2] = cz;
    }

#endif

    // Apply orientation
    if (! FPZERO(orientation[3]) )
      {
        if (nOrientation > 1)

#if defined(__ORBISNAP__)

        // rotate in the reverse angle
        {
          float ori[4];
          ori[0] = orientation[0];
          ori[1] = orientation[1];
          ori[2] = orientation[2];
          ori[3] = -orientation[3];
          Mrotation(om, ori);
        }

#else

          Mrotation( om, orientation );

#endif

        for (j = 0; j < nCrossSection; ++j) {
          float cx, cy, cz;
          cx = c[3*(ci+j)+0]*om[0][0]+c[3*(ci+j)+1]*om[1][0]+c[3*(ci+j)+2]*om[2][0];
          cy = c[3*(ci+j)+0]*om[0][1]+c[3*(ci+j)+1]*om[1][1]+c[3*(ci+j)+2]*om[2][1];
          cz = c[3*(ci+j)+0]*om[0][2]+c[3*(ci+j)+1]*om[1][2]+c[3*(ci+j)+2]*om[2][2];

          c[3*(ci+j)+0] = cx;
          c[3*(ci+j)+1] = cy;
          c[3*(ci+j)+2] = cz;
        }
      }

#if defined(__ORBISNAP__)

    // According to VRML specs, we rotate the cross section into SCP
    // and then orient it. However, it seems the appropriate matrix
    // multiplications must be done in reverse order. Is it a rule? Maybe :o)

    // Rotate cross section into SCP
    for (j = 0; j < nCrossSection; ++j) {
      float cx, cy, cz;
      cx = c[3*(ci+j)+0]*Xscp[0]+c[3*(ci+j)+1]*Yscp[0]+c[3*(ci+j)+2]*Zscp[0];
      cy = c[3*(ci+j)+0]*Xscp[1]+c[3*(ci+j)+1]*Yscp[1]+c[3*(ci+j)+2]*Zscp[1];
      cz = c[3*(ci+j)+0]*Xscp[2]+c[3*(ci+j)+1]*Yscp[2]+c[3*(ci+j)+2]*Zscp[2];

      c[3*(ci+j)+0] = cx;
      c[3*(ci+j)+1] = cy;
      c[3*(ci+j)+2] = cz;
    }

#endif

    // Translate cross section
    for (j = 0; j < nCrossSection; ++j) {
      c[3*(ci+j)+0] += spine[3*i+0];
      c[3*(ci+j)+1] += spine[3*i+1];
      c[3*(ci+j)+2] += spine[3*i+2];

      // Texture coords
      tc[3*(ci+j)+0] = ((float) j) / (nCrossSection-1);
      tc[3*(ci+j)+1] = ((float) i) / (nSpine-1);
      tc[3*(ci+j)+2] = 0.0;
    }

#if defined(__ORBISNAP__)

    // VRML specs say that nScale and nOrientation should be equal to nSpine,
    // or 1. However, we should not crash when we get something else.
    // So, if nScale or nOrientation is larger than 1 but lower than nSpine,
    // only the first value is used, ignoring the rest.

    if (nScale >= nSpine) scale += 2;
    if (nOrientation >= nSpine) orientation += 4;

#else

    if (nScale > 1) scale += 2;
    if (nOrientation > 1) orientation += 4;

#endif
  }

  // And compute face indices:
  if (faces)
    {
      int polyIndex = 0;
      for (i = 0, ci = 0; i < nSpine-1; ++i, ci+=nCrossSection) {
        for (j = 0; j < nCrossSection-1; ++j) {
          faces[polyIndex + 0] = ci+j;
          faces[polyIndex + 1] = ci+j+1;
          faces[polyIndex + 2] = ci+j+1 + nCrossSection;
          faces[polyIndex + 3] = ci+j + nCrossSection;
          faces[polyIndex + 4] = -1;
          polyIndex += 5;
        }
      }
    }
}


void Viewer::computeSphere(double radius,
                           int numLatLong,
                           float c[][3],
                           float tc[][3],
                           int *faces)
{
  double r, angle, x, y, z;
  int i, j, polyIndex;

  // Compute coordinates, texture coordinates:
  for (i = 0; i < numLatLong; ++i) {
    /*y = 2.0 * ( ((double)i) / (numLatLong-1) ) - 1.0;*/
    angle = ( i * M_PI / (numLatLong-1) ) - M_PI_2;
    y = sin( angle );
    r = sqrt( 1.0 - y*y );
    for (j = 0; j < numLatLong; ++j) {
      angle = 2 * M_PI * ((double)j) / numLatLong;
      x = - sin(angle)*r;
      z = - cos(angle)*r;
      c[i*numLatLong+j][0] = radius * x;
      c[i*numLatLong+j][1] = radius * y;
      c[i*numLatLong+j][2] = radius * z;
      if (tc)
        {
          tc[i*numLatLong+j][0] = ((float) j)/(numLatLong);
          tc[i*numLatLong+j][1] = ((float) i)/(numLatLong);
          tc[i*numLatLong+j][2] = 0.0;
        }
    }
  }

  // And compute indices:
  if (faces)
    for (i = 0; i < numLatLong-1; ++i) {
      for (j = 0; j < numLatLong; ++j) {
        polyIndex = 5*(i*numLatLong+j);
        faces[polyIndex + 0] = i*numLatLong+j;
        faces[polyIndex + 1] = i*numLatLong+(j+1)%numLatLong;
        faces[polyIndex + 2] = (i+1)*numLatLong+(j+1)%numLatLong;
        faces[polyIndex + 3] = (i+1)*numLatLong+j;
        faces[polyIndex + 4] = -1;  // quad
      }
    }
}


//
// Compute a target and up vector from position/orientation/distance.
//

void Viewer::computeView(const float position[3],    /* in */
                         float orientation[4],       /* in */   // [3] to [4] changed by __ORBISNAP__
                         float distance,             /* in */
                         float target[3],            /* out */
                         float up[3])                /* out */
{
  // Graphics Gems, p 466. Convert between axis/angle and rotation matrix
  double len = sqrt( orientation[0]*orientation[0] +
                     orientation[1]*orientation[1] +
                     orientation[2]*orientation[2] );
  if (len > 0.0)
    {
      orientation[0] /= len;
      orientation[1] /= len;
      orientation[2] /= len;
    }

  double s = sin(orientation[3]);
  double c = cos(orientation[3]);
  double t = 1.0 - c;

  // Transform [0,0,1] by the orientation to determine sight line
  target[0] = t * orientation[0] * orientation[2] + s * orientation[1];
  target[1] = t * orientation[1] * orientation[2] - s * orientation[0];
  target[2] = t * orientation[2] * orientation[2] + c;

  // Move along that vector the specified distance away from position[]
  target[0] = -distance*target[0] + position[0];
  target[1] = -distance*target[1] + position[1];
  target[2] = -distance*target[2] + position[2];

  // Transform [0,1,0] by the orientation to determine up vector
  up[0] = t * orientation[0] * orientation[1] - s * orientation[2];
  up[1] = t * orientation[1] * orientation[1] + c;
  up[2] = t * orientation[1] * orientation[2] + s * orientation[0];
}


void Viewer::setColor(float , float , float , float ) {}



/**
 *  For normal VRML97 use, this won't need to be overridden, but for
 *  systems with non-standard view volumes, this can be changed to
 *  cull as appropriate. Note that culling can be disabled by setting
 *  a flag in VrmlRenderContext. Since I don't have access to the
 *  appropriate cave/boom/whichever api's, I can't be sure that this
 *  is enough. If it isn't, please express any concerns to the
 *  libvrml97 developer's list, and it can be fixed...
 */
int Viewer::isectViewVolume(const VrmlBVolume& bv) const {
  int r = VrmlBVolume::BV_PARTIAL;
  const VrmlBSphere* bs = dynamic_cast<const VrmlBSphere*>(&bv);
  if (bs)
  {
    r = bs->isectFrustum(d_frust);
  }
  else
  {
    const VrmlAABox* ab = dynamic_cast<const VrmlAABox*>(&bv);
    if (ab)
      r = ab->isectFrustum(d_frust);
  }
  //cout << "Viewer::isectViewVolume(VrmlBSphere&)=" << r << endl;
  return r;
}


// comment here on how we're forcing everybody to carry around a frust
// whether they want it or not, and how it shouldn't be used except
// for debugging and stuff since it might not be valid in some
// implementations
//
const VrmlFrustum& Viewer::getFrustum() const
{
  return d_frust;
}

void Viewer::setFrustum(const VrmlFrustum& afrust) {
  d_frust = afrust; // copy
}

bool Viewer::isPrimary() const
{
  return(d_scene && (d_scene->getPrimaryViewer()==this));
}

void Viewer::setCurrent()
{
  d_currentViewer = this;
  if (d_scene) d_scene->setPrimaryViewer(this);
}

Viewer* Viewer::getCurrent()
{
  return d_currentViewer;
};


void Viewer::onBindableBound(VrmlNode* bindable)
{
  // only the primary viewer keeps its bindables in sync with the scene
  if (!isPrimary())
    return;

  // bindable can be a Viewpoint
  VrmlNodeViewpoint* newviewpoint = NodeDynamicCast<VrmlNodeViewpoint*>(bindable);
  if (newviewpoint!=NULL && newviewpoint!=viewpoint)
  {
    gotoViewpoint(newviewpoint);
    return;
  }

  // or a NavigationInfo
  VrmlNodeNavigationInfo* newnavinfo = NodeDynamicCast<VrmlNodeNavigationInfo*>(bindable);
  if (newnavinfo != NULL)
  {
    navinfo = newnavinfo;
    return;
  }

  // or a Background
  VrmlNodeBackground* newbackground = NodeDynamicCast<VrmlNodeBackground*>(bindable);
  if (newbackground != NULL)
  {
    background = newbackground;
    return;
  }

  // or a Fog
  VrmlNodeFog* newfog = NodeDynamicCast<VrmlNodeFog*>(bindable);
  if (newfog != NULL)
    fog = newfog;
}

void Viewer::onBindableUnbound(VrmlNode* bindable)
{
  // the primary viewer keeps its bindables in sync with the scene;
  // if our bindable was unbound, we must switch to the new one
  if (!isPrimary())
    return;

  // bindable can be a Viewpoint
  VrmlNodeViewpoint* oldviewpoint = NodeDynamicCast<VrmlNodeViewpoint*>(bindable);
  if (oldviewpoint != NULL && viewpoint == oldviewpoint)
  {
    viewpoint = d_scene->bindableViewpointTop();
    return;
  }

  // or a NavigationInfo
  VrmlNodeNavigationInfo* oldnavinfo = NodeDynamicCast<VrmlNodeNavigationInfo*>(bindable);
  if (oldnavinfo != NULL && navinfo == oldnavinfo)
  {
    navinfo = d_scene->bindableNavigationInfoTop();
    return;
  }

  // or a Background
  VrmlNodeBackground* oldbackground = NodeDynamicCast<VrmlNodeBackground*>(bindable);
  if (oldbackground != NULL && background == oldbackground)
  {
    background = d_scene->bindableBackgroundTop();
    return;
  }

  // or a Fog
  VrmlNodeFog* oldfog = NodeDynamicCast<VrmlNodeFog*>(bindable);
  if (oldfog != NULL && fog == oldfog)
    fog = d_scene->bindableFogTop();
}



//---------------------------------------------------------------------------
// world reloading callbacks
//---------------------------------------------------------------------------



void Viewer::onWorldReloadStarting()
{
  // store the cached pointers in the persistent way
  d_persistentInfo = new ViewerPersistentInfo();
  if (viewpoint == NULL) d_persistentInfo->viewpointIndex = -1;
  else
  {
    d_persistentInfo->viewpointName = viewpoint->name();
    d_persistentInfo->viewpointDesc = viewpoint->description();
    d_persistentInfo->viewpointIndex = d_scene->getViewpointIndex(viewpoint);
  }
  if (background == NULL) d_persistentInfo->backgroundIndex = -1;
  else
  {
    d_persistentInfo->backgroundName = background->name();
    d_persistentInfo->backgroundIndex = -1;        // not yet implemented
  }
  if (navinfo == NULL) d_persistentInfo->navinfoIndex = -1;
  else
  {
    d_persistentInfo->navinfoName = navinfo->name();
    d_persistentInfo->navinfoIndex = -1;        // not yet implemented
  }
  if (fog == NULL) d_persistentInfo->fogIndex = -1;
  else
  {
    d_persistentInfo->fogName = fog->name();
    d_persistentInfo->fogIndex = -1;        // not yet implemented
  }
  d_persistentInfo->camera = getCamera();
  d_persistentInfo->cameraAbs = getCameraAbsolute();
  d_persistentInfo->cameraBound = d_cameraBound;

  // discard the cached pointers
  viewpoint = NULL;
  navinfo = NULL;
  background = NULL;
  fog = NULL;

  // do not attempt to redraw scene until it is rebuilt
  d_renderSuppressed = true;
}



void Viewer::onWorldReloadComplete(bool preserveBindings)
{
  if (preserveBindings)
  {
    // restore the viewpoint if possible (note that viewpoint is NOT searched
    // by index, and if it is not found, it is kept unset).
    viewpoint = d_scene->findViewpointByDescription(
      d_persistentInfo->viewpointDesc.c_str());
    if (viewpoint == NULL)
      viewpoint = d_scene->findViewpointByName(
	d_persistentInfo->viewpointName.c_str());

    // navigation info, background and fog are not restored by now, they are
    // simply loaded from defaults
    navinfo = d_scene->bindableNavigationInfoTop();
    background = d_scene->bindableBackgroundTop();
    fog = d_scene->bindableFogTop();

    // load camera
    if (viewpoint == NULL) d_camera = d_persistentInfo->cameraAbs;
    else
    {
      if (d_persistentInfo->cameraBound)
	d_camera = d_persistentInfo->camera;
      else
	d_camera = d_persistentInfo->cameraAbs;
      d_cameraBound = d_persistentInfo->cameraBound;
    }

    // free the persistent info
    delete d_persistentInfo;
    d_persistentInfo = NULL;
  }
  else
  {
    viewpoint = d_scene->bindableViewpointTop();
    navinfo = d_scene->bindableNavigationInfoTop();
    background = d_scene->bindableBackgroundTop();
    fog = d_scene->bindableFogTop();

    setNavMode();

    if (navinfo)
      setHeadlight(navinfo->headlightOn());
  }

  // allow scene redraws
  d_renderSuppressed = false;

  // redraw it all
  requestRedraw();
}



void Viewer::onWorldLoadStarting()
{
  // reset cached node pointers
  viewpoint  = NULL;
  background = NULL;
  fog        = NULL;
  navinfo    = NULL;

  // do not attempt to redraw scene until it is rebuilt
  d_renderSuppressed = true;
}



void Viewer::onWorldLoadComplete()
{
  // get node pointers from defaults
  viewpoint  = d_scene->bindableViewpointTop();
  background = d_scene->bindableBackgroundTop();
  fog        = d_scene->bindableFogTop();
  navinfo    = d_scene->bindableNavigationInfoTop();

  // return to base viewpoint and reset navigation
  resetUserNavigation();

  // allow scene redraws
  d_renderSuppressed = false;
}



//---------------------------------------------------------------------------
// The camera
//---------------------------------------------------------------------------

void Viewer::setCameraChanged()
{
  d_cameraChanged = true;
  requestRedraw();
}

Camera Viewer::getCameraAbsolute() const
{
  Camera cam = d_camera;
  if (viewpoint && d_cameraBound)
    cam.transform(viewpoint->getCompoundTransformationMatrix());
  return cam;
}

void Viewer::setCameraAbsolute(const Camera &cam)
{
  Matrix4x4 vm;
  if (viewpoint && d_cameraBound)
    vm = viewpoint->getCompoundInverseTransformationMatrix();
  else
    vm.setToIdentity();
  Camera localCam = cam;
  localCam.transform(vm);
  setCamera(localCam);
}

void Viewer::getPosition(float *x, float *y, float *z)
{
  *x = d_camera.getPosition().getX();
  *y = d_camera.getPosition().getY();
  *z = d_camera.getPosition().getZ();
}

void Viewer::getOrientation(float *orientation)
{
  Vector3 axis = d_camera.getYawAxis();
  orientation[0] = axis.getX();
  orientation[1] = axis.getY();
  orientation[2] = axis.getZ();
  orientation[3] = d_camera.getYawAngle();
}

void Viewer::bindCamera(bool onoff)
{
  // do nothing if the camera is already bound/unbound
  if (onoff == d_cameraBound) return;

  if (onoff)
  {
    // convert camera to local coordinates of the current viewpoint (if any)
    Matrix4x4 m;
    if (viewpoint) m = viewpoint->getCompoundInverseTransformationMatrix();
    else m.setToIdentity();
    d_camera.transform(m);
    d_cameraBound = true;
  }
  else
  {
    // convert camera to global coordinates
    Matrix4x4 m;
    if (viewpoint) m = viewpoint->getCompoundTransformationMatrix();
    else m.setToIdentity();
    d_camera.transform(m);
    d_cameraBound = false;
  }
  setCameraChanged();
}

void Viewer::gotoViewpoint(VrmlNodeViewpoint* vp)
{
  if (vp!=NULL && vp->scene()==d_scene)
  {
    // set the viewpoint
    viewpoint = vp;

    // if this is the primary viewer, set the viewpoint also in the scene
    if (isPrimary())
      d_scene->setViewpoint(vp);
  }

  // clear the camera offset from the viewpoint and request redrawing
  resetUserNavigation();
}

void Viewer::leaveViewpoint()
{
  // convert camera to global coordinates
  Matrix4x4 m;
  if (viewpoint != NULL) m = viewpoint->getCompoundTransformationMatrix();
  else m.setToIdentity();
  d_camera.transform(m);

  // leave the viewpoint
  viewpoint = NULL;
  setCameraChanged();
}

void Viewer::gotoViewpoint(int n)
{
  if (d_scene != NULL) gotoViewpoint(d_scene->findViewpointByIndex(n));
}

void Viewer::gotoPrevViewpoint()
{
  if (d_scene != NULL) gotoViewpoint(d_scene->getPrevViewpoint(viewpoint));
}

void Viewer::gotoNextViewpoint()
{
  if (d_scene != NULL) gotoViewpoint(d_scene->getNextViewpoint(viewpoint));
}

void Viewer::pushCameraState()
{
  d_cameraStateStack.push(d_camera);
}

bool Viewer::popCameraState()
{
  if (d_cameraStateStack.empty()) return false;
  d_camera = d_cameraStateStack.top();
  d_cameraStateStack.pop();
  setCameraChanged();
  return true;
}

void Viewer::clearCameraStateStack()
{
  d_cameraStateStack.clear();
}

void Viewer::resetUserNavigation()
{
  d_camera.reset();
  d_cameraBound = true;
  clearCameraStateStack();

  // special case: if no viewpoint is defined in the scene,
  // use (0, 0, 10) as position (see VRML specification)
  if (d_scene && d_scene->bindableViewpointTop() == NULL)
    d_camera.setPosition(Point3(0, 0, 10));

  setCameraChanged();
}

void Viewer::step(double distance)
  { d_camera.step(distance); setCameraChanged(); }

void Viewer::slideH(double distance)
  { d_camera.slideH(distance); setCameraChanged(); }

void Viewer::slideV(double distance)
  { d_camera.slideV(distance); setCameraChanged(); }

void Viewer::rotateH(double angle)
  { d_camera.rotateH(angle); setCameraChanged(); }

void Viewer::rotateV(double angle)
  { d_camera.rotateV(angle); setCameraChanged(); }

void Viewer::rotateS(double angle)
  { d_camera.rotateS(angle); setCameraChanged(); }

void Viewer::orbitH(const Point3 &p, double angle)
  { d_camera.orbitH(p, angle); setCameraChanged(); }

void Viewer::orbitV(const Point3 &p, double angle)
  { d_camera.orbitV(p, angle); setCameraChanged(); }

double Viewer::getRecommendedMovementStep() const
{
  double result = 1;
  if (navinfo != NULL)
    result = navinfo->speed();
  return result;
}
