/*****************************************************************************
******************************************************************************
*
*               Abstract viewer base class
*
*               $Revision: 1.1.2.8 $
*               $Date: 2010/12/22 10:58:48 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright 1998-2010 HUMUSOFT s.r.o.
// Portions Copyright (C) 1998  Chris Morley

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#ifndef VIEWER_H
#define VIEWER_H

#include <stddef.h>
#include <stack>
#include <string>

class VrmlScene;
class VrmlBVolume;
class VrmlBSphere;
class VrmlAABox;

class VrmlNode;
class VrmlNodeViewpoint;
class VrmlNodeNavigationInfo;
class VrmlNodeBackground;
class VrmlNodeFog;

#include "VrmlFrustum.h"
#include "Math3d.h"
#include "Camera.h"
#include "System.h"  // OPENVRMLDLLEXPORT

#include "config.h"  // M_PI_x

#include "VrmlViewingSector.h"

/**
 * Modifications of standard stack and queue classes.
 *
 */
template<typename T> class clearable_stack : public std::stack<T>
{
public:
  void clear()
  { this->c.clear(); }
};


struct ViewerPersistentInfo;

// Maximum viewpoint fieldOfView allowed to render.
// We need this because of the Zoom function.
// Despite the VRML 97 standard allows FoV in (0;pi), user should not be allowed
// to zoom the scene out beyond some threshold that assures for most scenes
// still not totally distorted view.
// 3/4*pi
static const double maxFoV = M_PI_4 * 3.0;


/**
 * Map the scene graph to the underlying graphics library.
 *
 */
class OPENVRMLDLLEXPORT Viewer
{
friend class BindableStack;
friend class VrmlScene;

protected:

  /** Constructor.
    * @param scene Pointer to scene this viewer should initially look at
    * (NULL if no scene).
    */
  Viewer(VrmlScene *scene = 0);

public:

  virtual ~Viewer();

  /** Sets a new scene for this viewer to view.
    * @param scene Pointer to the scene (NULL if no scene).
    */
  virtual void setScene(VrmlScene* scene);

  // Options flags
  enum {
    MASK_NONE	= 0,
    MASK_CCW	= 1,
    MASK_CONVEX	= 2,
    MASK_SOLID	= 4,
    MASK_BOTTOM = 8,
    MASK_TOP	= 16,
    MASK_SIDE	= 32,
    MASK_COLOR_PER_VERTEX  = 64,
    MASK_NORMAL_PER_VERTEX = 128
  };

  /** Navigation modes. */
  enum NavMode
  {
      /** Walk (planar motion with gravity). */
      NAVMODE_WALK,

      /** Examine (orbit around the scene). */
      NAVMODE_EXAMINE,

      /** Fly (unlimited motion). */
      NAVMODE_FLY,

      /** No navigation (viewpoints only). */
      NAVMODE_NONE
  };

  // Object and texture keys. Don't mix them up.
  typedef long Object;
  typedef long TextureObject;

  //
  VrmlScene *scene() { return d_scene; }

  //-------------------------------------------------------------------------
  // Callbacks invoked from scene
  //-------------------------------------------------------------------------

  public:

    /**
     * This method is called when the world is going to be completely
     * cleared and loaded from another file. All pointers to nodes must be
     * considered invalid.
     */
    virtual void onWorldLoadStarting();

    /**
     * This method is called after the world loading is complete.
     * Typically the navigation should be reset.
     */
    virtual void onWorldLoadComplete();

  protected:

    /**
     * This method is called when the viewpoint list changes
     * (a viewpoint was added or removed).
     * @param gone True if the current viewpoint of this viewer
     * was removed; false if the change did not affect this viewpoint.
     * @note The default implementation of this method
     * returns to the default viewpoint if the current viewpoint
     * is gone, otherwise it does nothing.
     */
    virtual void onViewpointListChange(bool gone)
      { if (gone) leaveViewpoint(); }

    /**
     * This method is called when a bindable node (e.g. a viewpoint)
     * is bound.
     * @param bindable A pointer to the bindable node.
     */
    virtual void onBindableBound(VrmlNode* bindable);

    /**
     * This method is called when a bindable node
     * (e.g. a viewpoint) is unbound.
     * @param bindable A pointer to the bindable node.
     */
    virtual void onBindableUnbound(VrmlNode* bindable);

    /**
     * This method is called when the world is going to be reloaded.
     * This means that all pointers to nodes will be invalidated
     * and if some node reference should persist, it must be backed up
     * using node names or indexes.
     */
    virtual void onWorldReloadStarting();

    /**
     * This method is called when the world reload is complete.
     * Old node pointers must not be used anymore.
     */
    virtual void onWorldReloadComplete(bool preserveBindings = true);

    /**
     * This method is called when recording starts.
     * Called from VrmlScene::setTime(). Does nothing by default.
     */
    virtual void onRecordingStart() {};

    /**
     * This method is called when recording stops.
     * The viewer should stop recording after the upcoming frame
     * is rendered (not earlier because the upcoming frame still can be
     * recorded, and onRecordingFrame() may be called for it).
     * Called from VrmlScene::setTime(). Does nothing by default.
     */
    virtual void onRecordingStop() {};

    /**
     * This method is called when the upcoming animation frame should be
     * recorded.
     * Called from VrmlScene::setTime(). Does nothing by default.
     */
    virtual void onRecordingFrame() {};

    virtual void onRecordToVrmlFlagChanged() {};

    /**
     * This method is called when a new time is set to the scene.
     * Called from VrmlScene::setTime(). Does nothing by default.
     */
    virtual void onSceneTimeChange() {};

  //-------------------------------------------------------------------------
  // The camera
  //-------------------------------------------------------------------------

  private:

    /** The camera. */
    Camera d_camera;

    /**
     * Whether the camera is bound to the current viewpoint.
     * If this is true, the camera position and orientation are considered
     * to be relative to the current viewpoint.
     * If this is false, the camera position and orientation are considered
     * to be in world coordinates and the current viewpoint does not
     * affect it.
     */
    bool d_cameraBound;

    /** Whether the camera parameters have changed from the last redraw. */
    bool d_cameraChanged;

    /** The stack of remembered camera states. */
    clearable_stack<Camera> d_cameraStateStack;

    // existence flags
    bool valid;
    bool markedForDestruction;

  protected:

    /**
     * Marks the camera as not changed.
     * This is typically done after a redraw or camera recalculation.
     */
    void unsetCameraChanged() { d_cameraChanged = false; }

  public:

    /**
     * Checks whether the camera state changed from the last redraw.
     * @return True if the camera changed, false if it did not.
     */
    bool isCameraChanged() const { return d_cameraChanged; }

    /**
     * Marks the camera as changed, forcing recalculation of the related
     * information.
     */
    void setCameraChanged();

    /**
     * Gets the camera state.
     * @return The camera state.
     */
    const Camera& getCamera() const { return d_camera; }

    /**
     * Sets the camera state.
     * @param cam The new camera state.
     */
    void setCamera(const Camera &cam)
      { d_camera = cam; setCameraChanged(); }

    /**
     * Gets the camera state in absolute coordinates.
     * @return The camera state.
     */
    Camera getCameraAbsolute() const;

    /**
     * Sets the camera state so that its absolute coordinates
     * are as specified (i.e. converts the absolute camera coordinates
     * to the local coordinates and sets them).
     * @param cam The absolute camera state.
     */
    void setCameraAbsolute(const Camera &cam);

    /**
     * Gets the camera position.
     * @param x A pointer to variable to get the X-coordinate.
     * @param y A pointer to variable to get the Y-coordinate.
     * @param z A pointer to variable to get the Z-coordinate.
     * @deprecated New code should use getCamera() instead.
     */
    void getPosition(float *x, float *y, float *z);

    /**
     * Gets the camera orientation.
     * @param orientation A pointer to a 4-element array to receive
     * the orientation elements (x, y, z, angle).
     * @deprecated New code should use getCamera() instead.
     */
    void getOrientation(float *orientation);

    /**
     * Calculates the matrix representing the current viewing transformation
     * (representing the current viewpoint and camera state).
     * @returns The viewing transformation matrix.
     */
    Matrix4x4 getViewingTransformation() const
      { return getCameraAbsolute().computeViewingMatrix(); }

    /**
     * Checks whether the camera is bound to the current viewpoint
     * (i.e. camera position and orientation are relative to the viewpoint).
     * @returns True if the camera is bound, false if it is not.
     */
    bool isCameraBound() const { return d_cameraBound; }

    /**
     * Binds or unbinds the camera to the current viewpoint.
     * If the camera is bound to the viewpoint, its movement and orientation
     * are relative to that viewpoint. If the camera is not bound,
     * its movement and orientation are in world coordinate system.
     * @param onoff True to bind the camera, false to unbind it.
     */
    void bindCamera(bool onoff);

    /**
     * Returns a pointer to the current viewpoint.
     * @return The viewpoint pointer (may be NULL).
     */
    VrmlNodeViewpoint* getViewpoint() { return viewpoint; }

    /**
     * Moves the camera to the given viewpoint.
     * @param vp The viewpoint.
     * @note The method has no effect if the viewpoint does not belong
     * to the scene, or if the viewpoint pointer is NULL.
     */
    void gotoViewpoint(VrmlNodeViewpoint* vp);

    /**
     * Leaves the current viewpoint without assigning a new one.
     * The camera stays on its position but is placed to the global
     * coordinate system.
     */
    void leaveViewpoint();

    /**
     * Moves the camera to the given viewpoint.
     * @param n The viewpoint index.
     * @note The method has no effect if the viewpoint index
     * does not refer to any viewpoint in the scene.
     */
    void gotoViewpoint(int n);

    /**
     * Moves the camera to the previous viewpoint
     * in the circular viewpoint list.
     */
    void gotoPrevViewpoint();

    /**
     * Moves the camera to the next viewpoint
     * in the circular viewpoint list.
     */
    void gotoNextViewpoint();

    /**
     * Pushes the camera state to the undo stack.
     */
    void pushCameraState();

    /**
     * Pops the last camera state from the stack and sets it.
     * @return True on success, false on failure (when there are
     * no states on the stack).
     */
    bool popCameraState();

    /**
     * Clears the whole camera state stack.
     */
    void clearCameraStateStack();

    /**
     * Resets the camera to the last bound viewpoint.
     */
    virtual void resetUserNavigation();

    /**
     * Moves the camera forwards or backwards by the given distance.
     * @param distance The distance (negative to step backwards).
     */
    void step(double distance);

    /**
     * Slides the camera horizontally (to the left or to the right)
     * by the given distance.
     * @param distance The distance (positive to move to the right,
     * negative to move to the left).
     */
    void slideH(double distance);

    /**
     * Slides the camera vertically (up or down) by the given distance.
     * @param distance The distance (positive to move to the right,
     * negative to move to the left).
     */
    void slideV(double distance);

    /**
     * Rotates the camera in the horizontal plane (to the left
     * or to the right) by the given angle.
     * @param angle The angle.
     */
    void rotateH(double angle);

    /**
     * Rotates the camera in the vertical plane (up or down)
     * by the given angle.
     * @param angle The angle.
     */
    void rotateV(double angle);

    /**
     * Rotates the camera in the screen plane by the given angle.
     * @param angle The angle.
     */
    void rotateS(double angle);

    /**
     * Rotates the camera around the given center in the horizontal plane
     * by the given angle.
     * @param center The center point.
     * @param angle The angle.
     */
    void orbitH(const Point3 &p, double angle);

    /**
     * Rotates the camera around the given center in the vertical plane
     * by the given angle.
     * @param center The center point.
     * @param angle The angle.
     */
    void orbitV(const Point3 &p, double angle);

    /**
     * Returns the recommended step distance for easy movement
     * in the scene. The resulting value may depend on the size
     * of the scene and navigation info hints.
     * <p>
     * Child classes can override this method to provide more
     * custom-tailored values, e.g. to allow the user to switch
     * speeds.
     * @return The recommended step value.
     */
    virtual double getRecommendedMovementStep() const;

  //-------------------------------------------------------------------------
  // Rendering
  //-------------------------------------------------------------------------

  /**
   * Requests the underlying graphics subsystem to redraw.
   */
  virtual void requestRedraw() = 0;

  enum RenderMode {
    RENDER_MODE_DRAW,
    RENDER_MODE_PICK
  };

  // Return renderMode
  virtual RenderMode getRenderMode() = 0;


  /**
   * Render phases. When the scene is rendered, all non-transparent objects
   * are rendered first (non-transparent phase), then transparent ones
   * with Z-buffer set to readonly (transparent phase). This gives us
   * a quite good illusion of correct translucency :o)
   */
  enum RenderPhase {

    /** No rendering is currently occuring. */
    RENDER_PHASE_NONE,

    /** We are not really drawing but analyzing the scene. */
    RENDER_PHASE_ANALYSIS,

    /** Only non-transparent objects are drawn in this phase. */
    RENDER_PHASE_NON_TRANSPARENT,

    /** Only transparent objects are drawn in this phase. */
    RENDER_PHASE_TRANSPARENT,

    /** Only lights which illuminate the whole scene are drawn in this phase. */
    RENDER_PHASE_LIGHTS
  };

  /**
   * Returns the render phase the viewer is currently in (returns
   * RENDER_PHASE_NONE if the viewer is not rendering).
   */
  RenderPhase getRenderPhase() { return d_renderPhase; };

  virtual double getFrameRate() = 0;

  /**
   * Return user's navigation from the last bound viewpoint as a
   * MathUtils-format matrix. The navigation transformation should be
   * orthonormal.
   *
   * @param M returns navigation matrix in MathUtils format
   */
  virtual void getUserNavigation(double M[4][4]) = 0;

  // Open/close display lists
  virtual Object beginObject(const char *, bool = false) = 0;
  virtual void endObject() = 0;

  // Insert objects into the display list
  virtual Object insertBackground(size_t nGroundAngles = 0,
				  const float * groundAngle = 0,
				  const float * groundColor = 0,
				  size_t nSkyAngles = 0,
				  const float * skyAngle = 0,
				  const float * skyColor = 0,
				  int* whc = 0,
                                  unsigned char ** pixels = 0,
                                  Object texturesDisplayList = 0) = 0;

  virtual Object insertBox(float, float, float ) = 0;
  virtual Object insertCone(float, float, bool, bool) = 0;
  virtual Object insertCylinder(float, float, bool, bool, bool) = 0;

  virtual Object insertElevationGrid(unsigned int mask,
				     size_t nx,
				     size_t nz,
				     const float * height,
				     float dx,
				     float dz,
				     const float * tc,
				     const float * normals,
				     const float * colors) = 0;

  virtual Object insertExtrusion(unsigned int,
				 size_t nOrientation,
				 const float * orientation,
				 size_t nScale,
				 const float * scale,
				 size_t nCrossSection,
				 const float * crossSection,
				 size_t nSpine,
				 const float * spine,
                                 float creaseAngle) = 0;

  virtual Object insertLineSet(size_t nCoords, const float * coord,
			       size_t nCoordIndex, const long * coordIndex,
			       bool colorPerVertex,
			       const float * color,
			       size_t nColorIndex,
			       const long * colorIndex) = 0;

  virtual Object insertPointSet(size_t nv, const float * v, const float * c) = 0;

  virtual Object insertShell(unsigned int mask,
			     size_t npoints, const float * points,
			     size_t nfaces, const long * faces,
			     const float * tc,
			     size_t ntci, const long * tci,
			     const float * normal,
			     size_t nni, const long * ni,
			     const float * color,
			     size_t nci, const long * ci) = 0;

  virtual Object insertSphere(float /*radius*/) = 0;

  /** Font family identifiers. */
  enum FontFamily
  {
    FONT_FAMILY_SERIF = 0,
    FONT_FAMILY_SANS,
    FONT_FAMILY_TYPEWRITER
  };

  /** Font style identifiers. */
  enum FontStyle
  {
    FONT_STYLE_PLAIN = 0,
    FONT_STYLE_BOLD,
    FONT_STYLE_ITALIC,
    FONT_STYLE_BOLD_ITALIC
  };

  /**
   * Inserts a set of text lines to the scene.
   * @param family The font family (one of Viewer::FontFamily values).
   * @param style The font style (one of Viewer::FontStyle values).
   * @param justifications Array of values specifying how the text lines
   * should be justified (-1 to justify to the left, 0 to center,
   * +1 to justify to the right).
   * @param size Text height (in local coordinates).
   * @param spacing Text spacing (in local coordinates).
   * @param n Number of text lines in the array.
   * @param s Array of text lines, in form of C-style strings.
   */
  virtual Object insertText(FontFamily family,
                            FontStyle style,
                            int* justifications,
                            float size,
                            float spacing,
                            int n,
                            char const * const *s) = 0;

  // Lights
  virtual Object insertDirLight(float, float, const float [], const float []) = 0;

  virtual Object insertPointLight(float, const float [], const float [],
				  float, const float [], float ) = 0;

  virtual Object insertSpotLight( float /*ambientIntensity*/ ,
				  const float /*attenuation*/ [],
				  float /*beamWidth*/ ,
				  const float /*color*/ [],
				  float /*cutOffAngle*/ ,
				  const float /*direction*/ [],
				  float /*intensity*/ ,
				  const float /*location*/ [],
				  float /*radius*/ ) = 0;

  // Lightweight copy
  virtual Object insertReference(Object /*existingObject*/) = 0;

  // Remove an object from the display list
  virtual void removeObject(Object) = 0;

  virtual void enableLighting(bool) = 0;

  // Set attributes
  virtual void setFog(const float * /*color*/,
		      float   /*visibilityRange*/,
		      const char * /*fogType*/) = 0;

  virtual void setColor(float r, float g, float b, float a = 1.0) = 0;

  virtual void setMaterial(float, const float[],
			   const float[], float,
			   const float[], float) = 0;

  virtual void setMaterialMode( int nTexComponents, bool geometryColor ) = 0;

  virtual void setSensitive(void *object) = 0;


protected:
  /** Maximum texture size and hardware texture size */
  int d_maxTextureSize;

  /** Current navigation mode (one of the NavMode constants). */
  int d_navMode;

public:

  void setValid(bool arg = true) { valid = arg; };
  bool isValid() { return(valid); };

  /**
   * Returns the current navigation mode (one of the NavMode constants).
   * @returns Navigation mode currently set.
   */
  int getNavMode() { return d_navMode; };

  /**
   * Sets the navigation mode.
   * @param mode The new navigation mode (one of the NavMode constants).
   */
  virtual void setNavMode(int mode);

  /**
   * Reads navigation mode from scene and Sets it
   */
  virtual void setNavMode();

  /**
   * Sets headlight to viewer.
   */
  virtual void setHeadlight(bool headlight);

  /**
   * Return maximum texture size for this viewer.
   */
  int getMaxTextureSize()
  { return d_maxTextureSize; };

  /**
   * Set maximum texture size for this viewer.
   */
  virtual void setMaxTextureSize(int texsize)
  { d_maxTextureSize = texsize; };


  // Create a texture object
  virtual TextureObject insertTexture(size_t /*w*/, size_t /*h*/, size_t /*nc*/,
				      bool /*repeat_s*/,
				      bool /*repeat_t*/,
				      const unsigned char * /*pixels*/,
				      bool retainHint = false) = 0;

  // Reference/remove a texture object
  virtual void insertTextureReference(TextureObject, int) = 0;
  virtual void removeTextureObject(TextureObject) = 0;

  virtual void setTextureTransform(const float center[2],
				   float rotation,
				   const float scale[2],
				   const float translation[2]) = 0;

  virtual void setTransform(const float center[3],
			    const float rotation[4],
			    const float scale[3],
			    const float scaleOrientation[4],
			    const float translation[3]) = 0;

  // This is a hack to work around the glPushMatrix() limit (32 deep on Mesa).
  // It has some ugly disadvantages: it is slower and the resulting transform
  // after a setTransform/unsetTransform may not be identical to the original.
  // It might be better to just build our own matrix stack...
  virtual void unsetTransform(const float center[3],
			      const float rotation[4],
			      const float scale[3],
			      const float scaleOrientation[4],
			      const float translation[3]) = 0;

  virtual void setBillboardTransform(const float axisOfRotation[3]) = 0;

  virtual void unsetBillboardTransform(const float axisOfRotation[3]) = 0;

  virtual void setViewpoint(const float * /*position*/,
			    float * /*orientation*/,
			    float /*fieldOfView*/,
			    float /*avatarSize*/,
			    float /*visLimit*/) = 0;

  // The viewer knows the current viewpoint
  virtual void transformPoints(int nPoints, float *points) = 0;

  // still working on some navigation api issues, so don't depend on
  // thses yet. there's a default implementation in any case, so you
  // shouldn't have to worry about it.
  //
  virtual const VrmlFrustum& getFrustum() const;
  virtual void setFrustum(const VrmlFrustum& afrust);

  const VrmlViewingSector& getViewingSector() const { return d_sect; }
  void setViewingSector(const VrmlViewingSector& asect) { d_sect = asect; }

  /**
   * Intersect the given bounding volume with the view volume. This
   * goes into the viewer to provide a hook for systems that use
   * non-standard view volumes. Most subclasses should be able to use
   * the default implementation provided here. If your view volume is
   * so strange that there's no way to cull to is, then reimplement to
   * always return BV_INSIDE
   *
   * @param bv the bounding volume to intersect with the view volume
   * @return VrmlBVolume::INSIDE, OUTSIDE, or PARTIAL
   */
  virtual int isectViewVolume(const VrmlBVolume& bv) const;

  /**
   * Draw a bounding sphere. Used for debugging view culling. Probably
   * should be drawBVolume and handle aaboxes as well.
   *
   * @param bs a bounding sphere, if max, will not be drawn
   * @param flag one of the bvolume intersection test constants, or 4
   *             to draw in unique way. (useful for debugging)
   */
  virtual void drawBSphere(const VrmlBSphere& bs, int flag) = 0;

  /** Checks whether this viewer is a primary viewer for its scene.
    * @return True if the viewer is primary, false otherwise.
    */
  bool isPrimary() const;

  /** Sets this viewer as the current viewer. */
  void setCurrent();

  /** Returns the current viewer. */
  static Viewer* getCurrent();

private:

  /** Viewer that has current focus. */
  static Viewer* d_currentViewer;

protected:

  static void computeCylinder(double height,
		       double radius,
		       int numFacets,
		       float c[][3],
		       float tc[][3],
		       int faces[]);

  static void computeExtrusion( int  /*nOrientation*/,
			 const float * /*orientation*/,
			 int  /*nScale*/,
			 const float * /*scale*/,
			 int  /*nCrossSection*/,
			 const float * /*crossSection*/,
			 int   /*nSpine*/,
			 const float * /*spine*/,
			 float *c,   // OUT: coordinates
			 float *tc,  // OUT: texture coords
			 int *faces);    // OUT: face list

  static void computeSphere(double radius,
		     int numLatLong,
		     float c[][3],
		     float tc[][3],
		     int faces[]);

  static void computeView(const float position[3],
		   float orientation[4],	// [3] to [4] changed by __ORBISNAP__
		   float distance,
		   float target[3],
		   float up[3]);

  VrmlFrustum d_frust;
  VrmlViewingSector d_sect;

  /** A pointer to the scene. */
  VrmlScene* d_scene;

  /**
   * If true, the scene must not be rendered, probably because the scene
   * is being rebuilt and may not be ready. If the viewer is requested to redraw
   * and this field is set, it should draw only black screen (or some logo etc.).
   */
  bool d_renderSuppressed;

  /**
   * This field determines whether the headlight is enabled.
   * If this is false, the headlight will never appear.
   * If it is true, the headlight will be drawn if it is specified
   * by the world.
   */
  bool d_headlight;

  /** Current render phase. */
  RenderPhase d_renderPhase;

  /** Cached pointer to viewpoint node currently used. */
  VrmlNodeViewpoint* viewpoint;

  /** Cached pointer to navigation node currently used. */
  VrmlNodeNavigationInfo* navinfo;

  /** Cached pointer to background node currently used. */
  VrmlNodeBackground* background;

  /** Cached pointer to fog node currently used. */
  VrmlNodeFog* fog;

  /** Information kept during reload. */
  ViewerPersistentInfo* d_persistentInfo;

};



/**
 * This structure is used to hold information about nodes
 * the viewer was bound to, in form that can survive a reload.
 */
struct ViewerPersistentInfo
{
  std::string viewpointName;
  std::string viewpointDesc;
  int viewpointIndex;
  std::string backgroundName;
  int backgroundIndex;
  std::string navinfoName;
  int navinfoIndex;
  std::string fogName;
  int fogIndex;
  Camera camera;
  Camera cameraAbs;
  bool cameraBound;
};



#endif // VIEWER_H
