/*****************************************************************************
******************************************************************************
*
*               OpenGL viewer
*
*               $Revision: 1.1.4.1.2.36 $
*               $Date: 2012/02/21 11:35:03 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright 1998-2012 HUMUSOFT s.r.o.
// Portions Copyright (C) 1998  Chris Morley

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#include "wx/dcmemory.h"
#include "wx/image.h"

#if defined(__WXMSW__)
#include "wx/glcanvas.h"
#include <GL/glu.h>

#elif defined(__APPLE__)
#include <gl.h>
#include <glu.h>

#else // Linux
#include <GL/gl.h>
#include <GL/glu.h>
#endif

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "ViewerOpenGL.h"

#include <stdlib.h>		// atof
#include <math.h>

#include "MathUtils.h"
#include "System.h"
#include "VrmlScene.h"
#include "VrmlNodeNavigationInfo.h"
#include "VrmlBSphere.h"
#include "VrmlFrustum.h"

#include "OpenGLEvent.h"

#include "VrmlNodeViewpoint.h"
#include "VrmlNodeBackground.h"
#include "VrmlNodeExtrusion.h"
#include "VrmlNodeFog.h"
#include "VrmlNamespace.h"

#include "VrmlViewingSector.h"

// supported in OpenGL 1.2 and higher
#ifndef GL_CLAMP_TO_EDGE
#define GL_CLAMP_TO_EDGE 0x812F
#endif

// Put geometry into display lists.
// If geometry is not in display lists, performance will suffer,
// especially for non-convex geometries. You probably shouldn't
// change these unless you know what you are doing.

#define USE_GEOMETRY_DISPLAY_LISTS 1

// Textures are now done using OGL1.1 bindTexture API rather than
// display lists when this flag is set. Don't define this if you
// are still at OpenGL 1.0 (or get a newer OpenGL).

#define USE_TEXTURE_DISPLAY_LISTS 1

// if a texture size is not power of 2 and it is smaller
// in both directions than this constant, it is then upscaled by pixel
// replication algorithm (via wxImage)
#define PIXEL_REPLICATION_LIMIT	32

// if a texture is upscaled by pixel replication algorithm,
// the resulting size of such image is PIXEL_REPLICATION_SIZE_LIMIT x PIXEL_REPLICATION_SIZE_LIMIT
#define PIXEL_REPLICATION_SIZE 128


static void
matrix_to_glmatrix(double M[4][4], float GLM[16])
{
  GLM[0]  = M[0][0];
  GLM[1]  = M[1][0];
  GLM[2]  = M[2][0];
  GLM[3]  = M[3][0];
  GLM[4]  = M[0][1];
  GLM[5]  = M[1][1];
  GLM[6]  = M[2][1];
  GLM[7]  = M[3][1];
  GLM[8]  = M[0][2];
  GLM[9]  = M[1][2];
  GLM[10] = M[2][2];
  GLM[11] = M[3][2];
  GLM[12] = M[0][3];
  GLM[13] = M[1][3];
  GLM[14] = M[2][3];
  GLM[15] = M[3][3];
}

// initialize time intervals for checking sensitivities
const double ViewerOpenGL::MOUSE_MOVE_CHECK_SENSITIVE_INTERVAL = 0.1;
const double ViewerOpenGL::MOUSE_TIME_GENERATED_CHECK_SENSITIVE_INTERVAL = 0.25;

//  Construct a viewer for the specified scene. I'm not happy with the
//  mutual dependencies between VrmlScene/VrmlNodes and Viewers...
//  Maybe a static callback function pointer should be passed in rather
//  than a class object pointer. Currently, the scene is used to access
//  the VrmlScene::render() method. Also, the static VrmlScene::update
//  is called from the idle function. A way to pass mouse/keyboard sensor
//  events back to the scene is also needed.

ViewerOpenGL::ViewerOpenGL(VrmlScene *scene) : Viewer(scene),
  d_glVersion(1.1),
  d_GLinitialized(false), d_blend(false), d_lit(true), d_texture(true), d_dragging(false),
  d_wireframe(false), d_nObjects(0), d_nestedObjects(0), d_nSensitive(0),
  d_activeSensitive(0), d_overSensitive(0), d_selectMode(false),
  d_selectZ(0.0), d_tess(0),
  d_scale(1.0), d_translatex(0.0), d_translatey(0.0), d_translatez(0.0),
  d_rotating(false), d_scaling(false), d_translating(false),
  d_reportFPS(false), d_renderTime(1.0), d_renderTime1(1.0),
  d_drawBSpheres(false), d_cull(true), d_bottomReserved(0),
  d_antialias(false),
  d_motionStep(1), d_rotationStep(1), d_sensorCheckTime(-1),
  d_zNear(0.1), d_zFar(1000.0),
  planeClipMethod(PLANECLIP_OPENGL),
  d_maxTextureSizeHW(65536),
  d_mouseXCoord(0), d_mouseYCoord(0),
  checkSensitiveDisabled(false),
  d_selectionBufferSize(SELECTION_BUFFER_SIZE),
  width(0), height(0), newwidth(0), newheight(0)
{
  d_background[0] = d_background[1] = d_background[2] = 0.0;

  for (int i=0; i<MAX_LIGHTS; ++i)
    d_lightInfo[i].lightType = LIGHT_UNUSED;

  //d_rotationChanged = true;
  trackball(d_curquat, 0.0, 0.0, 0.0, 0.0);
}



ViewerOpenGL::~ViewerOpenGL()
{
#if GLU_VERSION_1_2
  if (d_tess) gluDeleteTess( d_tess );
#endif

  // unregister the viewer with scene
  // this must be called here before ViewerOpenGL is completely destroyed
  if (d_scene)
  {
    d_scene->unregisterViewer(this);
    d_scene = 0;
  }
}



void ViewerOpenGL::setScene(VrmlScene* scene)
{
  Viewer::setScene(scene);
  if (d_scene)
  {
    viewpoint = d_scene->bindableViewpointTop();
    navinfo = d_scene->bindableNavigationInfoTop();
    background = d_scene->bindableBackgroundTop();
    fog = d_scene->bindableFogTop();
  }
}


void ViewerOpenGL::initialize()
{
  d_glVersion = atof(reinterpret_cast<const char*>(glGetString(GL_VERSION)));

  // initialize the GL context
  glEnable(GL_DEPTH_TEST);
  glDepthFunc(GL_LEQUAL);

#if USE_STENCIL_SHAPE
  glEnable(GL_STENCIL_TEST);
  glStencilFunc(GL_ALWAYS, 1, 1);
  glStencilOp(GL_KEEP, GL_KEEP, GL_REPLACE);
#endif

  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
  glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, 1);
  float rgb[4] = { 0.0f, 0.0f, 0.0f, 0.0f };
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, rgb);

#if defined(__ORBISNAP__)

  // in many OpenGL implementations, GL_FASTEST means 'extremely ugly'
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

#endif

  // should only trigger this after a non-uniform scale is seen...
  // on my system this is <5% speedup though.
  glEnable(GL_NORMALIZE);

  // blending is only enabled if d_blend and a non-zero transparency value is
  // passed in
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

#if defined(__ORBISNAP__)

  // these should be disabled by default but this is not sure on
  // some machines (Sun)
  glDisable(GL_INDEX_LOGIC_OP);
  glDisable(GL_COLOR_LOGIC_OP);

#endif

  // read maximum texture size
  glGetIntegerv(GL_MAX_TEXTURE_SIZE, &d_maxTextureSizeHW);
  if (d_maxTextureSize > d_maxTextureSizeHW)
    d_maxTextureSize = d_maxTextureSizeHW;

  // EXPERIMENTAL: does alpha test help ameliorate the bad translucency problem?
  glEnable(GL_ALPHA_TEST);
  glAlphaFunc(GL_GREATER, 0.1);

  // check for errors
  checkOpenGLError("ViewerOpenGL::initialize()");

  d_GLinitialized = true;
}



//
//  beginObject/endObject should correspond to grouping nodes.
//  Group-level scoping for directional lights, anchors, sensors
//  are handled here. Display lists can optionally be created
//  (but the retain flag is just a hint, not guaranteed). Retained
//  objects can be referred to later to avoid duplicating geometry.
//  OpenGL doesn't allow nested objects. The top-down approach of
//  putting entire groups in display lists is faster for static
//  scenes but uses more memory and means that if anything is changed,
//  the whole object must be tossed.
//  The bottom-up model wraps each piece of geometry in a dlist but
//  requires traversal of the entire scene graph to reference each dlist.
//  The decision about what groups to stuff in an object is punted to
//  the object itself, as it can decide whether it is mutable.
//
//  The OpenGL viewer never puts objects in display lists, so the
//  retain hint is ignored.

Viewer::Object ViewerOpenGL::beginObject(const char *,
                                         bool /* retain */
                                         )
{
  d_nObjects++;         // increment total object count
  ++d_nestedObjects;    // increment the nesting count

  if (d_nObjects==1) glPushMatrix();

  // Increment nesting level for group-scoped lights
  for (int i=0; i<MAX_LIGHTS; ++i)
    if (d_lightInfo[i].lightType == LIGHT_DIRECTIONAL)
      ++d_lightInfo[i].nestingLevel;

  return 0;
}

// End of group scope

void ViewerOpenGL::endObject()
{
  // Decrement nesting level for group-scoped lights and get rid
  // of any defined at this level
  for (int i=0; i<MAX_LIGHTS; ++i)
    if (d_lightInfo[i].lightType == LIGHT_DIRECTIONAL)
      if (--d_lightInfo[i].nestingLevel < 0)
        {
          glDisable( (GLenum) (GL_LIGHT0 + i) );
          d_lightInfo[i].lightType = LIGHT_UNUSED;
        }

  if (--d_nestedObjects == 0)
    {
      //cout << "ViewerOpenGL::endObject():pop" << endl;
      glPopMatrix();
    }
}


// These attributes need to be reset per geometry. Any attribute
// modified in a geometry function should be reset here. This is
// called after Appearance/Material has been set. Any attribute
// that can be modified by an Appearance node should not be modified
// here since these settings will be put in dlists with the geometry.

void ViewerOpenGL::beginGeometry()
{
  glPushAttrib( GL_ENABLE_BIT );
}

// This should be called BEFORE ending any dlists containing geometry,
// otherwise, attributes changed in the geometry insertion will impact
// the geometries rendered after the dlist is executed.

void ViewerOpenGL::endGeometry()
{
  glPopAttrib();
  glCullFace( GL_BACK );
  glShadeModel( GL_SMOOTH );
  glFrontFace(GL_CCW);

  // if needed...
  glMatrixMode(GL_TEXTURE);
  glLoadIdentity();
  glMatrixMode(GL_MODELVIEW);
}

// Construct Billboard transformation matrix M
//
// get rid of this asap: change over to use the transformation matrix
// accumulated during render traversal. -cks
//
void ViewerOpenGL::getBillboardTransformMatrix(float M[4][4],
                                               const float axisOfRotation[3])
{
  GLfloat modelview[16],invert[9];
  glGetFloatv (GL_MODELVIEW_MATRIX, modelview);
  float pos[] = {-modelview[12], - modelview[13], -modelview[14]};
  float v[3],y[3];

  if(InvertMatrix3x3of4x4(modelview,invert))
  {
// Viewer position in local coordinate system
   v[0] = pos[0] * invert[0] + pos[1] * invert[3] + pos[2] * invert[6];
   v[1] = pos[0] * invert[1] + pos[1] * invert[4] + pos[2] * invert[7];
   v[2] = pos[0] * invert[2] + pos[1] * invert[5] + pos[2] * invert[8];
   Vnorm( v );
// Viewer-alignment
  if ( FPZERO(axisOfRotation[0]) &&
       FPZERO(axisOfRotation[1]) &&
       FPZERO(axisOfRotation[2]) )
    {
// Viewer up vector
          y[0] = invert[3];
          y[1] = invert[4];
          y[2] = invert[5];
      Vnorm( y );
// get x-vector from the cross product of Viewer's
// up vector and billboard-to-viewer vector.
     float x[3];
     Vcross( x, y, v );
         Vnorm( x );
     M[0][0] = x[0]; M[0][1] = x[1]; M[0][2] = x[2]; M[0][3] = 0.0;
     M[1][0] = y[0]; M[1][1] = y[1]; M[1][2] = y[2]; M[1][3] = 0.0;
     M[2][0] = v[0]; M[2][1] = v[1]; M[2][2] = v[2]; M[2][3] = 0.0,
     M[3][0] = M[3][1] = M[3][2] = 0.0; M[3][3] = 1.0;
    }

  // use axis of rotation
  else
    {
// axis of rotation will be the y-axis vector
     y[0] = axisOfRotation[0];
     y[1] = axisOfRotation[1];
     y[2] = axisOfRotation[2];
     Vnorm( y );

// Plane defined by the axisOfRotation and billboard-to-viewer vector
     float x[3];
     Vcross( x, y, v );
     Vnorm( x );
// Get Z axis vector from cross product of X and Y
     float z[3];
     Vcross(z, x, y);
// Transform Z axis vector of current coordinate system to new coordinate system.
     float nz[3];
     nz[0] = x[2]; nz[1] = y[2]; nz[2] = z[2];
// calculate the angle by which the Z axis vector of current coordinate system
// has to be rotated around the Y axis to new coordinate system.
     float angle = acos(nz[2])*180/M_PI;
     if(nz[0] > 0) angle = -angle;
     glPushMatrix();
     glLoadIdentity();
     glRotatef(angle,y[0],y[1],y[2]);
     glGetFloatv (GL_MODELVIEW_MATRIX,(float *) M);
     glPopMatrix();
  }
  }
}


Viewer::RenderMode ViewerOpenGL::getRenderMode()
{
  return d_selectMode ? RENDER_MODE_PICK : RENDER_MODE_DRAW;
}

double ViewerOpenGL::getFrameRate()
{
  return 1.0 / d_renderTime;
}

void ViewerOpenGL::resetUserNavigation()
{
  Viewer::resetUserNavigation();

  d_translatex = d_translatey = d_translatez = 0.0;

  d_activeSensitive = 0;
  wsSetCursor(CURSOR_ARROW);

#if !defined(__ORBISNAP__)
  d_target[0] = d_target[1] = d_target[2] = 0.0;
#endif

  trackball(d_curquat, 0.0, 0.0, 0.0, 0.0);
  //d_rotationChanged = true;
  double tmp[4][4];
  Midentity(tmp);
  matrix_to_glmatrix(tmp, (float*)d_rotationMatrix);

  requestRedraw();
}


void ViewerOpenGL::getUserNavigation(double M[4][4])
{
  //cout << "ViewerOpenGL::getuserNavigation()" << endl;

  Midentity(M);
  float pos_vec[3];
  pos_vec[0] = d_translatex;
  pos_vec[1] = d_translatey;
  pos_vec[2] = d_translatez;
  double pos_mat[4][4];
  Mtranslation(pos_mat, pos_vec);
  //cout << "(" << d_translatex << "," << d_translatey << "," << d_translatez << ")" << endl;

  double rot_mat[4][4];
  for(int i=0; i<4; i++) // oh good grief.
    for(int j=0; j<4; j++)
      rot_mat[i][j] = d_rotationMatrix[j][i];

  //MM((double[4][4])M, pos_mat, rot_mat);
  MM(M, rot_mat, pos_mat); // M = R * T
  //Mdump(cout, M) << endl;
}

// Generate a normal from 3 indexed points.

static void indexFaceNormal(int i1,
                            int i2,
                            int i3,
                            const float * p,
                            float *N)
{
  float V1[3], V2[3];

  Vdiff( V1, &p[i2], &p[i3] );
  Vdiff( V2, &p[i2], &p[i1] );
  Vcross( N, V1, V2 );
  Vnorm(N);                  // normalize the normal
}

//
//  Geometry insertion.
//

Viewer::Object ViewerOpenGL::insertBackground(size_t nGroundAngles,
                                              const float * groundAngle,
                                              const float * groundColor,
                                              size_t nSkyAngles,
                                              const float * skyAngle,
                                              const float * skyColor,
                                              int* whc,
                                              unsigned char** pixels,
                                              Object texturesDisplayList)
{
  GLuint glid = 0;

  float r = 0.0, g = 0.0, b = 0.0, a = 1.0;

  // Clear to last sky color
  if (skyColor != 0)
    {
      r = skyColor[3*nSkyAngles+0];
      g = skyColor[3*nSkyAngles+1];
      b = skyColor[3*nSkyAngles+2];
    }

  // clear screen (reserved area is protected by scissor setting
  // in ViewerOpenGL::redraw()).
  glClearColor(r, g, b, a);
  glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

  // Draw the background as big spheres centered at the view position
  if ( (nSkyAngles > 0 || nGroundAngles > 0 || pixels) )
    {

      glDisable(GL_DEPTH_TEST);
      glDisable(GL_LIGHTING);

      glPushMatrix();

      // Set a projection for rendering the background.
      // It is the same as the normal projection except that
      // the near and far plane distances are fixed constants.
      // This way we are sure that the background does not get
      // clipped between znear/zfar plane.
      glMatrixMode(GL_PROJECTION);
      glPushMatrix();
      glLoadIdentity();
      float fieldOfView = M_PI_4;
      if (viewpoint) fieldOfView = viewpoint->getFieldOfView().get();
      float restrictedFoV = 180.0 / M_PI * min (2.0 * atan(tan(fieldOfView/2.0)/getCamera().getZoom()), maxFoV);
      gluPerspective(
        restrictedFoV,
        ((float) width) / (height - d_bottomReserved),
        1.0, 1000.0);
      glMatrixMode(GL_MODELVIEW);

      /*
       * This is a bugfix for the badly placed Background.
       */

      /* get the current modeling matrix */
      float mat[16];
      glGetFloatv(GL_MODELVIEW_MATRIX, mat);

      /* background ignores translations, so clear the translation part */
      mat[12] = 0;
      mat[13] = 0;
      mat[14] = 0;
      mat[15] = 1;

      /* apply the modified matrix */
      glLoadMatrixf(mat);

      // set the background sphere as far as possible
      glScalef(10.0, 10.0, 10.0);

      // Sphere constants
      const int nCirc = 8;              // number of circumferential slices
      const double cd = 2.0 * M_PI / nCirc;

      double heightAngle0, heightAngle1 = 0.0;
      const float *c0, *c1 = skyColor;

      for (size_t nSky=0; nSky<nSkyAngles; ++nSky)
        {
          heightAngle0 = heightAngle1;
          heightAngle1 = skyAngle[nSky];
          c0 = c1;
          c1 += 3;

          double circAngle0, circAngle1 = 0.0;
          double sha0 = sin(heightAngle0), cha0 = cos(heightAngle0);
          double sha1 = sin(heightAngle1), cha1 = cos(heightAngle1);
          double sca0, cca0;
          double sca1 = sin(circAngle1), cca1 = cos(circAngle1);

          glBegin( GL_QUADS );
          for (int nc=0; nc<nCirc; ++nc)
            {
              circAngle0 = circAngle1;
              circAngle1 = (nc+1) * cd;
              sca0 = sca1;
              sca1 = sin(circAngle1);
              cca0 = cca1;
              cca1 = cos(circAngle1);

              glColor3fv( c1 );
              glVertex3f( sha1 * cca0, cha1, sha1 * sca0 );
              glVertex3f( sha1 * cca1, cha1, sha1 * sca1 );
              glColor3fv( c0 );
              glVertex3f( sha0 * cca1, cha0, sha0 * sca1 );
              glVertex3f( sha0 * cca0, cha0, sha0 * sca0 );
            }
          glEnd();
        }

      // Ground
      heightAngle1 = M_PI;
      c1 = groundColor;

      for (size_t nGround=0; nGround<nGroundAngles; ++nGround)
        {
          heightAngle0 = heightAngle1;
          heightAngle1 = M_PI - groundAngle[nGround];
          c0 = c1;
          c1 += 3;

          double circAngle0, circAngle1 = 0.0;
          double sha0 = sin(heightAngle0), cha0 = cos(heightAngle0);
          double sha1 = sin(heightAngle1), cha1 = cos(heightAngle1);
          double sca0, cca0;
          double sca1 = sin(circAngle1), cca1 = cos(circAngle1);

          glBegin( GL_QUADS );
          for (int nc=0; nc<nCirc; ++nc)
            {
              circAngle0 = circAngle1;
              circAngle1 = (nc+1) * cd;
              sca0 = sca1;
              sca1 = sin(circAngle1);
              cca0 = cca1;
              cca1 = cos(circAngle1);

              glColor3fv( c1 );
              glVertex3f( sha1 * cca1, cha1, sha1 * sca1 );
              glVertex3f( sha1 * cca0, cha1, sha1 * sca0 );
              glColor3fv( c0 );
              glVertex3f( sha0 * cca0, cha0, sha0 * sca0 );
              glVertex3f( sha0 * cca1, cha0, sha0 * sca1 );
            }
          glEnd();
        }

      if (d_texture && ! d_wireframe)
        {
	  if (texturesDisplayList!=0)
	    insertReference(texturesDisplayList);
	  else
	    {

#if USE_GEOMETRY_DISPLAY_LISTS
	      // Don't bother with a dlist if we aren't drawing anything
	      if (
		  (nSkyAngles > 0 || nGroundAngles > 0 || pixels) )
		{
		  glid = glGenLists(1);
		  glNewList( glid, GL_COMPILE_AND_EXECUTE );
		}
#endif

	      // Background textures are drawn on a non-transparent cube
	      if (pixels && d_texture && ! d_wireframe)
		{
		  float v[6][4][3] = {
		    {{1,-1,1}, {-1,-1,1}, {-1,1,1}, {1,1,1}},     // Back
		    {{-1,-1,1}, {1,-1,1}, {1,-1,-1}, {-1,-1,-1}}, // Bottom
		    {{-1,-1,-1}, {1,-1,-1}, {1,1,-1}, {-1,1,-1}}, // Front
		    {{-1,-1,1}, {-1,-1,-1}, {-1,1,-1}, {-1,1,1}}, // Left
		    {{1,-1,-1}, {1,-1,1}, {1,1,1}, {1,1,-1}},     // Right
		    {{-1,1,-1}, {1,1,-1}, {1,1,1}, {-1,1,1}}};    // Top

		  glColor4d(0.0, 0.0, 0.0, 1.0);
		  glScalef( 0.5, 0.5, 0.5 );
		  glEnable( GL_TEXTURE_2D );
		  glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE );

		  int t, lastT = -1;
		  for (t=0; t<6; ++t, whc+=3) {
		    // Check for non-zero width,height,coords and pixel data
		    if (whc[0] && whc[1] && whc[2] && pixels[t])
		      {
			// Optimize for the case where the same texture is used
			if (lastT == -1 || pixels[t] != pixels[lastT])
			  insertTexture(whc[0], whc[1], whc[2],
					false, false, pixels[t],
					false);  // Don't put the textures in dlists

			lastT = t;
			glBegin( GL_QUADS );

			//glTexCoord2f( 0.0, 1.0 );
			glTexCoord2f( 0.0, 0.0 );
			glVertex3fv( v[t][0] );

			//glTexCoord2f( 1.0, 1.0 );
			glTexCoord2f( 1.0, 0.0 );
			glVertex3fv( v[t][1] );

			//glTexCoord2f( 1.0, 0.0 );
			glTexCoord2f( 1.0, 1.0 );
			glVertex3fv( v[t][2] );

			//glTexCoord2f( 0.0, 0.0 );
			glTexCoord2f( 0.0, 1.0 );
			glVertex3fv( v[t][3] );

			glEnd();
		      }
		    }
		  glDisable( GL_TEXTURE_2D );
		}
	    }
        }
      //endGeometry();
      if (glid) glEndList();

      // Put everything back the way it was
      glPopMatrix();
      glMatrixMode(GL_PROJECTION);
      glPopMatrix();
      glMatrixMode(GL_MODELVIEW);

      if (d_lit) glEnable( GL_LIGHTING );
      glEnable( GL_DEPTH_TEST );
    }

  // Save bg color so we can choose a fg color (doesn't help bg textures...)
  d_background[0] = r;
  d_background[1] = g;
  d_background[2] = b;

  return (glid!=0) ? ((Object) glid) : texturesDisplayList;
}



Viewer::Object ViewerOpenGL::insertBox(float x, float y, float z)
{
  GLuint glid = 0;

#if USE_GEOMETRY_DISPLAY_LISTS
  glid = glGenLists(1);
  glNewList( glid, GL_COMPILE_AND_EXECUTE );
#endif

  static GLint faces[6][4] =
  {
    {0, 1, 2, 3},
    {1, 5, 6, 2},
    {5, 4, 7, 6},
    {4, 0, 3, 7},
    {2, 6, 7, 3},
    {0, 4, 5, 1}
  };

  static GLfloat n[6][3] =      // normals
  {
    {-1.0, 0.0, 0.0},
    {0.0, 0.0, 1.0},
    {1.0, 0.0, 0.0},
    {0.0, 0.0, -1.0},
    {0.0, 1.0, 0.0},
    {0.0, -1.0, 0.0}
  };

  GLfloat v[8][3];
  GLint i;

  v[0][0] = v[1][0] = v[2][0] = v[3][0] = -x / 2;
  v[4][0] = v[5][0] = v[6][0] = v[7][0] = x / 2;
  v[0][1] = v[1][1] = v[4][1] = v[5][1] = -y / 2;
  v[2][1] = v[3][1] = v[6][1] = v[7][1] = y / 2;
  v[0][2] = v[3][2] = v[4][2] = v[7][2] = -z / 2;
  v[1][2] = v[2][2] = v[5][2] = v[6][2] = z / 2;

  beginGeometry();
  glShadeModel( GL_FLAT );

  glBegin( GL_QUADS );
  for (i = 0; i < 6; ++i)
    {
      glNormal3fv(&n[i][0]);

      //glTexCoord2f( 0.0, 1.0 );
      glTexCoord2f( 0.0, 0.0 );
      glVertex3fv(&v[faces[i][0]][0]);

      //glTexCoord2f( 1.0, 1.0 );
      glTexCoord2f( 1.0, 0.0 );
      glVertex3fv(&v[faces[i][1]][0]);

      //glTexCoord2f( 1.0, 0.0 );
      glTexCoord2f( 1.0, 1.0 );
      glVertex3fv(&v[faces[i][2]][0]);

      //glTexCoord2f( 0.0, 0.0 );
      glTexCoord2f( 0.0, 1.0 );
      glVertex3fv(&v[faces[i][3]][0]);
    }
  glEnd();

  endGeometry();
  if (glid) glEndList();

  return (Object) glid;
}


Viewer::Object ViewerOpenGL::insertCone(float h,
                                        float r,
                                        bool bottom, bool side)
{
  GLuint glid = 0;

#if USE_GEOMETRY_DISPLAY_LISTS
  glid = glGenLists(1);
  glNewList( glid, GL_COMPILE_AND_EXECUTE );
#endif

  beginGeometry();

  if (! bottom || ! side )
    glDisable( GL_CULL_FACE );

  if (bottom || side)
    {
      const int nfacets = 24;           // Number of polygons for sides
      const int npts = 2 * nfacets;
      const int nfaces = nfacets * 5;

      float c[ npts ][ 3 ];             // coordinates
      float tc[ npts ][ 3 ];            // texture coordinates
      int faces[ nfaces ];              // face lists

      // should only compute tc if a texture is present...
      computeCylinder( h, r, nfacets, c, tc, faces);

      for (int i=0; i<nfacets; ++i)
        c[i][0] = c[i][2] = 0.0;

      if (side)
        {
          float Ny = r * r / h;
//          glBegin( GL_QUAD_STRIP );
          glBegin(GL_TRIANGLE_STRIP);
          for (int i = 0; i < nfacets; ++i)
            {
              float N[3] = { c[i+nfacets][0], Ny, c[i+nfacets][2] };
              Vnorm(N);              // normalize the normal
              glNormal3fv(N);
              glTexCoord2fv( &tc[i+nfacets][0] );
              glVertex3fv(   &c [i+nfacets][0] );
              glTexCoord2fv( &tc[i][0] );
              glVertex3fv(   &c [i][0] );
            }

          float N[3] = { c[nfacets][0], Ny, c[nfacets][2] };
          Vnorm(N);                  // normalize the normal
          glNormal3fv(N);
          glTexCoord2f( tc[nfacets][0]-1.0, tc[nfacets][1] );
          glVertex3fv(   &c [nfacets][0] );
          glTexCoord2f( tc[0][0]-1.0, tc[0][1] );
          glVertex3fv(   & c[0][0] );
          glEnd();
        }

      if (bottom)
        {
          glBegin( GL_TRIANGLE_FAN );
          glNormal3f( 0.0, -1.0, 0.0 );
          glTexCoord2f( 0.5, 0.5 );
          glVertex3f( 0.0, - 0.5 * h, 0.0 );

          float angle = 0.5 * M_PI; // First v is at max x
          float aincr = 2.0 * M_PI / (float) nfacets;
          for (int i = 0; i < nfacets; ++i, angle+=aincr )
            {
              glTexCoord2f( 0.5*(1.0+sin( angle )),
                            1.0-0.5*(1.+cos( angle )) );
              glVertex3fv( &c[i+nfacets][0] );
            }
          glTexCoord2f( 0.5*(1.0+sin( angle )),
                        1.0-0.5*(1.+cos( angle )) );
          glVertex3fv( &c[nfacets][0] );
          glEnd();
        }
    }

  endGeometry();
  if (glid) glEndList();

  return (Object) glid;
}


Viewer::Object ViewerOpenGL::insertCylinder(float h,
                                            float r,
                                            bool bottom, bool side, bool top)
{
  GLuint glid = 0;

#if USE_GEOMETRY_DISPLAY_LISTS
  glid = glGenLists(1);
  glNewList( glid, GL_COMPILE_AND_EXECUTE );
#endif

  beginGeometry();
  if (! bottom || ! side || ! top)
    glDisable( GL_CULL_FACE );

  if (bottom || side || top)
    {

#if defined(__ORBISNAP__)

      // 8 looks too rough for larger cylinders
      const int nfacets = 24;           // Number of polygons for sides

#else

      const int nfacets = 8;            // Number of polygons for sides

#endif

      const int npts = 2 * nfacets;
      const int nfaces = nfacets * 5;

      float c[ npts ][ 3 ];             // coordinates
      float tc[ npts ][ 3 ];            // texture coordinates
      int faces[ nfaces ];              // face lists

      // should only compute tc if a texture is present...
      computeCylinder( h, r, nfacets, c, tc, faces);

      if (side)
        {
          glBegin( GL_QUAD_STRIP );
          for (int i = 0; i < nfacets; ++i)
            {
              float N[3] = { c[i+nfacets][0], 0.0, c[i+nfacets][2] };
              Vnorm(N);                 // normalize the normal
              glNormal3fv(N);
              glTexCoord2fv( &tc[i+nfacets][0] );
              glVertex3fv(   &c [i+nfacets][0] );
              glTexCoord2fv( &tc[i][0] );
              glVertex3fv(   &c [i][0] );
            }

          float N[3] = { c[nfacets][0], 0.0, c[nfacets][2] };
          Vnorm(N);                     // normalize the normal
          glNormal3fv(N);
          glTexCoord2f( tc[nfacets][0]-1.0, tc[nfacets][1] );
          glVertex3fv(   &c [nfacets][0] );
          glTexCoord2f( tc[0][0]-1.0, tc[0][1] );
          glVertex3fv(   & c[0][0] );
          glEnd();
        }

      if (bottom)               // tex coords...
        {
          glBegin( GL_TRIANGLE_FAN );
          glNormal3f( 0.0, -1.0, 0.0);
          glTexCoord2f( 0.5, 0.5 );
          glVertex3f( 0.0, - 0.5 * h, 0.0 );

          float angle = 0.5 * M_PI; // First v is at max x
          float aincr = 2.0 * M_PI / (float) nfacets;
          for (int i = 0; i < nfacets; ++i, angle+=aincr)
            {
              glTexCoord2f( 0.5*(1.+sin( angle )),
                            1.0 - 0.5*(1.+cos( angle )) );
              glVertex3fv( &c[i+nfacets][0] );
            }
          glTexCoord2f( 0.5*(1.+sin( angle )),
                        1.0 - 0.5*(1.+cos( angle )) );
          glVertex3fv( &c[nfacets][0] );
          glEnd();
        }

      if (top)          // tex coords...
        {
          glBegin( GL_TRIANGLE_FAN );
          glNormal3f( 0.0, 1.0, 0.0);
          glTexCoord2f( 0.5, 0.5 );
          glVertex3f( 0.0, 0.5 * h, 0.0 );

          float angle = 0.75 * M_PI;
          float aincr = 2.0 * M_PI / (float) nfacets;
          for (int i = nfacets-1; i >= 0; --i, angle+=aincr)
            {
              glTexCoord2f( 0.5*(1.+sin( angle )),
                            1.0 - 0.5*(1.+cos( angle )) );
              glVertex3fv( &c[i][0] );
            }
          glTexCoord2f( 0.5*(1.+sin( angle )),
                        1.0 - 0.5*(1.+cos( angle )) );
          glVertex3fv( &c[nfacets-1][0] );
          glEnd();
        }
    }

  endGeometry();
  if (glid) glEndList();

  return (Object) glid;
}

// Compute a normal at vert i,j of an ElevationGrid.

static void elevationVertexNormal(int i, int j,
                                  int nx, int nz,
                                  float dx, float dz,
                                  const float * height,
                                  float N[])
{
  float Vx[3], Vz[3];

  if (i > 0 && i < nx-1)
    {
      Vx[0] = 2.0 * dx;
      Vx[1] = *(height+1) - *(height-1);
    }
  else if (i == 0)
    {
      Vx[0] = dx;
      Vx[1] = *(height+1) - *(height);
    }
  else
    {
      Vx[0] = dx;
      Vx[1] = *(height) - *(height-1);
    }
  Vx[2] = 0.0;

  Vz[0] = 0.0;
  if (j > 0 && j < nz-1)
    {
      Vz[1] = *(height+nx) - *(height-nx);
      Vz[2] = 2.0 * dz;
    }
  else if (j == 0)
    {
      Vz[1] = *(height+nx) - *(height);
      Vz[2] = dz;
    }
  else
    {
      Vz[1] = *(height) - *(height-nx);
      Vz[2] = dz;
    }

  Vcross( N, Vz, Vx );
  Vnorm(N);             // normalize the normal
}


Viewer::Object ViewerOpenGL::insertElevationGrid(unsigned int mask,
                                                 size_t nx,
                                                 size_t nz,
                                                 const float * height,
                                                 float dx,
                                                 float dz,
                                                 const float * texture_coords,
                                                 const float * normals,
                                                 const float * colors)
{
  size_t i, j;
  float x, z;

  GLuint glid = 0;

#if USE_GEOMETRY_DISPLAY_LISTS
  glid = glGenLists(1);
  glNewList( glid, GL_COMPILE_AND_EXECUTE );
#endif

  beginGeometry();

  // Face orientation & culling
  glFrontFace( (mask & MASK_CCW) ? GL_CCW : GL_CW );
  if (! (mask & MASK_SOLID) )
    glDisable( GL_CULL_FACE );

  // x varies fastest
  for (j=0; j<nz-1; ++j)
    {
      float s0, s1, t0 = 0.0, t1 = 0.0;

      z = dz * j;
      if (! texture_coords)
        {
          t0 = ((float) j) / (nz-1);
          t1 = ((float) j+1) / (nz-1);
        }

#if defined(__ORBISNAP__)

      // although we render groups of 4 vertices, triangle strips must be
      // used. Quad strips are dangerous because the four points of a quad
      // may not be coplanar, causing some quads to disappear
      // or change their shape unexpectedly, making the terrain look
      // "leaky" or "unstable".
      glBegin(GL_TRIANGLE_STRIP);

#else

      glBegin( GL_QUAD_STRIP );

#endif

      for (i=0; i<nx; ++i)
        {
          x = dx * i;

          if (colors &&
              ((mask & MASK_COLOR_PER_VERTEX) || (i < nx-1)))
            {
              glColor3fv( colors );
            }

          if (normals &&
              ((mask & MASK_NORMAL_PER_VERTEX) || (i < nx-1)))
            {
              glNormal3fv( normals );
            }
          else if (! normals)
            {
              float N[3];
               if (mask & MASK_NORMAL_PER_VERTEX)
                 {
                   elevationVertexNormal(i, j, nx, nz, dx, dz, height, N);
                   glNormal3fv( N );
                 }
               else if (i < nx-1)                // Normal per face
                 {
                  float Vx[3] = { dx, *(height+1) - *height, 0.0 };
                  float Vz[3] = { 0.0, *(height+nx) - *height, dz };
                  Vcross( N, Vx, Vz );
                  glNormal3fv( N );
                 }
            }

          if (texture_coords)
            {
              s0 = *(texture_coords);
              t0 = *(texture_coords+1);
              s1 = *(texture_coords+(nx*2));
              t1 = *(texture_coords+(nx*2)+1);
              texture_coords+=2;
            }
          else
            s0 = s1 = ((float) i) / (nx-1);

          glTexCoord2f( s0, t0 );
          //cout << "ViewerOpenGL::insertElevationGrid():(s0,t0)=" << s0 << "," << t0 << endl;
          glVertex3f( x, *height, z );
          //cout << "ViewerOpenGL::insertElevationGrid():(x,y,z)=" << x << "," << *height << "," << z << endl;

          // Vertex from next row
          if (colors && (mask & MASK_COLOR_PER_VERTEX))
            {
              glColor3fv( colors+3*nx );
            }

          if (mask & MASK_NORMAL_PER_VERTEX)
            {
              if (normals)
                {
                  glNormal3fv( normals+3*nx );
                }
              else
                {
                  float N[3];
                  elevationVertexNormal(i, j+1, nx, nz, dx, dz, height+nx, N);
                  glNormal3fv( N );
                }
            }

          glTexCoord2f( s1, t1 );
          //cout << "ViewerOpenGL::insertElevationGrid():(s1,t1)=" << s1 << "," << t1 << endl;
          glVertex3f( x, *(height+nx), z+dz );
          //cout << "ViewerOpenGL::insertElevationGrid():(x,y,z)=" << x << "," << *(height+nx) << "," << z+dz << endl;

          ++height;
          if ( colors ) colors += 3;
          if ( normals ) normals += 3;
        }

      glEnd();
    }

  endGeometry();
  if (glid) glEndList();
  return (Object) glid;
}


#if GLU_VERSION_1_2

// Tesselator callback

#if _WIN32
# define WINAPI __stdcall
#else
# define WINAPI
#endif
typedef GLvoid ( WINAPI *TessCB)();


// Extrusion cap tessellation for non-convex shapes

typedef struct {
  const float * c;                      // coordinates array [nVerts * 3]
  const float * crossSection;           // crossSection coordinates [nCrossSection * 2]
  float tcDeltaU, tcDeltaV;
  float tcScaleU, tcScaleV;
  int vOffset;
  float N[3];                   // Normal
} TessExtrusion;

static void WINAPI tessExtrusionBegin( GLenum type, void *pdata )
{
  TessExtrusion *p = (TessExtrusion *)pdata;
  glBegin( type );
  Vnorm(&p->N[0]);              // normalize the normal
  glNormal3fv( &p->N[0] );
}


static void WINAPI tessExtrusionVertex( void *vdata, void *pdata )
{
  ptrdiff_t j = (ptrdiff_t) vdata;
  TessExtrusion *p = (TessExtrusion *)pdata;

  glTexCoord2f( (p->crossSection[2*j] - p->tcDeltaU) * p->tcScaleU,
                (p->crossSection[2*j+1] - p->tcDeltaV) * p->tcScaleV );
  glVertex3fv( &(p->c[3 * (j + p->vOffset)]) );
}

#endif


void ViewerOpenGL::insertExtrusionCaps( unsigned int mask,
                                        size_t nSpine,
					const float * spine,
                                        const float * c,
                                        size_t nCrossSection,
                                        const float * cs )
{
  // Determine x,z ranges for top & bottom tex coords
  float xz[4] = { cs[0], cs[0], cs[1], cs[1] };
  const float * csp = cs;

  for (size_t nn=1; nn<nCrossSection; ++nn, csp += 2)
    {
      if (csp[0] < xz[0])      xz[0] = csp[0];
      else if (csp[0] > xz[1]) xz[1] = csp[0];
      if (csp[1] < xz[2])      xz[2] = csp[1];
      else if (csp[1] > xz[3]) xz[3] = csp[1];
    }

  float dx = xz[1] - xz[0];
  float dz = xz[3] - xz[2];
  if (! FPZERO(dx)) dx = 1.0 / dx;
  if (! FPZERO(dz)) dz = 1.0 / dz;

  if (! d_tess) d_tess = gluNewTess();

  gluTessCallback( d_tess, (GLenum) GLU_TESS_BEGIN_DATA,
		   (TessCB) tessExtrusionBegin );
  gluTessCallback( d_tess, (GLenum) GLU_TESS_VERTEX_DATA,
		   (TessCB) tessExtrusionVertex );
  gluTessCallback( d_tess, (GLenum) GLU_TESS_END,
		   (TessCB) glEnd );

  int last = 2*(nCrossSection-1);
  bool equalEndpts = FPEQUAL(cs[0], cs[last]) && FPEQUAL(cs[1], cs[last+1]);
  if (mask & MASK_BOTTOM)
  {
    TessExtrusion bottom = { c, cs, xz[0], xz[2], dx, dz, 0 };
    computeFaceNormal(false, c, nCrossSection, bottom.N);

    // is spine orientation changed relative to default orientation
    bool spineOrientationChanged = false;
    if ((mask & MASK_CCW) ^ VrmlNodeExtrusion::defaultSpineOrientationChanged(nSpine, spine))
    {
      spineOrientationChanged = true;
      bottom.N[0] = -bottom.N[0]; bottom.N[1] = -bottom.N[1]; bottom.N[2] = -bottom.N[2];
    }

    glFrontFace((spineOrientationChanged) ? GL_CCW : GL_CW);
    gluTessBeginPolygon( d_tess, &bottom );
    gluTessBeginContour( d_tess );
    GLdouble v[3];

    ptrdiff_t j = equalEndpts ? nCrossSection-2 : nCrossSection-1;
    for ( ; j>=0; --j)
      {
	v[0] = c[3*j];
	v[1] = c[3*j+1];
	v[2] = c[3*j+2];
	gluTessVertex( d_tess, v, (void*)j );
      }
    gluTessEndContour( d_tess );
    gluTessEndPolygon( d_tess );
  }

  if (mask & MASK_TOP)
  {
    int n = (nSpine - 1) * nCrossSection;
    TessExtrusion top = { c, cs, xz[0], xz[2], dx, dz, n };
    computeFaceNormal(false, &(c[3*(nSpine-1)*(nCrossSection)]), nCrossSection, top.N);

    // is spine orientation changed relative to default orientation
    bool spineOrientationChanged = false;

    if ((!(mask & MASK_CCW)) ^ VrmlNodeExtrusion::defaultSpineOrientationChanged(nSpine, spine))
    {
      spineOrientationChanged = true;
      top.N[0] = -top.N[0]; top.N[1] = -top.N[1]; top.N[2] = -top.N[2];
    }


    glFrontFace((spineOrientationChanged) ? GL_CCW : GL_CW);
    gluTessBeginPolygon( d_tess, &top );
    gluTessBeginContour( d_tess );

    GLdouble v[3];

    ptrdiff_t j = equalEndpts ? nCrossSection-2 : nCrossSection-1;
    for ( ; j>=0; --j)
      {
	v[0] = c[3*(j+n)];
	v[1] = c[3*(j+n)+1];
	v[2] = c[3*(j+n)+2];
	gluTessVertex( d_tess, v, (void*)j );
      }
    gluTessEndContour( d_tess );
    gluTessEndPolygon( d_tess );
  }
}



/**
 * If the angle between <em>v1</em> and <em>v2</em> is smaller than
 * <em>creaseAngle<em>, adds <em>v2</em> to <em>vsum</em>, otherwise
 * keeps <em>sum</em> unchanged.
 */
static void Vcondadd(float* vsum,
                     const float* v1,
                     const float* v2,
                     float creaseAngle)
{
  float v1n[3];
  float v2n[3];
  Vnorm(v1n, v1);
  Vnorm(v2n, v2);
  if (fabs(acos(Vdot(v1n, v2n))) <= creaseAngle)
  {
    vsum[0] += v2[0];
    vsum[1] += v2[1];
    vsum[2] += v2[2];
  }
}

// correct normal by finding its neighbor normal and applying crossAngle smoothing
static void correctNormal(float* normals, float* correctedNormals, int index, int bigStep, int smallStep, int normalsSize, float creaseAngle)
{
  int neighbor;
  if ((neighbor=index+bigStep+smallStep)>=0 && neighbor<normalsSize-1)
    Vcondadd(&correctedNormals[index],
		         &normals[index],
		         &normals[neighbor],
		         creaseAngle);
}

// functions which evaluate whether normal is at singular vertex or not
static bool firstIsBiggerThanSecond(int first, int second) { return first>second; }
static bool firstIsLesserThanSecond(int first, int second) { return first<second; }



Viewer::Object ViewerOpenGL::insertExtrusion(unsigned int mask,
                                             size_t nOrientation,
                                             const float * orientation,
                                             size_t nScale,
                                             const float * scale,
                                             size_t nCrossSection,
                                             const float * crossSection,
                                             size_t nSpine,
                                             const float * spine,
                                             float creaseAngle)
{
  float *c  = new float[nCrossSection * nSpine * 3];
  float *tc = new float[nCrossSection * nSpine * 3];

  // safety checks
  const float safetyScale[2] = { 1.0f, 1.0f };
  if (nScale < nSpine && nScale != 1)
  {
    // scale is invalid, use default
    nScale = 1;
    scale = safetyScale;
  }
  const float safetyOri[4] = { 0.0f, 1.0f, 0.0f, 0.0f };
  if (nOrientation < nSpine && nOrientation != 1)
  {
    // orientation is invalid, use default
    nOrientation = 1;
    orientation = safetyOri;
  }

  computeExtrusion( nOrientation, orientation,
                    nScale, scale,
                    nCrossSection, crossSection,
                    nSpine, spine,
                    c, tc, 0 );

  GLuint glid = 0;

#if USE_GEOMETRY_DISPLAY_LISTS
  glid = glGenLists(1);
  glNewList( glid, GL_COMPILE_AND_EXECUTE );
#endif

  beginGeometry();

  // Face orientation & culling. Note that the extrusion polygons
  // are faced OUTSIDE the extrusion (they must be visible from outside,
  // not from inside).
  glFrontFace((mask & MASK_CCW) ? GL_CCW : GL_CW);

  if (! (mask & MASK_SOLID) )
    glDisable( GL_CULL_FACE );


  // compute face normals using cross product of face vectors
  // (shell faces are convex so we can compute normal in this way)
  // each face has 4 normals
  int normalsSize = (nSpine-1)*(nCrossSection-1)*3*4;
  float* normals = new float[normalsSize];

  // compute both face normals and list of faces for each vertex
  for (size_t i =0, n=0, faceNum = 0; i<nSpine-1; i++, n+=nCrossSection)
  {
    for (size_t j=0; j<nCrossSection-1; ++j, ++faceNum)
    {
      float v1[3], v2[3], v3[3], v4[3], norm1[3], norm2[3], norm3[3], norm4[3];

      // compute face vertices indexes
      int ind0 = n+j;
      int ind1 = n+j+1;
      int ind2 = n+j+nCrossSection+1;
      int ind3 = n+j+nCrossSection;

      // compute difference vectors
      Vdiff( v1, &c[3*ind0], &c[3*ind1] ); Vnorm(v1);
      Vdiff( v2, &c[3*ind2], &c[3*ind1] ); Vnorm(v2);
      Vdiff( v3, &c[3*ind3], &c[3*ind2] ); Vnorm(v3);
      Vdiff( v4, &c[3*ind3], &c[3*ind0] ); Vnorm(v4);

      // first normal
      Vcross(norm4, v4, v1);
      Vnorm(norm4);
      memcpy(&normals[3*4*(faceNum)], norm4, 3*sizeof(float));

      // second normal
      Vcross(norm1, v2, v1);
      Vnorm(norm1);
      memcpy(&normals[3*4*(faceNum)+3], norm1, 3*sizeof(float));

      // third normal
      Vcross(norm2, v2, v3);
      Vnorm(norm2);
      memcpy(&normals[3*4*(faceNum)+6], norm2, 3*sizeof(float));

      // fourth normal
      Vcross(norm3, v4, v3);
      Vnorm(norm3);
      memcpy(&normals[3*4*(faceNum)+9], norm3, 3*sizeof(float));
    }
  }

  // true if cross section is enclosed polygon
  bool crossSectionClosed = (crossSection[0] == crossSection[2*nCrossSection-2] &&
			     crossSection[1] == crossSection[2*nCrossSection-1]);

  // compute corrected normals by smoothing every normal with its neighbours
  float* correctedNormals = new float[normalsSize];
  memcpy(correctedNormals, normals, normalsSize*sizeof(float));

  for (size_t i=0, n=0, faceNum = 0; i<nSpine-1; ++i, n+=nCrossSection)
  {
    for (size_t j=0; j<nCrossSection-1; ++j, ++faceNum)
    {
      // correct all 4 normals for each face
      int startIndex = 3*4*faceNum;

      // normal indices
      int normalIndices[4] = {startIndex, startIndex+3, startIndex+6, startIndex+9};

      // "floor" step
      int step = (nCrossSection-1)*3*4;

      // function pointers to functions
      bool (*fn[4]) (int a, int b) = {&firstIsBiggerThanSecond, &firstIsLesserThanSecond, &firstIsLesserThanSecond, &firstIsBiggerThanSecond};

      // second parameters to functions which evaluates singularity conditions
      int conditionParams[4] = {0, nCrossSection-2, nCrossSection-2, 0};

      // shifts for finding appropriate normals
      int bigSteps[4][5] = {{0, -step, step, 0, -step},
			    {-step, 0, -2*step, -step, -step},
			    {step, 0, 0, -step, step},
			    {step, 0, 2*step, step, step}};
      int smallSteps[4][3] = {{-9, -6 ,9},
			      {18, 9, 3},
			      {6, 15, -3},
			      {-18, -15, -9}};
      // correct normals
      for (int k = 0; k<4; ++k)
      {
	if ((fn[k] (j, conditionParams[k])))
	{
	  correctNormal(normals, correctedNormals, normalIndices[k], bigSteps[k][0], smallSteps[k][0], normalsSize, creaseAngle);
	  correctNormal(normals, correctedNormals, normalIndices[k], bigSteps[k][1], smallSteps[k][1], normalsSize, creaseAngle);
	}
	else if(crossSectionClosed)
	{
	  correctNormal(normals, correctedNormals, normalIndices[k], bigSteps[k][2], smallSteps[k][0], normalsSize, creaseAngle);
	  correctNormal(normals, correctedNormals, normalIndices[k], bigSteps[k][3], smallSteps[k][1], normalsSize, creaseAngle);
	}
	correctNormal(normals, correctedNormals, normalIndices[k], bigSteps[k][4], smallSteps[k][2], normalsSize, creaseAngle);
      }
    }
  }
  delete[] normals;
  normals = correctedNormals;

  // draw quaternions as triangles (they don't have to be co-planar)
  for (size_t i=0, n=0, faceNum=0; i<nSpine-1; ++i, n+=nCrossSection)
  {
    for (size_t j = 0; j < nCrossSection-1; ++j, ++faceNum)
    {
      // two triangles make up one face
      glBegin(GL_TRIANGLE_STRIP);

      float norm[3];

      int normalIndices[4] = {n+j, n+j+1, n+j+nCrossSection, n+j+nCrossSection+1};
      int normalShifts[4] = {0, 3, 9, 6};

      // use appropriate normals
      int normalsIndex = 3*4*(faceNum);

      for (int k=0; k<4; ++k)
      {
	// get normal
	memcpy(norm, &normals[normalsIndex+normalShifts[k]], 3*sizeof(float));
	Vnorm(norm); // normalize it

	// Invert the normal if the polygon vertices are in clockwise order

	if ((!(mask & MASK_CCW)))
	  { norm[0] = -norm[0]; norm[1] = -norm[1]; norm[2] = -norm[2]; }

	// do rendering
	glNormal3fv(norm);
	glTexCoord2fv( &tc[3*normalIndices[k]] );
	glVertex3fv(   &c [3*normalIndices[k]] );
      }

      glEnd();
    }
  }

  delete[] normals;

  // Draw caps. Convex can only impact the caps of an extrusion.
  if (mask & (MASK_BOTTOM | MASK_TOP))
    insertExtrusionCaps( mask, nSpine, spine, c, nCrossSection, crossSection );

  delete [] c;
  delete [] tc;

  endGeometry();
  if (glid) glEndList();
  return (Object) glid;
}


Viewer::Object ViewerOpenGL::insertLineSet(size_t npoints,
                                           const float * points,
                                           size_t nlines,
                                           const long * lines,
                                           bool colorPerVertex,
                                           const float * color,
                                           size_t nci,
                                           const long * ci)
{
  GLuint glid = 0;

  if (npoints < 2) return 0;

#if defined(__ORBISNAP__)

  // check for consistency between color indices and vertex indices
  if (colorPerVertex)
  {
    if (nci != 0 && nci < nlines) return 0;
  }
  else
  {
    size_t i, lineCnt = 0;
    for (i = 0; i < nlines; i++)
    {
      if (lines[i] == -1) lineCnt++;
    }
    if (nci != 0 && nci < lineCnt) return 0;
  }

#endif

#if USE_GEOMETRY_DISPLAY_LISTS
  glid = glGenLists(1);
  glNewList( glid, GL_COMPILE_AND_EXECUTE );
#endif

  beginGeometry();

  // Lighting, texturing don't apply to line sets
  glDisable( GL_LIGHTING );
  glDisable( GL_TEXTURE_2D );
  if (color && ! colorPerVertex )
    glShadeModel( GL_FLAT );

  glBegin( GL_LINE_STRIP );
  if (color && ! colorPerVertex)
    glColor3fv( &color[ (nci > 0) ? 3*ci[0] : 0 ] );

  int nl = 0;
  for (size_t i = 0; i<nlines; ++i)
    {
      if (lines[i] == -1)
        {
          glEnd();
          glBegin( GL_LINE_STRIP );
          ++nl;
          if ((i < nlines-1) && color && ! colorPerVertex)
            glColor3fv( &color[ (nci > 0) ? 3*ci[nl] : 3*nl ] );
        }
      else
        {
          if (color && colorPerVertex)
            glColor3fv( &color[ (nci > 0) ? 3*ci[i] : 3*lines[i] ] );
          glVertex3fv( &points[3*lines[i]] );
        }
    }

  glEnd();

  endGeometry();
  if (glid) glEndList();
  return (Object) glid;
}

Viewer::Object ViewerOpenGL::insertPointSet(size_t npoints,
                                            const float * points,
                                            const float * colors)
{
  GLuint glid = 0;

#if USE_GEOMETRY_DISPLAY_LISTS
  glid = glGenLists(1);
  glNewList( glid, GL_COMPILE_AND_EXECUTE );
#endif

  beginGeometry();

  // Lighting, texturing don't apply to points
  glDisable( GL_LIGHTING );
  glDisable( GL_TEXTURE_2D );

  glBegin( GL_POINTS );

  for (size_t i = 0; i<npoints; ++i)
    {
      if (colors)
        {
          glColor3fv( colors );
          colors += 3;
        }
      glVertex3fv( points );
      points += 3;
    }

  glEnd();
  endGeometry();
  if (glid) glEndList();

  return (Object) glid;
}

//

static void computeBounds(size_t npoints, const float * points, float * bounds)
{
  bounds[0] = bounds[1] = points[0]; // xmin, xmax
  bounds[2] = bounds[3] = points[1]; // ymin, ymax
  bounds[4] = bounds[5] = points[2]; // zmin, zmax

  for (size_t i=1; i<npoints; ++i)
    {
      points += 3;
      if (points[0] < bounds[0])      bounds[0] = points[0];
      else if (points[0] > bounds[1]) bounds[1] = points[0];
      if (points[1] < bounds[2])      bounds[2] = points[1];
      else if (points[1] > bounds[3]) bounds[3] = points[1];
      if (points[2] < bounds[4])      bounds[4] = points[2];
      else if (points[2] > bounds[5]) bounds[5] = points[2];
    }
}


void
texGenParams( float bounds[],   // xmin,xmax, ymin,ymax, zmin,zmax
              int axes[2],      // s, t
              float params[4] ) // s0, 1/sSize, t0, 1/tSize
{
  axes[0] = 0;
  axes[1] = 1;
  params[0] = params[1] = params[2] = params[3] = 0.0;

  for (int nb=0; nb<3; ++nb)
    {
      float db = bounds[2*nb+1]-bounds[2*nb];
      if ( db > params[1] )
        {
          axes[1] = axes[0];
          axes[0] = nb;
          params[2] = params[0];
          params[3] = params[1];
          params[0] = bounds[2*nb];
          params[1] = db;
        }
      else if ( db > params[3] )
        {
          axes[1] = nb;
          params[2] = bounds[2*nb];
          params[3] = db;
        }
    }

  // If two of the dimensions are zero, give up.
  if ( FPZERO( params[1] ) || FPZERO( params[3] )) return;

  params[1] = params[3] = 1.0 / params[1];
}


// Address of _ith entry of indexed triplet array _v. yummy
#define INDEX_VAL(_v,_i) \
&((_v).v[ 3*(((_v).ni > 0) ? (_v).i[_i] : (_i)) ])

#define INDEX_VTX_VAL(_v,_f,_i) \
 &((_v).v[ 3*(((_v).ni > 0) ? (_v).i[_i] : (_f)[_i]) ])

void
ViewerOpenGL::insertShellConvex( ShellData *s )
{
  float N[3];
  size_t i;
  size_t nf = 0;                     // Number of faces

  for (i = 0; i<s->nfaces; ++i)
    {
      if (i == 0 || s->faces[i] == -1)
        {
          if (i > 0) glEnd();
          if (i == s->nfaces-1) break;

          glBegin(GL_POLYGON);

          // Per-face attributes
          if (s->color.v && ! (s->mask & MASK_COLOR_PER_VERTEX))
            glColor3fv( INDEX_VAL(s->color, nf) );

          if (! (s->mask & MASK_NORMAL_PER_VERTEX))
            {
              int i1 = (i == 0) ? 0 : i+1;
              if (s->normal.v)
              {
                // normalize the normal
                float N[3];
                Vnorm(N, INDEX_VAL(s->normal, nf));

                glNormal3fv(N);
              }
              else if (i < s->nfaces - 4 &&
                       s->faces[i1] >= 0 &&
                       s->faces[i1+1] >= 0 && s->faces[i1+2] >= 0)
                {
                  indexFaceNormal( 3*s->faces[i1], 3*s->faces[i1+1],
                                   3*s->faces[i1+2], s->points, N );

                  Vnorm(N);        // normalize the normal

                  // Lukas: flip normal if primitiv-orientation is clockwise
                  if (!(s->mask & MASK_CCW))
                    for (int k=0;k<3;k++) // flip Normal
                      N[k] = -N[k];
                  glNormal3fv( N );
                }
            }
            ++nf;
        }

      if (s->faces[i] >= 0)
        {
          // Per-vertex attributes
          if (s->color.v && (s->mask & MASK_COLOR_PER_VERTEX) )
            glColor3fv( INDEX_VTX_VAL(s->color, s->faces, i) );

          if (s->mask & MASK_NORMAL_PER_VERTEX)
            {
              if (s->normal.v)
              {
                // normalize the normal
                float N[3];
                Vnorm(N, INDEX_VTX_VAL(s->normal, s->faces, i));
                                // flip normal if CCW FALSE
                                if (!(s->mask & MASK_CCW))
                                  { N[0] = -N[0]; N[1] = -N[1]; N[2] = -N[2]; }
                glNormal3fv(N);
              }
              else
                ; // Generate per-vertex normal here...
            }

          const float * v = &s->points[3*s->faces[i]];
          if (s->texCoord.v)
            {
              int tcindex;
              if (s->texCoord.ni > 0)
                tcindex = 2 * s->texCoord.i[i];
              else
                tcindex = 2 * s->faces[i];
              glTexCoord2f( s->texCoord.v[ tcindex ],
                            s->texCoord.v[ tcindex+1 ] );
            }
          else
            {
              float c0, c1;
              c0 = (v[s->texAxes[0]] - s->texParams[0]) * s->texParams[1];
              c1 = (v[s->texAxes[1]] - s->texParams[2]) * s->texParams[3];
              glTexCoord2f( c0, c1 );
            }

          glVertex3fv( v );
        }
    }

  // Watch out for no terminating -1 in face list
  // two ways to break out:
  //   i>0 && i==nfaces-1 && faces[i] == -1
  //   i==nfaces
  //
  if (i>=s->nfaces) {
    if (s->faces[i-1] >= 0) glEnd();
  } else {
    if (s->faces[i] >= 0) glEnd();
  }
}


#if GLU_VERSION_1_2

static void WINAPI tessShellBegin( GLenum type, void *pdata )
{
  ViewerOpenGL::ShellData *s = (ViewerOpenGL::ShellData *)pdata;
  float N[3];

  glBegin( type );

  // Per-face attributes
  if (s->color.v && ! (s->mask & Viewer::MASK_COLOR_PER_VERTEX))
    glColor3fv( INDEX_VAL(s->color, s->nf) );

  if (! (s->mask & Viewer::MASK_NORMAL_PER_VERTEX))
    {
      int i1 = s->i == 0 ? 0 : s->i-1;
      if (s->normal.v)
      {
        // normalize the normal
        float N[3];
        Vnorm(N, INDEX_VAL(s->normal, s->nf));

        glNormal3fv(N);
      }
      else if (s->i < s->nfaces - 4 &&
               s->faces[i1] >= 0 &&
               s->faces[i1+1] >= 0 && s->faces[i1+2] >= 0)
        {
          indexFaceNormal( 3*s->faces[i1], 3*s->faces[i1+1],
                           3*s->faces[i1+2], s->points, N );

          Vnorm(N);           // normalize the normal

          // Lukas: flip normal if primitiv-orientation is clockwise
          if (!(s->mask & Viewer::MASK_CCW))
            for (int k=0;k<3;k++) // flip Normal
              N[k] = -N[k];
          glNormal3fv( N );
        }
    }
}


static void WINAPI tessShellVertex( void *vdata, void *pdata )
{
  ptrdiff_t i = (ptrdiff_t) vdata;
  ViewerOpenGL::ShellData *s = (ViewerOpenGL::ShellData *)pdata;

  // Per-vertex attributes
  if (s->color.v && (s->mask & Viewer::MASK_COLOR_PER_VERTEX) )
    glColor3fv( INDEX_VTX_VAL(s->color, s->faces, i) );

  if (s->mask & Viewer::MASK_NORMAL_PER_VERTEX)
    {
      if (s->normal.v)
      {
        // normalize the normal
        float N[3];
        Vnorm(N, INDEX_VTX_VAL(s->normal, s->faces, i));
        // flip normal if CCW FALSE
                if (!(s->mask & Viewer::MASK_CCW))
                  { N[0] = -N[0]; N[1] = -N[1]; N[2] = -N[2]; }
        glNormal3fv(N);
      }
      else
        ; // Generate per-vertex normal here...
    }

  const float * v = &s->points[3*s->faces[i]];
  if (s->texCoord.v)
    {
      int tcindex;
      if (s->texCoord.ni > 0)
        tcindex = 2 * s->texCoord.i[i];
      else
        tcindex = 2 * s->faces[i];
      glTexCoord2f( s->texCoord.v[ tcindex ],
                    s->texCoord.v[ tcindex+1 ] );
    }
  else
    {
      float c0, c1;
      c0 = (v[s->texAxes[0]] - s->texParams[0]) * s->texParams[1];
      c1 = (v[s->texAxes[1]] - s->texParams[2]) * s->texParams[3];
      glTexCoord2f( c0, c1 );
    }

  glVertex3fv( v );
}


void
ViewerOpenGL::insertShellTess(ShellData *s)
{
  if (! d_tess) d_tess = gluNewTess();

  gluTessCallback( d_tess, (GLenum) GLU_TESS_BEGIN_DATA,
                   (TessCB) tessShellBegin );
  gluTessCallback( d_tess, (GLenum) GLU_TESS_VERTEX_DATA,
                   (TessCB) tessShellVertex );
  gluTessCallback( d_tess, (GLenum) GLU_TESS_END,
                   (TessCB) glEnd );

  size_t i;
  for (i = 0; i<s->nfaces; ++i)
    {
      if (i == 0 || s->faces[i] == -1)
        {
          if (i > 0)
            {
              gluTessEndContour( d_tess );
              gluTessEndPolygon( d_tess );
              ++ s->nf;
            }
          if (i == s->nfaces-1) break;
          gluTessBeginPolygon( d_tess, s );
          gluTessBeginContour( d_tess );
          s->i = i;
        }

      if (s->faces[i] >= 0)
        {
          GLdouble v[3];
          v[0] = s->points[3*s->faces[i]+0];
          v[1] = s->points[3*s->faces[i]+1];
          v[2] = s->points[3*s->faces[i]+2];
          gluTessVertex( d_tess, v, (void*)i );
        }
    }

  // Watch out for no terminating -1 in face list
  if (i > 1 && s->faces[i-1] >= 0)
    {
      gluTessEndContour( d_tess );
      gluTessEndPolygon( d_tess );
    }
}

#endif // GLU_VERSION_1_2


// There are too many arguments to this...

Viewer::Object
ViewerOpenGL::insertShell(unsigned int mask,
                          size_t npoints,
                          const float * points,
                          size_t nfaces,
                          const long * faces,    // face list (-1 ends each face)
                          const float * tc,     // texture coordinates
                          size_t ntci,      // # of texture coordinate indices
                          const long * tci,      // texture coordinate indices
                          const float * normal, // normals
                          size_t nni,       // # of normal indices
                          const long * ni,       // normal indices
                          const float *color,  // colors
                          size_t nci,
                          const long * ci)
{
  // at least 3 points (and possibly trailing -1) is needed
  if (nfaces < 3 || (nfaces==3 && faces[nfaces-1]==-1))
  {
    // produce an empty object
    GLuint glid = glGenLists(1);
    glNewList(glid, GL_COMPILE_AND_EXECUTE);
    beginGeometry();
    endGeometry();
    glEndList();
    return (Object) glid;
  }

  // Texture coordinate generation parameters.
  int texAxes[2];                       // Map s,t to x,y,z
  float texParams[4];           // s0, 1/sSize, t0, 1/tSize

  // Compute bounding box for texture coord generation and lighting.
  if ( ! tc )  // || any positional lights are active...
    {
      float bounds[6];          // xmin,xmax, ymin,ymax, zmin,zmax
      computeBounds( npoints, points, bounds );

      // do the bounds intersect the radius of any active positional lights...

      texGenParams( bounds, texAxes, texParams );
      if ( FPZERO( texParams[1] ) || FPZERO( texParams[3] )) return 0;
    }

  GLuint glid = 0;

#if USE_GEOMETRY_DISPLAY_LISTS
  glid = glGenLists(1);
  glNewList( glid, GL_COMPILE_AND_EXECUTE );
#endif

  beginGeometry();

  // Face orientation & culling
  glFrontFace( (mask & MASK_CCW) ? GL_CCW : GL_CW );
  if (! (mask & MASK_SOLID) )
    glDisable( GL_CULL_FACE );

  // -------------------------------------------------------

  // Generation of per vertex normals isn't implemented yet...
  if (! normal && (mask & MASK_NORMAL_PER_VERTEX))
    mask &= ~MASK_NORMAL_PER_VERTEX;

  // -------------------------------------------------------

  // Should build tri strips (probably at the VrmlNode level)...

  ShellData s = {
    mask, points, nfaces, faces,
    { tc, ntci, tci },
    { normal, nni, ni },
    { color, nci, ci },
    texAxes, texParams, 0
  };

#if GLU_VERSION_1_2
  // Handle non-convex polys
  if (! (mask & MASK_CONVEX))
    insertShellTess( &s );
  else
#endif
    insertShellConvex( &s );

  endGeometry();
  if (glid) glEndList();

  return (Object) glid;
}


Viewer::Object ViewerOpenGL::insertSphere(float radius)
{
  GLuint glid = 0;

#if USE_GEOMETRY_DISPLAY_LISTS
  glid = glGenLists(1);
  glNewList( glid, GL_COMPILE_AND_EXECUTE );
#endif

#if defined(__ORBISNAP__)

  /* I think 10 looks too rough for larger spheres */
  const int numLatLong = 24;

#else

  const int numLatLong = 10;

#endif

  const int npts = numLatLong * numLatLong;

  float c[ npts ][ 3 ];
  float tc[ npts ][ 3 ];

  // should only compute tc if a texture is present...
  computeSphere(radius, numLatLong, c, tc, 0);

  beginGeometry();

  for ( int i = 0; i < numLatLong-1; ++i)
    {
      int n = i * numLatLong;

      glBegin(GL_TRIANGLE_STRIP);

      float N[3];
      for ( int j = 0; j < numLatLong; ++j )
        {
          glTexCoord2f( tc[n+j+numLatLong][0], tc[n+j+numLatLong][1] );
          Vnorm(N, &c[n+j+numLatLong][0]);       // normalize the normal
          glNormal3fv(N);
          glVertex3fv( &c[n+j+numLatLong][0] );

          glTexCoord2f( tc[n+j][0], tc[n+j][1] );
          Vnorm(N, &c[n+j][0]);                  // normalize the normal
          glNormal3fv(N);
          glVertex3fv( &c[n+j][0] );
        }

      //glTexCoord2f( tc[n+numLatLong][0], tc[n+numLatLong][1] );
      //cout << "ViewerOpenGL::insertSphere():" << tc[n+numLatLong][0] << "," << tc[n+numLatLong][1] << endl;
      glTexCoord2f( 1.0, tc[n+numLatLong][1] );
      Vnorm(N, &c[n+numLatLong][0]);             // normalize the normal
      glNormal3fv(N);
      glVertex3fv( &c[n+numLatLong][0] );

      //glTexCoord2f( tc[n][0], tc[n][1] );
      //cout << "ViewerOpenGL::insertSphere():" << tc[n][0] << "," << tc[n][1] << endl;
      glTexCoord2f( 1.0, tc[n][1] );
      Vnorm(N, &c[n][0]);                        // normalize the normal
      glNormal3fv(N);
      glVertex3fv( &c[n][0] );

      glEnd();
    }

  endGeometry();
  if (glid) glEndList();

  return (Object) glid;
}

// Not fully implemented... need font, extents

Viewer::Object ViewerOpenGL::insertText(FontFamily family,
                                        FontStyle style,
                                        int *justify,
                                        float size,
                                        float spacing,
                                        int n, char const * const * s)
{
  GLuint glid = 0;

#if USE_GEOMETRY_DISPLAY_LISTS
  glid = glGenLists(1);
  glNewList( glid, GL_COMPILE_AND_EXECUTE );
#endif

  beginGeometry();

  this->text3(family, style, justify, size, spacing, n, s);

  endGeometry();
  if (glid) glEndList();

  return (Object) glid;
}


// Lights

Viewer::Object ViewerOpenGL::insertDirLight(float ambient,
                                            float intensity,
                                            const float rgb[],
                                            const float direction[])
{
  float amb[4] = { ambient * rgb[0],
                   ambient * rgb[1],
                   ambient * rgb[2],
                   1.0 };
  float dif[4] = { intensity * rgb[0],
                   intensity * rgb[1],
                   intensity * rgb[2],
                   1.0 };

#if defined(__ORBISNAP__)

  /* This is a bugfix for badly placed directional lights. */
  float pos[4] = { -direction[0], -direction[1], -direction[2], 0.0 };

#else

  float pos[4] = { direction[0], direction[1], -direction[2], 0.0 };

#endif

  // Find an unused light, give up if none left.
  int i;
  for (i=0; i<MAX_LIGHTS; ++i)
    if (d_lightInfo[i].lightType == LIGHT_UNUSED)
      break;
  if (i == MAX_LIGHTS)
    return 0;

  d_lightInfo[i].lightType = LIGHT_DIRECTIONAL;
  d_lightInfo[i].nestingLevel = 0;
  GLenum light = (GLenum) (GL_LIGHT0 + i);

  glEnable(light);
  glLightfv(light, GL_AMBIENT, amb);
  glLightfv(light, GL_DIFFUSE, dif);
  glLightfv(light, GL_SPECULAR, dif);

  glLightfv(light, GL_POSITION, pos);

  // Disable any old point/spot settings
  glLightf(light, GL_CONSTANT_ATTENUATION, 1.0);
  glLightf(light, GL_LINEAR_ATTENUATION, 0.0);
  glLightf(light, GL_QUADRATIC_ATTENUATION, 0.0);
  glLightf(light, GL_SPOT_CUTOFF, 180.0);
  glLightf(light, GL_SPOT_EXPONENT, 0.0);

  return 0;
}

//
//  Only objects within radius should be lit by each PointLight.
//  Test each object drawn against each point light and enable
//  the lights accordingly? Get light and geometry into consistent
//  coordinates first...
//

Viewer::Object ViewerOpenGL::insertPointLight(float ambient,
                                              const float attenuation[],
                                              const float rgb[],
                                              float intensity,
                                              const float location[],
                                              float radius)
{
  float amb[4] = { ambient * rgb[0],
                   ambient * rgb[1],
                   ambient * rgb[2],
                   1.0 };
  float dif[4] = { intensity * rgb[0],
                   intensity * rgb[1],
                   intensity * rgb[2],
                   1.0 };
  float pos[4] = { location[0], location[1], location[2], 1.0 };

  // Find an unused light, give up if none left.
  int i;
  for (i=0; i<MAX_LIGHTS; ++i)
    if (d_lightInfo[i].lightType == LIGHT_UNUSED)
      break;
  if (i == MAX_LIGHTS)
    return 0;

  d_lightInfo[i].lightType = LIGHT_POSITIONAL;
  d_lightInfo[i].location[0] = location[0];
  d_lightInfo[i].location[1] = location[1];
  d_lightInfo[i].location[2] = location[2];
  d_lightInfo[i].radiusSquared = radius * radius;

  GLenum light = (GLenum) (GL_LIGHT0 + i);

  // should be enabled/disabled per geometry based on distance & radius...
  glEnable(light);
  glLightfv(light, GL_AMBIENT, amb);
  glLightfv(light, GL_DIFFUSE, dif);
  glLightfv(light, GL_SPECULAR, dif);
  glLightfv(light, GL_POSITION, pos);

#if defined(__ORBISNAP__)

  // attenuation of (0, 0, 0) has no mathematical meaning, VRML spec says
  // it should be considered (1, 0, 0)
  if (attenuation[0] == 0.0 && attenuation[1] == 0.0
    && attenuation[2] == 0.0)
  {
    glLightf(light, GL_CONSTANT_ATTENUATION, 1.0);
    glLightf(light, GL_LINEAR_ATTENUATION, 0.0);
    glLightf(light, GL_QUADRATIC_ATTENUATION, 0.0);
  }

  // set attenuation
  else
  {
    glLightf(light, GL_CONSTANT_ATTENUATION, attenuation[0]);
    glLightf(light, GL_LINEAR_ATTENUATION, attenuation[1]);
    glLightf(light, GL_QUADRATIC_ATTENUATION, attenuation[2]);
  }

#else

  glLightf(light, GL_CONSTANT_ATTENUATION, attenuation[0]);
  glLightf(light, GL_LINEAR_ATTENUATION, attenuation[1]);
  glLightf(light, GL_QUADRATIC_ATTENUATION, attenuation[2]);

#endif

  // Disable old spot settings
  glLightf(light, GL_SPOT_CUTOFF, 180.0);
  glLightf(light, GL_SPOT_EXPONENT, 0.0);

  return 0;
}

// same comments as for PointLight apply here...
Viewer::Object ViewerOpenGL::insertSpotLight( float ambient,
                                              const float attenuation[],
                                              float beamWidth,
                                              const float rgb[],
                                              float cutOffAngle,
                                              const float direction[],
                                              float intensity,
                                              const float location[],
                                              float radius )
{
  float amb[4] = { ambient * rgb[0],
                   ambient * rgb[1],
                   ambient * rgb[2],
                   1.0 };
  float dif[4] = { intensity * rgb[0],
                   intensity * rgb[1],
                   intensity * rgb[2],
                   1.0 };
  float pos[4] = { location[0], location[1], location[2], 1.0 };


  // Find an unused light, give up if none left.
  int i;
  for (i=0; i<MAX_LIGHTS; ++i)
    if (d_lightInfo[i].lightType == LIGHT_UNUSED)
      break;
  if (i == MAX_LIGHTS)
    return 0;

  d_lightInfo[i].lightType = LIGHT_POSITIONAL;
  d_lightInfo[i].location[0] = location[0];
  d_lightInfo[i].location[1] = location[1];
  d_lightInfo[i].location[2] = location[2];
  d_lightInfo[i].radiusSquared = radius * radius;

  GLenum light = (GLenum) (GL_LIGHT0 + i);

  // should be enabled/disabled per geometry based on distance & radius...
  glEnable(light);
  glLightfv(light, GL_AMBIENT, amb);
  glLightfv(light, GL_DIFFUSE, dif);
  glLightfv(light, GL_SPECULAR, dif);
  glLightfv(light, GL_POSITION, pos);

  glLightf(light, GL_CONSTANT_ATTENUATION, attenuation[0]);
  glLightf(light, GL_LINEAR_ATTENUATION, attenuation[1]);
  glLightf(light, GL_QUADRATIC_ATTENUATION, attenuation[2]);

  glLightfv(light, GL_SPOT_DIRECTION, direction);
  glLightf(light, GL_SPOT_CUTOFF, cutOffAngle * 180.0 / M_PI);
  // The exponential dropoff is not right/spec compliant...
  glLightf(light, GL_SPOT_EXPONENT, beamWidth < cutOffAngle ? 1.0 : 0.0);

  return 0;
}


// Lightweight copy

Viewer::Object ViewerOpenGL::insertReference(Object existingObject)
{
  glCallList( (GLuint) existingObject );
  return 0;
}

// Remove an object from the display list

void ViewerOpenGL::removeObject(Object key)
{
#if USE_GEOMETRY_DISPLAY_LISTS
  glDeleteLists( (GLuint) key, 1 );
#endif
}


void ViewerOpenGL::enableLighting(bool lightsOn)
{
  if (lightsOn)
    {
      if (d_lit) glEnable(GL_LIGHTING);
    }
  else
    glDisable(GL_LIGHTING);
}

// Set attributes

void ViewerOpenGL::setColor(float r, float g, float b, float a)
{
  glColor4f(r,g,b,a);
}

void ViewerOpenGL::setFog(const float * color,
                          float   visibilityRange,
                          const char * fogType)
{
  GLfloat fogColor[4] = { color[0], color[1], color[2], 1.0 };

#if defined(__ORBISNAP__)

  // Fog node with visibilityRange == 0 does nothing
  if (visibilityRange == 0) return;

  // There is a problem here: VRML's equation for exponential fog is
  // f=exp(-z/(visibility-z)) where z is the distance from the viewer
  // and f is the blending factor. OpenGL uses f=exp(-density*z)
  // or f=exp(-density*z*z). So I produced this "approximation".

  GLint fogMode = (strcmp(fogType,"EXPONENTIAL") == 0) ? GL_EXP : GL_LINEAR;
  glEnable(GL_FOG);
  glFogf(GL_FOG_START, 0.0);
  glFogf(GL_FOG_END, visibilityRange);
  glFogf(GL_FOG_DENSITY, 1/visibilityRange);
  glFogi(GL_FOG_MODE, fogMode);
  glFogfv(GL_FOG_COLOR, fogColor);

#else

  GLint fogMode = (strcmp(fogType,"EXPONENTIAL") == 0) ? GL_EXP : GL_LINEAR;

  glEnable( GL_FOG );
  glFogf( GL_FOG_START, 1.5 );  // What should this be?...
  glFogf( GL_FOG_END, visibilityRange );
  glFogi( GL_FOG_MODE, fogMode );
  glFogfv( GL_FOG_COLOR, fogColor );

#endif
}

void ViewerOpenGL::setMaterial(float ambientIntensity,
                               const float diffuseColor[],
                               const float emissiveColor[],
                               float shininess,
                               const float specularColor[],
                               float transparency)
{
  float alpha = 1.0 - transparency;

  float ambient[4] = { ambientIntensity*diffuseColor[0],
                       ambientIntensity*diffuseColor[1],
                       ambientIntensity*diffuseColor[2],
                       alpha };

  float diffuse[4] = { diffuseColor[0],
                       diffuseColor[1],
                       diffuseColor[2],
                       alpha };

  float emission[4] = { emissiveColor[0],
                        emissiveColor[1],
                        emissiveColor[2],
                        alpha };

  float specular[4] = { specularColor[0],
                        specularColor[1],
                        specularColor[2],
                        alpha };

  // doesn't work right yet (need alpha render pass...)
  if (d_blend && ! FPZERO(transparency))
    glEnable(GL_BLEND);

  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, ambient);
  glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, diffuse);

#if defined(__ORBISNAP__)

  // OpenGL shininess range is <0, 128> while VRML shininess is <0, 1>.
  glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, shininess*128.0);

#else

  glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, shininess);

#endif

  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);

  if (ambientIntensity == 0. &&
      diffuse[0] == 0. && diffuse[1] == 0. && diffuse[2] == 0. &&
      specularColor[0] == 0. && specularColor[1] == 0. && specularColor[2] == 0.)
    {
      glDisable(GL_LIGHTING);
      glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, diffuse);
      glColor4fv( emission );
    }
  else
    {
      glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, emission);
      glColor4fv( diffuse );
    }
}


// This hack is necessary because setting the color mode needs to know
// about the appearance (presence & components of texture) and the geometry
// (presence of colors). Putting this stuff in either insertTexture or
// insert<geometry> causes problems when the texture or geometry node is
// USE'd with a different context.

void ViewerOpenGL::setMaterialMode (int textureComponents,
                                    bool colors)
{
  if (textureComponents && d_texture && ! d_wireframe)
    {
      glEnable( GL_TEXTURE_2D );

      // This is a hack: if per-{face,vertex} colors are specified,
      // they take precedence over textures with GL_MODULATE. The
      // textures won't be lit this way but at least they show up...
      if (textureComponents > 2 && colors)
        glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE );
      else
        glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
    }
  else
    glDisable( GL_TEXTURE_2D );

  if (colors && textureComponents < 3 /* && lighting enabled... */ )
    glEnable( GL_COLOR_MATERIAL );
  else
    glDisable( GL_COLOR_MATERIAL );
}

void ViewerOpenGL::setSensitive(void *object)
{
  if (object)
    {
    // push name, increment object count
    if (getRenderPhase() == RENDER_PHASE_ANALYSIS)
        {
      d_sensitiveObject.push_back(object);
      glPushName(++d_nSensitive); // array index+1
        }
    else if (getRenderPhase() == RENDER_PHASE_TRANSPARENT)
      ++d_nSensitive;
    }
  else if (getRenderPhase() == RENDER_PHASE_ANALYSIS)
    glPopName();
}



// Set maximum texture size
void ViewerOpenGL::setMaxTextureSize(int texsize)
{
  // texture size must be at most the maximum hardware texture size and at least 32
  if (texsize==0 || texsize>d_maxTextureSizeHW)
    texsize=d_maxTextureSizeHW;
  else if (texsize<32)
    texsize=32;

  // ensure the texture size is power of two
  int maxsize = d_maxTextureSizeHW;
  int newsize;
  for (newsize = maxsize; newsize > texsize; newsize>>=1);

  // store the new maximum texture size
  d_maxTextureSize = newsize;
  if (d_scene)
    d_scene->invalidateTextures();
}


//
// Pixels are lower left to upper right by row.
//
Viewer::TextureObject
ViewerOpenGL::insertTexture(size_t w, size_t h, size_t nc,
                            bool repeat_s,
                            bool repeat_t,
                            const unsigned char * pixels,
                            bool retainHint)
{
  GLenum fmt[] = { GL_LUMINANCE,        // single component
                   GL_LUMINANCE_ALPHA,  // 2 components
                   GL_RGB,              // 3 components
                   GL_RGBA              // 4 components
  };

  // immediately return in select mode, if no pixels supplied or if unknown format
  if (d_selectMode || !pixels || nc>4) return 0;

  // Texture dimensions are limited by max texture size and must be powers of two
  int neww, newh;
  for (neww = d_maxTextureSize; neww >= 2*w; neww>>=1);   // find next higher power of two
  for (newh = d_maxTextureSize; newh >= 2*h; newh>>=1);

  // rescale texture to the next higher power of two
  unsigned char* newPixels = 0;  // temporary pixel array
  if (w != neww || h != newh)
  {
    // rescale small textures using pixel replication
#ifdef __WXMSW__  // wxWidgets functionality available only on Windows - see g806188
    if ((w<PIXEL_REPLICATION_LIMIT || h<PIXEL_REPLICATION_LIMIT))
    {
      // scale images to at least PIXEL_REPLICATION_SIZE in both directions
      // to minimize aliasing effects when using pixel replication algorithm
      neww = std::max(neww, PIXEL_REPLICATION_SIZE);
      newh = std::max(newh, PIXEL_REPLICATION_SIZE);
      newPixels = new unsigned char[nc*neww*newh];

      // scale the texture using wxImage
      wxImage img(w, h, const_cast<unsigned char*>(pixels), true);  // we will not modify the original pixel data
      wxImage newimg = img.Scale(neww, newh);

      // copy scaled pixels to destination array
      memcpy(newPixels, newimg.GetData(), neww*newh*nc*sizeof(unsigned char));
      pixels = newPixels;
      w = neww;
      h = newh;
    }

    // rescale bigger textures using interpolation
    else
#endif // __WXMSW__
    {
      newPixels = new unsigned char[nc*neww*newh];
      glPixelStorei( GL_PACK_ALIGNMENT, 1 );
      glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );
      if (gluScaleImage(fmt[nc-1], w, h, GL_UNSIGNED_BYTE, pixels,
                        neww, newh, GL_UNSIGNED_BYTE, newPixels) == 0)
      {
        // use rescaled image on success
        w = neww;
        h = newh;
        pixels = newPixels;
      }
    }
  }

  // Microsoft GDI does not support RGB format
  // we always pass RGBA and set alpha to non-transparent
  if (nc == 3)
  {
    // add alpha channel to texture data
    nc = 4;
    unsigned char* rgbaPixels = new unsigned char[w*h*nc];
    const unsigned char* pixelsPos = pixels;
    unsigned char* newPixelsPos = rgbaPixels;
    for (int i=0; i<w*h; ++i)
    {
      *newPixelsPos++ = *pixelsPos++;
      *newPixelsPos++ = *pixelsPos++;
      *newPixelsPos++ = *pixelsPos++;
      *newPixelsPos++ = 255;
    }

    // delete any previous temporary pixel array and use the new one
    delete[] newPixels;
    pixels = newPixels = rgbaPixels;
  }

  // Enable blending if needed
  if (d_blend && (nc == 2 || nc == 4))
    glEnable(GL_BLEND);

  GLuint glid = 0;
#if USE_TEXTURE_DISPLAY_LISTS
  if (retainHint) glGenTextures(1, &glid);
  glBindTexture( GL_TEXTURE_2D, glid );
#endif

  // Texturing is enabled in setMaterialMode
  glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );

  glTexImage2D( GL_TEXTURE_2D, 0, nc, w, h, 0,
                fmt[nc-1], GL_UNSIGNED_BYTE, pixels);

  // delete any temporary pixel array
  delete[] newPixels;

  glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S,
                   repeat_s ? GL_REPEAT : (d_glVersion>=1.2) ? GL_CLAMP_TO_EDGE : GL_CLAMP );
  glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T,
                   repeat_t ? GL_REPEAT : (d_glVersion>=1.2) ? GL_CLAMP_TO_EDGE : GL_CLAMP );

  /* Apply texture antialiasing */
  glTexParameterf
  (
    GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,
    d_antialias ? GL_LINEAR : GL_NEAREST
  );
  glTexParameterf
  (
    GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
    d_antialias ? GL_LINEAR : GL_NEAREST
  );

  return (TextureObject) glid;
}

void ViewerOpenGL::insertTextureReference(TextureObject t, int nc)
{
#if USE_TEXTURE_DISPLAY_LISTS

  if (!glIsTexture((GLuint) t))
  {
    theSystem->warn(System::SEVERITY_HIGH, System::WARNINGTYPE_UNKNOWN, "Texture object %d not precompiled although it should be", (GLuint) t);
    return;
  }

  /* Apply texture antialiasing */
  glTexParameterf
  (
    GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,
    d_antialias ? GL_LINEAR : GL_NEAREST
  );
  glTexParameterf
  (
    GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
    d_antialias ? GL_LINEAR : GL_NEAREST
  );

  // Enable blending if needed
  if (d_blend && (nc == 2 || nc == 4))
    glEnable(GL_BLEND);

  glBindTexture( GL_TEXTURE_2D, (GLuint) t );

#endif
}


void ViewerOpenGL::removeTextureObject(TextureObject t)
{
#if USE_TEXTURE_DISPLAY_LISTS
  GLuint glid = (GLuint) t;
  glDeleteTextures( 1, &glid );
#endif
}

// Texture coordinate transform
// Tc' = -C x S x R x C x T x Tc

void ViewerOpenGL::setTextureTransform(const float center[2],
                                       float rotation,
                                       const float scale[2],
                                       const float translation[2])
{
  glMatrixMode(GL_TEXTURE);
  glLoadIdentity();

  if (center) glTranslatef(-center[0], -center[1], 0.0);
  if (scale) glScalef(scale[0], scale[1], 1.0);
  if (rotation != 0.0)
    glRotatef(rotation * 180.0 / M_PI, 0.0, 0.0, 1.0);

  if (center) glTranslatef(center[0], center[1], 0.0);
  if (translation) glTranslatef(translation[0], translation[1], 0.0);

  glMatrixMode(GL_MODELVIEW);
}

// Transforms
// P' = T x C x R x SR x S x -SR x -C x P

void ViewerOpenGL::setTransform(const float center[3],
                                const float rotation[4],
                                const float scale[3],
                                const float scaleOrientation[4],
                                const float translation[3])
{
  // put the previous matrix on the stack (we use our own transform stack
  // because the OpenGL stack is too limited in size)
  Matrix4x4 m;
  double data[16];
  glGetDoublev(GL_MODELVIEW_MATRIX, data);
  m.setColumns(data);
  transformStack.push(m);

  //cout << "ViewerOpenGL::setTransform()" << endl;
  glTranslatef(translation[0], translation[1], translation[2]);
  glTranslatef(center[0], center[1], center[2]);

  if (! FPZERO(rotation[3]) )
    glRotatef(rotation[3] * 180.0 / M_PI,
              rotation[0],
              rotation[1],
              rotation[2]);

  if (! FPEQUAL(scale[0], 1.0) ||
      ! FPEQUAL(scale[1], 1.0) ||
      ! FPEQUAL(scale[2], 1.0) )
    {
      if (! FPZERO(scaleOrientation[3]) )
        glRotatef(scaleOrientation[3] * 180.0 / M_PI,
                  scaleOrientation[0],
                  scaleOrientation[1],
                  scaleOrientation[2]);

      glScalef(scale[0], scale[1], scale[2]);

      if (! FPZERO(scaleOrientation[3]) )
        glRotatef(-scaleOrientation[3] * 180.0 / M_PI,
                  scaleOrientation[0],
                  scaleOrientation[1],
                  scaleOrientation[2]);
    }

  glTranslatef(-center[0], -center[1], -center[2]);
}

void ViewerOpenGL::unsetTransform(const float center[3],
                                  const float rotation[4],
                                  const float scale[3],
                                  const float scaleOrientation[4],
                                  const float translation[3])
{
  // restore the matrix from the stack
  Matrix4x4 m = transformStack.top();
  double data[16];
  m.getColumns(data);
  glLoadMatrixd(data);
  transformStack.pop();
}

// The matrix gets popped at endObject() - Not anymore. I added
// an explicit unsetBillboardTransform to work around the matrix
// depth limit of 32 in mesa. Now the limit only applies to
// nested billboards.

void ViewerOpenGL::setBillboardTransform(const float axisOfRotation[3])
{
// Following code is modified by S. K. Bose March 02/2000 to fix the Billboardtransform
// according to the specification.

  float M[4][4];
  getBillboardTransformMatrix(M, axisOfRotation);

  glPushMatrix();
  glMultMatrixf(&M[0][0]);

}

void ViewerOpenGL::unsetBillboardTransform(const float axisOfRotation[3])
{
  glPopMatrix();
}

void ViewerOpenGL::setViewpoint(const float *position,
                                float *orientation,
                                float fieldOfView,
                                float avatarSize,
                                float visibilityLimit)
{
  //cout << "ViewerOpenGL::setViewpoint()" << endl;
  //cout << "  p:(" << position[0] << ","<< position[1] << ","<< position[2] << ")" << endl;

  glMatrixMode( GL_PROJECTION );
  if (! d_selectMode) glLoadIdentity();

  float restrictedFoV = 180.0 / M_PI * min (2.0 * atan(tan(fieldOfView/2.0)/getCamera().getZoom()), maxFoV );

  float aspect = ((float) width) / (height - d_bottomReserved);

  float znear = (float) d_zNear;
  float zfar = (float) d_zFar;

  // calculate the perspective projection matrix
  gluPerspective(restrictedFoV, aspect, znear, zfar);

  VrmlFrustum frust(restrictedFoV, aspect, znear, zfar);
  this->setFrustum(frust);
  //frust.dump(cout) << endl;

  glMatrixMode(GL_MODELVIEW);

  // View modifiers are applied in first beginObject
  if (d_rotationChanged)
  {
    build_rotmatrix(d_rotationMatrix, d_curquat);
    d_rotationChanged = false;
  }
}


// The viewer knows the current viewpoint

void ViewerOpenGL::transformPoints(int np, float *p)
{
  float m[16];
  glGetFloatv (GL_MODELVIEW_MATRIX, m);

  float x, y, z;
  for (int i=0; i<np; ++i)
    {
      x = m[0]*p[0] + m[4]*p[1] + m[8]*p[2] + m[12];
      y = m[1]*p[0] + m[5]*p[1] + m[9]*p[2] + m[13];
      z = m[2]*p[0] + m[6]*p[1] + m[10]*p[2] + m[14];

      p[0] = x;
      p[1] = y;
      p[2] = z;
      p += 3;
    }

}

//
//  Viewer callbacks (called from window system specific functions)
//

// update is called from a timer callback and from checkSensitive
void ViewerOpenGL::update()
{
  if (d_scene->update())
    {
      checkOpenGLError("ViewerOpenGL::update()");
      requestRedraw();
    }
}


void ViewerOpenGL::drawHud()
{
}


void ViewerOpenGL::invalidateGL()
{
  d_GLinitialized = false;
}

void ViewerOpenGL::redraw()
{
  // initialize everything if not done yet
  if (!d_GLinitialized) initialize();

  // resize if necessary
  if (newwidth!=width || newheight!=height)
  {
    width = newwidth;
    height = newheight;
    glViewport(0, d_bottomReserved, width, height-d_bottomReserved);
  }

  // catch errors that could be accumulated from before
  checkOpenGLError("ViewerOpenGL::redraw(): at start");

  // get start time (for profiling purposes)
  double start = theSystem->time();

  // get the current drawing area size
  if (height<=0)  // status bar etc. may take up the whole window
    return;

  // use first bindable node in scene (if any) in editor mode
  if (d_scene && d_scene->isEdited())
  {
    if (viewpoint==NULL)
    {
      d_scene->bindSceneDefaultBindableNodes(VrmlScene::VRML_NODE_VIEWPOINT);
      viewpoint = d_scene->bindableViewpointTop();
    }

    if (navinfo==NULL)
    {
      d_scene->bindSceneDefaultBindableNodes(VrmlScene::VRML_NODE_NAVIGATION_INFO);
      navinfo = d_scene->bindableNavigationInfoTop();
    }

    if (background==NULL)
    {
      d_scene->bindSceneDefaultBindableNodes(VrmlScene::VRML_NODE_BACKGROUND);
      background = d_scene->bindableBackgroundTop();
    }

    if (fog==NULL)
    {
      d_scene->bindSceneDefaultBindableNodes(VrmlScene::VRML_NODE_FOG);
      fog = d_scene->bindableFogTop();
    }
  }


  // we can only draw the scene if we have one and at least one pixel height for it
  if (d_scene && height>d_bottomReserved && !d_renderSuppressed)
  {

    // determine avatar size and visibility limit
    float field = M_PI_4;
    float avatarSize = 0.25;
    float visLimit = 0.0;
    if (viewpoint) field = viewpoint->getFieldOfView().get();
    if (navinfo) avatarSize = navinfo->avatarSize()[0];
    if (navinfo) visLimit = navinfo->visibilityLimit();

    // Explicitly defined avatarSize and visibilityLimit tested here only
    // as different from default values, could be a flag in NavigationInfo.
    // If defined explicitly, no znear/zfar guessing should take place.
    // Other approach is to guess always, ignore explicit hints, but that needs
    // a perfect guessing algorithm. We can't ignore hints without having one.
    bool zNearExplicit = !FPEQUAL(avatarSize, 0.25);
    bool zFarExplicit = !FPZERO(visLimit);

    // visibilityLimit 0 indicates an infinite visibility, should be more than 10000
    if (visLimit == 0.0) visLimit = 10000.0;
    float restrictedFoV =  180.0 / M_PI * min (2.0 * atan(tan(field/2.0)/getCamera().getZoom()), maxFoV);
    float aspect = ((float) width) / (height - d_bottomReserved);

    // if the camera has changed, recalculate the near/far plane
    if (isCameraChanged() && !d_selectMode)
    {
      float nearDist = 0.5*avatarSize;
      float farDist = visLimit;
      Matrix4x4 m = getViewingTransformation();
      d_zNear = (m*Vector3(nearDist, 0.0, 0.0)).length();
      d_zFar = (m*Vector3(farDist, 0.0, 0.0)).length();

      // guess only if at least one of znear/zfar hints not explicitly defined,
      if (!zNearExplicit || !zFarExplicit)
      {
	double zNearGuessed = d_zNear;
	double zFarGuessed = d_zFar;
	if (planeClipMethod==PLANECLIP_BVOLUME)
	{
	  Camera camabs = getCameraAbsolute();
          setViewingSector(VrmlViewingSector(restrictedFoV, aspect, camabs.getYawAxis(), camabs.getYawAngle()));
	  guessNearFarPlaneWithTreeBrowsing(&zNearGuessed, &zFarGuessed, d_zNear, d_zFar, zNearExplicit, zFarExplicit);
        }
	else
	{
	  guessNearFarPlaneWithSelMode(&zNearGuessed, &zFarGuessed);
	  checkOpenGLError("ViewerOpenGL::redraw(): after near-far guess");
	}

	// accept only guessed values that are not defined explicitly
	// (it is possible that only one of znear/zfar hints is defined)
	if (!zNearExplicit) d_zNear = zNearGuessed;
	if (!zFarExplicit) d_zFar = zFarGuessed;
      }
    }

    if (!d_selectMode) resetOpenGLState();

    // lighting, shading and wireframe are irrelevant in select mode
    if (d_selectMode)
    {
      glDisable(GL_LIGHTING);
      glShadeModel(GL_FLAT);
      glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    }

    // protect the panel area from overdrawing (setting viewport is not enough
    // as it does not guarantee clipping).
    glScissor(0, d_bottomReserved, width, height-d_bottomReserved);
    glEnable(GL_SCISSOR_TEST);

    // reset object counters
    d_nObjects = 0;
    d_nestedObjects = 0;
    d_nSensitive = 0;
    d_sensitiveObject.clear();

    // Clean out any defined lights
    for (int i=0; i<MAX_LIGHTS; ++i)
    {
      d_lightInfo[i].lightType = LIGHT_UNUSED;
      GLenum light = (GLenum) (GL_LIGHT0 + i);
      glDisable( light );
    }

    if (!d_selectMode) resetOpenGLState();

    // catch errors during the state setup
    checkOpenGLError("ViewerOpenGL::redraw(): after pre-render setup");

    // calculate projection transformation
    glMatrixMode(GL_PROJECTION);
    if (!d_selectMode) glLoadIdentity();          // keep select mode xform
    gluPerspective(restrictedFoV, aspect, d_zNear, d_zFar);
    setFrustum(VrmlFrustum(restrictedFoV, aspect, d_zNear, d_zFar));

    // headlight is of no use in select mode
    if (!d_selectMode)
    {
      // insert headlight (this must be done before camera transformations)
      if (d_headlight)
      {
        float rgb[3] = { 1.0, 1.0, 1.0 };
        float xyz[3] = { 0.0, 0.0, -1.0 };
        float ambient = 0.0;
        insertDirLight(ambient, 1.0, rgb, xyz);
      }
    }

    // calculate viewing transformation
    Matrix4x4 m = getViewingTransformation();
    double glm[16];
    m.getColumns(glm);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glLoadMatrixd(glm);

    if (d_selectMode)
    {
      // no need for nice backgrounds in select mode, use default
      insertBackground();
    }
    else
    {
      // render background (use default if not specified)
      if (background) background->renderBindable(this);
      else insertBackground();
    }

    // fog is irrelevant in select mode
    if (!d_selectMode)
    {
      // render fog if specified
      if (fog) setFog(fog->color(), fog->visibilityRange(), fog->fogType());
    }

    // catch errors from preparation stages
    checkOpenGLError("ViewerOpenGL::redraw(): before rendering scene");


    // render the scene
    if (d_selectMode)
    {
      glDepthMask(GL_TRUE);
      d_renderPhase = RENDER_PHASE_ANALYSIS;
      d_scene->render(this, false);
      d_renderPhase = RENDER_PHASE_NONE;

      // reinitialize
      d_nObjects = 0;
      d_nestedObjects = 0;
    }
    else
    {
      GLboolean	lighting = GL_TRUE;
      glGetBooleanv(GL_LIGHTING, &lighting);
      if (lighting)
      {
        // first phase: draw lights
        d_renderPhase = RENDER_PHASE_LIGHTS;
        glDepthMask(GL_TRUE);
        d_scene->render(this, false);
        d_renderPhase = RENDER_PHASE_NONE;

        // reinitialize
        d_nObjects = 0;
        d_nestedObjects = 0;
      }

      // second phase: draw non-transparent objects
      d_renderPhase = RENDER_PHASE_NON_TRANSPARENT;
      glDepthMask(GL_TRUE);
      d_scene->render(this, false);
      d_renderPhase = RENDER_PHASE_NONE;

      // reinitialize
      d_nObjects = 0;
      d_nestedObjects = 0;

      // third phase: draw transparent objects
      d_renderPhase = RENDER_PHASE_TRANSPARENT;
      glDepthMask(GL_FALSE);
      d_scene->render(this, true);
      d_renderPhase = RENDER_PHASE_NONE;
      glDepthMask(GL_TRUE);
    }

    // catch errors from rendering
    checkOpenGLError("ViewerOpenGL::redraw(): after rendering");

    // remove the scissor rectangle so that HUD can be drawn
    glScissor(0, d_bottomReserved, width, height);
    glDisable(GL_SCISSOR_TEST);
  }


  else
  {
    // rendering is not possible now, clear to black color and return
    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
    glClear(GL_COLOR_BUFFER_BIT);
  }

  // in select mode, HUD and FPS are irrelevant and screen is not updated
  if (!d_selectMode)
  {
    // draw HUD (panel + nav zones + info bar)
    drawHud();

    wsSwapBuffers();

    // we have redrawn so camera is no more changed from the last redraw
    unsetCameraChanged();
  }

  // catch errors from post-rendering (mainly HUD drawing)
  checkOpenGLError("ViewerOpenGL::redraw(): before exit");
}

void ViewerOpenGL::resetOpenGLState()
{
  glDisable(GL_FOG);
  glDisable(GL_TEXTURE_2D);
  glEnable(GL_CULL_FACE);
  glFrontFace(GL_CCW);
  glCullFace(GL_BACK);
  if (d_lit) glEnable(GL_LIGHTING);
  glDisable(GL_COLOR_MATERIAL);
  glDisable(GL_BLEND);
  glShadeModel(GL_SMOOTH);
  glPolygonMode(GL_FRONT_AND_BACK, d_wireframe ? GL_LINE : GL_FILL);
  glViewport(0, d_bottomReserved, width, height-d_bottomReserved);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
}

void ViewerOpenGL::resizeCanvas(int w, int h)
{
  newwidth = std::max(w, 2);
  newheight = std::max(h, d_bottomReserved);
  requestRedraw();
}


void ViewerOpenGL::input( EventInfo *e )
{
  // reset the viewport
  if (height>d_bottomReserved)
    glViewport(0, d_bottomReserved, width, height-d_bottomReserved);

  switch (e->event)
  {
    case EVENT_KEY_DOWN:
      handleKey( e->what ); break;

    case EVENT_MOUSE_MOVE:
      (void) checkSensitive( e->x, e->y, EVENT_MOUSE_MOVE ); break;

    case EVENT_MOUSE_CLICK:
    case EVENT_MOUSE_RELEASE:
      handleButton( e ); break;
    case EVENT_MOUSE_DRAG:
    {
      initializeMouseDrag(e);
      handleMouseDrag( e->x, e->y );
      break;
    }
  }
}

void ViewerOpenGL::handleKey(int key)
{
  Camera cam = getCamera();

  switch (key)
  {
    // turn left/right
    case KEY_LEFT:
    case KEY_RIGHT:
    case 'o':
    case 'p':
    {
      setCamera(cam.rotateH(
        ((key == KEY_LEFT || key == 'o') ? 0.05 : -0.05)*d_rotationStep));
      break;
    }

    // look up/down
    case 'a':
    case 'z':
    {
      setCamera(cam.rotateV(
        ((key == 'a') ? 0.05 : -0.05)*d_rotationStep));
      break;
    }

    // tilt left/right
    case 'n':
    case 'm':
    {
      setCamera(cam.rotateS(
        ((key == 'n') ? 0.05 : -0.05)*d_rotationStep));
      break;
    }

    // straighten up
    case '0':
    {
      setCamera(cam.straightenUp());
      break;
    }

    // forward/backward
    case KEY_UP:
    case KEY_DOWN:
    {
      setCamera(cam.step(
        (key == KEY_UP) ? d_motionStep : -d_motionStep));
      break;
    }

    // slide left/right
    case 'O':
    case 'P':
    {
      setCamera(cam.slideH(
        (key == 'P') ? d_motionStep : -d_motionStep));
      break;
    }

    // slide up/down
    case 'A':
    case 'Z':
    {
      setCamera(cam.slideV(
        (key == 'A') ? d_motionStep : -d_motionStep));
      break;
    }

    // orbit left/right/up/down around the world centre
    case ',':
    case '.':
    case 's':
    case 'x':
    {
      Camera worldCam = getCameraAbsolute();
      if (key == ',') worldCam.orbitH(Point3(0, 0, 0), 0.05*d_rotationStep);
      else if (key == '.') worldCam.orbitH(Point3(0, 0, 0), -0.05*d_rotationStep);
      else if (key == 's') worldCam.orbitV(Point3(0, 0, 0), 0.05*d_rotationStep);
      else if (key == 'x') worldCam.orbitV(Point3(0, 0, 0), -0.05*d_rotationStep);
      setCameraAbsolute(worldCam);
      break;
    }

/*
    // bind/unbind the camera
    case '\\':
    {
      bindCamera(!isCameraBound());
      break;
    }
*/

    // zoom
    case 'f':
    case 'g':
    {
      if (key == 'f') cam.setZoom(cam.getZoom()*1.1);
      else cam.setZoom(cam.getZoom()/1.1);
      setCamera(cam);
      break;
    }

/*
    case KEY_PAGE_DOWN:
      gotoNextViewpoint();
      break;

    case KEY_PAGE_UP:
      gotoPrevViewpoint();
      break;

    case 'b':
      d_blend = ! d_blend;
      wsPostRedraw();
      break;

    case 'd':
      d_drawBSpheres = ! d_drawBSpheres;
      wsPostRedraw();
      break;

    case 'c':
      d_cull = ! d_cull;
      wsPostRedraw();
      break;

    case 'l':
      d_lit = ! d_lit;
      wsPostRedraw();
      break;

    case KEY_HOME:
    case 'r':                   // Reset view
      resetUserNavigation();
      break;
*/

    case KEY_ESCAPE:
    case 0x1b:
      gotoViewpoint(0);
      break;

/*
    case 't':
      d_texture = ! d_texture;
      wsPostRedraw();
      break;

    case 'w':                   // Wireframe (should disable texturing)
      d_wireframe = ! d_wireframe;
      glPolygonMode(GL_FRONT_AND_BACK, d_wireframe ? GL_LINE : GL_FILL);
      wsPostRedraw();
      break;
 */

    default:
      break;
    }
}

// Mouse button up/down

void ViewerOpenGL::handleButton( EventInfo *e)
{
  d_rotating = d_scaling = d_translating = false;

  // Check for a sensitive object first	(if SHIFT keyboard is down then sensitives' checking is disabled )
  if (e->what == MOUSE_BUTTON_LEFT && e->modifier != EVENT_SHIFT_DOWN && checkSensitive(e->x, e->y, e->event))
    return;

  d_activeSensitive = 0;

  // nothing under mouse, do navigation
  if (e->event == EVENT_MOUSE_RELEASE)
  {
    // mouse button was released
    d_dragging = false;
    d_dragZone = 0;

    // restore normal cursor
    wsSetCursor(CURSOR_ARROW);
  }
  else if (e->event == EVENT_MOUSE_CLICK && e->what == MOUSE_BUTTON_LEFT)
  {
    // Left mouse button was pressed -> potential start of mouse drag.
    initializeMouseDrag(e);
  }
}


// Mouse movement with button down

void ViewerOpenGL::handleMouseDrag(int x, int y)
{
  // check whether a sensor was activated
  if (d_activeSensitive)
  {
    (void) checkSensitive( x, y, EVENT_MOUSE_DRAG );
  }

  // no sensor was triggered, use mouse for navigation
  else
  {
    Camera cam = getCamera();

    // calculate deltas of the mouse pointer position
    int dx = x - d_prevx;
    int dy = y - d_prevy;

    // mouse navigation is different according to navigation mode
    if (d_navMode == NAVMODE_WALK)
    {
      // Walk mode.
      // Central & border zone:
      //     mouse motion left/right: turn left/right
      //     mouse motion up/down: move forwards/backwards
      // Left & right edge zone:
      //     mouse motion up/down: move up/down
      //     mouse motion left/right: move left/right
      if (d_dragZone == MOUSE_ZONE_CENTRAL
          || d_dragZone == MOUSE_ZONE_BORDER)
      {
        if (abs(dx) > abs(dy))
          setCamera(cam.rotateH(((dx>0) ? -0.02 : 0.02)*d_rotationStep));
        else if (abs(dy) > abs(dx))
          setCamera(cam.step((dy>0) ? -d_motionStep : d_motionStep));
      }
      else
      {
        if (abs(dy) > abs(dx))
          setCamera(cam.slideV((dy > 0) ? -d_motionStep : d_motionStep));
        else if (abs(dx) > abs(dy))
          setCamera(cam.slideH((dx > 0) ? d_motionStep : -d_motionStep));
      }
    }

    else if (d_navMode == NAVMODE_EXAMINE)
    {
      // Examine mode.
      // Central & border zone:
      //    mouse motion left/right: orbit left/right
      //    mouse motion up/down: orbit up/down
      // Left & right edge zone:
      //    mouse motion up/down: move forwards/backwards
      //    mouse motion left/right: move left/right
      if (d_dragZone == MOUSE_ZONE_CENTRAL
          || d_dragZone == MOUSE_ZONE_BORDER)
      {
        // orbit around the world centre
        Camera worldCam = getCameraAbsolute();
        if (abs(dx) > abs(dy))
          worldCam.orbitH(Point3(0, 0, 0), ((dx > 0) ? -0.02 : 0.02)*d_rotationStep);
        else
          worldCam.orbitV(Point3(0, 0, 0), ((dy > 0) ? -0.02 : 0.02)*d_rotationStep);
        setCameraAbsolute(worldCam);
      }
      else
      {
        if (abs(dy) > abs(dx))
          setCamera(cam.step((dy > 0) ? -d_motionStep : d_motionStep));
        else if (abs(dx) > abs(dy))
          setCamera(cam.slideH((dx > 0) ? d_motionStep : -d_motionStep));
      }
    }
    else if (d_navMode == NAVMODE_FLY)
    {
      // Fly mode.
      // Central zone:
      //    mouse motion left/right: turn left/right
      //    mouse motion up/down: forward/backward
      // Border zone:
      //    mouse motion left/right: turn left/right
      //    mouse motion up/down: turn up/down
      // Left edge zone:
      //    mouse motion up/down: tilt right/left
      // Right edge zone:
      //    mouse motion up/down: tilt left/right

      if (d_dragZone == MOUSE_ZONE_CENTRAL)
      {
        if (dy != 0)
          setCamera(cam.step((dy > 0) ? -d_motionStep : d_motionStep));
        if (dx != 0)
          setCamera(cam.rotateH(((dx > 0) ? -0.02 : 0.02)*d_rotationStep));
      }
      else if (d_dragZone == MOUSE_ZONE_LEFT_EDGE
          || d_dragZone == MOUSE_ZONE_RIGHT_EDGE)
      {
        if (abs(dy) > abs(dx) && dy != 0)
        {
          double delta = (dy > 0) ? 0.02 : -0.02;
          if (x>width/12) delta = -delta;
          setCamera(cam.rotateS(delta*d_rotationStep));
        }
      }
      else
      {
        if (dx != 0)
          setCamera(cam.rotateH(((dx > 0) ? -0.02 : 0.02)*d_rotationStep));
        if (dy != 0)
          setCamera(cam.rotateV(((dy > 0) ? -0.02 : 0.02)*d_rotationStep));
      }
    }
  }

  d_prevx = x;
  d_prevy = y;
}

void ViewerOpenGL::initializeMouseDrag(EventInfo* e)
{
  if (!d_dragging)
  {
    d_dragging = true;
    d_beginx = e->x;
    d_beginy = e->y;

    if (e->event == EVENT_MOUSE_CLICK && e->what == MOUSE_BUTTON_LEFT)
    {
      d_prevx = e->x;
      d_prevy = e->y;
    }

    // look which mouse zone it is
    if ((e->y > height/3) && (e->y < 2*(height/3))
        && (e->x > width/3) && (e->x < 2*(width/3)))
    {
      d_dragZone = MOUSE_ZONE_CENTRAL;
    }
    else if (e->x < width/12)
    {
      d_dragZone = MOUSE_ZONE_LEFT_EDGE;
    }
    else if (e->x > 11*(width/12))
    {
      d_dragZone = MOUSE_ZONE_RIGHT_EDGE;
    }
    else
    {
      d_dragZone = MOUSE_ZONE_BORDER;
    }

    // store the undo information so that we can revert the drag
    pushCameraState();

    // set crosshair cursor
    wsSetCursor(CURSOR_CROSSHAIR);
  }
}

bool ViewerOpenGL::guessNearFarPlaneWithSelMode(double* nearVal, double* farVal)
{
  // get viewport parameters
  GLint viewport[4];
  glGetIntegerv(GL_VIEWPORT, viewport);

  // prepare buffers
  GLuint selectBuffer[SELECTION_BUFFER_SIZE];
  glSelectBuffer(SELECTION_BUFFER_SIZE, selectBuffer);

  // render the scene to memory (with initial values of znear and zfar)
  // and remember which points are where
  (void) glRenderMode(GL_SELECT);
  d_selectMode = true;
  glPushAttrib(GL_ENABLE_BIT);
  glDisable(GL_FOG);
  glDisable(GL_TEXTURE_2D);
  glFrontFace(GL_CCW);
  glDisable(GL_CULL_FACE);
  glDisable(GL_LIGHTING);
  glDisable(GL_DEPTH_TEST);
  glDisable(GL_COLOR_MATERIAL);
  glShadeModel(GL_FLAT);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glMatrixMode(GL_MODELVIEW);
  glInitNames();
  glPushName(0);
  glLoadIdentity();
  d_nObjects = 0;
  d_nestedObjects = 0;
  d_nSensitive = 0;
  d_renderPhase = RENDER_PHASE_ANALYSIS;
  ViewerOpenGL::redraw();
  d_renderPhase = RENDER_PHASE_NONE;
  d_selectMode = false;

  // find lowest and highest z-value in the rendered scene
  GLuint minz = 0xFFFFFFFFU;
  GLuint maxz = 0;
  int hits = glRenderMode(GL_RENDER);

  // selectBuffer[] = { N1, znear, zfar, {list of N1 names}, N2, ... }
  GLuint *sel = selectBuffer;
  for (int nh=0; nh<hits; ++nh, sel+=(3+sel[0]))
  {
    if (sel[1] < minz) minz = sel[1];
    if (sel[1] > maxz) maxz = sel[1];
    if (sel[2] > maxz) maxz = sel[2];
  }

  // no objects in the view, sorry
  // this also takes care for ATI Radeon driver bug (version 8.1 and later)
  if (minz>=maxz)
  {
    glPopAttrib();
    return(false);
  }

  // calculate lowest and highest used Z coordinate (in screen coordinates)
  double nearPoint = ((double) minz)/0xFFFFFFFFU;
  double farPoint = ((double) maxz)/0xFFFFFFFFU;

  // add some tolerances
  double tol = (farPoint-nearPoint);
  nearPoint -= tol;
  farPoint += tol;
  farPoint = (farPoint > 1) ? 1 : farPoint;

  double znear, zfar;
  GLdouble modelview[16], projection[16];
  glGetDoublev (GL_PROJECTION_MATRIX, projection);

  // we are at the top level of scene graph, use identity as modelview
  memset(modelview, 0, sizeof(GLdouble)*16);
  modelview[0] = 1.0;
  modelview[5] = 1.0;
  modelview[10] = 1.0;
  modelview[15] = 1.0;

  // translate the z-values to the real distance
  GLdouble dx, dy, dz;
  gluUnProject(
    (GLdouble) 0.0, (GLdouble) 0.0,
    nearPoint, modelview, projection, viewport,
    &dx, &dy, &dz);
  znear = -dz;
  gluUnProject(
    (GLdouble) 0.0, (GLdouble) 0.0,
    farPoint, modelview, projection, viewport,
    &dx, &dy, &dz);
  zfar = -dz;

  // restore original settings
  glPopAttrib();

  // extend znear to zfar range for flat scenes to at least 100
  double zrange = zfar/znear;
  if (zrange < 100)
  {
    double zrscale = sqrt(100/zrange);
    znear /= zrscale;
    zfar *= zrscale;
  }

  // store computed znear and zfar
  if (nearVal) *nearVal = znear;
  if (farVal) *farVal = zfar;

  return true;
}

void ViewerOpenGL::guessNearFarPlaneWithTreeBrowsing(double* nearVal, double* farVal, double defaultNearVal, double defaultFarVal, bool zNearExplicit, bool zFarExplicit)
{
  // constants for zfar and znear guessing
  static const double Z_NEAR_INCEPTION = 524288;
  static const double Z_FAR_INCEPTION = 1000;
  static const double RANGE_LIMIT = 100;
  static const double Z_NEAR_PROPORTION = 1000;
  static const double Z_FAR_PROPORTION = 10000;

  // set starting (default) values
  double zNearGuessed = Z_NEAR_INCEPTION;
  double zFarGuessed = Z_FAR_INCEPTION;

  // browse the scene to get get near/far values
  d_scene->guessNearFarPlane(this, &zNearGuessed, &zFarGuessed);

  // add tolerance to znear
  if (zNearGuessed < 10000) zNearGuessed = zNearGuessed/Z_NEAR_PROPORTION;
  else if (zNearGuessed < 100000) zNearGuessed = zNearGuessed/100;
  else zNearGuessed = zNearGuessed/(Z_NEAR_PROPORTION*log10(zNearGuessed));

  // add tolerance to zfar
  if (zFarGuessed < 10000) zFarGuessed = zFarGuessed + zFarGuessed/2;
  else zFarGuessed = zFarGuessed + zFarGuessed/3;

  double zrange;
  if (!zNearExplicit && !zFarExplicit)
  {
    // both values are guessed
    if (zFarGuessed < zNearGuessed)
    {
      // no object in viewing sector, set values to default
      zNearGuessed = defaultNearVal;
      zFarGuessed = defaultFarVal;
    }
    zrange = zFarGuessed/zNearGuessed;
    if (zrange < RANGE_LIMIT)
    {
      zNearGuessed = zNearGuessed/Z_NEAR_PROPORTION;
    }
  }

  if (zNearExplicit)
  {
    // only zfar is guessed
    zNearGuessed = defaultNearVal;
    if (zFarGuessed < (zNearGuessed * Z_NEAR_PROPORTION))
    {
      zFarGuessed = zNearGuessed * Z_NEAR_PROPORTION;
    }
  }

  if (zFarExplicit)
  {
    // only znear is guessed
    zFarGuessed = defaultFarVal;
    if (zFarGuessed < zNearGuessed)
    {
      zNearGuessed = zFarGuessed/Z_FAR_PROPORTION;
    }
    zrange = zFarGuessed/zNearGuessed;
    if (zrange < RANGE_LIMIT)
    {
      zNearGuessed = zNearGuessed/Z_NEAR_PROPORTION;
    }
  }
  *nearVal = zNearGuessed;
  *farVal = zFarGuessed;
}


void ViewerOpenGL::checkSensitiveForCurrentMousePosition()
{
  if (!d_dragging)
    checkSensitive(getMouseCurrentXCoord(), getMouseCurrentYCoord(), EVENT_MOUSE_TIME_GENERATED);
}

// Check for pickable objects.

bool ViewerOpenGL::checkSensitive(int x, int y, EventType mouseEvent)
{
  // do not check sensitive objects in editor mode
  if (d_scene && d_scene->isEdited())
    return false;

  if (checkSensitiveDisabled)
  {
    wsSetCursor( CURSOR_ARROW );
    return false;
  }


#if defined(__ORBISNAP__)

  // Shortcut: do not check for sensitives if there are none.
  // The number of sensitives is set every time the scene is rendered
  // (the scene calls setSensitive() method) so we can safely return
  // if it is set to zero.
  if (d_nSensitive == 0) return false;

#endif

  // return if scene is not properly rendered yet.
  // this resolves 'textures disappearing on transparent shapes' problem during docking
  if (d_scene->isModified() || d_scene->texturesDirty()) return false;


  double timeNow = theSystem->time();


  // don't do this time-consuming test too often
  if ((mouseEvent == EVENT_MOUSE_MOVE) && (timeNow - d_sensorCheckTime < MOUSE_MOVE_CHECK_SENSITIVE_INTERVAL))
    return false;

  if ((mouseEvent == EVENT_MOUSE_TIME_GENERATED) && (timeNow - d_sensorCheckTime < MOUSE_TIME_GENERATED_CHECK_SENSITIVE_INTERVAL))
    return false;

  // reset opengl state	to ensure viewport does not contain nav. panel area
  resetOpenGLState();

  // initialize OpenGL select mode
  GLint viewport[4];
  glGetIntegerv( GL_VIEWPORT, viewport );
  std::vector<GLuint> selectBuffer(d_selectionBufferSize);
  glSelectBuffer(d_selectionBufferSize, &(selectBuffer.front()) );
  (void) glRenderMode( GL_SELECT );
  glInitNames();
  glPushName(0);
  glMatrixMode( GL_PROJECTION );
  glLoadIdentity();

  gluPickMatrix( (GLdouble) x, (GLdouble)(viewport[3] - y + d_bottomReserved),
                 2.0, 2.0, viewport );

  // render the scene in select mode, the hits will be stored
  d_selectMode = true;
  ViewerOpenGL::redraw();
  d_selectMode = false;

  // select closest hit
  GLuint closest = 0, minz = 0xffffffff;

  int selected = 0;             // nothing selected
  int hits = glRenderMode( GL_RENDER );

  if (hits > 0)
    {
      // selectBuffer[] = { N1, z1, z2, name1, ..., nameN1, N2, ... }
      GLuint *sel = &(selectBuffer.front());

      for (int nh=0; nh<hits; ++nh, sel+=(3+sel[0]))
        {
          if (sel[1] <= minz)
            {
              minz = sel[1];            // z1
              closest = sel[2+sel[0]];  // nameN (most deeply nested)
            }
        }

      if (closest > 0 && closest <= (GLuint)d_nSensitive)
        {
          selected = closest;
        }
    }
    else if (hits == -1)
    {
      // select buffer overflow -- extend it
      selectBuffer.clear();
      d_selectionBufferSize *= 2;
      // try again
      return checkSensitive(x,y,mouseEvent);
    }


  wsSetCursor( selected ? CURSOR_SENSOR : CURSOR_ARROW );

  // Compute & store the world coords of the pick if something
  // was already active or is about to become active. The window
  // Z coord is retained when a drag is started on a sensitive
  // so the drag stays in the same plane even if the mouse moves
  // off the original sensitive object.

  double selectCoord[3] = { 0.0, 0.0, 0.0 };

  if (d_activeSensitive || selected)
    {
      if (! d_activeSensitive) d_selectZ = minz / (double)0xffffffff;

      // set the viewing matrix according to the camera placement
      glMatrixMode(GL_MODELVIEW);
      Matrix4x4 m;
      m = getViewingTransformation();
      GLdouble modelview[16];
      m.getColumns(modelview);

      GLint viewport[4];
      GLdouble projection[16];
      glGetIntegerv (GL_VIEWPORT, viewport);
      glGetDoublev (GL_PROJECTION_MATRIX, projection);

      GLdouble dx, dy, dz;
      gluUnProject( (GLdouble) x,
        (GLdouble) (viewport[3] - y + d_bottomReserved), d_selectZ,
                modelview, projection, viewport,
                &dx, &dy, &dz);

      selectCoord[0] = dx;
      selectCoord[1] = dy;
      selectCoord[2] = dz;
    }

  bool wasActive = false;

  // Sanity check. This can happen when the world gets replaced
  // by clicking on an anchor - the current sensitive object goes
  // away, but these variables are not reset.
  if ( d_activeSensitive > d_nSensitive )
    d_activeSensitive = 0;
  if ( d_overSensitive > d_nSensitive )
    d_overSensitive = 0;

  // An active sensitive object "grabs" the mouse until button released
  if (d_activeSensitive)
    {
      if (mouseEvent == EVENT_MOUSE_RELEASE)
        {
          d_scene->sensitiveEvent( d_sensitiveObject[ d_activeSensitive-1 ],
                                   timeNow,
                                   selected == d_activeSensitive, false,
                                   selectCoord );
          d_activeSensitive = 0;
        }
      else                      // _DRAG
        {
          d_scene->sensitiveEvent( d_sensitiveObject[ d_activeSensitive-1 ],
                                   timeNow,
                                   selected == d_activeSensitive, true,
                                   selectCoord );
        }
      wasActive = true;
    }

  // A click down over a sensitive object initiates an active grab and
  // mouse over events are no longer relevant.
  else if (mouseEvent == EVENT_MOUSE_CLICK && selected != 0)
    {
      if (d_overSensitive && d_overSensitive != selected)
        {
          d_scene->sensitiveEvent( d_sensitiveObject[ d_overSensitive-1 ],
                                   timeNow,
                                   false, false, // isOver, isActive
                                   selectCoord );
          d_overSensitive = 0;
        }
      d_activeSensitive = selected;
      d_scene->sensitiveEvent( d_sensitiveObject[ d_activeSensitive-1 ],
                               timeNow,
                               true, true,  // isOver, isActive
                               selectCoord );
    }

  // Handle isOver events (coords are bogus)
  else if (mouseEvent == EVENT_MOUSE_MOVE || mouseEvent == EVENT_MOUSE_TIME_GENERATED)
    {
      if (d_overSensitive && d_overSensitive != selected)
        {
          d_scene->sensitiveEvent( d_sensitiveObject[ d_overSensitive-1 ],
                                   timeNow,
                                   false, false, // isOver, isActive
                                   selectCoord );
        }
      d_overSensitive = selected;
      if (d_overSensitive)
        d_scene->sensitiveEvent( d_sensitiveObject[ d_overSensitive-1 ],
                                 timeNow,
                                 true, false,  // isOver, isActive
                                 selectCoord );
    }


  // Was event handled here?
  if (d_activeSensitive || wasActive)
    update();

#ifdef __ORBISNAP__
  // remember last check time
  d_sensorCheckTime = timeNow;
#endif

  // Everything is handled except down clicks where nothing was selected
  // and up clicks where nothing was active.
  return d_activeSensitive || wasActive;
}

Point3 ViewerOpenGL::projectBeam(int x, int y)
{
  // initialize OpenGL select mode
  GLint viewport[4];
  glGetIntegerv( GL_VIEWPORT, viewport );
  GLuint selectBuffer[ SELECTION_BUFFER_SIZE ];
  glSelectBuffer( SELECTION_BUFFER_SIZE, selectBuffer );
  (void) glRenderMode( GL_SELECT );
  glInitNames();
  glPushName(0);
  glMatrixMode( GL_PROJECTION );
  glLoadIdentity();

  gluPickMatrix( (GLdouble) x, (GLdouble)(viewport[3] - y + d_bottomReserved),
                 2.0, 2.0, viewport );

  // render the scene in select mode, the hits will be stored
  d_selectMode = true;
  ViewerOpenGL::redraw();
  d_selectMode = false;

  // select closest hit
  GLuint closest = 0, minz = 0xffffffff;
  int selected = 0;             // nothing selected
  int hits = glRenderMode( GL_RENDER );
  if (hits > 0)
  {
    // selectBuffer[] = { N1, z1, z2, name1, ..., nameN1, N2, ... }
    GLuint *sel = selectBuffer;
    for (int nh=0; nh<hits; ++nh, sel+=(3+sel[0]))
    {
      if (sel[1] <= minz)
      {
        minz = sel[1];            // z1
        closest = sel[2+sel[0]];  // nameN (most deeply nested)
      }
    }
    if (closest > 0 && closest <= (GLuint)d_nSensitive) selected = closest;
  }

  // compute the world coordinates of the touch point
  double selectCoord[3] = { 0.0, 0.0, 0.0 };
  double selectZ = minz / (double)0xffffffff;

  // set the viewing matrix according to the camera placement
  glMatrixMode(GL_MODELVIEW);
  Matrix4x4 m;
  m = getViewingTransformation();
  GLdouble modelview[16];
  m.getColumns(modelview);

  GLdouble projection[16];
  glGetIntegerv (GL_VIEWPORT, viewport);
  glGetDoublev (GL_PROJECTION_MATRIX, projection);

  GLdouble dx, dy, dz;
  gluUnProject( (GLdouble) x,
    (GLdouble) (viewport[3] - y + d_bottomReserved), selectZ,
            modelview, projection, viewport,
            &dx, &dy, &dz);

  selectCoord[0] = dx;
  selectCoord[1] = dy;
  selectCoord[2] = dz;

  return Point3(selectCoord[0], selectCoord[1], selectCoord[2]);
}

// Text rendering

/**
 * Produces an R8G8B8A8 image containing the given text.
 * @param textLine Text to be drawn.
 * @param width The image width
 * (always a power of two).
 * @param height The image height
 * (always a power of two).
 * @param lmargin The margin to count for Drawtext drawing bold/italics
 * more to the left thank it should
 * @param result Pointer to variable to receive the pointer
 * to the newly created image data.
 * @param color The text color.
 */
#ifdef __WXMSW__  // wxWidgets functionality available only on Windows - see g806188
static void TextToRGBAData(wxFont font,
                           const char* textLine,
                           const unsigned int width,
                           const unsigned int height,
                           const unsigned int lmargin,
                           unsigned char** result,
                           const float* color)
{
  wxCoord w,h;
  // create the drawing context
  wxMemoryDC dc;
  // draw the text to a bitmap (grayscale, text is white),
  // convert it to RGB format
  wxBitmap bmp(width, height);
  dc.SelectObject(bmp);
  dc.SetFont(font);
  dc.SetTextForeground(*wxWHITE);
  dc.SetBackground(*wxBLACK_BRUSH);
  dc.Clear();
  dc.DrawText(wxString(textLine), lmargin, 0);
  wxImage img = bmp.ConvertToImage();
  img = img.Mirror(FALSE);

  // convert RGB to RGBA (black is transparent)
  w=(wxCoord) width;
  h=(wxCoord) height;
  unsigned char* data = (unsigned char*) img.GetData();
  unsigned char* rgba = new unsigned char[w*h*4];
  unsigned int i;
  for (i=0; i<w*h; i++)
  {
    rgba[4*i  ] = (int)(color[0] * 255);   // red
    rgba[4*i+1] = (int)(color[1] * 255);   // green
    rgba[4*i+2] = (int)(color[2] * 255);   // blue
    rgba[4*i+3] = data[3*i];  // alpha: black=transparent, white=opaque
  }

  *result = rgba;
}
#endif // __WXMSW__

void ViewerOpenGL::text3(FontFamily family, FontStyle style, int *justify,
                         float size, float spacing, int n, const char * const *s)
{
#ifdef __WXMSW__  // wxWidgets functionality available only on Windows - see g806188
  // prepare attributes
  glPushAttrib(GL_ENABLE_BIT);
  glDisable(GL_LIGHTING);
  glEnable(GL_TEXTURE_2D);
  glEnable(GL_BLEND);
  glEnable(GL_ALPHA_TEST);
  glAlphaFunc(GL_GREATER, 0.1);
  glBindTexture(GL_TEXTURE_2D, 0);

  // get the current color
  GLfloat color[4];
  glGetFloatv(GL_CURRENT_COLOR, color);

  GLfloat x = 0.0, y = 0.0;
  // height of the texture into which the text will be rendered (2^n)
  unsigned int txr_h = 64;
  // pointsize 42 in most cases leads to th<=64
  unsigned int pointsize = 42;
  // DrawText writes bold / italics text more to the left than it should
  // with th=64, bold / italics is drawn shifted ~4 pixels to the left
  unsigned int lmargin = 4;
  unsigned int maxtsize = getMaxTextureSize();

  // prepare the font
  int wxfamily = wxROMAN;
  int wxstyle = wxNORMAL;
  int wxweight = wxNORMAL;
  switch (family)
  {
    case FONT_FAMILY_SERIF:      wxfamily = wxROMAN; break;
    case FONT_FAMILY_SANS:       wxfamily = wxSWISS; break;
    case FONT_FAMILY_TYPEWRITER: wxfamily = wxMODERN; break;
  }
  switch (style)
  {
    case FONT_STYLE_BOLD:         wxweight = wxBOLD; break;
    case FONT_STYLE_ITALIC:       wxstyle = wxITALIC; break;
    case FONT_STYLE_BOLD_ITALIC:  wxweight = wxBOLD;
                                  wxstyle = wxITALIC; break;
  }

  wxFont font(pointsize, wxfamily, wxstyle, wxweight);

  // prepare the device context for getting the per line extent values
  wxMemoryDC dc;
  wxBitmap dummy(1, 1, -1);
  // text width, height, descent
  wxCoord tw, th, td;
  dc.SelectObject(dummy);
  dc.SetFont(font);

  // print the text lines
  for (int i=0; i<n; ++i, y-=size*spacing)
  {
    const char *tLine = s[i];
    if ( tLine )
    {
      // get the current text line extent
      dc.GetTextExtent(wxString(tLine), &tw, &th, &td);
      // ensure that font fits into the texture height
      while (th > txr_h)
      {
         pointsize -= 1;
         font.SetPointSize(pointsize);
         dc.SetFont(font);
         dc.GetTextExtent(wxString(tLine), &tw, &th, &td);
      }

      // compute the line shift depending on the major axis justification
      if (justify[0] < 0)
              x = - (double) (tw+lmargin) * size / (double) txr_h;
      else if (justify[0] == 0)
              x = - 0.5 * (double) (tw+lmargin) * size / (double) txr_h;
      else if (justify[0] > 0)
              x = - (double) lmargin * size / (double) txr_h;

      // move to the start of the line - texture shifted also down by text descent
      glPushMatrix();
      glTranslatef(x, ( td ? y - (double) td * size / (double) txr_h : y ), 0.0);
      glScalef(size, size, size);

      // draw the text lines by segments of MaxtTextureSize (soft, settable preference) width

      // Ensure that texture width is powers of two, >= than 32 and >= text line width extent
      // taking into account also lmargin
      unsigned int txr_w;
      for (txr_w = 32; txr_w < (tw+lmargin); txr_w<<=1);

      // prepare the text texture
      glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
      unsigned char* rgba;
      TextToRGBAData(font, tLine, txr_w, txr_h, lmargin, &rgba, color);

      if (txr_w <= maxtsize)
      {
        // Render the texture in one go
        glTexImage2D(GL_TEXTURE_2D, 0, 4, txr_w, txr_h, 0, GL_RGBA,
                     GL_UNSIGNED_BYTE, rgba);
        delete [] rgba;

        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, (d_glVersion>=1.2) ? GL_CLAMP_TO_EDGE : GL_CLAMP);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, (d_glVersion>=1.2) ? GL_CLAMP_TO_EDGE : GL_CLAMP);

        // render it
        glBegin(GL_QUADS);
        glColor3d(1.0, 1.0, 1.0);
        glTexCoord2d(0, 0);  glVertex3f(0, 0, 0);
        glTexCoord2d(1, 0);  glVertex3f((GLfloat) txr_w/(GLfloat) txr_h, 0, 0);
        glTexCoord2d(1, 1);  glVertex3f((GLfloat) txr_w/(GLfloat) txr_h, 1, 0);
        glTexCoord2d(0, 1);  glVertex3f(0, 1, 0);
        glEnd();
      }
      else
      {
        // Split the image into horizontal tiles w<= MaxTextureSize
        unsigned int tile_w = maxtsize;
        unsigned char *txr_tile = new unsigned char[4*txr_h*tile_w];

        for (unsigned int offset=0; offset<txr_w; offset+=tile_w)
        {
          // textures in memory in RGBA blocks,
          // lines of points L->R stacked bottom to top
          for (unsigned int line=0;line<txr_h;line++)
          {
             memcpy(txr_tile + 4*tile_w*line, rgba + 4*txr_w*line + 4*offset, 4*tile_w);
          }
            glTexImage2D(GL_TEXTURE_2D, 0, 4, tile_w, txr_h, 0, GL_RGBA,
                         GL_UNSIGNED_BYTE, txr_tile);

            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, (d_glVersion>=1.2) ? GL_CLAMP_TO_EDGE : GL_CLAMP);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, (d_glVersion>=1.2) ? GL_CLAMP_TO_EDGE : GL_CLAMP);

            // render it
            glBegin(GL_QUADS);
            glColor3d(1.0, 1.0, 1.0);
            glTexCoord2d(0, 0);  glVertex3f(0, 0, 0);
            glTexCoord2d(1, 0);  glVertex3f((GLfloat) tile_w/(GLfloat) txr_h, 0, 0);
            glTexCoord2d(1, 1);  glVertex3f((GLfloat) tile_w/(GLfloat) txr_h, 1, 0);
            glTexCoord2d(0, 1);  glVertex3f(0, 1, 0);
            glEnd();

            // move to the next horizontal tile position
            glTranslatef((GLfloat) tile_w / (GLfloat) txr_h, 0.0, 0.0);
          }
          delete [] txr_tile;
          delete [] rgba;
        }

        // return to the start of the line
        glPopMatrix();
        checkOpenGLError("text rendering");
    }
  }

  // clean up
  dc.Clear();
  glPopAttrib();
#endif // __WXMSW__
}

static GLfloat green[] = {0.25f, 1.0f, 0.25f, 1.0f};
static GLfloat red[] = {1.0f, 0.5f, 0.5f, 1.0f};
static GLfloat grey[] = {0.5f, 0.5f, 0.5f, 1.0f};

// argh: see <URL:http://www.debian.org/Bugs/db/44/44489-b.html> for
// an explanation. quick version: mesa doesn't work right with newver
// c++ compilers (c++ has recently become more strict about
// enums). should this go into configure.in? probably, since it might
// occur outside this file. the problem should go away with the next
// version of mesa.. -cks
//
#ifdef MESA
#if (GLU_VERSION_1_1 || GLU_VERSION_1_2)
#define MESA_GLU_ENUM_FIX (GLenum)
#else
#define MESA_GLU_ENUM_FIX
#endif
#else
#define MESA_GLU_ENUM_FIX
#endif


void
ViewerOpenGL::drawBSphere(const VrmlBSphere& bs, int flag)
{
  if (!d_drawBSpheres) return;
  //cout << "ViewerOpenGL::insertBSphere(" << bs.getRadius() << ")" << endl;
  if (bs.isMAX()) return;
  if (bs.getRadius()==-1.0) return;
  //bs.dump(cout);
  glEnable(GL_CULL_FACE);
  glCullFace(GL_BACK);
  //glEnable(GL_LIGHTING);
  glShadeModel(GL_FLAT);
  GLUquadricObj* sph = (GLUquadricObj*)0;
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  const float* c = bs.getCenter();
  glTranslatef(c[0], c[1], c[2]);
  sph = gluNewQuadric();
  if (flag == VrmlBVolume::BV_OUTSIDE) {
    //cout << "out" << endl;
    //glDisable(GL_LIGHTING);
    //glEnable(GL_LIGHTING);
    gluQuadricDrawStyle(sph, MESA_GLU_ENUM_FIX GLU_LINE);
    //gluQuadricDrawStyle(sph, GLU_POINT);
    glColor3f(0.5, 0.5, 0.5);
    glMaterialfv(GL_FRONT, GL_DIFFUSE, grey);
    gluSphere(sph, bs.getRadius(), 5, 5);
  } else if (flag == VrmlBVolume::BV_PARTIAL) {
    //cout << "par" << endl;
    //glEnable(GL_LIGHTING);
    gluQuadricNormals(sph, MESA_GLU_ENUM_FIX GLU_SMOOTH);
    //gluQuadricDrawStyle(sph, GLU_FILL);
    gluQuadricDrawStyle(sph, MESA_GLU_ENUM_FIX GLU_LINE);
    glColor3f(0.25, 1.0, 0.25);
    glMaterialfv(GL_FRONT, GL_DIFFUSE, green);
    gluSphere(sph, bs.getRadius(), 8, 8);
  } else if (flag == VrmlBVolume::BV_INSIDE) {
    //cout << "ins" << endl;
    //glEnable(GL_LIGHTING);
    gluQuadricNormals(sph, MESA_GLU_ENUM_FIX GLU_SMOOTH);
    //gluQuadricDrawStyle(sph, GLU_FILL);
    gluQuadricDrawStyle(sph, MESA_GLU_ENUM_FIX GLU_LINE);
    glMaterialfv(GL_FRONT, GL_DIFFUSE, red);
    glColor3f(1.0, 0.5, 0.5);
    gluSphere(sph, bs.getRadius(), 8, 8);
  } else {
    //cout << "oth" << endl;
    //glEnable(GL_LIGHTING);
    gluQuadricNormals(sph, MESA_GLU_ENUM_FIX GLU_SMOOTH);
    gluQuadricDrawStyle(sph, MESA_GLU_ENUM_FIX GLU_LINE);
    glMaterialfv(GL_FRONT, GL_DIFFUSE, grey);
    glColor3f(0.5, 0.5, 0.5);
    gluSphere(sph, bs.getRadius(), 8, 8);
  }
  gluDeleteQuadric(sph);

  glPopMatrix();

}



//---------------------------------------------------------------------------
// debugging functions
//---------------------------------------------------------------------------



const char* ViewerOpenGL::getOpenGLErrorMessage(int glErrno)
{
  switch (glErrno)
  {
    case GL_NO_ERROR:          return "No error";
    case GL_INVALID_ENUM:      return "Invalid enum";
    case GL_INVALID_VALUE:     return "Invalid value";
    case GL_INVALID_OPERATION: return "Invalid operation";
    case GL_STACK_OVERFLOW:    return "Stack overflow";
    case GL_STACK_UNDERFLOW:   return "Stack underflow";
    case GL_OUT_OF_MEMORY:     return "Out of memory";
    default:                   return "Unknown error";
  }
}

bool ViewerOpenGL::checkOpenGLError(const char* where)
{
  int glErrno = glGetError();
  if (glErrno == GL_NO_ERROR) return true;
  else
  {
    theSystem->warn(System::SEVERITY_HIGH, System::WARNINGTYPE_UNKNOWN, "%s: OpenGL error: %s\n",
                    where,
                    getOpenGLErrorMessage(glErrno));
    return false;
  }
}
