/*****************************************************************************
******************************************************************************
*
*               OpennGL viewer
*
*               $Revision: 1.1.4.1.2.18 $
*               $Date: 2010/12/25 14:24:28 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright 1998-2010 HUMUSOFT s.r.o.
// Portions Copyright (C) 1998  Chris Morley

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#ifndef VIEWEROPENGL_H
#define VIEWEROPENGL_H

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stack>

#include "Viewer.h"


// Use the stencil buffer to set the SHAPE mask.
#define USE_STENCIL_SHAPE 0


// Win32 needs to know that GL stuff lives in a DLL (OPENGL32.DLL)
#if defined(WIN32)
#include <wtypes.h>
#include <winbase.h>
#include <windef.h>
#include <wingdi.h>
#endif

#include "Math3d.h"
#include "VrmlNodeViewpoint.h"
#include "VrmlNodeNavigationInfo.h"
#include "VrmlNodeBackground.h"
#include "VrmlNodeFog.h"
#include "System.h"  // OPENVRMLDLLEXPORT

extern "C" class GLUtesselator;

/**
 * Abstract class for display of VRML models using OpenGL/Mesa.
 * A window-system specific subclass needs to redefine the pure
 * virtual methods.
 */
class OPENVRMLDLLEXPORT ViewerOpenGL : public Viewer
{
private:
  /**
   * Version of current OpenGL renderer
   */
  float d_glVersion;

protected:

  /**
   * Stack of transformation matrices. (We don't use OpenGL stack
   * here because it is too limited in depth.)
   */
  stack<Matrix4x4>   transformStack;

public:

  enum { MAX_LIGHTS = 8 };

  ViewerOpenGL(VrmlScene *);
//  ViewerOpenGL(ViewerOpenGL const &);
  virtual ~ViewerOpenGL();

  void setScene(VrmlScene* scene);

//  ViewerOpenGL & operator=(ViewerOpenGL const &);

  // get rid of this asap.
  virtual void getBillboardTransformMatrix(float [4][4],
                                           const float axisOfRotation[3]);

  virtual RenderMode getRenderMode();
  virtual double getFrameRate();

  //
  virtual void resetUserNavigation();
  virtual void getUserNavigation(double M[4][4]);

  // Scope dirlights, open/close display lists
  virtual Object beginObject(const char *name, bool retain);
  virtual void endObject();

  // Insert objects into the display list
  virtual Object insertBackground(size_t nGroundAngles = 0,
                                  const float * groundAngle = 0,
                                  const float * groundColor = 0,
                                  size_t nSkyAngles = 0,
                                  const float * skyAngle = 0,
                                  const float * skyColor = 0,
                                  int* whc = 0,
                                  unsigned char ** pixels = 0,
                                  Object texturesDisplayList = 0);


  virtual Object insertBox(float x, float y, float z);
  virtual Object insertCone(float h, float r, bool bottom, bool side);
  virtual Object insertCylinder(float h, float r, bool, bool, bool);

  virtual Object insertElevationGrid(unsigned int mask,
                                     size_t nx,
                                     size_t nz,
                                     const float * height,
                                     float dx,
                                     float dz,
                                     const float * tc,
                                     const float * normals,
                                     const float * colors);

  virtual Object insertExtrusion(unsigned int,
                                 size_t nOrientation,
                                 const float * orientation,
                                 size_t nScale,
                                 const float * scale,
                                 size_t nCrossSection,
                                 const float * crossSection,
                                 size_t nSpine,
                                 const float * spine,
                                 float creaseAngle);

  virtual Object insertLineSet(size_t, const float *, size_t, const long *,
                               bool colorPerVertex,
                               const float * color,
                               size_t nColorIndex,
                               const long * colorIndex);

  virtual Object insertPointSet(size_t npts, const float * points, const float * colors);
  virtual Object insertShell(unsigned int mask,
                             size_t npoints, const float * points,
                             size_t nfaces, const long * faces,
                             const float * tc,
                             size_t ntci, const long * tci,
                             const float * normal,
                             size_t nni, const long * ni,
                             const float * color,
                             size_t nci, const long * ci);

  virtual Object insertSphere(float radius);

  virtual Object insertText(FontFamily family,
                            FontStyle style,
                            int* justifications,
                            float size,
                            float spacing,
                            int n,
                            char const * const *s);

  // Lights
  virtual Object insertDirLight(float a, float i, const float rgb[], const float xyz[]);

  virtual Object insertPointLight( float ambientIntensity,
                                   const float attenuation[],
                                   const float color[],
                                   float intensity,
                                   const float location[],
                                   float radius );

  virtual Object insertSpotLight( float /*ambientIntensity*/ ,
                                  const float /*attenuation*/ [],
                                  float /*beamWidth*/ ,
                                  const float /*color*/ [],
                                  float /*cutOffAngle*/ ,
                                  const float /*direction*/ [],
                                  float /*intensity*/ ,
                                  const float /*location*/ [],
                                  float /*radius*/ );


  // Lightweight copy
  virtual Object insertReference(Object existingObject);

  // Remove an object from the display list
  virtual void removeObject(Object key);

  virtual void enableLighting(bool);

  // Set attributes
  virtual void setColor(float r, float g, float b, float a = 1.0);

  virtual void setFog(const float * /*color*/,
                      float   /*visibilityRange*/,
                      const char * /*fogType*/);

  virtual void setMaterial(float, const float[], const float[], float, const float[], float);

  virtual void setMaterialMode( int nTexComponents, bool geometryColor );

  virtual void setSensitive(void *object);

public:
  /**
   * Set maximum texture size for this viewer.
   */
  virtual void setMaxTextureSize(int texsize);

  virtual TextureObject insertTexture(size_t w, size_t h, size_t nc,
                                      bool repeat_s,
                                      bool repeat_t,
                                      const unsigned char *pixels,
                                      bool retainHint = false);

  // Reference/remove a texture object
  virtual void insertTextureReference(TextureObject, int);
  virtual void removeTextureObject(TextureObject);

  virtual void setTextureTransform(const float center[2],
                                   float rotation,
                                   const float scale[2],
                                   const float translation[2]);

  virtual void setTransform(const float center[3],
                            const float rotation[4],
                            const float scale[3],
                            const float scaleOrientation[4],
                            const float translation[3]);

  virtual void unsetTransform(const float center[3],
                              const float rotation[4],
                              const float scale[3],
                              const float scaleOrientation[4],
                              const float translation[3]);

  virtual void setBillboardTransform(const float axisOfRotation[3]);
  virtual void unsetBillboardTransform(const float axisOfRotation[3]);

  virtual void setViewpoint(const float *position,
                            float *orientation,
                            float fieldOfView,
                            float avatarSize,
                            float visLimit);

  // The viewer knows the current viewpoint
  virtual void transformPoints(int nPoints, float *points);

  virtual void drawBSphere(const VrmlBSphere& bs, int flag);



  //
  // Viewer callbacks (not for public consumption)

  // Update the model.
  void update();

  // invalidate GL initialization after context is lost
  virtual void invalidateGL();

  // Redraw the screen.
  virtual void redraw();

  // Draws the Head-up-display (the default implementation does nothing).
  // This is called after the scene is drawn so that the HUD is always
  // "above" the scene.
  virtual void drawHud();

/* resize API */
private:
  int width;
  int height;
  int newwidth;
  int newheight;

public:
  /**
   * Gets the size of the viewport.
   * This is the rendered size of the scene and does not include
   * any bottom reserved area.
   * @param w Reference to variable to receive the rendered scene width.
   * @param h Reference to variable to receive the rendered scene height.
   */
  void getViewport(int& w, int& h) { w = width; h = height-d_bottomReserved; }

  /**
   * Gets the size of the rendering area.
   * The size includes any bottom reserved area.
   * @param w Reference to variable to receive the rendering area width.
   * @param h Reference to variable to receive the rendering area height.
   */
  void getSize(int& w, int& h) { w = width; h = height; }

  /**
   * Resize the rendering area of the graphics subsystem.
   * @param w New width.
   * @param h New height.
   */
  virtual void resizeCanvas(int w, int h);


  // user interaction
#ifdef macintosh /* in order to be able to send 'keys' from menus */
  virtual void handleKey(int);
#endif

  // Event types
  typedef enum {
    EVENT_KEY_DOWN,
    EVENT_MOUSE_MOVE,
    EVENT_MOUSE_CLICK,
    EVENT_MOUSE_DRAG,
    EVENT_MOUSE_RELEASE,
    EVENT_MOUSE_TIME_GENERATED
  } EventType;

  enum {
    KEY_HOME = 1,
    KEY_LEFT,
    KEY_UP,
    KEY_RIGHT,
    KEY_DOWN,
    KEY_PAGE_UP,
    KEY_PAGE_DOWN,
    KEY_ESCAPE
  };

  typedef enum {
    EVENT_MODIFIER_UNKNOWN,
    EVENT_SHIFT_DOWN
  } EventModifier;

  enum MouseButton {
    MOUSE_BUTTON_LEFT = 1,
    MOUSE_BUTTON_MIDDLE,
    MOUSE_BUTTON_RIGHT
  };

  struct EventInfo
  {
    EventType event;
    int what;                   // key or button number
    int x, y;
    EventModifier modifier;

    // constructor
    EventInfo(EventType _event, int _what, int _x, int _y, EventModifier _modifier) :
      event(_event),
      what(_what),
      x(_x),
      y(_y),
      modifier(_modifier)
    { }
  };

  void input( EventInfo *);

  // Structs used in tesselator callbacks
  typedef struct IndexData {
    const float * v;                    // data values
    size_t ni; const long * i;                  // number of indices, index pointer
  } IndexData;

  struct ShellData {
    unsigned int mask;
    const float * points;
    size_t nfaces; const long * faces;          // face list
    IndexData texCoord;         // texture coordinates and indices
    IndexData normal;           // normals and indices
    IndexData color;            // colors and indices
    int *texAxes;
    float *texParams;
    size_t nf, i;
  };

  // Sun compiler requires this to be public
  typedef enum { LIGHT_UNUSED, LIGHT_DIRECTIONAL, LIGHT_POSITIONAL } LightType;

  typedef enum {
    // the following two use wxWidgets stock objects and must be first
    CURSOR_ARROW = 0,
    CURSOR_CROSSHAIR,

    CURSOR_INFO,
    CURSOR_CYCLE,
    CURSOR_UP_DOWN,

    /** Cursor used when the mouse touches a sensor. */
    CURSOR_SENSOR,

    CURSOR_FIRST_UNUSED
  } CursorStyle;

protected:
  // Window system specific methods
  virtual void wsSetCursor( CursorStyle c) = 0;
  virtual void wsSwapBuffers() = 0;

  // Initialize OpenGL state
  virtual void initialize();

  // Geometry insertion setup & cleanup methods
  void beginGeometry();
  void endGeometry();

  // Text rendering
  void text3(FontFamily family,
             FontStyle style,
             int *justify,
             float size,
             float spacing,
             int n,
             const char * const * s);

  // IndexedFaceSet helpers
  void insertShellConvex(ShellData *);
  void insertShellTess(ShellData *);

#ifndef macintosh
  virtual void handleKey(int);
#endif

  void handleButton(EventInfo*);
  void handleMouseDrag(int, int);

  void initializeMouseDrag(EventInfo* e);

  // time intervals for checking sensitivities
  static const double MOUSE_MOVE_CHECK_SENSITIVE_INTERVAL;
  static const double MOUSE_TIME_GENERATED_CHECK_SENSITIVE_INTERVAL;

  // GL status
  bool d_GLinitialized;
  bool d_blend;
  bool d_lit;
  bool d_texture;
  bool d_wireframe;

  int d_window;

  // reserved area in the bottom (for drawing a control panel/status bar)
  int d_bottomReserved;

  float d_background[3];

  // Groups used to be put in display lists, now these counters
  // trigger various initializations and clean ups
  int d_nObjects, d_nestedObjects;

  // Tessellation
  GLUtesselator * d_tess;
  void insertExtrusionCaps(unsigned int mask, size_t nSpine, const float * spine, const float * c,
                           size_t nCrossSection, const float * cs);

  // Check for pickable entity selection
  bool checkSensitive(int x, int y, EventType );

  /** check sensitive do not perform its routine */
  bool checkSensitiveDisabled;

  // Pickable entities
  int d_nSensitive;
  int d_activeSensitive;
  int d_overSensitive;

  static const size_t SELECTION_BUFFER_SIZE = 4096;

  size_t d_selectionBufferSize;

  std::vector<void*> d_sensitiveObject;

  bool d_selectMode;
  double d_selectZ;             // window Z coord of last selection

  // Lights
  typedef struct lightinfo {
    LightType lightType;
    int nestingLevel;
    float radiusSquared;
    float location[3];
  } LightInfo;

  LightInfo d_lightInfo[MAX_LIGHTS];

  // current mouse coordinates
  long  d_mouseXCoord;
  long  d_mouseYCoord;

  // View manipulation

  /** Initial mouse drag position (X coordinate) */
  int d_beginx;

  /** Initial mouse drag position (Y coordinate). */
  int d_beginy;

  /** Previous mouse drag position (X coordinate). */
  int d_prevx;

  /** Previous mouse drag position (Y coordinate). */
  int d_prevy;

  /** True if mouse drag is currently in progress. */
  bool d_dragging;

  /** Zones for mouse navigation. Effect of mouse click/drag
    * can be different in each zone.
    */
  enum MouseZone
  {
      /** Central zone (middle 1/3 of screen). */
      MOUSE_ZONE_CENTRAL = 1,

      /** Border zone (around the central zone). */
      MOUSE_ZONE_BORDER = 2,

      /** Left edge (left 1/12 of screen). */
      MOUSE_ZONE_LEFT_EDGE = 3,

      /** Right edge (right 1/12 of screen). */
      MOUSE_ZONE_RIGHT_EDGE = 4
  };

  /** Zone where the current mouse drag started
    * (one of MouseZone constants), or 0 if no drag is active.
    */
  int d_dragZone;

  float d_scale;
  float d_translatex, d_translatey, d_translatez;

  // quaternion representations of last, current rotation
  float d_lastquat[4], d_curquat[4];
  bool d_rotationChanged;

  // matrix representation of the user navigation
  float d_rotationMatrix[4][4];

  // current mouse navigation modes
  bool d_rotating, d_scaling, d_translating;

  bool d_drawBSpheres;
  bool d_cull;

  /**
   * True whether antialiasing should be used, false otherwise.
   * Currently, only textures are antialiased (by linear filtering),
   * we can add line and pixel antialiasing later.
   */
  bool d_antialias;

  bool d_reportFPS;
  double d_renderTime;
  double d_renderTime1;

  /** Motion step size (default is 1). */
  double d_motionStep;

  /** Rotation step size (default is 1). */
  double d_rotationStep;

  double d_zNear;
  double d_zFar;

  /** Maximum hardware texture size */
  int d_maxTextureSizeHW;

public:
  /**
   * depth values can be guessed either with opengl select mode
   * or by browsing vrml nodes tree
   */
  enum PlaneClipMethod
  {
    /** use opengl select mode */
    PLANECLIP_OPENGL,
    /** use vrml nodes bounding volume */
    PLANECLIP_BVOLUME
  };

private:
  /** contains current zfar and znear guessing method */
  PlaneClipMethod planeClipMethod;

public:
  void setPlaneClipMethod(PlaneClipMethod agm)
  {
    planeClipMethod = agm;
  }

  PlaneClipMethod getPlaneClipMethod() const
  {
    return planeClipMethod;
  }

  void disableCheckSensitive(bool disabled) { checkSensitiveDisabled = disabled; }


private:

  /** When were sensors checked last time
    */
  double d_sensorCheckTime;

public:

  double getSensorCheckTime() { return d_sensorCheckTime; }

  /**
   * Sets mouse current coordinates
   */
  void setMouseCurrentCoordinates(long mouseXCoord, long mouseYCoord)
  {
    d_mouseXCoord = mouseXCoord;
    d_mouseYCoord = mouseYCoord;
  }

  /**
   * Gets mouse current X coordinate
   */
  long getMouseCurrentXCoord()
  {
    return d_mouseXCoord;
  }

  /**
   * Gets mouse current Y coordinate
   */
  long getMouseCurrentYCoord()
  {
    return d_mouseYCoord;
  }

  // Check for pickable entity selection under current mouse position
  void checkSensitiveForCurrentMousePosition();

  /** Gets the identifier of the currently active dragging zone.
    * @returns The identifier of the dragging zone currently active,
    * or 0 if no zone is active.
    */
  int getDragZone() const { return d_dragZone; };

  /** Gets the identifier of the active sensitive object in the scene.
    * @returns The identifier of the active sensitive object,
    * or 0 if no sensitive object is currently active.
    */
  bool getActiveSensitive() const { return d_activeSensitive!=0; };

  /** Makes a guess about how should the near and far projection plane
    * be set to render this view correctly.
    * <p>
    * The guess may fail if the view does not contain any objects,
    * or no objects are visible with the initial setting of near and far
    * plane. In that case, the method returns false and values
    * of near and far plane stay unchanged.
    * @param nearVal Pointer to variable to receive the near plane value.
    * @param farVal Pointer to variable to receive the far plane value.
    * @returns True if successful, false if failed.
    */
  bool guessNearFarPlaneWithSelMode(double* nearVal, double* farVal);

  /** Alternative way how to obtain znear and zfar values.
   *  <p>
   *  This method guess depth-values without opengl select mode.
   *  Instead of it browses tree of Vrml nodes
   */
  void guessNearFarPlaneWithTreeBrowsing(double* nearVal, double* farVal, double defaultNearVal, double defaultFarVal, bool zNearExplicit, bool zFarExplicit);

  static const char* getOpenGLErrorMessage(int glErrno);

  static bool checkOpenGLError(const char* where);

  /**
   * Projects a beam from the given view coordinates
   * in the camera direction and returns the point (in world coordinates)
   * where it touches the geometry.
   * @param x The X-coordinate in the view.
   * @param y The Y-coordinate in the view.
   * @return The touch point.
   */
  Point3 projectBeam(int x, int y);


  void resetOpenGLState();
};

#endif // VIEWEROPENGL_H
