/*****************************************************************************
******************************************************************************
*
*               VRML97 parser
*
*               $Revision: 1.1.4.4.2.13 $
*               $Date: 2011/08/08 12:08:51 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright 1998-2011 HUMUSOFT s.r.o.

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include "Vrml97Parser.h"

#include <memory>
#include <cstdlib>
#include <assert.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>

#include "Doc.h"
#include "VrmlNamespace.h"
#include "VrmlNodeType.h"
#include "VrmlNode.h"
#include "VrmlNodeScript.h"
#include "VrmlField.h"



/** array of names of token types, indexed by token type values. */
static const char* VRMLTokenTypeNames[] =
{
  "end of file", "comment", "string", "number",
  "identifier", "keyword \"DEF\"", "keyword \"EXTERNPROTO\"",
  "keyword \"FALSE\"", "keyword \"IS\"", "keyword \"NULL\"",
  "keyword \"PROTO\"", "keyword \"ROUTE\"", "keyword \"TO\"",
  "keyword \"TRUE\"", "keyword \"USE\"", "keyword \"eventIn\"",
  "keyword \"eventOut\"", "keyword \"exposedField\"",
  "keyword \"field\"", "symbol '.'", "symbol '['",
  "symbol ']'", "symbol '{'", "symbol '}'", "symbol '('",
  "symbol ')'"
};

/** A table of VRML keywords. */
static const char* VRMLKeywords[] =
{
  "DEF", "EXTERNPROTO", "FALSE",
  "IS", "NULL", "PROTO", "ROUTE",
  "TO", "TRUE", "USE", "eventIn",
  "eventOut", "exposedField", "field"
};
static const int VRML_KEYWORD_COUNT = sizeof(VRMLKeywords)/sizeof(const char*);

/** A table of token type numbers assigned to VRML keywords. */
static int VRMLKeywordTokenTypes[] =
{
  VRML_KEYWORD_DEF, VRML_KEYWORD_EXTERNPROTO, VRML_KEYWORD_FALSE,
  VRML_KEYWORD_IS, VRML_KEYWORD_NULL, VRML_KEYWORD_PROTO, VRML_KEYWORD_ROUTE,
  VRML_KEYWORD_TO, VRML_KEYWORD_TRUE, VRML_KEYWORD_USE, VRML_KEYWORD_EVENT_IN,
  VRML_KEYWORD_EVENT_OUT, VRML_KEYWORD_EXPOSED_FIELD, VRML_KEYWORD_FIELD
};



/************************/
/*** helper functions ***/
/************************/

/** Creates a newly allocated (via 'new') string containing the same text
  * as the given string.
  */
static char* newString(const char* str)
{
  char* result;
  if (str == NULL)
  {
    result = new char[1];
    result[0] = '\0';
  }
  else
  {
    size_t len = strlen(str);
    result = new char[len+1];
    memcpy(result, str, len+1);
  }
  return result;
}



/*****************/
/*** VRMLToken ***/
/*****************/

std::string VRMLToken::desc()
{
  std::string retval(VRMLTokenTypeNames[type]);
  if (type == VRML_IDENTIFIER)
    retval.append(" \"").append(*this).append("\"");

  return(retval);
}



/************************/
/*** VrmlLoadingError ***/
/************************/

void VrmlLoadingError::creator::create(const char* filename,
                                       int line,
                                       int column,
                                       const char* fmt,
                                       va_list ap)
{
  // print filename if any
  if (filename && filename[0]!='\0')
    append(filename);

  // print line number if any
  char buf[256];
  if (line>0)
  {
    snprintf(buf, sizeof(buf), empty() ? "line %d" : ", line %d", line);
    append(buf);
  }

  // print colon if filename or line already printed
  if (!empty()) append(": ");

  // print the actual error message
  vsnprintf(buf, sizeof(buf), fmt, ap);
  append(buf);
}



/*****************/
/*** VRMLValue ***/
/*****************/

VRMLValue::~VRMLValue()
{
  for (std::vector<char*>::iterator it=strings.begin(); it != strings.end(); ++it)
    delete[] *it;
}




/********************/
/*** Vrml97Parser ***/
/********************/



Vrml97Parser::Vrml97Parser(istream& in, const char* filename)
  : input(in),
  columnNumber(1),
  lineNumber(1),
  tokenPrefetched(false),
  fileName(filename ? filename : ""),
  previousCommentTarget(0)
{ }


Vrml97Parser::~Vrml97Parser()
{ }

void Vrml97Parser::doError(const char* msg, ...)
{
  va_list ap;
  va_start(ap, msg);
  VrmlLoadingError err(VrmlLoadingError::UNKNOWN, fileName.c_str(), lineNumber, columnNumber, msg, ap);
  va_end(ap);
  throw err;
}

void Vrml97Parser::doError(VrmlLoadingError::errorType etype, char* msg, ...)
{
  va_list ap;
  va_start(ap, msg);
  VrmlLoadingError err(etype, fileName.c_str(), lineNumber, columnNumber, msg, ap);
  va_end(ap);
  throw err;
}



int Vrml97Parser::getNextCharacter()
{
  // character prefetched
  if (!prefetchedCharacters.empty())
  {
    int c = prefetchedCharacters.top();
    prefetchedCharacters.pop();
    return(c);
  }

  // read character from input stream
  else
  {
    int c = input.get();
    if (c < 0 || c > 255) return -1;
    columnNumber++;

    // be robust to CR/LF/CRLF - always return LF
    if (c == '\r')
    {
      c = '\n';
      int c2 = input.get();  // discard the LF from CRLF pair
      if (c2 != '\n')
        pushBackCharacter(c2);
    }
    if (c == '\n') { lineNumber++; columnNumber = 1; }
    return c;
  }
}



std::string Vrml97Parser::getNextLine()
{
  std::string result;
  int c;
  while ((c=getNextCharacter())>=0 && c!='\n') result.push_back(c);
  return result;
}



int Vrml97Parser::getNextCharacterSkipWhitespace()
{
  int c;
  while ((c=getNextCharacter())==' ' || c==',' || c=='\t' || c=='\r');
  return(c);
}



bool Vrml97Parser::isValidIdFirstChar(char c)
{
  return (!(
  ((c >= '0') && (c <= '9'))   // cannot start with number
    || (c == '+')
    || (c == '-')
    || !isValidIdRestChar(c)
  ));
}



bool Vrml97Parser::isValidIdRestChar(char c)
{
  return (!(
    (c <= ' ')                 // control characters and space
    || (c == '"') || (c == '#')
    || (c == '\'') || (c == ',') || (c == '.')
    || (c == '[') || (c == '\\') || (c == ']')
    || (c == '{') || (c == '}')
    || (c == 0x7f)           // Del character
  ));
}

bool Vrml97Parser::isValidId(const std::string& name)
{
  // we consider empty name to be ok
  if (name.empty())
    return(true);

  // treat first character
  if (!Vrml97Parser::isValidIdFirstChar(name.at(0)))
    return(false);

  // treat rest characters
  for (size_t i=1; i<name.size(); ++i)
  {
    if (!Vrml97Parser::isValidIdRestChar(name.at(i)))
      return(false);
  }

  // name must not be a VRML keyword
  for (size_t i=0; i<VRML_KEYWORD_COUNT; ++i)
  {
    if (strcmp(name.c_str(), VRMLKeywords[i]) == 0)
      return(false);
  }

  // we reached end of the function, name is valid
  return(true);
}



bool Vrml97Parser::isHexDigit(char c)
{
  return (isdigit(c) || (c >= 'A' && c <= 'F')
    || (c >= 'a' && c <= 'f'));
}



void Vrml97Parser::pushbackToken(const VRMLToken* token)
{
  prefetchedToken = *token;
  tokenPrefetched = true;
}



VRMLToken* Vrml97Parser::getNextToken(VRMLToken* token)
{
  // if a token is prefetched, return that token
  if (tokenPrefetched)
  {
    tokenPrefetched = false;
    *token = prefetchedToken;
    return(token);
  }

  // read a character, skipping whitespace, handling comments, EOL and EOF
  int c;
  while (true)
  {
    c = getNextCharacterSkipWhitespace();
    if (c=='#')       // comment
    {
      // add new comment to comments array
      comments.resize(comments.size()+1);
      std::string& comment(comments.back());

      // comment can span multiple lines if the first non-whitespace
      // char of the next line is '#'
      while ((c = getNextCharacter()) >= 0)  // test for EOF
      {
        comment.push_back(c);
        if (c == '\n')
        {
          c = getNextCharacterSkipWhitespace();
          if (c == '#') continue;    // comment block continues
          pushBackCharacter(c);      // done with comment block
          break;
        }
      }
      continue;
    }

    else if (c=='\n') // newline
    {
      // insert an empty string to the end of comments array if not there yet
      if (comments.empty() || !comments.back().empty())
        comments.resize(comments.size()+1);
      continue;
    }

    else if (c==-1)   // end of file
      return(&token->clear(VRML_END_OF_FILE));

    else              // pass all other characters below
      break;
  }

  // is it a keyword/identifier?
  if (isValidIdFirstChar(c))
  {
    token->clear(VRML_IDENTIFIER);
    while (c > 0 && isValidIdRestChar(c))
      { token->push_back(c); c = getNextCharacter(); }
    if (c > 0) pushBackCharacter(c);

    // is it a keyword?
    const char* word = token->c_str();
    int i;
    for (i=0; i<VRML_KEYWORD_COUNT; i++)
    {
      if (strcmp(word, VRMLKeywords[i]) == 0)
      {
        token->setType(VRMLKeywordTokenTypes[i]);
        return(token);
      }
    }
    return(token);
  }

  // is it a number?
  if (c == '.' || c == '+' || c == '-' || isdigit(c))
  {
    token->clear(VRML_NUMBER);

    // handle numbers beginning with dot (e.g. .97), distinguish them
    // from a single dot (which is a symbol).
    if (c == '.')
    {
      c = getNextCharacter();
      if (isdigit(c))
      {
        token->push_back('.');
      }
      else
      {
        if (c > 0) pushBackCharacter(c);
        token->clear(VRML_SYMBOL_DOT).push_back('.');
        return(token);
      }
    }

    // read digits
    bool wasDot = false;           // was a dot seen?
    bool wasExp = false;           // was an exponent sign seen?
    bool justAfterExp = false;     // are we just after the exponent sign?
    bool firstChar = true;         // are we at the very first character?
    while (c > 0 && (isdigit(c) || c == '.' || c == 'e' || c == 'E'
      || c == '+' || c == '-'))
    {
      // is it hexadecimal (in form 0xXXXX)?
      if (c == '0')
      {
        c = getNextCharacter();
        if (c == 'x' || c == 'X')
        {
          token->push_back('0');
          token->push_back('x');
          c = getNextCharacter();
          while (c > 0 && isHexDigit(c))
            { token->push_back(c); c = getNextCharacter(); }
          if (c > 0) pushBackCharacter(c);
          return(token);
        }
        else
        {
          pushBackCharacter(c);
          c = '0';
        }
      }

      // process special cases
      if (c == '.')
      {
        // two dots in a number are illegal
        if (wasDot) doError("Unrecognized token \"%s%c\"", token->c_str(), c);
        wasDot = true;
      }
      else if (c == 'e' || c == 'E')
      {
        // two exponent signs in a number are illegal
        if (wasExp) doError("Unrecognized token \"%s%c\"", token->c_str(), c);
        wasExp = true;
        justAfterExp = true;
      }
      else if (c == '-' || c == '+')
      {
        // a plus or minus sign anywhere else than on the beginning or after
        // the exponent character is already a part of the next token
        // (e.g. "12-3" [twenteen minus three]).
        if (!justAfterExp && !firstChar) break;
      }
      if (c != 'e' && c != 'E') justAfterExp = false;
      firstChar = false;

      token->push_back(c);
      c = getNextCharacter();
    }
    if (c>0) pushBackCharacter(c);
    return(token);
  }

  // is it a string?
  if (c == '"')
  {
    token->clear(VRML_STRING);
    c = getNextCharacter();

    int prevChar = '\0';

    // read until a '"' is found, but do not stop on escaped ones
    while (c != '"' || prevChar == '\\')
    {
      token->push_back(c);
      prevChar = c;
      c = getNextCharacter();
    }
    return(token);
  }

  // is it a symbol?
  if (c == '[' || c == ']' || c == '{' || c == '}')
  {
    switch (c)
    {
      case '[': token->clear(VRML_SYMBOL_LEFT_BRACKET); break;
      case ']': token->clear(VRML_SYMBOL_RIGHT_BRACKET); break;
      case '{': token->clear(VRML_SYMBOL_LEFT_BRACE); break;
      case '}': token->clear(VRML_SYMBOL_RIGHT_BRACE); break;
    }
    token->push_back(c);
    return(token);
  }

  // something unexpected was encountered
  doError("Unrecognized token \"%s\"", token->c_str());
  return(token);
}



bool Vrml97Parser::getNextTokenMatch(VRMLToken* token, int _type)
{
  getNextToken(token);
  return token->match(_type);
}



void Vrml97Parser::getNextTokenExpectType(VRMLToken* token, int type)
{
  getNextToken(token);
  if (token->getType() != type)
  {
    doError("Unexpected %s (should be %s)",
            token->desc().c_str(),
            VRMLTokenTypeNames[type]);
  }
}



void Vrml97Parser::getNextTokenExpectIdentifier(VRMLToken* token, const char* whatWasExpected)
{
  getNextToken(token);
  if (token->getType() != VRML_IDENTIFIER)
  {
    doError("Unexpected %s (should be %s)",
            token->desc().c_str(),
            whatWasExpected);
  }
}



void Vrml97Parser::loadRoute(VrmlNamespace* nmspc)
{
  VRMLToken token, srcNodeToken, srcFieldToken, destNodeToken, destFieldToken;

  // format: ROUTE sourcenode.sourcefield TO destnode.destfield
  getNextTokenExpectIdentifier(&srcNodeToken, "node name");
  getNextTokenExpectType(&token, VRML_SYMBOL_DOT);
  getNextTokenExpectIdentifier(&srcFieldToken, "field name");
  getNextTokenExpectType(&token, VRML_KEYWORD_TO);
  getNextTokenExpectIdentifier(&destNodeToken, "node name");
  getNextTokenExpectType(&token, VRML_SYMBOL_DOT);
  getNextTokenExpectIdentifier(&destFieldToken, "field name");

  // find the source and destination node
  VrmlNode* srcNode = nmspc->findNode(srcNodeToken.c_str());
  VrmlNode* destNode = nmspc->findNode(destNodeToken.c_str());
  if (srcNode == NULL)
  {
    doError("Can't find node \"%.64s\" referenced by ROUTE",
            srcNodeToken.c_str());
  }
  if (destNode == NULL)
  {
    doError("Can't find node \"%.64s\" referenced by ROUTE",
            destNodeToken.c_str());
  }

  // create the route and assign any comment to it
  assignComments(nmspc->addRoute(srcNode, srcFieldToken.c_str(), destNode, destFieldToken.c_str()));
}



void Vrml97Parser::loadTopLevel(VrmlMFNode* parent, VrmlNamespace* nmspc)
{
  // validate the header
  const char vrml97sig[] = "#VRML V2.0 utf8 ";
  const size_t vrml97siglen = sizeof(vrml97sig)-1;
  const char vrml1sig[] = "#VRML V1";
  std::string header(getNextLine());
  if (header.compare(0, vrml97siglen, vrml97sig, 0,
                     header.size()<vrml97siglen ? vrml97siglen-1 : vrml97siglen) != 0)
  {
    if (header.compare(0, sizeof(vrml1sig)-1, vrml1sig) == 0)
      doError("VRML 1.0 format is not supported. Please use V-Realm Builder or other conversion utility to convert into VRML97.");
    else
      doError("Invalid or unsupported VRML file format: %s", header.c_str());
  }

  // the remaining characters on the line are appended to the world comment
  if (header.size()>vrml97siglen)
    parent->appendComment(header.substr(vrml97siglen));
  previousCommentTarget = parent;

  // the token loop
  VRMLToken token;
  while (!getNextToken(&token)->match(VRML_END_OF_FILE))
  {
    if (token.match(VRML_KEYWORD_DEF))
      parent->addNode(loadNamedNode(nmspc, NULL));
    else if (token.match(VRML_IDENTIFIER))
      parent->addNode(loadUnnamedNode(nmspc, token.c_str(), NULL));
    else if (token.match(VRML_KEYWORD_ROUTE))
      loadRoute(nmspc);
    else if (token.match(VRML_KEYWORD_USE))
      parent->addNode(loadUsedNode(nmspc));
    else if (token.match(VRML_KEYWORD_PROTO))
      loadProto(nmspc);
    else if (token.match(VRML_KEYWORD_EXTERNPROTO))
      loadExternProto(nmspc);
    else
      doError("Unexpected %s", token.desc().c_str());
  }

  // assign the final comments to the last object
  assignComments(*previousCommentTarget);
}


void Vrml97Parser::assignComments(VrmlObject& obj)
{
  // the first comment block is treated as trailing comment of the previous object,
  // so append it there
  std::vector<std::string>::const_iterator it = comments.begin();
  if (previousCommentTarget!=0 && it!=comments.end())
    previousCommentTarget->appendComment(*(it++));

  // append all the remaining comment blocks to current object
  while (it!=comments.end())
    obj.appendComment(*(it++));
  comments.clear();

  // set the current object as previous comment target
  previousCommentTarget = &obj;
}


VrmlNode* Vrml97Parser::loadNamedNode(VrmlNamespace* nmspc,
  VrmlNodeType* parentProto)
{
  VRMLToken token;

  // get node name
  VRMLToken nodeNameToken;
  getNextTokenExpectIdentifier(&nodeNameToken, "node name");

  // get node type
  VRMLToken nodeTypeToken;
  getNextTokenExpectIdentifier(&nodeTypeToken, "node type");
  const VrmlNodeType* nodeType = nmspc->findType(nodeTypeToken.c_str());
  if (nodeType == NULL)
    doError("Unknown node type \"%.64s\"", nodeTypeToken.c_str());

  // check for opening brace
  getNextTokenExpectType(&token, VRML_SYMBOL_LEFT_BRACE);

  // create the node
  VrmlNode* node = nodeType->newNode();

  // add it to the namespace (check for double declaration first)
  if (nmspc->findNode(nodeNameToken.c_str()) == NULL)
    node->setName(nodeNameToken.c_str(), nmspc);
  else
  {
    // the name is already used, inform about it
    node->setName(NULL, nmspc);
    theSystem->warn(System::SEVERITY_LOW, System::WARNINGTYPE_UNKNOWN, "Another node named \"%s\" already exists", nodeNameToken.c_str());
  }

  // assign comments to node
  assignComments(*node);

  // load its contents
  loadNodeContents(nmspc, node, parentProto);

  // set the parent prototype pointer (if any)
  node->setEncapsulatingPrototype(parentProto);

  // if it is one of special-interest nodes, add it to appropriate list
  if (nodeTypeToken == "Viewpoint")
    viewpoints.push_back((VrmlNodeViewpoint*) node);
  else if (nodeTypeToken == "Background")
    backgrounds.push_back((VrmlNodeBackground*) node);
  else if (nodeTypeToken == "NavigationInfo")
    navinfos.push_back((VrmlNodeNavigationInfo*) node);
   else if (nodeTypeToken == "Fog")
    fogs.push_back((VrmlNodeFog*) node);

  return node;
}



VrmlNode* Vrml97Parser::loadUnnamedNode(VrmlNamespace* nmspc,
  const char* nodeTypeName, VrmlNodeType* parentProto)
{
  VRMLToken token;
  VrmlNode* node;

  const VrmlNodeType* nodeType = nmspc->findType(nodeTypeName);
  if (nodeType == NULL)
    doError("Unknown node type \"%.64s\"", nodeTypeName);
  getNextTokenExpectType(&token, VRML_SYMBOL_LEFT_BRACE);
  node = nodeType->newNode();
  node->setName(NULL, nmspc);

  // assign comments to node
  assignComments(*node);

  // load its contents
  loadNodeContents(nmspc, node, parentProto);

  // set the parent prototype pointer (if any)
  node->setEncapsulatingPrototype(parentProto);

  // if it is one of special-interest nodes, add it to appropriate list
  if (strcmp(nodeTypeName, "Viewpoint") == 0)
    viewpoints.push_back((VrmlNodeViewpoint*) node);
  else if (strcmp(nodeTypeName, "Background") == 0)
    backgrounds.push_back((VrmlNodeBackground*) node);
  else if (strcmp(nodeTypeName, "NavigationInfo") == 0)
    navinfos.push_back((VrmlNodeNavigationInfo*) node);
  else if (strcmp(nodeTypeName, "Fog") == 0)
    fogs.push_back((VrmlNodeFog*) node);

  return node;
}



VrmlNode* Vrml97Parser::loadUsedNode(VrmlNamespace* nmspc)
{
  VRMLToken token;
  getNextTokenExpectIdentifier(&token, "node name");
  VrmlNode* usedNode = nmspc->findNode(token.c_str());
  if (usedNode == NULL)
    doError("Can't find node \"%.64s\" referenced by USE", token.c_str());
  return usedNode;
}



void Vrml97Parser::loadValue(VrmlNamespace* nmspc, VrmlNodeType* parentProto,
  VRMLValue* value)
{
  VRMLToken token;

  // initialize the value
  value->strings.clear();
  value->numbers.clear();
  value->nodes.clear();
  value->hasStrings = value->hasNumbers = value->hasBoolean
    = value->hasNodes = value->hasEmptyVec = false;

  getNextToken(&token);

  // is it a string?
  if (token.getType() == VRML_STRING)
  {
    value->strings.push_back(newString(token.c_str()));
    value->hasStrings = true;
  }

  // is it a number, or an unbracketed array of numbers?
  else if (token.match(VRML_NUMBER))
  {
    // load numbers up to first non-number
    while (token.match(VRML_NUMBER))
    {
      char* endptr;
      value->numbers.push_back(strtod(token.c_str(), &endptr));
      getNextToken(&token);
    }
    value->hasNumbers = true;

    // the last token is not part of the value
    pushbackToken(&token);
  }

  // is it a boolean value?
  else if (token.match(VRML_KEYWORD_TRUE))
  {
    value->numbers.push_back(1);
    value->hasBoolean = true;
  }
  else if (token.match(VRML_KEYWORD_FALSE))
  {
    value->numbers.push_back(0);
    value->hasBoolean = true;
  }

  // is it a vector?
  else if (token.match(VRML_SYMBOL_LEFT_BRACKET))
  {
    getNextToken(&token);

    // is it a vector of strings?
    if (token.match(VRML_STRING))
    {
      // load strings up to the closing bracket
      while (true)
      {
        value->strings.push_back(newString(token.c_str()));
        getNextToken(&token);
        if (token.match(VRML_SYMBOL_RIGHT_BRACKET)) break;
        else if (!token.match(VRML_STRING))
          doError("Unexpected %s in a string list", token.desc().c_str());
      }
      value->hasStrings = true;
    }

    // is it a vector of numbers?
    else if (token.match(VRML_NUMBER))
    {
      // load numbers up to the closing bracket
      while (true)
      {
        char* endptr;
        value->numbers.push_back(strtod(token.c_str(), &endptr));
        getNextToken(&token);
        if (token.match(VRML_SYMBOL_RIGHT_BRACKET)) break;
        else if (!token.match(VRML_NUMBER))
          doError("Unexpected %s in a number list", token.desc().c_str());
      }
      value->hasNumbers = true;
    }

    // is it a vector of nodes?
    else if (token.match(VRML_KEYWORD_DEF, VRML_KEYWORD_USE,
      VRML_IDENTIFIER))
    {
      value->hasNodes = true;

      // load nodes up to the closing bracket
      while(true)
      {
        if (token.match(VRML_KEYWORD_DEF))
          value->nodes.push_back(loadNamedNode(nmspc, parentProto));
        else if (token.match(VRML_IDENTIFIER))
          value->nodes.push_back(loadUnnamedNode(nmspc, token.c_str(),
            parentProto));
        else if (token.match(VRML_KEYWORD_USE))
          value->nodes.push_back(loadUsedNode(nmspc));
        else if (token.match(VRML_KEYWORD_NULL))
          value->nodes.push_back(NULL);
        else if (token.match(VRML_SYMBOL_RIGHT_BRACKET)) break;
        else
          doError("Unexpected %s in a node list", token.desc().c_str());

        getNextToken(&token);
      }
    }

    // is it an empty vector?
    else if (token.match(VRML_SYMBOL_RIGHT_BRACKET))
      value->hasEmptyVec = true;

    else
      doError("Unexpected %s after '['", token.desc().c_str());
  }

  // is the value a named node?
  else if (token.match(VRML_KEYWORD_DEF))
  {
    value->nodes.push_back(loadNamedNode(nmspc, parentProto));
    value->hasNodes = true;
  }

  // is the value a NULL?
  else if (token.match(VRML_KEYWORD_NULL))
  {
    value->nodes.push_back(NULL);
    value->hasNodes = true;
  }

  // is the value an unnamed node?
  else if (token.match(VRML_IDENTIFIER))
  {
    value->nodes.push_back(loadUnnamedNode(nmspc, token.c_str(), parentProto));
    value->hasNodes = true;
  }

  // is it a node reference?
  else if (token.match(VRML_KEYWORD_USE))
  {
    value->nodes.push_back(loadUsedNode(nmspc));
    value->hasNodes = true;
  }

  else
    doError("Unexpected %s (should be a field value)", token.desc().c_str());
}



VrmlField *Vrml97Parser::convertValue(VRMLValue* value,
  VrmlField::VrmlFieldType fieldType)
{
  VrmlField* f = NULL;
  if (value->hasBoolean)
  {
    f = VrmlField::convert(fieldType, value->numbers.size(),
      value->numbers.empty() ? NULL : &value->numbers.front());
    if (!f)
    {
      doError("Invalid value (%d x boolean) for a %s field",
              value->numbers.size(),
              VrmlField::getTypeName(fieldType));
    }
  }
  else if (value->hasNumbers)
  {
    f = VrmlField::convert(fieldType, value->numbers.size(),
      value->numbers.empty() ? NULL : &value->numbers.front());
    if (!f)
    {
      doError("Invalid value (%d x number) for a %s field",
              value->numbers.size(),
              VrmlField::getTypeName(fieldType));
    }
  }
  else if (value->hasStrings)
  {
    f = VrmlField::convert(fieldType, value->strings.size(),
      value->strings.empty() ? NULL : (const char**) &value->strings.front());
    if (!f)
    {
      doError("Invalid value (%d x string) for a %s field",
              value->strings.size(),
              VrmlField::getTypeName(fieldType));
    }
  }
  else if (value->hasNodes)
  {
    f = VrmlField::convert(fieldType, value->nodes.size(),
      value->nodes.empty() ? NULL : &value->nodes.front());
    if (!f)
    {
      doError("Invalid value (%d x node) for a %s field",
              value->nodes.size(),
              VrmlField::getTypeName(fieldType));
    }
  }
  else if (value->hasEmptyVec)
  {
    f = VrmlField::convertEmptyVector(fieldType);
    if (!f)
    {
      doError("Invalid value (empty vector) for a %s field",
              VrmlField::getTypeName(fieldType));
    }
  }
  return f;
}



void Vrml97Parser::loadFieldValue(VrmlNamespace* nmspc, VrmlNode* node,
  const char* fieldName, VrmlNodeType* parentProto)
{
  VRMLToken token;
  VRMLValue value;

  loadValue(nmspc, parentProto, &value);

  // get field type
  VrmlField::VrmlFieldType fieldType = node->hasField(fieldName);
  if (fieldType == VrmlField::NO_FIELD)
  {
    doError("Unknown field \"%.64s\" for a %.64s node",
            fieldName,
            node->nodeType().getName());
  }
  else
  {
    VrmlField* f = convertValue(&value, fieldType);
    if (f) node->setField(fieldName, *f);
    delete f;
  }

  // assign comments to the field
  assignComments(*(node->getFieldStripDecorations(fieldName)));
}



void Vrml97Parser::loadNodeContents(VrmlNamespace* nmspc, VrmlNode* node,
  VrmlNodeType* parentProto)
{
  VRMLToken token;

  // read node contents in a loop
  while (true)
  {
    getNextToken(&token);

    // is it a field (or event) name?
    if (token.match(VRML_IDENTIFIER))
    {
      VRMLToken fieldNameToken(token);

      // the following may be an IS keyword, or a field value
      getNextToken(&token);
      if (token.match(VRML_KEYWORD_IS))
      {
        // load the IS binding
        getNextTokenExpectIdentifier(&token, "interface name");
        if (parentProto != NULL)
        {
            // !! TODO !! We should check whether the binding type is the same
            // as the field type (bad things can happen if they don't match).

          // add the IS binding
          parentProto->addIS(token.c_str(), node, fieldNameToken.c_str());
          node->addISBinding(fieldNameToken.c_str(), token.c_str());

          // load default value from the PROTO declaration (if specified
          // - events never have default values)
          VrmlField* defValue = parentProto->fieldDefault(token.c_str());
          if (defValue) node->setField(fieldNameToken.c_str(), *defValue);
        }
        else
          doError("Unexpected keyword \"IS\"");
      }
      else
      {
        // load the field value
        pushbackToken(&token);
        loadFieldValue(nmspc, node, fieldNameToken.c_str(), parentProto);
      }
    }

    // is it an additional event declaration?
    else if (token.match(VRML_KEYWORD_EVENT_IN, VRML_KEYWORD_EVENT_OUT))
    {
      bool isEventIn = (token.getType() == VRML_KEYWORD_EVENT_IN);

      if (strcmp(node->nodeType().getName(), "Script") != 0)
        doError("Events can't be added to a non-Script node");

      // get the event type
      getNextTokenExpectIdentifier(&token, "event type");
      VrmlField::VrmlFieldType eventType = VrmlField::fieldType(token.c_str());
      if (eventType == VrmlField::NO_FIELD)
        doError("Unknown event type \"%.64s\"", token.desc().c_str());

      // get the event name
      VRMLToken eventNameToken;
      getNextTokenExpectIdentifier(&eventNameToken, "event name");

      // add the event to the node's event list
      if (isEventIn)
        ((VrmlNodeScript*) node)->addEventIn(eventNameToken.c_str(), eventType);
      else
        ((VrmlNodeScript*) node)->addEventOut(eventNameToken.c_str(), eventType);

      // there may be an IS clause
      if (getNextTokenMatch(&token, VRML_KEYWORD_IS))
      {
        // load the IS binding
        if (getNextTokenMatch(&token, VRML_IDENTIFIER))
        {
          if (parentProto != NULL)
          {
            // !! TODO !! We should check whether the binding type is the same
            // as the field type (bad things can happen if they don't match).

            parentProto->addIS(token.c_str(), node, eventNameToken.c_str());
            node->addISBinding(eventNameToken.c_str(), token.c_str());
          }
          else
            doError("Unexpected keyword \"IS\"");
        }
        else
          doError("Unexpected %s (should be a field or event name)",
                  token.desc().c_str());
      }
      else pushbackToken(&token);
    }

    // is it an additional field declaration?
    else if (token.match(VRML_KEYWORD_FIELD))
    {
      if (strcmp(node->nodeType().getName(), "Script") != 0)
        doError("Fields can't be added to a non-Script node");

      // get the field type
      getNextTokenExpectIdentifier(&token, "field type");
      VrmlField::VrmlFieldType fieldType = VrmlField::fieldType(token.c_str());
      if (fieldType == VrmlField::NO_FIELD)
        doError("Unknown field type \"%s\"", token.c_str());

      // get the field name
      VRMLToken fieldNameToken;
      getNextTokenExpectIdentifier(&fieldNameToken, "field name");

      // add the field to the node
      ((VrmlNodeScript*) node)->addField(fieldNameToken.c_str(), fieldType);

      // the following may be: 1. an IS keyword, or 2. a normal field value,
      // or 3. a normal value followed by an IS
      if (getNextTokenMatch(&token, VRML_KEYWORD_IS)) goto loadIS;
      else
      {
        // load the field value
        pushbackToken(&token);
        loadFieldValue(nmspc, node, fieldNameToken.c_str(), parentProto);
      }

      // look if there is an IS keyword following
      if (getNextTokenMatch(&token, VRML_KEYWORD_IS))
      {

loadIS:

        // load the IS binding
        VRMLToken ifaceNameToken;
        if (getNextTokenMatch(&ifaceNameToken, VRML_IDENTIFIER))
        {
          if (parentProto != NULL)
          {
            // !! TODO !! We should check whether the binding type is the same
            // as the field type (bad things can happen if they don't match).

            parentProto->addIS(ifaceNameToken.c_str(), node, fieldNameToken.c_str());
            node->addISBinding(fieldNameToken.c_str(), ifaceNameToken.c_str());
          }
          else
            doError("Unexpected keyword \"IS\"");
        }
        else
          doError("Unexpected %s (should be a field/event name)",
                  token.desc().c_str());
      }
      else pushbackToken(&token);
    }

    // is it a route?
    else if (token.match(VRML_KEYWORD_ROUTE)) loadRoute(nmspc);

    // is it a closing brace?
    else if (token.match(VRML_SYMBOL_RIGHT_BRACE)) return;

    else
      doError("Unexpected %s (should be a field name or '}')",
              token.desc().c_str());
  }
}



void Vrml97Parser::loadProto(VrmlNamespace* nmspc)
{
  VRMLToken token, protoNameToken;
  getNextTokenExpectIdentifier(&protoNameToken, "PROTO name");
  getNextTokenExpectType(&token, VRML_SYMBOL_LEFT_BRACKET);

  // new node type we are defining
  VrmlNodeType* nodeType = new VrmlNodeType(protoNameToken.c_str());
  if (nmspc->findType(protoNameToken.c_str()) != NULL)
    doError("Prototype \"%.64s\" already exists", protoNameToken.c_str());

  // read PROTO declaration
  while (true)
  {
    getNextToken(&token);

    // is it a closing bracket?
    if (token.match(VRML_SYMBOL_RIGHT_BRACKET)) break;

    // is it an event declaration?
    else if (token.match(VRML_KEYWORD_EVENT_IN, VRML_KEYWORD_EVENT_OUT))
    {
      VRMLToken eventNameToken, eventTypeToken;
      getNextTokenExpectIdentifier(&eventTypeToken, "event type");
      getNextTokenExpectIdentifier(&eventNameToken, "event name");
      VrmlField::VrmlFieldType eventType = VrmlField::fieldType(
        eventTypeToken.c_str());

      if (eventType == VrmlField::NO_FIELD)
        doError("Unknown event type \"%.64s\"", eventTypeToken.c_str());

      if (token.match(VRML_KEYWORD_EVENT_IN))
        nodeType->addEventIn(eventNameToken.c_str(), eventType);
      else
        nodeType->addEventOut(eventNameToken.c_str(), eventType);
    }

    // is it a field declaration?
    else if (token.match(VRML_KEYWORD_FIELD, VRML_KEYWORD_EXPOSED_FIELD))
    {
      VRMLToken fieldNameToken, fieldTypeToken;
      getNextTokenExpectIdentifier(&fieldTypeToken, "field type");
      getNextTokenExpectIdentifier(&fieldNameToken, "field name");
      VrmlField::VrmlFieldType fieldType = VrmlField::fieldType(
        fieldTypeToken.c_str());

      if (fieldType == VrmlField::NO_FIELD)
        doError("Unknown field type \"%.64s\"", fieldTypeToken.c_str());

      if (token.match(VRML_KEYWORD_FIELD))
        nodeType->addField(fieldNameToken.c_str(), fieldType);
      else
        nodeType->addExposedField(fieldNameToken.c_str(), fieldType);

      // get the default field value
      loadPROTODefaultFieldValue(nmspc, nodeType, fieldNameToken.c_str());
    }

    else
      doError("Unexpected %s (should be a field name or '}')",
              token.desc().c_str());
  }

  // read PROTO implementation
  nodeType->setScope(*nmspc);
  getNextTokenExpectType(&token, VRML_SYMBOL_LEFT_BRACE);
  loadPROTOImplementation(nodeType->scope(), nodeType);

  // add PROTO to the namespace
  nmspc->addNodeType(nodeType);
}



void Vrml97Parser::loadPROTODefaultFieldValue(VrmlNamespace* nmspc,
  VrmlNodeType* proto, const char* fieldName)
{
  VRMLToken token;
  VRMLValue value;

  // load the value
  loadValue(nmspc, proto, &value);

  // convert it to field type and set it
  VrmlField::VrmlFieldType fieldType = proto->hasField(fieldName);
  if (fieldType == VrmlField::NO_FIELD)
  {
    // This can happen only if the field was somehow rejected by the PROTO
    // (probably because of an inappropriate name). Do nothing in this case.
  }
  else
  {
    VrmlField* f = convertValue(&value, fieldType);
    if (f == NULL)
    {
      // this can happen only if the value was incorrect and was rejected;
      // we provide a default value in this case
      f = VrmlField::create(fieldType);
    }
    proto->setFieldDefault(fieldName, f);
    delete(f);
  }
}



void Vrml97Parser::loadPROTOImplementation(VrmlNamespace* nmspc,
  VrmlNodeType* proto)
{
  VRMLToken token;

  // assign starting comments to PROTO itself
  assignComments(*proto);

  while (true)
  {
    getNextToken(&token);
    if (token.match(VRML_KEYWORD_DEF))
      proto->addNode(loadNamedNode(nmspc, proto));
    else if (token.match(VRML_IDENTIFIER))
      proto->addNode(loadUnnamedNode(nmspc, token.c_str(), proto));
    else if (token.match(VRML_KEYWORD_ROUTE))
      loadRoute(nmspc);
    else if (token.match(VRML_KEYWORD_USE))
      proto->addNode(loadUsedNode(nmspc));
    else if (token.match(VRML_KEYWORD_PROTO))
      loadProto(nmspc);
    else if (token.match(VRML_KEYWORD_EXTERNPROTO))
      loadExternProto(nmspc);
    else if (token.match(VRML_SYMBOL_RIGHT_BRACE))
      break;
    else
      doError("Unexpected %s", token.desc().c_str());
  }

  // assign remaining unassigned comments to PROTO itself
  assignComments(*proto);
}



void Vrml97Parser::loadExternProto(VrmlNamespace* nmspc)
{
  VRMLToken token, protoNameToken;
  getNextTokenExpectIdentifier(&protoNameToken, "EXTERNPROTO name");
  getNextTokenExpectType(&token, VRML_SYMBOL_LEFT_BRACKET);

  // new node type we are defining
  VrmlNodeType* nodeType = new VrmlNodeType(protoNameToken.c_str());
  if (nmspc->findType(protoNameToken.c_str()) != NULL)
    doError(VrmlLoadingError::PROTO_EXISTS, "Prototype \"%.64s\" already exists", protoNameToken.c_str());

  // read EXTERNPROTO declaration
  while (true)
  {
    getNextToken(&token);

    // is it a closing bracket?
    if (token.match(VRML_SYMBOL_RIGHT_BRACKET)) break;

    // is it an event declaration?
    else if (token.match(VRML_KEYWORD_EVENT_IN, VRML_KEYWORD_EVENT_OUT))
    {
      VRMLToken eventNameToken, eventTypeToken;
      getNextTokenExpectIdentifier(&eventTypeToken, "event type");
      getNextTokenExpectIdentifier(&eventNameToken, "event name");
      VrmlField::VrmlFieldType eventType = VrmlField::fieldType(
        eventTypeToken.c_str());

      if (eventType == VrmlField::NO_FIELD)
        doError("Unknown event type \"%.64s\"", eventTypeToken.c_str());

      if (token.match(VRML_KEYWORD_EVENT_IN))
        nodeType->addEventIn(eventNameToken.c_str(), eventType);
      else
        nodeType->addEventOut(eventNameToken.c_str(), eventType);
    }

    // is it a field declaration?
    else if (token.match(VRML_KEYWORD_FIELD, VRML_KEYWORD_EXPOSED_FIELD))
    {
      VRMLToken fieldNameToken, fieldTypeToken;
      getNextTokenExpectIdentifier(&fieldTypeToken, "field type");
      getNextTokenExpectIdentifier(&fieldNameToken, "field name");
      VrmlField::VrmlFieldType fieldType = VrmlField::fieldType(
        fieldTypeToken.c_str());

      if (fieldType == VrmlField::NO_FIELD)
        doError("Unknown field type \"%.64s\"", fieldTypeToken.c_str());

      if (token.match(VRML_KEYWORD_FIELD))
        nodeType->addField(fieldNameToken.c_str(), fieldType);
      else
        nodeType->addExposedField(fieldNameToken.c_str(), fieldType);

      // EXTERNPROTO fields DO NOT have default value
    }

    else
      doError("Unexpected %s (should be a field name or '}')",
              token.desc().c_str());
  }

  // read EXTERNPROTO file URLs
  std::vector<char*> urls;
  while (true)
  {
    getNextToken(&token);
    if (token.match(VRML_STRING)) urls.push_back(newString(token.c_str()));
    else
    {
      pushbackToken(&token);
      break;
    }
  }
  if (urls.size() > 0)
  {
    // set the URLs
    nodeType->setUrl(new VrmlMFString(urls.size(), &urls.front()),
      new Doc(fileName.c_str()));

    // clean up
    for (unsigned int i=0; i<urls.size(); i++) delete[] urls[i];
  }
  else
    doError("EXTERNPROTO \"%.64s\" has no URLs", protoNameToken.c_str());

  // set scope (nmspc is parent namespace)
  nodeType->setScope(*nmspc);

  // add the EXTERNPROTO to the namespace
  nmspc->addNodeType(nodeType);
}
