/*****************************************************************************
******************************************************************************
*
*               VRML97 parser
*
*               $Revision: 1.1.2.9 $
*               $Date: 2011/06/01 14:15:21 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright 1998-2011 HUMUSOFT s.r.o.

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#ifndef VRML97PARSER_H__
#define VRML97PARSER_H__

#include <stack>
#include <string>
#include <vector>

#include "VrmlField.h"
#include "VrmlException.h"

class VrmlNamespace;
class VrmlNode;
class VrmlNodeScript;
class VrmlNodeType;
class VrmlMFTime;
class VrmlMFNode;
class VrmlNodeViewpoint;
class VrmlNodeBackground;
class VrmlNodeNavigationInfo;
class VrmlNodeFog;



/** Token type values. */
enum VRMLTokenType
{
  VRML_UNKNOWN,
  VRML_END_OF_FILE,
  VRML_STRING,
  VRML_NUMBER,
  VRML_IDENTIFIER,
  VRML_KEYWORD_DEF,
  VRML_KEYWORD_EXTERNPROTO,
  VRML_KEYWORD_FALSE,
  VRML_KEYWORD_IS,
  VRML_KEYWORD_NULL,
  VRML_KEYWORD_PROTO,
  VRML_KEYWORD_ROUTE,
  VRML_KEYWORD_TO,
  VRML_KEYWORD_TRUE,
  VRML_KEYWORD_USE,
  VRML_KEYWORD_EVENT_IN,
  VRML_KEYWORD_EVENT_OUT,
  VRML_KEYWORD_EXPOSED_FIELD,
  VRML_KEYWORD_FIELD,
  VRML_SYMBOL_DOT,
  VRML_SYMBOL_LEFT_BRACKET,                 // '['
  VRML_SYMBOL_RIGHT_BRACKET,                // ']'
  VRML_SYMBOL_LEFT_BRACE,                   // '{'
  VRML_SYMBOL_RIGHT_BRACE,                  // '}'
  VRML_SYMBOL_LEFT_PARENTHESIS,             // '('
  VRML_SYMBOL_RIGHT_PARENTHESIS             // ')'
};


/** A VRML parsing token. */
class VRMLToken : public std::string
{
  private:

    /** Token type (one of VRMLTokenType values). */
    int type;

  public:

    /** Constructor. */
    VRMLToken(VRMLTokenType _type = VRML_UNKNOWN) :
      type(_type)
    { }

    /** Clears the token content and resets its type. */
    VRMLToken& clear(int _type)
    { type = _type; std::string::clear(); return(*this); }

    /** Gets the token type.
      * @return Token type value. */
    int getType() const
    { return type; };

    /** Sets the token type. */
    void setType(int _type)
    { type = _type; };

    /** Returns the string representation of this token.
      * @return String containing the description.
      */
    std::string desc();

    /** Returns true if the token type matches one of the given types,
      * false if it does not match any.
      */
    bool match(int _type1) const
    { return(type==_type1); };

    bool match(int _type1, int _type2) const
    { return(type==_type1 || type==_type2); };

    bool match(int _type1, int _type2, int _type3) const
    { return(type==_type1 || type==_type2 || type==_type3); };
};



/** This object describes a(ny) VRML value just loaded from a VRML file. */
class VRMLValue
{
  public:

    std::vector<char*> strings;
    std::vector<double> numbers;
    std::vector<VrmlNode*> nodes;
    bool hasStrings;
    bool hasNumbers;
    bool hasBoolean;
    bool hasNodes;
    bool hasEmptyVec;

    /**
     * Destructor.
     */
    ~VRMLValue();
};



/** VRML stream decoder. */
class OPENVRMLDLLEXPORT Vrml97Parser
{
  private:
    /** A reference to the input stream. */
    istream& input;

    /** Characters prefetched from the input stream. */
    std::stack< char, std::vector<char> > prefetchedCharacters;

    /** Last character read. */
    char lastCharacter;

    /** Current column number in the VRML file. */
    int columnNumber;

    /** Current line number in the VRML file. */
    int lineNumber;

    /** File name of the source file (empty if unknown). */
    std::string fileName;

    /** True if a token is prefetched. */
    bool tokenPrefetched;

    /** A token prefetched from the input stream. */
    VRMLToken prefetchedToken;

    /** Array of comments fetched so far.
      * Comments are divided to separate comments blocks,
      * empty string means empty line.
      */
    std::vector<std::string> comments;

    /** A token prefetched from the input stream. */
    VrmlObject* previousCommentTarget;

private:

    /** Reads a next character from the input stream.
      * @return The character, or -1 if an end of stream
      * or an I/O error occurred.
      */
    int getNextCharacter();

    /** Reads a next character from the input stream, skipping whitespace.
      * Newline (\n) is not treated as whitespace here.
      * @return The character, or -1 if an end of stream
      * or an I/O error occurred.
      */
    int getNextCharacterSkipWhitespace();

    /** Pushes the character back, to be read next time.
      * @param c Character to be pushed.
      */
    void pushBackCharacter(char c)
    {
      prefetchedCharacters.push(c);
    }

    /** Reads characters up to new line.
      * @return Characters up to newline; the newline is not included.
      */
    std::string getNextLine();

    /** Reads a next token from the input stream.
      * @param token A VRMLToken object to receive the new token.
      * @return VRMLToken object to receive the new token.
      */
    VRMLToken* getNextToken(VRMLToken* token);

    /** Reads a next token from the input stream
      * and compares its type with the specified types.
      * @param token A VRMLToken object to receive the new token.
      * @return True if the token is of the specified type,
      * false if it has a different type.
      */
    bool getNextTokenMatch(VRMLToken* token, int _type);

    /** Pushes a token back to the input stream to be read
      * by next call to getNextToken().
      * Only one token at a time can be pushed back.
      * Another pushed token will replace the previously pushed one.
      * @param token A token to be pushed back.
      */
    void pushbackToken(const VRMLToken* token);

    /** Same as getNextToken() but throws an exception
      * if the token just read is not of appropriate type.
      * @param token A VRMLToken object to receive the new token.
      * @param type Token type expected.
      * @throw VrmlLoadingError
      */
    void getNextTokenExpectType(VRMLToken* token, int type);

    /** Same as getNextToken() but throws an exception
      * if the token is not an identifier, reporting what
      * kind of identifier we were trying to read (for better
      * debugging).
      * @param token A VRMLToken object to receive the new token.
      * @param what What kind of identifier we expected.
      * @throw VrmlLoadingError
      */
    void getNextTokenExpectIdentifier(VRMLToken* token, const char* what);

    /** Same as getNextToken() but throws an exception
      * if the token just read is not one of specified types.
      * @param token A VRMLToken object to receive the new token.
      * @param type List of possible token types, terminated by -1.
      * @throw VrmlLoadingError
      */
    void getNextTokenExpectType(VRMLToken* token, const int* types);

    /** Loads a ROUTE declaration from the input stream.
      * The first token to be read must be the source node name
      * (i.e. the first token after the ROUTE keyword).
      * @param nmspc Namespace.
      * @throw VrmlLoadingError
      */
    void loadRoute(VrmlNamespace* nmspc);

    /** Loads a named node from the input stream.
      * The first token to be read must be the node name
      * (i.e. the first token after the DEF keyword).
      * @param nmspc Namespace.
      * @param parentProto Parent PROTO (NULL if we are at top level).
      * @return Newly allocated node, already added to the namespace.
      * @throw VrmlLoadingError
      */
    VrmlNode* loadNamedNode(VrmlNamespace* nmspc, VrmlNodeType* parentProto);

    /** Loads an unnamed node from the input stream.
      * The first token to be read must be the opening brace
      * (i.e. the first token after the node type token).
      * @param nmspc Namespace.
      * @param nodeTypeName Node type string.
      * @param parentProto Parent PROTO (NULL if we are at top level).
      * @return Newly allocated node, already added to the namespace.
      * @throw VrmlLoadingError
      */
    VrmlNode* loadUnnamedNode(VrmlNamespace* nmspc,
      const char* nodeTypeName, VrmlNodeType* parentProto);

    /** Loads a USE clause and returns a pointer to node it references.
      * The first token to be read must be the referenced node name
      * (i.e. the first token after the USE keyword).
      * @param nmspc Namespace.
      * @return Pointer to the referenced node (which already existed),
      * or NULL if the USE clause referenced an unknown node.
      * @throw VrmlLoadingError
      */
    VrmlNode* loadUsedNode(VrmlNamespace* nmspc);

    /** Loads a VRML value from the file (also checks its type).
      * The first token to be read must be the first token of the value.
      * @param nmspc Namespace.
      * @param parentProto Parent PROTO (NULL if we are at top level).
      * @param value Structure to receive the value read.
      * @throw VrmlLoadingError
      */
    void loadValue(VrmlNamespace* nmspc, VrmlNodeType* parentProto,
      VRMLValue* value);

    /** Converts a given value to the given VRML type (if possible).
      * @return The newly allocated VrmlField object, or NULL
      * if the conversion is not possible (also adds an appropriate
      * error to the error list).
      */
    VrmlField* convertValue(VRMLValue* value, VrmlField::VrmlFieldType type);

    /** Loads a field value and sets it.
      * @param nmspc Namespace.
      * @param node Node.
      * @param fieldName Name of the field to be set.
      * @param parentProto Parent PROTO (NULL if we are at top level).
      */
    void loadFieldValue(VrmlNamespace* nmspc, VrmlNode* node,
      const char* fieldName, VrmlNodeType* parentProto);

    /** Loads contents of a node.
      * The first token to be read must be the first token of the node
      * contents (i.e. first token after the opening curly brace).
      * @param nmspc Namespace.
      * @param node Node to be loaded.
      * @param parentProto Parent PROTO (NULL if we are at top level).
      * @throw VrmlLoadingError
      */
    void loadNodeContents(VrmlNamespace* nmspc, VrmlNode* node,
      VrmlNodeType* parentProto);

    /** Loads a prototype.
      * The first token to be read must be the PROTO name
      * (i.e. the first token after the PROTO keyword).
      * @param nmspc Namespace.
      * @throw VrmlLoadingError
      */
    void loadProto(VrmlNamespace* nmspc);

    /** Loads a default field value in a PROTO field declaration.
      * The first token to be read must be the first token
      * of the value (i.e. first token after the field name).
      * @param nmspc Namespace.
      * @param proto PROTO being loaded.
      * @param fieldName Field name.
      */
    void loadPROTODefaultFieldValue(VrmlNamespace* nmspc,
      VrmlNodeType* proto, const char* fieldName);

    /** Loads a PROTO implementation.
      * The first token to be read must be the first token
      * of the implementation body (i.e. the first token after
      * the opening brace).
      * @param nmspc Namespace.
      * @param proto PROTO being loaded.
      */
    void loadPROTOImplementation(VrmlNamespace* nmspc,
      VrmlNodeType* proto);

    /**
     * Produces the VrmlLoadingError object with appropriate error message
     * and throws it as an exception. This method does not return normally.
     * @param msg The message (formatting string as in printf).
     * @exception VrmlLoadingError
     */
    void doError(const char* msg, ...);

    /**
     * Produces the VrmlLoadingError object with appropriate error message
     * and throws it as an exception of given type. This method does not return normally.
     * @param etype The type of VrmlLoadingError
     * @param msg The message (formatting string as in printf).
     * @exception VrmlLoadingError
     */
    void doError(VrmlLoadingError::errorType etype, char* msg, ...);

    /** Character classifier.
      * @return True if the character is a valid identifier-start character,
      * false otherwise.
      */
    static bool isValidIdFirstChar(char c);

    /** Character classifier.
      * @return True if the character is a valid identifier character,
      * false otherwise.
      */
    static bool isValidIdRestChar(char c);

    /** Character classifier.
      * @return True if the character is a hex digit (0..9, a..f, A..F),
      * false otherwise.
      */
    static bool isHexDigit(char c);


  public:

    /** Creates a new decoder for the given input stream.
      * @param input Input stream.
      * @param filename Name of the file the input stream belongs to.
      * This is used for reporting errors and for loading other files
      * (EXTERNPROTOs) from the same directory. Can be NULL if
      * not needed.
      */
    explicit Vrml97Parser(std::istream& input, const char* filename = NULL);

    /** Destructor.
      */
    ~Vrml97Parser();

    /** Parses the input stream and produces a set of nodes
      * which are added to the given parent node.
      * @param parent Parent node to add new nodes to.
      * @param namespace Namespace node.
      * @throw VrmlLoadingError
      */
    void loadTopLevel(VrmlMFNode* parent, VrmlNamespace* nmspc);

    /** Assign comments to a VRML object
      * @param obj Object to assign commets to.
      */
    void assignComments(VrmlObject& obj);

    /** String classifier.
      * @return True if the string is a valid identifier,
      * false otherwise.
      */
    static bool isValidId(const std::string& name);

    /** Loads an extern prototype declaration.
      * The first token to be read must be the EXTERNPROTO name
      * (i.e. the first token after the EXTERNPROTO keyword).
      * @param nmspc Namespace.
      * @throw VrmlLoadingError
      */
    void loadExternProto(VrmlNamespace* nmspc);


    /** A list of viewpoint nodes in the file. */
    std::vector<VrmlNodeViewpoint*> viewpoints;

    /** A list of background nodes in the file. */
    std::vector<VrmlNodeBackground*> backgrounds;

    /** A list of navigation info nodes in the file. */
    std::vector<VrmlNodeNavigationInfo*> navinfos;

    /** A list of fog nodes in the file. */
    std::vector<VrmlNodeFog*> fogs;
};



#endif // VRML97PARSER_H__
