/*****************************************************************************
******************************************************************************
*
*               VRML exception
*
*               $Revision: 1.1.4.2.2.1 $
*               $Date: 2010/02/23 20:57:36 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright 1998-2010 HUMUSOFT s.r.o.

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//



#ifndef VRML_EXCEPTION_H
#define VRML_EXCEPTION_H

#include <cstdarg>
#include <stdexcept>
#include <string>


/** A base class for exceptions related to the VRML engine. */
class VrmlException : public std::runtime_error
{
protected:
  /** Protected constructor.
    * To be used as base class only.
    */
  explicit VrmlException(const std::string& msg) :
    runtime_error(msg)
  { }
};



/** An exception related to input/output operations. */
class SocketException : public VrmlException
{
public:
  /** Constructor. */
  explicit SocketException(const char* msg) :
    VrmlException(msg)
  { }
};



/**
 * This object describes an error that occurred during loading a VRML file.
 */
class VrmlLoadingError : public VrmlException
{
private:

  /* private class that creates the error message string */
  class creator : public std::string
  {
  private:
    void create(const char* filename, int line, int column, const char* fmt, va_list ap);

  public:
    creator(const char* filename, int line, int column, const char* fmt, ...)
    {
      va_list ap;
      va_start(ap, fmt);
      create(filename, line, column, fmt, ap);
      va_end(ap);
    }

    creator(const char* filename, int line, int column, const char* fmt, va_list ap)
    {
      create(filename, line, column, fmt, ap);
    }
  };

public:
  enum errorType
  {
    UNKNOWN,
    PROTO_EXISTS,
  } eType;

  /** Constructor.
    * @param type of VrmlLoadingError
    * @param filename Name of the offending file (NULL if
    * unknown/not relevant).
    * @param line Number of the offending line (-1 if
    * unknown/not relevant).
    * @param column Number of column where the offending phrase starts
    * (-1 if unknown/not relevant).
    * @param fmt Error message format.
    * @param ap Pointer to printf-like arguments.
  */
  VrmlLoadingError(errorType etype, const char* filename, int line, int column, const char* fmt, va_list ap) :
    VrmlException(creator(filename, line, column, fmt, ap)),
    eType(etype)
  { }

  /** Constructor.
    * @param msg Error message format.
    * @param filename Name of the offending file (NULL if
    * unknown/not relevant).
  */
  explicit VrmlLoadingError(const char* msg, const char* filename=0) :
    VrmlException(creator(filename, -1, -1, "%s", msg)),
    eType(UNKNOWN)
  { }

};



#endif
