/*****************************************************************************
******************************************************************************
*
*               VRML field (also contains all field types)
*
*               $Revision: 1.1.2.25 $
*               $Date: 2012/02/03 15:27:33 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright 1998-2012 HUMUSOFT s.r.o.
// Portions Copyright (C) 1998  Chris Morley

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#include <stdio.h>
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <algorithm>
#include <numeric>
#include <assert.h>
#include <string.h>
#include <stdarg.h>

#include <new>          // bad_alloc needs <new> - don't rely it is included indirectly

#include "VrmlField.h"
#include "MathUtils.h"
#include "VrmlNode.h"

#define NDEBUG


/** An entry for the field type table. */
class FieldTypeTableEntry
{
public:

  /** Field type name. */
  const char* name;

  /** Field type identifier. */
  VrmlField::VrmlFieldType id;
};



/** The field type table. */
static FieldTypeTableEntry fieldTypeTable[] =
{
  { "SFBool",        VrmlField::SFBOOL          },
  { "SFColor",       VrmlField::SFCOLOR         },
  { "SFFloat",       VrmlField::SFFLOAT         },
  { "SFImage",       VrmlField::SFIMAGE         },
  { "SFInt32",       VrmlField::SFINT32         },
  { "SFNode",        VrmlField::SFNODE          },
  { "SFRotation",    VrmlField::SFROTATION      },
  { "SFString",      VrmlField::SFSTRING        },
  { "SFTime",        VrmlField::SFTIME          },
  { "SFVec2f",       VrmlField::SFVEC2F         },
  { "SFVec3f",       VrmlField::SFVEC3F         },
  { "MFColor",       VrmlField::MFCOLOR         },
  { "MFFloat",       VrmlField::MFFLOAT         },
  { "MFInt32",       VrmlField::MFINT32         },
  { "MFNode",        VrmlField::MFNODE          },
  { "MFRotation",    VrmlField::MFROTATION      },
  { "MFString",      VrmlField::MFSTRING        },
  { "MFTime",        VrmlField::MFTIME          },
  { "MFVec2f",       VrmlField::MFVEC2F         },
  { "MFVec3f",       VrmlField::MFVEC3F         }
};


// on Mac, bool takes 4 bytes but we want it to be 1 byte in the external interface
// so we use extLogical instead
#ifdef __APPLE__
typedef unsigned char extLogical;
#else
typedef bool extLogical;
#endif



// ColorFloat class for floats limited to range 0 to 1
class ColorFloat
{
private:
  float f;

public:
  template <typename T>
  ColorFloat(T x) :
    f(x<=0 ? 0 : (x>=1 ? 1 : static_cast<float>(x)))
  { }

  ColorFloat(bool x) :
    f(x)
  { }
};



/*****************************************************************************
;*
;*		Conversion from any datatype to VRMLField data
;*
;****************************************************************************/

/* copy array of ST values to array of DT values, converting as necessary and skipping NaNs in ST, rounding ST to nearest */
template<typename ST, typename DT>
void Any2AnyTrNaNRound(DT* dst, size_t m, size_t n, const ST* src)
{
  size_t i, j, k;
  for (i=k=0; i<m; ++i)
    for (j=0; j<n; ++j, ++k)
    {
      const ST& val = src[i+j*m];
      if (!isnan(val))
	dst[k] = static_cast<DT>(val+ (val<0 ? -0.5 : 0.5));
    }
}

/* copy array of ST values to array of DT values, converting as necessary and skipping NaNs in ST */
template<typename ST, typename DT>
void Any2AnyTrNaN(DT* dst, size_t m, size_t n, const ST* src)
{
  size_t i, j, k;
  for (i=k=0; i<m; ++i)
    for (j=0; j<n; ++j, ++k)
    {
      const ST& val = src[i+j*m];
      if (!isnan(val))
	dst[k] = static_cast<DT>(val);
    }
}

/* Any2AnyTrNaN, specialized for DT=bool */
template<typename ST>
void Any2AnyTrNaN(bool* dst, size_t m, size_t n, const ST* src)
{
  size_t i, j, k;
  for (i=k=0; i<m; ++i)
    for (j=0; j<n; ++j, ++k)
    {
      const ST& val = src[i+j*m];
      if (!isnan(val))
	dst[k] = (val != 0);
    }
}

/* copy array of ST values to array of DT values, converting as necessary */
template<typename ST, typename DT>
void Any2AnyTr(DT* dst, size_t m, size_t n, const ST* src)
{
  size_t i, j, k;
  for (i=k=0; i<m; ++i)
    for (j=0; j<n; ++j, ++k)
      dst[k] = static_cast<DT>(src[i+j*m]);
}

/* Any2AnyTr, specialized for DT=bool */
template<typename ST>
void Any2AnyTr(bool* dst, size_t m, size_t n, const ST* src)
{
  size_t i, j, k;
  for (i=k=0; i<m; ++i)
    for (j=0; j<n; ++j, ++k)
      dst[k] = (src[i+j*m] != 0);
}


/* converts any numeric data type to SFImage represented by array of unsigned chars; serves only for ML2VRML transformation */
template<typename ST>
void Any2VRMLSFImageTr(unsigned char* dst, size_t m, size_t n, size_t p, const ST* src)
{
  const int UPPER_BOUND = 255;
  size_t k = 0;
  for (ptrdiff_t i=m-1; i>=0; --i) // i must not be size_t
    for (size_t j=0; j<n; ++j)
      for (size_t r=0; r<p; ++k, ++r)
      {
	ST val = src[i+(j*m)+r*m*n];
	if (val<=0)
	  dst[k] = 0;
	else
	  if (val > UPPER_BOUND)
	    dst[k] = UPPER_BOUND;
	  else
	    dst[k] = static_cast<unsigned char>(val);
      }
}

/* converts any numeric data type to SFImage represented by array of unsigned chars; serves only for ML2VRML transformation */
template<typename ST>
void Any2VRMLSFImageTrNaNRound(unsigned char* dst, size_t m, size_t n, size_t p, const ST* src)
{
  const int UPPER_BOUND = 255;
  size_t k = 0;
  for (ptrdiff_t i=m-1; i>=0; --i) // i must not be size_t
    for (size_t j=0; j<n; ++j)
      for (size_t r=0; r<p; ++k, ++r)
      {
        ST val = src[i+(j*m)+r*m*n];
        if (!isnan(val))
        {
          if (val<=0)
            dst[k] = 0;
          else
            if (val > UPPER_BOUND)
              dst[k] = UPPER_BOUND;
            else
              dst[k] = static_cast<unsigned char>(val+0.5);
        }
      }
}

/* converts bool data type to SFImage represented by array of unsigned chars; serves only for ML2VRML transformation */
template<>
void Any2VRMLSFImageTr(unsigned char* dst, size_t m, size_t n, size_t p, const bool* src)
{
  size_t k = 0;
  for (ptrdiff_t i=m-1; i>=0; --i) // i must not be size_t
    for (size_t j=0; j<n; ++j)
      for (size_t r=0; r<p; ++k, ++r)
        dst[k] = static_cast<unsigned char>(src[i+(j*m)+r*m*n]);
}

/* converts any numeric data type to SFImage represented by array of unsigned chars; serves only for ML2VRML transformation */
template<typename ST>
void VRMLSFImage2AnyTr(unsigned char* dst, size_t m, size_t n, size_t p, const ST* src)
{
  const int UPPER_BOUND = 255;
  size_t k = 0;
  for (ptrdiff_t i=m-1; i>=0; --i) // i must not be size_t
    for (size_t j=0; j<n; ++j)
      for (size_t r=0; r<p; ++k, ++r)
      {
	size_t index = i+(j*m)+r*m*n;
	ST val = src[k];
	if (val<=0)
	  dst[index] = 0;
	else
	  if (val > UPPER_BOUND)
	    dst[index] = UPPER_BOUND;
	  else
	    dst[index] = static_cast<unsigned char>(val);
      }
}

/* converts bool data type to SFImage represented by array of unsigned chars; serves only for ML2VRML transformation */
template<>
void VRMLSFImage2AnyTr(unsigned char* dst, size_t m, size_t n, size_t p, const bool* src)
{
  size_t k = 0;
  for (ptrdiff_t i=m-1; i>=0; --i) // i must not be size_t
    for (size_t j=0; j<n; ++j)
      for (size_t r=0; r<p; ++k, ++r)
        dst[i+(j*m)+r*m*n] = static_cast<unsigned char>(src[k]);
}

/* copy array of values of type specified by classid to array of DT values, converting as necessary */
template<typename DT>
bool MLData2VRML(DT* dst, extClassID classid, size_t m, size_t n, const void* src)
{
  switch(classid)
  {
    case extDOUBLE_CLASS  : Any2AnyTrNaN(dst, m, n, static_cast<const double*>(src)); break;
    case extSINGLE_CLASS  : Any2AnyTrNaN(dst, m, n, static_cast<const float*>(src)); break;
    case extLOGICAL_CLASS :    Any2AnyTr(dst, m, n, static_cast<const extLogical*>(src)); break;
    case extINT8_CLASS    :    Any2AnyTr(dst, m, n, static_cast<const signed char*>(src)); break;
    case extUINT8_CLASS   :    Any2AnyTr(dst, m, n, static_cast<const unsigned char*>(src)); break;
    case extINT16_CLASS   :    Any2AnyTr(dst, m, n, static_cast<const short*>(src)); break;
    case extUINT16_CLASS  :    Any2AnyTr(dst, m, n, static_cast<const unsigned short*>(src)); break;
    case extINT32_CLASS   :    Any2AnyTr(dst, m, n, static_cast<const int*>(src)); break;
    case extUINT32_CLASS  :    Any2AnyTr(dst, m, n, static_cast<const unsigned int*>(src)); break;
    case extINT64_CLASS   :    Any2AnyTr(dst, m, n, static_cast<const long long*>(src)); break;
    case extUINT64_CLASS  :    Any2AnyTr(dst, m, n, static_cast<const unsigned long long*>(src)); break;
    default : return(false);
  }
  return(true);
}

/* copy array of values of type specified by classid to array of DT values, converting as necessary */
template<typename DT>
bool MLData2VRMLRound(DT* dst, extClassID classid, size_t m, size_t n, const void* src)
{
  switch(classid)
  {
    case extDOUBLE_CLASS  : Any2AnyTrNaNRound(dst, m, n, static_cast<const double*>(src)); break;
    case extSINGLE_CLASS  : Any2AnyTrNaNRound(dst, m, n, static_cast<const float*>(src)); break;
    default : return(MLData2VRML(dst, classid, m, n, src));
  }
  return(true);
}

/* copy array of ST values to array of values of type specified by classid, converting as necessary */
template<typename ST>
bool VRMLData2ML(void* dst, extClassID classid, size_t m, size_t n, const ST* src)
{
  switch(classid)
  {
    // m and n must be swapped because they describe the (transposed} destination
    case extDOUBLE_CLASS  : Any2AnyTr(static_cast<double*>(dst), n, m, src); break;
    case extSINGLE_CLASS  : Any2AnyTr(static_cast<float*>(dst),  n, m, src); break;
    case extINT32_CLASS   : Any2AnyTr(static_cast<int*>(dst),    n, m, src); break;
    default : return(false);  // other types omitted to reduce code size, can be added when necessary
  }
  return(true);
}

bool MLData2VRMLSFImage(unsigned char* dst, extClassID classid, size_t m, size_t n, size_t k, const void* src)
{
  if (k>4)
    return(false);

  switch(classid)
  {
    case extDOUBLE_CLASS  : Any2VRMLSFImageTrNaNRound(dst, m, n, k, static_cast<const double*>(src)); break;
    case extSINGLE_CLASS  : Any2VRMLSFImageTrNaNRound(dst, m, n, k, static_cast<const float*>(src)); break;
    case extLOGICAL_CLASS : Any2VRMLSFImageTr(dst, m, n, k, static_cast<const bool*>(src)); break;
    case extINT8_CLASS    : Any2VRMLSFImageTr(dst, m, n, k, static_cast<const signed char*>(src)); break;
    case extUINT8_CLASS   : Any2VRMLSFImageTr(dst, m, n, k, static_cast<const unsigned char*>(src)); break;
    case extINT16_CLASS   : Any2VRMLSFImageTr(dst, m, n, k, static_cast<const short*>(src)); break;
    case extUINT16_CLASS  : Any2VRMLSFImageTr(dst, m, n, k, static_cast<const unsigned short*>(src)); break;
    case extINT32_CLASS   : Any2VRMLSFImageTr(dst, m, n, k, static_cast<const int*>(src)); break;
    case extUINT32_CLASS  : Any2VRMLSFImageTr(dst, m, n, k, static_cast<const unsigned int*>(src)); break;
    case extINT64_CLASS   : Any2VRMLSFImageTr(dst, m, n, k, static_cast<const long long*>(src)); break;
    case extUINT64_CLASS  : Any2VRMLSFImageTr(dst, m, n, k, static_cast<const unsigned long long*>(src)); break;

    default : return(false);
  }
  return(true);
}

bool VRMLSFImage2MLData(unsigned char* dst, extClassID classid, size_t m, size_t n, size_t k, const void* src)
{
  switch(classid)
  {
    case extUINT8_CLASS   : VRMLSFImage2AnyTr(dst, m, n, k, static_cast<const unsigned char*>(src)); break;
    default : return(false);  // other types omitted to reduce code size, can be added when necessary
  }
  return(true);
}



/*****************************************************************************
;*
;*		Generic MF float and double print function
;*
;****************************************************************************/

template <typename T>
static ostream& mfprint(ostream& os, const T* c, size_t n, size_t eltsize, size_t indent )
{
  if (n==1)
    for (size_t e=0; e<eltsize; ++e)
      os << c[e] << ((e < eltsize-1) ? " " : "");

  else
  {
    std::string indentString(indent, ' ');
    std::string indentIncrementedString(indent+VrmlNode::INDENT_INCREMENT, ' ');
    os << "[\n" << indentIncrementedString;
    for (size_t i=0; i<n; ++i, c+=eltsize)
    {
      for (size_t e=0; e<eltsize; ++e)
        os << c[e] << ((e < eltsize-1) ? " " : "");
      os << ((i < n-1) ? ",\n" + indentIncrementedString : " ");
    }
    os << "\n" << indentString << "]";
  }

  return os;
}



/*****************/
/*** VrmlField ***/
/*****************/



VrmlField::~VrmlField() {}



VrmlField::VrmlFieldType VrmlField::fieldType(const char* type)
{
  int i;
  for (i = 0; i<sizeof(fieldTypeTable)/sizeof(FieldTypeTableEntry); i++)
    if (strcmp(type, fieldTypeTable[i].name) == 0)
      return fieldTypeTable[i].id;
  return VrmlField::NO_FIELD;
}



const char* VrmlField::fieldTypeName() const
{
  return getTypeName(fieldType());
}



const char* VrmlField::getTypeName(int ftype)
{
  for (int i = 0; i<sizeof(fieldTypeTable)/sizeof(FieldTypeTableEntry); i++)
    if (ftype == fieldTypeTable[i].id) return fieldTypeTable[i].name;
  return NULL;
}



VrmlField* VrmlField::create(int ftype)
{
  switch (ftype)
  {
    case VrmlField::SFBOOL: return new VrmlSFBool();
    case VrmlField::SFFLOAT: return new VrmlSFFloat();
    case VrmlField::SFTIME: return new VrmlSFTime();
    case VrmlField::SFINT32: return new VrmlSFInt32();
    case VrmlField::SFVEC2F: return new VrmlSFVec2f();
    case VrmlField::SFVEC3F: return new VrmlSFVec3f();
    case VrmlField::SFCOLOR: return new VrmlSFColor();
    case VrmlField::SFROTATION: return new VrmlSFRotation();
    case VrmlField::SFSTRING: return new VrmlSFString();
    case VrmlField::SFIMAGE: return new VrmlSFImage();
    case VrmlField::SFNODE: return new VrmlSFNode();
    case VrmlField::MFFLOAT: return new VrmlMFFloat();
    case VrmlField::MFTIME: return new VrmlMFTime();
    case VrmlField::MFINT32: return new VrmlMFInt32();
    case VrmlField::MFVEC2F: return new VrmlMFVec2f();
    case VrmlField::MFVEC3F: return new VrmlMFVec3f();
    case VrmlField::MFCOLOR: return new VrmlMFColor();
    case VrmlField::MFROTATION: return new VrmlMFRotation();
    case VrmlField::MFSTRING: return new VrmlMFString();
    case VrmlField::MFNODE: return new VrmlMFNode();
    default: return NULL;
  }
}



VrmlField* VrmlField::convert(int ftype, size_t count, const double* values)
{
  switch (ftype)
  {
    case VrmlField::SFBOOL:
      return (count == 1) ? new VrmlSFBool(values[0] != 0.0) : NULL;

    case VrmlField::SFFLOAT:
      return (count == 1) ? new VrmlSFFloat((float) values[0]) : NULL;

    case VrmlField::SFTIME:
      return (count == 1) ? new VrmlSFTime((float) values[0]) : NULL;

    case VrmlField::SFINT32:
      return (count == 1) ? new VrmlSFInt32((int) values[0]) : NULL;

    case VrmlField::SFVEC2F:
      return (count == 2) ? new VrmlSFVec2f((float) values[0], (float) values[1]) : NULL;

    case VrmlField::SFCOLOR:
      return (count == 3) ? new VrmlSFColor((float) values[0], (float) values[1], (float) values[2])
        : NULL;

    case VrmlField::SFVEC3F:
      return (count == 3) ? new VrmlSFVec3f((float) values[0], (float) values[1], (float) values[2])
        : NULL;

    case VrmlField::SFROTATION:
      return (count == 4) ? new VrmlSFRotation((float) values[0], (float) values[1],
                                               (float) values[2], (float) values[3]) : NULL;

    case VrmlField::SFIMAGE:
    {
      if (count < 3) return NULL;
      int width  = (int) values[0];
      int height = (int) values[1];
      int compo  = (int) values[2];
      if (width < 0 || height < 0 || compo < 0 || compo > 4) return NULL;
      int bytes = width*height*compo;
      if (count - 3 != width*height) return NULL;
      unsigned char* data = new unsigned char[bytes];
      int i = 0;
      int j = 3;
      while (i<bytes)
      {
        unsigned int pixel = (unsigned int) values[j];
        switch (compo)
        {
          case 4:  data[i++] = (pixel >> 24) & 0xff;
                   // fall through
          case 3:  data[i++] = (pixel >> 16) & 0xff;
                   // fall through
          case 2:  data[i++] = (pixel >> 8)  & 0xff;
                   // fall through
          case 1:  data[i++] = (pixel)       & 0xff;
        }
	j++;
      }
      VrmlField* result = new VrmlSFImage(width, height, compo, data);
      delete[] data;
      return result;
    }

    case VrmlField::MFFLOAT:
    {
      VrmlMFFloat* field = new VrmlMFFloat(count, NULL);
      for (size_t i=0; i<count; i++) field->setElement(i, (float) values[i]);
      return field;
    }

    case VrmlField::MFTIME:
      return new VrmlMFTime(count, values);

    case VrmlField::MFINT32:
    {
      VrmlMFInt32* field = new VrmlMFInt32(count, NULL);
      for (size_t i=0; i<count; i++) field->setElement(i, (int) values[i]);
      return field;
    }

    case VrmlField::MFVEC2F:
    {
      if (count % 2 != 0) return NULL;
      VrmlMFVec2f* field = new VrmlMFVec2f(count/2, NULL);
      for (size_t i=0; i<count/2; i++)
        field->setElement(i, (float) values[2*i], (float) values[2*i+1]);
      return field;
    }

    case VrmlField::MFVEC3F:
    {
      if (count % 3 != 0) return NULL;
      VrmlMFVec3f* field = new VrmlMFVec3f(count/3, NULL);
      for (size_t i=0; i<count/3; i++)
        field->setElement(i, (float) values[3*i], (float) values[3*i+1], (float) values[3*i+2]);
      return field;
    }

    case VrmlField::MFCOLOR:
    {
      if (count % 3 != 0) return NULL;
      VrmlMFColor* field = new VrmlMFColor(count/3, NULL);
      for (size_t i=0; i<count/3; i++)
        field->setElement(i, (float) values[3*i], (float) values[3*i+1], (float) values[3*i+2]);
      return field;
    }

    case VrmlField::MFROTATION:
    {
      if (count % 4 != 0) return NULL;
      VrmlMFRotation* field = new VrmlMFRotation(count/4, NULL);
      for (size_t i=0; i<count/4; i++)
      {
        field->setElement(i, (float) values[4*i], (float) values[4*i+1], (float) values[4*i+2],
          (float) values[4*i+3]);
      }
      return field;
    }

    default:
      return NULL;
  }
}



VrmlField* VrmlField::convert(int ftype, size_t count, const char** values)
{
  switch (ftype)
  {
    case VrmlField::SFSTRING:
      return (count == 1) ? new VrmlSFString(values[0]) : NULL;

    case VrmlField::MFSTRING:
    {
      // special case: treat single empty string on input as empty MFString
      // this is to treat [] and "" (and also [""]) the same
      if ((count==1) && values[0][0]=='\0') count=0;

      return new VrmlMFString(count, values);
    }

    default:
      return NULL;
  }
}



VrmlField* VrmlField::convert(int ftype, size_t count, VrmlNode* const* values)
{
  switch (ftype)
  {
    case VrmlField::SFNODE:
      return (count == 1) ? new VrmlSFNode(values[0]) : NULL;

    case VrmlField::MFNODE:
      return new VrmlMFNode(count, values);

    default:
      return NULL;
  }
}

bool VrmlField::set(const VrmlField &src)
{
  return VrmlObject::set(src);
}



VrmlField* VrmlField::convertEmptyVector(int ftype)
{
  switch (ftype)
  {
    case VrmlField::MFCOLOR:         return new VrmlMFColor();
    case VrmlField::MFFLOAT:         return new VrmlMFFloat();
    case VrmlField::MFINT32:         return new VrmlMFInt32();
    case VrmlField::MFNODE:          return new VrmlMFNode();
    case VrmlField::MFROTATION:      return new VrmlMFRotation();
    case VrmlField::MFSTRING:        return new VrmlMFString();
    case VrmlField::MFTIME:          return new VrmlMFTime();
    case VrmlField::MFVEC2F:         return new VrmlMFVec2f();
    case VrmlField::MFVEC3F:         return new VrmlMFVec3f();
    default:                         return NULL;
  }
}



ostream& operator<<(ostream& os, const VrmlField& f)
{ return f.print(os); }


ostream& VrmlField::printAsUrl(ostream& os, int indent) const
{
  return os;
}


bool VrmlField::isTypeNumeric(int ftype)
{
  ftype &= ~MFFIELD;
  return (ftype == SFINT32 || ftype == SFFLOAT || ftype == SFTIME
    || ftype == SFVEC2F || ftype == SFVEC3F || ftype == SFCOLOR
    || ftype == SFROTATION || ftype == SFIMAGE);
}


bool VrmlField::IsExtSettable(int ftype)
{
  // all the numeric types and SFBOOL are settable
  return (isTypeNumeric(ftype) || ftype==SFBOOL);
}


bool VrmlField::isTypeSF(int ftype)
{
  return ((ftype & MFFIELD) == 0);
}


size_t VrmlField::getTypeWidth(int ftype)
{
  switch (ftype & ~MFFIELD)
  {
    case SFBOOL:     return 1;
    case SFFLOAT:    return 1;
    case SFINT32:    return 1;
    case SFTIME:     return 1;
    case SFVEC2F:    return 2;
    case SFVEC3F:    return 3;
    case SFCOLOR:    return 3;
    case SFROTATION: return 4;
  }
  return 0;                       // width not known
}


extClassID VrmlField::getTypeExtClassID(int ftype)
{
  switch (ftype & ~MFFIELD)
  {
    case SFBOOL:     return extLOGICAL_CLASS;
    case SFCOLOR:    return extSINGLE_CLASS;
    case SFFLOAT:    return extSINGLE_CLASS;
    case SFIMAGE:    return extUINT8_CLASS;
    case SFINT32:    return extINT32_CLASS;
    case SFNODE:     return extCHAR_CLASS;
    case SFROTATION: return extSINGLE_CLASS;
    case SFSTRING:   return extCHAR_CLASS;
    case SFTIME:     return extDOUBLE_CLASS;
    case SFVEC2F:    return extSINGLE_CLASS;
    case SFVEC3F:    return extSINGLE_CLASS;
  }
  return extUNKNOWN_CLASS;        // class not known
}


const char* VrmlField::getTypeExtClassName(int ftype)
{
  switch (ftype & ~MFFIELD)
  {
    case SFBOOL:     return "logical";
    case SFCOLOR:    return "single";
    case SFFLOAT:    return "single";
    case SFIMAGE:    return "uint8";
    case SFINT32:    return "int32";
    case SFNODE:     return "char";
    case SFROTATION: return "single";
    case SFSTRING:   return "char";
    case SFTIME:     return "double";
    case SFVEC2F:    return "single";
    case SFVEC3F:    return "single";
  }
  return "unknown";        // class name not known
}


bool VrmlField::getExtValueInWorldCoords(extClassID type, void* values, const Matrix4x4* tm) const
{
  return false;
}



/******************/
/*** VrmlSFBool ***/
/******************/



VrmlField * VrmlSFBool::clone(VrmlNamespace* ns) const
  { return new VrmlSFBool(*this); }

bool VrmlSFBool::equals(const VrmlField* f) const
{
  return (f->fieldType() == SFBOOL
          && ((VrmlSFBool*) f)->d_value == d_value);
}

bool VrmlSFBool::set(const VrmlField &src)
{
  if (src.fieldType() == SFBOOL)
  {
    d_value = static_cast<const VrmlSFBool &>(src).d_value;
    VrmlField::set(src);
    return true;
  }
  return false;
}


/* external value conversion API */

void VrmlSFBool::getExtValueType(extClassID* type, size_t* m, size_t* n, size_t* k) const
  { *type = extLOGICAL_CLASS; *m = 1; *n = 1; *k = 1; }

bool VrmlSFBool::getExtValue(extClassID type, void* values) const
{
  switch (type)
  {
    case extLOGICAL_CLASS :
      *((extLogical*) values) = d_value;
    break;

    case extCHAR_CLASS :
      *((const char **) values) = d_value ? "on" : "off";
    break;

    default :
      return(false);
    break;
  }

  return(true);
}

bool VrmlSFBool::setExtValue(extClassID type, size_t m, size_t n, size_t k, const void* values)
{
  if ((m!=1 || n!=1 || k!=1) && type!=extCHAR_CLASS)   // must be a scalar if it's not a string
    return(false);

  switch(type)
  {
    case extCHAR_CLASS :
    {
      const char* cval = (const char*) values;
      if (strcasecmp(cval, "FALSE")==0 || strcasecmp(cval, "off")==0 )
	d_value = false;
      else if (strcasecmp(cval, "TRUE")==0 || strcasecmp(cval, "on")==0 )
	d_value = true;
      else
        return(false);

      return(true);
    }
  }

  return(MLData2VRML(&d_value, type, m, n, values));
}


void VrmlSFBool::serialize(ostream& str) const
{
  StreamWriter(str).WriteBool(d_value);
};

bool VrmlSFBool::readSerialized(istream& str)
{
  d_value = StreamReader(str).ReadBool();
  return true;
};



/****************************/
/*** VrmlSingleFloatField ***/
/****************************/

/* external value conversion API */

void VrmlSingleFloatField::getExtValueType(extClassID* type, size_t* m, size_t* n, size_t* k) const
  { *type = extSINGLE_CLASS; *m = 1; *n = width; *k = 1; }

bool VrmlSingleFloatField::getExtValue(extClassID type, void* values) const
  { return(VRMLData2ML(values, type, 1, width, x)); }

bool VrmlSingleFloatField::setExtValue(extClassID type, size_t m, size_t n, size_t k, const void* values)
{
  if (m != 1 || n != width || k != 1)
    return(false);

  return(MLData2VRML(x, type, m, n, values));
}


void VrmlSingleFloatField::serialize(ostream& str) const
{
  for (size_t i=0; i<width; i++)
    StreamWriter(str).WriteFloat(x[i]);
};

bool VrmlSingleFloatField::readSerialized(istream& str)
{
  for (size_t i=0; i<width; i++)
    x[i] = StreamReader(str).ReadFloat();
  return true;
};



/*******************/
/*** VrmlSFColor ***/
/*******************/



// Conversion functions between RGB each in [0,1] and HSV with
// h in [0,360), s,v in [0,1]. From Foley, van Dam p615-616.

void VrmlSFColor::HSVtoRGB(const float hsv[3], float rgb[3])
{
    float h = hsv[0];
    if (hsv[1] == 0.0) {
        rgb[0] = rgb[1] = rgb[2] = hsv[2];
    } else {
        if (h >= 360.0) {
            h -= 360.0;
        }
        h /= 60.0;
        const float i = (float) floor(h);
        const float f = h - i;
        const float p = hsv[2] * (1.0F - hsv[1]);
        const float q = hsv[2] * (1.0F - hsv[1] * f);
        const float t = hsv[2] * (1.0F - hsv[1] * (1.0F - f));
        switch (static_cast<int>(i)) {
	    default:
	    case 0: rgb[0] = hsv[2]; rgb[1] = t; rgb[2] = p; break;
	    case 1: rgb[0] = q; rgb[1] = hsv[2]; rgb[2] = p; break;
	    case 2: rgb[0] = p; rgb[1] = hsv[2]; rgb[2] = t; break;
	    case 3: rgb[0] = p; rgb[1] = q; rgb[2] = hsv[2]; break;
	    case 4: rgb[0] = t; rgb[1] = p; rgb[2] = hsv[2]; break;
	    case 5: rgb[0] = hsv[2]; rgb[1] = p; rgb[2] = q; break;
	}
    }
}



void VrmlSFColor::RGBtoHSV(const float rgb[3], float hsv[3]) {
    const float maxrgb = *std::max_element(rgb, rgb + 3);
    const float minrgb = *std::min_element(rgb, rgb + 3);

    hsv[0] = 0.0;
    hsv[1] = (maxrgb > 0.0F) ? ((maxrgb - minrgb) / maxrgb) : 0.0F;
    hsv[2] = maxrgb;

    if (hsv[1] != 0.0) {
        const float rc = (maxrgb - rgb[0]) / (maxrgb - minrgb);
        const float gc = (maxrgb - rgb[1]) / (maxrgb - minrgb);
        const float bc = (maxrgb - rgb[2]) / (maxrgb - minrgb);

        if (rgb[0] == maxrgb) {
            hsv[0] = bc - gc;
        } else if (rgb[1] == maxrgb) {
            hsv[0] = 2 + rc - bc;
        } else {
            hsv[0] = 4 + gc - rc;
        }

        hsv[0] *= 60.0;
        if (hsv[0] < 0.0) {
            hsv[0] += 360.0;
        }
    }
}



void VrmlSFColor::setHSV(float h, float s, float v)
{
  const float hsv[3] = { h, s, v };
  HSVtoRGB(hsv, x);
}



void VrmlSFColor::getHSV(float hsv[3]) const
{
  RGBtoHSV(x, hsv);
}


/* external value conversion API */

void VrmlSFColor::getExtValueType(extClassID* type, size_t* m, size_t* n, size_t* k) const
  { *type = extSINGLE_CLASS; *m = 1; *n = 3; *k = 1; }

bool VrmlSFColor::getExtValue(extClassID type, void* values) const
  { return(VRMLData2ML(values, type, 1, 3, x)); }

bool VrmlSFColor::setExtValue(extClassID type, size_t m, size_t n, size_t k, const void* values)
{
  if (m != 1 || n != 3 || k != 1)
    return(false);

  return(MLData2VRML((ColorFloat*) x, type, m, n, values));
}



/*******************/
/*** VrmlSFImage ***/
/*******************/



VrmlSFImage::VrmlSFImage(): d_w(0), d_h(0), d_nc(0), d_pixels(0) {}

VrmlSFImage::VrmlSFImage(size_t width, size_t height, size_t components,
                         const unsigned char * pixels):
                       d_w(0L), d_h(0L), d_nc(0L), d_pixels(0L)
{
  reallocatePixels(width, height, components, pixels);
}

VrmlSFImage::VrmlSFImage(const VrmlSFImage& rhs): d_w(0), d_h(0), d_nc(0), d_pixels(0)
{
  reallocatePixels(rhs.d_w, rhs.d_h, rhs.d_nc, rhs.d_pixels);
}

VrmlSFImage::~VrmlSFImage()
{
  delete [] d_pixels;
}

VrmlSFImage & VrmlSFImage::operator=(const VrmlSFImage & rhs)
{
  if (this != &rhs)
  {
    reallocatePixels(rhs.d_w, rhs.d_h, rhs.d_nc, rhs.d_pixels);
    VrmlObject::operator=(rhs);
  }
  return *this;
}

ostream& VrmlSFImage::print(ostream& os, int indent) const
{
  os << d_w << " " << d_h << " " << d_nc << " \n";

  if (d_w*d_h*d_nc == 0)
    return os;

  unsigned char *p = d_pixels;
  std::string indentString(indent, ' ');
  for (size_t i=0; i<d_h; ++i)
  {
    os << indentString;
    for (size_t k=0; k<d_w; ++k)
    {
      unsigned int pixval = 0;
      for (size_t j=0; j<d_nc; ++j)
        pixval = (pixval << 8) | *p++;
      os << pixval << " ";
    }
    os << "\n";
  }
  return os;
}

VrmlField * VrmlSFImage::clone(VrmlNamespace* ns) const
{
  return new VrmlSFImage(*this);
}

bool VrmlSFImage::equals(const VrmlField* f) const
{
  if (f->fieldType() == SFIMAGE)
  {
    if (((VrmlSFImage*) f)->d_w == d_w
      && ((VrmlSFImage*) f)->d_h == d_h
      && ((VrmlSFImage*) f)->d_nc == d_nc)
    {
      size_t bytecnt = d_w*d_h*d_nc;
      for (size_t i=0; i<bytecnt; i++)
        if (((VrmlSFImage*) f)->d_pixels[i] != d_pixels[i]) return false;
      return true;
    }
  }
  return false;
}

VrmlField::VrmlFieldType VrmlSFImage::fieldType() const { return SFIMAGE; }

size_t VrmlSFImage::getWidth() const
{
  return this->d_w;
}

size_t VrmlSFImage::getHeight() const
{
  return this->d_h;
}

size_t VrmlSFImage::getComponents() const
{
  return this->d_nc;
}

const unsigned char * VrmlSFImage::getPixels() const
{
  return this->d_pixels;
}



void VrmlSFImage::set(size_t width, size_t height, size_t components,
                      const unsigned char * pixels)
{
  reallocatePixels(width, height, components, pixels);
}



bool VrmlSFImage::set(const VrmlField &src)
{
  if (src.fieldType() == SFIMAGE)
  {
    const VrmlSFImage& isrc = static_cast<const VrmlSFImage &>(src);
    reallocatePixels(isrc.d_w, isrc.d_h, isrc.d_nc, isrc.d_pixels);
    VrmlField::set(src);
    return true;
  }
  return false;
}

void VrmlSFImage::reallocatePixels(size_t width, size_t height, size_t components, const unsigned char* copySource)
{
  if (d_w!=width || d_h!=height || d_nc!=components)
  {
    delete[] d_pixels;
    d_pixels = 0;
    d_w = width;
    d_h = height;
    d_nc = components;

    const size_t pixelsSize = d_w*d_h*d_nc;
    if (pixelsSize)
    {
      d_pixels = new unsigned char[pixelsSize];
      if (copySource)
      {
	std::copy(copySource, copySource + pixelsSize, this->d_pixels);
      }
      else
      {
        memset(d_pixels, 0, pixelsSize);
      }
    }
  }
}

void VrmlSFImage::getExtValueType(extClassID* type, size_t* m, size_t* n, size_t* k) const
{
  *type = extUINT8_CLASS; *m = d_h; *n = d_w; *k = d_nc;
}

bool VrmlSFImage::getExtValue(extClassID type, void* values) const
{
  // we need to swap indexes for correct result
  return(VRMLSFImage2MLData(static_cast<unsigned char*> (values), type, d_h, d_w, d_nc, d_pixels));
}

bool VrmlSFImage::setExtValue(extClassID type, size_t m, size_t n, size_t k, const void* values)
{
  reallocatePixels(n, m, k, 0);
  return(MLData2VRMLSFImage(d_pixels, type, m, n, k, values));
}



/*******************/
/*** VrmlSFInt32 ***/
/*******************/



VrmlSFInt32::VrmlSFInt32(long value)
  : d_value(value) { }



VrmlField *VrmlSFInt32::clone(VrmlNamespace* ns) const
  { return new VrmlSFInt32(d_value); }



VrmlField::VrmlFieldType VrmlSFInt32::fieldType() const
  { return SFINT32; }



bool VrmlSFInt32::equals(const VrmlField* f) const
{
  if (f->fieldType() == SFINT32)
  {
    if (((VrmlSFInt32*) f)->d_value == d_value) return true;
  }
  return false;
}



long VrmlSFInt32::get() const
  { return d_value; }



void VrmlSFInt32::set(long value)
  { this->d_value = value; }



bool VrmlSFInt32::set(const VrmlField &src)
{
  if (src.fieldType() == SFINT32)
  {
    d_value = static_cast<const VrmlSFInt32 &>(src).d_value;
    VrmlField::set(src);
    return true;
  }
  return false;
}


/* external value conversion API */

void VrmlSFInt32::getExtValueType(extClassID* type, size_t* m, size_t* n, size_t* k) const
  { *type = extINT32_CLASS; *m = 1; *n = 1; *k = 1; }

bool VrmlSFInt32::getExtValue(extClassID type, void* values) const
  { return(VRMLData2ML(values, type, 1, 1, &d_value)); }

bool VrmlSFInt32::setExtValue(extClassID type, size_t m, size_t n, size_t k, const void* values)
{
  if (m != 1 || n != 1 || k != 1)
    return(false);

  return(MLData2VRMLRound(&d_value, type, m, n, values));
}


void VrmlSFInt32::serialize(ostream& str) const
{
  StreamWriter(str).WriteInt32(d_value);
};

bool VrmlSFInt32::readSerialized(istream& str)
{
  d_value = StreamReader(str).ReadInt32();
  return true;
};





/******************/
/*** VrmlSFNode ***/
/******************/



VrmlSFNode::VrmlSFNode(VrmlNode * node)
  : d_value(node) { if (d_value) d_value->reference(); }



VrmlSFNode::VrmlSFNode(const VrmlSFNode & rhs)
  : d_value(rhs.d_value) { if (d_value) d_value->reference(); }



VrmlSFNode::~VrmlSFNode()
  { if (d_value) d_value->dereference(); }



VrmlSFNode& VrmlSFNode::operator=(const VrmlSFNode& rhs)
{
  if (this == &rhs) return *this;
  if (d_value) d_value->dereference();
  d_value = rhs.d_value;
  if (d_value) d_value->reference();
  VrmlObject::operator=(rhs);
  return *this;
}


ostream& VrmlSFNode::toChain(ostream& os, bool editFlag) const
{
  if (d_value) d_value->toChain(os, editFlag);
  return os;
}



ostream& VrmlSFNode::print(ostream& os, int indent) const
{
  if (d_value) d_value->print(os, indent, !d_value->getComment().empty()); else os << "NULL";
  return os;
}

bool VrmlSFNode::hasChildren() const
{
	return (d_value != NULL);
}



VrmlField* VrmlSFNode::clone(VrmlNamespace* ns) const
{
  if (ns==0)
  {
    // create reference instead of a real copy
    return new VrmlSFNode(d_value);
  }
  else
  {
    VrmlNode* newNode = 0;
    if (d_value!=0)
    {
      newNode = d_value->clone(ns);
    }
    return new VrmlSFNode(newNode);
  }
}



VrmlField::VrmlFieldType VrmlSFNode::fieldType() const
  { return SFNODE; }



bool VrmlSFNode::equals(const VrmlField* f) const
{
  if (f->fieldType() == SFNODE)
  {
    if (((VrmlSFNode*) f)->d_value == d_value) return true;
  }
  return false;
}



VrmlNode * VrmlSFNode::get() const
  { return d_value; }



void VrmlSFNode::set(VrmlNode * node)
{
  if (d_value) d_value->dereference();
  if ((d_value = node) != 0) d_value->reference();
}



bool VrmlSFNode::set(const VrmlField &src)
{
  if (src.fieldType() == SFNODE)
  {
    if (&src == this) return true;            // be robust
    d_value->dereference();
    d_value = static_cast<const VrmlSFNode &>(src).d_value->reference();
    VrmlField::set(src);
    return true;
  }
  return false;
}

void VrmlSFNode::getExtValueType(extClassID* type, size_t* m, size_t* n, size_t* k) const
{
  *type = extCHAR_CLASS;
  *m = 1;
  *n = 1;

  // zero value indicates that strings should be treated as nodes
  *k = 0;
}

bool VrmlSFNode::getExtValue(extClassID type, void* values) const
{
  *((const char**) (values)) = (d_value) ? (d_value->name()) : "";
  return true;
}

bool VrmlSFNode::setExtValue(extClassID type, size_t m, size_t n, size_t k, const void* values)
{
  return false;
}




/**********************/
/*** VrmlSFRotation ***/
/**********************/



VrmlSFRotation::VrmlSFRotation(const VrmlSFVec3f & fromVector,
                               const VrmlSFVec3f & toVector) :
VrmlSingleFloatField(4)
{
  setAxis(fromVector.cross(toVector));
  x[3] = (float) ( acos(fromVector.dot(toVector) / (fromVector.length() * toVector.length())) );
}



static void multQuat(const float quat1[4], const float quat2[4], float result[4]) {
    result[3] = quat1[3]*quat2[3] - quat1[0]*quat2[0]
            - quat1[1]*quat2[1] - quat1[2]*quat2[2];

    result[0] = quat1[3]*quat2[0] + quat1[0]*quat2[3]
            + quat1[1]*quat2[2] - quat1[2]*quat2[1];

    result[1] = quat1[3]*quat2[1] + quat1[1]*quat2[3]
            + quat1[2]*quat2[0] - quat1[0]*quat2[2];

    result[2] = quat1[3]*quat2[2] + quat1[2]*quat2[3]
            + quat1[0]*quat2[1] - quat1[1]*quat2[0];
}



const VrmlSFRotation VrmlSFRotation::multiply(
  const VrmlSFRotation & rot) const
{
    // convert to quaternions
    float quatUS[4], quatVec[4];
    this->toQuaternion(quatUS);
    rot.toQuaternion(quatVec);

    // multiply quaternions
    float resultQuat[4];
    multQuat(quatUS, quatVec, resultQuat);

    // now convert back to axis/angle
    VrmlSFRotation result;
    result.fromQuaternion(resultQuat);
    return result;
}



void VrmlSFRotation::toQuaternion(float theQuat[4]) const
{
  float sintd2 = (float) sin(x[3] * 0.5);
  float len = (float) sqrt((x[0]*x[0]) + (x[1]*x[1]) + (x[2]*x[2]));
  float f = sintd2 / len;
  theQuat[3] = (float) cos(x[3] * 0.5);
  theQuat[0] = x[0] * f;
  theQuat[1] = x[1] * f;
  theQuat[2] = x[2] * f;
}



void VrmlSFRotation::fromQuaternion(const float theQuat[4])
{
  double sina2 = sqrt(theQuat[0]*theQuat[0] + theQuat[1]*theQuat[1]
                  + theQuat[2]*theQuat[2]);
  const double angle = (2.0*atan2(sina2, (double) theQuat[3]));
  if (sina2 >= 1e-8)
  {
    sina2 = 1.0/sina2;
    x[0] = (float) (theQuat[0] * sina2);
    x[1] = (float) (theQuat[1] * sina2);
    x[2] = (float) (theQuat[2] * sina2);
    x[3] = (float) angle;
  }
  else
  {
    x[0] = 0.0;
    x[1] = 1.0;
    x[2] = 0.0;
    x[3] = 0.0;
  }
}



const VrmlSFVec3f VrmlSFRotation::multVec(const VrmlSFVec3f & vec) const
  { return VrmlSFVec3f(); }



const VrmlSFRotation VrmlSFRotation::slerp(const VrmlSFRotation & destRotation,
                                           float t) const
{
    float fromQuat[4], toQuat[4];
    this->toQuaternion(fromQuat);
    destRotation.toQuaternion(toQuat);

    //
    // Calculate cosine.
    //
    double cosom = std::inner_product(fromQuat, fromQuat + 4, toQuat, 0.0);

    //
    // Adjust signs (if necessary).
    //
    float to1[4];
    if (cosom < 0.0) {
        cosom = -cosom;
        to1[0] = -toQuat[0];
        to1[1] = -toQuat[1];
        to1[2] = -toQuat[2];
        to1[3] = -toQuat[3];
    } else {
        std::copy(toQuat, toQuat + 4, to1);
    }

    //
    // Calculate coefficients.
    //
    double scale0, scale1;
    if ((1.0 - cosom) > FPTOLERANCE) {
        const double omega = acos(cosom);
        const double sinom = sin(omega);
        scale0 = sin((1.0 - t) * omega) / sinom;
        scale1 = sin(t * omega) / sinom;
    } else {
        //
        // "From" and "to" quaternions are very close, so do linear
        // interpolation.
        //
        scale0 = 1.0 - t;
        scale1 = t;
    }

    //
    // Calculate the final values.
    //
    float resultQuat[4];
    resultQuat[0] = (float) ( (scale0 * fromQuat[0]) + (scale1 * to1[0]) );
    resultQuat[1] = (float) ( (scale0 * fromQuat[1]) + (scale1 * to1[1]) );
    resultQuat[2] = (float) ( (scale0 * fromQuat[2]) + (scale1 * to1[2]) );
    resultQuat[3] = (float) ( (scale0 * fromQuat[3]) + (scale1 * to1[3]) );

    VrmlSFRotation result;
    result.fromQuaternion(resultQuat);

    return result;
}

bool VrmlSFRotation::getExtValueInWorldCoords(extClassID type, void* values, const Matrix4x4* tm) const
{
  const Matrix4x4& rotMat(*tm);

  Matrix4x4 thisMat;
  thisMat.setToRotation(getAngle(), Vector3(getX(), getY(), getZ()));

  Vector3 dir(0, 0, -1);
  Vector3 up(0, 1, 0);
  dir = thisMat*dir;
  up = thisMat*up;
  dir = rotMat*dir;
  up = rotMat*up;

  thisMat.setToViewing(Point3(0, 0, 0),
          dir,
          up);

  Quaternion thisQuat(thisMat);
  Vector3 axis(0, 0, 0);
  double angle = 0;
  thisQuat.toAxisAngle(axis, angle);
  VrmlSFRotation thisRot(VrmlSFVec3f(axis.getX(), axis.getY(), axis.getZ()), angle);
  float xret[] = { thisRot.getX(), thisRot.getY(), thisRot.getZ(), thisRot.getAngle() };

  // just convert supplied data
  return(VRMLData2ML(values, type, 1, width, xret));
}



/********************/
/*** VrmlSFString ***/
/********************/



VrmlSFString::VrmlSFString(const char *s)
{ if (s) { d_s = new char[strlen(s)+1]; strcpy(d_s,s); } else d_s = 0; }



VrmlSFString::VrmlSFString(const VrmlSFString &sfs)
{
  const char *s = sfs.get();
  if (s) { d_s = new char[strlen(s)+1]; strcpy(d_s,s); } else d_s = 0;
}



VrmlSFString::~VrmlSFString() { if (d_s) delete [] d_s; }



const char * VrmlSFString::get() const {
    return this->d_s;
}



void VrmlSFString::set(const char *s)
{
  if (d_s) delete [] d_s;
  if (s) { d_s = new char[strlen(s)+1]; strcpy(d_s,s); }
  else d_s = 0;
}



VrmlSFString& VrmlSFString::operator=(const VrmlSFString& rhs)
{
  if (this != &rhs)
  {
    set(rhs.d_s);
    VrmlObject::operator=(rhs);
  }
  return *this;
}



ostream& VrmlSFString::print(ostream& os, int indent) const
{
  if (d_s) os << '\"' << d_s << '\"';
  else os << "\"\"";
  return os;
}



VrmlField *VrmlSFString::clone(VrmlNamespace* ns) const { return new VrmlSFString(d_s); }



VrmlField::VrmlFieldType VrmlSFString::fieldType() const { return SFSTRING; }



bool VrmlSFString::equals(const VrmlField* f) const
{
  if (f->fieldType() == SFSTRING)
  {
    if (((VrmlSFString*) f)->d_s != NULL
      && d_s != NULL
      && strcmp(((VrmlSFString*) f)->d_s, d_s) == 0) return true;

    if (((VrmlSFString*) f)->d_s == NULL
      && d_s == NULL) return true;
  }
  return false;
}



bool VrmlSFString::set(const VrmlField &src)
{
  if (&src == this) return true;          // be robust
  if (src.fieldType() == SFSTRING)
  {
    set(static_cast<const VrmlSFString &>(src).d_s);
    VrmlField::set(src);
    return true;
  }
  return false;
}


/* external value conversion API */

void VrmlSFString::getExtValueType(extClassID* type, size_t* m, size_t* n, size_t* k) const
  { *type = extCHAR_CLASS; *m = 1; *n = 1; *k = 1; }

bool VrmlSFString::getExtValue(extClassID type, void* values) const
{
  *((const char **) values) = d_s;
  return(true);
}

bool VrmlSFString::setExtValue(extClassID type, size_t m, size_t n, size_t k, const void* values)
{
  switch(type)
  {
    case extPOINTER_CLASS :
      if (m!=1 || n!=1)
        return(false);
      values = *((void**) values);
      // and fallthrough

    case extCHAR_CLASS :
      set((const char*) values);
    break;

    default : return(false);
  }

  return(true);
}


void VrmlSFString::serialize(ostream& str) const
{
  StreamWriter(str).WriteString(d_s);
};

bool VrmlSFString::readSerialized(istream& str)
{
  delete[] d_s;
  d_s = StreamReader(str).ReadString();
  return true;
};





/******************/
/*** VrmlSFTime ***/
/******************/



VrmlSFTime::VrmlSFTime(double value) : d_value(value) {}

VrmlField * VrmlSFTime::clone(VrmlNamespace* ns) const {
    return new VrmlSFTime(*this);
}

VrmlField::VrmlFieldType VrmlSFTime::fieldType() const { return SFTIME; }

bool VrmlSFTime::equals(const VrmlField* f) const
{
  if (f->fieldType() == SFTIME)
  {
    if (((VrmlSFTime*) f)->d_value == d_value) return true;
  }
  return false;
}

double VrmlSFTime::get() const {
    return this->d_value;
}

void VrmlSFTime::set(double value) {
    this->d_value = value;
}


bool VrmlSFTime::set(const VrmlField &src)
{
  if (src.fieldType() == SFTIME)
  {
    d_value = static_cast<const VrmlSFTime &>(src).d_value;
    VrmlField::set(src);
    return true;
  }
  return false;
}



/* external value conversion API */

void VrmlSFTime::getExtValueType(extClassID* type, size_t* m, size_t* n, size_t* k) const
  { *type = extDOUBLE_CLASS; *m = 1; *n = 1; *k = 1;}

bool VrmlSFTime::getExtValue(extClassID type, void* values) const
  { return(VRMLData2ML(values, type, 1, 1, &d_value)); }

bool VrmlSFTime::setExtValue(extClassID type, size_t m, size_t n, size_t k, const void* values)
{
  if (m != 1 || n != 1 || k != 1)
    return(false);

  return(MLData2VRML(&d_value, type, m, n, values));
}


void VrmlSFTime::serialize(ostream& str) const
{
  StreamWriter(str).WriteDouble(d_value);
};

bool VrmlSFTime::readSerialized(istream& str)
{
  d_value = StreamReader(str).ReadDouble();
  return true;
};





/*******************/
/*** VrmlSFVec2f ***/
/*******************/



double VrmlSFVec2f::length() const
{
  return sqrt(x[0]*x[0] + x[1]*x[1]);
}



const VrmlSFVec2f VrmlSFVec2f::normalize() const
{
  float len = (float) this->length();
  if (FPZERO(len)) {
    return *this;
  }
  return VrmlSFVec2f(x[0]/len, x[1]/len);
}



/*******************/
/*** VrmlSFVec3f ***/
/*******************/



const VrmlSFVec3f VrmlSFVec3f::cross(const VrmlSFVec3f & vec) const
{
  VrmlSFVec3f result;
  result.x[0] = (x[1] * vec.x[2]) - (x[2] * vec.x[1]);
  result.x[1] = (x[2] * vec.x[0]) - (x[0] * vec.x[2]);
  result.x[2] = (x[0] * vec.x[1]) - (x[1] * vec.x[0]);
  return result;
}



double VrmlSFVec3f::length() const
{
  const double len = sqrt(x[0]*x[0] + x[1]*x[1] + x[2]*x[2]);
  return FPZERO(len) ? 0.0 : len;
}



const VrmlSFVec3f VrmlSFVec3f::normalize() const
{
  float len = (float) this->length();
  VrmlSFVec3f result(*this);
  if (!FPZERO(len))
  {
    result.x[0] /= len;
    result.x[1] /= len;
    result.x[2] /= len;
  }
  return result;
}

bool VrmlSFVec3f::getExtValueInWorldCoords(extClassID type, void* values, const Matrix4x4* tm) const
{
  Point3 oldpoint(x[0], x[1], x[2]);
  Point3 newpoint = (*tm)*oldpoint;
  float transformed[] = {newpoint.getX(), newpoint.getY(), newpoint.getZ()};
  return(VRMLData2ML(values, type, 1, width, transformed));
}



/***************************/
/*** VrmlMultiFloatField ***/
/***************************/

ostream& VrmlMultiFloatField::print(ostream& os, int indent) const
{
  return mfprint(os, get(), getLength(), width, indent);
}


/* external value conversion API */

void VrmlMultiFloatField::getExtValueType(extClassID* type, size_t* m, size_t* n, size_t* k) const
  { *type = extSINGLE_CLASS; *m = getLength(); *n = width; *k = 1; }

bool VrmlMultiFloatField::getExtValue(extClassID type, void* values) const
  { return(VRMLData2ML(values, type, getLength(), width, data->getValues())); }

bool VrmlMultiFloatField::setExtValue(extClassID type, size_t m, size_t n, size_t k, const void* values)
{
  if (!((n==width || (n==0 && m==0)) && k==1))  // accept also 0x0 empty
    return false;

  // create new data array if size doesn't match
  if (m != getLength())
  {
    RefCntArray<float>* newData = RefCntArray<float>::create(m*width, data, 0.0);
    // free the old data array
    data->decRef();
    data = newData;
  }

  // create an unique copy if data shared
  else
    data = data->makeUnique();

  return(MLData2VRML(data->getValues(), type, m, n, values));
}


void VrmlMultiFloatField::serialize(ostream& str) const
{
  size_t len = getLength();
  StreamWriter(str).WriteInt32(static_cast<int>(len));

  const float* x = data->getValues();
  for (size_t i=0; i<len*width; i++)
    StreamWriter(str).WriteFloat(x[i]);
};

bool VrmlMultiFloatField::readSerialized(istream& str)
{
  size_t len = StreamReader(str).ReadInt32();

  // create new data array
  RefCntArray<float>* newData = RefCntArray<float>::create(len*width, data, 0.0);
  // free the old data array
  data->decRef();
  data = newData;

  // fill the new array with data
  float* x = data->getValues();
  for (size_t i=0; i<len*width; i++)
    x[i] = (float) StreamReader(str).ReadFloat();
  return true;
};




/*******************/
/*** VrmlMFColor ***/
/*******************/



const float * VrmlMFColor::operator[](size_t index) const
  { return data->getValues() + 3*index; }



void VrmlMFColor::setElement(size_t index, float r, float g, float b)
{
  if (index >= getLength()) setLength(index+1);
  data = data->makeUnique();
  data->setValue(3*index, r);
  data->setValue(3*index+1, g);
  data->setValue(3*index+2, b);
}



/* external value conversion API */

void VrmlMFColor::getExtValueType(extClassID* type, size_t* m, size_t* n, size_t* k) const
  { *type = extSINGLE_CLASS; *m = getLength(); *n = 3; *k = 1; }

bool VrmlMFColor::getExtValue(extClassID type, void* values) const
  { return(VRMLData2ML(values, type, getLength(), 3, data->getValues())); }

bool VrmlMFColor::setExtValue(extClassID type, size_t m, size_t n, size_t k, const void* values)
{
  if (!((n==3 || (n==0 && m==0)) && k==1))  // accept also 0x0 empty
    return false;

  // create new data array if size doesn't match
  if (m != getLength())
  {
    RefCntArray<float>* newData = RefCntArray<float>::create(m*3, data, 0.0);
    // free the old data array
    data->decRef();
    data = newData;
  }

  // create an unique copy if data shared
  else
    data = data->makeUnique();

  return(MLData2VRML((ColorFloat*) data->getValues(), type, m, n, values));
}




/*******************/
/*** VrmlMFFloat ***/
/*******************/



const float VrmlMFFloat::operator[](size_t index) const
{
  return data->getValue(index);
}



void VrmlMFFloat::setElement(size_t index, float x)
{
  if (index >= getLength()) setLength(index+1);
  data = data->makeUnique();
  data->setValue(index, x);
}



ostream& VrmlMFFloat::print(ostream& os, int indent) const
{
  const size_t n = this->getLength();
  const float* c = get();

  if (n == 1) os << *c;
  else
  {
    const size_t elementsPerLine = 80;
    const bool multiLine = n>elementsPerLine;
    const char* lineSep = (multiLine) ? "\n" : "";

    std::string indentString = (multiLine) ? std::string(indent, ' ') : std::string();
    std::string indentIncrementedString = (multiLine) ? std::string(indent+VrmlNode::INDENT_INCREMENT, ' ') : std::string();

    os << "[" << lineSep << indentIncrementedString;
    for (size_t i=0; i<n; ++i, ++c)
    {
      os << *c;

      if (i < n-1)
        os << ", ";

      if ((i < n-1) && ((i+1) % elementsPerLine == 0))
        os <<  "\n" << indentIncrementedString;
    }
    os << lineSep << indentString << "]";
  }

  return os;
}


/*******************/
/*** VrmlMFInt32 ***/
/*******************/



VrmlMFInt32::VrmlMFInt32(size_t length, const long * numbers)
{
  data = RefCntArray<long>::create(length, numbers);
}



VrmlMFInt32::VrmlMFInt32(const VrmlMFInt32 &src)
{
  src.data->incRef();
  data = src.data;
}



VrmlMFInt32::~VrmlMFInt32()
{
  data->decRef();
}



bool VrmlMFInt32::set(const VrmlField &src)
{
  if (src.fieldType() == MFINT32)
  {
    *this = (const VrmlMFInt32 &) src;
    return true;
  }
  return false;
}



VrmlMFInt32 &VrmlMFInt32::operator=(const VrmlMFInt32 &src)
{
  if (&src == this) return *this;        // be robust
  data->decRef();
  src.data->incRef();
  data = src.data;
  VrmlObject::operator=(src);
  return *this;
}



long VrmlMFInt32::operator[](size_t index) const
{
  return data->getValue(index);
}



const long * VrmlMFInt32::get() const
{
  return data->getValues();
}



void VrmlMFInt32::set(size_t length, const long* numbers)
{
  RefCntArray<long>* newData = RefCntArray<long>::create(length, numbers);
  data->decRef();
  data = newData;
}



size_t VrmlMFInt32::getLength() const
{
  return data->getLength();
}



void VrmlMFInt32::setLength(size_t length)
{
  RefCntArray<long>* newData = RefCntArray<long>::create(length, data, 0);
  data->decRef();
  data = newData;
}



void VrmlMFInt32::setElement(size_t index, long n)
{
  if (index >= getLength()) setLength(index+1);
  data = data->makeUnique();
  data->setValue(index, n);
}



VrmlField * VrmlMFInt32::clone(VrmlNamespace* ns) const
{
  return new VrmlMFInt32(*this);
}



VrmlField::VrmlFieldType VrmlMFInt32::fieldType() const
{
  return MFINT32;
}



ostream& VrmlMFInt32::print(ostream& os, int indent) const
{
  size_t n = this->getLength();
  const long * c = get();

  if (n == 1) os << *c;
  else
  {
    std::string indentString(indent, ' ');
    std::string indentIncrementedString(indent+VrmlNode::INDENT_INCREMENT, ' ');
    os << "[\n" << indentIncrementedString;
    for (size_t i=0; i<n; ++i, ++c)
    {
      const char* sep = ((i < n-1) ? ", " : " ");
      (((*c == -1) && (i < n-1))   ? (os << *c << sep << "\n" << indentIncrementedString) : (os << *c << sep));
    }
    os << "\n" << indentString << "]";
  }

  return os;
}



bool VrmlMFInt32::equals(const VrmlField* f) const
{
  if (f->fieldType() != MFINT32) return false;
  return data->equals(((VrmlMFInt32*) f)->data);
}



/* external value conversion API */

void VrmlMFInt32::getExtValueType(extClassID* type, size_t* m, size_t* n, size_t* k) const
  { *type = extINT32_CLASS; *m = getLength(); *n = 1; *k = 1; }

bool VrmlMFInt32::getExtValue(extClassID type, void* values) const
  { return(VRMLData2ML(values, type, getLength(), 1, data->getValues())); }

bool VrmlMFInt32::setExtValue(extClassID type, size_t m, size_t n, size_t k, const void* values)
{
  if (!((n==1 || (n==0 && m==0)) && k==1))  // accept also 0x0 empty
    return false;

  // create new data array if size doesn't match
  if (m != getLength())
  {
    // create new data array
    RefCntArray<long>* newData = RefCntArray<long>::create(m, data, 0);
    // free the old data array
    data->decRef();
    data = newData;
  }

  // create an unique copy if data shared
  else
    data = data->makeUnique();

  return(MLData2VRMLRound(data->getValues(), type, m, n, values));
}



void VrmlMFInt32::serialize(ostream& str) const
{
  size_t len = getLength();
  StreamWriter(str).WriteInt32(static_cast<int>(len));

  const long* x = data->getValues();
  for (size_t i=0; i<len; i++)
    StreamWriter(str).WriteInt32(x[i]);
};

bool VrmlMFInt32::readSerialized(istream& str)
{
  int len = StreamReader(str).ReadInt32();

  // create new data array
  RefCntArray<long>* newData = RefCntArray<long>::create(len, data, 0);
  // free the old data array
  data->decRef();
  data = newData;

  // fill the new array with data
  long* x = data->getValues();
  for (int i=0; i<len; i++)
    x[i] = StreamReader(str).ReadInt32();
  return true;
};




/******************/
/*** VrmlMFNode ***/
/******************/

VrmlMFNode::VrmlMFNode(size_t length, VrmlNode * const * nodes):
        d_v(new VrmlNode *[length]), d_allocated(length), d_size(length) {
    if (nodes) {
        for (size_t i(0); i < length; ++i) {
            d_v[i] = nodes[i] ? nodes[i]->reference() : 0;
        }
    } else {
        std::fill(d_v, d_v + length, static_cast<VrmlNode *>(0));
    }
}

VrmlMFNode::VrmlMFNode(VrmlMFNode const & mfnode):
        d_v(new VrmlNode * [mfnode.d_size]), d_allocated(mfnode.d_size),
        d_size(mfnode.d_size) {
    for (size_t i(0); i < mfnode.d_size; ++i) {
        d_v[i] = mfnode.d_v[i] ? (mfnode.d_v[i]->reference()) : 0;
    }
}

VrmlMFNode::~VrmlMFNode()
{
    for (size_t i(0); i < d_size; ++i) {
        if (d_v[i]) d_v[i]->dereference();
    }
    delete [] d_v;
}

bool VrmlMFNode::set(const VrmlField &src)
{
  if (src.fieldType() == MFNODE)
  {
    *this = (const VrmlMFNode &) src;
    return true;
  }
  return false;
}


VrmlMFNode & VrmlMFNode::operator=(const VrmlMFNode & mfnode) {
    if (this != &mfnode) {
        size_t i(0);
        for (; i < d_size; ++i) {
            if (d_v[i]) {
                d_v[i]->dereference();
            }
        }

        if (d_allocated < mfnode.d_size) {
            delete [] d_v;
            d_size = d_allocated = 0;
            d_v = 0;
            d_v = new VrmlNode*[mfnode.d_size];
            d_allocated = mfnode.d_size;
        }

        d_size = mfnode.d_size;

        for (i = 0; i < d_size; ++i) {
            d_v[i] = mfnode.d_v[i] ? mfnode.d_v[i]->reference() : 0;
        }

	VrmlObject::operator=(mfnode);
    }

    return *this;
}

VrmlNode * VrmlMFNode::operator[](size_t index) const {
    return this->d_v[index];
}

VrmlNode * & VrmlMFNode::operator[](size_t index) {
    return this->d_v[index];
}

size_t VrmlMFNode::getLength() const {
    return this->d_size;
}

void VrmlMFNode::setLength(size_t length) {
    if (length < this->d_size) {
        for (size_t i = length; i < this->d_size; ++i) {
            this->d_v[i]->dereference();
            this->d_v[i] = 0;
        }
    } else if (length > this->d_allocated) {
        this->realloc(length);
    }
    for (size_t i = this->d_size; i < length; ++i) {
        this->d_v[i] = 0;
    }
    this->d_size = length;
}

void VrmlMFNode::realloc(size_t newSize) {
    VrmlNode * * newNodes = new VrmlNode *[newSize];
    std::copy(this->d_v, this->d_v + this->d_size, newNodes);
    delete [] this->d_v;
    this->d_v = newNodes;
    this->d_allocated = newSize;
}

bool VrmlMFNode::exists(const VrmlNode & node) const {
    for (size_t i(0); i < this->d_size; ++i) {
        if (this->d_v[i] == &node) {
            return true;
        }
    }

    return false;
}

bool VrmlMFNode::addNode(VrmlNode* node, int where)
{
  if (node)
  {
    if (where != -1 && where >= 0 && where < this->d_size)
    {
      if (this->d_allocated < d_size + 1)
	this->realloc(this->d_allocated + VrmlMFNode::REALLOC_SIZE);

      for (size_t j = this->d_size; j>where; --j)
	this->d_v[j] = this->d_v[j-1];

      this->d_v[where] = node->reference();
      this->d_size++;
    }
    else
    {
      if (this->d_allocated < d_size + 1)
      	this->realloc(this->d_allocated + VrmlMFNode::REALLOC_SIZE);

      this->d_v[this->d_size++] = node->reference();

      return true;
    }
  }

  return false;
}

bool VrmlMFNode::removeNode(VrmlNode * node, int option) {
    bool result = false;
    if (node)
    {
      for (size_t i = 0; i < this->d_size; ++i)
      {
	if (this->d_v[i] == node)
	{
	  d_v[i]->dereference();

	  for (size_t j = i; j < this->d_size - 1; ++j) {
	    this->d_v[j] = this->d_v[j + 1];
	  }
	  --this->d_size;
	  --i;
	  result |= true;
	  if (option == REMOVE_FIRST)
	    return result;
	}
      }
    }
    return result;
}

void VrmlMFNode::removeRange(int startPosition, int count)
{
  for (size_t i = startPosition, j = 0; i < this->d_size && j<count; ++i, ++j)
  {
    d_v[i]->dereference();

    for (size_t j = i; j < this->d_size - 1; ++j)
      this->d_v[j] = this->d_v[j + 1];

    --this->d_size;
    --i;
  }
}


VrmlField *VrmlMFNode::clone(VrmlNamespace* ns) const
{
  if (ns==0)
  {
    // create reference instead of a real copy
    return new VrmlMFNode(*this);
  }
  else
  {
    VrmlMFNode* clonedMFNode = new VrmlMFNode();
    for (size_t i=0; i<getLength(); ++i)
    {
      VrmlNode* cloned = (d_v[i])->clone(ns);
      clonedMFNode->addNode(cloned);
    }

    return(clonedMFNode);
  }
}

VrmlField::VrmlFieldType VrmlMFNode::fieldType() const { return MFNODE; }

ostream& VrmlMFNode::print(ostream& os, int indent) const
{
  size_t n = getLength();
  if (n == 0)
  {
    os << "[]" << std::endl;
  }
  else if (n==1)
  {
    d_v[0]->print(os, indent, !(d_v[0]->getComment().empty()));
  }
  else
  {
    os << '[';
    for (size_t i=0; i<n; ++i)
      d_v[i]->print(os, indent, true);

    os << std::endl;
    for (int j=0; j<indent-VrmlNode::INDENT_INCREMENT; j++) os << " ";     // indent
    os << ']';
  }

  return os;
}

ostream& VrmlMFNode::toChain(ostream& os, bool editFlag) const
{
  size_t n = getLength();
  for (size_t i=0; i<n; ++i)
  {
    d_v[i]->toChain(os, editFlag);
  }

  return os;
}

bool VrmlMFNode::hasChildren() const
{
	return (getLength() > 0);
}



ostream &VrmlMFNode::printUnbracketed(ostream &os, int indent) const
{
  size_t n = getLength();

  for (size_t i=0; i<n; ++i)
  {
    d_v[i]->print(os, indent, true);
  }
  return os;
}

bool VrmlMFNode::equals(const VrmlField* f) const
{
  if (f->fieldType() == MFNODE)
  {
    if (((VrmlMFNode*) f)->d_size == d_size)
    {
      for (size_t i=0; i<d_size; i++)
        if (((VrmlMFNode*) f)->d_v[i] != d_v[i]) return false;
      return true;
    }
  }
  return false;
}

void VrmlMFNode::getExtValueType(extClassID* type, size_t* m, size_t* n, size_t* k) const
{
  *type = extPOINTER_CLASS;
  *m = getLength();
  *n = 1;

  // zero value indicates that strings should be treated as nodes
  *k = 0;
}

bool VrmlMFNode::getExtValue(extClassID type, void* values) const
{
  d_names.clear();
  for (size_t i=0; i<d_size; ++i)
    d_names.push_back(d_v[i]->name());

  *((const char***) values) = (!d_names.empty()) ? &(d_names.front()) : 0;
  return true;
}

bool VrmlMFNode::setExtValue(extClassID type, size_t m, size_t n, size_t k, const void* values)
{
  return false;
}





/**********************/
/*** VrmlMFRotation ***/
/**********************/



const float * VrmlMFRotation::operator[](size_t index) const
{
  return data->getValues() + 4*index;
}



float* VrmlMFRotation::operator[](size_t index)
{
  return data->getValues() + 4*index;
}



void VrmlMFRotation::setElement(size_t index, float x, float y, float z,
                                float a)
{
  if (index >= getLength()) setLength(index+1);
  data = data->makeUnique();
  data->setValue(4*index, x);
  data->setValue(4*index+1, y);
  data->setValue(4*index+2, z);
  data->setValue(4*index+3, a);
}



/********************/
/*** VrmlMFString ***/
/********************/



VrmlMFString::VrmlMFString(size_t n, char const * const * v)
  : d_v(n ? new char *[n] : 0), d_allocated(n), d_size(n)
{
    if (v) {
        for (size_t i(0); i < n; ++i, ++v) {
            if (*v) {
                char * const s = new char[strlen(*v) + 1];
                strcpy(s, *v);
                d_v[i] = s;
            } else {
                d_v[i] = 0;
            }
        }
    } else {
        std::fill(d_v, d_v + n, static_cast<char *>(0));
    }
}


VrmlMFString::VrmlMFString(VrmlMFString const & rhs)
  : d_v(new char *[rhs.d_size]), d_allocated(rhs.d_size), d_size(rhs.d_size)
{
    for (size_t i(0); i < rhs.d_size; ++i) {
        if (rhs.d_v[i]) {
            char * const s = new char[strlen(rhs.d_v[i])+1];
            strcpy(s, rhs.d_v[i]);
            d_v[i] = s;
        } else {
            d_v[i] = 0;
        }
    }
}

VrmlMFString::~VrmlMFString()
{
    for (size_t i(0); i < d_size; ++i) {
        delete [] d_v[i];
    }
    delete [] d_v;
}

void VrmlMFString::set(size_t n, char const * const v[])
{
    for (size_t i = 0; i < d_size; ++i) {
        delete [] d_v[i];
    }

    if (d_allocated < n) {
        delete [] d_v;
        d_v = 0;
        d_allocated = d_size = 0;
        d_v = new char *[n];
        d_allocated = n;
    }
    d_size = n;

    for (size_t j = 0; j < n; ++j) {
        if (v[j]) {
            char * const s = new char[strlen(v[j]) + 1];
            strcpy(s, v[j]);
            d_v[j] = s;
        } else {
            d_v[j] = 0;
        }
    }
}



bool VrmlMFString::set(const VrmlField &src)
{
  if (src.fieldType() == MFSTRING)
  {
    *this = (const VrmlMFString &) src;
    return true;
  }
  return false;
}




VrmlMFString& VrmlMFString::operator=(const VrmlMFString& rhs)
{
  if (this != &rhs)
  {
    set( rhs.d_size, rhs.d_v );
    VrmlObject::operator=(rhs);
  }
  return *this;
}

VrmlField *VrmlMFString::clone(VrmlNamespace* ns) const
{ return new VrmlMFString(*this); }

VrmlField::VrmlFieldType VrmlMFString::fieldType() const { return MFSTRING; }

size_t VrmlMFString::getLength() const
{
    return d_size;
}

char const * const * VrmlMFString::get() const {
    return d_v;
}

char const * VrmlMFString::get(size_t index) const {
    assert(index < this->d_size);
    return this->d_v[index];
}

void VrmlMFString::set(size_t index, const char * str)
{
  // is reallocation needed?
  if (index >= d_size)
  {
    size_t size_new = index+1;
    char * * v_new = new char *[size_new];

    // copy data
    for (size_t i = 0; i < d_size; ++i)
    {
      char * const s = new char[strlen(d_v[i])+1];
      strcpy(s, d_v[i]);
      v_new[i] = s;
    }

    // append new data
    for (size_t i = d_size; i < size_new; ++i)
    {
      if (i == index && str != 0)
      {
	v_new[i] = new char[strlen(str)+1];
	strcpy(v_new[i], str);
      }
      else
      {
	v_new[i] = new char[1];
	strcpy(v_new[i], "\0");
      }
    }

    // destruct original data
    for (size_t i = 0; i < d_size; ++i)
    {
	delete [] d_v[i];
    }
    delete [] d_v;

    // assign new data
    d_v = v_new;
    d_allocated = d_size = size_new;
  }
  // no reallocation needed
  else
  {
    delete [] this->d_v[index];
    char * const s = new char[strlen(str) + 1];
    strcpy(s, str);
    this->d_v[index] = s;
  }
}

ostream& VrmlMFString::print(ostream& os, int indent) const
{
  size_t n = getLength();

  if (n != 1) os << '[';
  if (n >= 2) os << std::endl;

  for (size_t i=0; i<n; ++i)
  {
    if (n != 1) for (size_t j=0; j<indent+VrmlNode::INDENT_INCREMENT; ++j) os << ' ';
    os << '\"' << (d_v[i]) << "\" ";
    if (n != 1) os << std::endl;
  }

  if (n != 1)
  {
    for (size_t j=0; j<indent; ++j) os << ' ';
    os << ']';
  }

  return os;
}

ostream& VrmlMFString::printAsUrl(ostream& os, int indent) const
{
  size_t n = getLength();
  const std::string& libraryToken(theSystem->getLibraryToken());

  if (n != 1) os << '[';
  if (n >= 2) os << std::endl;

  for (size_t i=0; i<n; ++i)
  {
    if (n != 1) for (size_t j=0; j<indent+VrmlNode::INDENT_INCREMENT; ++j) os << ' ';
    std::string destUrl = theSystem->copyResource(d_v[i]);
    os << '\"' << destUrl << "\" ";
    if (n != 1) os << std::endl;
  }

  if (n != 1)
  {
    for (size_t j=0; j<indent; ++j) os << ' ';
    os << ']';
  }

  return os;
}

bool VrmlMFString::equals(const VrmlField* f) const
{
  if (f->fieldType() == MFSTRING)
  {
    if (((VrmlMFString*) f)->d_size == d_size)
    {
      for (size_t i=0; i<d_size; i++)
        if (strcmp(((VrmlMFString*) f)->d_v[i], d_v[i]) != 0) return false;
      return true;
    }
  }
  return false;
}


/* external value conversion API */

void VrmlMFString::getExtValueType(extClassID* type, size_t* m, size_t* n, size_t* k) const
  { *type = extPOINTER_CLASS; *m = d_size; *n = 1; *k = 1; }

bool VrmlMFString::getExtValue(extClassID type, void* values) const
{
  *((char ***) values) = d_v;
  return(true);
}

bool VrmlMFString::setExtValue(extClassID type, size_t m, size_t n, size_t k, const void* values)
{
  switch(type)
  {
    case extPOINTER_CLASS :
      if (m>1 && n>1)
        return(false);
      set(m*n, (const char**) values);
    break;

    case extCHAR_CLASS :
      set(1, (const char**) &values);
    break;

    default : return(false);
  }

  return(true);
}

void VrmlMFString::serialize(ostream& str) const
{
  size_t len = getLength();
  StreamWriter(str).WriteInt32(static_cast<int>(len));

  for (size_t i=0; i<len; i++)
    StreamWriter(str).WriteString(get(i));
};


bool VrmlMFString::readSerialized(istream& str)
{
  size_t i;

  // delete the old strings
  for (i=0; i<d_size; ++i)
    delete[] d_v[i];

  // reallocate pointer array if not enough allocated
  size_t len = StreamReader(str).ReadInt32();
  if (d_allocated < len)
  {
    delete[] d_v;
    d_v = new char*[len];
    d_allocated = len;
  }

  // store the strings read
  for (i=0; i<len; i++)
    d_v[i] = StreamReader(str).ReadString();
  d_size = len;

  return true;
};




/******************/
/*** VrmlMFTime ***/
/******************/



VrmlMFTime::VrmlMFTime(size_t length, const double * times)
{
  data = RefCntArray<double>::create(length, times);
}



VrmlMFTime::VrmlMFTime(const VrmlMFTime &src)
{
  src.data->incRef();
  data = src.data;
}



VrmlMFTime::~VrmlMFTime()
{
  data->decRef();
}



bool VrmlMFTime::set(const VrmlField &src)
{
  if (src.fieldType() == MFTIME)
  {
    *this = (const VrmlMFTime &) src;
    return true;
  }
  return false;
}



VrmlMFTime &VrmlMFTime::operator=(const VrmlMFTime &src)
{
  src.data->incRef();
  data->decRef();
  data = src.data;
  VrmlObject::operator=(src);
  return *this;
}



double VrmlMFTime::operator[](size_t index) const
{
  return data->getValue(index);
}



double &VrmlMFTime::operator[](size_t index)
{
  return *(data->getValues() + index);
}



const double * VrmlMFTime::get() const
{
  return data->getValues();
}



void VrmlMFTime::set(size_t length, const double * times)
{
  RefCntArray<double>* newData = RefCntArray<double>::create(length, times);
  data->decRef();
  data = newData;
}



size_t VrmlMFTime::getLength() const
{
  return data->getLength();
}



void VrmlMFTime::setLength(size_t length)
{
  RefCntArray<double>* newData = RefCntArray<double>::create(length, data, 0.0);
  data->decRef();
  data = newData;
}



VrmlField * VrmlMFTime::clone(VrmlNamespace* ns) const
{
  return new VrmlMFTime(*this);
}



VrmlField::VrmlFieldType VrmlMFTime::fieldType() const
{
  return MFTIME;
}



ostream & VrmlMFTime::print(ostream & os, int indent) const
{
  return mfprint(os, get(), getLength(), 1, indent);
}



bool VrmlMFTime::equals(const VrmlField* f) const
{
  if (f->fieldType() != MFTIME) return false;
  return data->equals(((VrmlMFTime*) f)->data);
}



/* external value conversion API */

void VrmlMFTime::getExtValueType(extClassID* type, size_t* m, size_t* n, size_t* k) const
  { *type = extDOUBLE_CLASS; *m = getLength(); *n = 1; *k = 1; }

bool VrmlMFTime::getExtValue(extClassID type, void* values) const
  { return(VRMLData2ML(values, type, getLength(), 1, data->getValues())); }

bool VrmlMFTime::setExtValue(extClassID type, size_t m, size_t n, size_t k, const void* values)
{
  if (!((n==1 || (n==0 && m==0)) && k==1))  // accept also 0x0 empty
    return false;

  // create new data array if size doesn't match
  if (m != getLength())
  {
    // create new data array
    RefCntArray<double>* newData = RefCntArray<double>::create(m, data, 0.0);
    // free the old data array
    data->decRef();
    data = newData;
  }

  // create an unique copy if data shared
  else
    data = data->makeUnique();

  return(MLData2VRML(data->getValues(), type, m, n, values));
}



void VrmlMFTime::serialize(ostream& str) const
{
  size_t len = getLength();
  StreamWriter(str).WriteInt32(static_cast<int>(len));

  const double* x = data->getValues();
  for (size_t i=0; i<len; i++)
    StreamWriter(str).WriteDouble(x[i]);
};

bool VrmlMFTime::readSerialized(istream& str)
{
  int len = StreamReader(str).ReadInt32();

  // create new data array
  RefCntArray<double>* newData = RefCntArray<double>::create(len, data, 0);
  // free the old data array
  data->decRef();
  data = newData;

  // fill the new array with data
  double* x = data->getValues();
  for (int i=0; i<len; i++)
    x[i] = StreamReader(str).ReadDouble();
  return true;
};



/*******************/
/*** VrmlMFVec2f ***/
/*******************/



const float* VrmlMFVec2f::operator[](size_t index) const
{
  return data->getValues() + 2*index;
}



float* VrmlMFVec2f::operator[](size_t index)
{
  return data->getValues() + 2*index;
}



void VrmlMFVec2f::setElement(size_t index, float x, float y)
{
  if (index >= getLength()) setLength(index+1);
  data = data->makeUnique();
  data->setValue(2*index, x);
  data->setValue(2*index+1, y);
}



/*******************/
/*** VrmlMFVec3f ***/
/*******************/



const float * VrmlMFVec3f::operator[](size_t index) const
{
  return data->getValues() + 3*index;
}



float * VrmlMFVec3f::operator[](size_t index)
{
  return data->getValues() + 3*index;
}



void VrmlMFVec3f::setElement(size_t index, float x, float y, float z)
{
  if (index >= getLength()) setLength(index+1);
  data = data->makeUnique();
  data->setValue(3*index, x);
  data->setValue(3*index+1, y);
  data->setValue(3*index+2, z);
}



/********************/
/*** StreamWriter ***/
/********************/


void StreamWriter::WriteBool(bool value)
{
  WriteByte(value ? 1 : 0);
}

void StreamWriter::WriteInt16(short n)
{
  WriteByte(n>>8);
  WriteByte((char) n);
}

void StreamWriter::WriteInt32(int n)
{
  for (int sh=3*8; sh>=0; sh-=8)
    WriteByte(static_cast<char>(n>>sh));
}

void StreamWriter::WriteFloat(float n)
{
  union { int i; float f; } val;

  val.f = n;
  WriteInt32(val.i);
}

void StreamWriter::WriteDouble(double n)
{
  union { long long i; double d; } val;
  val.d = n;

  for (int sh=7*8; sh>=0; sh-=8)
    WriteByte(static_cast<char>(val.i>>sh));
}


void StreamWriter::WriteString(const char *str)
{
  size_t strl = strlen(str);
  WriteInt16(static_cast<short>(strl));
  for (size_t i=0; i<strl; i++)
    WriteByte(str[i]);
}


void StreamWriter::WriteFormattedString(const char* format, ...)
{
  char buf[1024];
  va_list args;
  va_start(args, format);
  vsprintf(buf, format, args);
  va_end(args);
  WriteString(buf);
}



/********************/
/*** StreamReader ***/
/********************/


bool StreamReader::ReadBool()
{
  return (ReadByte()!=0);
}

short StreamReader::ReadInt16()
{
  short result = ReadByte()<<8;
  result |=    ReadByte();
  return result;
}

int StreamReader::ReadInt32()
{
  int val = 0;
  for (int sh=3*8; sh>=0; sh-=8)
    val |= ReadByte()<<sh;

  return val;
}

float StreamReader::ReadFloat()
{
  union { int i; float f; } val;

  val.i = ReadInt32();
  return val.f;
}

double StreamReader::ReadDouble()
{
  union { long long i; double d; } val;

  val.i = 0;
  for (int sh=7*8; sh>=0; sh-=8)
    val.i |= static_cast<long long>(ReadByte())<<sh;

  return val.d;
}

char *StreamReader::ReadString()
{
  char *buf;

  size_t slen = ReadInt16();
  buf = new char[slen+1];
  for (size_t i=0; i<slen; i++)
    buf[i] = ReadByte();
  buf[slen] = '\0';
  return buf;
}
