/*****************************************************************************
******************************************************************************
*
*               VRML field (also contains all field types)
*
*               $Revision: 1.1.2.16 $
*               $Date: 2012/02/03 15:27:33 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright 1998-2012 HUMUSOFT s.r.o.
// Portions Copyright (C) 1998  Chris Morley

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#ifndef VRMLFIELD_H
#define VRMLFIELD_H

#include <string.h>

#ifdef WIN32
#include <float.h>
#define isnan(n) _isnan(n)
#define strcasecmp(s1, s2) _stricmp(s1, s2)
#define strncasecmp(s1, s2, n) _strnicmp(s1, s2, n)
#else
#include <math.h>
#endif

#include "RefCntArray.h"
#include "System.h"   // OPENVRMLDLLEXPORT

#include "VrmlObject.h"

#include "Math3d.h"

#include <iostream>
#include <vector>
using namespace std;


/* external datatype class ID */
enum extClassID {
	extUNKNOWN_CLASS = 0,
	extPOINTER_CLASS,
        // class ID 2 is obsolete
	extLOGICAL_CLASS = 3,
	extCHAR_CLASS,
	extVOID_CLASS,
	extDOUBLE_CLASS,
	extSINGLE_CLASS,
	extINT8_CLASS,
	extUINT8_CLASS,
	extINT16_CLASS,
	extUINT16_CLASS,
	extINT32_CLASS,
	extUINT32_CLASS,
	extINT64_CLASS,
	extUINT64_CLASS
};



class VrmlNode;
class VrmlNamespace;


// Global output operator for fields

class VrmlField;
ostream& operator<<(ostream& os, const VrmlField& f);


/**
 * Abstract base class for VRML field values.
 */
class OPENVRMLDLLEXPORT VrmlField : public VrmlObject {

  friend ostream& operator<<(ostream& os, const VrmlField& f);

public:

  /**
   * Field type identifiers. SF... fields have identifiers in range
   * <0x0001..0x000b>. An appropriate MF... field identifier is obtained
   * by changing the first byte to 0x01. So, SFVEC3F is 0x000b,
   * MFVEC3F is 0x010b. Note that MFBOOL and MFIMAGE do not exist.
   */
  enum VrmlFieldType
  {
    NO_FIELD    = 0x00,
    SFBOOL      = 0x01,
    SFCOLOR     = 0x02,
    SFFLOAT     = 0x03,
    SFIMAGE     = 0x04,
    SFINT32     = 0x05,
    SFNODE      = 0x06,
    SFROTATION  = 0x07,
    SFSTRING    = 0x08,
    SFTIME      = 0x09,
    SFVEC2F     = 0x0a,
    SFVEC3F     = 0x0b,

    MFFIELD     = 0x80,
    MFCOLOR     = MFFIELD | SFCOLOR,
    MFFLOAT     = MFFIELD | SFFLOAT,
    MFINT32     = MFFIELD | SFINT32,
    MFNODE      = MFFIELD | SFNODE,
    MFROTATION  = MFFIELD | SFROTATION,
    MFSTRING    = MFFIELD | SFSTRING,
    MFTIME      = MFFIELD | SFTIME,
    MFVEC2F     = MFFIELD | SFVEC2F,
    MFVEC3F     = MFFIELD | SFVEC3F
  };

  virtual ~VrmlField();

  /**
   * Copy self.
   */
  virtual VrmlField *clone(VrmlNamespace* ns = 0) const = 0;

  /**
   * Writes the textual representation of the field value
   * to the given stream with appropriate indentation.
   * @param os The stream to write to.
   * @param indent Indentation value (used only for fields
   * where the textual representation is multilined).
   */
  virtual ostream& print(ostream& os, int indent = 0) const = 0;

  /**
   * Prints field as URL and exports resources from sl3d library (if any)
   */
  virtual ostream& printAsUrl(ostream& os, int indent = 0) const;

  /** Write simple scene description into the stream.
   * @param os target stream
   * @param editFlag true if chain is created for 3D World Editor purposes
   * @return resulting stream.
   */
  virtual ostream& toChain(ostream& os, bool editFlag) const {return os;};

  /**
   * Returns the field type (one of VrmlField::VrmlFieldType values).
   */
  virtual VrmlFieldType fieldType() const = 0;

  /**
   * Compares this field with the given field, returning true
   * if they have the same value, and false otherwise.
   */
  virtual bool equals(const VrmlField* f) const = 0;

  /** Find out whether current field has got any children.
   * @code return <code>true</code>, if the field has got
   * any children, otherwise, it returns <code>false</code>.
   */
  virtual bool hasChildren() const {return false;};

  /**
   * Returns the name of the type of this field value (e.g. "SFVec2f").
   */
  const char *fieldTypeName() const;

  /**
   * Returns the field type identifier for the given type,
   * specified by its name.
   */
  static VrmlFieldType fieldType(const char *fieldTypeName);

  /**
   * Creates a new VrmlField object of the given type, initialized
   * to default value.
   */
  static VrmlField* create(int ftype);

  /** Converts a set of numeric values to a VrmlField
    * of the given type.
    * @param ftype Field type.
    * @param count Number of values to set.
    * @param values Array of values.
    * @return A newly allocated VrmlField value of appropriate type,
    * or NULL if the conversion is not possible.
    */
  static VrmlField* convert(int ftype, size_t count, const double* values);

  /** Converts a set of strings to a VrmlField of the given type.
    * @param ftype Field type.
    * @param count Number of values to set.
    * @param values Array of values.
    * @return A newly allocated VrmlField value of appropriate type,
    * or NULL if the conversion is not possible.
    */
  static VrmlField* convert(int ftype, size_t count, const char** values);

  /** Converts a set of node pointers to a VrmlField of the given type.
    * @param ftype Field type.
    * @param count Number of values to set.
    * @param values Array of values.
    * @return A newly allocated VrmlField value of appropriate type,
    * or NULL if the conversion is not possible.
    */
  static VrmlField* convert(int ftype, size_t count, VrmlNode* const* values);

  /** Converts an empty vector to a VrmlField of the given type.
    * @param ftype Field type.
    * @return A newly allocated VrmlField value of appropriate type,
    * or NULL if the conversion is not possible.
    */
  static VrmlField* convertEmptyVector(int ftype);

  /**
   * Assignment operator.
   */
  virtual bool set(const VrmlField &src);

  //-------------------------------------------------------------------------
  // value conversion API
  //-------------------------------------------------------------------------

  /**
   * Gets external data type and dimensions of a field
   * @param type pointer to external data type
   * @param m pointer to number of rows
   * @param n pointer to number of columns
   * @param k pointer to number indicating size of third dimension of array (zero value distinguishes nodes from strings)
   */
  virtual void getExtValueType(extClassID* type, size_t* m, size_t* n, size_t* k) const
    { *type = extDOUBLE_CLASS; *m = *n = 0; *k = 1; }

  /**
   * Gets values in external data type format
   * @param type requested external data type
   * @param m pointer to values
   * @return True on success, false if the field can't return the value.
   */
  virtual bool getExtValue(extClassID type, void* values) const
    { return false; }

  /**
   * Gets values in external data type format
   * @param type requested external data type
   * @param m pointer to values
   * @param tm transformation matrix to transform field values
   * @return True on success, false if the field can't return the value.
   */
  virtual bool getExtValueInWorldCoords(extClassID type, void* values, const Matrix4x4* tm) const;

  /**
   * Sets the field to the given external data
   * The data are specified in column major order.
   * @param type external data type
   * @param m The number of rows.
   * @param n The number of columns.
   * @param k pointer to number indicating size of third dimension of array
   * @param values The values.
   * @return True on success, false if the field can't accept the value.
   */
  virtual bool setExtValue(extClassID type, size_t m, size_t n, size_t k, const void* values)
    { return false; }


  /**
   * Writes serial representation of the field into buffer.
   * @param str Stream to write serialized field to.
   */
  virtual void serialize(ostream& str) const
    { };


  /**
   * Reads serial representation of the field from buffer.
   * and sets the field to this value.
   * @param str Stream to read serialized field from.
   * @return True on success, false if the field can't be set.
   */
  virtual bool readSerialized(istream& str)
    { return false; };



  //-------------------------------------------------------------------------
  // type classification API
  //-------------------------------------------------------------------------

  /**
   * Checks whether the given field type is numeric (e.g. SFVec3f
   * is numeric, SFNode is not).
   * @param ftype Field type identifier.
   * @return True if the field type is numeric, false otherwise.
   */
  static bool isTypeNumeric(int ftype);

  /**
   * Checks whether the given field type can be set externally.
   * @param ftype Field type identifier.
   * @return True if the field type is numeric, false otherwise.
   */
  static bool IsExtSettable(int ftype);


  /**
   * Checks whether the given field type is SFxxx.
   * @param ftype Field type identifier.
   * @return True if the given field is a SF, false if it is not.
   */
  static bool isTypeSF(int ftype);

  /**
   * Returns a field type name for the given field type identifier.
   * @param ftype Field type identifier.
   * @return Pointer to statically allocated string containing
   * the name, or NULL if the field type identifier is not known.
   */
  static const char* getTypeName(int ftype);

  /**
   * Returns the width (number of columns in the matrix) of the given type
   * (e.g. SFVec3f or MFVec3f has width of 3, SFFloat has width of 1).
   * Returns 0 for non-matrix fields (e.g. SFString, MFNode, SFImage).
   * @param ftype Field type identifier.
   * @return The width of the given type.
   */
  static size_t getTypeWidth(int ftype);

  /**
   * Returns the external class type ID for the given type.
   * @param ftype Field type identifier.
   * @return External class ID for the given type.
   */
  static extClassID getTypeExtClassID(int ftype);

  /**
   * Returns the external class type name for the given type.
   * @param ftype Field type identifier.
   * @return External class name for the given type.
   */
  static const char* getTypeExtClassName(int ftype);
};


/**
 * A SFBool field value.
 */
class VrmlSFBool : public VrmlField {
public:

  /**
   * Constructor.
   * @param value The initial value.
   */
  VrmlSFBool(bool value = false) : d_value(value) { }

  /**
   * Get the value of this field.
   * @return The value of this SFBool, as a boolean.
   */
  bool get() const { return d_value; }

  /**
   * Set the value of this field.
   * @param value The new boolean value.
   */
  void set(bool value) { d_value = value; }

  bool set(const VrmlField &src);

  virtual ostream& print(ostream& os, int indent = 0) const
    { return (os << (d_value ? "TRUE" : "FALSE")); }

  virtual VrmlField *clone(VrmlNamespace* ns = 0) const;
  virtual VrmlFieldType fieldType() const { return SFBOOL; }
  virtual bool equals(const VrmlField* f) const;
  virtual void getExtValueType(extClassID* type, size_t* m, size_t* n, size_t* k) const;
  virtual bool getExtValue(extClassID type, void* values) const;
  virtual bool setExtValue(extClassID type, size_t m, size_t n, size_t k, const void* values);
  virtual void serialize(ostream& str) const;
  virtual bool readSerialized(istream& str);

private:
  bool d_value;
};



/**
 * This is a base class for fields that contain a single float
 * or a fixed number of floats (i.e. VrmlSFFloat, VrmlSFVec2f,
 * VrmlSFVec3f, VrmlSFColor, VrmlSFRotation).
 */
class OPENVRMLDLLEXPORT VrmlSingleFloatField : public VrmlField
{
protected:

  size_t width;

// we need at most four elements per field, so let's fix it instead of allocating dynamically
// some constructors are fixed to four values anyway
  float x[4];

public:

  /** Constructor. Field elements are initialized to zeros. */
  VrmlSingleFloatField(int _width)
    { width = _width; memset(x, 0, width*sizeof(float)); }

  /** Copy constructor. */
  VrmlSingleFloatField(const VrmlSingleFloatField &src)
    { width = src.width; memcpy(x, src.x, width*sizeof(float)); }

  /** Constructor. Field elements are initialized using the given array. */
  VrmlSingleFloatField(int _width, const float *values)
    { width = _width; memcpy(x, values, width*sizeof(float)); }

  /**
   * Constructor. Field elements are initialized to given numbers.
   * Excess parameters are ignored.
   */
  VrmlSingleFloatField(size_t _width, float a, float b = 0.0, float c = 0.0, float d = 0.0)
    {
	  width = _width;
	  switch (width)
	  {
		case 4: x[3] = d;   // and fallthrough to the next case
		case 3: x[2] = c;   // and fallthrough to the next case
		case 2: x[1] = b;   // and fallthrough to the next case
		case 1: x[0] = a;
	  }
      x[0] = a;
    }

  /** Element access operator (const version). */
  float operator[](size_t index) const
    { return (index < 0 || index >= width) ? x[0] : x[index]; }

  /** Element access operator (non-const version). */
  float& operator[](size_t index)
    { return (index < 0 || index >= width) ? x[0] : x[index]; }

  /** Sets the array elements to given values. */
  void set(const float* values)
    { size_t i; for (i=0; i<width; i++) x[i] = values[i]; }

  /**
   * Sets the field to the value of the given field if field types
   * are compatible. Returns true on success, false on failure.
   */
  bool set(const VrmlField &src)
    {
      if (src.fieldType() != fieldType()) return false;
      VrmlField::set(src);
      memcpy(x, ((const VrmlSingleFloatField &) src).x, width*sizeof(float));
      return true;
    }

  /** Returns true if field values are equal, false otherwise. */
  bool equals(const VrmlField* src) const
    {
      if (src->fieldType() != fieldType()) return false;
      return(memcmp(x, static_cast<const VrmlSingleFloatField*>(src)->x, width*sizeof(float)) == 0);
    }

  /** Prints the textual representation of the field to the stream. */
  ostream& print(ostream& os, int indent = 0) const
    {
      size_t i;
      for (i=0; i<width; i++)
      {
        os << x[i];
        if (i<width-1) os << ' ';
      }
      return os;
    }

  virtual void getExtValueType(extClassID* type, size_t* m, size_t* n, size_t* k) const;
  virtual bool getExtValue(extClassID type, void* values) const;
  virtual bool setExtValue(extClassID type, size_t m, size_t n, size_t k, const void* values);
  virtual void serialize(ostream& str) const;
  virtual bool readSerialized(istream& str);
};



/**
 * A SFColor field value.
 */
class VrmlSFColor : public VrmlSingleFloatField
{
public:

  /** Default constructor. Default value is (0, 0, 0). */
  VrmlSFColor()
    : VrmlSingleFloatField(3) { }

  /** Constructor. */
  VrmlSFColor(const float* rgb)
    : VrmlSingleFloatField(3, rgb) { }

  /** Constructor. */
  VrmlSFColor(float r, float g, float b)
    : VrmlSingleFloatField(3, r, g, b) { }

  /** Copy constructor. */
  VrmlSFColor(const VrmlSFColor &src)
    : VrmlSingleFloatField(src) { }

  virtual VrmlField *clone(VrmlNamespace* ns = 0) const { return new VrmlSFColor(*this); }
  virtual VrmlFieldType fieldType() const { return SFCOLOR; }

  /** Returns a pointer to the array of field elements. */
  const float* get() const { return x; }

  /** Get the red component value. */
  float getR() const { return x[0]; }

  /** Get the green component value. */
  float getG() const { return x[1]; }

  /** Get the blue component value. */
  float getB() const { return x[2]; }

  /**
   * @brief Convert a color from HSV to RGB.
   *
   * Convert from HSV (with(with <var>h</var> in [0,360), <var>s</var>,
   * <var>v</var> in [0,1]) to RGB (with each component in [0,1]).
   *
   * @param hsv a 3-element array comprising an HSV value
   * @retval rgb a 3-element array comprising an RGB value
   */
  static void HSVtoRGB(const float hsv[3], float rgb[3]);

  /**
   * @brief Convert a color from RGB to HSV.
   *
   * Convert from RGB (with each component in [0,1]) to HSV (with <var>h</var> in
   * [0,360), <var>s</var>, <var>v</var> in [0,1]).
   *
   * @param rgb a 3-element array comprising an RGB value
   * @retval hsv a 3-element array comprising an HSV value
   */
  static void RGBtoHSV(const float rgb[3], float hsv[3]);

  /**
   * Set the value using HSV.
   * @param h the hue component
   * @param s the saturation component
   * @param v the value component
   */
  void setHSV(float h, float s, float v);

  /**
   * @brief Get the value expressed in HSV.
   * @retval hsv a 3-element array comprising the HSV value.
   */
  void getHSV(float hsv[3]) const;

  virtual void getExtValueType(extClassID* type, size_t* m, size_t* n, size_t* k) const;
  virtual bool getExtValue(extClassID type, void* values) const;
  virtual bool setExtValue(extClassID type, size_t m, size_t n, size_t k, const void* values);
};



/**
 * A SFFloat value.
 */
class OPENVRMLDLLEXPORT VrmlSFFloat : public VrmlSingleFloatField
{
public:

  VrmlSFFloat(float value = 0.0)
    : VrmlSingleFloatField(1, value) { }

  virtual VrmlField *clone(VrmlNamespace* ns = 0) const
    { return new VrmlSFFloat(*this); }

  virtual VrmlFieldType fieldType() const
    { return SFFLOAT; }


  float get() const
    { return x[0]; }

  void set(float value)
    { x[0] = value; }
};



/**
 * @class VrmlSFImage
 * A single uncompressed 2-dimensional pixel image. The first
 * hexadecimal value is the lower left pixel and the last value is the
 * upper right pixel.Pixel values are limited to 256 levels of
 * intensity. A one-component image specifies one-byte greyscale
 * values. A two-component image specifies the intensity in the first
 * (high) byte and the alpha opacity in the second (low) byte. A
 * three-component image specifies the red component in the first
 * (high) byte, followed by the green and blue components.
 * Four-component images specify the alpha opacity byte after
 * red/green/blue.
 *
 */
class OPENVRMLDLLEXPORT VrmlSFImage : public VrmlField {
public:

  /**
   * Construct the default SFImage.
   */
  VrmlSFImage();

  /**
   * @brief Create an SFImage.
   *
   * Note that the pixels read from lower left to upper right, which
   * is a reflection around the y-axis from the "normal" convention.
   * <p>
   * Note also that width and height are specified in pixels, and a
   * pixel may be more than one byte wide. For example, an image with
   * a width and height of 16, and nc==3, would have a pixel array
   * w*h*nc = 16*16*3 = 768 bytes long. See the class intro above for
   * the interpretation of different pixel depths.
   *
   * @param width width in pixels
   * @param height height in pixels
   * @param components number of components/pixel (see above)
   * @param pixels the caller owns the bytes, so this ctr makes a copy
   *
   */
  VrmlSFImage(size_t w, size_t h, size_t nc, const unsigned char * pixels = 0);

  VrmlSFImage(const VrmlSFImage&);
  ~VrmlSFImage();

  /**
   * @brief Assignment.
   */
  VrmlSFImage& operator=(const VrmlSFImage& rhs);

  virtual ostream& print(ostream& os, int indent = 0) const;
  virtual VrmlField *clone(VrmlNamespace* ns = 0) const;
  virtual VrmlFieldType fieldType() const;
  virtual bool equals(const VrmlField* f) const;

  /**
   * Gets the image width.
   */
  size_t getWidth() const;

  /**
   * Gets the image height.
   */
  size_t getHeight() const;

  /**
   * Gets the number of color components.
   */
  size_t getComponents() const;

  /**
   * Get the pixel data.
   */
  const unsigned char * getPixels() const;

  /**
   * Set the image.
   * @param width width in pixels
   * @param height height in pixels
   * @param components number of components
   * @param pixels array of (width * height * components) bytes comprising
   * the image data.
   */
  void set(size_t width, size_t height, size_t components,
           const unsigned char * pixels);

  bool set(const VrmlField &src);

  void reallocatePixels(size_t width, size_t height, size_t components, const unsigned char* copySource);

  virtual void getExtValueType(extClassID* type, size_t* m, size_t* n, size_t* k) const;
  virtual bool getExtValue(extClassID type, void* values) const;
  virtual bool setExtValue(extClassID type, size_t m, size_t n, size_t k, const void* values);

private:

  size_t d_w, d_h, d_nc;
  unsigned char *d_pixels;	// nc bytes/pixel, lower left to upper right

};



/**
 * A SFInt32 value.
 */
class VrmlSFInt32 : public VrmlField {
public:

  VrmlSFInt32(long = 0);

  virtual ostream& print(ostream& os, int indent = 0) const
    { return (os << d_value); }

  virtual VrmlField *clone(VrmlNamespace* ns = 0) const;
  virtual VrmlFieldType fieldType() const;
  virtual bool equals(const VrmlField* f) const;
  long get() const;
  void set(long);
  bool set(const VrmlField &src);

  bool set(double value)
    { d_value = (long) value; return true; }

  virtual void getExtValueType(extClassID* type, size_t* m, size_t* n, size_t* k) const;
  virtual bool getExtValue(extClassID type, void* values) const;
  virtual bool setExtValue(extClassID type, size_t m, size_t n, size_t k, const void* values);
  virtual void serialize(ostream& str) const;
  virtual bool readSerialized(istream& str);

private:
  long d_value;

};



/**
 * A SFNode value.
 */
class OPENVRMLDLLEXPORT VrmlSFNode : public VrmlField {
public:

  VrmlSFNode(VrmlNode * node = 0);
  VrmlSFNode(const VrmlSFNode &n);
  ~VrmlSFNode();
  VrmlSFNode& operator=(const VrmlSFNode& rhs);
  virtual ostream& print(ostream& os, int indent = 0) const;
  /** Write simple scene description into the stream.
   * @param os target stream
   * @param editFlag true if chain is created for 3D World Editor purposes
   * @return resulting stream.
   */
  virtual ostream& toChain(ostream& os, bool editFlag) const;
  virtual VrmlField *clone(VrmlNamespace* ns = 0) const;
  virtual VrmlFieldType fieldType() const;
  virtual bool equals(const VrmlField* f) const;
  VrmlNode * get() const;

  /** Find out whether current field has got any children.
   * @code return <code>true</code>, if the field has got
   * any children, otherwise, it returns <code>false</code>.
   */
  virtual bool hasChildren() const;


  void set(VrmlNode * node);
  bool set(const VrmlField &src);

  virtual void getExtValueType(extClassID* type, size_t* m, size_t* n, size_t* k) const;
  virtual bool getExtValue(extClassID type, void* values) const;
  virtual bool setExtValue(extClassID type, size_t m, size_t n, size_t k, const void* values);

private:
  VrmlNode *d_value;
};



/**
 * A SFVec2f value.
 */
class VrmlSFVec2f : public VrmlSingleFloatField
{
public:

  /** Default constructor. Default value is (0, 0). */
  VrmlSFVec2f()
    : VrmlSingleFloatField(2) { }

  /** Constructor. */
  VrmlSFVec2f(const float vec[2])
    : VrmlSingleFloatField(2, vec) { }

  /** Constructor. */
  VrmlSFVec2f(float x, float y)
    : VrmlSingleFloatField(2, x, y) { }

  /** Copy constructor. */
  VrmlSFVec2f(const VrmlSFVec2f &src)
    : VrmlSingleFloatField(src) { }

  virtual VrmlField *clone(VrmlNamespace* ns = 0) const { return new VrmlSFVec2f(*this); }
  virtual VrmlFieldType fieldType() const { return SFVEC2F; }

  /** Get the x component. */
  float getX() const { return x[0]; }

  /** Get the y component. */
  float getY() const { return x[1]; }

  /** Set the x component. */
  void setX(float val) { x[0] = val; }

  /** Set the y component. */
  void setY(float val) { x[1] = val; }

  /**
   * Get the value of this vector.
   * @returns a pointer to a 2-element array.
   */
  const float * get() const { return x; }

  /**
   * Add two vectors.
   * @param vec the vector to add to this one.
   * @return a VrmlSFVec2f with a value that is the passed SFVec2f added,
   *         componentwise, to this object.
   */
  const VrmlSFVec2f add(const VrmlSFVec2f & vec) const
    { return VrmlSFVec2f(x[0] + vec.x[0], x[1] + vec.x[1]); }

  /**
   * Divide this vector by a scalar.
   * @param number a scalar value.
   * @return a VrmlSFVec2f with a value that is the object divided by the
   *         passed numeric value.
   */
  const VrmlSFVec2f divide(float number) const
    { return VrmlSFVec2f(x[0]/number, x[1]/number); }

  /**
   * Dot product.
   * @param vec
   * @return the dot product of this vector and vec.
   */
  double dot(const VrmlSFVec2f & vec) const
    { return x[0]*vec.x[0] + x[1]*vec.x[1]; }

  /**
   * Geometric length.
   * @return the length of this vector.
   */
  double length() const;

  /**
   * Multiply by a scalar.
   * @param number a scalar value
   * @return a VrmlSFVec2f with a value that is the object multiplied by the
   *         passed numeric value.
   */
  const VrmlSFVec2f multiply(float number) const
    { return VrmlSFVec2f(x[0]*number, x[1]*number); }

  /**
   * Negate.
   * @return a VrmlSFVec2f that the result of negating this vector.
   */
  const VrmlSFVec2f negate() const
    { return VrmlSFVec2f(-x[0], -x[1]); }

  /**
   * Normalize.
   * @return a VrmlSFVec2f that is this vector normalized.
   */
  const VrmlSFVec2f normalize() const;

  /**
   * Take the difference of two vectors.
   * @param vec the vector to subtract from this one
   * @return a VrmlSFVec2f that is the difference between this vector and vec
   */
  const VrmlSFVec2f subtract(const VrmlSFVec2f & vec) const
    { return VrmlSFVec2f(x[0]-vec.x[0], x[1]-vec.x[1]); }
};



/**
 * A SFVec3f value.
 */
class OPENVRMLDLLEXPORT VrmlSFVec3f : public VrmlSingleFloatField {
public:

  /** Default constructor. Default value is (0, 0, 0). */
  VrmlSFVec3f()
    : VrmlSingleFloatField(3) { }

  /** Constructor. */
  VrmlSFVec3f(const float vec[3])
    : VrmlSingleFloatField(3, vec) { }

  /** Constructor. */
  VrmlSFVec3f(float x, float y, float z)
    : VrmlSingleFloatField(3, x, y, z) { }

  /** Copy constructor. */
  VrmlSFVec3f(const VrmlSFVec3f &src)
    : VrmlSingleFloatField(src) { }

  virtual VrmlField * clone(VrmlNamespace* ns = 0) const { return new VrmlSFVec3f(*this); }
  virtual VrmlFieldType fieldType() const { return SFVEC3F; }

  /** Gets the <var>x</var>-component. */
  float getX() const { return x[0]; }

  /** Gets the <var>y</var>-component. */
  float getY() const { return x[1]; }

  /** Gets the <var>z</var>-component. */
  float getZ() const { return x[2]; }

  /** Sets the <var>x</var>-component. */
  void setX(float val) { x[0] = val; }

  /** Sets the <var>y</var>-component. */
  void setY(float val) { x[1] = val; }

  /** Sets the <var>z</var>-component. */
  void setZ(float val) { x[2] = val; }

  /**
   * Get the vector value.
   * @return a pointer to a 3-element array
   */
  const float * get() const { return x; }

  /**
   * Add this vector and vec component-wise.
   * @param vec
   */
  const VrmlSFVec3f add(const VrmlSFVec3f & vec) const
    { return VrmlSFVec3f(x[0]+vec.x[0], x[1]+vec.x[1], x[2]+vec.x[2]); }

  /**
   * Get the cross product of this vector and vec.
   * @param vec
   */
  const VrmlSFVec3f cross(const VrmlSFVec3f & vec) const;

  /**
   * Get the result of dividing this vector by number.
   * @param number
   */
  const VrmlSFVec3f divide(float number) const
    { return VrmlSFVec3f(x[0]/number, x[1]/number, x[2]/number); }

  /**
   * Get the dot product of this vector and vec.
   * @param vec
   */
  double dot(const VrmlSFVec3f & vec) const
    { return x[0]*vec.x[0] + x[1]*vec.x[1] + x[2]*vec.x[2]; }

  /**
   * Get the length of this vector.
   * @return the geometric length of the vector.
   */
  double length() const;

  /**
   * Multiply by a scalar.
   * @param number
   * @return the product
   */
  const VrmlSFVec3f multiply(float number) const
    { return VrmlSFVec3f(x[0]*number, x[1]*number, x[2]*number); }

  /**
   * Negate.
   * @return the negatation of this vector
   */
  const VrmlSFVec3f negate() const
    { return VrmlSFVec3f(-x[0], -x[1], -x[2]); }

  /**
   * Normalize.
   * @return a copy of this vector normalized
   */
  const VrmlSFVec3f normalize() const;

  /**
   * Subtract.
   * @param vec
   * @return the difference between this vector and vec
   */
  const VrmlSFVec3f subtract(const VrmlSFVec3f & vec) const
    { return VrmlSFVec3f(x[0]-vec.x[0], x[1]-vec.x[1], x[2]-vec.x[2]); }

  bool getExtValueInWorldCoords(extClassID type, void* values, const Matrix4x4* tm) const;
};



/**
 * A SFRotation value.
 */
class OPENVRMLDLLEXPORT VrmlSFRotation : public VrmlSingleFloatField
{
public:

  /** Default constructor. Default value is (0, 0, 1, 0). */
  VrmlSFRotation()
    : VrmlSingleFloatField(4, 0, 0, 1, 0) { }

  /** Constructor. */
  VrmlSFRotation(const float rotation[4])
    : VrmlSingleFloatField(4, rotation) { }

  /** Constructor. */
  VrmlSFRotation(float x, float y, float z, float angle)
    : VrmlSingleFloatField(4, x, y, z, angle) { }

  /** Constructor. */
  VrmlSFRotation(const VrmlSFVec3f & axis, float angle)
    : VrmlSingleFloatField(4, axis[0], axis[1], axis[2], angle) { }

  /**
   * @brief Construct a rotation between two vectors.
   *
   * Construct a VrmlSFRotation equal to the rotation between two different
   * vectors.
   *
   * @param fromVector the starting vector
   * @param toVector the ending vector
   */
  VrmlSFRotation(const VrmlSFVec3f & fromVector,
    const VrmlSFVec3f & toVector);

  virtual VrmlField *clone(VrmlNamespace* ns = 0) const { return new VrmlSFRotation(*this); }
  virtual VrmlFieldType fieldType() const { return SFROTATION; }

  /** Get the array of rotation elements. */
  const float * get() const { return x; }

  /** Returns the X-coordinate of the rotation axis. */
  float getX() const { return x[0]; }

  /** Returns the Y-coordinate of the rotation axis. */
  float getY() const { return x[1]; }

  /** Returns the Z-coordinate of the rotation axis. */
  float getZ() const { return x[2]; }

  /** Returns the rotation angle. */
  float getAngle() const { return x[3]; }

  /** Sets the X-coordinate of the rotation axis. */
  void setX(float val) { x[0] = val; }

  /** Sets the X-coordinate of the rotation axis. */
  void setY(float val) { x[1] = val; }

  /** Sets the X-coordinate of the rotation axis. */
  void setZ(float val) { x[2] = val; }

  /** Sets the rotation angle. */
  void setAngle(float val) { x[3] = val; }

  /** Gets the rotation axis. */
  const VrmlSFVec3f getAxis() const
    { return VrmlSFVec3f(x[0], x[1], x[2]); }

  /** Sets the rotation axis. */
  void setAxis(const VrmlSFVec3f &vec)
    { x[0] = vec[0]; x[1] = vec[1]; x[2] = vec[2]; }

  /**
   * @brief Get the inverse.
   * @return a VrmlSFRotation that is the inverse of this one
   */
  const VrmlSFRotation inverse() const
    { return VrmlSFRotation(x[0], x[1], x[2], -x[3]); }

  /**
   * @brief Multiply two rotations.
   * @param rotation the rotation by which to multiply this one
   * @return the result rotation
   */
  const VrmlSFRotation multiply(const VrmlSFRotation & rot) const;

  /**
   * @brief Multiply the matrix corresponding to this rotation by a vector.
   *
   * @todo IMPLEMENT ME!
   *
   * @param vec vector by which to multiply this rotation
   * @return the result of multiplying this rotation by vec
   */
  const VrmlSFVec3f multVec(const VrmlSFVec3f & vec) const;

  /**
   * @brief Perform a <b>S</b>pherical <b>L</b>inear Int<b>ERP</b>olation.
   *
   * @param destRotation the destination rotation
   * @param t the interval fraction
   */
  const VrmlSFRotation slerp(const VrmlSFRotation & destRotation,
                                   float t) const;

  bool getExtValueInWorldCoords(extClassID type, void* values, const Matrix4x4* tm) const;

private:
  void toQuaternion(float theQuat[4]) const;
  void fromQuaternion(const float theQuat[4]);
};



/**
 * A SFString value.
 */
class OPENVRMLDLLEXPORT VrmlSFString : public VrmlField {
public:

  VrmlSFString(const char *s = 0);
  VrmlSFString(const VrmlSFString&);
  ~VrmlSFString();

  VrmlSFString& operator=(const VrmlSFString& rhs);
  virtual ostream& print(ostream& os, int indent = 0) const;
  virtual VrmlField *clone(VrmlNamespace* ns = 0) const;
  virtual VrmlFieldType fieldType() const;
  virtual bool equals(const VrmlField* f) const;
  bool set(const VrmlField &src);
  virtual void getExtValueType(extClassID* type, size_t* m, size_t* n, size_t* k) const;
  virtual bool getExtValue(extClassID type, void* values) const;
  virtual bool setExtValue(extClassID type, size_t m, size_t n, size_t k, const void* values);
  virtual void serialize(ostream& str) const;
  virtual bool readSerialized(istream& str);

  /**
   * @brief Get value.
   * @return a C-style string
   */
  const char * get() const;

  /**
   * @brief Set value.
   * @param s a C-style string
   */
  void set(const char *s);

private:

  char *d_s;

};



/**
 * A SFTime value.
 */
class VrmlSFTime : public VrmlField {
public:

  VrmlSFTime(double value = 0.0);

  virtual ostream& print(ostream& os, int indent = 0) const
    { return (os << d_value); }

  virtual VrmlField *clone(VrmlNamespace* ns = 0) const;
  virtual VrmlFieldType fieldType() const;
  virtual bool equals(const VrmlField* f) const;
  bool set(const VrmlField &src);

  /**
   * Get the value.
   */
  double get() const;

  /**
   * Set the value.
   */
  void set(double value);

  virtual void getExtValueType(extClassID* type, size_t* m, size_t* n, size_t* k) const;
  virtual bool getExtValue(extClassID type, void* values) const;
  virtual bool setExtValue(extClassID type, size_t m, size_t n, size_t k, const void* values);
  virtual void serialize(ostream& str) const;
  virtual bool readSerialized(istream& str);

private:
  double d_value;
};



/**
 * This is a base class for fields that represent a fixed-width,
 * variable-height matrix of float numbers (currently VrmlMFFloat,
 * VrmlMFVec2f, VrmlMFVec3f, VrmlMFColor, VrmlMFRotation).
 */
class VrmlMultiFloatField : public VrmlField
{
private:

  size_t width;

protected:

  RefCntArray<float>* data;

public:

  /** Constructor. */
  explicit VrmlMultiFloatField(size_t _width, size_t length=0, const float* values = 0)
    { width = _width; data = RefCntArray<float>::create(length*width, values); }

  /** Copy constructor. */
  VrmlMultiFloatField(const VrmlMultiFloatField &src)
    {
	  width = src.width;
      src.data->incRef();
      data = src.data;
    }

  ~VrmlMultiFloatField()
    { data->decRef(); }

  /** Sets the field value to the given value if it is of the same type. */
  bool set(const VrmlField &src)
    {
      if (src.fieldType() == fieldType())
      {
        set((const VrmlMultiFloatField &) src);
	VrmlField::set(src);
        return true;
      }
      return false;
    }

  /** Sets the field value to the given value. */
  void set(const VrmlMultiFloatField &src)
    {
      VrmlField::set(src);
      src.data->incRef();
      data->decRef();
      data = src.data;
    }

  /** Returns the field data. */
  const float* get() const
    { return data->getValues(); }

  /** Sets the field data. */
  void set(size_t length, const float* values = 0)
    {
      RefCntArray<float>* newData = RefCntArray<float>::create(length*width,
                                                               values);
      data->decRef();
      data = newData;
    }

  /** Returns the length (the number of rows of the matrix). */
  size_t getLength() const
    { return data->getLength()/width; }

  /** Sets the length. */
  void setLength(size_t length)
    {
      RefCntArray<float>* newData = RefCntArray<float>::create(length*width,
                                                               data, 0.0);
      data->decRef();
      data = newData;
    }

  /** Compares this field with the given field. */
  bool equals(const VrmlField* f) const
    {
      if (f->fieldType() != fieldType()) return false;
      return data->equals(static_cast<const VrmlMultiFloatField*>(f)->data);
    }

  ostream& print(ostream& os, int indent = 0) const;


  virtual void getExtValueType(extClassID* type, size_t* m, size_t* n, size_t* k) const;
  virtual bool getExtValue(extClassID type, void* values) const;
  virtual bool setExtValue(extClassID type, size_t m, size_t n, size_t k, const void* values);
  virtual void serialize(ostream& str) const;
  virtual bool readSerialized(istream& str);

};



/**
 * A VrmlMFColor value.
 */
class VrmlMFColor : public VrmlMultiFloatField
{
public:

  explicit VrmlMFColor(size_t length = 0, float const * colors = 0)
    : VrmlMultiFloatField(3, length, colors) { }

  VrmlMFColor(const VrmlMFColor &source)
    : VrmlMultiFloatField(source) { }

  /** Assignment operator. */
  VrmlMFColor &operator=(const VrmlMFColor &src)
    { set(src); return *this; }

  /**
   * @brief Array element dereference operator.
   * @return a pointer to a 3-element array comprising an RGB triplet
   */
  const float * operator[](size_t index) const;

  /**
   * Sets an element of the array.
   */
  void setElement(size_t index, float r, float g, float b);

  virtual VrmlField *clone(VrmlNamespace* ns = 0) const { return new VrmlMFColor(*this); }
  virtual VrmlFieldType fieldType() const { return MFCOLOR; }

  virtual void getExtValueType(extClassID* type, size_t* m, size_t* n, size_t* k) const;
  virtual bool getExtValue(extClassID type, void* values) const;
  virtual bool setExtValue(extClassID type, size_t m, size_t n, size_t k, const void* values);
};



/**
 * A MFFloat value.
 */
class VrmlMFFloat : public VrmlMultiFloatField
{
public:

  /**
   * Constructor.
   * @param length the number of floats in the array
   * @param numbers a pointer to a float array
   */
  explicit VrmlMFFloat(size_t length = 0, float const * numbers = 0)
    : VrmlMultiFloatField(1, length, numbers) { }

  VrmlMFFloat(const VrmlMFFloat &src)
    : VrmlMultiFloatField(src) { }

  /** Assignment operator. */
  VrmlMFFloat& operator=(const VrmlMFFloat &src)
    { set(src); return *this; }

  /** Array element dereference operator. */
  const float operator[](size_t index) const;

  /** Sets an array element. */
  void setElement(size_t index, float f);

  //
  // VrmlField implementation
  //
  virtual ostream& print(ostream& os, int indent = 0) const;
  virtual VrmlField *clone(VrmlNamespace* ns = 0) const { return new VrmlMFFloat(*this); }
  virtual VrmlFieldType fieldType() const { return MFFLOAT; }
};



/**
 * A MFInt32 value.
 */
class VrmlMFInt32 : public VrmlField {
public:

  /**
   * Constructor.
   * @param length The number of integer values.
   * @param numbers Pointer to array of values.
   */
  explicit VrmlMFInt32(size_t length = 0, const long * numbers = 0);

  VrmlMFInt32(const VrmlMFInt32 & mfInt32);
  ~VrmlMFInt32();

  /** Assignment operator. */
  VrmlMFInt32 & operator=(const VrmlMFInt32 & mfInt32);

  /** Array access operator. */
  long operator[](size_t) const;

  bool set(const VrmlField &src);
  const long * get() const;
  void set(size_t length, const long * numbers);
  size_t getLength() const;

  /**
   * Set the length.
   *
   * If the new length is greater than the current length, the additional values
   * are initialized to the default (0.0). If the new length is less
   * than the current length, the array is truncated.
   *
   * @param length new length
   */
  void setLength(size_t length);

  void setElement(size_t index, long i);

  virtual ostream& print(ostream& os, int indent = 0) const;
  virtual VrmlField *clone(VrmlNamespace* ns = 0) const;
  virtual VrmlFieldType fieldType() const;
  virtual bool equals(const VrmlField* f) const;

  virtual void getExtValueType(extClassID* type, size_t* m, size_t* n, size_t* k) const;
  virtual bool getExtValue(extClassID type, void* values) const;
  virtual bool setExtValue(extClassID type, size_t m, size_t n, size_t k, const void* values);
  virtual void serialize(ostream& str) const;
  virtual bool readSerialized(istream& str);

private:

  RefCntArray<long>* data;
};



/**
 * A VrmlMFNode value.
 */
class OPENVRMLDLLEXPORT VrmlMFNode : public VrmlField {
public:
  /** Enumerates options for node occurrences removal */
  enum NodeRemovalOption
  {
    REMOVE_ALL,
    REMOVE_FIRST,
  };

  explicit VrmlMFNode(size_t length = 0, VrmlNode * const * nodes = 0);
  VrmlMFNode(VrmlMFNode const & mfnode);

  ~VrmlMFNode();

  bool set(const VrmlField &src);

  /** Assignment operator. */
  VrmlMFNode& operator=(const VrmlMFNode & mfnode);

  /** Array access operator. */
  VrmlNode * operator[](size_t index) const;

  /** Array access operator. */
  VrmlNode * & operator[](size_t index);

  /** Find out whether current field has got any children.
   * @code return <code>true</code>, if the field has got
   * any children, otherwise, it returns <code>false</code>.
   */
  virtual bool hasChildren() const;


  /** Gets the array length. */
  size_t getLength() const;

  /**
   * @brief Set the length.
   *
   * Set the length of the node array. If the new length is less than the
   * current length, the array is truncated. If the length is greater than
   * the current length, the new positions at the end of the array are set
   * to NULL.
   *
   * @param length the new length
   */
  void setLength(size_t length);

  /** Determine if a node exists in this MFNode. */
  bool exists(const VrmlNode & node) const;

  /**
   * @brief Add a node.
   *
   * Add <var>node</var> to the array if it isn't already part of this MFNode.
   * If <var>siblingNode</var> is <strong>not</strong> NULL and is part of this MFNode,
   * <var>node</node> is placed before it (and another nodes are shifted).
   * This method will <strong>not</strong> add NULLs.
   *
   * @param node a pointer to the node to add
   * @return <code>true</code> if a node was added, <code>false</code>
   *         otherwise
   */
  bool addNode(VrmlNode* node, int where = -1);

  /**
   * @brief Remove a node occurrence(s).
   *
   * Remove <var>node</var> occurrence(s) from the array, if it exists here. This method
   * will <strong>not</strong> remove NULLs.
   *
   * @param node a pointer to the node to remove
   * @param option for removal
   * @return <code>true</code> if a node was removed, <code>false</code>
   *         otherwise
   */
  bool removeNode(VrmlNode * node, int option = REMOVE_ALL);

  /**
   * @brief Remove nodes in a given range.
   * @param startPosition index from where removal is started
   * @param count number of elements/nodes to be removed
   */
  void removeRange(int startPosition, int count);

  /** Write simple scene description into the stream.
   * @param os target stream
   * @param editFlag true if chain is created for 3D World Editor purposes
   * @return resulting stream.
   */
  virtual ostream& toChain(ostream& os, bool editFlag) const;

  virtual ostream& print(ostream& os, int indent = 0) const;

  /**
   * Prints the nodes contained in this field without brackets.
   * This is used to print contents of containers like PROTOs
   * that use a MFNode value to store their nodes but they use
   * a different syntax.
   * @param os A reference to the output stream to be used.
   * @param indent How much indentation.
   * @return A reference to the output stream used.
   */
  virtual ostream &printUnbracketed(ostream &os, int indent) const;

  virtual VrmlField *clone(VrmlNamespace* ns = 0) const;
  virtual VrmlFieldType fieldType() const;
  virtual bool equals(const VrmlField* f) const;


  virtual void getExtValueType(extClassID* type, size_t* m, size_t* n, size_t* k) const;
  virtual bool getExtValue(extClassID type, void* values) const;
  virtual bool setExtValue(extClassID type, size_t m, size_t n, size_t k, const void* values);

private:

  static const size_t REALLOC_SIZE = 16;
  void realloc(size_t newSize);

  VrmlNode * * d_v;
  size_t d_allocated;
  size_t d_size;

  mutable std::vector<const char*> d_names;
};



/**
 * A VrmlMFRotation value.
 */
class VrmlMFRotation : public VrmlMultiFloatField
{
public:

  explicit VrmlMFRotation(size_t length = 0, const float * rotations = 0)
    : VrmlMultiFloatField(4, length, rotations) { }

  VrmlMFRotation(const VrmlMFRotation &src)
    : VrmlMultiFloatField(src) { }

  VrmlMFRotation &operator=(const VrmlMFRotation &src)
    { set(src); return *this; }

  const float * operator[](size_t index) const;
  float* operator[](size_t index);
  void setElement(size_t length, float x, float y, float z, float a);
  virtual VrmlField *clone(VrmlNamespace* ns = 0) const { return new VrmlMFRotation(*this); }
  virtual VrmlFieldType fieldType() const { return MFROTATION; }
};



/**
 * A MFString value.
 */
class OPENVRMLDLLEXPORT VrmlMFString : public VrmlField {
public:

  VrmlMFString(size_t n = 0, char const * const * values = 0);
  VrmlMFString(const VrmlMFString&);

  ~VrmlMFString();

  // Assignment. Just reallocate for now...
  void set(size_t n, char const * const v[]);

  bool set(const VrmlField &src);
  VrmlMFString& operator=(const VrmlMFString& rhs);
  virtual VrmlField *clone(VrmlNamespace* ns = 0) const;
  virtual VrmlFieldType fieldType() const;
  virtual ostream& print(ostream& os, int indent = 0) const;
  virtual ostream& printAsUrl(ostream& os, int indent = 0) const;
  virtual bool equals(const VrmlField* f) const;
  size_t getLength() const;
  char const * const * get() const;
  char const * get(size_t) const;
  void set(size_t, const char *);

  virtual void getExtValueType(extClassID* type, size_t* m, size_t* n, size_t* k) const;
  virtual bool getExtValue(extClassID type, void* values) const;
  virtual bool setExtValue(extClassID type, size_t m, size_t n, size_t k, const void* values);
  virtual void serialize(ostream& str) const;
  virtual bool readSerialized(istream& str);


private:

  char * * d_v;
  size_t d_allocated;
  size_t d_size;

};

/**
 * A MFTime value.
 */
class VrmlMFTime : public VrmlField {
private:

  RefCntArray<double>* data;

public:

  explicit VrmlMFTime(size_t length = 0, const double * times = 0);
  VrmlMFTime(const VrmlMFTime & mftime);
  ~VrmlMFTime();
  bool set(const VrmlField &src);
  VrmlMFTime & operator=(const VrmlMFTime & mftime);

  double operator[](size_t index) const;
  double & operator[](size_t index);

  const double * get() const;
  void set(size_t length, const double * times);

  size_t getLength() const;
  void setLength(size_t length);
  void setElement(size_t index, double t);

  //
  // Override from VrmlField
  //
  ostream & print(ostream &, int indent = 0) const;
  VrmlField * clone(VrmlNamespace* ns = 0) const;
  VrmlFieldType fieldType() const;
  bool equals(const VrmlField* f) const;

  virtual void getExtValueType(extClassID* type, size_t* m, size_t* n, size_t* k) const;
  virtual bool getExtValue(extClassID type, void* values) const;
  virtual bool setExtValue(extClassID type, size_t m, size_t n, size_t k, const void* values);
  virtual void serialize(ostream& str) const;
  virtual bool readSerialized(istream& str);
};



/**
 * A MFVec2f value.
 */
class VrmlMFVec2f : public VrmlMultiFloatField
{
public:

  explicit VrmlMFVec2f(size_t length = 0, const float * vecs = 0)
    : VrmlMultiFloatField(2, length, vecs) { }

  VrmlMFVec2f(const VrmlMFVec2f &src)
    : VrmlMultiFloatField(src) { }

  VrmlMFVec2f & operator=(const VrmlMFVec2f &src)
    { set(src); return *this; }

  const float * operator[](size_t index) const;
  float * operator[](size_t index);
  void setElement(size_t length, float x, float y);
  virtual VrmlField *clone(VrmlNamespace* ns = 0) const { return new VrmlMFVec2f(*this); }
  virtual VrmlFieldType fieldType() const { return MFVEC2F; }
};



/**
 * A VrmlMFVec3f value.
 */
class VrmlMFVec3f : public VrmlMultiFloatField
{
public:

  explicit VrmlMFVec3f(size_t length = 0, const float * vecs = 0)
    : VrmlMultiFloatField(3, length, vecs) { }

  VrmlMFVec3f(const VrmlMFVec3f &src)
    : VrmlMultiFloatField(src) { }

  VrmlMFVec3f &operator=(const VrmlMFVec3f &src)
    { set(src); return *this; }

  const float* operator[](size_t index) const;
  float* operator[](size_t index);
  void setElement(size_t index, float x, float y, float z);
  virtual VrmlField *clone(VrmlNamespace* ns = 0) const { return new VrmlMFVec3f(*this); }
  virtual VrmlFieldType fieldType() const { return MFVEC3F; }
};



/**
The StreamReader reads a block of encoded structured data.
The data format is compatible with Java DataOutputWriter.
*/
class OPENVRMLDLLEXPORT StreamReader
{
private:
  istream& str;

public:
  /**
  Creates a new BufferReader for reading for the input stream.
  */
  StreamReader(istream& _str) :
    str(_str)
  { };

  /**
  Reads a single byte from the buffer.
  */
  int ReadByte()
  { return str.eof() ? 0 : (unsigned char) str.get(); };

  /**
  Reads a single boolean value from the buffer.
  */
  bool ReadBool();

  /**
  Reads a 16-bit signed integer value from the buffer.
  */
  short ReadInt16();

  /**
  Reads a 32-bit signed integer value from the buffer.
  */
  int ReadInt32();

  /**
  Reads a float value from the buffer.
  */
  float ReadFloat();

  /**
  Reads a double value from the buffer.
  */
  double ReadDouble();

  /**
  Reads a string from the buffer and returns it as a sequence of bytes
  allocated by new[]. The caller is responsible for freeing the string
  when it's no longer needed.
  */
  char *ReadString();
};



/**
The StreamWriter is used to write structured data to a buffer.
The data format is compatible with Java DataOutputWriter.
*/
class OPENVRMLDLLEXPORT StreamWriter
{
private:
  ostream& str;

public:
  /**
  Creates a new StreamWriter for the output stream.
  */
  StreamWriter(ostream& _str) :
    str(_str)
  { };

  /**
  Writes a single byte to the buffer.
  */
  void WriteByte(char n)
  { str.put(n); };

  /**
  Writes a single boolean value to the buffer.
  */
  void WriteBool(bool value);

  /**
  Writes a given 16-bit signed integer number to the buffer.
  */
  void WriteInt16(short n);

  /**
  Writes a given 32-bit signed integer number to the buffer.
  */
  void WriteInt32(int n);

  /**
  Writes a given float value to the buffer.
  */
  void WriteFloat(float n);

  /**
  Writes a given double value to the buffer.
  */
  void WriteDouble(double n);

  /**
  Writes a given string to the buffer.
  */
  void WriteString(const char *str);

  /**
  Writes a given printf-like formatted string to the buffer.
  */
  void WriteFormattedString(const char* format, ...);
};


#endif
