/*****************************************************************************
******************************************************************************
*
*               Format-independent API for loading textures
*
*               $Revision: 1.1.2.10 $
*               $Date: 2012/02/27 09:13:05 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright (C) 1998  Chris Morley
// Modified by HUMUSOFT s.r.o.  Copyright 1998-2012 HUMUSOFT s.r.o.

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "Image.h"
#include "Doc.h"
#include "System.h"
#include "VrmlField.h"

// use Boost for reading JPEG
#include <boost/gil/extension/io/jpeg_io.hpp>

// Boost PNG library does not support paletted PNG, so use our own reader
#include "VrmlPNGRead.h"

#include <stdlib.h>		// free()
#include <string.h>


// gif support is builtin
#include "gifread.h"
#include "mpgread.h"

typedef enum {
  ImageFile_UNKNOWN,

  ImageFile_GIF,
  ImageFile_JPG,
  ImageFile_MPG,
  ImageFile_PNG

} ImageFileType;


static ImageFileType imageFileType(const char *, FILE *);


Image::Image(const char *url) :
  d_url(0),
  d_w(0), d_h(0), d_nc(0), d_nFrames(0), d_pixels(0), d_frame(0)
{
  if (url) (void) setURL(url);
}

Image::Image(const Image& other) :
  d_url(0), d_pixels(0), d_frame(0)
{
  *this = other;
}

Image::~Image()
{
  delete d_url;
  if (d_pixels) free(d_pixels); // assumes file readers use malloc...
  if (d_frame) free(d_frame);
}

Image& Image::operator=(const Image& other)
{
  d_w = other.d_w;
  d_h = other.d_h;
  d_nc = other.d_nc;
  d_nFrames = other.d_nc;

  delete d_url;
  d_url = (other.d_url) ? new Doc(other.d_url) : 0;

  if (d_pixels)
  {
    free(d_pixels);
    d_pixels = 0;
  }

  if (d_frame)
  {
    free(d_frame);
    d_frame = 0;
  }

  return *this;
}


bool Image::setURL(const char *url, Doc *relative)
{
  if (d_url) delete d_url;
  if (d_pixels) free(d_pixels); // assumes file readers use malloc...
  if (d_frame) free(d_frame);
  d_url = 0;
  d_pixels = 0;
  d_frame = 0;
  d_w = d_h = d_nc = d_nFrames = 0;
  if (!url) return true;

  d_url = new Doc(url, relative);

  //  theSystem->debug("Image: trying to create Doc(%s, %s)\n",
  //		   url, relative ? relative->url() : "");

  FILE *fp = d_url->fopen("rb");

  if (fp)
  {
    switch (imageFileType(url, fp))
    {
      case ImageFile_GIF:
        d_pixels = gifread(fp, &d_w, &d_h, &d_nc, &d_nFrames, &d_frame);
      break;

      case ImageFile_PNG:
        d_pixels = pngread(fp, &d_w, &d_h, &d_nc);
      break;

      case ImageFile_JPG:
      {
        // read the source image and create its constant view
        boost::gil::rgb8_image_t img;
        try
        {
          jpeg_read_and_convert_image(d_url->localName(), img);
        }
        catch(const std::ios_base::failure&)
        {
          // end now and let the code below report the error
          break;
        }
        boost::gil::rgb8c_view_t src = const_view(img);

        // create the destination image and create its view
        d_w = src.width();
        d_h = src.height();
        d_nc = src.num_channels();
        d_pixels = static_cast<unsigned char*>(malloc(d_w*d_h*d_nc));
        boost::gil::rgb8_view_t dst = boost::gil::interleaved_view(d_w, d_h, reinterpret_cast<boost::gil::rgb8_ptr_t>(d_pixels), d_w*d_nc);

        // copy the source view pixels to destination, flipping the source up/down
        copy_pixels(boost::gil::flipped_up_down_view(src), dst);
      }
      break;

      case ImageFile_MPG:
        d_pixels = mpgread(fp, &d_w, &d_h, &d_nc, &d_nFrames, &d_frame);
      break;

      default:
        theSystem->error(System::ERRORTYPE_UNKNOWN, "Error: unrecognized image file format (%s).\n", url);
      break;
    }

    if (! d_pixels)
      theSystem->error(System::ERRORTYPE_UNKNOWN, "Error reading image file \"%s%s\"\n", d_url->urlPath(), url);

    d_url->fclose();
  }

  return (d_pixels != 0);
}

bool Image::tryURLs(size_t nUrls, char const * const * urls, Doc * relative)
{
    size_t i(0);
    for (; i < nUrls; ++i) {  // Try each url until one succeeds
        if (urls[i] && setURL(urls[i], relative)) {
            break;
        }
    }

    return i < nUrls;
}

bool Image::tryURLs(const VrmlMFString & urls, Doc * relative) {
    size_t i(0);
    for (; i < urls.getLength(); ++i) {
        if (urls.get(i) && setURL(urls.get(i), relative)) {
            break;
        }
    }

    return (i < urls.getLength());
}

const char *Image::url() { return d_url ? d_url->url() : 0; }


// Could peek at file header...

static ImageFileType imageFileType(const char *url, FILE *)
{
  const char *suffix = strrchr(url, '.');
  if (suffix==0)
    return ImageFile_UNKNOWN;

  ++suffix;
  if (strcasecmp(suffix,"gif") == 0)
    return ImageFile_GIF;

  else if (strcasecmp(suffix,"jpg") == 0 ||
	   strcasecmp(suffix,"jpeg") == 0)
    return ImageFile_JPG;

  else if (strcasecmp(suffix,"mpg") == 0 ||
	   strcasecmp(suffix,"mpeg") == 0)
    return ImageFile_MPG;

  else if (strcasecmp(suffix,"png") == 0)
    return ImageFile_PNG;

  else
    return ImageFile_UNKNOWN;
}


unsigned char *Image::pixels(int frame)
{
  return (frame < d_nFrames) ? d_frame[ frame ] : 0;
}

