/*****************************************************************************
******************************************************************************
*
*               VRML namespace
*
*               $Revision: 1.1.2.11 $
*               $Date: 2011/11/30 13:27:59 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright (C) 1998  Chris Morley
// Modified by HUMUSOFT s.r.o.  Copyright 1998-2011 HUMUSOFT s.r.o.

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "VrmlNamespace.h"
#include "VrmlNodeType.h"
#include "System.h"
#include "VrmlSceneFromNodeType.h"

#include <cstdlib>
#include <string.h>

// This should at least be a sorted vector...
std::list< VrmlNodeType* > VrmlNamespace::builtInList;

// Ref count of namespaces so builtins can be freed
int VrmlNamespace::s_nNamespaces = 0;

VrmlNamespace::VrmlNamespace( VrmlNamespace *parent ) :
  d_parent(parent),
  uniqueSequence(1)
{
  ++s_nNamespaces;

  // Initialize typeList with built in nodes
  if (builtInList.size() == 0) defineBuiltIns();
}

VrmlNamespace::~VrmlNamespace()
{
  // Free nameList
#ifndef __ORBISNAP__     // do not dereference nodes while destroying namelist
  NodesList::iterator n;
  for (n = d_nameList.begin(); n != d_nameList.end(); ++n)
    (*n).nodePtr->dereference();
#endif

  // Free typeList
  std::list<VrmlNodeType*>::iterator i;
  for (i = d_typeList.begin(); i != d_typeList.end(); ++i)
    (*i)->dereference();

#ifndef __ORBISNAP__     //;!! temporarily disabled because of Unix crash - should it be here at all?
  // Free builtins
  if ( --s_nNamespaces == 0 )
    {
      for (i = builtInList.begin(); i != builtInList.end(); ++i)
        (*i)->dereference();
      builtInList.clear();
    }
#endif
}


//
//  Built in nodes.
//  This code replaces the reading of the "standardNodes.wrl" file
//  of empty PROTOs so I don't need to carry that file around.
//

void
VrmlNamespace::addBuiltIn( VrmlNodeType *type)
{
  builtInList.push_back( type->reference() );
}


#include "VrmlNodeAnchor.h"
#include "VrmlNodeAppearance.h"
#include "VrmlNodeAudioClip.h"
#include "VrmlNodeBackground.h"
#include "VrmlNodeBillboard.h"
#include "VrmlNodeBox.h"
#include "VrmlNodeCollision.h"
#include "VrmlNodeColor.h"
#include "VrmlNodeColorInt.h"
#include "VrmlNodeCone.h"
#include "VrmlNodeCoordinate.h"
#include "VrmlNodeCoordinateInt.h"
#include "VrmlNodeCylinder.h"
#include "VrmlNodeCylinderSensor.h"
#include "VrmlNodeDirLight.h"
#include "VrmlNodeElevationGrid.h"
#include "VrmlNodeExtrusion.h"
#include "VrmlNodeFog.h"
#include "VrmlNodeFontStyle.h"
#include "VrmlNodeGroup.h"
#include "VrmlNodeIFaceSet.h"
#include "VrmlNodeILineSet.h"
#include "VrmlNodeImageTexture.h"
#include "VrmlNodeInline.h"
#include "VrmlNodeLOD.h"
#include "VrmlNodeMaterial.h"
#include "VrmlNodeMovieTexture.h"
#include "VrmlNodeNavigationInfo.h"
#include "VrmlNodeNormal.h"
#include "VrmlNodeNormalInt.h"
#include "VrmlNodeOrientationInt.h"
#include "VrmlNodePixelTexture.h"
#include "VrmlNodePlaneSensor.h"
#include "VrmlNodePointLight.h"
#include "VrmlNodePointSet.h"
#include "VrmlNodePositionInt.h"
#include "VrmlNodeProto.h"
#include "VrmlNodeProximitySensor.h"
#include "VrmlNodeScalarInt.h"
#include "VrmlNodeScript.h"
#include "VrmlNodeShape.h"
#include "VrmlNodeSound.h"
#include "VrmlNodeSphere.h"
#include "VrmlNodeSphereSensor.h"
#include "VrmlNodeSpotLight.h"
#include "VrmlNodeSwitch.h"
#include "VrmlNodeText.h"
#include "VrmlNodeTextureCoordinate.h"
#include "VrmlNodeTextureTransform.h"
#include "VrmlNodeTimeSensor.h"
#include "VrmlNodeTouchSensor.h"
#include "VrmlNodeTransform.h"
#include "VrmlNodeViewpoint.h"
#include "VrmlNodeVisibilitySensor.h"
#include "VrmlNodeWorldInfo.h"


void VrmlNamespace::defineBuiltIns()
{
  addBuiltIn( VrmlNodeAnchor::defineType() );
  addBuiltIn( VrmlNodeAppearance::defineType() );
  addBuiltIn( VrmlNodeAudioClip::defineType() );
  addBuiltIn( VrmlNodeBackground::defineType() );
  addBuiltIn( VrmlNodeBillboard::defineType() );
  addBuiltIn( VrmlNodeBox::defineType() );
  addBuiltIn( VrmlNodeCollision::defineType() );
  addBuiltIn( VrmlNodeColor::defineType() );
  addBuiltIn( VrmlNodeColorInt::defineType() );
  addBuiltIn( VrmlNodeCone::defineType() );
  addBuiltIn( VrmlNodeCoordinate::defineType() );
  addBuiltIn( VrmlNodeCoordinateInt::defineType() );
  addBuiltIn( VrmlNodeCylinder::defineType() );
  addBuiltIn( VrmlNodeCylinderSensor::defineType() );
  addBuiltIn( VrmlNodeDirLight::defineType() );
  addBuiltIn( VrmlNodeElevationGrid::defineType() );
  addBuiltIn( VrmlNodeExtrusion::defineType() );
  addBuiltIn( VrmlNodeFog::defineType() );
  addBuiltIn( VrmlNodeFontStyle::defineType() );
  addBuiltIn( VrmlNodeGroup::defineType() );
  addBuiltIn( VrmlNodeIFaceSet::defineType() );
  addBuiltIn( VrmlNodeILineSet::defineType() );
  addBuiltIn( VrmlNodeImageTexture::defineType() );
  addBuiltIn( VrmlNodeInline::defineType() );
  addBuiltIn( VrmlNodeLOD::defineType() );
  addBuiltIn( VrmlNodeMaterial::defineType() );
  addBuiltIn( VrmlNodeMovieTexture::defineType() );
  addBuiltIn( VrmlNodeNavigationInfo::defineType() );
  addBuiltIn( VrmlNodeNormal::defineType() );
  addBuiltIn( VrmlNodeNormalInt::defineType() );
  addBuiltIn( VrmlNodeOrientationInt::defineType() );
  addBuiltIn( VrmlNodePixelTexture::defineType() );
  addBuiltIn( VrmlNodePlaneSensor::defineType() );
  addBuiltIn( VrmlNodePointLight::defineType() );
  addBuiltIn( VrmlNodePointSet::defineType() );
  addBuiltIn( VrmlNodePositionInt::defineType() );
  addBuiltIn( VrmlNodeProximitySensor::defineType() );
  addBuiltIn( VrmlNodeScalarInt::defineType() );
  addBuiltIn( VrmlNodeScript::defineType() );
  addBuiltIn( VrmlNodeShape::defineType() );
  addBuiltIn( VrmlNodeSound::defineType() );
  addBuiltIn( VrmlNodeSphere::defineType() );
  addBuiltIn( VrmlNodeSphereSensor::defineType() );
  addBuiltIn( VrmlNodeSpotLight::defineType() );
  addBuiltIn( VrmlNodeSwitch::defineType() );
  addBuiltIn( VrmlNodeText::defineType() );
  addBuiltIn( VrmlNodeTextureCoordinate::defineType() );
  addBuiltIn( VrmlNodeTextureTransform::defineType() );
  addBuiltIn( VrmlNodeTimeSensor::defineType() );
  addBuiltIn( VrmlNodeTouchSensor::defineType() );
  addBuiltIn( VrmlNodeTransform::defineType() );
  addBuiltIn( VrmlNodeViewpoint::defineType() );
  addBuiltIn( VrmlNodeVisibilitySensor::defineType() );
  addBuiltIn( VrmlNodeWorldInfo::defineType() );
}


// A safer version for reading PROTOs from files.

void
VrmlNamespace::addNodeType(VrmlNodeType* type, bool pushBack)
{
  if ( findType( type->getName() ) != NULL)
    theSystem->warn(System::SEVERITY_HIGH, System::WARNINGTYPE_UNKNOWN, "PROTO %s already defined\n",
                    type->getName() );
  else
    if (pushBack)
      d_typeList.push_back( type->reference() );
    else
      d_typeList.push_front( type->reference() );
}


const VrmlNodeType *
VrmlNamespace::findType( const char *name )
{
  // Look through the PROTO stack:
  const VrmlNodeType *nt = findPROTO(name);
  if (nt) return nt;

  // Look in parent scope for the type
  if (d_parent)
    return d_parent->findType( name );

  // Look through the built ins
  std::list<VrmlNodeType*>::iterator i;
  for (i = builtInList.begin(); i != builtInList.end(); ++i)
    {
      nt = *i;
      if (nt != NULL && strcmp(nt->getName(),name) == 0)
        return nt;
    }

  return NULL;
}

const VrmlNodeType *    // LarryD
VrmlNamespace::findPROTO(const char *name)
{
  // Look through the PROTO list ONLY:
  std::list<VrmlNodeType*>::iterator i;
  for (i = d_typeList.begin(); i != d_typeList.end(); ++i)
    {
      const VrmlNodeType *nt = *i;
      if (nt != NULL && strcmp(nt->getName(),name) == 0)
        return nt;
    }
  return NULL;
}

void VrmlNamespace::removePROTOandItsInstances(VrmlNodeType* typeToRemove, VrmlNodeType* fromWhereType)
{
  // PROTO instances can occur within another PROTO declaration
  for (std::list<VrmlNodeType*>::reverse_iterator ri = d_typeList.rbegin(); ri != d_typeList.rend(); ++ri)
    (*ri)->scope()->removePROTOandItsInstances(typeToRemove, (*ri));

  NodesList::iterator n = d_nameList.begin();
  while (n != d_nameList.end())
  {
    NodesList::iterator m = n;
    ++n;
    if (&((*m).nodePtr->nodeType())==typeToRemove)
    {
      if ((*m).nodePtr->scene())
        (*m).nodePtr->scene()->removeNode((*m).nodePtr);
      else
      {
	// create helper scene from node type
	VrmlSceneFromNodeType sceneFromNodeType(const_cast<VrmlNodeType*> (fromWhereType), 0);
	sceneFromNodeType.removeNode((*m).nodePtr);
      }
    }
  }
  d_typeList.remove(const_cast<VrmlNodeType*> (typeToRemove));
}

void VrmlNamespace::renamePROTOInstancesField(const VrmlNodeType* type, const char* fieldName, const char* newName)
{
  // PROTO instances can occur within another PROTO declaration
  for (std::list<VrmlNodeType*>::reverse_iterator ri = d_typeList.rbegin(); ri != d_typeList.rend(); ++ri)
    (*ri)->scope()->renamePROTOInstancesField(type, fieldName, newName);

  for (NodesList::iterator n = d_nameList.begin(); n != d_nameList.end(); ++n)
  {
    if (&((*n).nodePtr->nodeType())==type)
    {
      VrmlNodeProto::NameValueRec* nv = (dynamic_cast<VrmlNodeProto*> ((*n).nodePtr))->findField(fieldName);
      if (nv)
      {
	free(nv->name);
        nv->name = strdup(newName);
      }
    }
  }
}

void VrmlNamespace::reinstantiatePROTOInstances(VrmlNodeType* type)
{
  // PROTO instances can occur within another PROTO declaration
  // these instances need to be reinstantiated first (recursive call)
  for (std::list<VrmlNodeType*>::reverse_iterator ri = d_typeList.rbegin(); ri != d_typeList.rend(); ++ri)
    (*ri)->scope()->reinstantiatePROTOInstances(type);

  // reinstantiate instances in the given namespace
  for (NodesList::iterator n = d_nameList.begin(); n != d_nameList.end(); ++n)
  {
    if (&((*n).nodePtr->nodeType())==type)
      (*n).nodePtr->reinstantiate();

    type->clearModified();
  }
}

const VrmlNodeType *
VrmlNamespace::firstType()
{
  // Top of the PROTO stack (should make sure it has an implementation...)
  if (d_typeList.size() > 0)
    return d_typeList.front()->reference();
  return NULL;
}

void
VrmlNamespace::addNodeName( VrmlNode *namedNode )
{
  NodesList::nth_index<2>::type& node_index = d_nameList.get<2>();
  NodesList::nth_index<2>::type::iterator i = node_index.find(namedNode);
  if (i==node_index.end())
    d_nameList.push_back(NodeListValue(namedNode));
  else
    node_index.replace(i, NodeListValue(const_cast<VrmlNode*>((*i).nodePtr)));

  /*NodesList::nth_index<1>::type& name_index = d_nameList.get<1>();
  NodesList::nth_index<1>::type::iterator j = d_nameList.project<1>(i);
  if (j!=name_index.end())
    name_index.replace(j, NodeListValue(const_cast<VrmlNode*>((*j).nodePtr)));*/

}

void VrmlNamespace::removeNodeName( VrmlNode *node )
{
  NodesList::nth_index<2>::type::iterator i = d_nameList.get<2>().find(node);
  if (i!=d_nameList.get<2>().end())
  {
    NodesList::nth_index<2>::type& node_index = d_nameList.get<2>();
    node_index.erase(i);
  }

  // remove all routes from/to this node
  d_routeList.RemoveNodeNotify(node);

#ifndef __ORBISNAP__     // do not dereference node while removing from namelist
  namedNode->dereference();
#endif
}

VrmlNode * VrmlNamespace::findNode( const char *name, const VrmlNode* excludedNode ) const
{
  NodesList::nth_index<1>::type::iterator i = d_nameList.get<1>().find(name);
  if (i!=d_nameList.get<1>().end() && (*i).nodePtr!=const_cast<VrmlNode*>(excludedNode))
    return (*i).nodePtr;
  else
    return 0;
}

bool VrmlNamespace::containsNode(VrmlNode* node) const
{
  NodesList::nth_index<2>::type::iterator i = d_nameList.get<2>().find(node);
  return (i!=d_nameList.get<2>().end());
}

unsigned long long VrmlNamespace::getUniqueSequence()
{
  return uniqueSequence++;
}

Route& VrmlNamespace::addRoute(VrmlNode* fromNode, const char* fromEventOut, VrmlNode* toNode, const char* toEventIn)
{
  return(d_routeList[d_routeList.Add(fromNode, fromEventOut, toNode, toEventIn)]);
}

void VrmlNamespace::removeRoute(VrmlNode* fromNode, const char *fromEventOut, VrmlNode* toNode, const char* toEventIn)
{
  d_routeList.Remove(fromNode, fromEventOut, toNode, toEventIn);
}
