/*****************************************************************************
******************************************************************************
*
*               VRML namespace
*
*               $Revision: 1.1.2.10 $
*               $Date: 2011/11/30 13:27:59 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright (C) 1998  Chris Morley
// Modified by HUMUSOFT s.r.o.  Copyright 1998-2011 HUMUSOFT s.r.o.

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#ifndef VRMLNAMESPACE_H
#define VRMLNAMESPACE_H

//
// The VrmlNamespace class keeps track of defined nodes and prototypes.
//

#include "VrmlNode.h"
#include "VrmlRoute.h"
#include "System.h"   // OPENVRMLDLLEXPORT

#include <list>
using namespace std;

// disable serialization of boost container
#define BOOST_MULTI_INDEX_DISABLE_SERIALIZATION

#include <boost/multi_index_container.hpp>
#include <boost/multi_index/sequenced_index.hpp>
#include <boost/multi_index/ordered_index.hpp>
#include <boost/multi_index/member.hpp>

using namespace boost::multi_index;

struct NodeListValue
{
  VrmlNode* nodePtr;
  std::string name;

  NodeListValue(VrmlNode* aNodePtr) : nodePtr(aNodePtr), name(aNodePtr->name()) {}
};

typedef multi_index_container<NodeListValue,
	                      indexed_by<sequenced<>,
	                                 ordered_unique<BOOST_MULTI_INDEX_MEMBER(NodeListValue, std::string, name)>,
					 ordered_unique<BOOST_MULTI_INDEX_MEMBER(NodeListValue, VrmlNode*, nodePtr)>
	                                > > NodesList;

class VrmlNodeType;

class OPENVRMLDLLEXPORT VrmlNamespace {
public:

  VrmlNamespace( VrmlNamespace *parent = 0 );
  ~VrmlNamespace();

  // PROTO definitions add node types to the namespace.
  // PROTO implementations are a separate node type namespace,
  // and require that any nested PROTOs NOT be available outside
  // the PROTO implementation. PROTOs defined outside the current
  // namespace are available.

  // addNodeType will print an error if the given type
  // is already defined (spec says behavior is undefined).
  void addNodeType(VrmlNodeType* type, bool pushBack = true);

  // DEFd nodes add node names to the namespace.
  // Node names are only defined in the current name space. They
  // are not available outside of the PROTO they are defined in,
  // nor are they available inside of nested PROTOs.

  void addNodeName(VrmlNode *);
  void removeNodeName(VrmlNode *);

  // Find a node type, given a type name. Returns NULL if type is not defined.
  const VrmlNodeType *findType(const char *nm);

  // Larry Feb 10_99  Find a nodeType, given a PROTO name
  const VrmlNodeType *findPROTO(const char *nm);

  // removes VrmlNodeType from the type list (also delete its instances)
  void removePROTOandItsInstances(VrmlNodeType* typeToRemove, VrmlNodeType* fromWhereType = 0);

  // renames field in all PROTO instances of a given type
  void renamePROTOInstancesField(const VrmlNodeType* type, const char* fieldName, const char* newName);

  // reinstantiates all PROTO instances
  void reinstantiatePROTOInstances(VrmlNodeType* type);

  // Return the first node type in scope (default EXTERNPROTO implementation)
  const VrmlNodeType *firstType();

  // Find a node by name. Excluded node is not considered during the search.
  VrmlNode *findNode(const char *name, const VrmlNode* excludedNode = 0) const;

  // get information about node is in namespace
  bool containsNode(VrmlNode* node) const;

  // sets and gets parent namespace
  void setParent(VrmlNamespace* parent) { d_parent = parent; }
  VrmlNamespace* getParent() { return d_parent; }

  // return unique identifier
  unsigned long long getUniqueSequence();

  list<VrmlNodeType*> const* getTypeList() const
    { return &d_typeList; };


  NodesList const* getNamedNodes() const
    { return &d_nameList; }

  /** Adds a route between two events. */
  Route& addRoute(VrmlNode* fromNode, const char* fromEventOut,
                VrmlNode* toNode, const char* toEventIn);

  /** Removes a route. */
  void removeRoute(VrmlNode* fromNode, const char* fromEventOut,
                   VrmlNode* toNode, const char* toEventIn);

  /** Returns a first route in the route list. */
  RouteList& getRouteList()
    { return d_routeList; }

private:

  friend class VrmlNode;

  void addBuiltIn(VrmlNodeType *);
  void defineBuiltIns();

  // Builtin node types are stored (once) in this data structure:
  static list< VrmlNodeType* > builtInList;

  // Ref count of namespaces so builtins can be freed
  static int s_nNamespaces;

  // Defined node types (PROTOs) for this namespace
  list<VrmlNodeType*> d_typeList;

  // Defined nodes for this namespace
  NodesList d_nameList;

  // Parent namespace
  VrmlNamespace *d_parent;

  // Routes in this namespace
  RouteList d_routeList;

  unsigned long long uniqueSequence;
};

#endif
