/*****************************************************************************
******************************************************************************
*
*               VRML node
*
*               $Revision: 1.1.2.36 $
*               $Date: 2011/12/12 20:16:54 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright 1998-2011 HUMUSOFT s.r.o.
// Portions Copyright (C) 1998  Chris Morley

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "VrmlNode.h"
#include "VrmlNamespace.h"
#include "VrmlNodeType.h"
#include "VrmlScene.h"
#include "MathUtils.h"
#include "VrmlBVolume.h"
#include "VrmlBSphere.h"
#include "VrmlRenderContext.h"
#include "VrmlField.h"
#include "VrmlNodeProto.h"

#include <stdio.h>		// snprintf

#ifndef NDEBUG
//#   define VRML_NODE_DEBUG
#endif

#define TERMINAL_STRING '#'
#define SYNTH_NAME_BEGIN_STRING ','

VrmlNode::VrmlNode(VrmlScene *scene) :
  d_scene(scene),
  d_modified(false),
  d_refCount(0),
  d_name(0),
  d_proto(0),
  d_protoType(0),
  d_namespace(0),
  d_flag(false)
{
  this->setBVolumeDirty(true);
}

VrmlNode::VrmlNode( const VrmlNode & ) :
  d_scene(0), d_modified(true), d_refCount(0), d_name(0), d_proto(0), d_protoType(0),
  d_namespace(0), d_flag(false)
{
  this->setBVolumeDirty(true);
}


// Free name (if any) and route info.

VrmlNode::~VrmlNode()
{
  if (d_namespace) d_namespace->removeNodeName(this);

  // free the IS binding list
  for (int i=0; i<d_bindings.size(); i++)
  {
    if (d_protoType) d_protoType->removeIS(d_bindings[i]->getNode(), d_bindings[i]->getFieldName());
    delete d_bindings[i];
  }

  delete[] d_name;
}

VrmlNode* VrmlNode::newInstance(VrmlNamespace* ns)
{
  const char* validname = validName();

  VrmlNode* n;
  // use node
  if (validname && isFlagSet() && (n = ns->findNode(validname)) != NULL)
  {
    n->setFlag();
    return n;
  }

  n = this->cloneMe();
  if (n)
  {
    // check whether node with the same name already exists in target name space
    if (validname)
    {
      if (ns->findNode(validname))
      {
        n->setName(0, ns);
        theSystem->warn(System::SEVERITY_HIGH, System::WARNINGTYPE_UNKNOWN, "Node with name \"%s\" already exists in the target world.", validname);
      }
      else
        n->setName(validname, ns);

      setFlag();
    }
    else
    {
      if (ns->findNode(d_name))
        n->setName(0, ns); // node is "unnamed", do not inform about existence of node with the same name
      else
        n->setName(d_name, ns);
    }
  }

  return n;
}


VrmlNode *VrmlNode::clone(VrmlNamespace *ns)
{
  VrmlNode* n = newInstance(ns);
  if (n && !(n->isFlagSet()))
  {
    // node can be part of VrmlNodeType (part of PROTO definition)
    // we copy IS mappings
    for (int i=0; i<d_bindings.size(); ++i)
      n->d_bindings.push_back(new FieldISBinding(n, d_bindings[i]->getFieldName(), d_bindings[i]->getInterfaceName()));

    n->cloneChildren(ns);
  }

  return n;
}


void VrmlNode::cloneChildren( VrmlNamespace* ) {}


VrmlNode* VrmlNode::instantiate(VrmlNamespace* ns)
{
  VrmlNode* n = newInstance(ns);
  if (n && !(n->isFlagSet()))
    n->cloneChildren(ns);

  return n;
}

void VrmlNode::reinstantiate()
{}


VrmlNode *VrmlNode::reference()
{
 ++d_refCount;
 return this;
}

// Remove a reference to a node

void VrmlNode::dereference()
{
  if (--d_refCount == 0) delete this;
}

// Return the reference count
int VrmlNode::getRefCount()
{
  return d_refCount;
}

// Set the name of the node. Some one else (the parser) needs
// to tell the scene about the name for use in USE/ROUTEs.

void VrmlNode::setName(const char *nodeName, VrmlNamespace *ns)
{
  delete[] d_name;

  if (nodeName && *nodeName)
  {
    d_name = new char[strlen(nodeName)+1];
    strcpy(d_name, nodeName);
  }
  else
  {
    // use synthetic name
    const size_t SYNTHNAMESIZE = 18;
    d_name = new char[SYNTHNAMESIZE];
    if (ns)
      do
      {
        snprintf(d_name, SYNTHNAMESIZE, ",%016X", ns->getUniqueSequence()); //;!! SYNTH_NAME_BEGIN_STRING hardcoded here
      } while (ns->findNode(d_name, this));
  }

  // add node to namespace
  if (ns) ns->addNodeName(this);

  d_namespace = ns;
}

// Retrieve the name of this node.

const char *VrmlNode::name() const
{
  return d_name ? d_name : "";
}

const char *VrmlNode::validName() const
{
  return (d_name && d_name[0]!=SYNTH_NAME_BEGIN_STRING) ? d_name : 0;
}

// Add to scene

void VrmlNode::addToScene( VrmlScene *scene, const char * /* relativeUrl */ )
{
  d_scene = scene;
}


// Dirty bit - indicates node needs to be revisited for rendering.

void VrmlNode::setModified()
{
  d_modified = true;
  if (d_scene) d_scene->setModified();
  if (d_protoType && ((d_proto!=0 && d_proto->isInstantiated()) || d_proto==0)) d_protoType->setModified();
}

void VrmlNode::invalidate()
{
  setModified();
}

bool VrmlNode::isModified() const
{
  return d_modified;
}


void
VrmlNode::markPathModified(VrmlNodePath& path, bool mod, int flags) {
  VrmlNodePath::iterator i;
  VrmlNodePath::iterator end = path.end();
  if (flags & 0x001) {
    //cout << "VrmlNode[]::markPathModified():mod" << endl;
    for (i = path.begin(); i != end; ++i) {
      VrmlNode *c = *i;
      if (mod) {
	// do the proof that our invarient shows that this short
	// circuit is legal...
	//if (c->isModified())
	//break;
	c->setModified();
      } else
	c->clearModified();
    }
  }
  if (flags & 0x002) {
    //cout << "VrmlNode[]::markPathModified():bvol" << endl;
    for (i = path.begin(); i != end; ++i) {
      VrmlNode *c = *i;
      if (mod) {
	c->setBVolumeDirty(true);
      } else
	c->setBVolumeDirty(false);
    }
  }
}


void
VrmlNode::updateModified(VrmlNodePath& path, int flags)
{
  //cout << "VrmlNode[" << this << "]::updateModified()" << endl;
  if (this->d_modified||this->d_bvol_dirty) markPathModified(path, true, flags);
}


// note not virtual
//
void
VrmlNode::updateModified(int flags)
{
  VrmlNodePath path;
  updateModified(path, flags);
}


void VrmlNode::clearFlags()
{
  d_flag = false;
}

const VrmlBVolume*
VrmlNode::getBVolume() const
{
  //cout << "VrmlNode::getBVolume():inf" << endl;
  static VrmlBSphere* inf_bsphere = (VrmlBSphere*)0;
  if (!inf_bsphere) {
    inf_bsphere = new VrmlBSphere();
    inf_bsphere->setMAX();
  }
  ((VrmlNode*)this)->setBVolumeDirty(false);
  return inf_bsphere;
}


void
VrmlNode::setBVolume(const VrmlBVolume& v)
{
  cout << "VrmlNode::setBVolume():WARNING:not implemented" << endl;
}

void
VrmlNode::setBVolumeDirty(bool f)
{
  this->d_bvol_dirty = f;
  if (f && this->d_scene) // only if dirtying, not clearing
    this->d_scene->d_flags_need_updating = true;
}

bool
VrmlNode::isBVolumeDirty() const
{
  if (d_scene && d_scene->d_flags_need_updating) {
    d_scene->updateFlags();
    d_scene->d_flags_need_updating = false;
  }
  return this->d_bvol_dirty;
}

void VrmlNode::render(Viewer* v, VrmlRenderContext rc)
{
  //if (cull(v, c)) return;
  clearModified();
}

// Accumulate transformations for proper rendering of bindable nodes.

void VrmlNode::accumulateTransform(VrmlNode *node)
{
  ;
}

VrmlNode* VrmlNode::getParentTransform() { return 0; }

#if defined(__ORBISNAP__)

Matrix4x4 VrmlNode::getCompoundTransformationMatrix() const
{
    VrmlNode *parentTransform = ((VrmlNode *) this)->getParentTransform();
    if (parentTransform)
	return parentTransform->getCompoundTransformationMatrix();
    else
    {
	Matrix4x4 m;
	m.setToIdentity();
	return m;
    }
}

Matrix4x4 VrmlNode::getCompoundInverseTransformationMatrix() const
{
    VrmlNode *parentTransform = ((VrmlNode *) this)->getParentTransform();
    if (parentTransform)
	return parentTransform->getCompoundInverseTransformationMatrix();
    else
    {
	Matrix4x4 m;
	m.setToIdentity();
	return m;
    }
}

#endif

void VrmlNode::inverseTransform(Viewer *v)
{
  VrmlNode *parentTransform = getParentTransform();
  if (parentTransform)
    parentTransform->inverseTransform(v);
}

void VrmlNode::inverseTransform(double m[4][4])
{
  //cout << "VrmlNode::inverseTransform" << endl;
  VrmlNode *parentTransform = getParentTransform();
  if (parentTransform)
    parentTransform->inverseTransform(m);
  else
    Midentity(m);
}


// Pass a named event to this node.

void VrmlNode::eventIn(double timeStamp,
		       const char *eventName,
		       const VrmlField *fieldValue)
{
#ifdef VRML_NODE_DEBUG
  cout << "eventIn "
       << nodeType().getName()
       << "::"
       << (name() ? name() : "")
       << "."
       << eventName
       << " "
       << *fieldValue
       << endl;
#endif

  // Strip set_ prefix
  const char *origEventName = eventName;
  if ( strncmp(eventName, "set_", 4) == 0 )
    eventName += 4;

  // Handle exposedFields (issue both "xxx_changed" and "xxx" event)
  if ( nodeType().hasExposedField( eventName ) )
    {
      setField(eventName, *fieldValue);
      eventOut(timeStamp, eventName, *fieldValue);
      char eventOutName[256];
      snprintf(eventOutName, sizeof(eventOutName), "%s_changed", eventName);
      eventOut(timeStamp, eventOutName, *fieldValue);
      setModified();
    }

  // Handle set_field eventIn/field
  else if ( nodeType().hasEventIn( origEventName ) &&
	    nodeType().hasField( eventName ) )
    {
      setField(eventName, *fieldValue);
      setModified();
    }

  else
    theSystem->warn(System::SEVERITY_HIGH, System::WARNINGTYPE_UNKNOWN, "Error: unhandled eventIn %s::%s.%s\n", nodeType().getName(), name(), origEventName);

}


bool VrmlNode::eventOut(double timeStamp, const char* fieldName)
{
  // do not allow events in editor mode
  if (d_scene && d_scene->isEdited())
    return true;

  const VrmlField* field = getField(fieldName);
  char eventOutName[1024];
  snprintf(eventOutName, sizeof(eventOutName), "%s_changed", fieldName);
  eventOut(timeStamp, fieldName, *field);
  eventOut(timeStamp, eventOutName, *field);
  return true;
}


// Send a named event from this node.

void VrmlNode::eventOut(double timeStamp,
			const char *eventOut,
			const VrmlField &fieldValue)
{
  // do not allow events in editor mode
  if (d_scene && d_scene->isEdited())
    return;

  theSystem->debug("%s::%s %p eventOut %s\n",
	  nodeType().getName(), name(),
	  this, eventOut);

  // send the event through routes
  if (d_namespace)             // paranoia
    d_namespace->getRouteList().Dispatch(timeStamp, this, eventOut, fieldValue);

#if defined(__ORBISNAP__)

   // for nodes that are inside a PROTO, we notify the event as if it was
   // coming from the appropriate ISed field (if any)
   if (d_proto)
   {
     // get the list of ISed eventOuts for this PROTO
     VrmlNodeType::FieldList& eventOuts = d_proto->d_nodeType->eventOuts();
     VrmlNodeType::FieldList::iterator ev;

     // scan the list of ISed events, look for this event
     for (ev = eventOuts.begin(); ev != eventOuts.end(); ++ev)
     {
       // look which ISed fields this event is routed to
       VrmlNodeType::ISMap* ismap = &(*ev)->thisIS;
       VrmlNodeType::ISMap::const_iterator j;

       for (j = ismap->begin(); j != ismap->end(); ++j)
       {
	 if (strcmp((*j)->node->name(), d_name) == 0
	   && strcmp((*j)->fieldName, eventOut) == 0)
	 {
	   d_proto->setField((*ev)->name, fieldValue);
	   d_proto->eventOut(timeStamp, (*ev)->name, fieldValue);
           // fire event only if it differs from the previous one
	   if (strcmp(VrmlNode::stripVrmlDecors((*ev)->name), (*ev)->name)!=0)
	     d_proto->eventOut(timeStamp, VrmlNode::stripVrmlDecors((*ev)->name), fieldValue);
	 }
       }
     }
   }

#endif
}


ostream& operator<<(ostream& os, const VrmlNode& f)
{ return f.print(os, 0); }


void VrmlNode::clearModifiedAll()
{
  clearModified();

  // get the field list
  VrmlNodeType* nodetype = &(nodeType());
  VrmlNodeType::FieldList* flist = &(nodetype->fields());
  VrmlNodeType::FieldList::iterator iter;

  // cycle through all fields, clear the flag for the nodes
  // contained in SFNode and MFNode fields
  for (iter=flist->begin(); iter!=flist->end(); iter++)
  {
    VrmlField* field = (VrmlField*) getField((*iter)->name);
    VrmlSFNode* sfnode = dynamic_cast<VrmlSFNode*>(field);
    if (sfnode != NULL && sfnode->get() != NULL)
      sfnode->get()->clearModifiedAll();
    VrmlMFNode* mfnode = dynamic_cast<VrmlMFNode*>(field);
    if (mfnode != NULL)
    {
      unsigned int i;
      for (i=0; i<mfnode->getLength(); i++)
        if ((*mfnode)[i] != NULL) (*mfnode)[i]->clearModifiedAll();
    }
  }
}


ostream& VrmlNode::toChain(ostream& os, bool editFlag) const
{
  if (isFlagSet())
  {
    os << 'U' << d_name << TERMINAL_STRING;
    os << nodeType().getName() << TERMINAL_STRING;
    // terminate USE.
    os << TERMINAL_STRING;
    return os;
  }

  // 'N' is a mark for node
  os << 'N';
  // insert name of node and terminate the string
  const char *nm = name();
  if (nm && *nm) os << nm;
  os << TERMINAL_STRING;

  // insert name of the string withing parentheses and terminate the string
  os << nodeType().getName() << TERMINAL_STRING;

  VrmlNodeType* nodetype = &(nodeType());
  VrmlNodeType::FieldList::iterator i;
  VrmlNodeType::FieldList sortedList;


  reverseFieldList(&(nodetype->eventIns()), &sortedList);
  for (i=sortedList.begin(); i!=sortedList.end(); i++)
  {
    // do not print events derived from exposedFields, they are implicit
    if (nodetype->hasField((*i)->name)) continue;
    if (VrmlNodeType::strstarts((*i)->name, "set_"))
    {
      char* basename = VrmlNodeType::strtrim((*i)->name, 4, 0);
      if (nodetype->hasField(basename)) { delete[] basename; continue; }
      delete[] basename;
    }

    // event in
    os << 'I' << (*i)->name << TERMINAL_STRING;
    os << VrmlField::getTypeName((*i)->type) << TERMINAL_STRING;
    os << (VrmlField::IsExtSettable((*i)->type) ? '1' : '0') << TERMINAL_STRING; // just input
  }

  // fields without children
  sortedList.clear();
  getFieldsWithoutChildren(&(nodetype->fields()), &sortedList, editFlag);
  for (i=sortedList.begin(); i!=sortedList.end(); i++)
  {
    // exposed fields
    const char* fname = (*i)->name;
    if (nodetype->hasExposedField(fname) != VrmlField::NO_FIELD)
    {
      os << 'E' << TERMINAL_STRING;
      os << (isFieldValueDefault(fname) ? 'D' : 'C') << TERMINAL_STRING;
      os << fname << TERMINAL_STRING;
      os << VrmlField::getTypeName((*i)->type) << TERMINAL_STRING;
      // isFieldTransformable() is used only for exposed fields because it is not needed for fields yet
      os << (VrmlField::IsExtSettable((*i)->type) ? (this->isFieldTransformable((*i)->name)) ? '4' : '3' : '0') << TERMINAL_STRING;     // input and output
    }
  }

  for (i=sortedList.begin(); i!=sortedList.end(); i++)
  {
    // fields
    const char* fname = (*i)->name;
    if (nodetype->hasExposedField(fname) == VrmlField::NO_FIELD)
    {
      os << 'F' << TERMINAL_STRING;
      os << (isFieldValueDefault(fname) ? 'D' : 'C') << TERMINAL_STRING;
      os << fname << TERMINAL_STRING;
      os << VrmlField::getTypeName((*i)->type) << TERMINAL_STRING;
      os << (VrmlField::IsExtSettable((*i)->type) ? '3' : '0') << TERMINAL_STRING;     // input and output
    }
  }

  reverseFieldList(&(nodetype->eventOuts()), &sortedList);
  for (i=sortedList.begin(); i!=sortedList.end(); i++)
  {
    // do not print events derived from exposedFields, they are implicit
    if (nodetype->hasField((*i)->name)) continue;
    if (VrmlNodeType::strends((*i)->name, "_changed"))
    {
      char* basename = VrmlNodeType::strtrim((*i)->name, 0, 8);
      if (nodetype->hasField(basename)) { delete[] basename; continue; }
      delete[] basename;
    }

    // event out
    os << 'O' << (*i)->name << TERMINAL_STRING;
    os << VrmlField::getTypeName((*i)->type) << TERMINAL_STRING;
    os << (VrmlField::IsExtSettable((*i)->type) ? '2' : '0') << TERMINAL_STRING;     // just output
  }

  // extract children nodes
  sortedList.clear();
  getFieldsWithChildren(&(nodetype->fields()), &sortedList, editFlag);
  for (i=sortedList.begin(); i!=sortedList.end(); i++)
  {
    const char* fname = (*i)->name;
    os << ((nodetype->hasExposedField(fname) == VrmlField::NO_FIELD) ? 'F' : 'E') << TERMINAL_STRING;
    os << (isFieldValueDefault(fname) ? 'D' : 'C') << TERMINAL_STRING;
    os << fname << TERMINAL_STRING;
    os << VrmlField::getTypeName((*i)->type) << TERMINAL_STRING;
    os << 'S';
    const VrmlField* field = getField(fname);
		field->toChain(os, editFlag);
    os << TERMINAL_STRING;
  }

  os << TERMINAL_STRING;

  // this node was already printed
  ((VrmlNode*) this)->setFlag();

  return os;
}

void VrmlNode::reverseFieldList(VrmlNodeType::FieldList* sourceList,
	VrmlNodeType::FieldList* resultingList) const
{

   VrmlNodeType::FieldList::iterator sIter;

   for (sIter=sourceList->begin(); sIter!=sourceList->end(); sIter++)
    {
      	resultingList->push_front((*sIter));
    }
}

void VrmlNode::getFieldsWithoutChildren(VrmlNodeType::FieldList* sourceList,
	VrmlNodeType::FieldList* resultingList, bool editFlag) const
{

    VrmlNodeType::FieldList::iterator sIter;

    for (sIter=sourceList->begin(); sIter!=sourceList->end(); sIter++)
    {
      const VrmlField* field = getField((*sIter)->name);
      int type = (*sIter)->type;
      if ((type != VrmlField::SFNODE && type != VrmlField::MFNODE) ||
      	(!editFlag && !field->hasChildren()))
      {
      	resultingList->push_front((*sIter));
      }
    }

}

void VrmlNode::getFieldsWithChildren(VrmlNodeType::FieldList* sourceList,
	VrmlNodeType::FieldList* resultingList, bool editFlag) const
{
    VrmlNodeType::FieldList::iterator sIter;
    for (sIter=sourceList->begin(); sIter!=sourceList->end(); sIter++)
    {
	  int type = (*sIter)->type;
	  const VrmlField* field = getField((*sIter)->name);

      if ((type == VrmlField::SFNODE || type == VrmlField::MFNODE) && (editFlag || field->hasChildren()))
      {
      	resultingList->push_back((*sIter));
      }
    }

}


ostream& VrmlNode::print(ostream& os, int indent, bool breakLine) const
{
  if (isFlagSet())
  {
    if (validName() == NULL)
    {
      // a rare special case that can happen if the node was created
      // artificially, e.g. via vrmlscript interface, and kept unnamed
      theSystem->warn(System::SEVERITY_HIGH, System::WARNINGTYPE_UNKNOWN, "Reference to unnamed node can't be saved.");
      os << nodeType().getName() << " { }";
    }
    else
    {
      if (breakLine)
      {
	os << "\n";
        for (int i=0; i<indent; ++i) os << ' ';
      }
      os << "USE " << d_name;
    }
    return os;
  }

  printComment(os, indent, breakLine);

  // print node name if the node has valid name. there can be nodes
  // with synthetic names -- synthetic names are not valid
  const char *nm = validName();
  if (nm) os << "DEF " << nm << " ";

  // opening brace
  os << nodeType().getName() << " {";

  // print fields
  // EXPERIMENTAL: Let's try if we can print node fields in a common way using
  // the type info we have about the nodes. If it is feasible, we could
  // make the whole code a lot simpler (the nodes would not need special
  // printFields() implementation).
  {
    // get the field list
    VrmlNodeType* nodetype = &(nodeType());
    VrmlNodeType::FieldList* flist = &(nodetype->fields());
    VrmlNodeType::FieldList::iterator iter;

    // cycle through all fields, printing them to output
    for (iter=flist->begin(); iter!=flist->end(); ++iter)
    {
      printField(os, (*iter)->name, indent);
    }

    //////////////////////////////////////////////////////////////
    // we print eventIns and eventOuts only because of IS mappings
    //////////////////////////////////////////////////////////////
    flist = &(nodetype->eventOuts());
    // cycle through all eventOuts, printing them to output
    for (iter=flist->begin(); iter!=flist->end(); ++iter)
    {
      printField(os, (*iter)->name, indent);
    }
    flist = &(nodetype->eventIns());
    // cycle through all eventIns, printing them to output
    for (iter=flist->begin(); iter!=flist->end(); ++iter)
    {
      printField(os, (*iter)->name, indent);
    }
    //////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////
  }

  // ending brace
  os << "\n";
  for (int i=0; i<indent; ++i) os << ' ';  // indent
  os << "}";

  // mark this node as already printed; if it is encountered next time,
  // it means it is a USEd node and all other references should be printed
  // using USE.
  ((VrmlNode*) this)->setFlag();

  return os;
}

// This should probably generate an error...
// Might be nice to make this non-virtual (each node would have
// to provide a getField(const char* name) method and specify
// default values in the addField(). The VrmlNodeType class would
// have to make the fields list public.

ostream& VrmlNode::printFields(ostream& os, int /*indent*/)
{
  os << "# Error: " << nodeType().getName()
     << "::printFields unimplemented.\n";
  return os;
}


ostream& VrmlNode::printField(ostream& os,
			      int indent,
			      const char *name,
			      const VrmlField& f)
{
  os << name << ' ';

  if (theSystem->getExportDir() && f.fieldType() == VrmlField::MFSTRING && strcmp(name,"url"))
    f.printAsUrl(os, indent+INDENT_INCREMENT);
  else
    f.print(os, indent+INDENT_INCREMENT);

  return os;
}


bool VrmlNode::isFieldValueDefault(const char* fieldName) const
{
  const VrmlField* field = getEventOut(fieldName);
  VrmlField* defaultField = nodeType().fieldDefault(fieldName);

  return (defaultField!=NULL && field!=NULL && field->equals(defaultField));
}


void VrmlNode::setField(const char *fieldName, const VrmlField &fieldValue)
{
  char* name = stripVrmlDecors(fieldName);
  VrmlField* field = findField(name);
  delete[] name;
  if ( (field == &fieldValue) ||    // setting a field to itself
       (field == NULL) ||           // not found
       (field->fieldType() != fieldValue.fieldType()) )   // type mismatch
    return;
  field->set(fieldValue);
  setModified();
  onFieldChange(fieldName, fieldValue);
}


VrmlField* VrmlNode::getFieldStripDecorations(const char* fieldName)
{
  char* name = stripVrmlDecors(fieldName);
  const VrmlField* field = getField(name);
  delete[] name;
  return(const_cast<VrmlField*>(field));
}


void VrmlNode::onFieldChange(const char* fieldName, const VrmlField &fieldValue)
{ }


void VrmlNode::PROTOFieldChanged(const char *fieldName)
{ }


VrmlField::VrmlFieldType VrmlNode::hasField(const char* fieldName) const
{
  return nodeType().hasField(fieldName);
}


VrmlField::VrmlFieldType VrmlNode::hasEventIn(const char* fieldName) const
{
  return nodeType().hasEventIn(fieldName);
}


VrmlField::VrmlFieldType VrmlNode::hasEventOut(const char* fieldName) const
{
  return nodeType().hasEventOut(fieldName);
}



const VrmlField *VrmlNode::getField(const char *fieldname) const
{
  char* name = stripVrmlDecors(fieldname);
  VrmlField* field = const_cast<VrmlNode*>(this)->findField(name);
  delete[] name;
  return field;
}



// Retrieve a named eventOut/exposedField value.

const VrmlField *VrmlNode::getEventOut(const char *fieldName) const
{
  // try if the field name is a normal field name
  const VrmlField *result = getField(fieldName);
  if (result != 0) return result;

  // try if the user is not using the demented "_changed" syntax
  int rootLen = strlen(fieldName);
  if (rootLen > 8 && strcmp(fieldName+rootLen-8, "_changed") == 0)
  {
    // it is the "_changed" suffix, get rid of it
    char *shortName = new char[rootLen-7];
    strncpy(shortName, fieldName, rootLen-8);
    shortName[rootLen-8] = '\0';

    // try it without the suffix
    result = getField(shortName);
    delete[] shortName;
    return result;
  }

  return 0;
}


bool VrmlNode::isFieldTransformable(const std::string& fieldname) const
{
  return false;
}


std::string VrmlNode::stripWorldCoordsField(const std::string& fieldname) const
{
  // find if the "_abs" suffix is there
  const std::string suffix("_abs");
  const size_t suffixLength = suffix.length();
  ptrdiff_t suffixPos = fieldname.length()-suffixLength;
  if (suffixPos<0 || fieldname.compare(suffixPos, suffixLength, suffix)!=0)
    return(std::string());

  // strip the suffix, test if the field is transformable
  std::string stripped(fieldname, 0, suffixPos);
  if (!isFieldTransformable(stripped))
    stripped.clear();

  return(stripped);
}


Matrix4x4 VrmlNode::getWorldCoordsTransformationMatrix(const VrmlField* field)
{
  VrmlNode* node = this;
  if (node->getParentTransform()==NULL)
  {
    Matrix4x4 m; m.setToIdentity();
    return m;
  }
  node = node->getParentTransform();
  return node->getCompoundTransformationMatrix();
}

void VrmlNode::setEncapsulatingPrototype(VrmlNodeType* type)
{
  d_protoType = type;
}

void VrmlNode::setProto(VrmlNodeProto* proto)
{
  d_proto = proto;
}

VrmlNodeProto* VrmlNode::getProto()
{ return d_proto; }

void VrmlNode::addISBinding(const char* fname, const char* iname)
{
  d_bindings.push_back(new FieldISBinding(this, fname, iname));
}

void VrmlNode::deleteISBinding(const char* fieldName)
{
  std::vector<FieldISBinding*>::iterator i = d_bindings.begin();
  while (i != d_bindings.end())
  {
    if (strcmp((*i)->getFieldName(), fieldName) == 0)
    {
      delete (*i);
      i = d_bindings.erase(i);
    }
    else
      ++i;
  }
}

const FieldISBinding* VrmlNode::getISBinding(const char* fname) const
{
  for (int i=0; i<d_bindings.size(); ++i)
  {
    if (strcmp(d_bindings[i]->getFieldName(), fname) == 0)
      return d_bindings[i];
  }
  return NULL;
}

void VrmlNode::printField(ostream &os, const char* fieldname, int indent) const
{
  if (hasField(fieldname) || hasEventIn(fieldname) || hasEventOut(fieldname))                     // paranoia
  {
    // increase indent
    indent += INDENT_INCREMENT;

    // look if the field has an IS binding
    const FieldISBinding* binding = getISBinding(fieldname);

    // print eventOut or eventIn binding
    if (hasField(fieldname)==NULL)
    {
      if (binding != NULL)
      {
        std::string breakStr(indent+1, ' ');
        breakStr[0] = '\n';
        os << breakStr << fieldname << " IS " << binding->getInterfaceName();
      }

      return;
    }

    // print the field if its value is different from the default,
    // or if it has comment,
    // or if it has an IS binding
    const VrmlField* defaultValue = nodeType().fieldDefault(fieldname);
    const VrmlField* currentValue = getField(fieldname);
    const std::string& comment = currentValue->getComment();
    if ((defaultValue == NULL) || !(defaultValue->equals(currentValue)) ||
        !(comment.empty())|| (binding != NULL))
    {

      currentValue->printComment(os, indent, true);

      os << fieldname;

      // print the IS binding if defined
      if (binding != NULL)
      {
        os << " IS " << binding->getInterfaceName();
      }

      // otherwise print the value
      else
      {
        os << ' ';
	const VrmlField* field = getField(fieldname);
	VrmlNode* fieldContents = 0;
	if (
	    (field->fieldType() == VrmlField::MFNODE &&
	       (static_cast<const VrmlMFNode*>(field)->getLength()>1 ||
	       (static_cast<const VrmlMFNode*>(field)->getLength()==1 && !static_cast<const VrmlMFNode*>(field)->getComment().empty())))
	    ||
	    (field->fieldType() == VrmlField::SFNODE &&
	       (fieldContents = static_cast<const VrmlSFNode*>(field)->get()) &&
	       !fieldContents->getComment().empty()))

	  indent += INDENT_INCREMENT;

       if (theSystem->getExportDir() && field->fieldType() == VrmlField::MFSTRING &&
           (strcmp(fieldname,"url")==0 ||
            strcmp(fieldname,"backUrl")==0 ||
            strcmp(fieldname,"bottomUrl")==0 ||
            strcmp(fieldname,"frontUrl")==0 ||
            strcmp(fieldname,"leftUrl")==0 ||
            strcmp(fieldname,"rightUrl")==0 ||
            strcmp(fieldname,"topUrl")==0))
         field->printAsUrl(os, indent);
       else
         field->print(os, indent);
      }
    }
  }
}



char* VrmlNode::stripVrmlDecors(const char* str)
{
  char* result;

  // remove the stupid "set_" prefix
  if (strncmp(str, "set_", 4) == 0)
  {
    result = new char[strlen(str) - 3];
    strcpy(result, str+4);
    return result;
  }

  // remove the demented "_changed" suffix
  int len = strlen(str);
  if (len > 8 && strcmp(str + len - 8, "_changed") == 0)
  {
    result = new char[len - 7];
    strncpy(result, str, len - 8);
    result[len - 8] = '\0';
    return result;
  }

  // return an unchanged copy of the string
  result = new char[len + 1];
  strcpy(result, str);
  return result;
}

void VrmlNode::guessNearFarPlane(Viewer* viewer, double* nearVal, double* farVal)
{
}


void VrmlNode::adjustNearFarToViewer(Viewer* viewer, const Point3& worldCoord, double* nearVal, double* farVal)
{
  // compute vector between grouping node and camera
  VrmlSFVec3f coordToCam(worldCoord.getX(), worldCoord.getY(), worldCoord.getZ());
  Coordinates3 camPos = static_cast<Coordinates3>(viewer->getCameraAbsolute().getPosition());
  VrmlSFVec3f camPos3f(camPos.getX(), camPos.getY(), camPos.getZ());
  coordToCam = coordToCam.subtract(camPos3f);

  // compute norm of vector -- it is z-distance
  Coordinates3 transcoord = Coordinates3(coordToCam.getX(), coordToCam.getY(), coordToCam.getZ());
  Coordinates3 zeroPos(0,0,0);
  double z = zeroPos.euclidDist(&transcoord);

  // consider only objects which are in viewing sector or objects which are adequatly near to camera (in all directions)
  if (viewer->getViewingSector().isInSector(Vector3(coordToCam.getX(), coordToCam.getY(), coordToCam.getZ())) || z < 1000)
  {
    if (z < *nearVal) *nearVal = z;
    if (z > *farVal) *farVal = z;
  }
}



/**********************/
/*** FieldISBinding ***/
/**********************/



FieldISBinding::FieldISBinding(VrmlNode* n, const char* fname, const char* iname)
  : node(n)
{
  fieldName = new char[strlen(fname)+1];
  strcpy(fieldName, fname);
  interfaceName = new char[strlen(iname)+1];
  strcpy(interfaceName, iname);
}

FieldISBinding::~FieldISBinding()
{
  delete[] fieldName;
  delete[] interfaceName;
}
