/*****************************************************************************
******************************************************************************
*
*               VRML node
*
*               $Revision: 1.1.2.18 $
*               $Date: 2011/12/12 20:16:54 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright 1998-2011 HUMUSOFT s.r.o.
// Portions Copyright (C) 1998  Chris Morley

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//



#ifndef VRMLNODE_H__
#define VRMLNODE_H__

#include <string.h>
#include <list>
#include <iostream>
using namespace std;

#include "System.h"        // error handling
#include "VrmlObject.h"
#include "VrmlField.h"
#include "VrmlNodeType.h"
#include "Math3d.h"
#include "VrmlRenderContext.h"

class VrmlNode;
class Viewer;
class VrmlNamespace;
class VrmlNodeType;
class VrmlNodeProto;
class VrmlField;
class VrmlScene;
class VrmlBVolume;
class FieldISBinding;


// Global output operator for nodes
ostream& operator<<(ostream& os, const VrmlNode& f);


/**
 * A node in the scene graph.
 */
class OPENVRMLDLLEXPORT VrmlNode : public VrmlObject
{
friend ostream& operator<<(ostream& os, const VrmlNode& f);
friend class VrmlNodeProto;
friend class VrmlNodeGroup;
friend class Vrml97Parser;

private:

  /** Reference count. */
  int d_refCount;

  /** Name of this node if it is named, otherwise NULL. */
  char *d_name;

  /** True if a field changed since last render. */
  bool d_modified;

  /**
   * If this node is a part of a PROTO (or EXTERNPROTO) implementation
   * (either a part of an instantiated PROTO, or part of the PROTO
   * description), this is a pointer to the VrmlNodeType that describes
   * the prototype. Otherwise it is NULL.
   */
  VrmlNodeType* d_protoType;

protected:

  /** Scene this node belongs to. */
  VrmlScene *d_scene;

  bool d_flag;
  bool d_bvol_dirty;

  /**
   * If this node is a part of a PROTO (or EXTERNPROTO) instance,
   * this is a pointer to the appropriate object, otherwise NULL.
   */
  VrmlNodeProto* d_proto;

  /** Namespace this node belongs to. This pointer should be always valid
    * for completely loaded nodes.
    */
  VrmlNamespace* d_namespace;

  /**
   * List of IS bindings for nodes that are part of a PROTO implementation.
   */
  std::vector<FieldISBinding*> d_bindings;

  /**
   * Prints the field name, value, and IS binding (if present)
   * to the given output stream. Fields with default value
   * and no IS binding are skipped.
   * This is used when saving a world to VRML file.
   */
  void printField(ostream &os, const char* fieldName, int indent) const;

  /** Browse list of fields and add fields that has children into target list..
   * @param sourceList list of fields.
   * @param editFlag true if chain is created for 3D World Editor purposes.
   * @param resultingList list being used for storing result.
   *
   */
  void getFieldsWithChildren(VrmlNodeType::FieldList* sourceList,
	VrmlNodeType::FieldList* resultingList, bool editFlag) const;

  /** Browse list of fields and store fields without children into target list.
   * If the field has 'children', exclude it from the processing.
   * @param sourceList list of fields.
   * @param editFlag true if chain is created for 3D World Editor purposes.
   * @param resultingList list being used for storing result.
   *
   */
  void getFieldsWithoutChildren(VrmlNodeType::FieldList* sourceList,
	VrmlNodeType::FieldList* resultingList, bool editFlag) const;


   /** Store field of list in reversed order.
    * @param sourceList list of fields.
    * @param resultingList list being used for storing result.
    */

   void reverseFieldList(VrmlNodeType::FieldList* sourceList,
		VrmlNodeType::FieldList* resultingList) const;

  /**
   * Finds the internal object representing the given field
   * and returns a pointer to it.
   * Returns NULL if the specified field is not found.
   */
  virtual VrmlField* findField(const char* fieldname)
    { return NULL; }

  /**
   * Create a new instance of the node in the specified namespace.
   * This is done for the purpose of cloning a node or instantiating a PROTO.
   */
  VrmlNode* newInstance(VrmlNamespace* ns);

public:

  /**
   * Send a named event from this node.
   */
  void eventOut(double timeStamp,
                const char *eventName,
                const VrmlField &fieldValue);

  /**
   * Given a VrmlNodeType, add in the fields, exposedFields, eventIns
   * and eventOuts defined by the particular node implementation.
   * There's a great big method in VrmlNamespace that just calls
   * defineType for every built in node. Nodes that inherit from other
   * nodes (eg VrmlNodeIFaceSet) must call their parent class
   * defineType, it doesn't happen automatically. The defineType for
   * VrmlNode doesn't actually do anything, since there (currently)
   * aren't any base events to be defined. It's just here for the sake
   * of symmetry (and makes a nice place to park a comment)
   *
   * @param t to get the predefined cached type, pass a null,
   *	      otherwise the fields and events will be added to the
   *	      passed in vrmlnodetype object
   *
   * @see VrmlNamesspace::defineBuiltins()
   */
  static VrmlNodeType *defineType(VrmlNodeType *t)
    { return(t); };

  virtual VrmlNodeType & nodeType() const = 0;

  // VrmlNodes are reference counted, optionally named objects
  // The reference counting is manual (that is, each user of a
  // VrmlNode, such as the VrmlMFNode class, calls reference()
  // and dereference() explicitly). Should make it internal...

  VrmlNode ( VrmlScene *s );
  VrmlNode(  const VrmlNode& );
  virtual ~VrmlNode() = 0;

  // Copy the node, defining its name in the specified scope.
  // Uses the flag to determine whether the node is a USEd node.
  VrmlNode *clone( VrmlNamespace* );
  virtual VrmlNode *cloneMe() const = 0;
  virtual void cloneChildren( VrmlNamespace* );

  // method used when instantiating contents of VrmlNodeProto
  virtual VrmlNode* instantiate(VrmlNamespace*);

  // does nothing -- overriden for VrmlNodeProto
  virtual void reinstantiate();

  // Add/remove references to a VrmlNode. This is silly, as it
  // requires the users of VrmlNode to do the reference/derefs...
  VrmlNode *reference();
  void dereference();

  // Return the reference count
  int getRefCount();

  // Node DEF/USE/ROUTE name
  void setName(const char *nodeName, VrmlNamespace *ns = 0 );

  /**
   * Returns the node name (even if synthetic one is used)
   */
  const char *name() const;

  /**
   * Returns the node valid name
   */
  const char *validName() const;

  // Add to a scene. A node can belong to at most one scene for now.
  // If it doesn't belong to a scene, it can't be rendered.
  virtual void addToScene( VrmlScene *, const char *relativeUrl );

  /**
   * Prints the VRML code representing this node to the output stream.
   */
  virtual ostream& print(ostream& os, int indent, bool breakLine = false) const;

  /**
   * Store this node to output stream.
   */
  virtual ostream& toChain(ostream& os, bool editFlag) const;

  virtual ostream& printFields(ostream& os, int indent);
  static ostream& printField(ostream&, int, const char*, const VrmlField&);

  // works for both fields/exposedFields
  bool isFieldValueDefault(const char* fieldName) const;

  // Indicate that the node state has changed, need to re-render
  virtual void setModified();

  // Invalidates node and its children (if any)
  virtual void invalidate();

  /**
   * Clears the modified flag on this node.
   */
  void clearModified() { d_modified = false; }

  /**
   * Clears the modified flag on this node and all its descendants.
   */
  virtual void clearModifiedAll();

  virtual bool isModified() const;
  typedef std::list< VrmlNode* > VrmlNodePath; // duplicate from VrmlScene. argh.


  /**
   * Mark all the nodes in the path as (not) modified. Convenience
   * function used by updateModified.
   */
  static void markPathModified(VrmlNodePath& path, bool mod, int flags = 0x003);

  // do the work of updatemodified. move this to be protected
  //
  virtual void updateModified(VrmlNodePath& path, int flags = 0x003);

  /**
   * Propegate the bvolume dirty flag from children to parents. I
   * don't like this at all, but it's not worth making it pretty
   * because the need for it will go away when parent pointers are
   * implemented.
   *
   * @param path stack of ancestor nodes
   * @param mod set modified flag to this value
   * @param flags 1 indicates normal modified flag, 2 indicates the
   *		  bvolume dirty flag, 3 indicates both
   */
  void updateModified(int flags = 0x003);

  /**
   * Get this node bounding volume. Nodes that have no bounding
   * volume, or have a difficult to calculate bvolume (like, say,
   * extrusion or billboard) can just return an infinite bsphere. Note
   * that returning an infinite bvolume means that all the node
   * ancestors will also end up with an infinite bvolume, and will
   * never be culled.
   *
   * @return this node bounding volume
   */
  virtual const VrmlBVolume* getBVolume() const;

  /**
   * Override a node calculated bounding volume. Not implemented.
   */
  virtual void setBVolume(const VrmlBVolume& v);

  /**
   * Indicate that a node bounding volume needs to be recalculated
   * (or not). If a node bvolume is invalid, then the bvolumes of
   * all that node ancestors are also invalid. Normally, the node
   * itself will determine when its bvolume needs updating.
   */
  virtual void setBVolumeDirty(bool f);

  /**
   * Return true if the node bounding volume needs to be
   * recalculated.
   */
  virtual bool isBVolumeDirty() const;


  // A generic flag (typically used to find USEd nodes).
  void setFlag() { d_flag = true; }
  virtual void clearFlags();	// Clear childrens flags too.
  bool isFlagSet() const { return d_flag; }

  // Pass a named event to this node. This method needs to be overridden
  // to support any node-specific eventIns behaviors, but exposedFields
  // (should be) handled here...
  virtual void eventIn(double timeStamp,
		       const char *eventName,
		       const VrmlField *fieldValue);

  /**
   * Set a field (without emiting any events).
   * The node is automatically marked as changed.
   */
  virtual void setField(const char *fieldName, const VrmlField &fieldValue);

  /**
   * Return a field according to its name, stripping any VRML event decorations.
   */
  VrmlField* getFieldStripDecorations(const char *fieldName);

  /**
   * This method is invoked after a field is changed, allowing the node
   * to do any needed recalculations. The default implementation is empty.
   */
  virtual void onFieldChange(const char* fieldName, const VrmlField &fieldValue);

  /**
   * Implemented only for PPROTO/EXTERNPROTO instances.
   * Propagates field value from PROTO/EXTERNPROTO interafce to its body
   * (field IS mappings are updated with interface field value).
   */
  virtual void PROTOFieldChanged(const char *fieldName);

  /** Checks whether the node has the given field,
    * eventually returns the field type.
    * @param fieldName Field name.
    * @return Nonzero field type if the field exists,
    * or VrmlField::NO_FIELD (which has a zero value) otherwise.
    */
  virtual VrmlField::VrmlFieldType hasField(const char* fieldName) const;
  virtual VrmlField::VrmlFieldType hasEventIn(const char* fieldName) const;
  virtual VrmlField::VrmlFieldType hasEventOut(const char* fieldName) const;

  // Get a field or eventOut by name. getField is used by Script nodes
  // to access exposedFields. It does not allow access to private fields
  // (there tend to be specific access functions for each field for
  // programmatic access).
  virtual const VrmlField *getField(const char *fieldName) const;

  // Return an eventOut/exposedField value. Used by the script node
  // to access the node fields. This just strips the _changed suffix
  // and tries to access the field using getField.
  const VrmlField *getEventOut(const char *fieldName) const;


  /**
   * Returns true if matrix-transformations done on a field of the given
   * field name make sense. We allow such transformations only for some fields
   * of Transform and Viewpoint nodes at the moment.
   */
  virtual bool isFieldTransformable(const std::string& fieldname) const;

  /**
   * Returns a stripped name of a field that is to be read in world coordinates.
   * Such fields are currently allowed only for Transform and Viewpoint nodes
   * and they are ended up by "_abs" suffix.
   * Returns an empty string if field is not transformable or is not ended by "_abs".
   */
  std::string stripWorldCoordsField(const std::string& fieldname) const;

  /**
   * Returns transformation matrix which can be used
   * for transformation of a given field to world coordinates.
   */
  Matrix4x4 getWorldCoordsTransformationMatrix(const VrmlField* field);


  /**
   * Render this node. Actually, most of the rendering work is
   * delegated to the viewer, but this method is responsible for
   * traversal to the node renderable children, including
   * culling. Each node class needs to implement this routine
   * appropriately. It's not abstract since it doesn't make sense to
   * call render on some nodes. Alternative would be to break render
   * out into a seperate mixins class, but that's probably overkill.
   *
   * @param v viewer implementation responsible for actually doing the
   *	      drawing
   * @param rc generic context argument, holds things like the
   *	      accumulated modelview transform.
   */
  virtual void render(Viewer *, VrmlRenderContext rc);


  /**
   * Cache a pointer to one of the parent transforms. The resulting
   * pointer is used by getParentTransform. Grouping nodes need to
   * redefine this, the default implementation does nothing.
   *
   * @param p parent node. can be null.
   *
   * @deprecated This routine will go away once parent pointers
   * are implemented.
   */
  virtual void accumulateTransform(VrmlNode*);

  /**
   * Return the nearest ancestor node that affects the modelview
   * transform. Doesn't work for nodes with more than one parent.
   */
  virtual VrmlNode* getParentTransform();

#if defined(__ORBISNAP__)

  /** Calculates the compound transformation for this node,
    * defined by the transformation stack above this node
    * and possibly the node itself (if it is a transformation node).
    * The resulting matrix can be used to transform points and vectors
    * from local coordinates to world coordinates.
    * @returns The compound transformation in a matrix.
    */
  virtual Matrix4x4 getCompoundTransformationMatrix() const;

  /** Calculates the inverse of the compound transformation for this node.
    * The resulting matrix can be used to transform points and vectors
    * from world coordinates to local coordinates.
    * @returns The compound transformation in a matrix.
    */
  virtual Matrix4x4 getCompoundInverseTransformationMatrix() const;

#endif

  /**
   * Compute the inverse of the transform above a viewpoint node. Just
   * like the version that takes a matrix, but instead calls
   * Viewer::setTransform at each level. The idea is to call this
   * routine right before the start of a render traversal.
   *
   * @see getParentTransform
   *
   * @deprecated This method is (gradually) being replaces by
   * inverseTranform(double[4][4]) and should no longer be used.
   */
  virtual void inverseTransform(Viewer *);

  /**
   * Compute the inverse of the transform stack above a Viewpoint
   * node. This is safe since the behavior of multi-parented
   * Viewpoint nodes is undefined. May be called at any time.
   *
   * @param M return the accumulated inverse
   *
   * @see accumulateTransform
   * @see getParentTransform
   */
  virtual void inverseTransform(double [4][4]);

  VrmlScene *scene() { return d_scene; }

  /** Returns a pointer to the PROTO node this node belongs to,
    * or 0 if the node is not a part of any PROTO.
    */
  virtual VrmlNodeProto* getProto();

  /** Send an eventOut from this node, reporting that the specified field
    * has changed.
    * This is used when a field value was changed by other way
    * than by an eventIn (e.g. by a direct access from another module)
    * to notify the rest of the world about the change.
    * @param timeStamp Time stamp of the event to be generated.
    * @param fieldName Name of the field to be reported.
    * @returns True if succeeded, false if such a field does not exist.
    */
  bool eventOut(double timeStamp, const char* fieldName);

  /** Returns a pointer to namespace this node belongs to. */
  VrmlNamespace* getNamespace() { return d_namespace; }

  /** notifies this node that its related namespace does not exist any more */
  void setNamespaceDestructed() { d_namespace = 0; }

  /**
   * If this node is a part of a PROTO definition or a PROTO instance,
   * this method returns a pointer to VrmlNodeType object that represents
   * that PROTO type. Otherwise it returns NULL.
   */
  VrmlNodeType* getEncapsulatingPrototype() { return d_protoType; }

  /**
   * If this node is a part of a PROTO definition or a PROTO instance,
   * this method sets a pointer to VrmlNodeType object that represents
   * that PROTO type.
   */
  virtual void setEncapsulatingPrototype(VrmlNodeType* type);

  /** Sets a PROTO pointer for this node and for its children (if any).
    * This is used in PROTO instantiation (see VrmlNodeProto.cpp).
    */
  virtual void setProto(VrmlNodeProto* proto);

  /**
   * Adds a new IS binding to the list of bindings.
   * This is used by the parser when building a PROTO implementation.
   */
  void addISBinding(const char* fieldName, const char* interfaceName);

  /**
   * Deeletes an existing IS binding from the list of bindings.
   * This is used when editing a PROTO implementation via 3D World Editor.
   */
  void deleteISBinding(const char* fieldName);

  /**
   * Returns an IS binding for the given field, or NULL
   * if either the field or the binding don't exist.
   */
  const FieldISBinding* getISBinding(const char* field) const;

  /**
   * Removes the VRML "set_" and "_changed" decorations from the given
   * string. Returns a newly allocated string.
   */
  static char* stripVrmlDecors(const char* str);

 /**
  * Guess znear and zfar values
  */
  virtual void guessNearFarPlane(Viewer* viewer, double* nearVal, double* farVal);

 /**
  * Adjust near and far plane guess to viewer
  */
  void adjustNearFarToViewer(Viewer* viewer, const Point3& worldCoord, double* nearVal, double* farVal);
};


// Ugly macro used in printFields
#define PRINT_FIELD(_f) printField(os,indent+INDENT_INCREMENT,#_f,d_##_f)

// Ugly macros used in setField
#ifdef __GNUC__   // RTTI doesn't work very well with GNU C++
#define TRY_FIELD(_f, _t)                                              \
(strcmp(fieldName, #_f) == 0)				               \
{                                                                      \
  if (fieldValue.fieldType() == d_##_f.fieldType())		       \
    d_##_f = static_cast<const Vrml##_t&>(fieldValue);		       \
  else                                                                 \
    theSystem->error(System::ERRORTYPE_UNKNOWN, "Invalid type (%s) for %s field of %s node (expected %s).\n",  \
                     fieldValue.fieldTypeName(), #_f, nodeType().getName(), #_t);   \
}
#else
#define TRY_FIELD(_f, _t)                                              \
(strcmp(fieldName, #_f) == 0)				               \
{                                                                      \
  const Vrml##_t* fv = dynamic_cast<const Vrml##_t*>(&fieldValue);     \
  if (fv != NULL)                                                      \
    d_##_f = *fv;                                                      \
  else                                                                 \
    theSystem->error(System::ERRORTYPE_UNKNOWN, "Invalid type (%s) for %s field of %s node (expected %s).\n",  \
                     fieldValue.fieldTypeName(), #_f, nodeType().getName(), #_t);   \
}
#endif  // __GNUC__

// For SFNode fields. Allow un-fetched EXTERNPROTOs to succeed...
#define TRY_SFNODE_FIELD(_f, _n)                                       \
(strcmp(fieldName, #_f) == 0)                                          \
{                                                                      \
  const VrmlSFNode* x = dynamic_cast<const VrmlSFNode*>(&fieldValue);  \
  if (x != NULL &&                                                     \
      ( (x->get() == NULL) ||                                          \
         dynamic_cast<VrmlNode##_n*>(x->get()) ||                      \
         dynamic_cast<VrmlNodeProto*>(x->get()) ))                     \
    d_##_f = *x;                                                       \
  else                                                                 \
    theSystem->error(System::ERRORTYPE_UNKNOWN, "Invalid type (%s) for %s field of %s node (expected %s).\n",  \
                     fieldValue.fieldTypeName(), #_f, nodeType().getName(), #_n);   \
  }

#define TRY_SFNODE_FIELD2(_f, _n1, _n2)                                \
(strcmp(fieldName, #_f) == 0)                                          \
{                                                                      \
  const VrmlSFNode* x = dynamic_cast<const VrmlSFNode*>(&fieldValue);  \
  if (x != NULL &&                                                     \
      ( (x->get() == NULL) ||                                          \
         dynamic_cast<VrmlNode##_n1*>(x->get()) ||                     \
         dynamic_cast<VrmlNode##_n2*>(x->get()) ||                     \
         dynamic_cast<VrmlNodeProto*>(x->get()) ))                     \
    d_##_f = *x;                                                       \
  else                                                                 \
    theSystem->error(System::ERRORTYPE_UNKNOWN, "Invalid type (%s) for %s field of %s node (expected %s or %s).\n",   \
                     fieldValue.fieldTypeName(), #_f, nodeType().getName(), #_n1, #_n2);   \
  }



/**
 * This structure represents a binding of a local field to a prototype field.
 */
class FieldISBinding
{
  protected:

    /** Local node. */
    VrmlNode* node;

    /** Name of the local field. */
    char* fieldName;

    /** Name of the PROTO interface field. */
    char* interfaceName;

  public:

    /**
     * Constructor.
     */
    FieldISBinding(VrmlNode* node, const char* fieldName, const char* interfaceName);

    ~FieldISBinding();

    /**
     * Returns the local node this binding relates to.
     */
    const VrmlNode* getNode() const { return node; }

    /**
     * Returns the name of the local field of this binding.
     */
    const char* getFieldName() const { return fieldName; }

    /**
     * Returns the name of the interface field in the PROTO.
     */
    const char* getInterfaceName() const { return interfaceName; }
};



/** An entry to a node field table. */
class VrmlFieldTableEntry
{
public:

  /** Field name. */
  const char* name;

  /** Pointer to the variable representing this field. */
  VrmlField* VrmlNode::* field;
};



#endif // VRMLNODE_H__
