/*****************************************************************************
******************************************************************************
*
*               VRML AudioClip node
*
*               $Revision: 1.1.2.3 $
*               $Date: 2010/06/30 17:01:24 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright (C) 1999  Kumaran Santhanam
// Modified by HUMUSOFT s.r.o.  Copyright 1998-2010 HUMUSOFT s.r.o.

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#include "VrmlNodeAudioClip.h"
#include "VrmlNodeType.h"
#include "VrmlScene.h"
#include "MathUtils.h"
#include "Audio.h"
#include "Doc.h"
#include "sound.h"


static VrmlNode *creator( VrmlScene *s ) { return new VrmlNodeAudioClip(s); }


// Define the built in VrmlNodeType:: "AudioClip" fields

VrmlNodeType *VrmlNodeAudioClip::defineType(VrmlNodeType *t)
{
  static VrmlNodeType *st = 0;

  if (! t)
    {
      if (st) return st;
      t = st = new VrmlNodeType("AudioClip", creator);
      t->reference();
    }

  VrmlNode::defineType(t);	// Parent class
  t->addExposedField("description", VrmlField::SFSTRING, VrmlSFString());
  t->addExposedField("loop", VrmlField::SFBOOL, VrmlSFBool());
  t->addExposedField("pitch", VrmlField::SFFLOAT, VrmlSFFloat(1.0));
  t->addExposedField("startTime", VrmlField::SFTIME, VrmlSFTime());
  t->addExposedField("stopTime", VrmlField::SFTIME, VrmlSFTime());
  t->addExposedField("url", VrmlField::MFSTRING, VrmlMFString());

  t->addEventOut("duration_changed", VrmlField::SFTIME);
  t->addEventOut("isActive", VrmlField::SFBOOL);

  return t;
}


VrmlNodeType & VrmlNodeAudioClip::nodeType() const
{
    return *defineType(0);
}


VrmlNodeAudioClip::VrmlNodeAudioClip(VrmlScene *scene) :
  VrmlNode(scene),
  d_pitch(1.0),
  d_isActive(false),
  d_audio(0),
  d_url_modified(false),
  _audio_index(0),
  _audio_intensity(1.0),
  _audio_fd(-1)
{
  if (d_scene) d_scene->addAudioClip(this);
}


VrmlNodeAudioClip::VrmlNodeAudioClip(const VrmlNodeAudioClip &other) :
  VrmlNode(other),
  d_audio(0)
{
  *this = other;
}

VrmlNodeAudioClip::~VrmlNodeAudioClip()
{
  delete d_audio;
  if (d_scene) d_scene->removeAudioClip(this);
}

VrmlNodeAudioClip& VrmlNodeAudioClip::operator=(const VrmlNodeAudioClip& other)
{
  d_description = other.d_description;
  d_loop = other.d_loop;
  d_pitch = other.d_pitch;
  d_startTime = other.d_startTime;
  d_stopTime = other.d_stopTime;
  d_url = other.d_url;
  d_url_modified = other.d_url_modified;
  d_relativeUrl = other.d_relativeUrl;
  d_duration = other.d_duration;
  d_isActive = other.d_isActive;

  delete d_audio;
  d_audio = (other.d_audio) ? new Audio(*(other.d_audio)) : 0;

  _audio_index = other._audio_index;
  _audio_intensity = other._audio_intensity;
  _audio_fd = other._audio_fd;

  return *this;
}

VrmlNode *VrmlNodeAudioClip::cloneMe() const
{
  return new VrmlNodeAudioClip( *this);
}

void VrmlNodeAudioClip::addToScene(VrmlScene *s, const char *rel)
{
  if (d_scene != s && (d_scene = s) != 0) d_scene->addAudioClip(this);
  d_relativeUrl.set(rel);
}

ostream& VrmlNodeAudioClip::printFields(ostream& os, int indent)
{
  if (d_description.get()) PRINT_FIELD(description);
  if (d_loop.get()) PRINT_FIELD(loop);
  if (! FPEQUAL(d_pitch.get(),1.0)) PRINT_FIELD(pitch);
  if (! FPZERO(d_startTime.get())) PRINT_FIELD(startTime);
  if (! FPZERO(d_stopTime.get())) PRINT_FIELD(stopTime);
  if (d_url.getLength() > 0) PRINT_FIELD(url);
  return os;
}


void VrmlNodeAudioClip::update (VrmlSFTime &inTime)
{
    // If the URL has been modified, update the audio object
    if (d_url_modified)
    {
        Doc relDoc(d_relativeUrl.get(), static_cast<Doc const *>(0));
        delete d_audio;
        d_audio = new Audio(0);
        if (d_audio->tryURLs (d_url.getLength(),
                              d_url.get(),
			      &relDoc))
        {
            d_duration.set (d_audio->duration());
            eventOut (inTime.get(), "duration_changed", d_duration);
        }
        else
        {
#if HAVE_SOUND
          theSystem->error(System::ERRORTYPE_UNKNOWN, "Couldn't read AudioClip from URL %s\n", d_url.get(0));
#endif
            delete d_audio;
            d_audio = 0;
        }

        d_url_modified = false;
    }

    // If there's no audio or START <= 0, we don't play anything
    if (d_audio == 0 || d_startTime.get() <= 0)
        return;

    // Determine if this clip should be playing right now
    bool audible = false;

    // If START < STOP  and  START <= NOW < STOP
    if (d_stopTime.get() > d_startTime.get())
        audible = (d_startTime.get() <= inTime.get() &&
                   inTime.get() < d_stopTime.get());

    // If STOP < START  and  START <= NOW
    else
        audible = (inTime.get() >= d_startTime.get());

    // If the clip is not looping, it's not audible after
    // its duration has expired.
    if (d_loop.get() == false)
        if (inTime.get()-d_startTime.get() > d_audio->duration())
            audible = false;

    // If the clip is audible, play it.  Otherwise, stop it.
    if (audible)
    {
        // If the sound device is not already open, open it.
        if (_audio_fd < 0)
        {
            _audio_fd = openSound (d_audio->channels(),
                                   d_audio->bitsPerSample(),
                                   d_audio->samplesPerSec());

            // If we can't get a sound device, silently return
            if (_audio_fd < 0)
                    return;

            _audio_index = d_audio->getByteIndex
                               (inTime.get() - d_startTime.get());

            d_isActive.set (true);
            eventOut (inTime.get(), "isActive", d_isActive);
        }

        // Send out a sound buffer
        _audio_index = outputSoundChunk (d_audio->numBytes(),
                                         d_audio->samples(),
                                         d_audio->bitsPerSample(),
                                         _audio_index, d_loop.get(),
                                         _audio_intensity,
                                         _audio_fd);
    }

    // Otherwise, close the sound device
    else
    {
        if (_audio_fd >= 0)
        {
            _audio_fd = closeSound (_audio_fd);
            d_isActive.set (false);
            eventOut (inTime.get(), "isActive", d_isActive);
        }
    }
}

// Get the value of a field or eventOut.

const VrmlField *VrmlNodeAudioClip::getField(const char *fieldName) const
{
  // exposedFields
  if ( strcmp( fieldName, "description" ) == 0 )
    return &d_description;
  else if ( strcmp( fieldName, "loop" ) == 0 )
    return &d_loop;
  else if ( strcmp( fieldName, "pitch" ) == 0 )
    return &d_pitch;
  else if ( strcmp( fieldName, "startTime" ) == 0 )
    return &d_startTime;
  else if ( strcmp( fieldName, "stopTime" ) == 0 )
    return &d_stopTime;
  else if ( strcmp( fieldName, "url" ) == 0 )
    return &d_url;

  // eventOuts
  else if ( strcmp( fieldName, "duration" ) == 0 )
    return &d_duration;
  else if ( strcmp( fieldName, "isActive" ) == 0 )
    return &d_isActive;

  return VrmlNode::getField(fieldName); // Parent class
}


// Set the value of one of the node fields.

void VrmlNodeAudioClip::setField(const char *fieldName,
				 const VrmlField &fieldValue)
{
  if TRY_FIELD(description, SFString)
  else if TRY_FIELD(loop, SFBool)
  else if TRY_FIELD(pitch, SFFloat)
  else if TRY_FIELD(startTime, SFTime)
  else if TRY_FIELD(stopTime, SFTime)
  else if TRY_FIELD(url, MFString)
  else
    VrmlNode::setField(fieldName, fieldValue);

  if (strcmp ("url", fieldName) == 0)
  {
      d_url_modified = true;
      setModified();
  }
}
