/*****************************************************************************
******************************************************************************
*
*               VRML CylinderSensor node
*
*               $Revision: 1.1.2.1 $
*               $Date: 2009/01/22 13:25:18 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright (C) 1998  Chris Morley
// Modified by HUMUSOFT s.r.o.  Copyright 1998-2009 HUMUSOFT s.r.o.

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "VrmlNodeCylinderSensor.h"
#include "VrmlNodeType.h"
#include "MathUtils.h"
#include "System.h"
#include "Viewer.h"
#include "VrmlScene.h"

// CylinderSensor factory.

static VrmlNode *creator( VrmlScene *scene )
{
  return new VrmlNodeCylinderSensor(scene);
}


// Define the built in VrmlNodeType:: "CylinderSensor" fields

VrmlNodeType *VrmlNodeCylinderSensor::defineType(VrmlNodeType *t)
{
  static VrmlNodeType *st = 0;

  if (! t)
    {
      if (st) return st;    // Only define the type once.
      t = st = new VrmlNodeType("CylinderSensor", creator);
      t->reference();
    }

  VrmlNodeChild::defineType(t); // Parent class
  t->addExposedField("autoOffset", VrmlField::SFBOOL, VrmlSFBool(true));
  t->addExposedField("diskAngle", VrmlField::SFFLOAT, VrmlSFFloat(0.262));
  t->addExposedField("enabled", VrmlField::SFBOOL, VrmlSFBool(true));
  t->addExposedField("maxAngle", VrmlField::SFFLOAT, VrmlSFFloat(-1));
  t->addExposedField("minAngle", VrmlField::SFFLOAT, VrmlSFFloat(0));
  t->addExposedField("offset", VrmlField::SFFLOAT, VrmlSFFloat(0));
  t->addEventOut("isActive", VrmlField::SFBOOL);
  t->addEventOut("rotation_changed", VrmlField::SFROTATION);
  t->addEventOut("trackPoint_changed", VrmlField::SFVEC3F);

  return t;
}

VrmlNodeType & VrmlNodeCylinderSensor::nodeType() const
{
    return *defineType(0);
}


VrmlNodeCylinderSensor::VrmlNodeCylinderSensor( VrmlScene *scene ) :
  VrmlNodeChild(scene),
  d_autoOffset(true),
  d_diskAngle(0.262),
  d_enabled(true),
  d_maxAngle(-1.0),
  d_minAngle(0.0),
  d_offset(0.0),
  d_isActive(false),
  d_rotation(0.0, 1.0, 0.0, 0.0),
  d_parentTransform(0)
{
  setModified();
  d_activationMatrix.setToIdentity();
  d_diskmode = false;
}


VrmlNodeCylinderSensor::~VrmlNodeCylinderSensor()
{
}


VrmlNode *VrmlNodeCylinderSensor::cloneMe() const
{
  return new VrmlNodeCylinderSensor(*this);
}


ostream& VrmlNodeCylinderSensor::printFields(ostream& os, int indent)
{
  if (! d_autoOffset.get()) PRINT_FIELD(autoOffset);
  if (! FPEQUAL(d_diskAngle.get(), 0.262)) PRINT_FIELD(diskAngle);
  if (! d_enabled.get())    PRINT_FIELD(enabled);
  if (! FPEQUAL(d_maxAngle.get(), -1.0)) PRINT_FIELD(maxAngle);
  if (! FPZERO(d_minAngle.get())) PRINT_FIELD(minAngle);
  if (! FPZERO(d_offset.get())) PRINT_FIELD(offset);

  return os;
}

const VrmlField *VrmlNodeCylinderSensor::getField(const char *fieldName) const
{
  if ( strcmp( fieldName, "autoOffset" ) == 0 )
    return &d_autoOffset;
  else if ( strcmp( fieldName, "diskAngle" ) == 0 )
    return &d_diskAngle;
  else if ( strcmp( fieldName, "enabled" ) == 0 )
    return &d_enabled;
  else if ( strcmp( fieldName, "maxAngle" ) == 0 )
    return &d_maxAngle;
  else if ( strcmp( fieldName, "minAngle" ) == 0 )
    return &d_minAngle;
  else if ( strcmp( fieldName, "offset" ) == 0 )
    return &d_offset;

  // eventOuts
  else if ( strcmp( fieldName, "isActive" ) == 0 )
    return &d_isActive;
  else if ( strcmp( fieldName, "rotation" ) == 0 )
    return &d_rotation;
  else if ( strcmp( fieldName, "trackPoint" ) == 0 )
    return &d_trackPoint;

  return VrmlNodeChild::getField( fieldName );
}

// Set the value of one of the node fields.

void VrmlNodeCylinderSensor::setField(const char *fieldName,
              const VrmlField &fieldValue)
{
  if TRY_FIELD(autoOffset, SFBool)
  else if TRY_FIELD(diskAngle, SFFloat)
  else if TRY_FIELD(enabled, SFBool)
  else if TRY_FIELD(maxAngle, SFFloat)
  else if TRY_FIELD(minAngle, SFFloat)
  else if TRY_FIELD(offset, SFFloat)
  else
    VrmlNodeChild::setField(fieldName, fieldValue);
}

// Cache a pointer to (one of the) parent transforms for converting
// hits into local coords.

void VrmlNodeCylinderSensor::accumulateTransform( VrmlNode *parent )
{
  d_parentTransform = parent;
}

VrmlNode* VrmlNodeCylinderSensor::getParentTransform() { return d_parentTransform; }

void VrmlNodeCylinderSensor::activate( double timeStamp,
            bool isActive,
            double *p )
{
   // becoming active
   if (isActive && !d_isActive.get())
   {
      // update status flag
      d_isActive.set(isActive);

      // store the current world->local transformation matrix
      d_activationMatrix = getCompoundInverseTransformationMatrix();

      // compute the activation point in local coordinates
      Point3 touchPoint = d_activationMatrix*Point3(p[0], p[1], p[2]);
      d_activationPoint.setX(touchPoint.getX());
      d_activationPoint.setY(touchPoint.getY());
      d_activationPoint.setZ(touchPoint.getZ());

      // bearing vector in local coordinates
      Vector3 bearing = Vector3(d_activationMatrix.getElement(2,0),
                                d_activationMatrix.getElement(2,1),
                                d_activationMatrix.getElement(2,2));
      bearing = bearing.normalize();
      Vector3 up = Vector3(0.0, 1.0, 0.0);
      double angle = acos(bearing.dot(up));
      if (angle > M_PI_2) { angle = M_PI - angle; }
      // set sensor behaviour
      d_diskmode = (angle < d_diskAngle.get());

      // generate event
      eventOut(timeStamp, "isActive", d_isActive);
   }

   // becoming inactive
   else if (!isActive && d_isActive.get())
   {
      // update status flag
      d_isActive.set(isActive);

      // generate event
      eventOut(timeStamp, "isActive", d_isActive);

      // set new offset if autooffset enabled
      if (d_autoOffset.get())
      {
          d_offset = d_rotation.getAngle();
          eventOut(timeStamp, "offset_changed", d_offset);
      }
    }

   // dragging
   else if (isActive)
   {
      // compute the activation point in local coordinates
      Point3 touchPoint = d_activationMatrix*Point3(p[0], p[1], p[2]);
      d_trackPoint.setX(touchPoint.getX());
      d_trackPoint.setY(touchPoint.getY());
      d_trackPoint.setZ(touchPoint.getZ());

      // issue event
      eventOut(timeStamp, "trackPoint_changed", d_trackPoint);

      // calculate the rotation
      Vector3 dif1 = Vector3(touchPoint.getX(), 0.0, touchPoint.getZ());
      double radius = d_diskmode ? 1.0 : dif1.length();
      dif1 = dif1.normalize();
      Vector3 dif2 = Vector3(d_activationPoint.getX(),
                             0.0,
                             d_activationPoint.getZ());
      dif2 = dif2.normalize();
      Vector3 cx = dif2.cross(dif1);
      if (cx.length() == 0.0) { return; }
      cx = cx.normalize();
      double angle = radius * acos(dif2.dot(dif1));
      if FPEQUAL(cx.getY(), -1.0) { angle = -angle; }

      // add offset
      angle = angle + d_offset.get();

      // test min/max angle
      if (d_minAngle.get() <= d_maxAngle.get())
      {
         angle = (angle < d_minAngle.get()) ? d_minAngle.get() :
                 ((angle > d_maxAngle.get()) ? d_maxAngle.get() : angle);
      }

      d_rotation = VrmlSFRotation(0.0f, 1.0f, 0.0f, angle);
      // Alternative way of setting this, only the angle changes
      // d_rotation.setAngle(angle);

      // issue an appropriate event
      eventOut(timeStamp, "rotation_changed", d_rotation);

   }
}
