/*****************************************************************************
******************************************************************************
*
*               VRML ElevationGrid node
*
*               $Revision: 1.1.4.4.2.1 $
*               $Date: 2011/07/19 14:53:20 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright (C) 1998  Chris Morley
// Modified by HUMUSOFT s.r.o.  Copyright 1998-2011 HUMUSOFT s.r.o.

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <cstdlib>

#include "VrmlNodeElevationGrid.h"
#include "VrmlNodeType.h"
#include "VrmlNodeColor.h"
#include "VrmlNodeNormal.h"
#include "VrmlNodeProto.h"
#include "VrmlNodeTextureCoordinate.h"
#include "Viewer.h"

#include <stdlib.h>


static VrmlNode *creator( VrmlScene *s ) {
  return new VrmlNodeElevationGrid(s);
}


// Define the built in VrmlNodeType:: "ElevationGrid" fields

VrmlNodeType *VrmlNodeElevationGrid::defineType(VrmlNodeType *t)
{
  static VrmlNodeType *st = 0;

  if (! t)
    {
      if (st) return st;
      t = st = new VrmlNodeType("ElevationGrid", creator);
      t->reference();
    }

  VrmlNodeGeometry::defineType(t);	// Parent class

  t->addEventIn("set_height", VrmlField::MFFLOAT);
  t->addExposedField("color", VrmlField::SFNODE, VrmlSFNode());
  t->addExposedField("normal", VrmlField::SFNODE, VrmlSFNode());
  t->addExposedField("texCoord", VrmlField::SFNODE, VrmlSFNode());
  t->addField("height", VrmlField::MFFLOAT, VrmlMFFloat());
  t->addField("ccw", VrmlField::SFBOOL, VrmlSFBool(true));
  t->addField("colorPerVertex", VrmlField::SFBOOL, VrmlSFBool(true));
  t->addField("creaseAngle", VrmlField::SFFLOAT, VrmlSFFloat(0));
  t->addField("normalPerVertex", VrmlField::SFBOOL, VrmlSFBool(true));
  t->addField("solid", VrmlField::SFBOOL, VrmlSFBool(true));
  t->addField("xDimension", VrmlField::SFINT32, VrmlSFInt32(0));
  t->addField("xSpacing", VrmlField::SFFLOAT, VrmlSFFloat(1.0));
  t->addField("zDimension", VrmlField::SFINT32, VrmlSFInt32(0));
  t->addField("zSpacing", VrmlField::SFFLOAT, VrmlSFFloat(1.0));

  return t;
}

VrmlNodeType & VrmlNodeElevationGrid::nodeType() const
{
    return *defineType(0);
}


VrmlNodeElevationGrid::VrmlNodeElevationGrid(VrmlScene *scene) :
  VrmlNodeGeometry(scene),
  d_ccw(true),
  d_colorPerVertex(true),
  d_normalPerVertex(true),
  d_solid(true),
  d_xDimension(0),
  d_xSpacing(1.0f),
  d_zDimension(0),
  d_zSpacing(1.0f)
{
}

VrmlNodeElevationGrid::~VrmlNodeElevationGrid()
{
}

VrmlNode *VrmlNodeElevationGrid::cloneMe() const
{
  return new VrmlNodeElevationGrid(*this);
}

void VrmlNodeElevationGrid::cloneChildren(VrmlNamespace *ns)
{
  if (d_color.get())
    d_color.set(d_color.get()->clone(ns));
  if (d_normal.get())
    d_normal.set(d_normal.get()->clone(ns));
  if (d_texCoord.get())
    d_texCoord.set(d_texCoord.get()->clone(ns));
}


bool VrmlNodeElevationGrid::isModified() const
{
  return ( VrmlNodeGeometry::isModified() ||
	   (d_color.get() && d_color.get()->isModified()) ||
	   (d_normal.get() && d_normal.get()->isModified()) ||
	   (d_texCoord.get() && d_texCoord.get()->isModified()) );
}

void VrmlNodeElevationGrid::updateModified(VrmlNodePath& path)
{
  if (this->isModified()) markPathModified(path, true);
  path.push_front(this);
  if (d_color.get()) d_color.get()->updateModified(path);
  if (d_normal.get()) d_normal.get()->updateModified(path);
  if (d_texCoord.get()) d_texCoord.get()->updateModified(path);
  path.pop_front();
}


void VrmlNodeElevationGrid::clearFlags()
{
  VrmlNode::clearFlags();
  if (d_color.get()) d_color.get()->clearFlags();
  if (d_normal.get()) d_normal.get()->clearFlags();
  if (d_texCoord.get()) d_texCoord.get()->clearFlags();
}

void VrmlNodeElevationGrid::addToScene( VrmlScene *s, const char *rel )
{
  d_scene = s;
  if (d_color.get()) d_color.get()->addToScene(s, rel);
  if (d_normal.get()) d_normal.get()->addToScene(s, rel);
  if (d_texCoord.get()) d_texCoord.get()->addToScene(s, rel);
}

ostream& VrmlNodeElevationGrid::printFields(ostream& os, int indent)
{
  if (d_color.get()) PRINT_FIELD(color);
  if (d_normal.get()) PRINT_FIELD(normal);
  if (d_texCoord.get()) PRINT_FIELD(texCoord);

  if (! d_ccw.get()) PRINT_FIELD(ccw);
  if (! d_colorPerVertex.get()) PRINT_FIELD(colorPerVertex);
  if (! d_normalPerVertex.get()) PRINT_FIELD(normalPerVertex);
  if (! d_solid.get()) PRINT_FIELD(solid);

  if (d_creaseAngle.get() != 0.0) PRINT_FIELD(creaseAngle);
  if (d_height.getLength() > 0) PRINT_FIELD(height);

  if (d_xDimension.get() != 0) PRINT_FIELD(xDimension);
  if (d_xSpacing.get() != 0) PRINT_FIELD(xSpacing);
  if (d_zDimension.get() != 0) PRINT_FIELD(zDimension);
  if (d_zSpacing.get() != 0) PRINT_FIELD(zSpacing);

  return os;
}

VrmlNodeColor *VrmlNodeElevationGrid::color()
{
  return NodeDynamicCast<VrmlNodeColor*>(d_color.get());
}


Viewer::Object VrmlNodeElevationGrid::insertGeometry(Viewer *viewer, VrmlRenderContext rc)
{
  Viewer::Object obj = 0;

  // determine the node name to be printed in warnings
  const char* nodename = (*(name()) == '\0') ? "(unnamed ElevationGrid)" : name();

  if ( d_height.getLength() > 0 )
    {
      int x_dimension = d_xDimension.get();
      if (x_dimension < 0)
      {
	theSystem->warn(System::SEVERITY_LOW, System::WARNINGTYPE_UNKNOWN,
	                "Value in %s.xDimension must be nonnegative", nodename);
	x_dimension = abs(x_dimension);
      }

      int z_dimension = d_zDimension.get();
      if (z_dimension < 0)
      {
	theSystem->warn(System::SEVERITY_LOW, System::WARNINGTYPE_UNKNOWN,
	                "Value in %s.zDimension must be nonnegative", nodename);
	z_dimension = abs(z_dimension);
      }

      float x_spacing = d_xSpacing.get();
      if (x_spacing<=0)
	theSystem->warn(System::SEVERITY_LOW, System::WARNINGTYPE_UNKNOWN,
	                "Value in %s.xSpacing must be positive", nodename);

      float z_spacing = d_zSpacing.get();
      if (z_spacing<=0)
	theSystem->warn(System::SEVERITY_LOW, System::WARNINGTYPE_UNKNOWN,
	                "Value in %s.zSpacing must be positive", nodename);

      bool colorPerVertex = d_colorPerVertex.get();
      const float* colors = 0;
      VrmlNodeColor* cnode = NodeDynamicCast<VrmlNodeColor*>(d_color.get());
      if (cnode != NULL)
      {
	colors = cnode->color()[0];
	long ncolors = cnode->color().getLength();
	if (!colorPerVertex && ncolors>0 && ncolors<((z_spacing-1)*(x_spacing-1)))
          theSystem->warn(System::SEVERITY_LOW, System::WARNINGTYPE_UNKNOWN,
                          "Not enough values for %s.color.color: %d (%d is required)",
                           nodename, ncolors,(z_spacing-1)*(x_spacing-1));

	if (colorPerVertex && ncolors>0 && ncolors<(z_spacing*x_spacing))
          theSystem->warn(System::SEVERITY_LOW, System::WARNINGTYPE_UNKNOWN,
                          "Not enough values for %s.color.color: %d (%d is required)",
                           nodename, ncolors, z_spacing*x_spacing);
      }

      bool normalPerVertex = d_normalPerVertex.get();
      const float *normals = 0;
      VrmlNodeNormal* nnode = NodeDynamicCast<VrmlNodeNormal*>(d_normal.get());
      if (nnode != NULL)
      {
        normals = nnode->normal()[0];
	long nnormals = nnode->normal().getLength();
	if (!normalPerVertex && nnormals>0 && nnormals<((z_spacing-1)*(x_spacing-1)))
          theSystem->warn(System::SEVERITY_LOW, System::WARNINGTYPE_UNKNOWN,
                          "Not enough values for %s.normal.normal: %d (%d is required)",
                           nodename, nnormals,(z_spacing-1)*(x_spacing-1));
	if (normalPerVertex && nnormals>0 && nnormals<(z_spacing*x_spacing))
          theSystem->warn(System::SEVERITY_LOW, System::WARNINGTYPE_UNKNOWN,
                          "Not enough values for %s.normal.normal: %d (%d is required)",
                           nodename, nnormals, z_spacing*x_spacing);
      }

      const float *tc = 0;
      VrmlNodeTextureCoordinate* tcnode = NodeDynamicCast<VrmlNodeTextureCoordinate*>(d_texCoord.get());
      if (tcnode != NULL)
      {
        tc = tcnode->coordinate()[0];
	long ntexcoords = tcnode->coordinate().getLength();
	if (ntexcoords<(z_spacing*x_spacing))
          theSystem->warn(System::SEVERITY_LOW, System::WARNINGTYPE_UNKNOWN,
                          "Not enough values for %s.texCoord.point %d (%d is required)",
                           nodename, ntexcoords, z_spacing*x_spacing);
      }

      // insert geometry
      unsigned int optMask = 0;
      if (d_ccw.get()) optMask |= Viewer::MASK_CCW;
      if (d_solid.get()) optMask |= Viewer::MASK_SOLID;
      if (colorPerVertex) optMask |= Viewer::MASK_COLOR_PER_VERTEX;
      if (normalPerVertex) optMask |= Viewer::MASK_NORMAL_PER_VERTEX;

      obj = viewer->insertElevationGrid(optMask,
					x_dimension,
					z_dimension,
					d_height.get(),
					x_spacing,
					z_spacing,
					tc,
					normals,
					colors);
    }

  if (d_color.get()) d_color.get()->clearModified();
  if (d_normal.get()) d_normal.get()->clearModified();
  if (d_texCoord.get()) d_texCoord.get()->clearModified();

  return obj;
}

// Get the value of one of the exposedFields or eventOuts.

const VrmlField *VrmlNodeElevationGrid::getField(const char *fieldName) const
{
  if ( strcmp( fieldName, "color" ) == 0 )
    return &d_color;
  else if ( strcmp( fieldName, "normal" ) == 0 )
    return &d_normal;
  else if ( strcmp( fieldName, "texCoord" ) == 0 )
    return &d_texCoord;

#if defined(__ORBISNAP__)

  else if (strcmp(fieldName, "height") == 0)
    return &d_height;
  else if (strcmp(fieldName, "ccw") == 0)
    return &d_ccw;
  else if (strcmp(fieldName, "colorPerVertex") == 0)
    return &d_colorPerVertex;
  else if (strcmp(fieldName, "creaseAngle") == 0)
    return &d_creaseAngle;
  else if (strcmp(fieldName, "normalPerVertex") == 0)
    return &d_normalPerVertex;
  else if (strcmp(fieldName, "solid") == 0)
    return &d_solid;
  else if (strcmp(fieldName, "xDimension") == 0)
    return &d_xDimension;
  else if (strcmp(fieldName, "xSpacing") == 0)
    return &d_xSpacing;
  else if (strcmp(fieldName, "zDimension") == 0)
    return &d_zDimension;
  else if (strcmp(fieldName, "zSpacing") == 0)
    return &d_zSpacing;

#endif

  return VrmlNodeGeometry::getField( fieldName );
}

// Set the value of one of the node fields.

void VrmlNodeElevationGrid::setField(const char *fieldName,
				     const VrmlField &fieldValue)
{
  if TRY_SFNODE_FIELD(color, Color)
  else if TRY_SFNODE_FIELD(normal, Normal)
  else if TRY_SFNODE_FIELD(texCoord, TextureCoordinate)
  else if TRY_FIELD(ccw, SFBool)
  else if TRY_FIELD(colorPerVertex, SFBool)
  else if TRY_FIELD(creaseAngle, SFFloat)
  else if TRY_FIELD(height, MFFloat)
  else if TRY_FIELD(normalPerVertex, SFBool)
  else if TRY_FIELD(solid, SFBool)
  else if TRY_FIELD(xDimension, SFInt32)
  else if TRY_FIELD(xSpacing, SFFloat)
  else if TRY_FIELD(zDimension, SFInt32)
  else if TRY_FIELD(zSpacing, SFFloat)
  else
    VrmlNodeGeometry::setField(fieldName, fieldValue);
}

VrmlNode* VrmlNodeElevationGrid::getNormal()  // LarryD Mar 09/99
{   return d_normal.get(); }

VrmlNode* VrmlNodeElevationGrid::getTexCoord()  // LarryD Mar 09/99
{   return d_texCoord.get(); }

const VrmlMFFloat& VrmlNodeElevationGrid::getHeight() const  // LarryD Mar 09/99
{   return d_height; }

void VrmlNodeElevationGrid::setProto(VrmlNodeProto* proto)
{
  if (d_color.get()) d_color.get()->setProto(proto);
  if (d_normal.get()) d_normal.get()->setProto(proto);
  if (d_texCoord.get()) d_texCoord.get()->setProto(proto);
  VrmlNodeGeometry::setProto(proto);
}

void VrmlNodeElevationGrid::setEncapsulatingPrototype(VrmlNodeType* type)
{
  if (d_color.get()) d_color.get()->setEncapsulatingPrototype(type);
  if (d_normal.get()) d_normal.get()->setEncapsulatingPrototype(type);
  if (d_texCoord.get()) d_texCoord.get()->setEncapsulatingPrototype(type);
  VrmlNodeGeometry::setEncapsulatingPrototype(type);
}
