/*****************************************************************************
******************************************************************************
*
*               VRML Extrusion node
*
*               $Revision: 1.1.2.3 $
*               $Date: 2010/07/21 10:17:23 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright (C) 1998  Chris Morley
// Modified by HUMUSOFT s.r.o.  Copyright 1998-2009 HUMUSOFT s.r.o.

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#include "VrmlNodeExtrusion.h"
#include "VrmlNodeType.h"
#include "Viewer.h"


static VrmlNode *creator( VrmlScene *s ) { return new VrmlNodeExtrusion(s); }


// Define the built in VrmlNodeType:: "Extrusion" fields

VrmlNodeType *VrmlNodeExtrusion::defineType(VrmlNodeType *t)
{
  static VrmlNodeType *st = 0;

  if (! t)
    {
      if (st) return st;
      t = st = new VrmlNodeType("Extrusion", creator);
      t->reference();
    }

  VrmlNodeGeometry::defineType(t);	// Parent class

  t->addEventIn("set_crossSection", VrmlField::MFVEC2F);
  t->addEventIn("set_orientation", VrmlField::MFROTATION);
  t->addEventIn("set_scale", VrmlField::MFVEC2F);
  t->addEventIn("set_spine", VrmlField::MFVEC3F);

  t->addField("beginCap", VrmlField::SFBOOL, VrmlSFBool(true));
  t->addField("ccw", VrmlField::SFBOOL, VrmlSFBool(true));
  t->addField("convex", VrmlField::SFBOOL, VrmlSFBool(true));
  t->addField("creaseAngle", VrmlField::SFFLOAT, VrmlSFFloat(0));
  float cross[10] = { 1, 1, 1, -1, -1, -1, -1, 1, 1, 1 };
  t->addField("crossSection", VrmlField::MFVEC2F, VrmlMFVec2f(5, cross));
  t->addField("endCap", VrmlField::SFBOOL, VrmlSFBool(true));
  float ori[4] = { 0, 0, 1, 0 };
  t->addField("orientation", VrmlField::MFROTATION, VrmlMFRotation(1, ori));
  float sc[2] = { 1, 1 };
  t->addField("scale", VrmlField::MFVEC2F, VrmlMFVec2f(1, sc));
  t->addField("solid", VrmlField::SFBOOL, VrmlSFBool(true));
  float sp[6] = { 0, 0, 0, 0, 1, 0 };
  t->addField("spine", VrmlField::MFVEC3F, VrmlMFVec3f(2, sp));

  return t;
}


VrmlNodeType & VrmlNodeExtrusion::nodeType() const
{
    return *defineType(0);
}


VrmlNodeExtrusion::VrmlNodeExtrusion(VrmlScene * scene):
        VrmlNodeGeometry(scene), d_beginCap(true), d_ccw(true), d_convex(true),
        d_creaseAngle(0), d_crossSection(5), d_endCap(true), d_orientation(1),
        d_scale(1), d_solid(true), d_spine(2) {
    this->d_crossSection[0][0] = 1.0f;
    this->d_crossSection[0][1] = 1.0f;
    this->d_crossSection[1][0] = 1.0f;
    this->d_crossSection[1][1] = -1.0f;
    this->d_crossSection[2][0] = -1.0f;
    this->d_crossSection[2][1] = -1.0f;
    this->d_crossSection[3][0] = -1.0f;
    this->d_crossSection[3][1] = 1.0f;
    this->d_crossSection[4][0] = 1.0f;
    this->d_crossSection[4][1] = 1.0f;

    this->d_scale[0][0] = 1.0f;
    this->d_scale[0][1] = 1.0f;

    this->d_spine[0][0] = 0.0f;
    this->d_spine[0][1] = 0.0f;
    this->d_spine[0][2] = 0.0f;
    this->d_spine[1][0] = 0.0f;
    this->d_spine[1][1] = 1.0f;
    this->d_spine[1][2] = 0.0f;

#if defined(__ORBISNAP__)

    // we must initialize this as well, the constructor only specifies size
    // but does not fill the array with values!
    d_orientation[0][0] = 0.0f;
    d_orientation[0][1] = 0.0f;
    d_orientation[0][2] = 1.0f;
    d_orientation[0][3] = 0.0f;

#endif
}

VrmlNodeExtrusion::~VrmlNodeExtrusion()
{
}


VrmlNode *VrmlNodeExtrusion::cloneMe() const
{
  return new VrmlNodeExtrusion(*this);
}


ostream& VrmlNodeExtrusion::printFields(ostream& os, int indent)
{
  if (! d_beginCap.get()) PRINT_FIELD(beginCap);
  if (! d_endCap.get()) PRINT_FIELD(endCap);
  if (! d_ccw.get()) PRINT_FIELD(ccw);
  if (! d_convex.get()) PRINT_FIELD(convex);
  if (! d_solid.get()) PRINT_FIELD(solid);

  if (d_creaseAngle.get() != 0.0) PRINT_FIELD(creaseAngle);
  PRINT_FIELD(crossSection);
  PRINT_FIELD(orientation);
  PRINT_FIELD(scale);
  PRINT_FIELD(spine);

  return os;
}


Viewer::Object VrmlNodeExtrusion::insertGeometry(Viewer *viewer, VrmlRenderContext rc)
{
  Viewer::Object obj = 0;
  if ( d_crossSection.getLength() > 0 &&
       d_spine.getLength() > 1 )
    {
      unsigned int optMask = 0;
      if (d_ccw.get()) optMask |= Viewer::MASK_CCW;
      if (d_convex.get()) optMask |= Viewer::MASK_CONVEX;
      if (d_solid.get()) optMask |= Viewer::MASK_SOLID;
      if (d_beginCap.get()) optMask |= Viewer::MASK_BOTTOM;
      if (d_endCap.get()) optMask |= Viewer::MASK_TOP;

      long spineLength = d_spine.getLength();
      long orientationLength = d_orientation.getLength();
      long scaleLength = d_scale.getLength();

      // determine the node name to be printed in warnings
      const char* nodename = (*(name()) == '\0') ? "(unnamed Extrusion)" : name();

      if (orientationLength !=1 && orientationLength < spineLength)
        theSystem->warn(System::SEVERITY_LOW, System::WARNINGTYPE_UNKNOWN,
                   "There must be %d or %d orientation vectors in %s.orientation", 1, spineLength, nodename);

      if (scaleLength !=1 && scaleLength < spineLength)
	theSystem->warn(System::SEVERITY_LOW, System::WARNINGTYPE_UNKNOWN,
	                "There must be %d or %d scale vectors in %s.scale", 1, spineLength, nodename);

      for (long i = 0; i < scaleLength; i++)
	if (d_scale[i][0] < 0 || d_scale[i][1] < 0)
	  theSystem->warn(System::SEVERITY_LOW, System::WARNINGTYPE_UNKNOWN,
	                  "Values in %s.scale must be nonnegative", nodename);

      obj = viewer->insertExtrusion(optMask,
				    orientationLength,
				    d_orientation.get(),
				    scaleLength,
				    d_scale.get(),
				    d_crossSection.getLength(),
				    d_crossSection.get(),
				    spineLength,
				    d_spine.get(),
                                    d_creaseAngle.get());
    }

  return obj;
}

#if defined(__ORBISNAP__)

const VrmlField* VrmlNodeExtrusion::getField(const char *fieldName) const
{
  if (strcmp(fieldName, "beginCap") == 0)
    return &d_beginCap;
  else if (strcmp(fieldName, "ccw") == 0)
    return &d_ccw;
  else if (strcmp(fieldName, "convex") == 0)
    return &d_convex;
  else if (strcmp(fieldName, "creaseAngle") == 0)
    return &d_creaseAngle;
  else if (strcmp(fieldName, "crossSection") == 0)
    return &d_crossSection;
  else if (strcmp(fieldName, "endCap") == 0)
    return &d_endCap;
  else if (strcmp(fieldName, "orientation") == 0)
    return &d_orientation;
  else if (strcmp(fieldName, "scale") == 0)
    return &d_scale;
  else if (strcmp(fieldName, "solid") == 0)
    return &d_solid;
  else if (strcmp(fieldName, "spine") == 0)
    return &d_spine;
  return VrmlNodeGeometry::getField( fieldName );
}

#endif

// Set the value of one of the node fields.

void VrmlNodeExtrusion::setField(const char *fieldName,
				 const VrmlField &fieldValue)
{
  if TRY_FIELD(beginCap, SFBool)
  else if TRY_FIELD(ccw, SFBool)
  else if TRY_FIELD(convex, SFBool)
  else if TRY_FIELD(creaseAngle, SFFloat)
  else if TRY_FIELD(crossSection, MFVec2f)
  else if TRY_FIELD(endCap, SFBool)
  else if TRY_FIELD(orientation, MFRotation)
  else if TRY_FIELD(scale, MFVec2f)
  else if TRY_FIELD(solid, SFBool)
  else if TRY_FIELD(spine, MFVec3f)
  else
    VrmlNodeGeometry::setField(fieldName, fieldValue);
}

bool VrmlNodeExtrusion::defaultSpineOrientationChanged(size_t nSpine, const float * spine)
{
  bool orientationChanged = false;
  if (nSpine>1)
  {
    Vector3 spine1(spine);
    Vector3 spine2(spine + 3);
    Vector3 spineOri = spine2 - spine1;
    spineOri = spineOri.normalize();
    Vector3 upVec(0,1,0);

    if (fabs(acos(spineOri.dot(upVec))) > M_PI_2)
      orientationChanged = true;
  }
  return orientationChanged;
}
