/*****************************************************************************
******************************************************************************
*
*               VRML IndexedFaceSet node
*
*               $Revision: 1.1.2.8 $
*               $Date: 2011/07/19 14:53:20 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright 1998-2011 HUMUSOFT s.r.o.
// Portions Copyright (C) 1998  Chris Morley

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#include "VrmlNodeIFaceSet.h"
#include "VrmlNodeType.h"
#include "VrmlNodeColor.h"
#include "VrmlNodeCoordinate.h"
#include "VrmlNodeNormal.h"
#include "VrmlNodeProto.h"
#include "VrmlNodeTextureCoordinate.h"
#include "Viewer.h"

#include "MathUtils.h"

static VrmlNode *creator( VrmlScene *s ) { return new VrmlNodeIFaceSet(s); }

// Define the built in VrmlNodeType:: "IndexedFaceSet" fields

VrmlNodeType *VrmlNodeIFaceSet::defineType(VrmlNodeType *t)
{
  static VrmlNodeType *st = 0;

  if (! t)
    {
      if (st) return st;
      t = st = new VrmlNodeType("IndexedFaceSet", creator);
      t->reference();
    }

  // VrmlNodeIndexedSet::defineType(t);	// Parent class
  // don't call parent and rather add new fields directly in this class.
  // It should help to keep order of fields.


  t->addEventIn("set_colorIndex", VrmlField::MFINT32);
  t->addEventIn("set_coordIndex", VrmlField::MFINT32);
  t->addEventIn("set_normalIndex", VrmlField::MFINT32);
  t->addEventIn("set_texCoordIndex", VrmlField::MFINT32);

  t->addExposedField("color", VrmlField::SFNODE, VrmlSFNode());
  t->addExposedField("coord", VrmlField::SFNODE, VrmlSFNode());
  t->addExposedField("normal", VrmlField::SFNODE, VrmlSFNode());
  t->addExposedField("texCoord", VrmlField::SFNODE, VrmlSFNode());

  t->addField("ccw", VrmlField::SFBOOL, VrmlSFBool(true));
  t->addField("colorIndex", VrmlField::MFINT32, VrmlMFInt32());
  t->addField("colorPerVertex", VrmlField::SFBOOL, VrmlSFBool(true));
  t->addField("convex", VrmlField::SFBOOL, VrmlSFBool(true));
  t->addField("coordIndex", VrmlField::MFINT32, VrmlMFInt32());
  t->addField("creaseAngle", VrmlField::SFFLOAT, VrmlSFFloat(0));
  t->addField("normalIndex", VrmlField::MFINT32, VrmlMFInt32());
  t->addField("normalPerVertex", VrmlField::SFBOOL, VrmlSFBool(true));
  t->addField("solid", VrmlField::SFBOOL, VrmlSFBool(true));
  t->addField("texCoordIndex", VrmlField::MFINT32, VrmlMFInt32());

  return t;
}


VrmlNodeType & VrmlNodeIFaceSet::nodeType() const
{
    return *defineType(0);
}


VrmlNodeIFaceSet::VrmlNodeIFaceSet(VrmlScene *scene) :
  VrmlNodeIndexedSet(scene),
  d_ccw(true),
  d_convex(true),
  d_creaseAngle(0.0),
  d_normalPerVertex(true),
  d_solid(true)
{
  this->setBVolumeDirty(true);
}

VrmlNodeIFaceSet::~VrmlNodeIFaceSet()
{
}


VrmlNode *VrmlNodeIFaceSet::cloneMe() const
{
  return new VrmlNodeIFaceSet(*this);
}

void VrmlNodeIFaceSet::cloneChildren(VrmlNamespace* ns)
{
  if (d_color.get())
    d_color.set(d_color.get()->clone(ns));
  if (d_coord.get())
    d_coord.set(d_coord.get()->clone(ns));
  if (d_normal.get())
    d_normal.set(d_normal.get()->clone(ns));
  if (d_texCoord.get())
    d_texCoord.set(d_texCoord.get()->clone(ns));
}


bool VrmlNodeIFaceSet::isModified() const
{
  return ( VrmlNodeIndexedSet::isModified() ||
	   (d_color.get() && d_color.get()->isModified()) ||
	   (d_coord.get() && d_coord.get()->isModified()) ||
	   (d_normal.get() && d_normal.get()->isModified()) ||
	   (d_texCoord.get() && d_texCoord.get()->isModified()) );
}

void VrmlNodeIFaceSet::updateModified(VrmlNodePath& path, int flags)
{
  //cout << "VrmlNodeIFaceSet::updateModified()" << endl;
  if (this->isModified()) markPathModified(path, true, flags);
  path.push_front(this);
  if (d_color.get()) d_color.get()->updateModified(path, flags);
  if (d_coord.get()) d_coord.get()->updateModified(path, flags);
  if (d_normal.get()) d_normal.get()->updateModified(path, flags);
  if (d_texCoord.get()) d_texCoord.get()->updateModified(path, flags);
  path.pop_front();
}

void VrmlNodeIFaceSet::clearFlags()
{
  VrmlNode::clearFlags();
  if (d_color.get()) d_color.get()->clearFlags();
  if (d_coord.get()) d_coord.get()->clearFlags();
  if (d_normal.get()) d_normal.get()->clearFlags();
  if (d_texCoord.get()) d_texCoord.get()->clearFlags();
}

void VrmlNodeIFaceSet::addToScene( VrmlScene *s, const char *rel )
{
  d_scene = s;
  if (d_color.get()) d_color.get()->addToScene(s, rel);
  if (d_coord.get()) d_coord.get()->addToScene(s, rel);
  if (d_normal.get()) d_normal.get()->addToScene(s, rel);
  if (d_texCoord.get()) d_texCoord.get()->addToScene(s, rel);
}

ostream& VrmlNodeIFaceSet::printFields(ostream& os, int indent)
{
  if (! d_ccw.get()) PRINT_FIELD(ccw);
  if (! d_convex.get()) PRINT_FIELD(convex);
  if (! d_normalPerVertex.get()) PRINT_FIELD(normalPerVertex);
  if (! d_solid.get()) PRINT_FIELD(solid);

  if (d_creaseAngle.get() != 0.0) PRINT_FIELD(creaseAngle);
  if (d_normal.get()) PRINT_FIELD(normal);
  if (d_normalIndex.getLength() > 0) PRINT_FIELD(normalIndex);
  if (d_texCoord.get()) PRINT_FIELD(texCoord);
  if (d_texCoordIndex.getLength() > 0) PRINT_FIELD(texCoordIndex);

  VrmlNodeIndexedSet::printFields(os, indent);

  return os;
}



// TO DO: stripify, crease angle, generate normals ...

Viewer::Object VrmlNodeIFaceSet::insertGeometry(Viewer *viewer,
                                                VrmlRenderContext rc)
{
  Viewer::Object obj = 0;

  // these fields hold indices to the coordinate/color/normal tables
  const long*  coordIndices       = d_coordIndex.get();
  long         nCoordIndices      = d_coordIndex.getLength();
  const long*  texCoordIndices    = d_texCoordIndex.get();
  long         nTexCoordIndices   = d_texCoordIndex.getLength();
  const long*  colorIndices       = d_colorIndex.get();
  long         nColorIndices      = d_colorIndex.getLength();
  const long*  normalIndices      = d_normalIndex.get();
  long         nNormalIndices     = d_normalIndex.getLength();
  long         nFaces             = 0;
  long         maxCoordIndex      = -1;
  long         i;
  bool         normalPerVertex    = d_normalPerVertex.get();

  // the tables are stored in child nodes (some may be missing)
  VrmlNodeCoordinate*         coordNode;
  VrmlNodeTextureCoordinate*  texCoordNode;
  VrmlNodeColor*              colorNode;
  VrmlNodeNormal*             normalNode;
  coordNode    = NodeDynamicCast<VrmlNodeCoordinate*>(d_coord.get());
  texCoordNode = NodeDynamicCast<VrmlNodeTextureCoordinate*>(d_texCoord.get());
  colorNode    = NodeDynamicCast<VrmlNodeColor*>(d_color.get());
  normalNode   = NodeDynamicCast<VrmlNodeNormal*>(d_normal.get());
  const float* vertices  = coordNode    ? coordNode->coordinate().get() : 0;
  const float* texCoords = texCoordNode ? texCoordNode->coordinate().get() : 0;
  const float* colors    = colorNode    ? colorNode->color().get() : 0;
  const float* normals   = normalNode   ? normalNode->normal().get() : 0;
  long nVertices  = coordNode    ? coordNode->coordinate().getLength() : 0;
  long nTexCoords = texCoordNode ? texCoordNode->coordinate().getLength() : 0;
  long nColors    = colorNode    ? colorNode->color().getLength() : 0;
  long nNormals   = normalNode   ? normalNode->normal().getLength() : 0;

  // determine the node name to be printed in warnings
  const char* nodename = (*(name()) == '\0') ? "(unnamed IndexedFaceSet)" : name();

  //:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
  // check coord indices and count faces
  //:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

  for (i=0; i<nCoordIndices; i++)
  {
    if (coordIndices[i] >= nVertices || coordIndices[i] < -1)
    {
      theSystem->warn(System::SEVERITY_LOW, System::WARNINGTYPE_UNKNOWN,
        "Invalid value for %s.coordIndex(%d): %d (max is %d)",
        nodename, i, coordIndices[i], nVertices - 1);
      isValid = false;
      return(obj);
    }

    // count faces
    if (coordIndices[i] == -1) nFaces++;

    // remember the highest coord index
    if (maxCoordIndex < coordIndices[i]) maxCoordIndex = coordIndices[i];
  }

  // last face does not have to have "-1" separator at the end
  if (nCoordIndices > 0 && coordIndices[i-1] != -1) nFaces++;

  //:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
  // check texture coordinates
  //:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

#define DISCARD_TEXCOORDS \
  { nTexCoords = 0; nTexCoordIndices = 0; texCoords = 0; texCoordIndices = 0; }

  if (nTexCoordIndices > 0 && nTexCoords == 0)
  {
    // common and harmless case (possibly a temporary state), do not complain
    DISCARD_TEXCOORDS;
  }
  if (nTexCoordIndices == 0 && nTexCoords != 0 && nTexCoords <= maxCoordIndex)
  {
    theSystem->warn(System::SEVERITY_LOW, System::WARNINGTYPE_UNKNOWN,
      "Not enough values for %s.texCoord.point: %d (%d is required)",
      nodename, nTexCoords, maxCoordIndex + 1);
    DISCARD_TEXCOORDS;
  }
  if (nTexCoordIndices > 0 && nTexCoordIndices < nCoordIndices)
  {
    theSystem->warn(System::SEVERITY_LOW, System::WARNINGTYPE_UNKNOWN,
      "Not enough values for %s.texCoordIndex: %d (%d is required)",
      nodename, nTexCoordIndices, nCoordIndices);
    DISCARD_TEXCOORDS;
  }

  for (i=0; i<nTexCoordIndices; i++)
  {
    if (texCoordIndices[i] >= nTexCoords || texCoordIndices[i] < -1)
    {
      theSystem->warn(System::SEVERITY_LOW, System::WARNINGTYPE_UNKNOWN,
        "Invalid value for %s.texCoordIndex(%d): %d (max is %d)",
        nodename, i, texCoordIndices[i], nTexCoords - 1);
      DISCARD_TEXCOORDS;
      break;
    }
    if (texCoordIndices[i] == -1 && coordIndices[i] != -1)
    {
      theSystem->warn(System::SEVERITY_LOW, System::WARNINGTYPE_UNKNOWN,
        "Invalid value for %s.texCoordIndex(%d): -1",
        nodename, i);
      DISCARD_TEXCOORDS;
      break;
    }
  }

  checkIndexSetSemantics(nodename,
                         coordIndices,
                         nCoordIndices,
                         colorIndices,
                         nColorIndices,
                         nVertices,
                         colors,
                         nColors,
                         nFaces,
                         maxCoordIndex);

  //:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
  // check normals and normal indices
  //:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

#define DISCARD_NORMALS \
  { nNormalIndices = 0; nNormals = 0; normalIndices = 0; normals = 0; }

  if (nNormalIndices > 0 && nNormals == 0)
  {
    // common and harmless case (possibly a temporary state), do not complain
    DISCARD_NORMALS;
  }
  long nNormalIndicesRequired = normalPerVertex ? nCoordIndices : nFaces;
  if (nNormalIndices != 0 && nNormalIndices < nNormalIndicesRequired)
  {
    theSystem->warn(System::SEVERITY_LOW, System::WARNINGTYPE_UNKNOWN,
      "Not enough values for %s.normalIndex: %d (%d is required)",
      nodename, nNormalIndices, nNormalIndicesRequired);
    DISCARD_NORMALS;
  }
  else if (nNormalIndices == 0 && nNormals != 0 && (normalPerVertex ?
           (nNormals <= maxCoordIndex) : (nNormals < nFaces)))
  {
    theSystem->warn(System::SEVERITY_LOW, System::WARNINGTYPE_UNKNOWN,
      "Not enough values for %s.normal.normal: %d (%d is required)",
      nodename, nNormals,
      normalPerVertex ? (maxCoordIndex + 1) : nFaces);
    DISCARD_NORMALS;
  }
  for (i=0; i<nNormalIndices; i++)
  {
    if (normalIndices[i] >= nNormals || normalIndices[i] < -1 ||
        (!normalPerVertex && normalIndices[i] < 0))
    {
      theSystem->warn(System::SEVERITY_LOW, System::WARNINGTYPE_UNKNOWN,
        "Invalid value for %s.normalIndex(%d): %d (max is %d)",
        nodename, i, normalIndices[i], nNormals - 1);
      DISCARD_NORMALS;
      break;
    }
    if (normalIndices[i] == -1 && (!normalPerVertex || coordIndices[i] != -1))
    {
      theSystem->warn(System::SEVERITY_LOW, System::WARNINGTYPE_UNKNOWN,
        "Invalid value for %s.normalIndex(%d): -1",
        nodename, i);
      DISCARD_NORMALS;
      break;
    }
  }

  //:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
  // calculate face pointers and normals
  //:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

  std::vector<float> autoNormals;
  std::vector<long>  autoNormalIndices;
  if (nCoordIndices > 0 && nNormals == 0)
  {
    calcFacePointers();

    std::vector<Vector3> newNormals;
    std::vector<Vector3> correctedFaceNormals;

    long facenum = 0;
    // counts number of vertices in one face
    long verticesCount = 0;
    autoNormalIndices.reserve(nCoordIndices);
    for (i=0; i<nCoordIndices; i++)
    {
      autoNormalIndices.push_back(i);

      long vertexIndex;
      vertexIndex = coordIndices[i];
      if (vertexIndex == -1)
      {
	// there must be at least three vertices in one face
	// if not skip the incomplete face
	if (verticesCount>2)
	{
	  // add corrected normals
	  newNormals.insert(newNormals.end(), correctedFaceNormals.begin(), correctedFaceNormals.end());

          // no normal at face separator
	  newNormals.push_back(Vector3(0, 0, 0));

	  facenum++;
	}

	// separator found -- clear per-face container
        correctedFaceNormals.clear();

	// reset vertices count
	verticesCount = 0;
      }
      else
      {
	// increase vertices count
	verticesCount++;
        // get the analytic normal of the face
        const Vector3* normal = getFaceNormal(facenum);
        Vector3 sumNormal = *normal;

        // combine the normal with normals of adjacent faces if their angles
        // do not differ more than the crease angle
        long componentCount = 1;

        size_t j;
        for (j=0; j<d_vertexFaces[vertexIndex].size(); j++)
        {
          if (d_vertexFaces[vertexIndex][j] == facenum) continue;
          const Vector3* compNormal = getFaceNormal(d_vertexFaces[vertexIndex][j]);
          if (fabs(acos(normal->dot(*compNormal))) < d_creaseAngle.get())
          {
            sumNormal = sumNormal + *compNormal;
            componentCount++;
          }
        }

        correctedFaceNormals.push_back(sumNormal.normalize());
      }
    }

    // use the newly computed normals
    normalPerVertex = true;
    autoNormals.reserve(newNormals.size() * 3);

    size_t j;
    for (j=0; j<newNormals.size(); j++)
    {
      autoNormals.push_back((float) newNormals[j].getX());
      autoNormals.push_back((float) newNormals[j].getY());
      autoNormals.push_back((float) newNormals[j].getZ());
    }
    normals = (autoNormals.size()>0) ? &(autoNormals[0]) : 0;
    nNormalIndices = autoNormalIndices.size();
    normalIndices = (nNormalIndices>0) ? &(autoNormalIndices[0]) : 0;

  }

  //:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
  // rendering
  //:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

  // draw bounding sphere if requested
  if (rc.getDrawBSpheres()) {
    const VrmlBSphere* bs = (VrmlBSphere*)this->getBVolume();
    viewer->drawBSphere(*bs, 4);
  }

  // render the object
  unsigned int optMask = 0;
  if (d_ccw.get())             optMask |= Viewer::MASK_CCW;
  if (d_convex.get())          optMask |= Viewer::MASK_CONVEX;
  if (d_solid.get())           optMask |= Viewer::MASK_SOLID;
  if (d_colorPerVertex.get())  optMask |= Viewer::MASK_COLOR_PER_VERTEX;
  if (normalPerVertex)         optMask |= Viewer::MASK_NORMAL_PER_VERTEX;
  obj = viewer->insertShell(optMask,
                            nVertices, vertices,
                            nCoordIndices, coordIndices,
                            texCoords, nTexCoordIndices, texCoordIndices,
                            normals, nNormalIndices, normalIndices,
                            colors, nColorIndices, colorIndices);

  // remove the modification flag from involved nodes
  if (d_color.get()) d_color.get()->clearModified();
  if (d_coord.get()) d_coord.get()->clearModified();
  if (d_normal.get()) d_normal.get()->clearModified();
  if (d_texCoord.get()) d_texCoord.get()->clearModified();

  return obj;
}

// Get the value of one of the exposedFields

const VrmlField *VrmlNodeIFaceSet::getField(const char *fieldName) const
{
  // exposedFields
  if ( strcmp( fieldName, "normal" ) == 0 )
    return &d_normal;
  else if ( strcmp( fieldName, "texCoord" ) == 0 )
    return &d_texCoord;

#if defined(__ORBISNAP__)

  else if (strcmp(fieldName, "ccw") == 0)
    return &d_ccw;
  else if (strcmp(fieldName, "convex") == 0)
    return &d_convex;
  else if (strcmp(fieldName, "creaseAngle") == 0)
    return &d_creaseAngle;
  else if (strcmp(fieldName, "normalIndex") == 0)
    return &d_normalIndex;
  else if (strcmp(fieldName, "normalPerVertex") == 0)
    return &d_normalPerVertex;
  else if (strcmp(fieldName, "solid") == 0)
    return &d_solid;
  else if (strcmp(fieldName, "texCoordIndex") == 0)
    return &d_texCoordIndex;

#endif

  return VrmlNodeIndexedSet::getField(fieldName); // Parent class
}

// Set the value of one of the node fields.

void VrmlNodeIFaceSet::setField(const char *fieldName,
				const VrmlField &fieldValue)
{
  if TRY_FIELD(ccw, SFBool)
  else if TRY_FIELD(convex, SFBool)
  else if TRY_FIELD(creaseAngle, SFFloat)
  else if TRY_SFNODE_FIELD(normal, Normal)
  else if TRY_FIELD(normalIndex, MFInt32)
  else if TRY_FIELD(normalPerVertex, SFBool)
  else if TRY_FIELD(solid, SFBool)
  else if TRY_FIELD(normalIndex, MFInt32)
  else if TRY_SFNODE_FIELD(texCoord, TextureCoordinate)
  else if TRY_FIELD(texCoordIndex, MFInt32)
  else
    VrmlNodeIndexedSet::setField(fieldName, fieldValue);

  // overly conservative: changing the creaseAngle doesn't really
  // change the bvolume, but the TRY_FIELD macro makes it hard to
  // case things out. sigh. "clarity over optimization", but it
  // still rankles...
  //
  this->setBVolumeDirty(true);
}


VrmlNode* VrmlNodeIFaceSet::getNormal()
{   return d_normal.get(); }

const VrmlMFInt32 & VrmlNodeIFaceSet::getNormalIndex() const
{   return d_normalIndex; }

VrmlNode* VrmlNodeIFaceSet::getTexCoord()
{   return d_texCoord.get(); }

const VrmlMFInt32 & VrmlNodeIFaceSet::getTexCoordIndex() const
{   return d_texCoordIndex; }

void VrmlNodeIFaceSet::recalcBSphere()
{
  //cout << "VrmlNodeIFaceSet::recalcBSphere()" << endl;

  // take the bvolume of all the points. technically, we should figure
  // out just which points are used by the index and just use those,
  // but for a first pass this is fine (also: if we do it this way
  // then we don't have to update the bvolume when the index
  // changes). motto: always do it the simple way first...
  //

  // The indexed face set does not have to have a Coordinate node...
  VrmlNodeCoordinate* coordn = NodeDynamicCast<VrmlNodeCoordinate*>(d_coord.get());
  if (coordn == NULL)
  {
    d_bsphere.reset();
    this->setBVolumeDirty(false);
    return;
  }

  VrmlMFVec3f &coord = coordn->coordinate();
  const float* p = coord.get();
  int n = coord.getLength();
  d_bsphere.reset();
  d_bsphere.enclose(p, n);
  //int nvert = coord.size();
  //for(int i=0; i<nvert; i++) {
  //float* vi = coord[i]; // vi[3]
  //d_bsphere.extend(vi);
  //}
  //d_bsphere.dump(cout);
  this->setBVolumeDirty(false);
}


const VrmlBVolume* VrmlNodeIFaceSet::getBVolume() const
{
  //cout << "VrmlNodeIFaceSet::getBVolume():" << this->isBVolumeDirty() << endl;
  if (this->isBVolumeDirty())
    ((VrmlNodeIFaceSet*)this)->recalcBSphere();
  return &d_bsphere; // hmmm, const?
}



bool VrmlNodeIFaceSet::calcFacePointers()
{
  VrmlNodeCoordinate* coordNode = NodeDynamicCast<VrmlNodeCoordinate*>(d_coord.get());
  const float* vertices  = coordNode ? coordNode->coordinate().get() : 0;
  if (!vertices) return false;
  long nVertices = coordNode->coordinate().getLength();

  long nCoordIndices = d_coordIndex.getLength();
  d_facePointers.clear();
  d_facePointers.reserve(nCoordIndices/4);
  d_faceNormals.clear();
  d_faceNormals.reserve(nCoordIndices/4);
  d_vertexFaces.clear();
  d_vertexFaces.resize(nVertices);

  for (int i=0, j=0; i<nCoordIndices;)
  {
    // hold values of vertices
    std::vector<float> faceVertices;
    // beginning of face
    int iFace = i;
    // read vertices in face
    while (i<nCoordIndices && d_coordIndex[i] != -1)
    {
      // read vertex
      for (int k = 0; k<3; k++)
        faceVertices.insert(faceVertices.end(), *(vertices + 3*(d_coordIndex[i])+k));

      // increase index
      i++;
    }

    // there must be at least three vertices
    // (otherwise skip incomplete face)
    if (faceVertices.size()>=9)
    {
      // compute face normal
      float N[3];
      computeFaceNormal(false, &(*(faceVertices.begin())), faceVertices.size()/3, N);
      Vector3 normal(N);
      // save it
      d_faceNormals.push_back(normal);
      // increase faces count
      d_facePointers.push_back(j++);

      while (iFace<nCoordIndices && d_coordIndex[iFace] != -1)
        // for each vertex, build a list of faces that reference it
        d_vertexFaces[(d_coordIndex[iFace++])].push_back(d_facePointers.size()-1);
    }
    else
      // increase index
      i++;
  }

  return true;
}



const Vector3* VrmlNodeIFaceSet::getFaceNormal(int faceNum)
{
  return (d_faceNormals.size()>faceNum) ? &(d_faceNormals[faceNum]) : 0;
}

void VrmlNodeIFaceSet::setProto(VrmlNodeProto* proto)
{
  if (d_normal.get()) d_normal.get()->setProto(proto);
  VrmlNodeIndexedSet::setProto(proto);
}

void VrmlNodeIFaceSet::setEncapsulatingPrototype(VrmlNodeType* type)
{
  if (d_normal.get()) d_normal.get()->setEncapsulatingPrototype(type);
  VrmlNodeIndexedSet::setEncapsulatingPrototype(type);
}
