/*****************************************************************************
******************************************************************************
*
*               VRML ImageTexture node
*
*               $Revision: 1.1.2.9 $
*               $Date: 2010/10/30 18:28:06 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright (C) 1998  Chris Morley
// Modified by HUMUSOFT s.r.o.  Copyright 1998-2010 HUMUSOFT s.r.o.

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#include "VrmlNodeImageTexture.h"
#include "Image.h"
#include "VrmlNodeType.h"
#include "VrmlScene.h"
#include "Doc.h"

static VrmlNode *creator( VrmlScene *s )
{ return new VrmlNodeImageTexture(s); }

const VrmlMFString& VrmlNodeImageTexture::getUrl() const
{   return d_url; }

// Define the built in VrmlNodeType:: "ImageTexture" fields

VrmlNodeType *VrmlNodeImageTexture::defineType(VrmlNodeType *t)
{
  static VrmlNodeType *st = 0;

  if (! t)
    {
      if (st) return st;
      t = st = new VrmlNodeType("ImageTexture", creator);
      t->reference();
    }

  VrmlNodeTexture::defineType(t);	// Parent class

  t->addExposedField("url", VrmlField::MFSTRING, VrmlMFString());
  t->addField("repeatS", VrmlField::SFBOOL, VrmlSFBool(true));
  t->addField("repeatT", VrmlField::SFBOOL, VrmlSFBool(true));

  return t;
}

VrmlNodeImageTexture::VrmlNodeImageTexture(VrmlScene *scene) :
  VrmlNodeTexture(scene),
  d_repeatS(true),
  d_repeatT(true),
  d_image(0),
  d_texObject(0)
{
}

VrmlNodeImageTexture::VrmlNodeImageTexture(const VrmlNodeImageTexture& other) :
  VrmlNodeTexture(other),
  d_image(0),
  d_texObject(0)
{
  *this = other;
}


VrmlNodeImageTexture& VrmlNodeImageTexture::operator=(const VrmlNodeImageTexture& other)
{
  d_url = other.d_url;
  d_repeatS = other.d_repeatS;
  d_repeatT = other.d_repeatT;

  delete d_image;
  d_image = (other.d_image) ? new Image(*(other.d_image)) : 0;

  invalidate();

  return *this;
}

VrmlNodeType & VrmlNodeImageTexture::nodeType() const
{
    return *defineType(0);
}

VrmlNodeImageTexture::~VrmlNodeImageTexture()
{
  invalidate();

  // delete the image
  delete d_image;
}


void VrmlNodeImageTexture::invalidate()
{
  // delete precompiled texObject
  if (d_texObject && d_scene)
  {
    Viewer* primary = d_scene->getPrimaryViewer();
    if (primary) primary->removeTextureObject(d_texObject);
    d_texObject = 0;
  }
  VrmlNode::invalidate();
}


VrmlNode *VrmlNodeImageTexture::cloneMe() const
{
  return new VrmlNodeImageTexture(*this);
}


ostream& VrmlNodeImageTexture::printFields(ostream& os, int indent)
{
  if (d_url.get()) PRINT_FIELD(url);
  if (! d_repeatS.get()) PRINT_FIELD(repeatS);
  if (! d_repeatT.get()) PRINT_FIELD(repeatT);
  return os;
}

void VrmlNodeImageTexture::addToScene(VrmlScene *scene, const char *relativeUrl)
{
  VrmlNodeTexture::addToScene(scene, relativeUrl);
  if (scene != NULL) load();
}

void VrmlNodeImageTexture::load()
{
  if (!d_image && d_url.getLength() > 0)
  {
    const char *relUrl = d_relativeUrl.get() ? d_relativeUrl.get() :
      d_scene->urlDoc()->url();
    Doc relDoc(relUrl);
    d_image = new Image;
    if (!d_image->tryURLs(d_url.getLength(), d_url.get(), &relDoc))
    {
      theSystem->error(System::ERRORTYPE_UNKNOWN, "Couldn't read ImageTexture from URL %s\n",
        (const char*) d_url.get(0));
    }
  }
}

void VrmlNodeImageTexture::render(Viewer *viewer, VrmlRenderContext rc)
{
  // remove image from cache
  if ((isModified() || d_scene->texturesDirty()) && d_texObject)
  {
    viewer->removeTextureObject(d_texObject);
    d_texObject = 0;
  }

  // reload image
  if (isModified())
  {
    // discard the old image both from pixmap and from the cache
    if (d_image)
    {
      delete d_image;
      d_image = 0;
    }

    // reload
    load();
  }

  // Check texture cache
  if (d_texObject && d_image)
    {
      viewer->insertTextureReference(d_texObject, d_image->nc());
    }
  else
    {
      if (d_image)
      {
	// Insert the texture into scene
        d_texObject = viewer->insertTexture(d_image->w(),
                                            d_image->h(),
					    d_image->nc(),
					    d_repeatS.get(),
					    d_repeatT.get(),
					    d_image->pixels(),
					    true);
      }
    }
  clearModified();
}


size_t VrmlNodeImageTexture::nComponents()
{
  return d_image ? d_image->nc() : 0;
}

size_t VrmlNodeImageTexture::width()
{
  return d_image ? d_image->w() : 0;
}

size_t VrmlNodeImageTexture::height()
{
  return d_image ? d_image->h() : 0;
}

size_t VrmlNodeImageTexture::nFrames()
{
  return 0;
}

const unsigned char * VrmlNodeImageTexture::pixels()
{
  return d_image ? d_image->pixels() : 0;
}


// Get the value of a field or eventOut.

const VrmlField *VrmlNodeImageTexture::getField(const char *fieldName) const
{
  // exposedFields
  if ( strcmp( fieldName, "url" ) == 0 )
    return &d_url;

#if defined(__ORBISNAP__)

  else if (strcmp(fieldName, "repeatS") == 0)
    return &d_repeatS;
  else if (strcmp(fieldName, "repeatT") == 0)
    return &d_repeatT;

#endif

  return VrmlNode::getField(fieldName); // Parent class
}

// Set the value of one of the node fields.

void VrmlNodeImageTexture::setField(const char *fieldName,
				    const VrmlField &fieldValue)
{
  if TRY_FIELD(url, MFString)
  else if TRY_FIELD(repeatS, SFBool)
  else if TRY_FIELD(repeatT, SFBool)
  else
    VrmlNode::setField(fieldName, fieldValue);
}

bool VrmlNodeImageTexture::getRepeatS() const
{
    return d_repeatS.get();
}

bool VrmlNodeImageTexture::getRepeatT() const
{
    return d_repeatT.get();
}
