/*****************************************************************************
******************************************************************************
*
*               VRML indexed set nodes base class
*
*               $Revision: 1.1.2.4 $
*               $Date: 2011/07/19 14:53:20 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright (C) 1998  Chris Morley
// Modified by HUMUSOFT s.r.o.  Copyright 1998-2011 HUMUSOFT s.r.o.

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#include "VrmlNodeIndexedSet.h"
#include "VrmlNodeType.h"
#include "VrmlNodeColor.h"
#include "VrmlNodeCoordinate.h"
#include "VrmlNodeProto.h"


// Define the built in VrmlNodeType:: "Indexed*Set" fields

VrmlNodeType *VrmlNodeIndexedSet::defineType(VrmlNodeType *t)
{
  VrmlNodeGeometry::defineType(t);	// Parent class

  t->addEventIn("set_colorIndex", VrmlField::MFINT32);
  t->addEventIn("set_coordIndex", VrmlField::MFINT32);
  t->addExposedField("color", VrmlField::SFNODE, VrmlSFNode());
  t->addExposedField("coord", VrmlField::SFNODE, VrmlSFNode());
  t->addField("colorIndex", VrmlField::MFINT32, VrmlMFInt32());
  t->addField("colorPerVertex", VrmlField::SFBOOL, VrmlSFBool(true));
  t->addField("coordIndex", VrmlField::MFINT32, VrmlMFInt32());

  return t;
}

VrmlNodeIndexedSet::VrmlNodeIndexedSet(VrmlScene *scene) :
  VrmlNodeGeometry(scene),
  d_colorPerVertex(true)
{
}

VrmlNodeIndexedSet::~VrmlNodeIndexedSet()
{
}


bool VrmlNodeIndexedSet::isModified() const
{
  return ( VrmlNodeGeometry::isModified() ||
	   (d_color.get() && d_color.get()->isModified()) ||
	   (d_coord.get() && d_coord.get()->isModified()) );
}

void VrmlNodeIndexedSet::updateModified(VrmlNodePath& path)
{
  if (this->isModified()) markPathModified(path, true);
  path.push_front(this);
  if (d_color.get()) d_color.get()->updateModified(path);
  if (d_coord.get()) d_coord.get()->updateModified(path);
  path.pop_front();
}

void VrmlNodeIndexedSet::clearFlags()
{
  VrmlNode::clearFlags();
  if (d_color.get()) d_color.get()->clearFlags();
  if (d_coord.get()) d_coord.get()->clearFlags();
}

void VrmlNodeIndexedSet::addToScene( VrmlScene *s, const char *rel )
{
  d_scene = s;
  if (d_color.get()) d_color.get()->addToScene(s, rel);
  if (d_coord.get()) d_coord.get()->addToScene(s, rel);
}

ostream& VrmlNodeIndexedSet::printFields(ostream& os, int indent)
{
  if (d_color.get()) PRINT_FIELD(color);
  if (d_colorIndex.getLength() > 0) PRINT_FIELD(colorIndex);
  if (! d_colorPerVertex.get()) PRINT_FIELD(colorPerVertex);
  if (d_coord.get()) PRINT_FIELD(coord);
  if (d_coordIndex.getLength() > 0) PRINT_FIELD(coordIndex);
  return os;
}

VrmlNodeColor* VrmlNodeIndexedSet::color()
{
  return NodeDynamicCast<VrmlNodeColor*>(d_color.get());
}


VrmlNode* VrmlNodeIndexedSet::getCoordinate()
{   return d_coord.get(); }

const VrmlMFInt32 & VrmlNodeIndexedSet::getCoordIndex() const
{   return d_coordIndex; }

const VrmlMFInt32 & VrmlNodeIndexedSet::getColorIndex() const   // LarryD Feb 18/99
{   return d_colorIndex; }

// Get the value of one of the exposedFields

const VrmlField *VrmlNodeIndexedSet::getField(const char *fieldName) const
{
  // exposedFields
  if ( strcmp( fieldName, "color" ) == 0 )
    return &d_color;
  else if ( strcmp( fieldName, "coord" ) == 0 )
    return &d_coord;

#if defined(__ORBISNAP__)
  else if (strcmp(fieldName, "colorIndex") == 0)
    return &d_colorIndex;
  else if (strcmp(fieldName, "coordIndex") == 0)
    return &d_coordIndex;
  else if (strcmp(fieldName, "colorPerVertex") == 0)
    return &d_colorPerVertex;
#endif

  return VrmlNodeGeometry::getField(fieldName); // Parent class
}

// Set the value of one of the node fields.

void VrmlNodeIndexedSet::setField(const char *fieldName,
				  const VrmlField &fieldValue)
{
  if TRY_SFNODE_FIELD(color, Color)
  else if TRY_FIELD(colorIndex, MFInt32)
  else if TRY_FIELD(colorPerVertex, SFBool)
  else if TRY_SFNODE_FIELD(coord, Coordinate)
  else if TRY_FIELD(coordIndex, MFInt32)
  else
    VrmlNodeGeometry::setField(fieldName, fieldValue);
}


void VrmlNodeIndexedSet::checkIndexSetSemantics( const char* nodename,
                                                 const long*& coordIndices,
                                                 long& nCoordIndices,
                                                 const long*& colorIndices,
                                                 long& nColorIndices,
                                                 long& nVertices,
                                                 const float*& colors,
                                                 long& nColors,
                                                 long& nShapes,
                                                 long& maxCoordIndex)
{
  //:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
  // check colors and color indices
  //:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

#define DISCARD_COLORS \
  { nColorIndices = 0; nColors = 0; colorIndices = 0; colors = 0; }

  if (nColorIndices > 0 && nColors == 0)
  {
    // common and harmless case (possibly a temporary state), do not complain
    DISCARD_COLORS;
  }
  long nColorIndicesRequired = d_colorPerVertex.get() ? nCoordIndices : nShapes;
  if (nColorIndices != 0 && nColorIndices < nColorIndicesRequired)
  {
    theSystem->warn(System::SEVERITY_LOW, System::WARNINGTYPE_UNKNOWN,
      "Not enough values for %s.colorIndex: %d (%d is required)",
      nodename, nColorIndices, nColorIndicesRequired);
    DISCARD_COLORS;
  }
  else if (nColorIndices == 0 && nColors != 0 && (d_colorPerVertex.get() ?
           (nColors <= maxCoordIndex) : (nColors < nShapes)))
  {
    theSystem->warn(System::SEVERITY_LOW, System::WARNINGTYPE_UNKNOWN,
      "Not enough values for %s.color.color: %d (%d is required)",
      nodename, nColors,
      d_colorPerVertex.get() ? (maxCoordIndex + 1) : nShapes);
    DISCARD_COLORS;
  }
  for (long i=0; i<nColorIndices; i++)
  {
    if (colorIndices[i] >= nColors || colorIndices[i] < -1 ||
        (!d_colorPerVertex.get() && colorIndices[i] < 0))
    {
      theSystem->warn(System::SEVERITY_LOW, System::WARNINGTYPE_UNKNOWN,
        "Invalid value for %s.colorIndex(%d): %d (max is %d)",
        nodename, i, colorIndices[i], nColors - 1);
      DISCARD_COLORS;
      break;
    }
    if (colorIndices[i] == -1 && (!d_colorPerVertex.get()
        || coordIndices[i] != -1))
    {
      theSystem->warn(System::SEVERITY_LOW, System::WARNINGTYPE_UNKNOWN,
        "Invalid value for %s.colorIndex(%d): -1",
        nodename, i);
      DISCARD_COLORS;
      break;
    }
  }
}

void VrmlNodeIndexedSet::setProto(VrmlNodeProto* proto)
{
  if (d_color.get()) d_color.get()->setProto(proto);
  if (d_coord.get()) d_coord.get()->setProto(proto);
  VrmlNodeGeometry::setProto(proto);
}

void VrmlNodeIndexedSet::setEncapsulatingPrototype(VrmlNodeType* type)
{
  if (d_color.get()) d_color.get()->setEncapsulatingPrototype(type);
  if (d_coord.get()) d_coord.get()->setEncapsulatingPrototype(type);
  VrmlNodeGeometry::setEncapsulatingPrototype(type);
}
