/*****************************************************************************
******************************************************************************
*
*               VRML Inline node
*
*               $Revision: 1.1.2.18 $
*               $Date: 2011/09/20 10:08:52 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright (C) 1998  Chris Morley
// Modified by HUMUSOFT s.r.o.  Copyright 1998-2011 HUMUSOFT s.r.o.

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include "VrmlNodeInline.h"
#include "VrmlException.h"
#include "VrmlNamespace.h"
#include "VrmlNodeType.h"
#include "Doc.h"
#include "MathUtils.h"
#include "VrmlScene.h"

static VrmlNode *creator( VrmlScene *scene )
{
  return new VrmlNodeInline(scene);
}


// Define the built in VrmlNodeType:: "Inline" fields

VrmlNodeType *VrmlNodeInline::defineType(VrmlNodeType *t)
{
  static VrmlNodeType *st = 0;
  if (! t)
    {
      if (st) return st;
      t = st = new VrmlNodeType("Inline", creator);
      t->reference();
    }

  VrmlNodeChild::defineType(t);	// Parent of class
  t->addExposedField("url", VrmlField::MFSTRING, VrmlMFString());
  t->addField("bboxCenter", VrmlField::SFVEC3F, VrmlSFVec3f(0, 0, 0));
  t->addField("bboxSize", VrmlField::SFVEC3F, VrmlSFVec3f(-1, -1, -1));

  return t;
}

VrmlNodeType & VrmlNodeInline::nodeType() const
{
    return *defineType(0);
}


VrmlNodeInline::VrmlNodeInline(VrmlScene *scene) :
  VrmlNodeGroup(scene),
  d_scope(0)
{
  this->setBVolumeDirty(true);
}

VrmlNodeInline::VrmlNodeInline(const VrmlNodeInline& other) :
  VrmlNodeGroup(other),
  d_scope(0)
{
  *this = other;
}

VrmlNodeInline::~VrmlNodeInline()
{
  // remove children nodes (we have loaded them so we have to remove them)
  d_children.setLength(0);

  // remove the namespace (this must be done after the nodes were cleared)
  delete d_scope;
}

VrmlNodeInline& VrmlNodeInline::operator=(const VrmlNodeInline& other)
{
  d_url = other.d_url;

  delete d_scope;
  d_scope = 0;

  return *this;
}


VrmlNode *VrmlNodeInline::cloneMe() const
{
  return new VrmlNodeInline(*this);
}

void VrmlNodeInline::cloneChildren(VrmlNamespace* ns)
{
  /* do not clone Inline children, load them from URL instead */
  removeChildren();
}


// Inlines are loaded during addToScene traversal

void VrmlNodeInline::addToScene(VrmlScene *s, const char *relativeUrl)
{
  d_scene = s;

  if (s || relativeUrl) load(relativeUrl);

  // add descendants to scene (if there are some)
  if (getNodes()->getLength()>0 && d_relative.get())
    VrmlNodeGroup::addToScene(s, d_relative.get());
}



// We implement our own version of print() because the Inline node
// is implemented as a child of a Group node and so it has the "children"
// field which must not be printed because it is not part of the VRML specs.
// So our print() does the same as VrmlNode::print() except that it hides
// this field.

ostream& VrmlNodeInline::print(ostream& os, int indent, bool breakLine) const
{
  int i;

  // if this node was already printed, issue just the USE reference
  if (isFlagSet() && validName())
  {
    if (breakLine)
    {
      os << "\n";
      for (int i=0; i<indent; ++i) os << ' ';
    }
    os << "USE " << name();
    return os;
  }

  printComment(os, indent, breakLine);

  // node name (if node has valid name)
  const char *nm = validName();
  if (nm) os << "DEF " << nm << " ";

  // opening brace
  os << "Inline {";

  // print fields
  printField(os, "bboxCenter", indent);
  printField(os, "bboxSize", indent);
  printField(os, "url", indent);

  // ending brace
  os << "\n";
  for (i=0; i<indent; ++i) os << ' ';  // indent
  os << "}";

  // this node was already printed
  ((VrmlNode*) this)->setFlag();

  return os;

}


ostream& VrmlNodeInline::printFields(ostream& os, int indent)
{
  if ( !FPZERO(d_bboxCenter.getX()) ||
       !FPZERO(d_bboxCenter.getY()) ||
       !FPZERO(d_bboxCenter.getZ()) )
    PRINT_FIELD(bboxCenter);

  if ( !FPEQUAL(d_bboxSize.getX(), -1) ||
       !FPEQUAL(d_bboxSize.getY(), -1) ||
       !FPEQUAL(d_bboxSize.getZ(), -1) )
    PRINT_FIELD(bboxSize);

  if (d_url.get()) PRINT_FIELD(url);

  return os;
}

VrmlField* VrmlNodeInline::findField(const char* fieldname)
{
  if (strcmp(fieldname, "url") == 0)
    return &d_url;
  else if (strcmp(fieldname, "children") == 0 ||
           strcmp(fieldname, "addChildren") == 0 ||
           strcmp(fieldname, "removeChildren") == 0)
    return 0;
  else
    return VrmlNodeGroup::findField(fieldname);
}

// Set the value of one of the exposed fields.

const VrmlField *VrmlNodeInline::getField(const char *fieldName) const
{
  if ( strcmp( fieldName, "url" ) == 0 )
    return &d_url;

  return VrmlNodeGroup::getField(fieldName);
}

// Set the value of one of the node fields.

void VrmlNodeInline::setField(const char *fieldName,
			      const VrmlField &fieldValue)
{
  if TRY_FIELD(url, MFString)
  else
    VrmlNodeGroup::setField(fieldName, fieldValue);
}

void VrmlNodeInline::setModified()
{
  // do nothing if the modified flag is already set
  if (isModified()) return;

  // set the modified flag
  VrmlNode::setModified();

#if defined(__ORBISNAP__)

  // reload
  if (d_relative.get())
    load(d_relative.get());


#endif
}

//  Load the children from the URL

void VrmlNodeInline::load(const char *relativeUrl)
{
  if (d_url.getLength() > 0)
  {
    this->setBVolumeDirty(true);

    VrmlNamespace *ns = new VrmlNamespace();
    VrmlMFNode *kids = 0;
    Doc url;
    int i, n = d_url.getLength();
    for (i=0; i<n; ++i)
    {
      Doc relDoc(relativeUrl);
      theSystem->debug("Trying to read url '%s' (relative %s)\n",
        d_url.get(i), d_relative.get() ? d_relative.get() : "<null>");

      url.seturl( d_url.get(i), &relDoc );

      try
      {
        kids = VrmlScene::readWrl( &url, ns );
        // we found first valid entry in url array
        // escape the loop
        break;
      }
      catch (VrmlLoadingError &error)
      {
        theSystem->warn(System::SEVERITY_HIGH, System::WARNINGTYPE_UNKNOWN, "Could not load Inline URL: %s", error.what());
      }
    }

    if ( kids )
    {
      removeChildren();
      delete d_scope;

      d_scope = ns;
      d_relative.set(url.url()); // children will be relative to this url
      addChildren( *kids );      // check for nested Inlines
      delete kids;
    }
    else
    {
      d_relative.set(relativeUrl);
      delete ns;
    }
  }
  else
    d_relative.set(relativeUrl);
}

