/*****************************************************************************
******************************************************************************
*
*               VRML LOD node
*
*               $Revision: 1.1.2.3 $
*               $Date: 2011/07/19 14:53:20 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright (C) 1998  Chris Morley
// Modified by HUMUSOFT s.r.o.  Copyright 1998-2011 HUMUSOFT s.r.o.

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#include "VrmlNodeLOD.h"
#include "VrmlNodeType.h"
#include "MathUtils.h"
#include "Viewer.h"

// Return a new VrmlNodeLOD
static VrmlNode *creator( VrmlScene *s ) { return new VrmlNodeLOD(s); }


// Define the built in VrmlNodeType:: "LOD" fields

VrmlNodeType *VrmlNodeLOD::defineType(VrmlNodeType *t)
{
  static VrmlNodeType *st = 0;

  if (! t)
    {
      if (st) return st;
      t = st = new VrmlNodeType("LOD", creator);
      t->reference();
    }

  VrmlNodeChild::defineType(t); // Parent class
  t->addExposedField("level", VrmlField::MFNODE, VrmlMFNode());
  t->addField("center", VrmlField::SFVEC3F, VrmlSFVec3f(0, 0, 0));
  t->addField("range", VrmlField::MFFLOAT, VrmlMFFloat());

  return t;
}


VrmlNodeType & VrmlNodeLOD::nodeType() const
{
    return *defineType(0);
}


VrmlNodeLOD::VrmlNodeLOD(VrmlScene *scene) :
  VrmlNodeChild(scene),
  d_center(0.0, 0.0, 0.0),
  d_parentTransform(0)
{
  this->setBVolumeDirty(true); // lazy calc of bvolume
}

VrmlNodeLOD::~VrmlNodeLOD()
{
}


VrmlNode *VrmlNodeLOD::cloneMe() const
{
  return new VrmlNodeLOD(*this);
}

void VrmlNodeLOD::cloneChildren(VrmlNamespace * ns) {
    for (size_t i = 0; i < this->d_level.getLength(); ++i) {
        if (! this->d_level[i]) {
            continue;
        }
        VrmlNode * const newKid = this->d_level[i]->clone(ns)->reference();
        this->d_level[i]->dereference();
        this->d_level[i] = newKid;
    }
}

bool VrmlNodeLOD::isModified() const
{
  if (VrmlNodeChild::isModified()) return true;

  int n = d_level.getLength();

  // This should really check which range is being rendered...
  for (int i = 0; i<n; ++i)
    if (d_level[i]->isModified())
      return true;

  return false;
}

// what happens if one of the other children suddenly becomes the one
// selected? to be safe: check them all. this potentially means some
// extra work, but it's a lot easier to reason about.
//
void VrmlNodeLOD::updateModified(VrmlNodePath& path)
{
  if (this->isModified()) markPathModified(path, true);
  path.push_front(this);
 int n = d_level.getLength();
  for (int i = 0; i<n; ++i)
    d_level[i]->updateModified(path);
  path.pop_front();
}

void VrmlNodeLOD::clearFlags()
{
  VrmlNode::clearFlags();
  int n = d_level.getLength();
  for (int i = 0; i<n; ++i)
    d_level[i]->clearFlags();
}

void VrmlNodeLOD::addToScene( VrmlScene *s, const char *rel )
{
  d_scene = s;

  int n = d_level.getLength();

  for (int i = 0; i<n; ++i)
    d_level[i]->addToScene(s, rel);
}

VrmlNode* VrmlNodeLOD::getParentTransform() { return d_parentTransform; }

void VrmlNodeLOD::accumulateTransform(VrmlNode *parent)
{
  d_parentTransform = parent;

  int n = d_level.getLength();
    for (int i = 0; i<n; ++i)
    {
      VrmlNode *kid = d_level[i];
      kid->accumulateTransform(parent);
    }
}

ostream& VrmlNodeLOD::printFields(ostream& os, int indent)
{
  if (d_level.getLength() > 0) PRINT_FIELD(level);
  if (! FPZERO(d_center.getX()) ||
      ! FPZERO(d_center.getY()) ||
      ! FPZERO(d_center.getZ()) )
    PRINT_FIELD(center);

  if (d_range.getLength() > 0) PRINT_FIELD(range);

  return os;
}

int VrmlNodeLOD::getChosenLevel(Viewer* viewer)
{
  if (d_level.getLength() <= 0) return -1;

  // get inverse transformation matrix for this node
  Matrix4x4 im = getCompoundInverseTransformationMatrix();

  // get the current camera position in world coordinates
  Camera cam = viewer->getCameraAbsolute();
  Point3 pos = cam.getPosition();

 // calculate current camera position in LOD local coordinates
  Point3 localpos = im*pos;

  double dx = localpos.getX() - d_center.getX();
  double dy = localpos.getY() - d_center.getY();
  double dz = localpos.getZ() - d_center.getZ();
  float d2 = (float) dx*dx + dy*dy + dz*dz;

  int i, n = d_range.getLength();
  for (i=0; i<n; ++i)
    if (d2 < d_range[i] * d_range[i])
      break;

  // Should choose an "optimal" level...
  if (d_range.getLength() == 0) i = d_level.getLength() - 1;

  // Not enough levels...
  if (i >= (int) d_level.getLength()) i = d_level.getLength() - 1;

  return i;


}

// Render one of the children

void VrmlNodeLOD::render(Viewer *viewer, VrmlRenderContext rc)
{
  clearModified();

  int i = getChosenLevel(viewer);
  if (i == -1) return;

  d_level[i]->render(viewer, rc);

  // Don't re-render on their accounts
  int n = d_level.getLength();
  for (i = 0; i<n; ++i)
    d_level[i]->clearModified();
}

void VrmlNodeLOD::invalidate()
{
  int n = d_level.getLength();
  for (int i = 0; i<n; ++i)
    d_level[i]->invalidate();

  VrmlNode::invalidate();
}

void VrmlNodeLOD::guessNearFarPlane(Viewer* viewer, double* nearVal, double* farVal)
{
  int i = getChosenLevel(viewer);
  if (i == -1) return;

  (d_level[i])->guessNearFarPlane(viewer, nearVal, farVal);
}

// Get the value of one of the node fields.
const VrmlField *VrmlNodeLOD::getField(const char *fieldName) const
{
  if ( strcmp( fieldName, "level" ) == 0 )
    return &d_level;

#if defined(__ORBISNAP__)

  else if (strcmp(fieldName, "center") == 0)
    return &d_center;
  else if (strcmp(fieldName, "range") == 0)
    return &d_range;

#endif

  return VrmlNodeChild::getField( fieldName );
}

// Set the value of one of the node fields.
void VrmlNodeLOD::setField(const char *fieldName,
                           const VrmlField &fieldValue)
{
  if TRY_FIELD(level, MFNode)
  else if TRY_FIELD(center, SFVec3f)
  else if TRY_FIELD(range, MFFloat)
  else
    VrmlNodeChild::setField(fieldName, fieldValue);
  setBVolumeDirty(true); // lazy calc of bvolume
}

const VrmlMFFloat& VrmlNodeLOD::getRange() const
{   return d_range; }

const VrmlSFVec3f& VrmlNodeLOD::getCenter() const
{  return d_center; }


const VrmlBVolume* VrmlNodeLOD::getBVolume() const
{
  //cout << "VrmlNodeLOD[" << this << "]::getBVolume()" << endl;
  if (this->isBVolumeDirty())
    ((VrmlNodeLOD*)this)->recalcBSphere();
  return &d_bsphere;
}


void
VrmlNodeLOD::recalcBSphere()
{
  //cout << "VrmlNodeLOD[" << this << "]::recalcBSphere()" << endl;
  d_bsphere.reset();

  // let's say our bsphere is the union of the bspheres of all the
  // levels. we could have said it was just the bsphere of the current
  // level, but the current level depends on the viewer position, and
  // we'd like to make the calculation idependent of that. we could do
  // some sort of trick where we reset the bsphere during render, but
  // that seems like overkill unless this simpler method proves to be
  // a bottleneck.
  //
  // hmm: just thought of a problem: one of the uses of the lod is to
  // switch in delayed-load inlines. this would necessarily switch
  // them in all at once. live with it for now.
  //
  for(int i=0; i<(int) d_level.getLength(); i++) {
    const VrmlBVolume* ci_bv = d_level[i]->getBVolume();
    d_bsphere.extend(*ci_bv);

  }
  this->setBVolumeDirty(false);
}

void VrmlNodeLOD::setProto(VrmlNodeProto* proto)
{
  int i;
  for (i=0; i<d_level.getLength(); i++)
  {
    VrmlNode* ci = d_level[i];
    ci->setProto(proto);
  }
  VrmlNodeChild::setProto(proto);
}

void VrmlNodeLOD::setEncapsulatingPrototype(VrmlNodeType* type)
{
  for (int i=0; i<d_level.getLength(); i++)
  {
    VrmlNode* ci = d_level[i];
    ci->setEncapsulatingPrototype(type);
  }
  VrmlNodeChild::setEncapsulatingPrototype(type);
}
