/*****************************************************************************
******************************************************************************
*
*               VRML MovieTexture node
*
*               $Revision: 1.1.2.6 $
*               $Date: 2010/10/30 18:28:06 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright (C) 1998  Chris Morley
// Modified by HUMUSOFT s.r.o.  Copyright 1998-2010 HUMUSOFT s.r.o.

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#include "VrmlNodeMovieTexture.h"
#include "Image.h"
#include "MathUtils.h"
#include "System.h"
#include "VrmlNodeType.h"
#include "VrmlScene.h"
#include "Viewer.h"
#include "Doc.h"

static VrmlNode *creator( VrmlScene *s )
{ return new VrmlNodeMovieTexture(s); }


// Define the built in VrmlNodeType:: "MovieTexture" fields

VrmlNodeType *VrmlNodeMovieTexture::defineType(VrmlNodeType *t)
{
  static VrmlNodeType *st = 0;

  if (! t)
    {
      if (st) return st;
      t = st = new VrmlNodeType("MovieTexture", creator);
      t->reference();
    }

  VrmlNodeTexture::defineType(t);	// Parent class

  t->addExposedField("loop", VrmlField::SFBOOL, VrmlSFBool(true));
  t->addExposedField("speed", VrmlField::SFFLOAT, VrmlSFFloat(1.0));
  t->addExposedField("startTime", VrmlField::SFTIME, VrmlSFTime(0));
  t->addExposedField("stopTime", VrmlField::SFTIME, VrmlSFTime(0));
  t->addExposedField("url", VrmlField::MFSTRING, VrmlMFString());
  t->addField("repeatS", VrmlField::SFBOOL, VrmlSFBool(true));
  t->addField("repeatT", VrmlField::SFBOOL, VrmlSFBool(true));
  t->addEventOut("duration_changed", VrmlField::SFTIME);
  t->addEventOut("isActive", VrmlField::SFBOOL);

  return t;
}


VrmlNodeType & VrmlNodeMovieTexture::nodeType() const
{
    return *defineType(0);
}


VrmlNodeMovieTexture::VrmlNodeMovieTexture(VrmlScene *scene) :
  VrmlNodeTexture(scene),
  d_loop(false),
  d_speed(1.0),
  d_repeatS(true),
  d_repeatT(true),
  d_image(0),
  d_frame(0),
  d_lastFrame(-1),
  d_lastFrameTime(-1.0),
  d_texObject(0)
{
  if (d_scene) d_scene->addMovie( this );
}

VrmlNodeMovieTexture::VrmlNodeMovieTexture(const VrmlNodeMovieTexture& other) :
  VrmlNodeTexture(other),
  d_image(0),
  d_texObject(0)
{
  *this = other;
}

VrmlNodeMovieTexture::~VrmlNodeMovieTexture()
{
  invalidate();

  // delete the movie
  if (d_scene) d_scene->removeMovie(this);
  delete d_image;
}


void VrmlNodeMovieTexture::invalidate()
{
  // delete precompiled texObject
  if (d_texObject && d_scene)
  {
    Viewer* primary = d_scene->getPrimaryViewer();
    if (primary) primary->removeTextureObject(d_texObject);
    d_texObject = 0;
  }
  VrmlNode::invalidate();
}

VrmlNodeMovieTexture& VrmlNodeMovieTexture::operator=(const VrmlNodeMovieTexture& other)
{
  d_loop = other.d_loop;
  d_speed = other.d_speed;
  d_startTime = other.d_startTime;
  d_stopTime = other.d_stopTime;

  d_url = other.d_url;
  d_repeatS = other.d_repeatS;
  d_repeatT = other.d_repeatT;

  d_duration = other.d_duration;
  d_isActive = other.d_isActive;

  delete d_image;
  d_image = (other.d_image) ? new Image(*(other.d_image)) : 0;

  d_frame = other.d_frame;
  d_lastFrame = other.d_lastFrame;
  d_lastFrameTime = other.d_lastFrameTime;

  invalidate();

  return *this;
}

VrmlNode *VrmlNodeMovieTexture::cloneMe() const
{
  return new VrmlNodeMovieTexture(*this);
}


void VrmlNodeMovieTexture::addToScene(VrmlScene *s, const char *rel)
{
  if (d_scene != s && (d_scene = s) != 0) d_scene->addMovie(this);
  VrmlNodeTexture::addToScene(s, rel);
}


ostream& VrmlNodeMovieTexture::printFields(ostream& os, int indent)
{
  if (d_loop.get()) PRINT_FIELD(loop);
  if (!FPEQUAL(d_speed.get(), 1.0)) PRINT_FIELD(speed);
  if (!FPZERO(d_startTime.get())) PRINT_FIELD(startTime);
  if (!FPZERO(d_stopTime.get())) PRINT_FIELD(stopTime);
  if (d_url.get()) PRINT_FIELD(url);
  if (! d_repeatS.get()) PRINT_FIELD(repeatS);
  if (! d_repeatT.get()) PRINT_FIELD(repeatT);
  return os;
}


void VrmlNodeMovieTexture::update( VrmlSFTime &timeNow )
{
  if ( isModified() )
    {
      if (d_image)
	{
	  const char *imageUrl = d_image->url();
	  int imageLen = strlen(imageUrl);
	  int i, nUrls = d_url.getLength();
	  for (i=0; i<nUrls; ++i)
	    {
	      int len = strlen(d_url.get(i));

	      if ((strcmp(imageUrl, d_url.get(i)) == 0) ||
		  (imageLen > len &&
		   strcmp(imageUrl+imageLen-len, d_url.get(i)) == 0))
		break;
	    }

	  // if (d_image->url() not in d_url list) ...
	  if (i == nUrls)
	    {
	      delete d_image;
	      d_image = 0;
	    }
	}
    }

  // Load the movie if needed (should check startTime...)
  if (! d_image && d_url.getLength() > 0)
    {
      Doc relDoc(d_relativeUrl.get());
      Doc * rel = d_relativeUrl.get() ? &relDoc : d_scene->urlDoc();
      d_image = new Image;
      if ( ! d_image->tryURLs( d_url.getLength(), d_url.get(), rel ) )
        theSystem->error(System::ERRORTYPE_UNKNOWN, "Couldn't read MovieTexture from URL %s\n", d_url.get(0));

      int nFrames = d_image->nFrames();
      d_duration = (nFrames >= 0) ? nFrames : -1;
      eventOut( timeNow.get(), "duration_changed", d_duration );
      d_frame = (d_speed.get() >= 0) ? 0 : nFrames-1;

      //theSystem->debug("MovieTexture.%s loaded %d frames\n", name(), nFrames);
    }

  // No pictures to show
  if (! d_image || d_image->nFrames() == 0) return;

  // Become active at the first tick at or after startTime if either
  // the valid stopTime hasn't passed or we are looping.
  if (! d_isActive.get() &&
      d_startTime.get() <= timeNow.get() &&
      d_startTime.get() >= d_lastFrameTime &&
      ( (d_stopTime.get() < d_startTime.get() || // valid stopTime
	 d_stopTime.get() > timeNow.get()) ||    // hasn't passed
	d_loop.get() ))
    {
      //theSystem->debug("MovieTexture.%s::isActive TRUE\n", name());
      d_isActive.set(true);
      eventOut( timeNow.get(), "isActive", d_isActive );
      d_lastFrameTime = timeNow.get();
      d_frame = (d_speed.get() >= 0) ? 0 : d_image->nFrames() - 1;
      setModified();
    }

  // Check whether stopTime has passed
  else if ( d_isActive.get() &&
	    (( d_stopTime.get() > d_startTime.get() &&
	       d_stopTime.get() <= timeNow.get() ) ||
	     d_frame < 0))
    {
      //theSystem->debug("MovieTexture.%s::isActive FALSE\n", name());
      d_isActive.set(false);
      eventOut( timeNow.get(), "isActive", d_isActive );
      setModified();
    }

  // Check whether the frame should be advanced
  else if ( d_isActive.get() &&
	    d_lastFrameTime + fabs(1/d_speed.get()) <= timeNow.get() )
    {
      if (d_speed.get() < 0.0)
	--d_frame;
      else
	++d_frame;
      //theSystem->debug("MovieTexture.%s::frame %d\n", name(), d_frame);
      d_lastFrameTime = timeNow.get();
      setModified();
    }

}


// Ignore set_speed when active.

void VrmlNodeMovieTexture::eventIn(double timeStamp,
				   const char *eventName,
				   const VrmlField *fieldValue)
{
  const char *origEventName = eventName;
  if ( strncmp(eventName, "set_", 4) == 0 )
    eventName += 4;

  // Ignore set_speed when active
  if ( strcmp(eventName,"speed") == 0 )
    {
      if (! d_isActive.get())
	{
	  setField(eventName, *fieldValue);
	  eventOut(timeStamp, "speed_changed", *fieldValue);
	  setModified();
	}
    }

  // Let the generic code handle the rest.
  else
    VrmlNode::eventIn( timeStamp, origEventName, fieldValue );
}


// Render a frame if there is one available.

void VrmlNodeMovieTexture::render(Viewer *viewer, VrmlRenderContext rc)
{
  //theSystem->debug("MovieTexture.%s::render frame %d\n", name(), d_frame);

  if ( ! d_image || d_frame < 0 ) return;

  unsigned char *pix = d_image->pixels( d_frame );

  if ( d_frame != d_lastFrame && d_texObject )
    {
      viewer->removeTextureObject( d_texObject );
      d_texObject = 0;
    }

  if ( ! pix )
    {
      d_frame = -1;
    }
  else if ( d_texObject )
    {
      viewer->insertTextureReference( d_texObject, d_image->nc() );
    }
  else
    {
      // Insert the texture into scene
      d_texObject = viewer->insertTexture(d_image->w(),
					  d_image->h(),
					  d_image->nc(),
					  d_repeatS.get(),
					  d_repeatT.get(),
					  pix,
					  ! d_isActive.get() );
    }

  d_lastFrame = d_frame;
  clearModified();
}


size_t VrmlNodeMovieTexture::nComponents()
{
  return d_image ? d_image->nc() : 0;
}

size_t VrmlNodeMovieTexture::width()
{
  return d_image ? d_image->w() : 0;
}

size_t VrmlNodeMovieTexture::height()
{
  return d_image ? d_image->h() : 0;
}

size_t VrmlNodeMovieTexture::nFrames()
{
  return d_image ? d_image->nFrames() : 0;
}

const unsigned char * VrmlNodeMovieTexture::pixels()
{
  return d_image ? d_image->pixels() : 0;
}



// Get the value of a field or eventOut.

const VrmlField *VrmlNodeMovieTexture::getField(const char *fieldName) const
{
  // exposedFields
  if ( strcmp( fieldName, "loop" ) == 0 )
    return &d_loop;
  else if ( strcmp( fieldName, "speed" ) == 0 )
    return &d_speed;
  else if ( strcmp( fieldName, "startTime" ) == 0 )
    return &d_startTime;
  else if ( strcmp( fieldName, "stopTime" ) == 0 )
    return &d_stopTime;
  else if ( strcmp( fieldName, "url" ) == 0 )
    return &d_url;

  // eventOuts
  else if ( strcmp( fieldName, "duration" ) == 0 )
    return &d_duration;
  else if ( strcmp( fieldName, "isActive" ) == 0 )
    return &d_isActive;

#if defined(__ORBISNAP__)

  else if (strcmp(fieldName, "repeatS") == 0)
    return &d_repeatS;
  else if (strcmp(fieldName, "repeatT") == 0)
    return &d_repeatT;

#endif

  return VrmlNode::getField(fieldName); // Parent class
}

// Set the value of one of the node fields.

void VrmlNodeMovieTexture::setField(const char *fieldName,
				    const VrmlField &fieldValue)
{
  if TRY_FIELD(loop, SFBool)
  else if TRY_FIELD(speed, SFFloat)
  else if TRY_FIELD(startTime, SFTime)
  else if TRY_FIELD(stopTime, SFTime)
  else if TRY_FIELD(url, MFString)
  else if TRY_FIELD(repeatS, SFBool)
  else if TRY_FIELD(repeatT, SFBool)
  else
    VrmlNode::setField(fieldName, fieldValue);
}

