/*****************************************************************************
******************************************************************************
*
*               VRML PixelTexture node
*
*               $Revision: 1.1.2.5 $
*               $Date: 2010/10/30 18:28:06 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright (C) 1998  Chris Morley
// Modified by HUMUSOFT s.r.o.  Copyright 1998-2010 HUMUSOFT s.r.o.

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#include "VrmlNodePixelTexture.h"
#include "VrmlNodeType.h"
#include "VrmlScene.h"

static VrmlNode *creator( VrmlScene *s ) {
  return new VrmlNodePixelTexture(s);
}

// Define the built in VrmlNodeType:: "PixelTexture" fields

VrmlNodeType *VrmlNodePixelTexture::defineType(VrmlNodeType *t)
{
  static VrmlNodeType *st = 0;

  if (! t)
    {
      if (st) return st;
      t = st = new VrmlNodeType("PixelTexture", creator);
      t->reference();
    }

  VrmlNodeTexture::defineType(t);	// Parent class

  t->addExposedField("image", VrmlField::SFIMAGE, VrmlSFImage());
  t->addField("repeatS", VrmlField::SFBOOL, VrmlSFBool(true));
  t->addField("repeatT", VrmlField::SFBOOL, VrmlSFBool(true));
  return t;
}

VrmlNodeType & VrmlNodePixelTexture::nodeType() const
{
    return *defineType(0);
}


VrmlNodePixelTexture::VrmlNodePixelTexture(VrmlScene *scene) :
  VrmlNodeTexture(scene),
  d_repeatS(true),
  d_repeatT(true),
  d_texObject(0)
{
}

VrmlNodePixelTexture::~VrmlNodePixelTexture()
{
  invalidate();
}


void VrmlNodePixelTexture::invalidate()
{
  // delete precompiled texObject
  if (d_texObject && d_scene)
  {
    Viewer* primary = d_scene->getPrimaryViewer();
    if (primary) primary->removeTextureObject(d_texObject);
    d_texObject = 0;
  }
  VrmlNode::invalidate();
}


VrmlNode *VrmlNodePixelTexture::cloneMe() const
{
  return new VrmlNodePixelTexture(*this);
}


ostream& VrmlNodePixelTexture::printFields(ostream& os, int indent)
{
  if (! d_repeatS.get()) PRINT_FIELD(repeatS);
  if (! d_repeatT.get()) PRINT_FIELD(repeatT);
  if (d_image.getWidth() > 0 &&
      d_image.getHeight() > 0 &&
      d_image.getComponents() > 0 &&
      d_image.getPixels() != 0)
    PRINT_FIELD(image);
  return os;
}


void VrmlNodePixelTexture::render(Viewer* viewer, VrmlRenderContext rc)
{
  // replace precompiled object if modified
  if (isModified() && d_texObject!=0)
  {
    viewer->removeTextureObject(d_texObject);
    d_texObject = 0;
  }

  // use precompiled object if available
  if (d_texObject)
  {
    viewer->insertTextureReference(d_texObject, d_image.getComponents());
  }

  // precompile a new object from source data
  else
  {
    const unsigned char* pixelsToRender = d_image.getPixels();
    if (pixelsToRender != 0)
    {
      d_texObject = viewer->insertTexture(d_image.getWidth(),
                                          d_image.getHeight(),
                                          d_image.getComponents(),
                                          d_repeatS.get(),
                                          d_repeatT.get(),
                                          pixelsToRender,
                                          true);
    }
  }
  clearModified();
}


size_t VrmlNodePixelTexture::nComponents()
{
  return d_image.getComponents();
}

size_t VrmlNodePixelTexture::width()
{
  return d_image.getWidth();
}

size_t VrmlNodePixelTexture::height()
{
  return d_image.getHeight();
}

size_t VrmlNodePixelTexture::nFrames()
{
  return 0;
}

const unsigned char * VrmlNodePixelTexture::pixels()
{
  return d_image.getPixels();
}


// Get the value of a field or eventOut.

const VrmlField *VrmlNodePixelTexture::getField(const char *fieldName) const
{
  // exposedFields
  if ( strcmp( fieldName, "image" ) == 0 )
    return &d_image;

#if defined(__ORBISNAP__)

  else if (strcmp(fieldName, "repeatS") == 0)
    return &d_repeatS;
  else if (strcmp(fieldName, "repeatT") == 0)
    return &d_repeatT;

#endif

  return VrmlNode::getField(fieldName); // Parent class
}

// Set the value of one of the node fields.

void VrmlNodePixelTexture::setField(const char *fieldName,
				    const VrmlField &fieldValue)
{
  if TRY_FIELD(image, SFImage)
  else if TRY_FIELD(repeatS, SFBool)
  else if TRY_FIELD(repeatT, SFBool)
  else
    VrmlNode::setField(fieldName, fieldValue);

  //if ( strcmp(fieldName, "image") == 0 )
  //{
  //theSystem->debug("image w h c %dx%dx%d\n", d_image.width(), d_image.height(), d_image.nComponents());
  //unsigned char* p= d_image.pixels();
  //for (int y=0; y<d_image.height(); ++y) {
  //for (int y=0; y<1; ++y) {
  //for (int x=0; x<d_image.width(); ++x) {
  //for (int nc=0; nc<d_image.nComponents(); ++nc, ++p) {
  //if (*p)
  //cout << x << "," << y << "," << nc << ":" << (int)*p << endl;
  //}
  //}
  //}
  //theSystem->debug("pixel %d, %d byte %d: 0x%x",
  //x, y, nc, *p);
  //}
}

const VrmlSFImage & VrmlNodePixelTexture::getImage() const {
    return this->d_image;
}

const VrmlSFBool & VrmlNodePixelTexture::getRepeatS() const {
    return this->d_repeatS;
}

const VrmlSFBool & VrmlNodePixelTexture::getRepeatT() const {
    return this->d_repeatT;
}
