/*****************************************************************************
******************************************************************************
*
*               VRML Proto node
*
*               $Revision: 1.1.4.5.2.10 $
*               $Date: 2011/12/12 20:17:00 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright 1998-2011 HUMUSOFT s.r.o.
// Portions Copyright (C) 1998  Chris Morley

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


/**
 * @class VrmlNodeProto
 *
 * @brief A VrmlNodeProto object represents an instance of a PROTOd node.
 *
 * The definition of the PROTO is stored in a VrmlNodeType object;
 * the VrmlNodeProto object stores a local copy of the implementation
 * nodes.
 */
#include "VrmlNodeProto.h"
#include "VrmlNamespace.h"
#include "VrmlBSphere.h"
#include "VrmlScene.h"
#include "Doc.h"

#include <cstdlib>
#include <string.h>

//#ifndef NDEBUG
//#define VRML_NODE_PROTO_DEBUG
//#endif

//
// Instances of PROTOs clone the implementation nodes stored
// in a VrmlNodeType object. The only tricky parts are to
// make sure ROUTEs are properly copied (the DEF name map is
// not available) and that fields are copied properly (the
// MF* guys currently share data & need to be made copy-on-
// write for this to be correct). Flags are set as each node
// is cloned so that USEd nodes are referenced rather than
// duplicated.
//
// ROUTEs: Build a temp namespace as each (named) implementation
// node is cloned, then traverse the implementation nodes again,
// reproducing the routes in the cloned nodes using the temp ns.
// I think that the addToScene() method is the right place to
// download EXTERNPROTO implementations. Should also check that
// the first node matches the expected type (need to store the
// expected type if first node is null when one of the type
// tests is run).
//
// Events between nodes in the PROTO implementation are handled
// by the ROUTE copying described above. For eventIns coming into
// the proto, when the implementation nodes are copied, a list
// of eventIns/exposedFields along with their IS mappings should
// be constructed.
// EventOuts from an implementation node to a node outside the
// PROTO can be directly replaced at copy time.
//

VrmlNodeType & VrmlNodeProto::nodeType() const
{
    return *d_nodeType;
}


VrmlNodeProto::VrmlNodeProto(VrmlNodeType *nodeDef, VrmlScene *scene) :
  VrmlNode(scene),
  d_nodeType(nodeDef->reference()),
  d_instantiated(false),
  d_scope(0),
  d_nodes(0),
  d_viewerObject(0)
{
}

VrmlNodeProto::VrmlNodeProto(const VrmlNodeProto &n) :
  VrmlNode(0),
  d_nodeType(n.nodeType().reference()),
  d_instantiated(false),
  d_scope(0),
  d_nodes(0),
  d_viewerObject(0)
{
  // Copy fields.
  std::list<NameValueRec*>::const_iterator i;
  for (i = n.d_fields.begin(); i != n.d_fields.end(); i++)
    {
      VrmlField *value = (*i)->value;
      setField( (*i)->name, *value );
    }
  this->setBVolumeDirty(true); // lazy calc of bvolume
}

VrmlNodeProto::~VrmlNodeProto()
{
  // Free strings & values
  std::list<NameValueRec*>::iterator i;

  for (i = d_fields.begin(); i != d_fields.end(); i++) {
    NameValueRec *r = *i;
    free(r->name);
    delete r->value;
    delete r;
  }

  EventDispatchList::iterator e;
  for (e = d_eventDispatch.begin(); e != d_eventDispatch.end(); ++e)
    {
      EventDispatch *ed = *e;
      delete [] ed->name;
      VrmlNodeType::ISMap::iterator j;
      for (j = ed->ismap.begin(); j != ed->ismap.end(); ++j)
	{
	  VrmlNodeType::NodeFieldRec *nf = (*j);
	  delete [] nf->fieldName;
	  delete nf;
	}
      delete ed;
    }

  // delete precompiled viewerObject
  if (d_viewerObject && d_scene)
  {
    Viewer* primary = d_scene->getPrimaryViewer();
    if (primary) primary->removeObject(d_viewerObject);
  }

  delete d_nodes;
  delete d_scope;
  d_nodeType->dereference();
}


// Note that the copy constructor doesn't copy the implementation
// nodes, so they will need to be instantiated.

VrmlNode *VrmlNodeProto::cloneMe() const
{
  return new VrmlNodeProto(*this);
}

void VrmlNodeProto::cloneChildren(VrmlNamespace* ns)
{
  // we break rule that proto instances share interface fields (d_fields),
  // but it should not case any memory lack issues

  VrmlNodeType* newNodeType = 0;
  // check whether node type exists
  if (!(newNodeType=const_cast<VrmlNodeType*> (ns->findType(d_nodeType->getName()))))
  {
    newNodeType = new VrmlNodeType(*d_nodeType, ns);
  }

  // we created deep copy, dereference previous node type
  // (in case that namespaces are equal we do not have to worry about d_nodeType deletion
  // as there are at least two instances at the moment)
  d_nodeType->dereference();

  // use target namespace type or newly created one
  d_nodeType = newNodeType;

  d_nodeType->reference();

  // add new node type to parent namespace if not exists there
  if (!ns->findType(d_nodeType->getName()))
    ns->addNodeType(d_nodeType, false);


  // Deep copy fields.
  std::list<NameValueRec*>::iterator i;
  for (i = d_fields.begin(); i != d_fields.end();)
  {
    if (d_nodeType->hasInterface((*i)->name))
    {
      VrmlField *value = (*i)->value->clone(ns);
      setField( (*i)->name, *value );
      ++i;
    }
    else
    {
      NameValueRec *nv = findField((*i)->name);

      // remove it from list
      i=d_fields.erase(i);

      // delete interface item
      free(nv->name);
      delete nv->value;
      delete nv;
    }
  }

  this->setBVolumeDirty(true); // lazy calc of bvolume
}


VrmlNode* VrmlNodeProto::instantiate(VrmlNamespace* ns)
{
  return newInstance(ns);
}


void VrmlNodeProto::reinstantiate()
{
  instantiate();
  addImplementationNodesToScene();
}

// Instantiate a local copy of the implementation nodes.
// EXTERNPROTOs are actually loaded here. We don't want
// *references* (DEF/USE) to the nodes in the PROTO definition,
// we need to actually clone new nodes...

void VrmlNodeProto::instantiate()
{
  //////////////////////////////////////////////////////////////////////////////
  // it can be reinstantiation -- delete nodes and IS mappings first
  //////////////////////////////////////////////////////////////////////////////
  // PROTOs can be nested
  if (getEncapsulatingPrototype()) getEncapsulatingPrototype()->setModified();

  EventDispatchList::iterator e;
  for (e = d_eventDispatch.begin(); e != d_eventDispatch.end(); ++e)
  {
    EventDispatch *ed = *e;
    delete [] ed->name;
    VrmlNodeType::ISMap::iterator j;
    for (j = ed->ismap.begin(); j != ed->ismap.end(); ++j)
      {
	VrmlNodeType::NodeFieldRec *nf = (*j);
	delete [] nf->fieldName;
	delete nf;
      }
    delete ed;
  }
  d_eventDispatch.clear();

  // delete precompiled viewerObject
  if (d_viewerObject && d_scene)
  {
    Viewer* primary = d_scene->getPrimaryViewer();
    if (primary) primary->removeObject(d_viewerObject);
  }

  delete d_nodes;
  d_nodes = 0;
  delete d_scope;
  d_scope = 0;
  //////////////////////////////////////////////////
  // previous implementation (if any) was destructed
  //////////////////////////////////////////////////


  //////////////////////////////////////////////////
  // create PROTO implementation ///////////////
  //////////////////////////////////////////////////

  VrmlMFNode *protoNodes = d_nodeType->getImplementationNodes();

  // copy default values from node type to PROTO instance
  // this must be done after EXTERNPROTO is loaded above
  // do not copy any existing fields because they have their values already set
  // by VRMLDecoder::loadFieldValue when parsing the VRML world
  {
    VrmlNodeType::FieldList &flist = d_nodeType->fields();
    for (VrmlNodeType::FieldList::const_iterator i=flist.begin(); i!=flist.end(); ++i)
    {
      if (getField((*i)->name) == NULL)
      {
	VrmlField* value = (*i)->defaultValue;
	if (value != NULL)
	  setField((*i)->name, *value);
      }
    }
  }

  // remove fields which do not exist any more
  std::list<NameValueRec*>::iterator it = d_fields.begin();
  while (it != d_fields.end())
  {
    NameValueRec *r = *it;
    if (!d_nodeType->hasField(r->name) && !d_nodeType->hasExposedField(r->name))
    {
      free(r->name);
      delete r->value;
      delete r;

      it = d_fields.erase(it);
    }
    else
      ++it;
  }

  // copy eventOuts from the definition
  {
    VrmlNodeType::FieldList &flist = d_nodeType->eventOuts();
    for (VrmlNodeType::FieldList::const_iterator i=flist.begin(); i!=flist.end(); ++i)
    {
      VrmlField* value = (*i)->defaultValue;

      // be careful not to overwrite already existing fields
      if (getField((*i)->name) == NULL)
      {
	if (value == NULL)
	{
	  value = VrmlField::create((*i)->type);
	  setField((*i)->name, *value);
	  delete value;
	}
	else setField((*i)->name, *value);
      }
    }
  }

  int nNodes = protoNodes ? protoNodes->getLength() : 0;
  int i;
  d_scope = new VrmlNamespace();

  // Clear all flags - encountering a set flag during cloning
  // indicates a USEd node, which should be referenced rather
  // than cloned.
  for (i=0; i<nNodes; ++i)
    (*protoNodes)[i]->clearFlags();

  // Instantiate nodes
  d_nodes = new VrmlMFNode(nNodes);
  Doc relUrl;
  for (i=0; i<nNodes; ++i)
  {
    (*this->d_nodes)[i] = (*protoNodes)[i]->instantiate(d_scope)->reference();
    ((*d_nodes)[i])->setProto(this);
  }

  // Copy internal (to the PROTO implementation) ROUTEs.
  {
    RouteList& rl = d_nodeType->scope()->getRouteList();
    RouteList::iterator r;
    for (r=rl.begin(); r!=rl.end(); ++r)
    {
      d_scope->addRoute(
	d_scope->findNode(r->second.fromNode()->name()),
	r->second.fromEventOut(),
	d_scope->findNode(r->second.toNode()->name()),
	r->second.toEventIn());
    }
  }

  // Collect eventIns coming from outside the PROTO.
  // A list of eventIns along with their maps to local
  // nodes/eventIns is constructed for each instance.
  VrmlNodeType::FieldList &eventIns = d_nodeType->eventIns();
  VrmlNodeType::FieldList::iterator ev;
  VrmlNodeType::ISMap *ismap;
  VrmlNodeType::ISMap::const_iterator j;

  for (ev = eventIns.begin(); ev != eventIns.end(); ++ev)
  {
    EventDispatch *ed = new EventDispatch;
    char *eventName = (*ev)->name;

    ed->name = new char[strlen(eventName)+1];
    strcpy(ed->name, eventName);
    ismap = &(*ev)->thisIS;
    d_eventDispatch.push_front(ed);

    for (j = ismap->begin(); j != ismap->end(); ++j)
    {
      VrmlNodeType::NodeFieldRec *nf = new VrmlNodeType::NodeFieldRec;
      nf->node = d_scope->findNode((*j)->node->name());
      nf->fieldName = new char[strlen((*j)->fieldName)+1];
      strcpy(nf->fieldName, (*j)->fieldName);
      ed->ismap.push_front(nf);
    }
  }

  // Set IS'd field values in the implementation nodes to
  // the values specified in the instantiation.
  std::list<NameValueRec*>::iterator ifld;
  for (ifld = d_fields.begin(); ifld != d_fields.end(); ++ifld)
  {
    VrmlField *value = (*ifld)->value;

    if (!value) continue;
    if ((ismap = d_nodeType->getFieldISMap( (*ifld)->name )) != 0)
    {
      for (j = ismap->begin(); j != ismap->end(); ++j)
      {
	VrmlNode *n = d_scope->findNode((*j)->node->name());
	if (n) n->setField( (*j)->fieldName, *value );
      }
    }
  }

  d_instantiated = true;
  this->setBVolumeDirty(true); // lazy calc of bvolume
}


void VrmlNodeProto::addToScene(VrmlScene *s, const char *relUrl)
{
  d_scene = s;

  // Make sure my nodes are here (reinstantiation possible here)
  instantiate();

  // add the implementation nodes to the scene.
  addImplementationNodesToScene(relUrl);
}

void VrmlNodeProto::addImplementationNodesToScene(const char *relUrl)
{
  // add the implementation nodes to the scene.
  if (d_nodes)
  {
    const char *rel = d_nodeType->url();
    int j, n = d_nodes->getLength();
    for (j=0; j<n; ++j)
      (*d_nodes)[j]->addToScene(d_scene, rel ? rel : relUrl);
  }
}

void VrmlNodeProto::accumulateTransform( VrmlNode *n )
{
  // Make sure my nodes are here
  if (! d_instantiated)
    {
      theSystem->debug("VrmlNodeProto::%s accumTrans before instantiation\n",
		       name());
      instantiate();
    }

  // ... and process the implementation nodes.
  if (d_nodes)
    {
      int i, j = d_nodes->getLength();
      for (i=0; i<j; ++i)
	(*d_nodes)[i]->accumulateTransform(n);
    }
}


int VrmlNodeProto::size()
{   return d_nodes ? d_nodes->getLength() : 0;	}


VrmlNode *VrmlNodeProto::child(int index)
{
  if (d_nodes && index >= 0 && index < (int) d_nodes->getLength())
    return (*d_nodes)[index];
  return 0;
}

// Print the node type, instance vars

ostream& VrmlNodeProto::printFields(ostream& os, int )
{
  std::list<NameValueRec*>::iterator i;

  for (i = d_fields.begin(); i != d_fields.end(); i++) {
    NameValueRec *r = *i;
    os << r->name;
    os << r->value;
  }

  return os;
}


// Use the first node to check the type

VrmlNode *VrmlNodeProto::firstNode() const
{
  return ((d_nodes && d_nodes->getLength())
	  ? (*d_nodes)[0]
	  : d_nodeType->firstNode());
}


void VrmlNodeProto::render(Viewer *viewer, VrmlRenderContext rc)
{
  if (! d_instantiated)
    {
      theSystem->debug("VrmlNodeProto::%s render before instantiation\n",
		       name());
      instantiate();
    }

  if (d_viewerObject)
    viewer->insertReference(d_viewerObject);

  else if (d_nodes)
    {
      d_viewerObject = viewer->beginObject( name() );

      // render the nodes with the new values
      int n = d_nodes->getLength();
      for (int j = 0; j<n; ++j)
	    (*d_nodes)[j]->render(viewer, rc);
      viewer->endObject();
    }

  clearModified();
}

void VrmlNodeProto::invalidate()
{
  if (d_nodes)
  {
    int n = d_nodes->getLength();
    for (int j = 0; j<n; ++j)
      (*d_nodes)[j]->invalidate();
  }
  VrmlNode::invalidate();
}

void VrmlNodeProto::guessNearFarPlane(Viewer* viewer, double* nearVal, double* farVal)
{
  if (d_nodes)
  {
    int n = d_nodes->getLength();
    for (int j = 0; j<n; ++j)
      child(j)->guessNearFarPlane(viewer, nearVal, farVal);
  }
}

void VrmlNodeProto::eventIn(double timeStamp,
			    const char *eventName,
			    const VrmlField *fieldValue)
{
  if (! d_instantiated)
  {
    theSystem->debug("VrmlNodeProto::%s eventIn before instantiation\n",
		     name());
    instantiate();
  }

  EventDispatchList::iterator i;
  const char* evname = VrmlNodeType::strstarts(eventName, "set_") ? eventName+4 : eventName;
  for (i=d_eventDispatch.begin(); i != d_eventDispatch.end(); ++i)
  {
    const char* edname = VrmlNodeType::strstarts((*i)->name, "set_") ? (*i)->name+4 : (*i)->name;
    if (strcmp(evname, edname) == 0)
    {
      VrmlNodeType::ISMap *ismap = & ((*i)->ismap);
      VrmlNodeType::ISMap::iterator j;
      for (j = ismap->begin(); j != ismap->end(); ++j)
	(*j)->node->eventIn( timeStamp, (*j)->fieldName, fieldValue );
      return;
    }
  }

  // Let the generic code handle errors.
  VrmlNode::eventIn( timeStamp, eventName, fieldValue );
  this->setBVolumeDirty(true); // lazy calc of bvolume
}



VrmlNodeProto::NameValueRec *VrmlNodeProto::findField(const char *fieldName) const
{
  std::list<NameValueRec*>::const_iterator i;
  for (i = d_fields.begin(); i != d_fields.end(); ++i) {
    NameValueRec *nv = *i;
    if (nv != NULL && strcmp(nv->name, fieldName) == 0) {
      return nv;
    }
  }
  return NULL;
}


// Set the value of one of the node fields (creates the field if
// it doesn't exist - is that necessary?...)

void VrmlNodeProto::setField(const char *fieldName,
			     const VrmlField &fieldValue)
{
  NameValueRec *nv = findField(fieldName);

  if (!nv)
  {
    nv = new NameValueRec;
    nv->name = strdup(fieldName);
    d_fields.push_front(nv);
  }
  else
  {
    if (nv->value==&fieldValue) return;    // setting a field to itself
    delete nv->value;
  }

  nv->value = fieldValue.clone();
  this->setBVolumeDirty(true); // lazy calc of bvolume
}


const VrmlField *VrmlNodeProto::getField(const char *fieldName) const
{
  NameValueRec *nv = findField(fieldName);
  if (nv) return nv->value;
  return 0;
}


void VrmlNodeProto::PROTOFieldChanged(const char *fieldName)
{
  for (std::list<NameValueRec*>::const_iterator i = d_fields.begin(); i != d_fields.end(); ++i)
  {
    if (strcmp((*i)->name, fieldName) == 0)
    {
      VrmlNodeType::ISMap* ismap = d_nodeType->getFieldISMap(fieldName);
      if (ismap)
      {
	for (VrmlNodeType::ISMap::iterator j = ismap->begin(); j != ismap->end(); ++j)
	{
	  VrmlNode *n = d_scope->findNode((*j)->node->name());
	  if (n)
	  {
	    n->setField( (*j)->fieldName, *((*i)->value) );
	    n->setModified();
	  }
	}
      }
      break;
    }
  }
}


const VrmlBVolume*
VrmlNodeProto::getBVolume() const
{
  //cout << "VrmlNodeProto::getBVolume() {" << endl;
  if (!d_instantiated) {
    cout << "VrmlNodeProto::getBVolume():WARNING:not instantiated" << endl;
    return (VrmlBVolume*)0; // shouldn't happen
  }
  VrmlNode* base = firstNode();
  if (!base) {
    cout << "VrmlNodeProto::getBVolume():WARNING:!base" << endl;
    return (VrmlBVolume*)0; // shouldn't happen
  }
  const VrmlBVolume* bv = base->getBVolume();
  //cout << "}";
  //bv->dump(cout) << endl;
  return bv;
}

void
VrmlNodeProto::updateModified(VrmlNodePath& path, int flags)
{
  //cout << "VrmlNodeProto::updateModified()" << endl;
  if (this->isModified()) markPathModified(path, true);
  path.push_front(this);
  VrmlNode* base = firstNode();
  if (base) base->updateModified(path, flags);
  path.pop_front();
}

void VrmlNodeProto::clearModifiedAll()
{
  // first, clear the flag of all descendants
  VrmlNode::clearModifiedAll();

  // clear the flag for the implementation nodes
  unsigned int i;
  for (i=0; i<d_nodes->getLength(); i++)
    (*d_nodes)[i]->clearModifiedAll();
}

void VrmlNodeProto::setProto(VrmlNodeProto* proto)
{
  if (d_nodes)
  {
    for (size_t i=0; i<d_nodes->getLength(); i++)
    {
      VrmlNode* ci = (*d_nodes)[i];
      ci->setProto(proto);
    }
  }
  VrmlNode::setProto(proto);
}

void VrmlNodeProto::setEncapsulatingPrototype(VrmlNodeType* type)
{
  // PROTO instance resides inside another PROTO declaration
  // instantiate it to ensure that instance will not remain uninstantiated for ever
  if (type != 0 && !d_instantiated) instantiate();

  // add the implementation nodes to the scene.
  addImplementationNodesToScene();

  if (d_nodes)
  {
    for (size_t i=0; i<d_nodes->getLength(); i++)
    {
      VrmlNode* ci = (*d_nodes)[i];
      ci->setEncapsulatingPrototype(type);
    }
  }
  VrmlNode::setEncapsulatingPrototype(type);
}
