/*****************************************************************************
******************************************************************************
*
*               VRML ProximitySensor node
*
*               $Revision: 1.1.2.1 $
*               $Date: 2009/01/22 12:57:53 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright (C) 1998  Chris Morley
// Modified by HUMUSOFT s.r.o.  Copyright 1998-2009 HUMUSOFT s.r.o.

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "VrmlNodeProximitySensor.h"
#include "VrmlNodeType.h"
#include "MathUtils.h"
#include "System.h"
#include "Viewer.h"
#include "VrmlScene.h"

// ProximitySensor factory.

static VrmlNode *creator( VrmlScene *scene )
{
  return new VrmlNodeProximitySensor(scene);
}


// Define the built in VrmlNodeType:: "ProximitySensor" fields

VrmlNodeType *VrmlNodeProximitySensor::defineType(VrmlNodeType *t)
{
  static VrmlNodeType *st = 0;

  if (! t)
    {
      if (st) return st;                // Only define the type once.
      t = st = new VrmlNodeType("ProximitySensor", creator);
      t->reference();
    }

  VrmlNodeChild::defineType(t); // Parent class
  t->addExposedField("center", VrmlField::SFVEC3F, VrmlSFVec3f(0, 0, 0));
  t->addExposedField("size", VrmlField::SFVEC3F, VrmlSFVec3f(0, 0, 0));
  t->addExposedField("enabled", VrmlField::SFBOOL, VrmlSFBool(true));
  t->addEventOut("enterTime", VrmlField::SFTIME);
  t->addEventOut("exitTime", VrmlField::SFTIME);
  t->addEventOut("isActive", VrmlField::SFBOOL);
  t->addEventOut("orientation_changed", VrmlField::SFROTATION);
  t->addEventOut("position_changed", VrmlField::SFVEC3F);

  return t;
}

VrmlNodeType & VrmlNodeProximitySensor::nodeType() const
{
    return *defineType(0);
}


VrmlNodeProximitySensor::VrmlNodeProximitySensor( VrmlScene *scene ) :
  VrmlNodeChild(scene),
  d_center(0.0, 0.0, 0.0),
  d_enabled(true),
  d_size(0.0, 0.0, 0.0),
  d_isActive(false),
  d_position(0.0, 0.0, 0.0),
  d_enterTime(0.0),
  d_exitTime(0.0),
  d_parentTransform(0)
{
  setModified();
}

VrmlNodeProximitySensor::~VrmlNodeProximitySensor()
{
}

VrmlNode *VrmlNodeProximitySensor::cloneMe() const
{
  return new VrmlNodeProximitySensor(*this);
}


ostream& VrmlNodeProximitySensor::printFields(ostream& os, int indent)
{
  if (! FPZERO(d_center.getX()) ||
      ! FPZERO(d_center.getY()) ||
      ! FPZERO(d_center.getZ()) )
    PRINT_FIELD(center);
  if (! d_enabled.get())
    PRINT_FIELD(enabled);
  if (! FPZERO(d_size.getX()) ||
      ! FPZERO(d_size.getY()) ||
      ! FPZERO(d_size.getZ()) )
    PRINT_FIELD(size);

  return os;
}

// Cache a pointer to (one of the) parent transforms for converting
// hits into local coords.

void VrmlNodeProximitySensor::accumulateTransform( VrmlNode *parent )
{
  d_parentTransform = parent;
}

VrmlNode* VrmlNodeProximitySensor::getParentTransform() { return d_parentTransform; }

//
// Generate proximity events. If necessary, events prior to the current
// time are generated due to interpolation of enterTimes and exitTimes.
// The timestamp should never be increased.
//
// This is in a render() method since the it needs the viewer position
// with respect to the local coordinate system.
// Could do this with VrmlNode::inverseTransform(double [4][4]) now...
//
// The positions and times are not interpolated to report the exact
// place and time of entries and exits from the sensor regions as
// required by the spec, since that would require the last viewer position
// to be stored in the node, which is problematic in the presence of
// DEF/USEd nodes...
// I suppose the scene could keep the last viewer position in the global
// coordinate system and it could be transformed all the way down the
// scenegraph, but that sounds painful.

void VrmlNodeProximitySensor::render(Viewer *viewer, VrmlRenderContext rc)
{
  // if the sensor is disabled, or if its size is zero, do nothing
  if
  (
    !d_enabled.get()
    || (d_size.getX() <= 0.0 && d_size.getY() <= 0.0 && d_size.getZ() <= 0.0)
    || (viewer->getRenderMode() != Viewer::RENDER_MODE_DRAW)
  )
  {
    clearModified();
    return;
  }

  // get the current time
  double timeNow = d_scene->getTime();

  // get transformation matrices for this node
  Matrix4x4 m = getCompoundTransformationMatrix();
  Matrix4x4 im = getCompoundInverseTransformationMatrix();

  // calculate the sensor center and size in world coordinates
  Point3 center = m*Point3(d_center.get());
  Vector3 size = m*Vector3(d_size.get());

  // get the current camera position and orientation in world coordinates
  Camera cam = viewer->getCameraAbsolute();
  Point3 pos = cam.getPosition();
  Vector3 dir = cam.getDirection();

  // is viewer inside the box?
  bool inside = (fabs(pos.getX()-center.getX()) <= 0.5*size.getX()
    && fabs(pos.getY()-center.getY()) <= 0.5*size.getY()
    && fabs(pos.getZ()-center.getZ()) <= 0.5*size.getZ());
  bool wasIn = d_isActive.get();

  // check if the viewer has just entered the box
  if (inside && !wasIn)
  {
    d_isActive.set(true);
    eventOut(timeNow, "isActive", d_isActive);
    d_enterTime.set(timeNow);
    eventOut(timeNow, "enterTime", d_enterTime);
  }

  // check if the viewer has just left the box
  else if (!inside && wasIn)
  {
    d_isActive.set(false);
    eventOut(timeNow, "isActive", d_isActive );
    d_exitTime.set(timeNow);
    eventOut(timeNow, "exitTime", d_exitTime);
  }

  // movement inside the box
  else if (wasIn || inside)
  {
    // calculate viewer position in local coordinates
    Point3 localpos = im*pos;

    // has the position changed?
    if (!FPEQUAL(d_position.getX(), localpos.getX())
        || !FPEQUAL(d_position.getY(), localpos.getY())
        || !FPEQUAL(d_position.getZ(), localpos.getZ()))
    {
      d_position.setX(localpos.getX());
      d_position.setY(localpos.getY());
      d_position.setZ(localpos.getZ());
      eventOut(timeNow, "position_changed", d_position);
    }

    // Calculate viewer orientation in local coordinates.
    // We need to calculate a rotation (an axis and an angle) that transforms
    // the default direction (0, 0, 1) to the current direction
    double angle;
    Vector3 rotAxis;
    Quaternion quat(viewer->getCameraAbsolute().computeViewingMatrix()*im);
    quat.toAxisAngle(rotAxis, angle);
    rotAxis = m*rotAxis;

    // has the orientation changed?
    if (!FPEQUAL(d_orientation.getX(), rotAxis.getX())
        || !FPEQUAL(d_orientation.getY(), rotAxis.getY())
        || !FPEQUAL(d_orientation.getZ(), rotAxis.getZ())
        || !FPEQUAL(d_orientation.getAngle(), angle))
    {
      d_orientation.setX(rotAxis.getX());
      d_orientation.setY(rotAxis.getY());
      d_orientation.setZ(rotAxis.getZ());
      d_orientation.setAngle(angle);
      eventOut(timeNow, "orientation_changed", d_orientation);
    }
  }
}


const VrmlField *VrmlNodeProximitySensor::getField(const char *fieldName) const
{
  if ( strcmp( fieldName, "center" ) == 0 )
    return &d_center;
  else if ( strcmp( fieldName, "enabled" ) == 0 )
    return &d_enabled;
  else if ( strcmp( fieldName, "size" ) == 0 )
    return &d_size;

  // eventOuts
  else if ( strcmp( fieldName, "isActive" ) == 0 )
    return &d_isActive;
  else if ( strcmp( fieldName, "position" ) == 0 )
    return &d_position;
  else if ( strcmp( fieldName, "orientation" ) == 0 )
    return &d_orientation;
  else if ( strcmp( fieldName, "enterTime" ) == 0 )
    return &d_enterTime;
  else if ( strcmp( fieldName, "exitTime" ) == 0 )
    return &d_exitTime;

  return VrmlNodeChild::getField( fieldName );
}

// Set the value of one of the node fields.

void VrmlNodeProximitySensor::setField(const char *fieldName,
                                       const VrmlField &fieldValue)
{
  if TRY_FIELD(center, SFVec3f)
  else if TRY_FIELD(enabled, SFBool)
  else if TRY_FIELD(size, SFVec3f)
  else
    VrmlNodeChild::setField(fieldName, fieldValue);
}

