/*****************************************************************************
******************************************************************************
*
*               VRML Script node
*
*               $Revision: 1.1.2.17 $
*               $Date: 2011/09/30 05:54:18 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright 1998-2011 HUMUSOFT s.r.o.
// Portions Copyright (C) 1998  Chris Morley

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#include <string.h> // strncpy, strcat
#include <stdio.h> //  snprintf

#include "VrmlNodeScript.h"
#include "VrmlNodeType.h"
#include "ScriptObject.h"
#include "VrmlScene.h"
#include "VrmlNamespace.h"

#if defined(WIN32)
// Disable warning messages about forcing value to bool 'true' or 'false'
#pragma warning (disable:4800)
#endif


#define TERMINAL_STRING '#'
#define SYNTH_NAME_BEGIN_STRING ','


// Per-script field type
VrmlNodeScript::ScriptField::ScriptField(VrmlField::VrmlFieldType type_, VrmlField* value_) :
  value(value_),
  type(type_),
  modified(false)
{ }

VrmlNodeScript::ScriptField::ScriptField() :
  value(0),
  type(VrmlField::NO_FIELD),
  modified(false)
{}

VrmlNodeScript::ScriptField::ScriptField(const ScriptField& other) :
  value(0)
{
  *this = other;
}


VrmlNodeScript::ScriptField::~ScriptField()
{ delete value; }


VrmlNodeScript::ScriptField& VrmlNodeScript::ScriptField::operator=(const ScriptField& other)
{
  type = other.type;
  modified = false;

  delete value;
  value = (other.value) ? other.value->clone() : 0;

  return *this;
}


// Script factory. Add each Script to the scene for fast access.

static VrmlNode *creator( VrmlScene *scene )
{
  return new VrmlNodeScript(scene);
}


// Define the built in VrmlNodeType:: "Script" fields

VrmlNodeType *VrmlNodeScript::defineType(VrmlNodeType *t)
{
  static VrmlNodeType *st = 0;

  if (! t)
    {
      if (st) return st;		// Only define the type once.
      t = st = new VrmlNodeType("Script", creator);
      t->reference();
    }

  VrmlNodeChild::defineType(t);	// Parent class
  t->addExposedField("url", VrmlField::MFSTRING, VrmlMFString());
  t->addField("directOutput", VrmlField::SFBOOL, VrmlSFBool(false));
  t->addField("mustEvaluate", VrmlField::SFBOOL, VrmlSFBool(false));
  return t;
}

// Should subclass NodeType and have each Script maintain its own type...

VrmlNodeType& VrmlNodeScript::nodeType() const
{
  return *defineType(0);
}


VrmlNodeScript::VrmlNodeScript( VrmlScene *scene ) :
  VrmlNodeChild(scene),
  d_directOutput(false),
  d_mustEvaluate(false),
  d_script(0),
  d_eventsReceived(0),
  d_eventInsSequence(0), d_eventOutsSequence(0), d_fieldsSequence(0)
{
  if (d_scene) d_scene->addScript(this);
}

VrmlNodeScript::VrmlNodeScript(const VrmlNodeScript& other) :
  VrmlNodeChild(0),
  d_directOutput(other.d_directOutput),
  d_mustEvaluate(other.d_mustEvaluate),
  d_url(other.d_url),
  d_script(0),		// Force creation of a distinct script object
  d_eventsReceived(0),

  d_eventInsSequence(other.d_eventInsSequence),
  d_eventOutsSequence(other.d_eventOutsSequence),
  d_fieldsSequence(other.d_fieldsSequence)
{
  d_eventIns = other.d_eventIns;
  d_eventOuts = other.d_eventOuts;
  d_fields = other.d_fields;
}


VrmlNodeScript::~VrmlNodeScript()
{
  // removeScript ought to call shutdown...
  if (d_scene)
  {
    d_scene->removeScript(this);
    shutdown(d_scene->getTime());
  }

  delete d_script;
}


VrmlNode *VrmlNodeScript::cloneMe() const
{
  return new VrmlNodeScript(*this);
}

//
// Any SFNode or MFNode fields need to be cloned as well.
//

void VrmlNodeScript::cloneChildren(VrmlNamespace* ns)
{
  FieldList* containers[] = {&d_fields, &d_eventIns, &d_eventOuts};
  for(int i=0; i<sizeof(containers)/sizeof(FieldList*); ++i)
  {
    for (FieldList::iterator it = (containers[i])->begin(); it != (containers[i])->end(); ++it)
    {
      if (it->second.value)
      {
	VrmlField* copy = it->second.value->clone(ns);
	delete it->second.value;
	it->second.value = copy;
      }
    }
  }
}



void VrmlNodeScript::addToScene(VrmlScene *s, const char *relUrl)
{
  theSystem->debug("VrmlNodeScript::%s %p addToScene %p\n",
		   name(), this, s);

  d_relativeUrl.set(relUrl);
  if (d_scene == s) return;
  d_scene = s;
  if (d_scene != 0)
  {
    initialize(d_scene->getTime());
    d_scene->addScript(this);
  }
}

ostream& VrmlNodeScript::fixFieldsToChain(ostream& os, bool editFlag) const
{
  // write 'url' field
  const char* fname = "url";

  os << 'F' << TERMINAL_STRING;
  os << (isFieldValueDefault(fname) ? 'D' : 'C') << TERMINAL_STRING;
  os << fname << TERMINAL_STRING;
  os << "MFString" << TERMINAL_STRING;
  os << "0" << TERMINAL_STRING;

  // write 'directOutput' field
  fname = "directOutput";

  os << 'F' << TERMINAL_STRING;
  os << (isFieldValueDefault(fname) ? 'D' : 'C') << TERMINAL_STRING;
  os << fname << TERMINAL_STRING;
  os << "SFBool" << TERMINAL_STRING;
  os << "1" << TERMINAL_STRING;

  // write 'mustEvaluate' field
  fname = "mustEvaluate";

  os << 'F' << TERMINAL_STRING;
  os << (isFieldValueDefault(fname) ? 'D' : 'C') << TERMINAL_STRING;
  os << fname << TERMINAL_STRING;
  os << "SFBool" << TERMINAL_STRING;
  os << "1" << TERMINAL_STRING;

  return os;
}



ostream& VrmlNodeScript::toChain(ostream& os, bool editFlag) const
{
  // if this node is marked, it means it was already printed in this round;
  // that means a USE reference was encountered
  if (d_flag)
  {
    os << 'U' << name() << TERMINAL_STRING;
    os << "Script" << TERMINAL_STRING;
    // terminate USE.
    os << TERMINAL_STRING;
    return os;
  }

  // 'N' is a mark for node
  os << "N";
  // insert name of node and terminate the string with tabular
  const char *nm = name();
  if (nm && *nm) os << nm;
  os << TERMINAL_STRING;

  // insert name of the string withing parenthesses and terminate the string
  os << nodeType().getName() << TERMINAL_STRING;

  // write fixed fields at first.
  fixFieldsToChain(os, editFlag);

  // print user-added fields and events
  const FieldISBinding* binding;
  for (FieldList::const_iterator it=d_fields.begin(); it!=d_fields.end(); ++it)
  {
    os << "F" << TERMINAL_STRING;
    os << (isFieldValueDefault(it->first.c_str()) ? 'D' : 'C') << TERMINAL_STRING;
    // FIELD name
    os << it->first << TERMINAL_STRING;
    // FIELD type
    const VrmlField* field = getField(it->first.c_str());
    int type = it->second.type;
    os << VrmlField::getTypeName(type) << TERMINAL_STRING;

    binding = getISBinding(it->first.c_str());
    if (binding == NULL)
    {
      os << "S";
      field->toChain(os, editFlag);
    }
    else
      os << (VrmlField::IsExtSettable(type) ? "3" : "0") << TERMINAL_STRING;

    os << TERMINAL_STRING;
  }

  for (FieldList::const_iterator it=d_eventOuts.begin(); it!=d_eventOuts.end(); ++it)
  {
    // FIELD name
    os << "O" << it->first << TERMINAL_STRING;
    const VrmlField* field = getField(it->first.c_str());
    int type = it->second.type;
    os << VrmlField::getTypeName(type) << TERMINAL_STRING;
    os << (VrmlField::IsExtSettable(type) ? "2" : "0") << TERMINAL_STRING;
  }

  for (FieldList::const_iterator it=d_eventIns.begin(); it!=d_eventIns.end(); ++it)
  {
    os << "I" << it->first << TERMINAL_STRING;
    // FIELD type
    const VrmlField* field = getField(it->first.c_str());
    int type = it->second.type;
    os << VrmlField::getTypeName(type) << TERMINAL_STRING;
    os << (VrmlField::IsExtSettable(type) ? "1" : "0") << TERMINAL_STRING;
  }

  os << TERMINAL_STRING;

  // mark this node as already printed; if it is encountered next time,
  // it means it is a USEd node and all other references should be printed
  // using USE.
  ((VrmlNode*) this)->setFlag();

  return os;
}


void VrmlNodeScript::printFieldGroup(ostream& os, const char* fieldclass, const FieldList& recs, int indent) const
{
  for (FieldList::const_iterator it=recs.begin(); it!=recs.end(); ++it)
  {

    it->second.value->printComment(os, indent+VrmlNode::INDENT_INCREMENT, true);

    os << fieldclass << " ";
    os << VrmlField::getTypeName(it->second.type) << " ";
    os << it->first << " ";
    const FieldISBinding* binding = getISBinding(it->first.c_str());
    if (binding != NULL)
      os << "IS " << binding->getInterfaceName();
    else
    {
      const VrmlField* field = get(d_fields, it->first.c_str());
      if (field) field->print(os);
    }
  }
}


ostream& VrmlNodeScript::print(ostream& os, int indent, bool breakLine) const
{
  // if this node is marked, it means it was already printed in this round;
  // that means a USE reference was encountered
  if (isFlagSet() && validName())
  {
    if (breakLine)
    {
      os << "\n";
      for (int i=0; i<indent; ++i) os << ' ';
    }
    os << "USE " << name();
    return os;
  }

  printComment(os, indent, breakLine);

  // node name (if node has valid name)
  const char *nm = validName();
  if (nm) os << "DEF " << nm << " ";

  os << nodeType().getName() << " {";

  // print fields and events first
  printFields(os, indent);

  // ending brace
  os << "\n";
  for (int j=0; j<indent; ++j) os << ' ';  // indent
  os << "}";

  // mark this node as already printed; if it is encountered next time,
  // it means it is a USEd node and all other references should be printed
  // using USE.
  ((VrmlNode*) this)->setFlag();

  return os;
}

ostream& VrmlNodeScript::printFields(ostream& os, int indent) const
{
  // print user-added fields and events first
  printFieldGroup(os, "eventIn",  d_eventIns,  indent);
  printFieldGroup(os, "field",    d_fields,    indent);
  printFieldGroup(os, "eventOut", d_eventOuts, indent);

  // print fixed fields
  if (d_directOutput.get() || !d_directOutput.getComment().empty())
  {
    d_directOutput.printComment(os, indent+INDENT_INCREMENT, true);
    PRINT_FIELD(directOutput);
  }

  if (d_mustEvaluate.get() || !d_mustEvaluate.getComment().empty())
  {
    d_mustEvaluate.printComment(os, indent+INDENT_INCREMENT, true);
    PRINT_FIELD(mustEvaluate);
  }

  // URL may be very long if it contains the whole script, so print it last
  if (d_url.getLength() > 0 || !d_url.getComment().empty())
  {
    d_url.printComment(os, indent+INDENT_INCREMENT, true);
    PRINT_FIELD(url);
  }

  return os;
}


void VrmlNodeScript::initialize( double ts )
{
  theSystem->debug("Script.%s::initialize\n", name());

  if (d_script) return;	       // How did I get here? Letting the days go by...
  d_eventsReceived = 0;
  if (d_url.getLength() > 0)
    {
      d_script = ScriptObject::create( this, d_url );
      if (d_script)
	d_script->activate( ts, "initialize", 0, 0 );
    }
}

void VrmlNodeScript::shutdown( double ts )
{
  if (d_script)
    d_script->activate( ts, "shutdown", 0, 0 );
}

void VrmlNodeScript::update( VrmlSFTime &timeNow )
{
  if (d_eventsReceived > 0)
    {
      //theSystem->debug("Script.%s::update\n", name());
      if (d_script)
	d_script->activate( timeNow.get(), "eventsProcessed", 0, 0 );
      d_eventsReceived = 0;
    }
}

//

void VrmlNodeScript::eventIn(double timeStamp,
			     const char *eventName,
			     const VrmlField *fieldValue)
{
  if (! d_script ) initialize( timeStamp );
  if (! d_script ) return;

  const char *origEventName = eventName;
  bool valid = hasEventIn( eventName );
  if (! valid && strncmp(eventName, "set_", 4) == 0 )
  {
    eventName += 4;
    valid = hasEventIn( eventName );
  }
  if (valid)
  {
    setEventIn( eventName, fieldValue );

    VrmlSFTime ts( timeStamp );
    const VrmlField *args[] = { fieldValue, &ts };

    d_script->activate( timeStamp, eventName, 2, args );

    // For each modified eventOut, send an event
    for (FieldList::iterator it = d_eventOuts.begin(); it != d_eventOuts.end(); ++it)
      if (it->second.modified)
      {
        // reset modified flag first so we don't call eventOut again on re-enter
        // then call eventOut (which can re-enter this procedure)
        it->second.modified = false;
        eventOut( timeStamp, it->first.c_str(), *(it->second.value) );

	// send the 'xxx_changed' form of the event
        eventOut(timeStamp, (it->first+ "_changed").c_str(), *(it->second.value));

        ++d_eventsReceived;	// call eventsProcessed later
      }

    }

  // Let the generic code handle the rest.
  else
    VrmlNode::eventIn( timeStamp, origEventName, fieldValue );

  // Scripts shouldn't generate redraws.
  clearModified();
}


const char* VrmlNodeScript::addInterfaceItem(const char* accessType, VrmlField::VrmlFieldType type)
{
  char interfaceName[32];
  const char* itemName = 0;
  if (strcmp(accessType, "field")==0)
  {
    snprintf(interfaceName, sizeof(interfaceName), "field_%u", ++d_fieldsSequence);
    addField(interfaceName, type);
    itemName = d_fields.find(interfaceName)->first.c_str();
  }
  else if (strcmp(accessType, "eventIn")==0)
  {
    snprintf(interfaceName, sizeof(interfaceName), "eventIn_%u", ++d_eventInsSequence);
    addEventIn(interfaceName, type);
    itemName = d_eventIns.find(interfaceName)->first.c_str();
  }
  else if (strcmp(accessType, "eventOut")==0)
  {
    snprintf(interfaceName, sizeof(interfaceName), "eventOut_%u", ++d_eventOutsSequence);
    addEventOut(interfaceName, type);
    itemName = d_eventOuts.find(interfaceName)->first.c_str();
  }

  return itemName;
}

void VrmlNodeScript::removeInterfaceItem(const char* name, const char* accessType)
{
  std::string key(name);
  if (strcmp(accessType, "field")==0)
  {
    d_fields.erase(key);
  }
  else if (strcmp(accessType, "eventIn")==0)
  {
    d_eventIns.erase(key);
  }
  else if (strcmp(accessType, "eventOut")==0)
  {
    d_eventOuts.erase(key);
  }
}

bool VrmlNodeScript::renameInterfaceItem(const char* name, const char* newName, const char* accessType)
{
  bool result = false;

  // old and new key
  std::string key(name);
  std::string newKey(newName);

  // field with the name does not exist yet
  FieldList::iterator it;
  ScriptField interfaceItem;
  FieldList* list = 0;
  if (hasInterface(newName)==VrmlField::NO_FIELD)
  {
    if (strcmp(accessType, "field")==0)
    {
      it = d_fields.find(key);
      list = &d_fields;
    }
    else if (strcmp(accessType, "eventIn")==0)
    {
      it = d_eventIns.find(key);
      list = &d_eventIns;
    }
    else if (strcmp(accessType, "eventOut")==0)
    {
      it = d_eventOuts.find(key);
      list = &d_eventOuts;
    }

    if (list!=0 && it!=list->end())
    {
      interfaceItem = it->second;
      list->erase(it);
      (*list)[newKey] = interfaceItem;
      result = true;
    }
  }

  return result;
}


// add events/fields

void VrmlNodeScript::addEventIn(const char* ename, VrmlField::VrmlFieldType t)
{
  add(d_eventIns, ename, t);
}

void VrmlNodeScript::addEventOut(const char* ename, VrmlField::VrmlFieldType t)
{
  add(d_eventOuts, ename, t);
}

void VrmlNodeScript::addField(const char* ename, VrmlField::VrmlFieldType t,
			      const VrmlField* val)
{
  add(d_fields, ename, t);
  if (val) set(d_fields, ename, val);
}

void VrmlNodeScript::add(FieldList& recs,
			 const char* ename,
			 VrmlField::VrmlFieldType type)
{
  recs[ename] = ScriptField(type, VrmlField::create(type));
}

VrmlField* VrmlNodeScript::get(const FieldList& recs, const char* fname) const
{
  FieldList::const_iterator item = recs.find(fname);
  return(item != recs.end() ? item->second.value : 0);
}

// has

VrmlField::VrmlFieldType VrmlNodeScript::hasEventIn(const char *ename) const
{
  VrmlField::VrmlFieldType ft = has(d_eventIns, ename);
  return(ft!=VrmlField::NO_FIELD ? ft : nodeType().hasEventIn(ename));
}

VrmlField::VrmlFieldType VrmlNodeScript::hasEventOut(const char *ename) const
{
  VrmlField::VrmlFieldType ft = has(d_eventOuts, ename);
  return(ft!=VrmlField::NO_FIELD ? ft : nodeType().hasEventOut(ename));
}

VrmlField::VrmlFieldType VrmlNodeScript::hasField(const char *ename) const
{
  // handle fixed fields
  if (strcmp(ename, "url") == 0) return VrmlField::MFSTRING;
  else if (strcmp(ename, "directOutput") == 0) return VrmlField::SFBOOL;
  else if (strcmp(ename, "mustEvaluate") == 0) return VrmlField::SFBOOL;

  // handle additional fields
  else return has(d_fields, ename);
}

VrmlField::VrmlFieldType VrmlNodeScript::hasInterface(const char * id) const
{
  VrmlField::VrmlFieldType fieldType = VrmlField::NO_FIELD;

  if ((fieldType = this->hasField(id)) != VrmlField::NO_FIELD)
    return fieldType;

  if ((fieldType = this->hasEventIn(id)) != VrmlField::NO_FIELD)
      return fieldType;

  if ((fieldType = this->hasEventOut(id)) != VrmlField::NO_FIELD)
      return fieldType;

  return fieldType;
}

VrmlField::VrmlFieldType VrmlNodeScript::has(const FieldList& recs, const char* ename) const
{
  FieldList::const_iterator item = recs.find(ename);
  return(item != recs.end() ? item->second.type : VrmlField::NO_FIELD);
}

// Get the value of a field or eventOut.

const VrmlField *VrmlNodeScript::getField(const char *fieldName) const
{
  // exposedFields
  if ( strcmp( fieldName, "url" ) == 0 )
    return &d_url;

  // fields
  if ( strcmp( fieldName, "directOutput" ) == 0 )
    return &d_directOutput;

  if ( strcmp( fieldName, "mustEvaluate" ) == 0 )
    return &d_mustEvaluate;

  VrmlField* result;
  result = get(d_fields, fieldName);
  if (result != NULL) return result;
  result = get(d_eventOuts, fieldName);
  if (result != NULL) return result;
  result = get(d_eventIns, fieldName);
  if (result != NULL) return result;

  return VrmlNodeChild::getField( fieldName );
}


// Set the value of one of the node fields/events.
// setField is public so the parser can access it.

void VrmlNodeScript::setField(const char *fieldName,
			      const VrmlField &fieldValue)
{
  VrmlField::VrmlFieldType ft;

  if TRY_FIELD(url, MFString)	// need to re-initialize() if url changes...
  else if TRY_FIELD(directOutput, SFBool)
  else if TRY_FIELD(mustEvaluate, SFBool)
  else if ( (ft = hasField(fieldName)) != 0 )
    {
      if (ft == VrmlField::fieldType( fieldValue.fieldTypeName() ))
	set( d_fields, fieldName, &fieldValue );
      else
	theSystem->error(System::ERRORTYPE_UNKNOWN, "Invalid type (%s) for %s field of Script node.\n",
		      fieldValue.fieldTypeName(), fieldName );
    }
  else
    VrmlNodeChild::setField(fieldName, fieldValue);
}

void
VrmlNodeScript::setEventIn(const char *fname, const VrmlField *value)
{
  theSystem->debug("Script: %s, setEventIn: %s\n", name(), fname);
  set(d_eventIns, fname, value);
}

void
VrmlNodeScript::setEventOut(const char *fname, const VrmlField *value)
{
  theSystem->debug("Script: %s, setEventOut: %s\n", name(), fname);
  set(d_eventOuts, fname, value);
}

void VrmlNodeScript::set(FieldList& recs,
                         const char* fname,
                         const VrmlField* value)
{
  FieldList::iterator item = recs.find(fname);
  if (item != recs.end())
  {
    VrmlField* oldvalue = item->second.value;
    if (oldvalue != value)   // protect against setting a field to itself
    {
      item->second.value = value->clone();
      item->second.modified = true;
      delete oldvalue;
    }
  }
}
