/*****************************************************************************
******************************************************************************
*
*               VRML Script node
*
*               $Revision: 1.1.2.6 $
*               $Date: 2011/09/30 05:54:18 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright 1998-2011 HUMUSOFT s.r.o.
// Portions Copyright (C) 1998  Chris Morley

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#ifndef VRMLNODESCRIPT_H__
#define VRMLNODESCRIPT_H__

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string>
#include <map>

#include "VrmlNodeChild.h"
#include "VrmlNodeType.h"
#include "VrmlField.h"

class Doc;
class ScriptObject;
class VrmlScene;


class OPENVRMLDLLEXPORT VrmlNodeScript : public VrmlNodeChild
{
public:

  // Define the fields of Script nodes
  static VrmlNodeType *defineType(VrmlNodeType *t = 0);
  virtual VrmlNodeType & nodeType() const;

  VrmlNodeScript( VrmlScene *scene = 0 );
  VrmlNodeScript( const VrmlNodeScript& other);
  virtual ~VrmlNodeScript();

  virtual VrmlNode *cloneMe() const;
  virtual void cloneChildren(VrmlNamespace*);

  virtual void addToScene( VrmlScene *s, const char *relUrl );

  /** Export node script to stream in simplified hierarchy format.
   * @param os target stream.
   * @param editFlag true if chain is created for 3D World Editor purposes.
   * @return filled stream.
   */
  virtual ostream& toChain(ostream& os, bool editFlag) const;
  /** Export fix fields of the NodeScript into stream in simplified hierarchy
   * @param os target stream.
   * @param editFlag true if chain is created for 3D World Editor purposes.
   * @return filled stream.
   **/
  virtual ostream& fixFieldsToChain(ostream& os, bool editFlag) const;

  virtual ostream& print(ostream &os, int indent, bool breakLine) const;
  virtual ostream& printFields(ostream& os, int indent) const;


  virtual void eventIn(double timeStamp,
		       const char *eventName,
		       const VrmlField *fieldValue);

  virtual const VrmlField *getField(const char *fieldName) const;
  virtual void setField(const char *fieldName, const VrmlField &fieldValue);

  // Script processing methods
  void initialize( double timeStamp );
  void update( VrmlSFTime &now );
  void shutdown( double timeStamp );

  // Methods for adding eventIns/Outs/fields to this script
  void addEventIn(const char *name, VrmlField::VrmlFieldType type);
  void addEventOut(const char *name, VrmlField::VrmlFieldType type);
  void addField(const char *name, VrmlField::VrmlFieldType type,
		const VrmlField * defaultVal = 0);

  // methods for manipulating with Script node interface
  const char* addInterfaceItem(const char* accessType, VrmlField::VrmlFieldType type);
  void removeInterfaceItem(const char* name, const char* accessType);
  bool renameInterfaceItem(const char* name, const char* newName, const char* accessType);

  // Access to eventIns/Outs/fields for ScriptObjects

  // Per-script field type
  struct ScriptField
  {
    VrmlField* value;
    VrmlField::VrmlFieldType type;
    bool modified;

    ScriptField();
    ScriptField(VrmlField::VrmlFieldType type_, VrmlField* value_);
    ScriptField(const ScriptField& other);
    ~ScriptField();

    ScriptField& operator=(const ScriptField& other);
  };

  // Tests for specific fields/events
  VrmlField::VrmlFieldType hasEventIn(const char *name) const;
  VrmlField::VrmlFieldType hasEventOut(const char *name) const;
  VrmlField::VrmlFieldType hasField(const char *name) const;
  VrmlField::VrmlFieldType hasInterface(char const *) const;

  // Set field/event values
  void setEventIn(const char *, const VrmlField *);
  void setEventOut(const char *, const VrmlField *);
  // setField declared above as virtual

  // container which holds fields
  typedef std::map<std::string, ScriptField> FieldList;

  // Fields and events defined for this Script
  const FieldList& eventIns() const
  { return d_eventIns; }
  const FieldList& eventOuts() const
  { return d_eventOuts; }
  const FieldList& fields() const
  { return d_fields; }

  // Access to browser functions for ScriptObjects
  VrmlScene *browser() { return d_scene; }

private:

  // Fields
  VrmlSFBool d_directOutput;
  VrmlSFBool d_mustEvaluate;
  VrmlMFString d_url;

  VrmlSFString d_relativeUrl;

  // The script language-dependent part
  ScriptObject *d_script;

  // Fields and events defined for this Script
  FieldList d_eventIns;
  FieldList d_eventOuts;
  FieldList d_fields;

  // sequences for generation of appropriate synthetic/gerneric names
  size_t d_eventInsSequence, d_eventOutsSequence, d_fieldsSequence;

  // Generic field/event add/test/value methods
  void add(FieldList&, const char*, VrmlField::VrmlFieldType);
  VrmlField::VrmlFieldType has(const FieldList&, const char*) const;
  VrmlField* get(const FieldList&, const char*) const;
  void set(FieldList &, const char*, const VrmlField*);

  // Print helper
  void printFieldGroup(ostream& os, const char* fieldclass, const FieldList& recs, int indent) const;

  int d_eventsReceived;
};

#endif // VRMLNODESCRIPT_H__
