/*****************************************************************************
******************************************************************************
*
*               VRML Shape node
*
*               $Revision: 1.1.2.4 $
*               $Date: 2011/07/19 14:53:29 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright (C) 1998  Chris Morley
// Modified by HUMUSOFT s.r.o.  Copyright 1998-2011 HUMUSOFT s.r.o.

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#include "VrmlNodeShape.h"
#include "VrmlNodeAppearance.h"
#include "VrmlNodeGeometry.h"
#include "VrmlNodeProto.h"
#include "VrmlNodeTexture.h"
#include "VrmlNodeType.h"
#include "Viewer.h"
#include "VrmlBVolume.h"
#include "VrmlNodeMaterial.h"

#include "VrmlScene.h"

static VrmlNode *creator( VrmlScene *s ) { return new VrmlNodeShape(s); }


// Define the built in VrmlNodeType:: "Shape" fields

VrmlNodeType *VrmlNodeShape::defineType(VrmlNodeType *t)
{
  static VrmlNodeType *st = 0;

  if (! t)
    {
      if (st) return st;
      t = st = new VrmlNodeType("Shape", creator);
      t->reference();
    }

  VrmlNodeChild::defineType(t);	// Parent class
  t->addExposedField("appearance", VrmlField::SFNODE, VrmlSFNode());
  t->addExposedField("geometry", VrmlField::SFNODE, VrmlSFNode());

  return t;
}


VrmlNodeType & VrmlNodeShape::nodeType() const
{
    return *defineType(0);
}


VrmlNodeShape::VrmlNodeShape(VrmlScene *scene) :
  VrmlNodeChild(scene),
  d_viewerObject(0)
{
}

VrmlNodeShape::~VrmlNodeShape()
{
  // delete precompiled viewerObject
  if (d_viewerObject && d_scene)
  {
    Viewer* primary = d_scene->getPrimaryViewer();
    if (primary) primary->removeObject(d_viewerObject);
  }
}


VrmlNode *VrmlNodeShape::cloneMe() const
{
  return new VrmlNodeShape(*this);
}

void VrmlNodeShape::cloneChildren(VrmlNamespace *ns)
{
  if (d_appearance.get())
    d_appearance.set(d_appearance.get()->clone(ns));
  if (d_geometry.get())
    d_geometry.set(d_geometry.get()->clone(ns));
}


bool VrmlNodeShape::isModified() const
{
  return ( VrmlNodeChild::isModified() ||
	   (d_geometry.get() && d_geometry.get()->isModified()) ||
	   (d_appearance.get() && d_appearance.get()->isModified()) );
}

void VrmlNodeShape::updateModified(VrmlNodePath& path, int flags)
{
  //cout << "VrmlNodeShape[" << this << "]::updateModified()" << endl;
  if (this->isModified()) markPathModified(path, true, flags);
  path.push_front(this);
  if (d_appearance.get()) d_appearance.get()->updateModified(path, flags);
  if (d_geometry.get()) d_geometry.get()->updateModified(path, flags);
  path.pop_front();
}

void VrmlNodeShape::clearFlags()
{
  VrmlNode::clearFlags();
  if (d_appearance.get()) d_appearance.get()->clearFlags();
  if (d_geometry.get()) d_geometry.get()->clearFlags();
}

void VrmlNodeShape::addToScene(VrmlScene *s, const char *relUrl)
{
  d_scene = s;
  if (d_appearance.get()) d_appearance.get()->addToScene(s,relUrl);
  if (d_geometry.get()) d_geometry.get()->addToScene(s,relUrl);
}

ostream& VrmlNodeShape::printFields(ostream& os, int indent)
{
  if (d_appearance.get()) PRINT_FIELD(appearance);
  if (d_geometry.get()) PRINT_FIELD(geometry);

  return os;
}

void VrmlNodeShape::render(Viewer *viewer, VrmlRenderContext rc)
{
  //cout << "VrmlNodeShape::render()" << endl;

  if (viewer->getRenderPhase()==Viewer::RENDER_PHASE_LIGHTS)
    return;

  VrmlNodeGeometry* g = NodeDynamicCast<VrmlNodeGeometry*>(d_geometry.get());
  VrmlNodeAppearance* a = NodeDynamicCast<VrmlNodeAppearance*>(d_appearance.get());

  float trans = 0.0;
  if (a)
  {
    const VrmlSFNode* sfnval = dynamic_cast<const VrmlSFNode*>(a->getField("material"));
    VrmlNodeMaterial *m = sfnval ? NodeDynamicCast<VrmlNodeMaterial*>(sfnval->get()) : 0;
    if (m) trans = m->transparency();
  }

  // transparent objects must not be drawn in non-transparent phase
  if (trans>0.0)
  {
    if (viewer->getRenderPhase()==Viewer::RENDER_PHASE_NON_TRANSPARENT)
      return;
  }

  // non-transparent objects must not be drawn in transparent phase
  else
  {
    if (viewer->getRenderPhase()==Viewer::RENDER_PHASE_TRANSPARENT) return;
  }

  if ( d_viewerObject && isModified() )
    {
      viewer->removeObject(d_viewerObject);
      d_viewerObject = 0;
    }

  if (d_viewerObject)
    viewer->insertReference(d_viewerObject);

  else if ( g )
    {
      d_viewerObject = viewer->beginObject( name() );

      int nTexComponents = 0;

      VrmlNodeAppearance* a = NodeDynamicCast<VrmlNodeAppearance*>(d_appearance.get());
      if (a != NULL)
      {
        a->render(viewer, rc);

	VrmlNodeTexture* texturen = NodeDynamicCast<VrmlNodeTexture*>(a->texture());
        if (texturen != NULL)
          nTexComponents = texturen->nComponents();
      }
      else
        {
          viewer->setColor(1.0, 1.0, 1.0); // default object color
          viewer->enableLighting(false);  // turn lighting off
        }

      // hack for opengl material mode
      viewer->setMaterialMode( nTexComponents, g->color() != 0 );

      g->render(viewer, rc);

      viewer->endObject();
    }

  else if (d_appearance.get())
    d_appearance.get()->clearModified();

  clearModified();
}

void VrmlNodeShape::invalidate()
{
  VrmlNodeGeometry* g = NodeDynamicCast<VrmlNodeGeometry*>(d_geometry.get());
  VrmlNodeAppearance* a = NodeDynamicCast<VrmlNodeAppearance*>(d_appearance.get());

  if (g) g->invalidate();
  if (a) a->invalidate();

  if (d_appearance.get())
    d_appearance.get()->invalidate();

  VrmlNode::invalidate();
}

// Get the value of a field or eventOut.

const VrmlField *VrmlNodeShape::getField(const char *fieldName) const
{
  // exposedFields
  if ( strcmp( fieldName, "appearance" ) == 0 )
    return &d_appearance;
  else if ( strcmp( fieldName, "geometry" ) == 0 )
    return &d_geometry;

  return VrmlNodeChild::getField(fieldName); // Parent class
}

// Set the value of one of the node fields.

void VrmlNodeShape::setField(const char *fieldName,
			     const VrmlField &fieldValue)
{
  if TRY_SFNODE_FIELD(appearance, Appearance)
  else if TRY_SFNODE_FIELD(geometry, Geometry)
  else
    VrmlNodeChild::setField(fieldName, fieldValue);
  this->setBVolumeDirty(true); // case out: just if geom set!
}


// just pass off to the geometry's getbvolume() method
//
const VrmlBVolume* VrmlNodeShape::getBVolume() const
{
  //cout << "VrmlNodeShape::getBVolume() {" << endl;
  const VrmlBVolume* r = (VrmlBVolume*)0;
  VrmlNode* geom = d_geometry.get();
  if (geom) {
    r = geom->getBVolume();
  }
  //cout << "}:" << r << ":";
  //if (r) r->dump(cout);
  //cout << endl;
  ((VrmlNodeShape*)this)->setBVolumeDirty(false);
  return r;
}

void VrmlNodeShape::setProto(VrmlNodeProto* proto)
{
  VrmlNodeGeometry* g = NodeDynamicCast<VrmlNodeGeometry*>(d_geometry.get());
  VrmlNodeAppearance* a = NodeDynamicCast<VrmlNodeAppearance*>(d_appearance.get());

  if (g) g->setProto(proto);
  if (a) a->setProto(proto);

  if (d_appearance.get())
    d_appearance.get()->setProto(proto);

  VrmlNodeChild::setProto(proto);
}

void VrmlNodeShape::setEncapsulatingPrototype(VrmlNodeType* type)
{
  VrmlNodeGeometry* g = NodeDynamicCast<VrmlNodeGeometry*>(d_geometry.get());
  VrmlNodeAppearance* a = NodeDynamicCast<VrmlNodeAppearance*>(d_appearance.get());

  if (g) g->setEncapsulatingPrototype(type);
  if (a) a->setEncapsulatingPrototype(type);

  if (d_appearance.get())
    d_appearance.get()->setEncapsulatingPrototype(type);

  VrmlNodeChild::setEncapsulatingPrototype(type);
}
