/*****************************************************************************
******************************************************************************
*
*               VRML Text node
*
*               $Revision: 1.1.2.3 $
*               $Date: 2011/07/19 14:53:30 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright (C) 1998  Chris Morley
// Modified by HUMUSOFT s.r.o.  Copyright 1998-2011 HUMUSOFT s.r.o.

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#include "VrmlNodeText.h"
#include "VrmlNodeType.h"
#include "VrmlNodeFontStyle.h"
#include "VrmlNodeProto.h"
#include "MathUtils.h"
#include "Viewer.h"

// Make a VrmlNodeText

static VrmlNode *creator( VrmlScene *s ) { return new VrmlNodeText(s); }


// Define the built in VrmlNodeType:: "Text" fields

VrmlNodeType *VrmlNodeText::defineType(VrmlNodeType *t)
{
  static VrmlNodeType *st = 0;

  if (! t)
    {
      if (st) return st;	// Define type only once.
      t = st = new VrmlNodeType("Text", creator);
      t->reference();
    }

  VrmlNodeGeometry::defineType(t);	// Parent class
  t->addExposedField("string", VrmlField::MFSTRING, VrmlMFString());
  t->addExposedField("fontStyle", VrmlField::SFNODE, VrmlSFNode());
  t->addExposedField("length", VrmlField::MFFLOAT, VrmlMFFloat());
  t->addExposedField("maxExtent", VrmlField::SFFLOAT, VrmlSFFloat(0.0));

  return t;
}

VrmlNodeType & VrmlNodeText::nodeType() const
{
    return *defineType(0);
}


VrmlNodeText::VrmlNodeText(VrmlScene *scene) :
  VrmlNodeGeometry(scene)
{
}

VrmlNodeText::~VrmlNodeText()
{
}


VrmlNode *VrmlNodeText::cloneMe() const
{
  return new VrmlNodeText(*this);
}

void VrmlNodeText::cloneChildren(VrmlNamespace *ns)
{
  if (d_fontStyle.get())
    d_fontStyle.set(d_fontStyle.get()->clone(ns));
}

bool VrmlNodeText::isModified() const
{
  return (VrmlNode::isModified() ||
         (d_fontStyle.get() && d_fontStyle.get()->isModified()));
}

void VrmlNodeText::updateModified(VrmlNodePath& path)
{
  if (this->isModified()) markPathModified(path, true);
  path.push_front(this);
  if (d_fontStyle.get()) d_fontStyle.get()->updateModified(path);
  path.pop_front();
}


void VrmlNodeText::clearFlags()
{
  VrmlNode::clearFlags();
  if (d_fontStyle.get()) d_fontStyle.get()->clearFlags();
}

void VrmlNodeText::addToScene( VrmlScene *s, const char *relUrl )
{
  d_scene = s;
  if (d_fontStyle.get()) d_fontStyle.get()->addToScene(s, relUrl);
}

ostream& VrmlNodeText::printFields(ostream& os, int indent)
{
  if (d_string.getLength() > 0) PRINT_FIELD(string);
  if (d_fontStyle.get()) PRINT_FIELD(fontStyle);
  if (d_length.getLength() > 0) PRINT_FIELD(length);
  if (! FPZERO(d_maxExtent.get()) ) PRINT_FIELD(maxExtent);
  return os;
}


Viewer::Object VrmlNodeText::insertGeometry(Viewer *viewer, VrmlRenderContext rc)
{
    char const * const * s = d_string.get();

    if (s) {
        int justify[2] = { 1, 1 };
        float size = 1.0;
        float spacing = 1.0;
        VrmlNodeFontStyle* f = NodeDynamicCast<VrmlNodeFontStyle*>(d_fontStyle.get());

        // load font properties from the FontStyle node
        Viewer::FontFamily family = Viewer::FONT_FAMILY_SERIF;
        Viewer::FontStyle style = Viewer::FONT_STYLE_PLAIN;
        if (f) {

            // justification
            VrmlMFString & j = f->justify();
            size_t i;
            for (i=0; i<j.getLength(); ++i) {
                if (strcmp(j.get(i), "END") == 0) {
                    justify[i] = -1;
                } else if (strcmp(j.get(i), "MIDDLE") == 0) {
                    justify[i] = 0;
                }
            }
            size = f->size();
            spacing = f->spacing();

            // font family
            const VrmlMFString* fam = dynamic_cast<const VrmlMFString*>(f->getField("family"));
            for (i=0; i<fam->getLength(); ++i)
            {
              if (strcmp(fam->get(i), "SANS") == 0)
                family = Viewer::FONT_FAMILY_SANS;
              else if (strcmp(fam->get(i), "TYPEWRITER") == 0)
                family = Viewer::FONT_FAMILY_TYPEWRITER;
            }

            // font style
            const VrmlSFString* st = dynamic_cast<const VrmlSFString*>(f->getField("style"));
            if (strcmp(st->get(), "BOLD") == 0)
              style = Viewer::FONT_STYLE_BOLD;
            else if (strcmp(st->get(), "ITALIC") == 0)
              style = Viewer::FONT_STYLE_ITALIC;
            else if (strcmp(st->get(), "BOLDITALIC") == 0)
              style = Viewer::FONT_STYLE_BOLD_ITALIC;
        }

        return viewer->insertText(family, style, justify, size, spacing, d_string.getLength(), s);
    }

    return 0;
}

// Get the value of a field or eventOut.

const VrmlField *VrmlNodeText::getField(const char *fieldName) const
{
  // exposedFields
  if ( strcmp( fieldName, "string" ) == 0 )
    return &d_string;
  else if ( strcmp( fieldName, "fontStyle" ) == 0 )
    return &d_fontStyle;
  else if ( strcmp( fieldName, "length" ) == 0 )
    return &d_length;
  else if ( strcmp( fieldName, "maxExtent" ) == 0 )
    return &d_maxExtent;

  return VrmlNodeGeometry::getField(fieldName); // Parent class
}

// Set the value of one of the node fields.

void VrmlNodeText::setField(const char *fieldName,
			    const VrmlField &fieldValue)
{
  if TRY_FIELD(string, MFString)
  else if TRY_SFNODE_FIELD(fontStyle, FontStyle)
  else if TRY_FIELD(length, MFFloat)
  else if TRY_FIELD(maxExtent, SFFloat)
  else
    VrmlNodeGeometry::setField(fieldName, fieldValue);
}

void VrmlNodeText::setProto(VrmlNodeProto* proto)
{
  if (d_fontStyle.get())
    d_fontStyle.get()->setProto(proto);

  VrmlNodeGeometry::setProto(proto);
}

void VrmlNodeText::setEncapsulatingPrototype(VrmlNodeType* type)
{
  if (d_fontStyle.get())
    d_fontStyle.get()->setEncapsulatingPrototype(type);

  VrmlNodeGeometry::setEncapsulatingPrototype(type);
}
