/*****************************************************************************
******************************************************************************
*
*               VRML TouchSensor node
*
*               $Revision: 1.1.2.2 $
*               $Date: 2009/11/02 14:49:53 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright (C) 1998  Chris Morley
// Modified by HUMUSOFT s.r.o.  Copyright 1998-2009 HUMUSOFT s.r.o.

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "VrmlNodeTouchSensor.h"
#include "VrmlNodeType.h"
#include "MathUtils.h"
#include "System.h"
#include "Viewer.h"
#include "VrmlScene.h"

// TouchSensor factory.

static VrmlNode *creator( VrmlScene *scene )
{
  return new VrmlNodeTouchSensor(scene);
}


// Define the built in VrmlNodeType:: "TouchSensor" fields

VrmlNodeType *VrmlNodeTouchSensor::defineType(VrmlNodeType *t)
{
  static VrmlNodeType *st = 0;

  if (! t)
    {
      if (st) return st;		// Only define the type once.
      t = st = new VrmlNodeType("TouchSensor", creator);
      t->reference();
    }

  VrmlNodeChild::defineType(t);	// Parent class
  t->addExposedField("enabled", VrmlField::SFBOOL, VrmlSFBool(true));
  t->addEventOut("hitNormal_changed", VrmlField::SFVEC3F);
  t->addEventOut("hitPoint_changed", VrmlField::SFVEC3F);
  t->addEventOut("hitTexCoord_changed", VrmlField::SFVEC2F);
  t->addEventOut("isActive", VrmlField::SFBOOL);
  t->addEventOut("isOver", VrmlField::SFBOOL);
  t->addEventOut("touchTime", VrmlField::SFTIME);

  return t;
}

VrmlNodeType & VrmlNodeTouchSensor::nodeType() const
{
    return *defineType(0);
}


VrmlNodeTouchSensor::VrmlNodeTouchSensor( VrmlScene *scene ) :
  VrmlNodeChild(scene),
  d_enabled(true),
  d_isActive(false),
  d_isOver(false),
  d_touchTime(0.0),
  d_parentTransform(0)
{
  setModified();
}


VrmlNodeTouchSensor::~VrmlNodeTouchSensor()
{
}


VrmlNode *VrmlNodeTouchSensor::cloneMe() const
{
  return new VrmlNodeTouchSensor(*this);
}


ostream& VrmlNodeTouchSensor::printFields(ostream& os, int indent)
{
  if (! d_enabled.get())
    PRINT_FIELD(enabled);
  return os;
}

// Cache a pointer to (one of the) parent transforms for converting
// hits into local coords.

void VrmlNodeTouchSensor::accumulateTransform( VrmlNode *parent )
{
  d_parentTransform = parent;
}

VrmlNode* VrmlNodeTouchSensor::getParentTransform() { return d_parentTransform; }

void VrmlNodeTouchSensor::activate(double timeStamp,
    bool isOver, bool isActive, double *touchPos)
{
  if (isOver && !isActive && d_isActive.get())
    {
      d_touchTime.set(timeStamp);
      eventOut( timeStamp, "touchTime", d_touchTime);
      //theSystem->debug("TouchSensor.%s touchTime\n", name());
    }

  if (isOver != d_isOver.get())
    {
      d_isOver.set(isOver);
      eventOut( timeStamp, "isOver", d_isOver );
    }

  if (isActive != d_isActive.get())
    {
      d_isActive.set(isActive);
      eventOut( timeStamp, "isActive", d_isActive );
    }

  // if (isOver && any routes from eventOuts)
  //   generate xxx_changed eventOuts...

#if defined(__ORBISNAP__)

    Matrix4x4 m = getCompoundInverseTransformationMatrix();
    Point3 point(touchPos[0], touchPos[1], touchPos[2]);
    Point3 transformedPoint = m*point;
    float hit[3];
    hit[0] = transformedPoint.getX();
    hit[1] = transformedPoint.getY();
    hit[2] = transformedPoint.getZ();
    d_hitPoint.set(hit);
    eventOut(timeStamp, "hitPoint_changed", d_hitPoint);

    // TO DO: hitTexCoord_changed, hitNormal_changed

#endif

}



VrmlField *VrmlNodeTouchSensor::findField(const char *fieldName)
{
  if      (strcmp(fieldName, "enabled") == 0)         return &d_enabled;
  else if (strcmp(fieldName, "hitNormal") == 0)       return &d_hitNormal;
  else if (strcmp(fieldName, "hitPoint") == 0)        return &d_hitPoint;
  else if (strcmp(fieldName, "hitTexCoord") == 0)     return &d_hitTexCoord;
  else if (strcmp(fieldName, "isActive") == 0)        return &d_isActive;
  else if (strcmp(fieldName, "isOver") == 0)          return &d_isOver;
  else if (strcmp(fieldName, "touchTime") == 0)       return &d_touchTime;
  else                                                return NULL;
}
