/*****************************************************************************
******************************************************************************
*
*               VRML node type class
*
*               $Revision: 1.1.4.6.2.13 $
*               $Date: 2011/08/08 12:08:12 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright 1998-2011 HUMUSOFT s.r.o.
// Portions Copyright (C) 1998  Chris Morley

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


// The VrmlNodeType class is responsible for storing information about node
// or prototype types.
//

#include "VrmlNodeType.h"
#include "VrmlField.h"
#include "VrmlNode.h"
#include "VrmlNamespace.h"
#include "Doc.h"
#include "VrmlNodeProto.h"	// Instances of PROTOtyped nodes
#include "VrmlScene.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <cstdlib>

#define TERMINAL_STRING "#"

size_t VrmlNodeType::d_uniqueNameSequence = 0;

VrmlNodeType::NodeFieldRec::NodeFieldRec() :
  fieldName(0),
  node(0)
{ }

VrmlNodeType::NodeFieldRec::NodeFieldRec(const VrmlNodeType::NodeFieldRec& source, VrmlNamespace* ns)
{
  //;!!	do we need to assign existing node?
  node = (source.node!=NULL) ? ns->findNode(source.node->name()) : 0;

  // copy fieldname
  if (source.fieldName!=NULL)
  {
    size_t len = strlen(source.fieldName);
    fieldName = new char[len + 1];
    memcpy(fieldName, source.fieldName, len + 1);
  }
}

VrmlNodeType::ProtoField::ProtoField()  :
  VrmlObject(),
  name(0),
  type(VrmlField::NO_FIELD),
  defaultValue(0)
{ }

VrmlNodeType::ProtoField::ProtoField(const VrmlNodeType::ProtoField& source, VrmlNamespace* ns) :
  VrmlObject(source),
  name(0),
  type(source.type),
  defaultValue(VrmlField::create(source.type))
{
  // copy fieldname
  if (source.name!=NULL)
  {
    size_t len = strlen(source.name);
    name = new char[len + 1];
    memcpy(name, source.name, len + 1);
  }

  // clone field
  if (source.defaultValue!=NULL)
    defaultValue = source.defaultValue->clone(ns);

  // create copy
  VrmlNodeType::ISMap::const_iterator j;
  for (j = source.thisIS.begin(); j != source.thisIS.end(); ++j)
  {
    VrmlNodeType::NodeFieldRec* nf = new VrmlNodeType::NodeFieldRec(**j, ns);
    thisIS.push_front(nf);
  }
}

VrmlNodeType::VrmlNodeType(const char *nm, VrmlNode* (*creator)(VrmlScene *)) :
  d_refCount(0),
  d_namespace(0),
  d_url(0),
  d_actualUrl(0),
  d_relative(0),
  d_creator(creator),
  d_implementation(0),
  d_fieldsInitialized(false),
  d_eventInsSequence(0), d_eventOutsSequence(0), d_fieldsSequence(0), d_exposedFieldsSequence(0),
  d_modified(false)
{
  if (nm)
    d_name = strdup(nm);
  else
  {
    char nodeTypeName[32];
    snprintf(nodeTypeName, sizeof(nodeTypeName), "PROTO_%u", ++(VrmlNodeType::d_uniqueNameSequence));
    d_name = strdup(nodeTypeName);
  }
}


VrmlNodeType::VrmlNodeType(const VrmlNodeType& type, VrmlNamespace* parentNamespace) :
  d_refCount(0),
  d_namespace(0),
  d_url(0),
  d_actualUrl(0),
  d_relative(0),
  d_creator(type.d_creator),
  d_implementation(0),
  d_fieldsInitialized(false),

  d_eventInsSequence(type.d_eventInsSequence),
  d_eventOutsSequence(type.d_eventOutsSequence),
  d_fieldsSequence(type.d_fieldsSequence),
  d_exposedFieldsSequence(type.d_exposedFieldsSequence),
  d_modified(false)
{
  // create new namespace
  d_namespace = new VrmlNamespace(parentNamespace);

  // we need to copy VrmlNamespace inner types
  const list<VrmlNodeType*>* typeList = type.d_namespace->getTypeList();
  if (typeList!=NULL)
  {
    list<VrmlNodeType*>::const_iterator i;
    for (i = typeList->begin(); i != typeList->end(); ++i)
    {
      VrmlNodeType* newType = new VrmlNodeType(**i);
      d_namespace->addNodeType(newType);
    }
  }

  // clone implementation nodes
  if (type.d_implementation!=NULL)
  {
    VrmlMFNode* mfnode = type.d_implementation->getNodes();
    for (int i=0; i<mfnode->getLength(); ++i)
      ((*mfnode)[i])->clearFlags();

    d_implementation = dynamic_cast<VrmlNodeGroup*> (type.d_implementation->clone(d_namespace));
    d_implementation->setEncapsulatingPrototype(this);
  }

  if (type.d_namespace!=NULL)
  {
    // copy routes
    const RouteList& rl = type.d_namespace->getRouteList();
    for (RouteList::const_iterator r=rl.begin(); r!=rl.end(); ++r)
    {
      VrmlNode* fromNode = d_namespace->findNode(r->second.fromNode()->name());
      VrmlNode* toNode = d_namespace->findNode(r->second.toNode()->name());
      if (fromNode!=NULL && toNode!=NULL)
      {
	d_namespace->addRoute(fromNode, r->second.fromEventOut(), toNode, r->second.toEventIn());
      }
    }
  }

  // (deep) copy containers
  FieldList::const_iterator i;
  for (i = type.d_fields.begin(); i != type.d_fields.end(); ++i)
  {
    ProtoField* newField = new ProtoField(**i, d_namespace);
    d_fields.push_front(newField);
  }

  for (i = type.d_eventIns.begin(); i != type.d_eventIns.end(); ++i)
  {
    ProtoField* newField = new ProtoField(**i, d_namespace);
    d_eventIns.push_front(newField);
  }

  for (i = type.d_eventOuts.begin(); i != type.d_eventOuts.end(); ++i)
  {
    ProtoField* newField = new ProtoField(**i, d_namespace);
    d_eventOuts.push_front(newField);
  }

  // copy d_name
  d_name = strdup(type.d_name);

  // copy d_url
  if (type.d_url!=NULL)
    d_url = new VrmlMFString(*(type.d_url));

  // copy d_actualUrl
  if (type.d_actualUrl!=NULL)
    d_actualUrl = new VrmlSFString(*(type.d_actualUrl));

  // copy d_relative
  if (type.d_relative!=NULL)
    d_relative = new Doc(type.d_relative);
}

void destructFieldList( VrmlNodeType::FieldList &f )
{
  VrmlNodeType::FieldList::iterator i;
  for (i = f.begin(); i != f.end(); ++i) {
    VrmlNodeType::ProtoField *r = *i;
    free(r->name);
    if (r->defaultValue) delete r->defaultValue;

    // free NodeFieldRec* s in r->thisIS;
    VrmlNodeType::ISMap::const_iterator j;
    for (j = r->thisIS.begin(); j != r->thisIS.end(); ++j)
    {
      VrmlNodeType::NodeFieldRec *nf = *j;
      free(nf->fieldName);
      delete nf;
    }

    delete r;
  }
}

void destructFieldListItem( VrmlNodeType::FieldList &f, const char* itemName, bool propagateToISMappings=false )
{
  VrmlNodeType::FieldList::iterator i = f.begin();
  while (i != f.end())
  {
    if (strcmp((*i)->name, itemName)==0)
    {
      free((*i)->name);
      if ((*i)->defaultValue) delete (*i)->defaultValue;

      // free NodeFieldRec* s in r->thisIS;
      VrmlNodeType::ISMap::const_iterator j;
      for (j = (*i)->thisIS.begin(); j != (*i)->thisIS.end(); ++j)
      {
	VrmlNodeType::NodeFieldRec *nf = *j;
	if (propagateToISMappings) nf->node->deleteISBinding(nf->fieldName);
	free(nf->fieldName);
	delete nf;
      }
      delete (*i);
      i = f.erase(i);
    }
    else
      ++i;
  }
}

void renameFieldListItem( VrmlNodeType::FieldList &f, const char* itemName, const char* itemNewName, bool propagateToISMappings=false )
{
  VrmlNodeType::FieldList::iterator i = f.begin();
  while (i!=f.end())
  {
    if (strcmp((*i)->name, itemName)==0)
    {
      free((*i)->name);
      (*i)->name = strdup(itemNewName);

      // propagate change to nodes
      VrmlNodeType::ISMap::const_iterator j;
      for (j = (*i)->thisIS.begin(); j != (*i)->thisIS.end(); ++j)
      {
	VrmlNodeType::NodeFieldRec *nf = *j;
	if (propagateToISMappings)
	{
	  nf->node->deleteISBinding(nf->fieldName);
	  nf->node->addISBinding(nf->fieldName, (*i)->name);
	}
      }
    }
    ++i;
  }
}

VrmlNodeType::~VrmlNodeType()
{
  if (d_implementation) d_implementation->setEncapsulatingPrototype(0);

  // delete the implementation nodes
  delete d_implementation;

  free(d_name);

  // Free strings & defaults duplicated when fields/eventIns/eventOuts added:
  destructFieldList( d_eventIns );
  destructFieldList( d_eventOuts );
  destructFieldList( d_fields );

  delete d_namespace;
  delete d_url;
  delete d_actualUrl;
  delete d_relative;
}

void VrmlNodeType::dereference()
{
  --d_refCount;
  if (d_refCount==1) d_modified = false;
  if (d_refCount==0) delete this;

}

void VrmlNodeType::setModified()
{
  if (d_refCount<=1 || !d_fieldsInitialized) return;
  d_modified = true;
}

void VrmlNodeType::rename(const char* newname)
{
  free(d_name);
  d_name = strdup(newname);
}

void VrmlNodeType::setScope(VrmlNamespace & createdIn)
{
  d_namespace = new VrmlNamespace(&createdIn);
}


void VrmlNodeType::setUrl(VrmlMFString * url, Doc const * relative)
{
  // delete url
  delete d_url;
  d_url = url;

  // delete implementation nodes
  delete d_implementation;
  d_implementation = 0;

  // relative and d_relative can be the same
  Doc* thisRelative = d_relative;
  // create new relative
  d_relative = relative ? new Doc(relative) : 0;
  // delete old one
  delete thisRelative;
}

void VrmlNodeType::setActualUrl(const char *url)
{
  if (d_actualUrl) delete d_actualUrl;
  d_actualUrl = url ? new VrmlSFString(url): 0;
}

VrmlNodeGroup* VrmlNodeType::nodes()
{
  // nodes requested -- if it is not EXTERNPROTO and nodes do not exist yet, create them
  if (!d_url && !d_implementation)
  {
    d_implementation = new VrmlNodeGroup();
    d_implementation->setName(0, d_namespace);
    d_implementation->setEncapsulatingPrototype(this);
  }

  return d_implementation;
}


const char* VrmlNodeType::addEventIn(const char *ename, VrmlField::VrmlFieldType type)
{
  char interfaceName[32];
  if (ename==0)
    do {
      snprintf(interfaceName, sizeof(interfaceName), "eventIn_%u", ++d_eventInsSequence);
      ename = interfaceName;
    } while (hasInterface(ename));

  add(d_eventIns, ename, type);
  // event in added to the front of the container
  return d_eventIns.front()->name;
}


const char* VrmlNodeType::addEventOut(const char *ename, VrmlField::VrmlFieldType type)
{
  char interfaceName[32];
  if (ename==0)
    do {
      snprintf(interfaceName, sizeof(interfaceName), "eventOut_%u", ++d_eventOutsSequence);
      ename = interfaceName;
    } while (hasInterface(ename));

  add(d_eventOuts, ename, type);
  // event out added to the front of the container
  return d_eventOuts.front()->name;
}


const char* VrmlNodeType::addField(const char *ename, VrmlField::VrmlFieldType type)
{
  char interfaceName[32];
  if (ename==0)
    do {
      snprintf(interfaceName, sizeof(interfaceName), "field_%u", ++d_fieldsSequence);
      ename = interfaceName;
    } while (hasInterface(ename));

  add(d_fields, ename, type);

  // set an initial default value (may be changed later)
  VrmlField* defval = VrmlField::create(type);
  setFieldDefault(ename, defval);
  delete defval;
  // field added to the front of the container
  return d_fields.front()->name;
}

void
VrmlNodeType::addField(const char *name,
                       VrmlField::VrmlFieldType type,
                       const VrmlField& defaultValue)
{
  add(d_fields, name, type);
  setFieldDefault(name, &defaultValue);
}


const char* VrmlNodeType::addExposedField(const char *ename,
                              VrmlField::VrmlFieldType type)
{
  char interfaceName[32];
  if (ename==0)

    do {
      snprintf(interfaceName, sizeof(interfaceName), "exposedField_%u", ++d_exposedFieldsSequence);
      ename = interfaceName;
    } while (hasInterface(ename));

  // add it as a common field
  addField(ename, type);

  // add appropriate events
  char tmp[256];
  snprintf(tmp, sizeof(tmp), "set_%s", ename);
  add(d_eventIns, tmp, type);
  snprintf(tmp, sizeof(tmp), "%s_changed", ename);
  add(d_eventOuts, tmp, type);

  // exposed field added to the front of the container
  return d_fields.front()->name;
}


void
VrmlNodeType::addExposedField(const char* ename,
                              VrmlField::VrmlFieldType type,
                              const VrmlField& defaultValue)
{
  addExposedField(ename, type);
  setFieldDefault(ename, &defaultValue);
}



void
VrmlNodeType::add(FieldList &recs, const char *ename, VrmlField::VrmlFieldType type)
{
  ProtoField *r = new ProtoField;
  r->name = strdup(ename);
  r->type = type;
  r->defaultValue = 0;
  recs.push_front(r);
  setModified();
}

void
VrmlNodeType::setFieldDefault(const char *fname, VrmlField const * defaultValue)
{
  FieldList::const_iterator i;

  for (i = d_fields.begin(); i != d_fields.end(); ++i)
    if (strcmp((*i)->name, fname) == 0)
      {
	if ((*i)->defaultValue) delete (*i)->defaultValue;
	(*i)->defaultValue = defaultValue ? defaultValue->clone() : 0;
	return;
      }

  theSystem->error(System::ERRORTYPE_UNKNOWN, "setFieldDefault for field %s of %s failed: no such field.",
		   fname, getName());
}


// Download the EXTERNPROTO definition

void VrmlNodeType::fetchImplementation()
{
  // Get the PROTO def from the url (relative to original scene url).
  VrmlNodeType *proto = VrmlScene::readPROTO( d_url, d_relative );
  if (proto)
    {
      // remember original parent namespace
      proto->d_namespace->setParent(d_namespace->getParent());

      // take the namespace (remember to protect it from deletion!)
      d_namespace = proto->d_namespace;

      //;!! ??? check type of first node...
      // steal nodes without cloning
      d_implementation = proto->d_implementation;
      proto->d_implementation = 0;
      // change encapsulating prototype
      d_implementation->setEncapsulatingPrototype(this);

      // Make sure we get all the IS maps by using the field
      // lists from the implementation rather than just the
      // interface.
      destructFieldList( d_eventIns );
      destructFieldList( d_eventOuts );
      destructFieldList( d_fields );

      d_eventIns = proto->d_eventIns;
      d_eventOuts = proto->d_eventOuts;
      d_fields = proto->d_fields;

      // protect the field lists from deletion (hack)
      proto->d_eventIns.erase(proto->d_eventIns.begin(),
			      proto->d_eventIns.end());
      proto->d_eventOuts.erase(proto->d_eventOuts.begin(),
			      proto->d_eventOuts.end());
      proto->d_fields.erase(proto->d_fields.begin(),
			    proto->d_fields.end());

      setActualUrl(proto->url());

      // protect the namespace from deletion (hack)
      proto->d_namespace = NULL;

      delete proto;
    }
  else
    {
      theSystem->warn(System::SEVERITY_HIGH, System::WARNINGTYPE_UNKNOWN, "Couldn't read EXTERNPROTO %s\n", getName());
    }
}


VrmlMFNode *VrmlNodeType::getImplementationNodes()
{
  if (! d_implementation && d_url)
    fetchImplementation();

  // Now that the nodes are here, initialize any IS'd fields
  // to the default values (could do it at instantiation...)
  // Also, any IS'd field needs a name because of the stupid
  // IS mapping implementation (when I copy the implementation
  // nodes for each instance of the PROTO, the IS mappings
  // need to copied too). So make up a unique name for those
  // without.
  if (! d_fieldsInitialized)
    {
      FieldList::iterator i;

      for (i = d_fields.begin(); i != d_fields.end(); ++i)
	{
	  ISMap &ismap = (*i)->thisIS;
	  ISMap::iterator j;
	  for (j = ismap.begin(); j != ismap.end(); ++j)
	    {
	      VrmlNode *n = (*j)->node;

	      if ((*i)->defaultValue)
		n->setField( (*j)->fieldName, *((*i)->defaultValue) );

	    }
	}
      d_fieldsInitialized = true;
    }

  return d_implementation ? d_implementation->getNodes() : 0;
}


// This will NOT fetch the implementation of an EXTERNPROTO.
// This method is used in VrmlNodeProto to check the type of
// SFNode fields in toXXX() node downcasts. Type checking
// of EXTERNPROTOs is deferred until the implementation is
// actually downloaded. (not actually done yet...)

VrmlNode *VrmlNodeType::firstNode()
{
  return (d_implementation && d_implementation->getNodes()->getLength()) ? (*(d_implementation->getNodes()))[0] : 0;
}

VrmlField::VrmlFieldType VrmlNodeType::hasEventIn(const char *ename) const
{
  VrmlField::VrmlFieldType ft;

  // first try the event name as it is
  ft = has(d_eventIns, ename);
  if (ft!=VrmlField::NO_FIELD || strstarts(ename, "set_"))
    return(ft);

  // then add "set_" prefix if not yet there (for exposedFields)
  char tmp[256];
  snprintf(tmp, sizeof(tmp), "set_%s", ename);
  return(has(d_eventIns, tmp));
}

VrmlField::VrmlFieldType VrmlNodeType::hasEventOut(const char *ename) const
{
  VrmlField::VrmlFieldType ft;

  // first try the event name as it is
  ft = has(d_eventOuts, ename);
  if (ft!=VrmlField::NO_FIELD || strends(ename, "_changed"))
    return(ft);

  // then add "_changed" suffix if not yet there (for exposedFields)
  char tmp[1000];
  sprintf(tmp, "%s_changed", ename);
  return has(d_eventOuts, tmp);
}

VrmlField::VrmlFieldType VrmlNodeType::hasField(const char *ename) const
{
  return has(d_fields, ename);
}

VrmlField::VrmlFieldType VrmlNodeType::hasExposedField(const char *ename) const
{
  // Must have field "name", eventIn "set_name", and eventOut
  // "name_changed", all with same type. "set_ and "_changed" are
  // handled by hasEventIn and hasEventOut
  VrmlField::VrmlFieldType ft;

  if ( (ft = has(d_fields, ename)) == VrmlField::NO_FIELD)
    return VrmlField::NO_FIELD;
  if (ft != hasEventIn(ename))
    return VrmlField::NO_FIELD;
  if (ft != hasEventOut(ename))
    return VrmlField::NO_FIELD;

  return ft;
}

VrmlField::VrmlFieldType VrmlNodeType::hasInterface(char const * id) const
{
    VrmlField::VrmlFieldType fieldType = VrmlField::NO_FIELD;

    if ((fieldType = this->hasField(id)) != VrmlField::NO_FIELD) {
        return fieldType;
    }

    if ((fieldType = this->hasEventIn(id)) != VrmlField::NO_FIELD) {
        return fieldType;
    }

    if ((fieldType = this->hasEventOut(id)) != VrmlField::NO_FIELD) {
        return fieldType;
    }

    return fieldType;
}

VrmlField::VrmlFieldType
VrmlNodeType::has(const FieldList &recs, const char *ename) const
{
  FieldList::const_iterator i;
  for (i = recs.begin(); i != recs.end(); ++i) {
    if (strcmp((*i)->name, ename) == 0)
      return (*i)->type;
  }
  return VrmlField::NO_FIELD;
}

VrmlField*
VrmlNodeType::fieldDefault(const char *fname) const
{
  FieldList::const_iterator i;
  for (i = d_fields.begin(); i != d_fields.end(); ++i) {
    if (strcmp((*i)->name, fname) == 0)
      return (*i)->defaultValue;
  }
  return 0;
}

void VrmlNodeType::deleteInterface(const char* name)
{
  // possibly destruct item (if exists)
  destructFieldListItem(d_fields, name, true);
  destructFieldListItem(d_eventIns, name);
  destructFieldListItem(d_eventOuts, name);

  // in case of	exposed field don't forget destruct its events
  char tmp[256];
  snprintf(tmp, sizeof(tmp), "set_%s", name);
  destructFieldListItem(d_eventIns, tmp);
  snprintf(tmp, sizeof(tmp), "%s_changed", name);
  destructFieldListItem(d_eventOuts, tmp);
}

bool VrmlNodeType::renameInterface(const char* name, const char* newname)
{
  if (hasInterface(newname))
    return false;

  // possibly rename item (if exists)
  renameFieldListItem(d_fields, name, newname, true);
  renameFieldListItem(d_eventIns, name, newname);
  renameFieldListItem(d_eventOuts, name, newname);

  // in case of	exposed field don't forget rename its events
  char tmp[256];
  char tmpnew[256];
  snprintf(tmp, sizeof(tmp), "set_%s", name);
  snprintf(tmpnew, sizeof(tmpnew), "set_%s", newname);
  renameFieldListItem(d_eventIns, tmp, tmpnew);

  snprintf(tmp, sizeof(tmp), "%s_changed", name);
  snprintf(tmpnew, sizeof(tmpnew), "%s_changed", newname);
  renameFieldListItem(d_eventOuts, tmp, tmpnew);

  return hasInterface(newname);
}

void VrmlNodeType::addNode(VrmlNode * node)
{
    // add node to list of implementation nodes
    if (d_implementation) {
        d_implementation->getNodes()->addNode(node);
    } else {
        d_implementation = new VrmlNodeGroup();
	d_implementation->setName(0, d_namespace);
        d_implementation->setField("children", VrmlMFNode(1, &node));
	d_implementation->setEncapsulatingPrototype(this);
    }
}

void VrmlNodeType::addIS(const char *isFieldName,
			 const VrmlNode *implNode,
			 const char *implFieldName)
{
  FieldList::iterator i;

  theSystem->debug("%s::addIS(%s, %s::%s.%s)\n",
		   getName(),
		   isFieldName,
		   implNode->nodeType().getName(),
		   implNode->name(),
		   implFieldName);

  // Fields
  for (i = d_fields.begin(); i != d_fields.end(); ++i)
  {
    if (strcmp((*i)->name, isFieldName) == 0)
    {
      NodeFieldRec *nf = new NodeFieldRec;
      nf->node = (VrmlNode*) implNode; // oops...
      nf->fieldName = strdup(implFieldName);
      (*i)->thisIS.push_front(nf);
      break;
    }
  }

  // EventIns
  {
    const char* isname = strstarts(isFieldName, "set_") ? isFieldName+4 : isFieldName;
    for (i = d_eventIns.begin(); i != d_eventIns.end(); ++i)
    {
      char* implname = strstarts((*i)->name, "set_") ? (*i)->name+4 : (*i)->name;
      if (strcmp(implname, isname) == 0)
      {
	NodeFieldRec *nf = new NodeFieldRec;
	nf->node = (VrmlNode*) implNode; // oops...
	nf->fieldName = strdup(implFieldName);
	(*i)->thisIS.push_front(nf);
	break;
      }
    }
  }

  // EventOuts
  {
    const char* isname = strends(isFieldName, "_changed") ? strtrim(isFieldName, 0, 8) : isFieldName;
    for (i = d_eventOuts.begin(); i != d_eventOuts.end(); ++i)
    {
      char* implname = strends((*i)->name, "_changed") ? strtrim((*i)->name, 0, 8) : (*i)->name;
      if (strcmp(implname, isname) == 0)
      {
	NodeFieldRec *nf = new NodeFieldRec;
	nf->node = (VrmlNode*) implNode; // oops...
	nf->fieldName = strdup(implFieldName);
	(*i)->thisIS.push_front(nf);
	if (implname != (*i)->name) delete[] implname;
	break;
      }
      if (implname != (*i)->name) delete[] implname;
    }
    if (isname != isFieldName) delete[] isname;
  }

}

void VrmlNodeType::removeIS(const VrmlNode *implNode,
			    const char *implFieldName)
{
  // Fields
  for (FieldList::iterator i = d_fields.begin(); i != d_fields.end(); ++i)
  {
    ISMap::iterator j = (*i)->thisIS.begin();
    while (j != (*i)->thisIS.end())
    {
      VrmlNodeType::NodeFieldRec *nf = *j;
      if (nf->node==implNode && strcmp(implFieldName, nf->fieldName)==0)
      {
        free(nf->fieldName);
        delete nf;

	j = (*i)->thisIS.erase(j);
      }
      else
        ++j;
    }
  }

  // EventIns
  for (FieldList::iterator i = d_eventIns.begin(); i != d_eventIns.end(); ++i)
  {
    ISMap::iterator j = (*i)->thisIS.begin();
    while (j != (*i)->thisIS.end())
    {
      VrmlNodeType::NodeFieldRec *nf = *j;
      if (nf->node==implNode && strcmp(implFieldName, nf->fieldName)==0)
      {
        free(nf->fieldName);
        delete nf;

	j = (*i)->thisIS.erase(j);
      }
      else
        ++j;
    }
  }

  // EventOuts
  for (FieldList::iterator i = d_eventOuts.begin(); i != d_eventOuts.end(); ++i)
  {
    ISMap::iterator j = (*i)->thisIS.begin();
    while (j != (*i)->thisIS.end())
    {
      VrmlNodeType::NodeFieldRec *nf = *j;
      if (nf->node==implNode && strcmp(implFieldName, nf->fieldName)==0)
      {
        free(nf->fieldName);
        delete nf;

	j = (*i)->thisIS.erase(j);
      }
      else
        ++j;
    }
  }
}


VrmlNodeType::ISMap *VrmlNodeType::getFieldISMap( const char *fieldName )
{

  FieldList::iterator i;
  for (i = d_fields.begin(); i != d_fields.end(); ++i)
  {
    if (strcmp((*i)->name, fieldName) == 0)
      return & ((*i)->thisIS);
  }

#if defined(__ORBISNAP__)

  for (i = d_eventIns.begin(); i != d_eventIns.end(); ++i)
  {
    if (strcmp((*i)->name, fieldName) == 0)
      return & ((*i)->thisIS);
  }

  for (i = d_eventOuts.begin(); i != d_eventOuts.end(); ++i)
  {
    if (strcmp((*i)->name, fieldName) == 0)
      return & ((*i)->thisIS);
  }

#endif

  return 0;
}


// VrmlNode factory: create a new instance of a node of this type.
// Built in nodes have a creator function specified, while instances
// of PROTOs are constructed by VrmlNodeProto.

VrmlNode* VrmlNodeType::newNode( VrmlScene *scene ) const
{
  if (d_creator)
    return (*d_creator)( scene );

  return new VrmlNodeProto( (VrmlNodeType *)this, scene );
}

/** Returns nonzero if the string "str" ends with the substring "sub". */
bool VrmlNodeType::strends(const char* str, const char* sub)
{
  int len = strlen(str);
  int sublen = strlen(sub);
  return(len>=sublen && strcmp(str+len-sublen, sub)==0);
}

/** Returns nonzero if the string "str" starts with the substring "sub". */
bool VrmlNodeType::strstarts(const char* str, const char* sub)
{
  return(strncmp(str, sub, strlen(sub)) == 0);
}

/** Removes first "start" characters from the start of the string
  * and "end" characters from the end and returns the result
  * as a newly allocated (via "new[]") string.
  */
char *VrmlNodeType::strtrim(const char* str, int start, int end)
{
  int len = strlen(str);
  char* result = new char[len-start-end+1];
  strncpy(result, str+start, len-start-end);
  result[len-start-end] = '\0';
  return result;
}

ostream& VrmlNodeType::print(ostream& os, int indent)
{
  // clear the temporary flag - we'll use it for marking USEd nodes
  // it is useful only for PROTOs because EXTERNPROTOs implementation is not saved
  if (!d_url)
  {
    VrmlMFNode *protoNodes = getImplementationNodes();
    int nNodes = protoNodes ? protoNodes->getLength() : 0;
    int i;
    for (i=0; i<nNodes; ++i)
      (*protoNodes)[i]->clearFlags();
  }

  printComment(os, indent, true);

  // PROTO or EXTERNPROTO header
  ((d_url) ? os <<  "EXTERNPROTO " : os << "PROTO ") << d_name << " [";

  // PROTO definition (fields and events)
  FieldList::iterator i;
  for (i=d_fields.begin(); i!=d_fields.end(); i++)
  {
    (*i)->printComment(os, indent+VrmlNode::INDENT_INCREMENT, true);

    os << (hasExposedField((*i)->name) ? "exposedField " : "field ");
    os << VrmlField::getTypeName((*i)->type) << " ";
    os << (*i)->name;
    // do not print defaults for EXTERNPROTO
    if (!d_url)
    {
      os << " ";
      (*i)->defaultValue->print(os);
    }
  }
  for (i=d_eventOuts.begin(); i!=d_eventOuts.end(); i++)
  {
    // do not print events derived from exposedFields, they are implicit
    const char* ename = (*i)->name;
    if (hasField(ename)) continue;
    if (strends(ename, "_changed"))
    {
      char* basename = strtrim(ename, 0, 8);
      if (hasField(basename)) { delete[] basename; continue; }
      delete[] basename;
    }

    (*i)->printComment(os, indent+VrmlNode::INDENT_INCREMENT, true);

    os << "eventOut ";
    os << VrmlField::getTypeName((*i)->type) << " ";
    os << (*i)->name << " ";
  }
  for (i=d_eventIns.begin(); i!=d_eventIns.end(); i++)
  {
    // do not print events derived from exposedFields, they are implicit
    const char* ename = (*i)->name;
    if (hasField(ename)) continue;
    if (strstarts(ename, "set_") && hasField(ename+4)) continue;

    (*i)->printComment(os, indent+VrmlNode::INDENT_INCREMENT, true);

    os << "eventIn ";
    os << VrmlField::getTypeName((*i)->type) << " ";
    os << (*i)->name << " ";
  }

  // end previous line
  os << endl;
  // indentation
  int j;
  for (j=0; j<indent; j++) os << " ";

  // end of header
  os << "]";

  // it is EXTERNPROTO
  if (d_url)
  {
    d_url->printComment(os, indent, true);
    d_url->print(os, indent);

    // end of EXTERNPROTO
    os << "\n";
  }
  // otherwise it is PROTO
  else
  {
    // new line
    os << "\n";

    // indentation
    for (j=0; j<indent; j++) os << " ";

    // begin of body
    os << "{";

    // print nested PROTOs and EXTERNPROTOs
    std::list<VrmlNodeType*> const* typeslist = d_namespace->getTypeList();
    for (std::list<VrmlNodeType*>::const_iterator it=typeslist->begin(); it!=typeslist->end(); ++it)
    {
      (*it)->print(os, indent+INDENT_INCREMENT);
    }

    // body
    if (d_implementation)
    {
      d_implementation->getNodes()->printUnbracketed(os, indent+VrmlNode::INDENT_INCREMENT);
    }

    // ROUTEs
    RouteList& rl = d_namespace->getRouteList();
    RouteList::iterator r;
    for (r=rl.begin(); r!=rl.end(); ++r)
    {
      // print route
      r->second.print(os, indent+VrmlNode::INDENT_INCREMENT);
    }

    // end of previous line
    os << endl;
    // ending brace
    for (j=0; j<indent; j++) os << " ";
    os << "}" << endl;
  }

  return os;
}



ostream& VrmlNodeType::toChain(ostream& os)
{
  // clear the temporary flag - we'll use it for marking USEd nodes
  {
    VrmlMFNode *protoNodes = getImplementationNodes();
    int nNodes = protoNodes ? protoNodes->getLength() : 0;
    int i;
    for (i=0; i<nNodes; ++i)
      (*protoNodes)[i]->clearFlags();
  }

  // PROTO identification mark
  os << "P";
  os << "PROTO " << d_name << TERMINAL_STRING;

  // there is no child of PROTO node.
  os << TERMINAL_STRING;

  return os;
}
