/*****************************************************************************
******************************************************************************
*
*               VRML node type class
*
*               $Revision: 1.1.2.10 $
*               $Date: 2011/08/08 12:08:12 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright 1998-2011 HUMUSOFT s.r.o.
// Portions Copyright (C) 1998  Chris Morley

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#ifndef VRMLNODETYPE_H
#define VRMLNODETYPE_H

//
// The VrmlNodeType class is responsible for storing information about node
// or prototype types.
//

#include <list>
#include <map>

#include "System.h"  // OPENVRMLDLLEXPORT
#include "VrmlObject.h"
#include "VrmlField.h"

class Doc;
class VrmlNamespace;
class VrmlNode;
class VrmlNodeGroup;
class VrmlScene;


class OPENVRMLDLLEXPORT VrmlNodeType : public VrmlObject {
public:
  // Constructor.  Takes name of new type (e.g. "Transform" or "Box")
  // Copies the string given as name.
  VrmlNodeType(const char *nm,
	       VrmlNode* (*creator)(VrmlScene *scene) = 0);

  // Copy constructor
  VrmlNodeType(const VrmlNodeType& type, VrmlNamespace* parentNamespace = 0);

  // Deallocate storage for name and PROTO implementations
  ~VrmlNodeType();

  VrmlNodeType *reference() { ++d_refCount; return this; }
  void dereference();

  int getRefCount() const { return d_refCount; }

  // Create a node of this type in the specified scene
  VrmlNode *newNode( VrmlScene* s = 0 ) const;

  // Set/get scope of this type (namespace is NULL for builtins).
  VrmlNamespace *scope() { return d_namespace; }
  void setScope(VrmlNamespace &);

  // Routines for adding/getting eventIns/Outs/fields to this type
  const char* addEventIn(const char *name, VrmlField::VrmlFieldType type);
  const char* addEventOut(const char *name, VrmlField::VrmlFieldType type);
  const char* addField(const char *name, VrmlField::VrmlFieldType type);
  void addField(const char *name, VrmlField::VrmlFieldType type,
                const VrmlField& defaultVal);
  const char* addExposedField(const char *name, VrmlField::VrmlFieldType type);
  void addExposedField(const char *name, VrmlField::VrmlFieldType type,
                       const VrmlField& defaultVal);

  void setFieldDefault(const char *name, VrmlField const * value);

  VrmlField::VrmlFieldType hasEventIn(const char *name) const;
  VrmlField::VrmlFieldType hasEventOut(const char *name) const;
  VrmlField::VrmlFieldType hasField(const char *name) const;
  VrmlField::VrmlFieldType hasExposedField(const char *name) const;
  VrmlField::VrmlFieldType hasInterface(char const *) const;

  VrmlField *fieldDefault(const char *name) const;

  // deletes interface item (and its mappings)
  void deleteInterface(const char* name);

  // renames interface (mappings taken into account, of course)
  bool renameInterface(const char* name, const char* newname);

  // Set the URL to retrieve an EXTERNPROTO implementation from.
  void setUrl(VrmlMFString * url, Doc const * relative = 0);

  // Get the URL to retrieve an EXTERNPROTO implementation from
  VrmlMFString* getUrl() const { return d_url; }

  void setActualUrl(const char *url);

  Doc* getRelativeUrlDoc() { return d_relative; }

  // Retrieve the actual URL the PROTO was retrieved from.
  const char *url() { return d_actualUrl ? d_actualUrl->get() : 0; }

  // Add a node to a PROTO implementation
  void addNode(VrmlNode * implNode);

  // Add an IS linkage to one of the PROTO interface fields/events.
  void addIS(const char *isFieldName,
	     const VrmlNode *implNode,
	     const char *implFieldName);

  // Remove an IS linkage for one of fields/events from implementation.
  void removeIS(const VrmlNode *implNode,
	        const char *implFieldName);

  const char *getName() const { return d_name; }

  void rename(const char* newname);

  VrmlMFNode *getImplementationNodes();

  VrmlNodeGroup* nodes();

  VrmlNode *firstNode();

  struct NodeFieldRec {
    VrmlNode *node;
    char *fieldName;

    NodeFieldRec();
    NodeFieldRec(const NodeFieldRec& source,VrmlNamespace* ns);
  };

  typedef std::list<NodeFieldRec*> ISMap;

  class ProtoField : public VrmlObject {
    public:
      char *name;
      VrmlField::VrmlFieldType type;
      VrmlField *defaultValue;
      ISMap thisIS;

      ProtoField();
      ProtoField(const ProtoField& source, VrmlNamespace* ns);
  };

  typedef std::list<ProtoField*> FieldList;

  ISMap *getFieldISMap( const char *fieldName );

  const FieldList& eventIns() const
    { return d_eventIns; }
  const FieldList& eventOuts() const
    { return d_eventOuts; }

  FieldList &eventIns() { return d_eventIns; }
  FieldList &eventOuts() { return d_eventOuts; }

#if defined(__ORBISNAP__)
  FieldList &fields() { return d_fields; }
#endif

  ostream& print(ostream& os, int indent=0);
 /** Generate simplified description of VRML scene and store it
   * to the given output stream.
   * @param os an output stream where the description should be stored.
   * @return resulting stream.
   */
  ostream& toChain(ostream& os);

  // functions for manipulating d_modified flag
  void setModified();
  void clearModified() { d_modified = false; }
  bool isModified() { return d_modified; }

  static bool strends(const char* str, const char* sub);
  static bool strstarts(const char* str, const char* sub);
  static char *strtrim(const char* str, int start, int end);

private:

  // Grab the implementation of an EXTERNPROTO
  void fetchImplementation();

  void add(FieldList &, const char *, VrmlField::VrmlFieldType);
  VrmlField::VrmlFieldType has(const FieldList &, const char *) const;

  int d_refCount;

  bool d_modified; // PROTO instances need to be reinstantiated

  char *d_name;

  VrmlNamespace *d_namespace;

  VrmlMFString *d_url;		// Where the EXTERNPROTO could be.
  VrmlSFString *d_actualUrl;	// The URL actually used.
  Doc *d_relative;

  VrmlNodeGroup* d_implementation;	// The PROTO implementation nodes

  // Pointer to function to create instances
  VrmlNode* (*d_creator)( VrmlScene* );

  // Fields defined for this node type
  FieldList d_eventIns;
  FieldList d_eventOuts;
  FieldList d_fields;

  static size_t d_uniqueNameSequence;

  // sequences for generation of appropriate synthetic/gerneric names
  size_t d_eventInsSequence, d_eventOutsSequence, d_fieldsSequence, d_exposedFieldsSequence;

  bool d_fieldsInitialized;
};
#endif // VRMLNODETYPE_H
