/*****************************************************************************
******************************************************************************
*
*               VRML scene (aka world) class
*
*               $Revision: 1.1.2.37 $
*               $Date: 2011/09/30 05:53:23 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright 1998-2011 HUMUSOFT s.r.o.
// Portions Copyright (C) 1998  Chris Morley

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include <stdio.h>
#include <string.h>

#include <sstream>
#include <fstream>
using namespace std;

#include "VrmlScene.h"

#include "Doc.h"
#include "Viewer.h"
#include "System.h"
#include "MathUtils.h"
#include "Vrml97Parser.h"
#include "VrmlNamespace.h"
#include "VrmlNodeType.h"
#include "VrmlRenderContext.h"


// Handle clicks on Anchor nodes
#include "VrmlNodeAnchor.h"

#include "VrmlNodeScalarInt.h"
#include "VrmlNodePositionInt.h"
#include "VrmlNodeOrientationInt.h"
#include "VrmlNodeColorInt.h"

#define TERMINAL_STRING "#"

/**
 * Describes a change in a field value (e.g. for logging purposes).
 */
struct FieldChangeLogEntry
{
  /** Time of the change. */
  size_t timeIndex;

  /** The new value of the field. */
  VrmlField* value;

  /**
   * Constructor.
   */
  FieldChangeLogEntry(size_t timeIndex, const VrmlField* value);

  ~FieldChangeLogEntry();
};



/**
 * A log of changes to a field.
 */
typedef std::vector<FieldChangeLogEntry*> FieldChangeLog;



/**
 * This structure describes changes that were made to specific fields
 * during some time. It can be used to hold a VRML animation
 * for later reconstruction.
 */
struct SceneChangeLog
{
  static const char* EXPOSED_FIELD_TOKEN;

  /**
   * Names of the fields in form "nodename.fieldname".
   */
  std::vector<std::string> fieldNames;

  std::vector<double> logTimes;

  /**
   * Logs of changes to the named fields.
   */
  std::vector<FieldChangeLog*> fieldChanges;

  ~SceneChangeLog();
  void addEntry(size_t timeIndex, const char* field, const VrmlField* value);
};

// initialize static member
const char* SceneChangeLog::EXPOSED_FIELD_TOKEN = "#EXF";


// initialize static member to NULL
VrmlScene* VrmlScene::clipboardOwner = NULL;

//
// Create a VrmlScene from a URL (optionally loading from a local copy,
// so I can run as a netscape helper but still retrieve embedded urls).
//

VrmlScene::VrmlScene(const char *sceneUrl, bool edited) :
  d_flags_need_updating(false),
  openCount(0),
  d_time(0),
  d_realTimeSynced(false),
  d_recordingStartTime(0),
  d_recordingStopTime(0),
  d_recordFileName(0),
  d_recordingScheduled(false),
  d_recordToVRML(false),
  d_url(0),
  d_namespace(0),
  d_edited(edited),
  d_markedForDestruction(false),

  d_pendingUrl(0),
  d_pendingParameters(0),
  d_pendingNodes(0),
  d_pendingScope(0),
  d_recordLog(0)

{
  if (sceneUrl!=NULL)
    d_url = new Doc(sceneUrl);
  d_accessFlags.Viewable = true;
  setRecordFileName("%f_anim_%n.wrl");
}

VrmlScene::~VrmlScene()
{
  // close the scene unconditionally
  if (openCount>1) openCount = 1;
  close();

  // there is no namespace available at this moment
  d_nodes.setNamespaceDestructed();

  // delete the URL
  delete[] d_recordFileName;

  // delete the URL
  delete d_url;
}

void VrmlScene::open()
{
  if (openCount==0)
  {
    // initialize members
    d_flags_need_updating = false;
    d_namespace = 0;
    d_modified = false;
    d_pendingUrl = 0;
    d_pendingParameters = 0;
    d_pendingNodes = 0;
    d_pendingScope = 0;
    d_frameRate = 0.0,
    d_texturesDirty = true;
    d_lastUpdateTime = -1;  // must be behind d_time
    d_recording = false;
    d_recordLog = NULL;
    d_recordRealFileName = NULL;
    d_lastRecordingStartTime = 0;
    d_lastRecordingStopTime = 0;
    primaryViewer = NULL;
    viewers.clear();
    setTime(0);

    d_nodes.addToScene(this, d_url->url());
    load();
  }
  openCount++;
}

void VrmlScene::close()
{
  // sanity check
  if (openCount<=0)
  {
    openCount=0;
    return;
  }

  // decrement the open count
  openCount--;

  // if the world is closing, remove it from memory
  if (openCount<=0)
  {
    // remove all viewers
    ViewerList::iterator it;
    for (it=viewers.begin(); it!=viewers.end(); )
    {
      ViewerList::iterator itd = it++;
      delete(*itd);
    }
    viewers.clear();

    // finalize recording if not stopped yet (also ensures
    // that appropriate files are closed)
    stopRecording();

    // delete clipboard contents
    d_clipboard.set(0);

    // if it is clipboard owner, set it to NULL
    if (VrmlScene::getClipboardOwner()==this)
      VrmlScene::setClipboardOwner(0);

    // remove the nodes themselves
    d_nodes.removeChildren();
    d_nodes.addToScene( 0, 0 );

    // destroy related scenes
    relatedScenesFromNodeTypeContainer.destroyall();

    delete d_pendingUrl;
    delete d_pendingParameters;
    delete d_pendingNodes;
    delete d_pendingScope;

    delete d_namespace;

    delete d_recordLog;

    // clear containers in the end
    // clear stacks of bindables
    d_backgroundStack.clear();
    d_fogStack.clear();
    d_navigationInfoStack.clear();
    d_viewpointStack.clear();

    // clear stacks of scene-level nodes
    d_backgrounds.clear();
    d_fogs.clear();
    d_navigationInfos.clear();
    d_viewpoints.clear();
    d_scopedLights.clear();
    d_scripts.clear();
    d_timers.clear();
    d_movies.clear();
    d_audioClips.clear();
  }

}

Doc* VrmlScene::urlDoc()
{ return d_url; }

VrmlScene* VrmlScene::getParent()
{ return this; }

VrmlScene* VrmlScene::getTopLevelParent()
{ return this; }

VrmlNodeType* VrmlScene::getNodeType()
{ return 0; }

VrmlNamespace* VrmlScene::scope()
{ return d_namespace; }

VrmlNodeGroup* VrmlScene::nodes()
{ return &d_nodes; }

bool VrmlScene::isEdited()
{ return d_edited; }



// Load a (possibly non-VRML) file...

void VrmlScene::loadUrl(VrmlMFString const * url, VrmlMFString const * parameters)
{
    VrmlLoadingError primaryError("No usable URL found");
    bool primaryUrlFailed = false;

    if (!url) throw primaryError;

    size_t np = parameters ? parameters->getLength() : 0;
    char const * const * params = parameters ? parameters->get() : 0;

    // try each url until we find one we can handle
    size_t i(0);
    char const * const * urls = url->get();
    for (; i < url->getLength(); ++i)
    {
        // ignore empty strings
        if (! urls[i]) continue;

        // if URL is in form "#Viewpoint", simply go to the viewpoint
        if (*urls[i] == '#')
        {
            load(urls[i]);
            return;
        }

        // Load .wrl's, or pass off to system
        else {
            // Check mime type...
            char const * tail = strrchr(urls[i], SLASH);
            if (! tail) {
                tail = urls[i];
            }
            char const * mod = strchr(tail, '#');
            if (! mod) {
                mod = urls[i] + strlen(urls[i]);
            }
            if (mod-tail > 4
                && (   strncmp( mod-4, ".wrl", 4) == 0
                    || strncmp( mod-4, ".wrz", 4) == 0
                    || strncmp( mod-4, ".WRL", 4) == 0
                    || strncmp( mod-4, ".WRZ", 4) == 0
                    || (mod-tail > 7 && strncmp( mod-7, ".wrl.gz", 7) == 0 )))
            {
                // try loading the VRML file, return immediately on success
                try
                {
                  load(urls[i]);
                  return;
                }
                catch (VrmlLoadingError &err)
                {
                  // remember the error that caused the primary URL to fail
                  if (!primaryUrlFailed)
                  {
                    primaryError = err;
                    primaryUrlFailed = true;
                  }
                }

            }
        }
    }

    // all URLs failed, throw an error
    throw primaryError;
}

// Replace nodes

void VrmlScene::replaceWorld(VrmlMFNode & nodes, VrmlNamespace * ns,
                             Doc * url)
{
  // Clear bindable stacks.
  d_backgroundStack.removeAll();
  d_fogStack.removeAll();
  d_navigationInfoStack.removeAll();
  d_viewpointStack.removeAll();

  // Get rid of current world nodes.
  d_nodes.removeChildren();

  // remove namespaces and URLs
  delete d_namespace;
  delete d_url;

  d_namespace = ns;

  // URL can be NULL if the scene was loaded from a string
  if (url) d_url = url; else d_url = new Doc();

  // Do this to set the relative URL
  d_nodes.addToScene( (VrmlScene *) this, urlDoc()->url() );

  // Add the nodes to a Group and put the group in the scene.
  // This will load EXTERNPROTOs and Inlines.
  d_nodes.addChildren( nodes );

  // set comment to root
  d_nodes.setComment(nodes.getComment());

  // set new name to root node
  d_nodes.setName(0, d_namespace);

  // Send initial set_binds to bindable nodes
  bindSceneDefaultBindableNodes(VRML_NODE_BACKGROUND |
                                     VRML_NODE_FOG |
                                     VRML_NODE_NAVIGATION_INFO |
                                     VRML_NODE_VIEWPOINT);

  setModified();
}


/**
 * Load the scene using the associated URL.
 * @exception VrmlLoadingError
 */
void VrmlScene::load()
{
  // allocate new namespace and new root MFNode
  VrmlNamespace* newScope = new VrmlNamespace();
  VrmlMFNode* newNodes;

  // read the scene from associated URL
  try
  {
    newNodes = readWrl(d_url, newScope);
  }
  catch (VrmlLoadingError&)
  {
    delete newScope;
    throw;
  }

  d_namespace = newScope;

  // notify viewers that world is about to be loaded
  for (ViewerList::iterator it = viewers.begin(); it != viewers.end(); ++it)
    (*it)->onWorldLoadStarting();

  // set the relative URL to all nodes
  d_nodes.addToScene( (VrmlScene *) this, d_url->url() );

  // add the nodes to a Group and put the group in the scene
  // this will load EXTERNPROTOs and INLINEs
  d_nodes.addChildren(*newNodes);
  // set comment to root
  d_nodes.setComment(newNodes->getComment());
  delete newNodes;

  // set name to root node
  d_nodes.setName(0, d_namespace);

  // send initial set_binds to bindable nodes
  bindSceneDefaultBindableNodes(VRML_NODE_BACKGROUND |
                                     VRML_NODE_FOG |
                                     VRML_NODE_NAVIGATION_INFO |
                                     VRML_NODE_VIEWPOINT);

  // mark the new world as modified
  setModified();

  // notify viewers that world loading has finished
  for (ViewerList::iterator it = viewers.begin(); it != viewers.end(); ++it)
    (*it)->onWorldLoadComplete();
}



// Read a VRML97 file.
// This is only for [*.wrl][#viewpoint] url loading (no parameters).

void VrmlScene::load(const char *url, const char *localCopy)
{
    // Look for '#Viewpoint' syntax. There ought to be a current
    // scene if this format is used.
    VrmlSFBool flag(true);
    if (*url == '#') {
        VrmlNode *vp = d_namespace ? d_namespace->findNode(url+1) : 0;

        // spec: ignore if named viewpoint not found
        if (vp) {
            vp->eventIn(d_time, "set_bind", &flag );
            setModified();
        }
        return;
    }

    // Try to load a file. Prefer a local copy if available.
    Doc * tryUrl = 0;
    if (localCopy) {
        tryUrl = new Doc(localCopy, 0);
    } else {
        tryUrl = new Doc(url, d_url);
    }

    VrmlNamespace * newScope = new VrmlNamespace();

    VrmlMFNode* newNodes = NULL;
    try
    {
      newNodes = readWrl(tryUrl, newScope);
    }
    catch (VrmlLoadingError&)
    {
      delete tryUrl;
      delete newScope;
      throw;
    }

    Doc * sourceUrl = tryUrl;
    if (localCopy) {
        sourceUrl = new Doc( url );
    }

    // notify viewers
    for (ViewerList::iterator it = viewers.begin(); it != viewers.end(); ++it)
      (*it)->onWorldLoadStarting();

    replaceWorld(*newNodes, newScope, sourceUrl);

    delete newNodes;

    // Look for '#Viewpoint' syntax
    if (sourceUrl->urlModifier())
    {
        VrmlNode * vp = d_namespace->findNode(sourceUrl->urlModifier() + 1);
        if (vp)
        {
          vp->eventIn(d_time, "set_bind", &flag);
        }
    }

  // notify viewers
  for (ViewerList::iterator it = viewers.begin(); it != viewers.end(); ++it)
    (*it)->onWorldLoadComplete();
}

bool VrmlScene::reload(bool preserveBindings)
{
  // notify viewers
  for (ViewerList::iterator it = viewers.begin(); it != viewers.end(); ++it)
    (*it)->onWorldReloadStarting();

  // underlying node types will be re-created
  relatedScenesFromNodeTypeContainer.destroyall();

  // create a new namespace and new tree node
  VrmlNamespace* newScope = new VrmlNamespace();
  VrmlMFNode* newNodes = new VrmlMFNode();
  Doc* doc = new Doc(d_url->url());
  // do not try to read empty url
  if (strcmp(d_url->url(), "")!=0)
  {
    DocInputStream istm(*doc);
    if (!istm.good()) return false;
    Vrml97Parser decoder(istm, d_url->url());
    try
    {
      decoder.loadTopLevel(newNodes, newScope);
    }
    catch (VrmlLoadingError&)
    {
      delete doc;
      delete newScope;
      return false;
    }
  }

  // replace the old nodes
  replaceWorld(*newNodes, newScope, doc);
  delete newNodes;

  // notify viewers
  for (ViewerList::iterator it = viewers.begin(); it != viewers.end(); ++it)
    (*it)->onWorldReloadComplete(preserveBindings);

  // NOTE: after the world is reloaded, the default viewpoint is bound
  // while the primary viewer may rest on a different one. Is this
  // a problem? If it is, we'll have to resynchronize them somehow
  // (it is not possible to simply bound that viewpoint because
  // this would reset the camera)

  return true;
}



VrmlMFNode* VrmlScene::readWrl(Doc* tryUrl, VrmlNamespace* ns)
{
  VrmlMFNode* newnodes = new VrmlMFNode();
  // do not try to read empty url
  if (strcmp(tryUrl->url(), "")==0)
    return (newnodes);

  // try reading the URL
  DocInputStream istm(*tryUrl);
  if (!istm.good())
    throw VrmlLoadingError("Can't open file for reading",
			   tryUrl->url());

  // parse the file
  Vrml97Parser decoder(istm, tryUrl->url());
  try
  {
    decoder.loadTopLevel(newnodes, ns);
  }
  catch (VrmlLoadingError&)
  {
    delete newnodes;
    throw;
  }

  return(newnodes);
}



VrmlMFNode* VrmlScene::readWrl(VrmlMFString* urls, Doc* relative, VrmlNamespace* ns)
{
  Doc url;
  VrmlLoadingError primaryError("No usable URL found");
  bool primaryUrlFailed = false;

  // cycle through all URLs
  int n = urls->getLength();
  for (int i = 0; i < n; ++i)
  {
    url.seturl(urls->get(i), relative);
    try
    {
      VrmlMFNode* result = readWrl(&url, ns);
      return result;
    }
    catch (VrmlLoadingError &error)
    {
      if (!primaryUrlFailed)
      {
        primaryError = error;
        primaryUrlFailed = true;
      }
    }
  }

  // all URLs failed, throw error
  throw primaryError;
}



// Read VRML from a string and return the (valid) nodes.

VrmlMFNode* VrmlScene::readString(const char* vrmlString, VrmlNamespace* ns)
{
    //
    // Hmm. It looks like passing a zero pointer for the namespace argument of
    // this function is not an option. Perhaps it should be a reference rather
    // than a pointer?
    // -- Braden McDaniel <braden@endoframe.com> 1 Apr, 2000
    //

    VrmlMFNode* result = 0;
    if (vrmlString == NULL)
      throw VrmlLoadingError("No VRML string specified");

    istringstream istrstm(vrmlString);
    Vrml97Parser decoder(istrstm);
    result = new VrmlMFNode();
    try { decoder.loadTopLevel(result, ns); }
    catch (VrmlLoadingError&)
    {
      delete result;
      throw;
    }
    return result;
}


// Read a PROTO from a URL to get the implementation of an EXTERNPROTO.
// This should read only PROTOs and return when the first/specified PROTO
// is read...

VrmlNodeType * VrmlScene::readPROTO(VrmlMFString * urls, Doc const * relative)
{
    // This is a problem. The nodeType of the EXTERNPROTO has a namespace
    // that refers back to this namespace (protos), which will be invalid
    // after we exit this function. I guess it needs to be allocated and
    // ref counted too...
    //VrmlNamespace protos;
    VrmlNamespace * protos = new VrmlNamespace();  // leak...
    Doc urlDoc(static_cast<char *>(0));
    VrmlNodeType * def = 0;
//    int i, n = urls->size();

    for (size_t i(0); i < urls->getLength(); ++i) {

        theSystem->debug("Trying to read EXTERNPROTO from url '%s'\n",
                          urls->get(i));

        urlDoc.seturl( urls->get(i), relative );
        VrmlMFNode* kids = NULL;
        try
        {
          kids = VrmlScene::readWrl(&urlDoc, protos);
        }
        catch (VrmlLoadingError &error)
        {
          theSystem->warn(System::SEVERITY_HIGH, System::WARNINGTYPE_UNKNOWN,
                          "Could not read URL '%s': %s",
                          urls->get(i), error.what());
        }
        delete kids;

        // Grab the specified PROTO, or the first one.
        const char * whichProto = urlDoc.urlModifier();
        if (whichProto && *whichProto) {
            def = (VrmlNodeType*) protos->findType( whichProto+1 );
        } else {
            def = (VrmlNodeType*) protos->firstType();
        }

        if (def) {
            def->setActualUrl( urlDoc.url() );
            break;
        }
    }

    return def;
}



void VrmlScene::save(ostream& os)
{
  // print VRML2.0 header
  os << "#VRML V2.0 utf8\n";

  // print world comment
  d_nodes.printComment(os, 0, false);


  // print PROTOs and EXTERNPROTOs
  std::list<VrmlNodeType*> const* typeslist = d_namespace->getTypeList();
  for (std::list<VrmlNodeType*>::const_iterator it=typeslist->begin(); it!=typeslist->end(); ++it)
  {
    (*it)->print(os, 0);
  }

  // clear temporary flags for all nodes, we'll use them for marking
  // USEd nodes
  d_nodes.clearFlags();

  // print nodes
  VrmlMFNode* nodelist = (VrmlMFNode*) d_nodes.getField("children");
  for (int i=0; i<nodelist->getLength(); i++)
  {
    (nodelist->operator[](i))->print(os, 0, true);
  }

  // print ROUTEs
  const RouteList& rl = d_namespace->getRouteList();
  for (RouteList::const_iterator r=rl.begin(); r!=rl.end(); ++r)
  {
    // print route
    r->second.print(os,0);
  }
}



void VrmlScene::saveSimpleChain(ostream &os, bool editFlag)
{

  os << desc() << TERMINAL_STRING;

  // save ROOT
  os << "R";


  // print PROTOs and EXTERNPROTOs
  std::list<VrmlNodeType*> const* typeslist = scope()->getTypeList();
  for (std::list<VrmlNodeType*>::const_iterator it=typeslist->begin(); it!=typeslist->end(); ++it)
  {
    (*it)->toChain(os);
  }

  // clear temporary flags for all nodes, we'll use them for marking
  // USEd nodes
  nodes()->clearFlags();

  // print nodes
  VrmlMFNode* nodelist = (VrmlMFNode*) nodes()->getField("children");
  for (int i=0; i<nodelist->getLength(); i++)
  {
    (nodelist->operator[](i))->toChain(os, editFlag);
  }


  // print ROUTEs
  RouteList& rl = scope()->getRouteList();
  for (RouteList::const_iterator r=rl.begin(); r!=rl.end(); ++r)
  {
    if (r->second.fromNode() && r->second.toNode() && r->second.fromEventOut() && r->second.toEventIn())
    {
      os << "T" << "ROUTE " << r->second.fromNode()->name() << "." << r->second.fromEventOut();
      os << " TO " << r->second.toNode()->name() << "." << r->second.toEventIn() << TERMINAL_STRING;

      // no further sub-items
      os << TERMINAL_STRING;
    }
  }

  // no further sub-item of ROOT
  os << TERMINAL_STRING;
}


bool VrmlScene::save(const char* filename)
{
  // create backup of the original file
  theSystem->renameToBackup(filename);

  // create the output stream
  // the Win32 special case is needed only for VC 8 and locale not set (as in R2007a)
#ifdef _WIN32
  FILE* fptr = fopen(filename, "w");
  if (!fptr) return(false);
  ofstream f(fptr);
#else
  ofstream f(filename);
#endif

  // save it
  save(f);
  f.close();
  return (f.good());
}



//
// Script node API functions
//
const char *VrmlScene::getName() { return "OpenVRML"; }

const char *VrmlScene::getVersion()
{
  static char vs[32];
  snprintf(vs, sizeof(vs), "%d.%d.%d", OPENVRML_MAJOR_VERSION, OPENVRML_MINOR_VERSION, OPENVRML_MICRO_VERSION);
  return vs;
}

double VrmlScene::getFrameRate() { return d_frameRate; }

// Queue an event to load URL/nodes (async so it can be called from a node)

void VrmlScene::queueLoadUrl( VrmlMFString *url, VrmlMFString *parameters )
{
  if (! d_pendingNodes && ! d_pendingUrl)
    {
      d_pendingUrl = dynamic_cast<VrmlMFString*>(url->clone());
      d_pendingParameters = dynamic_cast<VrmlMFString*>(parameters->clone());
    }
}

void VrmlScene::queueReplaceNodes( VrmlMFNode *nodes, VrmlNamespace *ns )
{
  if (! d_pendingNodes && ! d_pendingUrl)
    {
      d_pendingNodes = dynamic_cast<VrmlMFNode*>(nodes->clone());
      d_pendingScope = ns;
    }
}

// Called by the viewer when the cursor passes over, clicks, drags, or
// releases a sensitive object (an Anchor or another grouping node with
// an enabled TouchSensor child).

void VrmlScene::sensitiveEvent( void *object,
                                double timeStamp,
                                bool isOver, bool isActive,
                                double *point )
{
  VrmlNode *n = (VrmlNode *)object;

  if (n)
  {
    VrmlNodeAnchor* a = NodeDynamicCast<VrmlNodeAnchor*>(n);
    if (a)
    {
      // This should really be (isOver && !isActive && n->wasActive)
      // (ie, button up over the anchor after button down over the anchor)
      if (isActive && isOver)
	a->activate();
      else if (isOver)
      {
	const char *description = a->description();
	const char *url = a->url();

	if (description && url)
	  theSystem->debug("%s (%s)", description, url);
	else if (description || url)
	  theSystem->debug("%s", description ? description : url);
      }
    }

  // The parent grouping node is registered for Touch/Drag Sensors
  else
  {
    VrmlNodeGroup* g = NodeDynamicCast<VrmlNodeGroup*>(n);
    if (g)
    {
      g->activate(timeStamp, isOver, isActive, point);
      setModified();
    }
  }
}

}


/**
 * Switches the automatical updating of the scene time on or off.
 */
void VrmlScene::setRealTimeSynced(bool value)
{
  if (!d_realTimeSynced && value)
    d_realTimeOffset = theSystem->time() - d_time;
  d_realTimeSynced = value;
}


//
// The update method is where the events are processed. It should be
// called after each frame is rendered.
//
bool VrmlScene::update()
{
  // don't try to update closed scene
  if (openCount<=0)
    return(false);

  // auto-increment the time counter with real time if requested
  if (d_realTimeSynced)
    setTime(theSystem->time() - d_realTimeOffset);

  // call time-dependent event only once in given time
  if (d_lastUpdateTime != d_time)
  {
    VrmlSFTime now(d_time);

    // Update each of the timers.
    for (std::list<VrmlNodeTimeSensor*>::iterator tsi = d_timers.begin(); tsi != d_timers.end(); ++tsi)
      (*tsi)->update(now);

    // Update each of the clips.
    for (std::list<VrmlNodeAudioClip*>::iterator aci = d_audioClips.begin(); aci != d_audioClips.end(); ++aci)
      (*aci)->update(now);

    // Update each of the scripts.
    for (std::list<VrmlNodeScript*>::iterator si = d_scripts.begin(); si != d_scripts.end(); ++si)
      (*si)->update(now);

    // Update each of the movies.
    for (std::list<VrmlNodeMovieTexture*>::iterator mi = d_movies.begin(); mi != d_movies.end(); ++mi)
      (*mi)->update(now);

    d_lastUpdateTime = d_time;
  }

  if (d_pendingNodes)
    {
      replaceWorld( *d_pendingNodes, d_pendingScope );
      delete d_pendingNodes;
      d_pendingNodes = 0;
      d_pendingScope = 0;
    }
  else if (d_pendingUrl)
    {
      (void) loadUrl( d_pendingUrl, d_pendingParameters );
      delete d_pendingUrl;
      delete d_pendingParameters;
      d_pendingUrl = 0;
      d_pendingParameters = 0;
    }

  // Signal a redisplay if necessary
  return isModified();
}


/**
 * Request redraw for all viewers of this scene.
 */
void VrmlScene::redrawViewers()
{
  // request redraw for all viewers
  for (ViewerList::iterator it = viewers.begin(); it != viewers.end(); ++it)
  {
    if ((*it)->isValid())
      (*it)->requestRedraw();
  }
}


bool VrmlScene::headlightOn()
{
  VrmlNodeNavigationInfo *navInfo = bindableNavigationInfoTop();
  if (navInfo)
    return navInfo->headlightOn();
  return true;
}


// Draw this scene into the specified viewer

void VrmlScene::render(Viewer *viewer, bool final)
{
  // do not try to render closed scene
  if (openCount<=0) return;

  //if no lights exist do not render
  if (viewer->getRenderPhase()==Viewer::RENDER_PHASE_LIGHTS && d_scopedLights.size()==0)
    return;

  double M[4][4];
  Midentity(M);
  VrmlRenderContext rc(VrmlBVolume::BV_INSIDE, M);

  // Render the top level group
  d_nodes.render( viewer, rc );

  if (final)
  {
    d_texturesDirty = false;
  }

  // This is actually one frame late...
  d_frameRate = viewer->getFrameRate();
}

void VrmlScene::setModified()
{ d_modified = true; }

void VrmlScene::clearModified()
{ d_modified = false; }

bool VrmlScene::isModified() const
{ return d_modified; }

void VrmlScene::invalidate()
{
  // invalidated textures allow proper return from viewer checkSensitive() method
  d_texturesDirty = true;

  d_nodes.invalidate();
}

//
//  Bindable children node stacks. For the CS purists out there, these
//  aren't really stacks as they allow arbitrary elements to be removed
//  (not just the top).
//

void BindableStack::push(VrmlNode* node)
{
  list<VrmlNode*>::remove(node);
  push_front(node);
  node->scene()->setModified();

  // notify the viewers
  ViewerList::iterator end = node->scene()->viewers.end();
  for (ViewerList::iterator it = node->scene()->viewers.begin(); it != end; ++it)
    (*it)->onBindableBound(node);
}

void BindableStack::remove(VrmlNode* node)
{
  list<VrmlNode*>::remove(node);
  node->scene()->setModified();

  // notify the viewers
  ViewerList::iterator end = node->scene()->viewers.end();
  for (ViewerList::iterator it = node->scene()->viewers.begin(); it != end; ++it)
    (*it)->onBindableUnbound(node);
}


// Viewpoint
void VrmlScene::addViewpoint(VrmlNodeViewpoint* n)
{
  // add the viewpoint to the list
  d_viewpoints.push_back(n);

  // notify viewers about the change
  for (ViewerList::iterator it = viewers.begin(); it != viewers.end(); ++it)
    (*it)->onViewpointListChange(false);
}

void VrmlScene::removeViewpoint(VrmlNodeViewpoint* n)
{
    // remove the viewpoint from the viewpoint list
    d_viewpoints.remove(n);

    // remove bindable from the stack
    bindableRemove(n);

  // notify viewers about the change
  for (ViewerList::iterator it = viewers.begin(); it != viewers.end(); ++it)
    (*it)->onViewpointListChange(((*it)->getViewpoint() == n));
}

VrmlNodeViewpoint* VrmlScene::getNextViewpoint(VrmlNodeViewpoint* vp)
{
  // return always NULL if there are no viewpoints in the scene
  if (d_viewpoints.begin() == d_viewpoints.end()) return NULL;

  // if the pointer given is NULL, return the first viewpoint
  if (vp == NULL) return NodeDynamicCast<VrmlNodeViewpoint*>(*(d_viewpoints.begin()));

  // otherwise find the specified viewpoint and the next viewpoint
  for (std::list<VrmlNode*>::iterator it = d_viewpoints.begin(); it != d_viewpoints.end(); ++it)
  {
    if ((*it) == vp)
    {
      if (++it == d_viewpoints.end()) it = d_viewpoints.begin();
      return NodeDynamicCast<VrmlNodeViewpoint*>(*it);
    }
  }
  return NULL;
}

VrmlNodeViewpoint* VrmlScene::getPrevViewpoint(VrmlNodeViewpoint* vp)
{
  // return always NULL if there are no viewpoints in the scene
  if (d_viewpoints.begin() == d_viewpoints.end()) return NULL;

  // if NULL is specified, return the last viewpoint
  if (vp == NULL) return NodeDynamicCast<VrmlNodeViewpoint*>(*(--d_viewpoints.end()));

  // otherwise find the specified viewpoint and the previous viewpoint
  for (std::list<VrmlNode*>::iterator it = d_viewpoints.begin(); it != d_viewpoints.end(); ++it)
  {
    if ((*it) == vp)
    {
      if (it == d_viewpoints.begin()) it = d_viewpoints.end();
      return NodeDynamicCast<VrmlNodeViewpoint*>(*(--it));
    }
  }
  return NULL;
}



int VrmlScene::getViewpointIndex(VrmlNodeViewpoint* vp)
{
  int n = 0;

  for (std::list<VrmlNode*>::iterator it = d_viewpoints.begin(); it != d_viewpoints.end(); ++it, ++n)
    if (dynamic_cast<VrmlNodeViewpoint*>(*it) == vp)
      return n;

  return -1;
}



VrmlNodeViewpoint* VrmlScene::findViewpointByDescription(const char* desc)
{
  if (desc == NULL || strlen(desc)==0) return NULL;
  for (std::list<VrmlNode*>::iterator it = d_viewpoints.begin(); it != d_viewpoints.end(); ++it)
  {
    VrmlNodeViewpoint* viewpoint = dynamic_cast<VrmlNodeViewpoint*>(*it);
    if (strcmp(desc, viewpoint->description()) == 0)
      return viewpoint;
  }
  return NULL;
}



VrmlNodeViewpoint* VrmlScene::findViewpointByName(const char* name)
{
  if (name == NULL) return NULL;
  for (std::list<VrmlNode*>::iterator it = d_viewpoints.begin(); it != d_viewpoints.end(); ++it)
  {
    VrmlNodeViewpoint* viewpoint = dynamic_cast<VrmlNodeViewpoint*>(*it);
    if (strcmp(name, viewpoint->name()) == 0)
      return viewpoint;
  }
  return NULL;
}



VrmlNodeViewpoint* VrmlScene::findViewpointByIndex(int nvp)
{
  int n = 0;

  for (std::list<VrmlNode*>::iterator it = d_viewpoints.begin(); it != d_viewpoints.end(); ++it, ++n)
    if (n == nvp)
      return dynamic_cast<VrmlNodeViewpoint*>(*it);

  return NULL;
}



void VrmlScene::setViewpoint(VrmlNodeViewpoint* vp)
{
  if (vp == NULL || vp->scene() != this) return;
  VrmlSFBool flag(true);
  vp->eventIn(d_time, "set_bind", &flag );
  setViewpointChanged(true);
}

void VrmlScene::bindSceneDefaultBindableNodes(int bindableNodesMask)
{
  VrmlSFBool flag(true);
  VrmlNode *bindable = 0;       // compiler warning

  if ((VRML_NODE_BACKGROUND & bindableNodesMask) && d_backgrounds.size() > 0 && (bindable = d_backgrounds.front()) != 0)
    bindable->eventIn(d_time, "set_bind", &flag);

  if ((VRML_NODE_FOG & bindableNodesMask) && d_fogs.size() > 0 && (bindable = d_fogs.front()) != 0)
    bindable->eventIn(d_time, "set_bind", &flag);

  if ((VRML_NODE_NAVIGATION_INFO & bindableNodesMask) && d_navigationInfos.size() > 0 && (bindable = d_navigationInfos.front()) != 0)
    bindable->eventIn(d_time, "set_bind", &flag);

  if ((VRML_NODE_VIEWPOINT & bindableNodesMask) && d_viewpoints.size() > 0 && (bindable = d_viewpoints.front()) != 0)
    bindable->eventIn(d_time, "set_bind", &flag);
}

//---------------------------------------------------------------------------

void VrmlScene::updateFlags()
{
  dynamic_cast<VrmlNode*>(&d_nodes)->updateModified(0x002);
}

void VrmlScene::registerViewer(Viewer* viewer)
{
  // do nothing for closed scene
  if (openCount<=0) return;

  viewers.push_back(viewer);
}

void VrmlScene::unregisterViewer(Viewer* viewer)
{
  // set another primary viewer
  // if none, this invalidates precompiled objects
  setAnotherPrimaryViewer(viewer);

  // remove the viewer
  viewers.remove(viewer);
}

void VrmlScene::setAnotherPrimaryViewer(Viewer* viewer)
{
  // just reset the primary viewer if we are destroying the scene and we are being called
  // as a side-effect of destroying all its viewers
  if (openCount<=0)
  {
    primaryViewer = NULL;
    return;
  }

  // if this is the primary viewer, set another one as primary
  if (viewer == primaryViewer)
  {
    Viewer* other = getAnyOtherViewer(viewer);
    // if there is valid other viewer and the scene owns the global current viewer
    // change both primary and global current viewer
    if (other && Viewer::getCurrent()==viewer)
      other->setCurrent();
    // otherwise change just the primary viewer for this scene (even to NULL)
    else
      setPrimaryViewer(other);
  }
}

/**
 * Returns the number of viewers looking at this scene.
 */
size_t VrmlScene::getViewerCount()
{
  size_t count = 0;
  for (ViewerList::iterator it=viewers.begin(); it!=viewers.end(); ++it)
    if ((*it)->isValid())
      count++;
  return(count);
}

Viewer* VrmlScene::getAnyOtherViewer(Viewer* v)
{
  for (ViewerList::iterator it=viewers.begin(); it!=viewers.end(); ++it)
    if (*it != v && (*it)->isValid())
      return(*it);
  return(NULL);
}

void VrmlScene::deleteRelatedVrmlSceneFromNodeType(const std::string& id)
{
  RelatedScenesFromNodeTypeContainer::iterator it = relatedScenesFromNodeTypeContainer.begin();
  while ( it != relatedScenesFromNodeTypeContainer.end() )
    if (id.compare(it->second->getNodeType()->getName())==0)
    {
      delete it->second;
      relatedScenesFromNodeTypeContainer.erase(it);
      break;
    }
    else
      ++it;
}

VrmlScene* VrmlScene::getRelatedVrmlSceneFromNodeType(const std::string& id)
{
  RelatedScenesFromNodeTypeContainer::const_iterator it;
  for (it = relatedScenesFromNodeTypeContainer.begin(); it != relatedScenesFromNodeTypeContainer.end(); ++it)
    if (id.compare(it->second->getNodeType()->getName())==0)
      return it->second;

  return 0;
}

void VrmlScene::setTime(double newTime)
{
  // store offset to real time
  if (d_realTimeSynced)
    d_realTimeOffset = theSystem->time() - newTime;

  // update time
  d_prevTime = d_time;         // previous time
  d_time = newTime;

  if (d_recordLog != NULL)
    d_recordLog->logTimes.push_back(d_time);

  // notify the viewers
  for (ViewerList::iterator it = viewers.begin(); it != viewers.end(); ++it)
    (*it)->onSceneTimeChange();

  // start scheduled recording if requested
  if (d_recordingScheduled
      && d_time == d_recordingStartTime
      || (d_time > d_recordingStartTime && d_prevTime < d_recordingStartTime))
  {
    startRecording();
  }

  // if recording, notify the viewers to record this frame
  if (d_recording)
  {
    for (ViewerList::iterator it = viewers.begin(); it != viewers.end(); ++it)
      (*it)->onRecordingFrame();
  }

  // stop scheduled recording if requested (there exists possibility to record last time step of simulation)
  if (d_recordingScheduled
      && (d_time >= d_recordingStopTime && d_prevTime < d_recordingStopTime))
  {
    stopRecording();
  }
}



void VrmlScene::setPrimaryViewer(Viewer* viewer)
{
  // do nothing for closed scene
  if (openCount<=0) return;

  if (viewer==primaryViewer) return;  // already primary

  // invalidate scene if closing the primary viewer (new primary viewer is 0)
  if (viewer==0) invalidate();

  primaryViewer = viewer;

  // change viewpoint to that of the new primary viewer
  if (primaryViewer && primaryViewer->viewpoint)
  {
    VrmlSFBool flag(true);
    primaryViewer->viewpoint->eventIn(d_time, "set_bind", &flag);
  }
}



//---------------------------------------------------------------------------
// node management
//---------------------------------------------------------------------------



void VrmlScene::removeNodeRefs(VrmlNode *parent, VrmlNode *victim)
{
  VrmlNodeType *type;
  std::list<VrmlNodeType::ProtoField*>* fieldList;

  // scan the fields of the parent, remove references to the victim node,
  // call this method recursively for all children of the parent
  type = &parent->nodeType();
  fieldList = &type->fields();
  for (std::list<VrmlNodeType::ProtoField*>::iterator it = fieldList->begin(); it != fieldList->end(); ++it)
  {
    if ((*it)->type == VrmlField::SFNODE)
    {
      VrmlSFNode *field;

      field = (VrmlSFNode *) parent->getField((*it)->name);
      if (field->get() == victim) field->set(NULL);
      else if (field->get() != NULL) removeNodeRefs(field->get(), victim);
    }
    else if ((*it)->type == VrmlField::MFNODE)
    {
      VrmlMFNode *field;

      field = (VrmlMFNode *) parent->getField((*it)->name);
      field->removeNode(victim);
      int size = field->getLength();
      for (int i=0; i<size; ++i) removeNodeRefs((*field)[i], victim);
    }
  }
}



VrmlNode* VrmlScene::findNode(const char* name)
{
  if (name == NULL) return NULL;
  return (d_namespace != NULL) ? d_namespace->findNode(name) : NULL;
}



void VrmlScene::removeNode(VrmlNode* node)
{
  if (node == NULL) return;

  // reference the node so it doesn't get deleted during RemoveNodeRefs
  node->reference();

  // remove all references to this node in the whole tree
  removeNodeRefs(&d_nodes, node);

  // remove the node itself
  node->dereference();
}



//---------------------------------------------------------------------------
// recording
//---------------------------------------------------------------------------



void VrmlScene::startRecording()
{
  // do nothing if recording is already in progress
  if (d_recording) return;

  // set the flag
  d_recording = true;

  // remember the simulation time
  d_lastRecordingStartTime = d_time;

  if (d_recordToVRML)
  {
    // open the file
    d_recordRealFileName = theSystem->generateFile(d_recordFileName,
                                                   d_url->url());
    if (d_recordRealFileName==NULL)
    {
      theSystem->warn(System::SEVERITY_HIGH, System::WARNINGTYPE_UNKNOWN,
                      "Can't open VRML record file, recording suppressed.");
    }
  }

  // inform the viewers
  for (ViewerList::iterator it = viewers.begin(); it != viewers.end(); ++it)
    (*it)->onRecordingStart();
}



void VrmlScene::stopRecording()
{
  // do nothing if no recording is done
  if (!d_recording) return;

  // set the flag
  d_recording = false;

  // remember the simulation time
  d_lastRecordingStopTime = d_time;

  // store the VRML record if any
  if (d_recordToVRML)
  {
    // save the record
    saveRecord();
    clearRecord();

    delete[] d_recordRealFileName;
    d_recordRealFileName = NULL;
  }

  // inform the viewers
  for (ViewerList::iterator it = viewers.begin(); it != viewers.end(); ++it)
    (*it)->onRecordingStop();
}



void VrmlScene::logChange(const VrmlNode* node, const char* fieldname,
  const VrmlField* value)
{
  if (d_recording && d_recordToVRML)
  {
    if (d_recordLog == NULL) d_recordLog = new SceneChangeLog();

    std::string logEntryStr(node->name());
    logEntryStr.append(".");
    if (node->nodeType().hasExposedField(fieldname))
      logEntryStr.append(SceneChangeLog::EXPOSED_FIELD_TOKEN);	// remember that it is exposed field
    logEntryStr.append(fieldname);

    d_recordLog->addEntry(d_recordLog->logTimes.size(), logEntryStr.c_str(), value);
  }
}



void VrmlScene::saveRecord(ostream &os) const
{
  // Save the scene itself
  ((VrmlScene*) this)->save(os);

  // do nothing else if nothing was recorded
  if (d_recordLog == NULL) return;

  // calculate the duration of the animation (the last event time
  // is used instead of the current time, because saveRecord() can be called
  // after the time is already set for the new frame)
  double lastRecordedTime = d_recordLog->logTimes.empty() ? 0 : d_recordLog->logTimes.back();
  double duration = lastRecordedTime - d_lastRecordingStartTime;
  if (duration <= 0) return;  // can be negative if stopped in the same frame

  os << endl << "###### Recorded animation START ######" << endl << endl;

  // Build a timer that controls the replay of the recorded data.
  VrmlNodeTimeSensor ts;
  ts.setName("Replay_control");
  ts.setField("cycleInterval", VrmlSFTime(duration));
  ts.setField("loop", VrmlSFBool(true));
  os << ts << endl;

  // Scan through all logged changes. For every field that was changed
  // we produce an interpolator that contains the list of values
  // that field had through the recorded period, and a route
  // for replaying these values to the field.
  for (int i=0; i<d_recordLog->fieldNames.size(); ++i)
  {
    // look which type the field is
    FieldChangeLog* log = d_recordLog->fieldChanges[i];
    FieldChangeLogEntry* entry = (*log)[0];
    VrmlField::VrmlFieldType type = entry->value->fieldType();

    // generate a name for the interpolator
    std::string ipolName(d_recordLog->fieldNames[i]);
    ipolName.append("_recorded");
    size_t found = 0;
    while ((found=ipolName.find('.', found))!=std::string::npos)
      ipolName.replace(found,1,1,'_');

    if ((found=ipolName.find(SceneChangeLog::EXPOSED_FIELD_TOKEN))!=std::string::npos)
      ipolName.erase(found,strlen(SceneChangeLog::EXPOSED_FIELD_TOKEN));


    // replace exposed field token by "set_" prefix (it results in eventIn)
    if ((found=d_recordLog->fieldNames[i].find(SceneChangeLog::EXPOSED_FIELD_TOKEN))!=std::string::npos)
      d_recordLog->fieldNames[i].replace(found,strlen(SceneChangeLog::EXPOSED_FIELD_TOKEN), "set_");

    // build a list of time values when the value should change
    VrmlMFFloat keys;
    keys.setLength(log->size());
    for (int j=0; j<log->size(); ++j)
    {
      // use indexed time or last indexed time if index does not exist
      size_t timeIndex = std::min((*log)[j]->timeIndex, d_recordLog->logTimes.size()-1);
      keys.setElement(j, (d_recordLog->logTimes[timeIndex] - d_lastRecordingStartTime)/duration);
    }

    // build an appropriate interpolator
    VrmlNode* ipol = NULL;
    if (type == VrmlField::SFFLOAT)
    {
      ipol = new VrmlNodeScalarInt();
      VrmlMFFloat keyValues;
      keyValues.setLength(log->size());
      for (int j=0; j<log->size(); ++j)
      {
        keyValues.setElement(j, ((VrmlSFFloat*) (*log)[j]->value)->get());
      }
      ipol->setField("keyValue", keyValues);
    }
    else if (type == VrmlField::SFINT32)
    {
      // interpolators for ints do not exist so we use a scalar interpolator
      // and add a small script that converts the values (see below)
      ipol = new VrmlNodeScalarInt();
      VrmlMFFloat keyValues;
      keyValues.setLength(log->size());
      for (int j=0; j<log->size(); j++)
      {
        keyValues.setElement(j, (float) ((VrmlSFInt32*) (*log)[j]->value)->get());
      }
      ipol->setField("keyValue", keyValues);
    }
    else if (type == VrmlField::SFBOOL)
    {
      // again, there is no such thing as interpolator for booleans,
      // we use a scalar interpolator and convert later
      ipol = new VrmlNodeScalarInt();
      VrmlMFFloat keyValues;
      keyValues.setLength(log->size());
      for (int j=0; j<log->size(); ++j)
      {
        keyValues.setElement(j, (float) ((VrmlSFBool*) (*log)[j]->value)->get());
      }
      ipol->setField("keyValue", keyValues);
    }
    else if (type == VrmlField::SFVEC3F)
    {
      ipol = new VrmlNodePositionInt();
      VrmlMFVec3f keyValues;
      keyValues.setLength(log->size());
      for (int j=0; j<log->size(); ++j)
      {
        keyValues.setElement(j,
          ((VrmlSFVec3f*) (*log)[j]->value)->getX(),
          ((VrmlSFVec3f*) (*log)[j]->value)->getY(),
          ((VrmlSFVec3f*) (*log)[j]->value)->getZ());
      }
      ipol->setField("keyValue", keyValues);
    }
    else if (type == VrmlField::SFCOLOR)
    {
      ipol = new VrmlNodeColorInt();
      VrmlMFColor keyValues;
      keyValues.setLength(log->size());
      for (int j=0; j<log->size(); ++j)
      {
        keyValues.setElement(j,
          ((VrmlSFColor*) (*log)[j]->value)->getR(),
          ((VrmlSFColor*) (*log)[j]->value)->getG(),
          ((VrmlSFColor*) (*log)[j]->value)->getB());
      }
      ipol->setField("keyValue", keyValues);
    }
    else if (type == VrmlField::SFROTATION)
    {
      ipol = new VrmlNodeOrientationInt();
      VrmlMFRotation keyValues;
      keyValues.setLength(log->size());
      for (int j=0; j<log->size(); ++j)
      {
        keyValues.setElement(j,
          ((VrmlSFRotation*) (*log)[j]->value)->getX(),
          ((VrmlSFRotation*) (*log)[j]->value)->getY(),
          ((VrmlSFRotation*) (*log)[j]->value)->getZ(),
          ((VrmlSFRotation*) (*log)[j]->value)->getAngle());
      }
      ipol->setField("keyValue", keyValues);
    }

    // skip fields that are unsupported
    if (ipol == NULL)
    {
      theSystem->warn(System::SEVERITY_HIGH, System::WARNINGTYPE_UNKNOWN,
                      "Recording for field %s not supported.",
                      d_recordLog->fieldNames[i].c_str());
      continue;
    }

    // finish the interpolator, print it out, then remove it
    ipol->setName(ipolName.c_str());
    ipol->setField("key", keys);
    os << *ipol << endl;
    delete ipol;


    // add a ROUTE to the replay control
    os << "ROUTE Replay_control.fraction_changed TO ";
    os << ipolName << ".set_fraction" << endl;

    // Add an appropriate ROUTE between the interpolator and the field.
    // Sometimes we don't have an interpolator for the value type we need
    // (e.g. ints, booleans) so we have to use a different type and convert.
    // In that case we use a small convertor script.
    if (type == VrmlField::SFINT32)
    {
      // add a script that converts floats to ints
      os << "DEF " << ipolName << "_converter Script {" << endl;
      os << "  eventIn SFFloat i" << endl;
      os << "  eventOut SFInt32 o" << endl;
      os << "  url \"vrmlscript: function i(value) { o = value; }\"" << endl;
      os << "}" << endl;

      // add ROUTEs using the script for conversion
      os << "ROUTE " << ipolName << ".value_changed TO ";
      os << ipolName << "_converter.i" << endl;
      os << "ROUTE " << ipolName << "_converter.o TO ";
      os << d_recordLog->fieldNames[i] << endl;
    }
    else if (type == VrmlField::SFBOOL)
    {
      // add a script that converts floats to booleans
      os << "DEF " << ipolName << "_converter Script {" << endl;
      os << "  eventIn SFFloat i" << endl;
      os << "  eventOut SFBool o" << endl;
      os << "  url \"vrmlscript: function i(value) { o = (value!=0); }\"" << endl;
      os << "}" << endl;

      // add ROUTEs using the script for conversion
      os << "ROUTE " << ipolName << ".value_changed TO ";
      os << ipolName << "_converter.i" << endl;
      os << "ROUTE " << ipolName << "_converter.o TO ";
      os << d_recordLog->fieldNames[i] << endl;
    }
    else
    {
      // no conversion needed, use a simple ROUTE
      os << "ROUTE " << ipolName << ".value_changed TO ";
      os << d_recordLog->fieldNames[i] << endl;
    }
  }

  os << endl << "###### Recorded animation END ######" << endl << endl;
}



bool VrmlScene::saveRecord()
{
  // do nothing if the record file name is not specified
  if (d_recordRealFileName == NULL) return false;

  // create the output stream
  // the Win32 special case is needed only for VC 8 and locale not set (as in R2007a)
#ifdef _WIN32
  FILE* fptr = fopen(d_recordRealFileName, "w");
  if (!fptr) return(false);
  ofstream f(fptr);
#else
  ofstream f(d_recordRealFileName);
#endif

  // save it
  saveRecord(f);
  f.close();
  return (f.good());
}



void VrmlScene::clearRecord()
{
  delete d_recordLog;
  d_recordLog = NULL;
}



void VrmlScene::setRecordFileName(const char* name)
{
  delete[] d_recordFileName;
  if (name == NULL) d_recordFileName = NULL;
  else
  {
    int len = strlen(name);
    d_recordFileName = new char[len+1];
    strcpy(d_recordFileName, name);
  }
}

void VrmlScene::guessNearFarPlane(Viewer* viewer, double* nearVal, double* farVal)
{
  // step into the top level group
  d_nodes.guessNearFarPlane(viewer, nearVal, farVal );
}

void VrmlScene::setRecordToVRML(bool value)
{
  d_recordToVRML = value;
  for (ViewerList::iterator it = viewers.begin(); it != viewers.end(); ++it)
    (*it)->onRecordToVrmlFlagChanged();
}



// SceneChangeLog and friends



static char* newstr(const char* str)
{
  int len = strlen(str);
  char* result = new char[len+1];
  strcpy(result, str);
  return result;
}



FieldChangeLogEntry::FieldChangeLogEntry(size_t _timeIndex, const VrmlField* _value)
  : timeIndex(_timeIndex)
{
  value = _value->clone();
}



FieldChangeLogEntry::~FieldChangeLogEntry()
{
  delete value;
}



SceneChangeLog::~SceneChangeLog()
{
}



void SceneChangeLog::addEntry(size_t timeIndex, const char* field,
  const VrmlField* value)
{
  int i;

  /*
   * Look if the field name is already in the list.
   */
  for (i=0; i<fieldNames.size(); i++)
  {
    if (fieldNames[i].compare(field) == 0)
    {
      /*
       * It is in the list, add the new entry to the existing log.
       */
      fieldChanges[i]->push_back(new FieldChangeLogEntry(timeIndex, value));
      return;
    }
  }

  /*
   * The field is not in the list, add it and create a new log.
   */
  fieldNames.push_back(field);
  fieldChanges.push_back(new FieldChangeLog());
  fieldChanges[i]->push_back(new FieldChangeLogEntry(timeIndex, value));
}
