/*****************************************************************************
******************************************************************************
*
*               VRML scene (aka world) class
*
*               $Revision: 1.1.2.27 $
*               $Date: 2012/03/11 21:21:52 $
*               $Author: Xjhouska $
*
******************************************************************************
*****************************************************************************/

// Copyright 1998-2012 HUMUSOFT s.r.o.
// Portions Copyright (C) 1998  Chris Morley

//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#ifndef VRMLSCENE_H
#define VRMLSCENE_H

#include <list>
#include <string>
#include <sstream>
using namespace std;

#include "RecordingInfo.h"
#include "System.h" // OPENVRMLDLLEXPORT
#include "HandleFactory.h"

// The loaders fill in a Group node
#include "VrmlNodeGroup.h"
#include "VrmlNodeProto.h"
#include "VrmlEvent.h"

// Bindable children nodes
#include "VrmlNodeBackground.h"
#include "VrmlNodeFog.h"
#include "VrmlNodeNavigationInfo.h"
#include "VrmlNodeViewpoint.h"

// Non-bindable scene-level children nodes
#include "VrmlNodePointLight.h"
#include "VrmlNodeSpotLight.h"
#include "VrmlNodeMovieTexture.h"
#include "VrmlNodeAudioClip.h"
#include "VrmlNodeScript.h"
#include "VrmlNodeTimeSensor.h"

// Container for viewers
typedef std::list<Viewer*> ViewerList;

class VrmlScene;

// Container for related VrmlSceneFromNodeType instances
typedef HandleFactory<VrmlScene> RelatedScenesFromNodeTypeContainer;

class Doc;
class Viewer;

// PROTO definitions
class VrmlNamespace;
class VrmlNodeType;

// Bindable children node types
class VrmlNodeBackground;
class VrmlNodeFog;
class VrmlNodeNavigationInfo;
class VrmlNodeViewpoint;

// Lists of other node types needed for rendering or updating the scene
class VrmlNodeAudioClip;
class VrmlNodeLight;
class VrmlNodeMovieTexture;
class VrmlNodeScript;
class VrmlNodeTimeSensor;

struct SceneChangeLog;

/**
 * Access flags for the scene.
 */
struct SceneAccessFlags
{
  bool Viewable : 1;
  bool RemoteAccess : 1;
  bool RedrawLock : 1;

  SceneAccessFlags() :
    Viewable(false), RemoteAccess(false), RedrawLock(false)
  {};
};


// Bindable children nodes (pseudo-stacks - allow random access deletion).
class BindableStack : public std::list<VrmlNode*>
{
  public:

  // return a node on top of the stack
  VrmlNode* top()
  { return(empty() ? 0 : front()); };

  // push a node on stack top
  void push(VrmlNode* node);

  // remove a node from stack
  void remove(VrmlNode* node);

  // remove all nodes from stack
  void removeAll()
  { clear(); }
};


/**
 * This object represents a VRML scene graph.
 */
class OPENVRMLDLLEXPORT VrmlScene
{
friend class BindableStack;
friend class Viewer;
public:

  /**
   * VRML bindable nodes enumeration
   */
  enum BindableNodes {
    VRML_NODE_BACKGROUND      = 1,
    VRML_NODE_FOG             = 2,
    VRML_NODE_NAVIGATION_INFO = 4,
    VRML_NODE_VIEWPOINT       = 8,
  };

  /**
   * Loads VRML nodes from one of the given URLs.
   * All URLs are tried in the given order until one succeeds.
   * The method throws an error only if all specified URLs fail
   * (the error then describes why the first, preferred URL has failed).
   * @param url A list of URLs.
   * @param relative The relative URL.
   * @param ns The namespace to register the newly loaded nodes in.
   * @return The newly loaded nodes.
   * @exception VrmlLoadingError
   */
  static VrmlMFNode * readWrl(VrmlMFString * url, Doc * relative,
                              VrmlNamespace * ns);

  /**
   * Reads a VRML file from the given URL, creating the appropriate node tree
   * and adding the nodes to the given namespace.
   * @return The root of the new node tree.
   * @exception VrmlLoadingError
   */
  static VrmlMFNode * readWrl(Doc * url, VrmlNamespace * ns );

  /**
   * Reads a VRML file from the given string, creating the appropriate node
   * tree and adding the nodes to the given namespace.
   * @return The root of the new node tree.
   * @exception VrmlLoadingError
   */
  static VrmlMFNode * readString(char const * vrmlString, VrmlNamespace * ns);

  static VrmlNodeType * readPROTO(VrmlMFString * url, Doc const * relative = 0);

  /**
   * Constructor.
   * @param url The URL to read the VRML file from.
   * @exception VrmlLoadingError
   */
  VrmlScene( const char *url = 0, bool edited = false);

  virtual ~VrmlScene();         // Allow overriding of Script API methods

  // Destroy world (just passes destroy request up to client)
  void destroyWorld();

  // Replace world with nodes, recording url as the source URL.
  void replaceWorld(VrmlMFNode & nodes, VrmlNamespace * ns,
                    Doc * url=0);

  /**
   * Loads a file specified by the given URLs.
   * If the file is a VRML file, the scene contents are replaced
   * with the contents of that file. Otherwise the processing depends
   * on file type.
   * All URLs are tried in the given order until one succeeds.
   * The method throws an error only if all specified URLs fail
   * (the error then describes why the first, preferred URL has failed).
   * @param url A list of URLs.
   * @param parameters
   * @exception VrmlLoadingError
   */
  void loadUrl(VrmlMFString const * url, VrmlMFString const * parameters = 0 );

  /**
   * Load the scene from a VRML file, replacing original contents (if any).
   * @param url The URL.
   * @param localCopy A file name of the local copy (if any).
   * @exception VrmlLoadingError
   */
  void load(const char *url, const char *localCopy = 0);

  /**
   * Load the scene using the associated URL.
   * @exception VrmlLoadingError
   */
  void load();

  /**
   * Reloads the scene from the original URL.
   */
  virtual bool reload(bool preserveBindings = true);

  /** Save the scene in VRML format to the given output stream. */
  void save(ostream& os);

  /** Generate simplified description of VRML scene and store it
   * to the given output stream.
   * @param os an output stream where the description should be stored.
   * @param editFlag true if chain is created for 3D World Editor purposes
   */
  void saveSimpleChain(ostream &os, bool editFlag);

  /**
   * Saves the scene to the given file.
   * @return True on success, false on failure.
   */
  bool save(const char* filename);

  /**
   * Returns a pointer to a Doc object representing the URL this scene
   * was loaded from.
   */
  virtual Doc* urlDoc();

  /**
   * Overriden in VrmlSceneFromNodeType.
   * It returns direct parent of the scene.
   * For VrmlScene instances returns this poitner.
   */
  virtual VrmlScene* getParent();


  /**
   * Overriden in VrmlSceneFromNodeType.
   * It returns topmost parent of the scene.
   * For VrmlScene instances returns this poitner.
   */
  virtual VrmlScene* getTopLevelParent();

  /**
   * Returns null for VrmlScene instances,
   * overriden in VrmlSceneFromNodeType where it returns
   * name of underlying VrmlNodeType.
   */
  virtual VrmlNodeType* getNodeType();

  /**
   * Returns a pointer to the root namespace of the scene.
   * The pointer is valid until the scene is reloaded or a new VRML file
   * is loaded.
   */
  virtual VrmlNamespace *scope();

  // Queue an event to load URL/nodes (async so it can be called from a node)
  void queueLoadUrl( VrmlMFString *url, VrmlMFString *parameters );
  void queueReplaceNodes( VrmlMFNode *nodes, VrmlNamespace *ns );

  void sensitiveEvent( void *object, double timeStamp,
                       bool isOver, bool isActive, double *point );

  // Script node API support functions. Can be overridden if desired.
  virtual const char *getName();
  virtual const char *getVersion();
  double getFrameRate();

  /**
   * Updates the scene, re-running scripts, dispatching events, etc.
   * This method should be called periodically to make the scene live.
   * Returns true if the scene views should be redrawn, false
   * if it is not necessary (no visible changes).
   */
  bool update();

  /**
   * Request redraw for all viewers of this scene.
   */
  void redrawViewers();

  void render(Viewer* viewer, bool final);

  // Indicate that the scene state has changed, need to re-render
  virtual void setModified();
  virtual void clearModified();
  virtual bool isModified() const;

  // Invalidates all nodes in scene
  void invalidate();

  // Indicate that the textures need to re-render
  void invalidateTextures() { d_texturesDirty = true; setModified(); }
  bool texturesDirty() { return d_texturesDirty; }


  // Bindable children nodes can be referenced via a list or bindable stacks.
  // Define for each bindableType:
  //    addBindableType(bindableType *);
  //    removeBindableType(bindableType *);
  //    bindableType *bindableTypeTop();
  //    void bindablePush(VrmlNodeType *);
  //    void bindableRemove(VrmlNodeType *);

  // Background
  void addBackground(VrmlNodeBackground* n)
  { d_backgrounds.push_back(n); }

  void removeBackground(VrmlNodeBackground* n)
  {
    bindableRemove(n);
    d_backgrounds.remove(n);
  }

  VrmlNodeBackground* bindableBackgroundTop()
  { return NodeDynamicCast<VrmlNodeBackground*>(d_backgroundStack.top()); }

  void bindablePush(VrmlNodeBackground* n)
  { d_backgroundStack.push(n); }

  void bindableRemove(VrmlNodeBackground* n)
  { d_backgroundStack.remove(n); }

  // Fog
  void addFog(VrmlNodeFog* n)
  { d_fogs.push_back(n); }

  void removeFog(VrmlNodeFog* n)
  {
    bindableRemove(n);
    d_fogs.remove(n);
  }

  VrmlNodeFog* bindableFogTop()
  { return NodeDynamicCast<VrmlNodeFog*>(d_fogStack.top()); }

  void bindablePush(VrmlNodeFog* n)
  { d_fogStack.push(n); }

  void bindableRemove(VrmlNodeFog* n)
  { d_fogStack.remove(n); }

  // NavigationInfo
  void addNavigationInfo(VrmlNodeNavigationInfo* n)
  { d_navigationInfos.push_back(n); }

  void removeNavigationInfo(VrmlNodeNavigationInfo* n)
  {
    bindableRemove(n);
    d_navigationInfos.remove(n);
  }

  VrmlNodeNavigationInfo* bindableNavigationInfoTop()
  { return NodeDynamicCast<VrmlNodeNavigationInfo*>(d_navigationInfoStack.top()); }

  void bindablePush(VrmlNodeNavigationInfo* n)
  { d_navigationInfoStack.push(n); }

  void bindableRemove(VrmlNodeNavigationInfo* n)
  { d_navigationInfoStack.remove(n); }

  //-------------------------------------------------------------------------
  // Viewpoints and other bindables
  //-------------------------------------------------------------------------

  /**
   * @internal
   * Adds the viewpoint to the internal viewpoint list.
   * This is called from the VrmlNodeViewpoint constructor
   * and from its addToScene() method.
   * @param vp The viewpoint.
   */
  void addViewpoint(VrmlNodeViewpoint* vp);

  /**
   * @internal
   * Removes the viewpoint from the internal viewpoint list.
   * This is called from the VrmlNodeViewpoint destructor
   * and from its addToScene() method.
   * @param vp The viewpoint.
   */
  void removeViewpoint(VrmlNodeViewpoint* vp);

  VrmlNodeViewpoint* bindableViewpointTop()
  { return NodeDynamicCast<VrmlNodeViewpoint*>(d_viewpointStack.top()); }

  void bindablePush(VrmlNodeViewpoint* n)
  { d_viewpointStack.push(n); }

  void bindableRemove(VrmlNodeViewpoint* n)
  { d_viewpointStack.remove(n); }

  /**
   * Finds a viewpoint by its description.
   * @param name The description.
   * @return The viewpoint pointer, or NULL if not found.
   * @note If description is NULL, the method returns NULL.
   */
  VrmlNodeViewpoint* findViewpointByDescription(const char* desc);

  /**
   * Finds a viewpoint by its VRML name.
   * @param name The VRML name.
   * @return The viewpoint pointer, or NULL if not found.
   * @note If name is NULL, the method returns NULL.
   */
  VrmlNodeViewpoint* findViewpointByName(const char* name);

  /**
   * Finds a viewpoint by its index.
   * @param index The index.
   * @return The viewpoint pointer, or NULL if not found.
   */
  VrmlNodeViewpoint* findViewpointByIndex(int n);

  /**
   * Returns the number of viewpoints in the scene.
   * @return The number of viewpoints.
   */
  size_t nViewpoints() { return d_viewpoints.size(); }

  /**
   * Returns the index of the viewpoint in the scene.
   * @param vp The viewpoint.
   * @return The index of the viewpoint, or -1 if not found.
   */
  int getViewpointIndex(VrmlNodeViewpoint* vp);

  /**
   * Returns the next viewpoint in the viewpoint list.
   * The list is circular so the first viewpoint follows after the last one.
   * @param vp Pointer to the viewpoint, or NULL to get the first viewpoint.
   * @return Pointer to the next viewpoint in the list, or NULL
   * if the given viewpoint is not found in the viewpoint list,
   * or if there are no viewpoints in the scene.
   */
  VrmlNodeViewpoint* getNextViewpoint(VrmlNodeViewpoint* vp);

  /**
   * Returns the previous viewpoint in the viewpoint list.
   * The list is circular so the last viewpoint precedes the first one.
   * @param vp Pointer to the viewpoint, or NULL to get the last viewpoint.
   * @return Pointer to the previous viewpoint in the list, or NULL
   * if the given viewpoint is not found in the viewpoint list,
   * or if there are no viewpoints in the scene.
   */
  VrmlNodeViewpoint* getPrevViewpoint(VrmlNodeViewpoint* vp);

  /**
   * Activates the given viewpoint (moves it to the top of the stack).
   * @param vp A pointer to the viewpoint.
   * @note The method has no effect if the given viewpoint is NULL
   * or does not belong to the scene.
   */
  void setViewpoint(VrmlNodeViewpoint* vp);

  //-------------------------------------------------------------------------

  // Other (non-bindable) node types that the scene needs access to:

  // Scene-scoped lights
  void addScopedLight(VrmlNodeLight* light)
  { d_scopedLights.push_back(light); }

  void removeScopedLight(VrmlNodeLight* light)
  { d_scopedLights.remove(light); }

  // Scripts
  void addScript(VrmlNodeScript* script)
  { d_scripts.push_back(script); }

  void removeScript(VrmlNodeScript* script)
  { d_scripts.remove(script); }

  // TimeSensors
  void addTimeSensor(VrmlNodeTimeSensor* timer)
  { d_timers.push_back(timer); }

  void removeTimeSensor(VrmlNodeTimeSensor* timer)
  { d_timers.remove(timer); }

  // AudioClips
  void addAudioClip( VrmlNodeAudioClip* audio_clip)
  { d_audioClips.push_back(audio_clip); }

  void removeAudioClip( VrmlNodeAudioClip* audio_clip)
  { d_audioClips.remove(audio_clip); }

  // MovieTextures
  void addMovie(VrmlNodeMovieTexture* movie)
  { d_movies.push_back(movie); }

  void removeMovie(VrmlNodeMovieTexture* movie)
  { d_movies.remove(movie); }

  /**
   * True if the bvolume dirty flag has been set on a node in the
   * scene graph, but has not yet been propegated to that node
   * ancestors. Set by VrmlNode::setBVolumeDirty on any node in this
   * scene graph, cleared by updateFlags()
   */
  bool d_flags_need_updating;

  /**
   * Propegate the bvolume dirty flag from children to ancestors. The
   * invariant is that if a node bounding volume is out of date,
   * then the bounding volumes of all that nodes ancestors must be
   * out of date. However, VrmlNode does not maintain a parent
   * pointer. So we must do a traversal of the entire scene graph to
   * do the propegation.
   *
   * @see VrmlNode::setBVolumeDirty
   * @see VrmlNode::isBVolumeDirty
   */
  void updateFlags();

  bool headlightOn();

  /** Checks whether the viewpoint was changed.
    * @returns True if viewpoint was changed, false otherwise.
    */
  bool wasViewpointChanged() const { return d_newView; };

  /** Sets or resets the flag that specifies whether a viewpoint
    * was changed. The flag is reset when GUI approperiately changed.
    * @param onoff True to set the flag, false to reset it.
    */
  void setViewpointChanged(bool onoff) { d_newView = onoff; };

  /** Sets primary viewer.
    * Primary viewer is the only viewer that can affect the scene,
    * e.g. trigger sensors, change viewpoint stack etc.
    * Other viewers just look at the scene but don't affect it.
    * @param viewer Pointer to viewer to be set as primary,
    * or NULL to leave the scene without a primary viewer.
    */
  void setPrimaryViewer(Viewer* viewer);

  /**
   * Returns a pointer to a primary viewer currently in use.
   * Returns NULL if there is no primary viewer.
   */
  Viewer* getPrimaryViewer() { return primaryViewer; };

  /**
   * Returns the number of viewers looking at this scene.
   */
  size_t getViewerCount();

  /** Returns a pointer to another viewer looking at this scene.
    * (I admit this one *is* a bit weird.)
    */
  Viewer* getAnyOtherViewer(Viewer* v);

  /**
   * Sets primary viewer distinct from given one
   */
  void setAnotherPrimaryViewer(Viewer* viewer);

  /** Returns the open count of the scene. If it is 0,
    * scene contents may not be loaded in memory.
    */
  int getOpenCount() const { return openCount; }

  /** Increments the scene open count by one.
    * Overriden by VrmlSceneFromNodeType. */
  virtual void open();

  /** Decrements the scene open count by one.
    * Overriden by VrmlSceneFromNodeType. */
  virtual void close();

  /** Returns the scene description. */
  const char* desc()
    { return description.c_str(); };

  /** Sets scene description. */
  void setDesc(const char* newDesc)
    { description.assign(newDesc); };

  /**
   * Returns the current value of access bits.
   * Access bits can be used by an external security monitor to limit
   * access to the scene. The scene itself does not use them.
   * The access bits value is set to Viewable when the scene is created.
   */
  SceneAccessFlags getAccessFlags() const { return d_accessFlags; }

  /**
   * Sets access bits to the given value.
   */
  void setAccessFlags(SceneAccessFlags value) { d_accessFlags = value; }

  /**
   * Returns the pointer to the root of the nodes hierarchy.
   */
  virtual VrmlNodeGroup* nodes();

  /**
   * Returns the current scene time.
   */
  double getTime() const { return d_time; }

  /**
   * Sets the scene time.
   * @param newTime The new time.
   */
  void setTime(double newTime);

  /**
   * Returns true if the scene time is automatically updated
   * in sync with the real time, false if it is not.
   */
  bool isRealTimeSynced() const { return d_realTimeSynced; }

  /**
   * Switches the automatical updating of the scene time on or off.
   */
  void setRealTimeSynced(bool value);

  /**
   * Trigger which triggers "set_bind" events for default bindable nodes in the scene
   * (nodes which were added to the scene first).
   * @param bindableNodesMask enumerates which bindable nodes should be bind
   */
   void bindSceneDefaultBindableNodes(int bindableNodesMask);

private:

  void doCallbacks( int reason );

  // Document URL
  Doc * d_url;

  // Scene graph
  VrmlNodeGroup d_nodes;

  // Nodes and node types defined in this scope
  VrmlNamespace *d_namespace;

  /** Scene description string, used for better identification of the scene
    * to the user. This may be an empty string but not NULL.
    */
  std::string description;

  /**
   * Initial viewpoint
   */
  std::string d_initialViewpoint;

  // Need render
  bool d_modified;

  // true if scene open in editting mode
  bool d_edited;

  /** True if textures need re-rendering (resolution or antialiasing
    * have changed).
    */
  bool d_texturesDirty;

  // New viewpoint has been bound
  bool d_newView;

  // Offset to system time
  double d_realTimeOffset;

  // Allow requests to load urls, nodes to wait until events are processed
  VrmlMFString *d_pendingUrl;
  VrmlMFString *d_pendingParameters;

  VrmlMFNode *d_pendingNodes;
  VrmlNamespace *d_pendingScope;

  // frame rate
  double d_frameRate;

  //   Background
  std::list<VrmlNode*> d_backgrounds;           // All backgrounds
  BindableStack        d_backgroundStack;       // Background stack

  //   Fog
  std::list<VrmlNode*> d_fogs;                  // All fog nodes
  BindableStack        d_fogStack;              // Fog stack

  //   NavigationInfo
  std::list<VrmlNode*> d_navigationInfos;       // All navigation info nodes
  BindableStack        d_navigationInfoStack;   // Navigation info stack

  //   Viewpoint
  std::list<VrmlNode*> d_viewpoints;            // All viewpoint nodes
  BindableStack        d_viewpointStack;        // Viewpoint stack

  /** Pointer to the primary viewer. The primary viewer controls the scene,
    * i.e. it can change the scene viewpoint stack, affect sensors, etc.
    * Other viewers do not affect the scene, they only look at it.
    * This field can be NULL if there is currently no primary viewer.
    */
  Viewer* primaryViewer;

  /** Set of viewers currently viewing this scene.
    */
  ViewerList viewers;

  // An event has a value and a destination, and is associated with a time
  typedef struct {
    double timeStamp;
    VrmlField *value;
    VrmlNode *toNode;
    char *toEventIn;
  } Event;

protected:

  // Scene which is current clipboard owner
  static VrmlScene* clipboardOwner;

  // Clipboard for cut/copy/paste opeations in scene
  VrmlSFNode d_clipboard;

  /** Set of related VrmlSceneFromNodeType instances.
    */
  RelatedScenesFromNodeTypeContainer  relatedScenesFromNodeTypeContainer;

  /** Counts how many times the scene was opened.
    * If it is 0, scene contents may not be loaded in memory.
    */
  int openCount;

private:

  // Scene-scoped lights (PointLights and SpotLights)
  std::list<VrmlNodeLight*> d_scopedLights;

  // Scripts in this scene
  std::list<VrmlNodeScript*> d_scripts;

  // Time sensors in this scene
  std::list<VrmlNodeTimeSensor*> d_timers;

  // Audio clips in this scene
  std::list<VrmlNodeAudioClip*> d_audioClips;

  // Movies in this scene
  std::list<VrmlNodeMovieTexture*> d_movies;

  // true if the scene is marked for destruction and is about to be deleted
  // after it lost clipboard ownership
  bool d_markedForDestruction;

  /**
   * Access bits for use by an external security manager.
   * Not used by any method except getAccessBits() and setAccessBits().
   */
  SceneAccessFlags d_accessFlags;

  /**
   * Current time for the scene.
   */
  double d_time;

  /**
   * Previous time for the scene (time before setTime() was last called).
   */
  double d_prevTime;

  /**
   * Time of the last update, in real world time.
   */
  double d_lastUpdateTime;

  /**
   * True if the scene time should be automatically updated in sync with
   * the real time, false if not. If this is false, the external code
   * is responsible to update the time as needed.
   */
  bool d_realTimeSynced;

public:

  /**
   * Returns true if scene is opened in scene editor
   */
  virtual bool isEdited();

  /**
   * Getter for InitialViewpoint.
   * This property is not saved to the VRML file
   * and serves only for SL3D purposes.
   * If InitialViewpoint is not empty, then it is bound
   * as initial viewpoint when the scene is opened for the first time.
   */
  const std::string& getInitialViewpoint() const { return d_initialViewpoint; }

  /**
   * Setter and setter for InitialViewpoint.
   */
  void setInitialViewpoint(const std::string& initialViewpoint)
  { d_initialViewpoint = initialViewpoint; }

  //-------------------------------------------------------------------------
  // node management
  //-------------------------------------------------------------------------

  protected:

    /**
     * Removes all references to the specified node, starting
     * with the given parent node.
     * @param parent The parent node.
     * @param victim The node to remove the references to.
     */
    void removeNodeRefs(VrmlNode *parent, VrmlNode *victim);

  public:

    /**
     * Finds a node by its name.
     * @param name The node name.
     * @return A pointer to the node, or NULL if the node was not found.
     * @note If the node name is NULL, the method returns NULL.
     */
    virtual VrmlNode* findNode(const char* name);

    /**
     * Removes the given node from the scene.
     * @param node A pointer to the node.
     * @note If the node pointer specified is NULL, the method does nothing.
     */
    virtual void removeNode(VrmlNode* node);

  //-------------------------------------------------------------------------
  // recording API
  //-------------------------------------------------------------------------

  private:

    /** True if recording is currently active, false if not. */
    bool d_recording;

    /** True if recording to VRML is enabled, false if not. */
    bool d_recordToVRML;

    /** True if recording should automatically start and stop. */
    bool d_recordingScheduled;

    /** Simulation time when recording should be started automatically. */
    double d_recordingStartTime;

    /** Simulation time when recording should be stopped automatically. */
    double d_recordingStopTime;

    /** Simulation time when recording last started. */
    double d_lastRecordingStartTime;

    /** Simulation time when recording last stopped. */
    double d_lastRecordingStopTime;

    /** A log that holds all scene changes for VRML recording purposes. */
    SceneChangeLog* d_recordLog;

    /**
     * Name of file to store recorded scene animation (NULL if unset).
     * The file name can contain tokens like '%d' which will be replaced
     * by appropriate values when the file will be created.
     */
    char* d_recordFileName;

    /**
     * The record file name with all tokens resolved. This field
     * is initialized when the appropriate file is created,
     * on the start of VRML recording and is valid only during the recording.
     */
    char* d_recordRealFileName;

    /**
     * Clears the log, removing the stored animation.
     */
    void clearRecord();

  public:

    /**
      *	Returns contents of scene clipboard.
      */
    VrmlNode* getClipboardContents() { return d_clipboard.get(); }

    /**
      *	Sets new contents to clipboard.
      */
    void setClipboardContents(VrmlNode* contents) { d_clipboard.set(contents); }

    /**
     *	Gets current clipboard owner
     */
    static VrmlScene* getClipboardOwner() { return clipboardOwner; }

    /**
     *	Sets current clipboard owner (and possibly delete previous clipboard owner)
     */
    static void setClipboardOwner(VrmlScene* newClipboardOwner) { clipboardOwner = newClipboardOwner; }

    /**
     * Mark scene as to be deleted after clipboard ownership lost
     */
    void markForDestruction() { d_markedForDestruction = true; }

    /**
     * Returns true when the scene is marked for destruction
     */
    bool isMarkedForDestruction() { return d_markedForDestruction; }

    /** Adds a new viewer to the viewer list. This is called from
      * the Viewer constructor.
      */
    void registerViewer(Viewer* viewer);

    /** Removes a viewer from the viewer list. This is called from
      * the Viewer destructor.
      */
    void unregisterViewer(Viewer* viewer);

    /** Returns VrmlScene pointer identified by given handle. */
    VrmlScene* getSceneForHandle(unsigned handle)
    { return relatedScenesFromNodeTypeContainer.get(handle); }

    /** Returns handle for given VrmlScene. */
    unsigned getHandleForScene(VrmlScene* scene)
    { return relatedScenesFromNodeTypeContainer.gethandle(scene); }

    /** Adds a scene to container which holds
      *	related VrmlSceneFromNodeType instances.
      */
    unsigned addRelatedVrmlSceneFromNodeType(VrmlScene* scene)
    { return relatedScenesFromNodeTypeContainer.add(scene); }

    /** Removes a scene from container which holds
      *	related VrmlSceneFromNodeType instances deletes it.
      */
    void deleteRelatedVrmlSceneFromNodeType(const std::string& id);

    /** Gets related VrmlSceneFromNodeType instance from
      * the container (if it is already there).
      */
    VrmlScene* getRelatedVrmlSceneFromNodeType(const std::string& id);


    /**
     * Logs the given field change.
     * @param nodename Name of the changed node.
     * @param fieldname Name of the changed field.
     * @param value The new value of the field.
     */
    void logChange(const VrmlNode* node,
                   const char* fieldname,
                   const VrmlField* value);

    /**
     * Checks whether recording is in progress right now.
     * @return True if recording is in progress, false if not.
     */
    bool isRecording() const { return d_recording; }

    /**
     * Checks whether recording to VRML is enabled.
     * @return True if recording to VRML is enabled, false if not.
     */
    bool getRecordToVRML() const { return d_recordToVRML; }

    /**
     * Enables or disabled recording to VRML.
     * @param value True to enable VRML recording, false to disable it.
     */
    void setRecordToVRML(bool value);

    /**
     * Checks whether scheduled recording is enabled.
     * @return True if scheduled recording is enabled, false if not.
     */
    bool isRecordingScheduled() const { return d_recordingScheduled; }

    /**
     * Enables or disables scheduled recording.
     * @param value True to enable scheduled recording, false to disable it.
     */
    void setRecordingScheduled(bool value) { d_recordingScheduled = value; }

    /**
     * Returns the scheduled recording start time.
     * @return The scheduled recording start time.
     */
    double getRecordingStartTime() const { return d_recordingStartTime; }

    /**
     * Sets the scheduled recording start time.
     * @param value The new value.
     */
    void setRecordingStartTime(double value) { d_recordingStartTime = value; }

    /**
     * Returns the scheduled recording stop time.
     * @return The scheduled recording stop time.
     */
    double getRecordingStopTime() const { return d_recordingStopTime; }

    /**
     * Sets the scheduled recording stop time.
     * @param value The new value.
     */
    void setRecordingStopTime(double value) { d_recordingStopTime = value; }

    /**
     * Resets last recording start time to zero.
     * Called when last recording start time was previously
     * set to inappropriate value in startRecording() function.
     */
    void resetLastRecordingStartTime() { d_lastRecordingStartTime = 0; }

    /**
     * Starts recording. Does nothing if recording is already in progress.
     */
    void startRecording();

    /**
     * Stops recording. If VRML recording is enabled, the recorded animation
     * is automatically saved. Does nothing if recording is not in progress.
     */
    void stopRecording();

    /**
     * Saves the recorded animation of the scene as a VRML file.
     * @param os A stream to send the output to.
     */
    void saveRecord(ostream &os) const;

    /**
     * Returns the file name of the file to store the recorded animation.
     * @return The file name.
     */
    const char* getRecordFileName() const { return d_recordFileName ? d_recordFileName : ""; }

    /**
     * Sets a new file name for storing recorded animations.
     * @param name The file name.
     */
    void setRecordFileName(const char* name);

    /**
     * Saves the recorded animation (to the file mentioned in d_recordFile).
     * @return True on success and false on failure.
     */
    bool saveRecord();

  public:
     /**
      *	Guess znear and zfar values
      */
    void guessNearFarPlane(Viewer* viewer, double* nearVal, double* farVal);
};

#endif // VRMLSCENE_H
